Require Import Machine.
Require Import StepStar.
Require Import ContextSemantics.
Require Import world_model.
Require Import msl.msl_standard.
Require Import SfLib.
Require Import LibTactics.
Require Import WorldLogic.
Require Import StepStarCs.
Require Import HoareLogic.
Require Import IterationBounds.
Require Export BisimulationCs.

Section Definitions.

  Import Machine.Notation.
  Import ContextSemantics.Notation.
  Open Scope context_scope.
  Open Scope wterm_scope.


  Definition bisimulation (R: state->wterm->wterm->Prop) :=
    forall s b1 b2,
    R s b1 b2 ->
    (halted b1 -> step_star s b2 s b1 /\ R s b1 b1)
    /\
    (halted b2 -> step_star s b1 s b2 /\ R s b2 b2)
    /\
    (forall s', s_iter s s' -> R s' b1 b2)
    /\ (* context stepping *)
    (forall b s' b', step s b s' b' -> R s' b1 b2)
    /\
    (forall s' b1',
      step s b1 s' b1' ->
      exists b2',
        step_star s b2 s' b2' /\
        R s' b1' b2')
    /\
    (forall s' b2',
      step s b2 s' b2' ->
      exists b1',
        step_star s b1 s' b1' /\
        R s' b1' b2')
  .

  Definition bisimilar s b1 b2 :=
    exists R, bisimulation R /\ R s b1 b2.

  Definition bisimulation_symm (R: state->wterm->wterm->Prop) :=
    forall s b1 b2,
    R s b1 b2 ->
    (halted b1 -> step_star s b2 s b1 /\ R s b1 b1)
    /\
    (forall s', s_iter s s' -> R s' b1 b2)
    /\
    (forall b s' b', step s b s' b' -> R s' b1 b2)
    /\
    (forall s' b1',
      step s b1 s' b1' ->
      exists b2',
        step_star s b2 s' b2' /\
        R s' b1' b2')
    /\
    R s b2 b1
  .

  Definition bisimilar_symm s ba bb :=
    exists R, bisimulation_symm R /\ R s ba bb.



End Definitions.

Section BisimulationRelations.
  Import Machine.Notation.
  Open Scope wterm_scope.

  Lemma bisim_refl: forall s b,
    bisimilar s b b.
  Proof. 
    intros.
    exists (fun (sa:state) (ba:wterm) bb => ba=bb).
    split; auto.
    clear s b.
    intros s b_ b ?; subst b_.
    splits; intros; auto; try solve [split; constructor | eexists; split; auto; eapply step_cons; eauto; apply step_nil].
  Qed.

  Lemma bisim_symm: forall s ba bb,
    bisimilar s ba bb ->
    bisimilar s bb ba.
  Proof.
    intros.
    destruct H as [Rab[??]].
    exists (fun s bb ba => Rab s ba bb).
    split; eauto.
    clear s ba bb H0.
    intros s ba bb ?.
    apply H in H0; destruct H0 as [?[?[?[?[??]]]]]; eauto.
    splits; auto.
  Qed.

  Lemma step_bisim1: forall s ba bb s' ba',
    step s ba s' ba' ->
    bisimilar s ba bb ->
    exists bb',
      step_star s bb s' bb' /\ bisimilar s' ba' bb'.
  Proof.
    introv H0 H; destruct H as [R[??]].
    apply H in H1; destruct H1 as [_[_[_[_[? _]]]]].
    edestruct H1 as [bb'[??]]; eauto.
    exists bb'; split; auto.
    exists R; split; auto.
  Qed.

  Lemma step_bisim2: forall s ba bb s' bb',
    step s bb s' bb' ->
    bisimilar s ba bb ->
    exists ba',
      step_star s ba s' ba' /\ bisimilar s' ba' bb'.
  Proof.
    intros.
    apply bisim_symm in H0.
    edestruct step_bisim1 as [ba'[??]]; eauto.
    exists ba'; split; auto.
    apply bisim_symm; auto.
  Qed.

  Lemma bisim_context: forall s ba bb b s' b',
    step s b s' b' ->
    bisimilar s ba bb ->
    bisimilar s' ba bb.
  Proof.
    introv H0 H; destruct H as [R[??]].
    apply H in H1; destruct H1 as [_[_[_[? _]]]].
    exists R; eauto.
  Qed.

  Lemma cstep_bisim1: forall C C' s ba bb s' ba',
    cstep s C ba s' C' ba' ->
    bisimilar s ba bb ->
    exists bb',
      cstep_star s C bb s' C' bb' /\ bisimilar s' ba' bb'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    edestruct step_bisim1 as [bb'[??]]; eauto.
    exists bb'; split; auto.
    apply step_star_add_context_cstep_star; auto.
    subst ba'.
    exists bb; split.
    eapply cstep_context_cstep_star. eauto. apply cstep_nil.
    eapply bisim_context; eauto.
    apply step_simulates_cstep; eauto.
  Qed.

  Lemma cstep_bisim2: forall C C' s ba bb s' bb',
    cstep s C bb s' C' bb' ->
    bisimilar s ba bb ->
    exists ba',
      cstep_star s C ba s' C' ba' /\ bisimilar s' ba' bb'.
  Proof.
    intros.
    apply bisim_symm in H0.
    edestruct cstep_bisim1 as [ba'[??]]; eauto.
    exists ba'; split; auto.
    apply bisim_symm; auto.
  Qed.

  Lemma bisim_s_iter: forall s s' ba bb,
    s_iter s s' ->
    bisimilar s ba bb ->
    bisimilar s' ba bb.
  Proof.
    intros.
    destruct H0 as [R[??]].
    apply H0 in H1; destruct H1 as [_[_[? _]]].
    exists R; eauto.
  Qed.

  Lemma step_star_bisim1: forall s ba bb s' ba',
    step_star s ba s' ba' ->
    bisimilar s ba bb ->
    exists bb',
      step_star s bb s' bb' /\ bisimilar s' ba' bb'.
  Proof.
    intros.
    revert bb H0.
    induction H; intros.
    exists bb; split; auto; apply step_nil.
    rename b into ba. rename b' into ba'. rename b'' into ba''.
    edestruct step_bisim1 as [bb'[??]]; eauto.
    edestruct IHstep_star as [bb''[??]]; clear IHstep_star; eauto.
    exists bb''; split; auto.
      eapply step_star_app; eauto.
    rename b into ba. rename b'' into ba''.
    edestruct IHstep_star as [bb''[??]]; clear IHstep_star; eauto.
    eapply bisim_s_iter; eauto.
    exists bb''; split; auto.
      eapply st_iteration; eauto.
  Qed.

  Lemma step_star_bisim2: forall s ba bb s' bb',
    step_star s bb s' bb' ->
    bisimilar s ba bb ->
    exists ba',
      step_star s ba s' ba' /\ bisimilar s' ba' bb'.
  Proof.
    intros.
    apply bisim_symm in H0.
    edestruct step_star_bisim1 as [ba'[??]]; eauto.
    exists ba'; split; auto. apply bisim_symm; auto.
  Qed.

  Lemma cstep_star_bisim1: forall C C' s ba bb s' ba',
    cstep_star s C ba s' C' ba' ->
    bisimilar s ba bb ->
    exists bb',
      cstep_star s C bb s' C' bb' /\ bisimilar s' ba' bb'.
  Proof.
    introv H0 H.
    revert bb H.
    induction H0; intros.
    exists bb; split; auto. apply cstep_nil.
    rename b into ba. rename b' into ba'. rename b'' into ba''.
    edestruct cstep_bisim1 as [bb'[??]]; eauto.
    edestruct IHcstep_star as [bb''[??]]; clear IHcstep_star; eauto.
    exists bb''; split; auto.
      eapply cstep_star_app; eauto.
    rename b into ba. rename b'' into ba''.
    edestruct IHcstep_star as [bb''[??]]; clear IHcstep_star; eauto.
    eapply bisim_s_iter; eauto.
    exists bb''; split; auto.
      eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_star_bisim2: forall C C' s ba bb s' bb',
    cstep_star s C bb s' C' bb' ->
    bisimilar s ba bb ->
    exists ba',
      cstep_star s C ba s' C' ba' /\ bisimilar s' ba' bb'.
  Proof.
    intros.
    apply bisim_symm in H0.
    edestruct cstep_star_bisim1 as [ba'[??]]; eauto.
    exists ba'; split; auto. apply bisim_symm; auto.
  Qed.

  Lemma bisim_halted1: forall s ba bb,
    halted ba ->
    bisimilar s ba bb ->
    step_star s bb s ba.
  Proof.
    intros.
    destruct H0 as [R[??]].
    apply H0 in H1; destruct H1 as [[??] _]; auto.
  Qed.

  Lemma bisim_halted2: forall s ba bb,
    halted bb ->
    bisimilar s ba bb ->
    step_star s ba s bb.
  Proof.
    intros.
    destruct H0 as [R[??]].
    apply H0 in H1; destruct H1 as [_ [[??] _]]; auto.
  Qed.

  Lemma bisim_halted1': forall s ba bb C,
    halted ba ->
    bisimilar s ba bb ->
    cstep_star s C bb s C ba.
  Proof. intros; apply step_star_add_context_cstep_star; apply bisim_halted1; auto. Qed.

  Lemma bisim_halted2': forall s C ba bb,
    halted bb ->
    bisimilar s ba bb ->
    cstep_star s C ba s C bb.
  Proof. intros; apply step_star_add_context_cstep_star; apply bisim_halted2; auto. Qed.

  Lemma cstep_star_bisim1': forall C C' s ba bb s' ba',
    cstep_star s C ba s' C' ba' ->
    bisimilar s ba bb ->
    exists bb',
      cstep_star s C bb s' C' bb' /\ (halted ba' -> bb'=ba') /\ bisimilar s' ba' bb'.
  Proof.
    intros.
    edestruct cstep_star_bisim1 as [bb'[??]]; eauto.
    destruct (halted_dec ba').
    exists ba'; splits; auto.
      eapply cstep_star_app; eauto.
      apply step_star_add_context_cstep_star.
      eapply bisim_halted1; eauto.
      apply bisim_refl.
    exists bb'; splits; auto.
      contradiction.
  Qed.

  Lemma cstep_star_bisim2': forall C C' s ba bb s' bb',
    cstep_star s C bb s' C' bb' ->
    bisimilar s ba bb ->
    exists ba',
      cstep_star s C ba s' C' ba' /\ (halted bb' -> ba'=bb') /\ bisimilar s' ba' bb'.
  Proof.
    introv H0 H.
    apply bisim_symm in H.
    edestruct cstep_star_bisim1' as [ba'[?[??]]]; eauto.
    exists ba'; splits; auto. apply bisim_symm; auto.
  Qed.

  Lemma bisim_halted_converge1': forall s ba bb s' ba',
    step_star s ba s' ba' ->
    bisimilar s ba bb ->
    halted ba' ->
    step_star s bb s' ba'.
  Proof.
    introv H0 H H1.
    edestruct cstep_star_bisim1 with (C:=cx_hole) as [bb'[??]]; eauto.
    apply step_star_add_context_cstep_star; eauto.
    eapply step_star_app.
    change bb with (set_context cx_hole bb).
    apply step_star_simulates_cstep_star; eauto.
    simpl.
    apply bisim_halted1; auto.
  Qed.

  Lemma bisim_halted_converge1: forall C C' s ba bb s' ba',
    cstep_star s C ba s' C' ba' ->
    bisimilar s ba bb ->
    halted ba' ->
    cstep_star s C bb s' C' ba'.
  Proof.
    introv H0 H H1.
    edestruct cstep_star_bisim1 as [bb'[??]]; eauto.
    eapply cstep_star_app; eauto.
    apply step_star_add_context_cstep_star.
    apply bisim_halted1; auto.
  Qed.

  Lemma bisim_halted_converge2': forall s ba bb s' bb',
    step_star s bb s' bb' ->
    bisimilar s ba bb ->
    halted bb' ->
    step_star s ba s' bb'.
  Proof.
    introv H0 H H1.
    edestruct cstep_star_bisim2 with (C:=cx_hole) as [ba'[??]]; eauto.
    apply step_star_add_context_cstep_star; eauto.
    eapply step_star_app.
    change ba with (set_context cx_hole ba).
    apply step_star_simulates_cstep_star; eauto.
    simpl.
    apply bisim_halted2; auto.
  Qed.

  Lemma bisim_halted_converge2: forall C C' s ba bb s' bb',
    cstep_star s C bb s' C' bb' ->
    bisimilar s ba bb ->
    halted bb' ->
    cstep_star s C ba s' C' bb'.
  Proof.
    introv H0 H H1.
    edestruct cstep_star_bisim2 as [ba'[??]]; eauto.
    eapply cstep_star_app; eauto.
    apply step_star_add_context_cstep_star.
    apply bisim_halted2; auto.
  Qed.

  Lemma bisim_trans: forall s ba bb bc,
    bisimilar s ba bb ->
    bisimilar s bb bc ->
    bisimilar s ba bc.
  Proof.
    intros.
    exists (fun s ba bc => exists bb, bisimilar s ba bb /\ bisimilar s bb bc).
    split; eauto.
    clear s ba bb bc H H0.
    intros s ba bc [bb[??]].
    splits.

    intro; split.
    eapply bisim_halted_converge1'; eauto.
    eapply bisim_halted1; eauto.
    eexists; split; apply bisim_refl.

    intro; split.
    eapply bisim_halted_converge2'; eauto.
    eapply bisim_halted2; eauto.
    eexists; split; apply bisim_refl.

    intros; exists bb; split; eapply bisim_s_iter; eauto.

    intros.
    exists bb; split; eapply bisim_context; eauto.

    intros s' ba' ?.
    edestruct step_bisim1 as [bb'[??]]; eauto.    
    edestruct step_star_bisim1 as [bc'[??]]; eauto.

    intros s' bc' ?.
    edestruct step_bisim2 as [bb'[??]]; eauto.    
    edestruct step_star_bisim2 as [ba'[??]]; eauto.
  Qed.


  Lemma bisimulation_symm_equiv: forall R,
    bisimulation_symm R -> bisimulation R.
  Proof.
    intros R ? s ba bb ?.
    splits.

    intros; eapply H; eauto.
    intros; eapply H; eauto. eapply H; eauto.

    intros. apply H in H0; destruct H0 as [_[? _]]. eauto.

    intros; apply H in H0; destruct H0 as [_[_[? _]]]; eauto.

    intros s' ba' ?.
    eapply H; eauto.

    intros s' bb' ?.
    apply H in H0; destruct H0 as [_[_[_[_ ?]]]].
    apply H in H0; destruct H0 as [_[_[_[? _]]]].
    edestruct H0 as [ba'[??]]; clear H0; eauto.
    exists ba'; splits; auto.
    apply H; auto.
  Qed.

  Lemma bisim_exists_symm_R: forall s ba bb,
    bisimilar s ba bb -> bisimilar_symm s ba bb.
  Proof.
    intros.
    exists bisimilar.
    split; auto.
    clear s ba bb H.
    intros s ba bb ?.
    splits.
    intro; split. apply bisim_halted1; auto. apply bisim_refl.
    intros. eapply bisim_s_iter; eauto.
    intros b s' b' ?. eapply bisim_context; eauto.
    intros s' ba' ?.
    edestruct step_bisim1 as [bb'[??]]; eauto.
    apply bisim_symm; auto.
  Qed.

  Lemma bisim_symm_equiv: forall s ba bb,
    bisimilar_symm s ba bb <-> bisimilar s ba bb.
  Proof.
    split; intros.
    destruct H as [R[??]].
    apply bisimulation_symm_equiv in H. exists R; auto.
    apply bisim_exists_symm_R; auto.
  Qed.

  Lemma bisim_Cs_bisim: forall b1 b2,
    bisimilar_Cs b1 b2 -> forall s, bisimilar s b1 b2.
  Proof.
    intros ba bb ? s.
    apply bisim_symm_equiv.
    exists (fun (s:state) b1 b2 => bisimilar_Cs b1 b2).
    split; eauto.
    clear ba bb H s.
    intros s ba bb ?.
    splits.

    split. apply bisim_Cs_halted1; auto. apply bisim_Cs_refl.
    intros; auto.

    intros; auto.

    intros s' ba' ?.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    apply bisim_Cs_symm; auto.    
  Qed.


End BisimulationRelations.

(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)

Module Notation.
  Notation "C '|-b' ( s1 , b1 ) '~' ( s2 , b2 )" := (bisimilar C s1 b1%wterm s2 b2%wterm)(at level 40) : bsim_scope.

  Module U.
  End U.
End Notation.

(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)

Module Morphisms.
  Require Import Setoid.
  Require Export Morphisms.
  Export SetoidTactics.
  Require Export SetoidClass.
  Import Machine.Notation.
  Import Notation.
  Open Scope wterm_scope.
  Export BisimulationCs.Morphisms.

  Add Parametric Morphism (s:state): (bisimilar s)
  with signature equiv ==> equiv ==> equiv
  as bisim_Cs_bisim_m.
    intros; split; intros.
    eapply bisim_trans.
    apply bisim_Cs_bisim.
    apply bisim_Cs_symm; apply H.
    eapply bisim_trans. eauto.
    apply bisim_Cs_bisim. auto.
    eapply bisim_trans.
    apply bisim_Cs_bisim. eauto.
    eapply bisim_trans. eauto.
    apply bisim_symm.
    apply bisim_Cs_bisim. auto.
  Qed.


End Morphisms.


(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)

Section Congruence.
  Import Machine.Notation.
  Import Notation.
  Import Notation.U.
  Import ContextSemantics.Notation.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import Morphisms.
  Open Scope forces_pred.
  Open Scope context_scope.
  Open Scope bsim_scope.
  Open Scope wterm_scope.

  Lemma bisim_par_swap: forall s b1 b2,
    bisimilar s (b1||b2) (b2||b1).
  Proof.
    intros.
    apply bisim_Cs_bisim.
    apply bisim_Cs_par_swap.
  Qed.

  Lemma bisim_seq: forall s ba bb ta tb,
    bisimilar s ba bb ->
    (forall C s' C' w', cstep_star s C ba s' C' (w':tm_skip) ->bisimilar s' (w':ta) (w':tb)) ->
    bisimilar s (ba;; ta) (bb;; tb).
  Proof.
    intros.
    exists (fun s bat bbt =>
      (exists ba ta bb tb, bat=ba;;ta /\ bbt=bb;;tb /\ (forall C s' C' w', cstep_star s C ba s' C' (w':tm_skip) ->bisimilar s' (w':ta) (w':tb)) /\ bisimilar s ba bb)
      \/ (bisimilar s bat bbt))%wterm.
    split.
    apply bisimulation_symm_equiv.
    2: left; jauto.
    clear s ba ta bb tb H H0.
    intros s bat bbt [[ba[ta[bb[tb[?[?[??]]]]]]] | ?].
    subst bat bbt.
    splits.

    intros [?]; discriminate.

    intros. left; exists ba ta bb tb; splits; auto.
    intros. eapply H1; eauto. eapply cstep_iteration; eauto.
    eapply bisim_s_iter; eauto.

    intros.
    left; exists ba ta bb tb; splits; intros; auto.
    eapply H1 with (C:=(cx_par1 C b)).
    eapply cstep_context_cstep_star.
    apply cst_par1_step2; eauto.
    apply cstep_star_par_cx_par1; eauto.
    apply cstep_star_par1_step; eauto.
    eapply bisim_context; eauto.

    intros sa' bat' ?.
    inverts H.
    edestruct step_bisim1 as [bb'[??]]; eauto.
    exists (bb';; tb); splits; auto.
      apply step_star_seq_app; auto.
      left; exists b1' ta bb' tb; splits; auto.
      intros.
      eapply H1 with (C:= C).
      eapply step_cstep_star; eauto.
    exists (w: tb); splits; auto.
      eapply step_star_seq_join.
      eapply bisim_halted1; eauto. eexists; eauto.
      apply step_nil.
      right.
      eapply H1 with (C:=cx_hole) (s':=sa').
      apply cstep_nil.

    left; exists bb tb ba ta; splits; auto.
    intros; apply bisim_symm.
    eapply H1 with (C:=C).
    eapply bisim_halted_converge2; eauto. eexists; eauto.
    apply bisim_symm; auto.

    splits; intros.
    split. apply bisim_halted1; auto. right; apply bisim_refl.
    right; eapply bisim_s_iter; eauto.
    right; eapply bisim_context; eauto.
    edestruct step_bisim1 as [bb'[??]]; eauto.
    right; apply bisim_symm; auto.
  Qed.

  Lemma bisim_seq1: forall s ba bb t,
    bisimilar s ba bb ->
    bisimilar s (ba;; t) (bb;; t).
  Proof.
    intros; eapply bisim_seq; auto; intros; apply bisim_refl.
  Qed.

  Lemma bisim_seq2: forall s b ta tb,
    (forall C s' C' w', cstep_star s C b s' C' (w':tm_skip) ->bisimilar s' (w':ta) (w':tb)) ->
    bisimilar s (b;; ta) (b;; tb).
  Proof.
    intros; eapply bisim_seq; auto; intros; apply bisim_refl.
  Qed.

  Lemma cstep_par_compose_contexts1: forall s C b1 b2 s' C' b1' b2',
    cstep s C (b1 || b2) s' C' (b1' || b2') ->
    cstep s (compose_contexts C (cx_par1 cx_hole b2)) b1 s' (compose_contexts C' (cx_par1 cx_hole b2')) b1'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto; clear H.
    subst C'.
    inverts H0.
    apply cstep_simulates_step; apply step_add_context; auto.
    apply cstep_composed_context_step2 with gwemp.
    apply cst_par1_step2; auto.
    inverts H1.
    apply cstep_composed_context_step1 with gwemp; auto.
  Qed.

  Lemma bisim_par1: forall s ba bb b2,
    bisimilar s ba bb ->
    bisimilar s (ba || b2) (bb || b2).
  Proof.
    intros.
    exists (fun (s:state) ba2 bb2 => (exists ba bb b2, ba2=ba||b2 /\ bb2=bb||b2 /\ bisimilar s ba bb) \/ ba2=bb2)%wterm.
    split.
    2: left; jauto.
    eapply bisimulation_symm_equiv.
    clear s ba bb b2 H.
    intros s ba2 bb2 [[ba[bb[b2[?[??]]]]] | ?].
    subst ba2 bb2.
    splits.

    intros [?]; discriminate.
    intros; left; exists ba bb b2; splits; auto.
    eapply bisim_s_iter; eauto.
    intros; left; exists ba bb b2; splits; auto.
    eapply bisim_context; eauto.

    intros s' ba2' ?.
    inverts H.
    edestruct step_bisim1 as [bb'[??]]; eauto.
    exists (bb'||b2); split; eauto.
      apply step_star_par1_step; auto.
      left; exists b1' bb' b2; splits; auto.
    exists (bb||b2'); split; eauto.
      apply step_star_par2_step; auto. eapply step_cons; eauto; apply step_nil.
      left; exists ba bb b2'; splits; auto.
      eapply bisim_context; eauto.
    exists (w:tm_skip); split; eauto.
      eapply step_star_app.
      apply step_star_par1_step.
      eapply bisim_halted1; eauto. eexists; eauto.
      eapply step_cons; [apply st_par_join; eauto | apply step_nil].

    left; exists bb ba b2; splits; auto.
    apply bisim_symm; auto.

    subst.
    splits; intros; auto.
    split; intros; auto. apply step_nil.
    exists b1'; split; auto. eapply step_cons; eauto; apply step_nil.
  Qed.

  Lemma bisim_par2: forall s ba bb b1,
    bisimilar s ba bb ->
    bisimilar s (b1 || ba) (b1 || bb).
  Proof.
    intros.
    apply bisim_trans with (bb||b1).
    apply bisim_trans with (ba||b1).
    apply bisim_par_swap.
    apply bisim_par1; auto.
    apply bisim_par_swap.
  Qed.

  Lemma bisim_par: forall s ba bc bb bd,
    bisimilar s ba bb ->
    bisimilar s bc bd ->
    bisimilar s (ba || bc) (bb || bd).
  Proof.
    intros. eapply bisim_trans.
    apply bisim_par1; eauto.
    apply bisim_par2; eauto.
  Qed.

  Lemma bisim_tm_seq_seq: forall s w t1 t2,
    bisimilar s (w : tm_seq t1 t2) (w : t1;; t2).
  Proof. intros; apply bisim_Cs_bisim; apply bisim_Cs_tm_seq_seq. Qed.

  Lemma bisim_seq_assoc: forall s b1 t2 t3,
    bisimilar s (b1;; (t2;;; t3)) ((b1;; t2);; t3).
  Proof. intros; apply bisim_Cs_bisim; apply bisim_Cs_seq_assoc. Qed.

  Lemma bisim_tm_par_par: forall s w w1 w2 t1 t2,
    gjoin w1 w2 w ->
    bisimilar s (w: tm_par (exactly w1) t1 (exactly w2) t2) (w1:t1 || w2:t2).
  Proof. intros; apply bisim_Cs_bisim; apply bisim_Cs_tm_par_par; auto. Qed.

  Lemma bisim_par_skip: forall s w1 w2 w,
    gjoin w1 w2 w ->
    bisimilar s (w1:tm_skip || w2:tm_skip) (w: tm_skip).
  Proof. intros; apply bisim_Cs_bisim; apply bisim_Cs_par_skip; auto. Qed.

  Lemma bisim_if: forall s w e ta1 ta2 tb1 tb2,
    (evalTrue e w -> bisimilar s (w: ta1) (w: tb1)) ->
    (evalFalse e w -> bisimilar s (w: ta2) (w: tb2)) ->
    bisimilar s (w: tm_if e ta1 ta2) (w: tm_if e tb1 tb2).
  Proof.
    intros.
    destruct (evalTrueFalse_dec e w) as [[?|?]|[??]].
    setoid_rewrite bisim_Cs_if_true; auto.
    setoid_rewrite bisim_Cs_if_false; auto.
    apply bisim_Cs_bisim.
    apply bisim_Cs_if_noeval; auto.
  Qed.


  Lemma bisim_while: forall s w e z t1 t2 P,
    w |= P ->
    (forall s' C' w', w'|= P -> evalTrue e w' -> halted_assert s' C' (w':t1) P) ->
    (forall s' C' w', w'|= P -> evalTrue e w' -> halted_assert s' C' (w':t2) P) ->
    (forall C s' C' w' tt bb, cstep_star s C (w: tm_while_max e z t1) s' C' (w': tt;; bb) -> w'|= P -> evalTrue e w' -> bisimilar s' (w': t1) (w': t2)) ->
    bisimilar s (w: tm_while_max e z t1) (w: tm_while_max e z t2).
  Proof.
    intros.
    rename s into s0.
    rename w into w0.
    rename z into z0.
    exists (fun s ba bb =>
      (
        (exists z w, ba=w:tm_while_max e z t1 /\ bb=w:tm_while_max e z t2 /\ w|=P) \/
        (exists z bba bbb, ba=bba;; tm_while_max e z t1 /\ bb=bbb;; tm_while_max e z t2 /\ bisimilar s bba bbb /\ (forall C, halted_assert s C bba P)) \/
        ba=bb
      ) /\
      exists C0 C, cstep_star s0 C0 (w0: tm_while_max e z0 t1) s C ba).
    split.
    2: split; eauto; [ left; eauto | exists cx_hole cx_hole; apply cstep_nil].
    rename H into HPw0.
    intros s ba bb [[[z[w[?[??]]]]| [ [z[bba[bbb[?[?[??]]]]]] | ?] ] ?]; subst.
    splits.

    intros [?]; discriminate.
    intros [?]; discriminate.

    intros; split.
    left; exists z w; splits; auto.
    edestruct H5 as [C0[C ?]].
    exists C0 C.
    eapply cstep_star_iter_snoc; eauto.

    intros; split.
    left; exists z w; splits; auto.
    edestruct H5 as [C0[C0' ?]].
    exists (cx_par1 C0 b) (cx_par1 C0' b').
      eapply cstep_star_snoc.
      eapply cstep_star_par_cx_par1; eauto.
      apply cstep_star_par1_step; eauto.
      eapply cst_par1_step2; eauto.

    intros s' ba' ?.
    destruct H5 as [C0[C0' H7]].
    inverts H.
    exists (w: t2;; tm_while_max e z' t2); splits; auto.
      apply step_single; apply st_while_true; auto.
      right; left; exists z' (w: t1) (w:t2); splits; auto.
      eapply H2 with (C:=C0); eauto.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_true; eauto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_true; auto.
    exists (w: tm_skip); splits; auto.
      apply step_single; apply st_while_false; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_false; auto.
    exists (w: tm_while_max e z1 t2); splits; auto.
      apply step_single; apply st_while_go; auto.
      left; exists z1 w; splits; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_go; auto.
    exists (w: tm_skip); splits; auto.
      apply step_single; apply st_while_limit; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_limit; auto.

    intros s' bb' ?.
    destruct H5 as [C0[C0' H7]].
    inverts H.
    exists (w: t1;; tm_while_max e z' t1); splits; auto.
      apply step_single; apply st_while_true; auto.
      right; left; exists z' (w: t1) (w:t2); splits; auto.
      eapply H2 with (C:=C0); eauto.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_true; eauto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_true; auto.
    exists (w: tm_skip); splits; auto.
      apply step_single; apply st_while_false; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_false; eauto.
    exists (w: tm_while_max e z1 t1); splits; auto.
      apply step_single; apply st_while_go; auto.
      left; exists z1 w; splits; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_go; eauto.
    exists (w: tm_skip); splits; auto.
      apply step_single; apply st_while_limit; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_limit; eauto.

    splits.
    intros [?]; discriminate.
    intros [?]; discriminate.

    intros; split.
    right; left; exists z bba bbb; splits; auto.
    eapply bisim_s_iter; eauto.
    intro C; intro; intros; eapply H5.
    eapply cstep_iteration; eauto.
    edestruct H6 as [C0[C ?]].
    exists C0 C.
    eapply cstep_star_iter_snoc; eauto.

    intros; split.
    right; left; exists z bba bbb; splits; auto.
    eapply bisim_context; eauto.
    intro C; intro; intros; eapply H5 with (C:=cx_par1 C b).
    eapply cstep_cons.
    apply cst_par1_step2; eauto.
    apply cstep_star_par_cx_par1; apply cstep_star_par1_step; eauto.
    edestruct H6 as [C0[C0' ?]].
    exists (cx_par1 C0 b) (cx_par1 C0' b').
      eapply cstep_star_snoc.
      eapply cstep_star_par_cx_par1; eauto.
      apply cstep_star_par1_step; eauto.
      eapply cst_par1_step2; eauto.

    intros s' ba' ?.
    destruct H6 as [C0[C0' H7]].
    inverts H.
    edestruct step_bisim1 with (ba:=bba) as [bbb'[??]]; eauto.
    exists (bbb';; tm_while_max e z t2); splits; auto.
      apply step_star_seq_app; auto.
      right; left; exists z b1' bbb'; splits; auto.
      intros C1; intro; intros; eapply H5.
      eapply step_cstep_star; eauto.
      exists C0 C0'.
      eapply cstep_star_snoc; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_seq1; auto.
    exists (w: tm_while_max e z t2); splits; auto.
      eapply step_star_seq_join; eauto.
      eapply bisim_halted1; eauto. eexists; eauto.
      apply step_nil.
      left; exists z w; splits; auto.
      eapply H5; eauto. apply cstep_nil with (C:=cx_hole).
      exists C0 C0'.
      eapply cstep_star_app; eauto. apply step_single_cstep_star; apply st_seq_skip; auto.

    intros s' bb' ?.
    destruct H6 as [C0[C0' H7]].
    inverts H.
    edestruct step_bisim2 with (bb:=bbb) as [bba'[??]]; eauto.
    exists (bba';; tm_while_max e z t1); splits; auto.
      apply step_star_seq_app; auto.
      right; left; exists z bba' b1'; splits; auto.
      intros C1; intro; intros; eapply H5.
      eapply cstep_star_app; eauto.
      apply step_star_add_context_cstep_star; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_star_add_context_cstep_star; apply step_star_seq_app; auto.
    exists (w: tm_while_max e z t1); splits; auto.
      eapply step_star_seq_join; eauto.
      eapply bisim_halted2; eauto. eexists; eauto.
      apply step_nil.
      left; exists z w; splits; auto.
      eapply H5; eauto.
      apply step_star_add_context_cstep_star with (C:=cx_hole).
      eapply bisim_halted2; eauto. eexists; eauto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      eapply cstep_star_seq_join; eauto.
      apply step_star_add_context_cstep_star.
      eapply bisim_halted2; eauto. eexists; eauto.
      apply cstep_nil.

    destruct H3 as [C0[C0' ?]].
    splits; auto.
    intros.
    split; auto.
    apply step_nil.
    intros.
    split; eauto.
    intro; split.
    apply step_nil.
    intros; splits; eauto.
    intros; splits; eauto.
    exists C0 C0'; eapply cstep_star_iter_snoc; eauto.
    intros; splits; eauto.
    exists (cx_par1 C0 b) (cx_par1 C0' b').
      eapply cstep_star_snoc.
      eapply cstep_star_par_cx_par1; eauto.
      apply cstep_star_par1_step; eauto.
      eapply cst_par1_step2; eauto.
    intros; exists b1'; splits; eauto.
    apply step_single; auto.
    exists C0 C0'.
    eapply cstep_star_snoc; eauto. apply cstep_simulates_step; apply step_add_context; auto.
    intros; exists b2'; splits; eauto.
    apply step_single; auto.
    exists C0 C0'.
    eapply cstep_star_snoc; eauto. apply cstep_simulates_step; apply step_add_context; auto.
  Qed.


  Lemma bisim_step': forall s ba bb,
    (forall bO s' bO', step_star s bO s' bO' -> halted ba -> step_star s' bb s' ba) ->
    (forall bO s' bO', step_star s bO s' bO' -> halted bb -> step_star s' ba s' bb) ->
    (forall bO s'0 bO' ba' s', step_star s bO s'0 bO' -> step s'0 ba s' ba' -> ba'<>ba -> exists bb', step_star s'0 bb s' bb' /\ bisimilar s' ba' bb') ->
    (forall bO s'0 bO' bb' s', step_star s bO s'0 bO' -> step s'0 bb s' bb' -> bb'<>bb -> exists ba', step_star s'0 ba s' ba' /\ bisimilar s' ba' bb') ->
    bisimilar s ba bb.
  Proof.
    intros.
    rename s into s0.
    exists (fun s ba bb => (exists b0 b, step_star s0 b0 s b) /\
      (((forall bO s' bO', step_star s bO s' bO' -> halted ba -> step_star s' bb s' ba) /\
      (forall bO s' bO', step_star s bO s' bO' -> halted bb -> step_star s' ba s' bb) /\
      (forall bO s'0 bO' ba' s', step_star s bO s'0 bO' -> step s'0 ba s' ba' -> ba'<>ba -> exists bb', step_star s'0 bb s' bb' /\ bisimilar s' ba' bb') /\
      (forall bO s'0 bO' bb' s', step_star s bO s'0 bO' -> step s'0 bb s' bb' -> bb'<>bb -> exists ba', step_star s'0 ba s' ba' /\ bisimilar s' ba' bb')
      ) \/ bisimilar s ba bb)).
    split.
    eapply bisimulation_symm_equiv.
    clear ba bb H H0 H1 H2.
    intros s ba bb [[b0[b?]][[?[?[??]]]|?]]; subst.
    * splits; eauto.

    - intro; splits.
    eapply H0; eauto. apply step_nil with (b:= gwemp:tm_skip).
    eauto.
    right; apply bisim_refl.

    - intros; split.
    exists b0 b.
      eapply step_star_iter_snoc; eauto.
    left; splits; intros; auto.
    eapply H0; eauto. eapply st_iteration; eauto.
    eapply H1; eauto. eapply st_iteration; eauto.
    edestruct H2 as [bb'[??]]; eauto.
    eapply st_iteration; eauto.
    edestruct H3 as [ba'[??]]; eauto.
    eapply st_iteration; eauto.

    - intros; split.
    exists (b0||b1) (b||b').
      eapply step_star_snoc.
      apply step_star_par1_step; eauto.
      apply st_par2; auto.
    left; splits; intros; auto.
    eapply H0; eauto. eapply step_cons.
    apply st_par1; eauto. apply step_star_par2_step; eauto.
    eapply H1; eauto. eapply step_cons.
    apply st_par1; eauto. apply step_star_par2_step; eauto.
    edestruct H2 as [bb'[??]]; eauto.
    eapply step_cons.
    apply st_par2; eauto.
    apply step_star_par1_step; eauto.
    edestruct H3 as [ba'[??]]; eauto.
    eapply step_cons.
    apply st_par2; eauto.
    apply step_star_par1_step; eauto.

    - intros s' ba' ?.
    destruct (wterm_eq_dec ba' ba).
    subst ba'.
    assert (s'=s); subst; eauto. eapply step_spin_no_change; eauto.
    exists bb; splits; auto. apply step_nil.
      exists b0 b; auto.
    edestruct H2 as [bb'[??]]; clear H2; eauto.
      apply step_nil with (b:=ba).
    exists bb'; splits; auto.
      exists (b0||bb) (b||bb'); auto.
      eapply step_star_app.
      apply step_star_par1_step; eauto.
      apply step_star_par2_step; auto.

    - left; splits; intros; auto.
    eapply H1; eauto.
    eapply H0; eauto.
    edestruct H3 as [bb'[??]]; eauto.
    exists bb'; split; auto. apply bisim_symm; auto.
    edestruct H2 as [ba'[??]]; eauto.
    exists ba'; split; auto. apply bisim_symm; auto.

    * splits; intros; auto.
    splits; intros; eauto.
    apply bisim_halted1; auto.
    right; apply bisim_refl.
    splits; intros; eauto.
    exists b0 b. eapply step_star_iter_snoc; eauto.
    right; eapply bisim_s_iter; eauto.
    splits; intros; eauto.
    exists (b0||b1) (b||b'); auto.
      eapply step_star_snoc.
      apply step_star_par1_step; eauto.
      apply st_par2; auto.
    right; eapply bisim_context; eauto.
    edestruct step_bisim1 as [bb'[??]]; eauto.
    exists bb'; splits; auto.
    exists (b0||ba) (b||b1'); auto.
      eapply step_star_snoc.
      apply step_star_par1_step; eauto.
      apply st_par2; auto.
    eauto.
    right; apply bisim_symm; auto.

    * splits; intros; eauto.
    exists (gwemp:tm_skip) (gwemp:tm_skip); apply step_nil.
  Qed.

  Lemma bisim_step: forall s ba bb,
    (forall s : state, halted ba -> step_star s bb s ba) ->
    (forall s : state, halted bb -> step_star s ba s bb) ->
    (forall b s'0 b'0 s' ba', step_star s b s'0 b'0 -> step s'0 ba s' ba' -> ba'<>ba -> exists bb', step_star s'0 bb s' bb' /\ bisimilar s' ba' bb') ->
    (forall b s'0 b'0 s' bb', step_star s b s'0 b'0 -> step s'0 bb s' bb' -> bb'<>bb -> exists ba', step_star s'0 ba s' ba' /\ bisimilar s' ba' bb') ->
    bisimilar s ba bb.
  Proof.
    intros.
    rename s into s0.
    exists (fun s ba bb => (exists b0 b, step_star s0 b0 s b) /\
      (((forall s : state, halted ba -> step_star s bb s ba) /\ (forall s : state, halted bb -> step_star s ba s bb) /\
      (forall b s'0 b'0 s' ba', step_star s b s'0 b'0 -> step s'0 ba s' ba' -> ba'<>ba -> exists bb', step_star s'0 bb s' bb' /\ bisimilar s' ba' bb') /\
      (forall b s'0 b'0 s' bb', step_star s b s'0 b'0 -> step s'0 bb s' bb' -> bb'<>bb -> exists ba', step_star s'0 ba s' ba' /\ bisimilar s' bb' ba')
      ) \/ bisimilar s ba bb)).
    split.
    eapply bisimulation_symm_equiv.
    clear ba bb H H0 H1 H2.
    intros s ba bb [[b0[b?]][[?[?[??]]]|?]]; subst.
    splits; eauto.

    intro; splits.
    apply H0; auto.
    eauto.
    right; apply bisim_refl.

    intros; split.
    exists b0 b.
      eapply step_star_iter_snoc; eauto.
    left; splits; intros; auto.
    edestruct H2 as [bb'[??]]; eauto.
    eapply st_iteration; eauto.
    edestruct H3 as [ba'[??]]; eauto.
    eapply st_iteration; eauto.

    intros; split.
    exists (b0||b1) (b||b').
      eapply step_star_snoc.
      apply step_star_par1_step; eauto.
      apply st_par2; auto.
    left; splits; intros; auto.
    edestruct H2 as [bb'[??]]; eauto.
    eapply step_cons.
    apply st_par2; eauto.
    apply step_star_par1_step; eauto.
    edestruct H3 as [ba'[??]]; eauto.
    eapply step_cons.
    apply st_par2; eauto.
    apply step_star_par1_step; eauto.

    intros s' ba' ?.
    destruct (wterm_eq_dec ba' ba).
    subst ba'.
    assert (s'=s); subst; eauto. eapply step_spin_no_change; eauto.
    exists bb; splits; auto. apply step_nil.
      exists b0 b; auto.
    edestruct H2 as [bb'[??]]; clear H2; eauto.
      apply step_nil with (b:=ba).
    exists bb'; splits; auto.
      exists (b0||bb) (b||bb'); auto.
      eapply step_star_app.
      apply step_star_par1_step; eauto.
      apply step_star_par2_step; auto.

    splits; intros; auto.
    splits; intros; eauto.
    apply bisim_halted1; auto.
    right; apply bisim_refl.
    splits; intros; eauto.
    exists b0 b. eapply step_star_iter_snoc; eauto.
    right; eapply bisim_s_iter; eauto.
    splits; intros; eauto.
    exists (b0||b1) (b||b'); auto.
      eapply step_star_snoc.
      apply step_star_par1_step; eauto.
      apply st_par2; auto.
    right; eapply bisim_context; eauto.
    edestruct step_bisim1 as [bb'[??]]; eauto.
    exists bb'; splits; auto.
    exists (b0||ba) (b||b1'); auto.
      eapply step_star_snoc.
      apply step_star_par1_step; eauto.
      apply st_par2; auto.
    eauto.
    right; apply bisim_symm; auto.
    splits; intros; eauto.
    exists (gwemp:tm_skip) (gwemp:tm_skip); apply step_nil.
    left; splits; auto.
    intros.
    edestruct H2 as [ba'[??]]; eauto.
    exists ba'; split; auto.
    apply bisim_symm; auto.
  Qed.

  Lemma bisim_tm_par: forall s w P1 ta1 tb1 P2 ta2 tb2,
    (forall b s' b' w1, step_star s b s' b' -> w1|=gsep_con P1 TT -> bisimilar s' (w1: ta1) (w1:tb1)) ->
    (forall b s' b' w2, step_star s b s' b' -> w2|=gsep_con P2 TT -> bisimilar s' (w2: ta2) (w2:tb2)) ->
    bisimilar s (w: tm_par P1 ta1 P2 ta2) (w: tm_par P1 tb1 P2 tb2).
  Proof.
    intros.
    apply bisim_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H2.
    exists (w1: tb1 || w2: tb2); split.
      apply step_single.
      apply st_par; auto.
      apply bisim_par.
      eapply H; eauto.
      eapply H0; eauto.
    intros.
    inverts H2.
    exists (w1: ta1 || w2: ta2); split.
      apply step_single.
      apply st_par; auto.
      apply bisim_par.
      eapply H; eauto.
      eapply H0; eauto.
  Qed.

  Lemma bisim_s_iter_star: forall s s' ba bb,
    s_iter_star s s' ->
    bisimilar s ba bb ->
    bisimilar s' ba bb.
  Proof.
    intros.
    induction H; auto.
    apply IHs_iter_star.
    eapply bisim_s_iter; eauto.
  Qed.

  Lemma bisim_step_star_context: forall s b s' b' ba bb,
    step_star s b s' b' ->
    bisimilar s ba bb ->
    bisimilar s' ba bb.
  Proof.
    intros.
    induction H; auto.
    apply IHstep_star; eapply bisim_context; eauto.
    apply IHstep_star; eapply bisim_s_iter; eauto.
  Qed.

  Lemma bisim_assert_intro: forall w P s t1 t2,
    (w|= gsep_con P TT -> bisimilar s (w: t1) (w: t2)) ->
    bisimilar s (w: tm_assert P;; t1) (w: tm_assert P;; t2).
  Proof.
    intros.
    apply bisim_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H1. inverts H7.
    exists (w: t2)%wterm; split.
      eapply step_cons.
      apply st_seq1; apply st_assert; eauto.
      apply step_single.
      apply st_seq_skip.
      setoid_rewrite bisim_Cs_seq_skip1'.
      eapply bisim_step_star_context; eauto.
    intros.
    inverts H1. inverts H7.
    exists (w: t1)%wterm; split.
      eapply step_cons.
      apply st_seq1; apply st_assert; eauto.
      apply step_single.
      apply st_seq_skip.
      setoid_rewrite bisim_Cs_seq_skip1'.
      eapply bisim_step_star_context; eauto.
  Qed.

  Lemma bisim_seq1_rewrite: forall s b1 b1' b2 b2' t3 t4,
    bisimilar s b1 b1' ->
    bisimilar s b2 b2' ->
    bisimilar s (b1';; t3) (b2';; t4) ->
    bisimilar s (b1;; t3) (b2;; t4).
  Proof.
    intros.
    eapply bisim_trans.
    apply bisim_seq1; eauto.
    eapply bisim_trans; eauto.
    apply bisim_symm.
    apply bisim_seq1; auto.
  Qed.

  Lemma bisim_seq1_rewrite': forall s b1 b2 t3 t4,
    bisimilar s b1 b2 ->
    bisimilar s (b2;; t3) (b2;; t4) ->
    bisimilar s (b1;; t3) (b1;; t4).
  Proof.
    intros.
    eapply bisim_seq1_rewrite; eauto.
  Qed.


  Lemma halted_assert_bisim: forall s C ba bb P,
    halted_assert s C ba P ->
    bisimilar s ba bb ->
    halted_assert s C bb P.
  Proof.
    intros.
    intro; intros.
    eapply H.
    eapply bisim_halted_converge1; eauto.
    apply bisim_symm; auto.
    eexists; eauto.
  Qed.

  Lemma bisim_tm_local: forall s w e1 e2 f1 f2,
    e1 == e2 ->
    (forall x w' v, eval e1 v w -> gjoin w (new_local x v) w' -> bisimilar s (w': f1 x) (w': f2 x)) ->
    bisimilar s (w: tm_local e1 f1) (w: tm_local e2 f2).
  Proof.
    intros.
    apply bisim_step.
    - introv [?]; discriminate.
    - introv [?]; discriminate.
    - intros.
    inverts H2.
    eexists; split.
    step_cons 1. eapply st_local; eauto.
    setoid_rewrite<- H; eauto.
    eapply bisim_step_star_context; eauto.
    - intros.
    inverts H2.
    eexists; split.
    step_cons 1. eapply st_local; eauto.
    setoid_rewrite H; eauto.
    eapply bisim_step_star_context; eauto.
    eapply H0; eauto.
    setoid_rewrite H; eauto.
  Qed.

  Lemma bisim_tm_local': forall s w e1 e2 f1 f2,
    e1 == e2 ->
    (forall x w' v, eval e1 v w -> gjoin w (new_local x v) w' -> bisimilar s (w': f1 x) (w': f2 x)) ->
    bisimilar s (w: local x:=e1 in f1 x) (w: local x:=e2 in f2 x).
  Proof.
    intros.
    apply bisim_tm_local; auto.
    intros.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_seq1; eauto.
  Qed.


End Congruence.


















