Require Import Machine.
Require Import StepStar.
Require Import ContextSemantics.
Require Import world_model.
Require Import msl.msl_standard.
Require Import SfLib.
Require Import LibTactics.
Require Import WorldLogic.
Require Import StepStarCs.
Require Import HoareLogic.
Require Import IterationBounds.

Section Definitions.
  Import Machine.Notation.
  Import ContextSemantics.Notation.

  Open Scope wterm_scope.


  Definition bisimulation_Cs (R: wterm->wterm->Prop) :=
    forall b1 b2,
    R b1 b2 ->
    (halted b1 -> (forall s, step_star s b2 s b1) /\ R b1 b1)
    /\
    (halted b2 -> (forall s, step_star s b1 s b2) /\ R b2 b2)
    /\
    (forall s s' b1',
      step s b1 s' b1' ->
      exists b2',
        step_star s b2 s' b2' /\
        R b1' b2')
    /\
    (forall s s' b2',
      step s b2 s' b2' ->
      exists b1',
        step_star s b1 s' b1' /\
        R b1' b2')
  .

  Definition bisimilar_Cs b1 b2 :=
    exists R, bisimulation_Cs R /\ R b1 b2.

  Definition bisimulation_Cs_symm (R: wterm->wterm->Prop) :=
    forall b1 b2,
    R b1 b2 ->
    (halted b1 -> (forall s, step_star s b2 s b1) /\ R b1 b1)
    /\
    (forall s s' b1',
      step s b1 s' b1' ->
      exists b2',
        step_star s b2 s' b2' /\
        R b1' b2')
    /\
    R b2 b1
  .

  Definition bisimilar_Cs_symm ba bb :=
    exists R, bisimulation_Cs_symm R /\ R ba bb.

  Definition term_bisimilar_Cs t1 t2 := forall w, bisimilar_Cs (w:t1) (w:t2).
  Hint Unfold term_bisimilar_Cs.

  Definition deterministic_Cs_wterm b := forall b',
    step_star_Cs b b' -> bisimilar_Cs b b'.

End Definitions.


Section BasicRelations.
  Import Machine.Notation.
  Open Scope wterm_scope.


  Lemma bisim_Cs_refl: forall b,
    bisimilar_Cs b b.
  Proof.
    intros b.
    exists (fun (ba:wterm) bb => bb=ba).
    split; eauto.
    clear b.
    intros b b_ ?; subst.
    splits; intros; auto.
    split; auto. intros; apply step_nil.
    split; auto. intros; apply step_nil.
    exists b1'; split; auto. apply step_single; auto.
    exists b2'; split; auto. apply step_single; auto.
  Qed.

  Lemma bisim_Cs_symm: forall b1 b2,
    bisimilar_Cs b1 b2 -> bisimilar_Cs b2 b1.
  Proof.
    intros ba bb [R[??]]. 
    exists (fun ba bb => R bb ba).
    split; eauto.
    clear ba bb H0.
    intros ba bb ?.
    splits; intros; auto.
    apply H in H0; destruct H0 as [_[? _]]; eauto.
    apply H in H0; destruct H0 as [? _]; eauto.
    apply H in H0; destruct H0 as [_[_[_ ?]]]; eauto.
    apply H in H0; destruct H0 as [_[_[?  _]]]; eauto.
  Qed.


  Lemma step_bisim_Cs1: forall s ba bb s' ba',
    step s ba s' ba' ->
    bisimilar_Cs ba bb ->
    exists bb',
      step_star s bb s' bb' /\ bisimilar_Cs ba' bb'.
  Proof.
    intros; destruct H0 as [R[??]].
    apply H0 in H1; destruct H1 as [_[_[? _]]].
    edestruct H1 as [bb'[??]]; eauto.
    exists bb'; split; auto.
      exists R; auto.
  Qed.

  Lemma step_bisim_Cs2: forall s ba bb s' bb',
    step s bb s' bb' ->
    bisimilar_Cs ba bb ->
    exists ba',
      step_star s ba s' ba' /\ bisimilar_Cs ba' bb'.
  Proof.
    intros; edestruct step_bisim_Cs1 as [ba'[??]]; eauto.
    apply bisim_Cs_symm; eauto.
    eexists; split; eauto.
    apply bisim_Cs_symm; eauto.
  Qed.

  Lemma skip_step_bisim_Cs_respects_state: forall s w b s' b',
    bisimilar_Cs b (w: tm_skip) ->
    step s b s' b' ->
    s' = s.
  Proof.
    intros.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    apply step_star_skip_resolve in H1.
    destruct H1; subst.
    eapply s_iter_star_step_eq; eauto.
  Qed.

  Lemma step_star_bisim_Cs1: forall s ba bb s' ba',
    step_star s ba s' ba' ->
    bisimilar_Cs ba bb ->
    exists bb',
      step_star s bb s' bb' /\ bisimilar_Cs ba' bb'.
  Proof.
    introv H0 H.
    revert bb H.
    induction H0; intros.
    exists bb; split; auto. apply step_nil.
    rename b into ba. rename b' into ba'. rename b'' into ba''.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    edestruct IHstep_star as [bb''[??]]; clear IHstep_star; eauto.
    exists bb''; split; auto.
      eapply step_star_app; eauto.
    rename b into ba. rename b'' into ba''.
    edestruct IHstep_star as [bb''[??]]; clear IHstep_star; eauto.
    exists bb''; split; auto.
      eapply st_iteration; eauto.
  Qed.

  Lemma step_star_bisim_Cs2: forall s ba bb s' bb',
    step_star s bb s' bb' ->
    bisimilar_Cs ba bb ->
    exists ba',
      step_star s ba s' ba' /\ bisimilar_Cs ba' bb'.
  Proof.
    intros; edestruct step_star_bisim_Cs1 as [ba'[??]]; eauto.
    apply bisim_Cs_symm; eauto.
    eexists; split; eauto.
    apply bisim_Cs_symm; eauto.
  Qed.

  Lemma bisim_Cs_halted1: forall s ba bb,
    halted ba ->
    bisimilar_Cs ba bb ->
    step_star s bb s ba.
  Proof.
    intros.
    destruct H0 as [R[??]].
    apply H0 in H1; destruct H1 as [[??] _]; auto.
  Qed.

  Lemma bisim_Cs_halted2: forall s ba bb,
    halted bb ->
    bisimilar_Cs ba bb ->
    step_star s ba s bb.
  Proof.
    intros.
    destruct H0 as [R[??]].
    apply H0 in H1; destruct H1 as [_ [[??] _]]; auto.
  Qed.

  Lemma bisim_Cs_halted_converge1: forall s ba bb s' ba',
    step_star s ba s' ba' ->
    bisimilar_Cs ba bb ->
    halted ba' ->
    step_star s bb s' ba'.
  Proof.
    introv H0 H H1.
    edestruct step_star_bisim_Cs1 as [bb'[??]]; eauto.
    eapply step_star_app; eauto.
    apply bisim_Cs_halted1; eauto.
  Qed.

  Lemma bisim_Cs_halted_converge2: forall s ba bb s' bb',
    step_star s bb s' bb' ->
    bisimilar_Cs ba bb ->
    halted bb' ->
    step_star s ba s' bb'.
  Proof.
    intros; eapply bisim_Cs_halted_converge1; eauto; apply bisim_Cs_symm; auto.
  Qed.

  Lemma bisim_Cs_trans: forall ba bb bc,
    bisimilar_Cs ba bb ->
    bisimilar_Cs bb bc ->
    bisimilar_Cs ba bc.
  Proof.
    intros ba bb bc. 
    exists (fun ba bc => exists bb, bisimilar_Cs ba bb /\ bisimilar_Cs bb bc).
    split; eauto.
    clear ba bb bc H H0.
    intros ba bc [bb[??]].
    splits.

    intro; split; intros.
    eapply bisim_Cs_halted_converge1; eauto.
    apply bisim_Cs_halted1; auto.
    exists ba; split; apply bisim_Cs_refl.

    intro; split; intros.
    eapply bisim_Cs_halted_converge2; eauto.
    apply bisim_Cs_halted2; auto.
    exists bc; split; apply bisim_Cs_refl.

    intros s s' ba' ?.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    edestruct step_star_bisim_Cs1 as [bc'[??]]; eauto.

    intros s s' bc' ?.
    edestruct step_bisim_Cs2 as [bb'[??]]; eauto.
    edestruct step_star_bisim_Cs2 as [ba'[??]]; eauto.
  Qed.


  Lemma cstep_bisim_Cs1: forall C C' s ba bb s' ba',
    cstep s C ba s' C' ba' ->
    bisimilar_Cs ba bb ->
    exists bb',
      cstep_star s C bb s' C' bb' /\ bisimilar_Cs ba' bb'.
  Proof.
    introv H0 H.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    exists bb'; split; auto.
    apply step_star_add_context_cstep_star; auto.
    subst ba'.
    exists bb; split; auto.
    eapply cstep_context_cstep_star; eauto. apply cstep_nil.
  Qed.

  Lemma cstep_bisim_Cs2: forall C C' s ba bb s' bb',
    cstep s C bb s' C' bb' ->
    bisimilar_Cs ba bb ->
    exists ba',
      cstep_star s C ba s' C' ba' /\ bisimilar_Cs ba' bb'.
  Proof.
    intros; edestruct cstep_bisim_Cs1 as [ba'[??]]; eauto.
    apply bisim_Cs_symm; eauto.
    eexists; split; eauto.
    apply bisim_Cs_symm; eauto.
  Qed.

  Lemma cstep_star_bisim_Cs1: forall C C' s ba bb s' ba',
    cstep_star s C ba s' C' ba' ->
    bisimilar_Cs ba bb ->
    exists bb',
      cstep_star s C bb s' C' bb' /\ bisimilar_Cs ba' bb'.
  Proof.
    introv H0 H.
    revert bb H.
    induction H0; intros.
    exists bb; split; auto. apply cstep_nil.
    rename b into ba. rename b' into ba'. rename b'' into ba''.
    edestruct cstep_bisim_Cs1 as [bb'[??]]; eauto.
    edestruct IHcstep_star as [bb''[??]]; clear IHcstep_star; eauto.
    exists bb''; split; auto.
      eapply cstep_star_app; eauto.
    rename b into ba. rename b'' into ba''.
    edestruct IHcstep_star as [bb''[??]]; clear IHcstep_star; eauto.
    exists bb''; split; auto.
      eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_star_bisim_Cs2: forall C C' s ba bb s' bb',
    cstep_star s C bb s' C' bb' ->
    bisimilar_Cs ba bb ->
    exists ba',
      cstep_star s C ba s' C' ba' /\ bisimilar_Cs ba' bb'.
  Proof.
    intros; edestruct cstep_star_bisim_Cs1 as [ba'[??]]; eauto.
    apply bisim_Cs_symm; eauto.
    eexists; split; eauto.
    apply bisim_Cs_symm; eauto.
  Qed.

  Lemma bisim_Cs_halted_converge1': forall C C' s ba bb s' ba',
    cstep_star s C ba s' C' ba' ->
    bisimilar_Cs ba bb ->
    halted ba' ->
    cstep_star s C bb s' C' ba'.
  Proof.
    introv H0 H H1.
    edestruct cstep_star_bisim_Cs1 as [bb'[??]]; eauto.
    eapply cstep_star_app; eauto.
    apply step_star_add_context_cstep_star.
    apply bisim_Cs_halted1; auto.
  Qed.

  Lemma bisim_Cs_halted_converge2': forall C C' s ba bb s' bb',
    cstep_star s C bb s' C' bb' ->
    bisimilar_Cs ba bb ->
    halted bb' ->
    cstep_star s C ba s' C' bb'.
  Proof.
    intros; eapply bisim_Cs_halted_converge1'; eauto; apply bisim_Cs_symm; auto.
  Qed.


  Lemma cstep_star_bisim_Cs1': forall C C' s ba bb s' ba',
    cstep_star s C ba s' C' ba' ->
    bisimilar_Cs ba bb ->
    exists bb',
      cstep_star s C bb s' C' bb' /\ (halted ba' -> bb'=ba') /\ bisimilar_Cs ba' bb'.
  Proof.
    intros.
    edestruct cstep_star_bisim_Cs1 as [bb'[??]]; eauto.
    destruct (halted_dec ba').
    exists ba'; splits; auto.
      eapply cstep_star_app; eauto.
      apply step_star_add_context_cstep_star.
      eapply bisim_Cs_halted1; eauto.
      apply bisim_Cs_refl.
    exists bb'; splits; auto.
      contradiction.
  Qed.

  Lemma cstep_star_bisim_Cs2': forall C C' s ba bb s' bb',
    cstep_star s C bb s' C' bb' ->
    bisimilar_Cs ba bb ->
    exists ba',
      cstep_star s C ba s' C' ba' /\ (halted bb' -> ba'=bb') /\ bisimilar_Cs ba' bb'.
  Proof.
    introv H0 H.
    apply bisim_Cs_symm in H.
    edestruct cstep_star_bisim_Cs1' as [ba'[?[??]]]; eauto.
    exists ba'; splits; auto. apply bisim_Cs_symm; auto.
  Qed.


  Lemma step_star_Cs_bisim_Cs1: forall ba ba' bb,
    step_star_Cs ba ba' ->
    bisimilar_Cs ba bb ->
    exists bb',
      step_star_Cs bb bb' /\ bisimilar_Cs ba' bb'.
  Proof.
    intros.
    revert bb H0.
    induction H; intros.
    exists bb; split; auto. apply step_star_Cs_nil.
    edestruct cstep_star_bisim_Cs1 as [bb'[??]]; eauto.
    edestruct IHstep_star_Cs as [bb''[??]]; eauto.
    exists bb''; split; auto.
    eapply step_star_Cs_cons; eauto.
  Qed.

  Lemma step_star_Cs_bisim_Cs2: forall ba bb bb',
    step_star_Cs bb bb' ->
    bisimilar_Cs ba bb ->
    exists ba',
      step_star_Cs ba ba' /\ bisimilar_Cs ba' bb'.
  Proof.
    intros.
    revert ba H0.
    induction H; rename ba into bb; intros.
    exists ba; split; auto. apply step_star_Cs_nil.
    rename ba' into bb'; rename ba'' into bb''.
    edestruct cstep_star_bisim_Cs2 as [ba'[??]]; eauto.
    edestruct IHstep_star_Cs as [ba''[??]]; eauto.
    exists ba''; split; auto.
    eapply step_star_Cs_cons; eauto.
  Qed.

  Lemma step_star_Cs_bisim_Cs1': forall ba bb ba',
    step_star_Cs ba ba' ->
    bisimilar_Cs ba bb ->
    exists bb',
      step_star_Cs bb bb' /\
      (halted ba' -> bb'=ba') /\
      bisimilar_Cs ba' bb'.
  Proof.
    intros.
    revert bb H0.
    induction H; intros.
    destruct (halted_dec ba).
    exists ba; splits; auto.
      eapply step_star_Cs_cons; [| apply step_star_Cs_nil].
      eapply step_star_add_context_cstep_star with (C:=cx_hole) (s:=State.empty_state).
      eapply bisim_Cs_halted1; eauto.
      apply bisim_Cs_refl.
    exists bb; splits; auto.
      apply step_star_Cs_nil.
      contradiction.
    edestruct cstep_star_bisim_Cs1 as [bb'[??]]; eauto.
    edestruct IHstep_star_Cs with (bb:=bb') as [bb''[?[??]]]; clear IHstep_star_Cs; eauto.
    exists bb''; splits; eauto.
    eapply step_star_Cs_cons; eauto.
  Qed.



  Lemma bisimulation_Cs_symm_equiv: forall R,
    bisimulation_Cs_symm R -> bisimulation_Cs R.
  Proof.
    intros R ? ba bb ?.
    splits.

    eapply H; eauto.

    eapply H; eauto.
    eapply H; eauto.

    intros ss' ba' ?.
    eapply H; eauto.

    intros s s' bb' ?.
    apply H in H0; destruct H0 as [_[_ ?]].
    apply H in H0; destruct H0 as [_[? _]].
    edestruct H0 as [ba'[??]]; clear H0; eauto.
    exists ba'; splits; auto.
    apply H; auto.
  Qed.

  Lemma bisim_Cs_exists_symm_R: forall ba bb,
    bisimilar_Cs ba bb -> bisimilar_Cs_symm ba bb.
  Proof.
    intros.
    exists bisimilar_Cs.
    split; auto.
    clear ba bb H.
    intros ba bb ?.
    splits.
    split; intros.
    eapply bisim_Cs_halted1; eauto.
    apply bisim_Cs_refl.

    intros s s' ba' ?.
    eapply step_bisim_Cs1; eauto.

    apply bisim_Cs_symm; auto.
  Qed.

  Lemma bisim_Cs_symm_equiv: forall ba bb,
    bisimilar_Cs_symm ba bb <-> bisimilar_Cs ba bb.
  Proof.
    split; intros.
    destruct H as [R[??]].
    apply bisimulation_Cs_symm_equiv in H. exists R; auto.
    apply bisim_Cs_exists_symm_R; auto.
  Qed.

  Lemma bisim_Cs_halted_converge1'': forall ba bb ba',
    step_star_Cs ba ba' ->
    bisimilar_Cs ba bb ->
    halted ba' ->
    step_star_Cs bb ba'.
  Proof.
    intros.
    edestruct step_star_Cs_bisim_Cs1' as [bb'[?[??]]]; eauto.
    apply H3 in H1.
    subst; auto.
  Qed.

  Lemma bisim_Cs_halted_converge2'': forall ba bb bb',
    step_star_Cs bb bb' ->
    bisimilar_Cs ba bb ->
    halted bb' ->
    step_star_Cs ba bb'.
  Proof.
    intros.
    apply bisim_Cs_symm in H0.
    eapply bisim_Cs_halted_converge1''; eauto.
  Qed.


  Lemma term_bisim_Cs_refl: forall t, term_bisimilar_Cs t t.
  Proof. repeat intro; apply bisim_Cs_refl. Qed.
  Lemma term_bisim_Cs_symm: forall t1 t2, term_bisimilar_Cs t1 t2 -> term_bisimilar_Cs t2 t1.
  Proof. repeat intro; apply bisim_Cs_symm; auto. Qed.
  Lemma term_bisim_Cs_trans: forall t1 t2 t3, term_bisimilar_Cs t1 t2 -> term_bisimilar_Cs t2 t3 -> term_bisimilar_Cs t1 t3.
  Proof. repeat intro; eapply bisim_Cs_trans; eauto. Qed.

  (******************
    Lemmas to help prove bisimulations
  ******************)

  (*
    This lemma is used to prove bisimulation of ba and bb by letting them each take a single step,
    then proving that the resulting states are bisimilar. This approach is useful if
    ba or bb normalize into simpler (or equal) forms.

    ba is bisimilar to bb if, for any way that ba or bb may step, then there is a matching step such that the resulting states are bisimilar 
  *)
  Lemma bisim_Cs_step: forall ba bb,
    (forall s, halted ba -> step_star s bb s ba) ->   
    (forall s, halted bb -> step_star s ba s bb) ->   
    (forall s s' ba', step s ba s' ba' -> ba'<>ba -> exists bb', step_star s bb s' bb' /\ bisimilar_Cs ba' bb') ->
    (forall s s' bb', step s bb s' bb' -> bb'<>bb -> exists ba', step_star s ba s' ba' /\ bisimilar_Cs ba' bb') ->
    bisimilar_Cs ba bb.
  Proof.
    intros.
    exists (fun ba bb => ((forall s, halted ba -> step_star s bb s ba) /\ (forall s, halted bb -> step_star s ba s bb) /\ (forall s s' ba', step s ba s' ba' -> ba'<>ba -> exists bb', step_star s bb s' bb' /\ bisimilar_Cs ba' bb') /\ (forall s s' bb', step s bb s' bb' -> bb'<>bb -> exists ba', step_star s ba s' ba' /\ bisimilar_Cs bb' ba')) \/ bisimilar_Cs ba bb).
    split; eauto.
    eapply bisimulation_Cs_symm_equiv.
    clear ba bb H H0 H1 H2.
    intros ba bb [[?[?[??]]]|?]; subst.
    splits; eauto.

    split; intros; auto.
    right; apply bisim_Cs_refl.

    intros s s' ba' ?.
    destruct (wterm_eq_dec ba' ba).
    subst ba'.
    assert (s'=s); subst; eauto. eapply step_spin_no_change; eauto.
    exists bb; splits; auto. apply step_nil.
    edestruct H1 as [bb'[??]]; clear H1; eauto.

    splits; intros; auto.
    split; intros.
    apply bisim_Cs_halted1; auto.
    right; apply bisim_Cs_refl.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    right; apply bisim_Cs_symm; auto.
    left; splits; auto.
    intros.
    edestruct H2 as [ba'[??]]; eauto.
    exists ba'; split; auto.
    apply bisim_Cs_symm; auto.
  Qed.

  (* b' is bisimilar to b if b nondeterministically steps to b' *)
  Lemma bisim_Cs_tao_step: forall b b',
    (forall s, step s b s b') ->
    (forall s s'' b'', step s b s'' b'' -> b''<>b -> (s''=s/\b''=b')) ->
   bisimilar_Cs b b'.
  Proof.
    intros.
    apply bisim_Cs_step.

    introv [?] ?; subst.
    spec H s; inverts H.

    intros.
    apply step_single; auto.

    intros.
    edestruct H0; eauto; subst s' ba'.
    eexists; split. apply step_nil. apply bisim_Cs_refl.

    intros.
    eexists; split.
    step_cons 2; eauto.
    apply bisim_Cs_refl.
  Qed.

  (* b' is bisimilar to b if b nondeterministically steps b' (there may be multiple ways to step, but they are all bisimilar) *)
  Lemma bisim_Cs_tao_step': forall b b',
    (forall s, step s b s b') ->
    (forall s s1' s2' b1' b2', step s b s1' b1' -> step s b s2' b2' -> b1'<>b -> b2'<>b -> s1'=s/\s2'=s/\bisimilar_Cs b1' b2') ->
    bisimilar_Cs b b'.
  Proof.
    intros.
    apply bisim_Cs_step.

    introv [?] ?; subst.
    spec H s; inverts H.

    intros.
    apply step_single; auto.

    intros.
    edestruct H0 with (b1':=ba') (b2':=ba') as [?[??]]; eauto.
    subst s'. clear H4.
    destruct (wterm_eq_dec b' b).
    subst b'.
    eexists; split.
      apply step_single; eauto. auto.
    edestruct H0 with (b1':=b') (b2':=ba') as [?[??]]; eauto.
    eexists; split.
      apply step_nil; eauto. apply bisim_Cs_symm; auto.

    intros.
    eexists; split.
      step_cons 2; eauto.
      apply bisim_Cs_refl.
  Qed.

End BasicRelations.


(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)

Add Parametric Relation : wterm bisimilar_Cs
    reflexivity proved by bisim_Cs_refl
    symmetry proved by bisim_Cs_symm
    transitivity proved by bisim_Cs_trans
    as bisimilar_Cs_rel.

Add Parametric Relation : term term_bisimilar_Cs
    reflexivity proved by term_bisim_Cs_refl
    symmetry proved by term_bisim_Cs_symm
    transitivity proved by term_bisim_Cs_trans
    as term_bisimilar_Cs_rel.

(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)


Section Congruence.
  Import Machine.Notation.
  Import Notation.
  Import Notation.U.
  Import ContextSemantics.Notation.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import Morphisms.
  Import world_model.Morphisms.
  Open Scope forces_pred.
(*  Open Scope context_scope. *)
  Open Scope wterm_scope.
(*  Import WorldLogic.Notation WorldLogic.Notation.U WorldLogic.Tactics. *)
(*  Open Scope join_scope. *)

  Lemma bisim_Cs_seq: forall ba bb ta tb,
    bisimilar_Cs ba bb ->
    (forall w', step_star_Cs ba (w':tm_skip) -> bisimilar_Cs (w':ta) (w':tb)) ->
    bisimilar_Cs (ba;; ta) (bb;; tb).
  Proof.
    intros.
    apply bisim_Cs_exists_symm_R in H.
    destruct H as [R[??]].
    exists (fun bat bbt =>
      (exists ba ta bb tb, bat=ba;;ta /\ bbt=bb;;tb /\ (forall w', step_star_Cs ba (w':tm_skip) -> bisimilar_Cs (w':ta) (w':tb)) /\ bisimilar_Cs ba bb)
      \/ (bisimilar_Cs bat bbt))%wterm.
    split.
    apply bisimulation_Cs_symm_equiv.
    2: left; do 4 eexists; splits; eauto; exists R; split; auto; apply bisimulation_Cs_symm_equiv; auto.
    clear R H ba ta bb tb H1 H0.
    intros bat bbt [[ba[ta[bb[tb[?[?[??]]]]]]] | ?].
    subst bat bbt. splits.

    intros [?]; discriminate.

    intros s s' bat' ?.
    edestruct step_seq_inversion as [[ba'[??]] | [w[?[??]]]]; eauto.
    subst bat'.
    clear H.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    exists (bb';; tb); splits; auto.
      apply step_star_seq_app; eauto.
      left; exists ba' ta bb' tb; splits; intros; auto.
      apply H1.
      eapply step_star_Cs_cons with (C:=cx_hole); eauto.
      apply step_star_add_context_cstep_star.
      eapply step_single; eauto.
    subst s' ba bat'.
    exists (w: tb); splits; auto.
      eapply step_star_seq_join.
      eapply bisim_Cs_halted1; eauto.
      eexists; eauto.
      apply step_nil.
      right; apply H1. apply step_star_Cs_nil.

    left; exists bb tb ba ta; splits; auto.
      intros; apply bisim_Cs_symm; apply H1.
      edestruct step_star_Cs_bisim_Cs1' with (ba:=bb) as [bb'[?[??]]]; eauto.
      symmetry; eauto.
      spec H3; subst. eexists; eauto.
      apply H0.
      symmetry; auto.

    splits.
    split; intros.
    apply bisim_Cs_halted1; auto.
    right; apply bisim_Cs_refl.
    intros s s' bat' ?.
      edestruct step_bisim_Cs1 as [bbt'[??]]; eauto.
    right; apply bisim_Cs_symm; auto.
  Qed.

  Lemma bisim_Cs_seq1: forall ba bb t,
    bisimilar_Cs ba bb ->
    bisimilar_Cs (ba;; t) (bb;; t).
  Proof.
    intros; eapply bisim_Cs_seq; auto; intros; reflexivity.
  Qed.

  Lemma bisim_Cs_seq2: forall b ta tb,
    (forall w', step_star_Cs b (w':tm_skip) -> bisimilar_Cs (w':ta) (w':tb)) ->
    bisimilar_Cs (b;; ta) (b;; tb).
  Proof.
    intros; eapply bisim_Cs_seq; auto; intros; reflexivity.
  Qed.

  Lemma bisim_Cs_par_swap: forall b1 b2,
    bisimilar_Cs (b1||b2) (b2||b1).
  Proof.
    intros.
    exists (fun ba2 bb2 => (exists ba bb, ba2=ba||bb /\ bb2=bb||ba) \/ ba2=bb2)%wterm.
    split; eauto.
    eapply bisimulation_Cs_symm_equiv.
    clear b1 b2.
    intros ba bb [[b1[b2[??]]] | ?].
    subst ba bb. splits.

    intros [?]; discriminate.

    intros s s' b' ?.
    inverts H.
    exists (b2||b1'); split; eauto.
      apply step_star_par2_step; auto.
      apply step_single; auto.
    exists (b2'||b1); split; eauto.
      apply step_star_par1_step; auto.
      apply step_single; auto.
    exists (w:tm_skip); split; eauto.
      apply step_single; apply st_par_join; auto.
      apply join_com; auto.
    eauto.

    subst.
    splits; intros; auto.
    split; intros; auto. apply step_nil.
    exists b1'; split; auto.
    apply step_single; auto.
  Qed.

  Lemma bisim_Cs_par1: forall ba bb b2,
    bisimilar_Cs ba bb ->
    bisimilar_Cs (ba || b2) (bb || b2).
  Proof.
    intros.
    apply bisim_Cs_exists_symm_R in H.
    exists (fun ba2 bb2 => (exists ba bb b2, ba2=ba||b2 /\ bb2=bb||b2 /\ bisimilar_Cs ba bb) \/ ba2=bb2)%wterm.
    split.
    2: left; jauto.
    eapply bisimulation_Cs_symm_equiv.
    clear ba bb b2 H.
    intros ba2 bb2 [[ba[bb[b2[?[??]]]]] | ?].
    subst ba2 bb2.
    splits.

    intros [?]; discriminate.

    intros s s' ba2' ?.
    inverts H.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    exists (bb'||b2); split; eauto.
      apply step_star_par1_step; auto.
      left; exists b1' bb' b2; splits; auto.
    exists (bb||b2'); split; eauto.
      apply step_star_par2_step; auto.
      apply step_single; auto.
      left; exists ba bb b2'; splits; auto.
    exists (w:tm_skip); split; eauto.
      eapply step_star_app.
      apply step_star_par1_step.
      eapply bisim_Cs_halted1; eauto. eexists; eauto.
      apply step_single; apply st_par_join; auto.

    left; exists bb ba b2; splits; auto.
    apply bisim_Cs_symm; auto.

    subst.
    splits; intros; auto.
    split; intros; auto. apply step_nil.
    exists b1'; split; auto. apply step_single; auto.

    exists ba bb b2; splits; auto. apply bisim_Cs_symm_equiv; auto.
  Qed.

  Lemma bisim_Cs_par2: forall ba bb b1,
    bisimilar_Cs ba bb ->
    bisimilar_Cs (b1 || ba) (b1 || bb).
  Proof.
    intros.
    apply bisim_Cs_trans with (bb||b1).
    apply bisim_Cs_trans with (ba||b1).
    apply bisim_Cs_par_swap.
    apply bisim_Cs_par1; auto.
    apply bisim_Cs_par_swap.
  Qed.


  Lemma bisim_Cs_par: forall ba1 ba2 bb1 bb2,
    bisimilar_Cs ba1 bb1 ->
    bisimilar_Cs ba2 bb2 ->
    bisimilar_Cs (ba1 || ba2) (bb1 || bb2).
  Proof.
    intros.
    eapply bisim_Cs_trans.
    apply bisim_Cs_par1; eauto.
    apply bisim_Cs_par2; auto.
  Qed.


  Lemma bisim_Cs_tm_seq_seq: forall w t1 t2,
    bisimilar_Cs (w : tm_seq t1 t2) (w : t1;; t2).
  Proof.
    intros.
    apply bisim_Cs_tao_step.
    intros. apply st_seq.
    intros.
    inverts H; auto.
  Qed.

  Lemma bisim_Cs_while: forall (e1 e2:expr) z ta tb,
    expr_bool_equiv e1 e2 ->
    (forall w, evalTrue e1 w -> bisimilar_Cs (w: ta) (w: tb)) ->
    term_bisimilar_Cs (tm_while_max e1 z ta) (tm_while_max e2 z tb).
  Proof.
    introv He_equiv ?.
    intro w.
    exists (fun ba bb => exists e1 ta e2 tb,
      expr_bool_equiv e1 e2 /\
      (forall w, evalTrue e1 w -> bisimilar_Cs (w: ta) (w: tb)) /\
      ((exists bat bbt z,
        ba=bat;; tm_while_max e1 z ta /\
        bb=bbt;; tm_while_max e2 z tb /\
        bisimilar_Cs bat bbt) \/
      (exists w z,
        ba=w: tm_while_max e1 z ta /\
        bb=w: tm_while_max e2 z tb) \/ ba=bb))%wterm.
    split.
    apply bisimulation_Cs_symm_equiv.
    clear e1 e2 ta tb He_equiv H w z.
    intros ba bb [e1[ta[e2[tb[He_equiv[? [ [bba[bbt[z[?[??]]]]] | [ [w[z[??]]] | ?]]]]]]]]; subst.

    splits.
    intros [?]; discriminate.
    intros s s' ba' ?.
    inverts H0.
    edestruct step_bisim_Cs1 as [bbt'[??]]; eauto.
    exists (bbt';; tm_while_max e2 z tb)%wterm; split.
      apply step_star_seq_app; auto.
      exists e1 ta e2 tb; splits; auto; left; jauto.
    exists (w:tm_while_max e2 z tb)%wterm; split.
      eapply step_star_seq_join.
      eapply bisim_Cs_halted1; eauto. eexists; eauto.
      apply step_nil.
      exists e1 ta e2 tb; splits; auto; right; jauto.
    exists e2 tb e1 ta; splits.
    split; intros; split; intros; apply He_equiv; auto.
    intros; symmetry; apply H; apply He_equiv; auto.
    left; exists bbt bba z;splits; auto.
    symmetry; auto.

    splits.
    intros [?]; discriminate.
    intros s s' ba' ?.
    inverts H0.
    exists (w: tb;; tm_while_max e2 z' tb)%wterm; split.
      apply step_single; apply st_while_true; auto.
      apply He_equiv; auto.
      exists e1 ta e2 tb; splits; auto.
      left; exists (w:ta)%wterm (w:tb)%wterm z'; splits; auto.
    exists (w:tm_skip)%wterm; split.
      apply step_single; apply st_while_false; auto.
      apply He_equiv; auto.
      exists e1 ta e2 tb; splits; auto.
    exists (w: tm_while_max e2 z0 tb)%wterm; split; iauto.
      apply step_single; apply st_while_go; auto.
      exists e1 ta e2 tb; split; iauto.
    exists (w:tm_skip)%wterm; split; iauto.
      apply step_single; apply st_while_limit; auto.
      exists e1 ta e2 tb; split; auto.
    exists e2 tb e1 ta; splits.
      split; intros; split; intros; apply He_equiv; auto.
      intros; symmetry; auto.
      apply H. apply He_equiv; auto.
      right; left; exists w z; splits; auto.

    splits.
    split; intros.
    apply step_nil.
    exists e1 ta e2 tb; eauto.
    intros.
    exists b1'; split. apply step_single; auto.
    exists e1 ta e2 tb; split; eauto.
    exists e1 ta e2 tb; split; eauto.

    exists e1 ta e2 tb; splits; intros; auto.
    right; left; exists w z; split; auto.
  Qed.

  Lemma bisim_Cs_if_true: forall w e t1 t2,
    evalTrue e w ->
    bisimilar_Cs (w: tm_if e t1 t2) (w: t1).
  Proof.
    intros.
    eapply bisim_Cs_tao_step.
    intros.
    apply st_if_true; auto.
    intros.
    inverts H0; auto.
    false; eapply evalTrue_evalFalse; eauto.
  Qed.

  Lemma bisim_Cs_if_false: forall w e t1 t2,
    evalFalse e w ->
    bisimilar_Cs (w: tm_if e t1 t2) (w: t2).
  Proof.
    intros.
    eapply bisim_Cs_tao_step.
    intros.
    apply st_if_false; auto.
    intros.
    inverts H0; auto.
    false; eapply evalTrue_evalFalse; eauto.
  Qed.

  Lemma bisim_Cs_stuck_assert_FF: forall b w,
    ~halted b ->
    (forall s s' b', step s b s' b' -> False) ->
    bisimilar_Cs b (w: tm_assert FF).
  Proof.
    intros.
    apply bisim_Cs_step.
    intros; contradiction.
    introv [?]; discriminate.
    intros.
    false; eauto.
    intros.
    false. inverts H1.
    destruct H7 as [?[?[?[??]]]].
    eapply H3.
  Qed.

  Lemma bisim_Cs_stuck_assert_FF': forall t,
    (forall w, ~halted (w:t)) ->
    (forall s w s' b', step s (w:t) s' b' -> False) ->
    term_bisimilar_Cs t (tm_assert FF).
  Proof.
    intros. intro w.
    apply bisim_Cs_stuck_assert_FF; eauto.
  Qed.

  Lemma bisim_Cs_assert_FF_seq: forall w t,
    bisimilar_Cs (w: tm_assert FF) (w: tm_assert FF;; t).
  Proof.
    intros.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros. false. inverts H.
    destruct H5 as [?[?[?[??]]]].
    eapply H1.
    intros. false. inverts H. inverts H5.
    destruct H4 as [?[?[?[??]]]].
    eapply H1.
  Qed.

  Lemma bisim_Cs_assert_FF_seq': forall t,
    term_bisimilar_Cs (tm_assert FF) (tm_assert FF;;; t).
  Proof.
    intros. intro w.
    eapply bisim_Cs_trans.
    eapply bisim_Cs_assert_FF_seq with (t:=t); auto.
    apply bisim_Cs_symm.
    apply bisim_Cs_tm_seq_seq.
  Qed.

  Lemma bisim_Cs_if_noeval: forall w e t1 t2 t3 t4,
    ~evalTrue e w ->
    ~evalFalse e w ->
    bisimilar_Cs (w: tm_if e t1 t2) (w: tm_if e t3 t4).
  Proof.
    intros.
    eapply bisim_Cs_trans.
    apply bisim_Cs_stuck_assert_FF with (w:=w).
    introv [?]; discriminate.
    intros. inverts H1; contradiction.
    eapply bisim_Cs_trans.
    apply bisim_Cs_assert_FF_seq with (t:=tm_if e t3 t3).
    apply bisim_Cs_symm.
    eapply bisim_Cs_trans.
    apply bisim_Cs_stuck_assert_FF with (w:=w).
    introv [?]; discriminate.
    intros. inverts H1; contradiction.
    apply bisim_Cs_assert_FF_seq.
  Qed.

  Lemma bisim_Cs_if: forall w e1 e2 ta1 ta2 tb1 tb2,
    expr_bool_equiv e1 e2 ->
    (evalTrue e1 w -> bisimilar_Cs (w: ta1) (w: tb1)) ->
    (evalFalse e1 w -> bisimilar_Cs (w: ta2) (w: tb2)) ->
    bisimilar_Cs (w: tm_if e1 ta1 ta2) (w: tm_if e2 tb1 tb2).
  Proof.
    intros.
    destruct (evalTrueFalse_dec e1 w) as [[?|?]|[??]].
    setoid_rewrite bisim_Cs_if_true; auto.
    apply H; auto.
    setoid_rewrite bisim_Cs_if_false; auto.
    apply H; auto.
    setoid_rewrite bisim_Cs_if_noeval with (t1:=ta1) (t2:=ta2) (t3:=tb1) (t4:=tb2); eauto.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H4; try contradiction.
    intros.
    inverts H4.
    apply H in H12; contradiction.
    apply H in H12; contradiction.
  Qed.

  Lemma bisim_Cs_tm_seq: forall w ta1 tb1 ta2 tb2,
    bisimilar_Cs (w: ta1) (w: tb1) ->
    (forall w', step_star_Cs (w:ta1) (w':tm_skip) -> bisimilar_Cs (w':ta2) (w':tb2)) ->
    bisimilar_Cs (w: tm_seq ta1 ta2) (w: tm_seq tb1 tb2).
  Proof.
    intros.
    eapply bisim_Cs_trans.
    apply bisim_Cs_tm_seq_seq.
    eapply bisim_Cs_trans.
    apply bisim_Cs_seq; eauto.
    apply bisim_Cs_symm.
    apply bisim_Cs_tm_seq_seq.
  Qed.

  Lemma bisim_Cs_tm_seq1: forall w ta1 tb1 t2,
    bisimilar_Cs (w: ta1) (w: tb1) ->
    bisimilar_Cs (w: tm_seq ta1 t2) (w: tm_seq tb1 t2).
  Proof.
    intros.
    apply bisim_Cs_tm_seq; auto.
    intros; apply bisim_Cs_refl.
  Qed.

  Lemma bisim_Cs_tm_seq1': forall ta1 tb1 t2,
    term_bisimilar_Cs ta1 tb1 ->
    term_bisimilar_Cs (tm_seq ta1 t2) (tm_seq tb1 t2).
  Proof.
    intros; intro w.
    apply bisim_Cs_tm_seq1; auto.
  Qed.

  Lemma bisim_Cs_tm_seq2: forall w t1 ta2 tb2,
    (forall w', step_star_Cs (w:t1) (w':tm_skip) -> bisimilar_Cs (w':ta2) (w':tb2)) ->
    bisimilar_Cs (w: tm_seq t1 ta2) (w: tm_seq t1 tb2).
  Proof.
    intros.
    apply bisim_Cs_tm_seq; auto.
    apply bisim_Cs_refl.
  Qed.

  Lemma bisim_Cs_tm_par: forall w Pa1 Pb1 ta1 tb1 Pa2 Pb2 ta2 tb2,
    Pa1 == Pb1 -> Pa2 == Pb2 ->
    (forall w1, w1|=gsep_con Pa1 TT -> bisimilar_Cs (w1:ta1) (w1:tb1)) ->
    (forall w2, w2|=gsep_con Pa2 TT -> bisimilar_Cs (w2:ta2) (w2:tb2)) ->
    bisimilar_Cs (w: tm_par Pa1 ta1 Pa2 ta2) (w: tm_par Pb1 tb1 Pb2 tb2).
  Proof.
    intros.
    eapply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H3.
    exists (w1: tb1 || w2: tb2); split.
      apply step_single; apply st_par; auto.
      rewrite<- H; auto.
      rewrite<- H0; auto.
      apply bisim_Cs_par; eauto.
    intros.
    inverts H3.
    rewrite<- H in H12; rewrite<- H0 in H13.
    exists (w1: ta1 || w2: ta2); split.
      apply step_single; apply st_par; auto.
      apply bisim_Cs_par; eauto.
  Qed.

  Lemma bisim_Cs_tm_assert: forall w Pa Pb,
    (w|=gsep_con Pa TT <-> w|=gsep_con Pb TT) ->
    bisimilar_Cs (w: tm_assert Pa) (w: tm_assert Pb).
  Proof.
    intros.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H0.
    eexists; split.
    apply step_single; apply st_assert; eauto.
    apply H; auto.
    reflexivity.
    intros.
    inverts H0.
    eexists; split.
    apply step_single; apply st_assert; eauto.
    apply H; auto.
    reflexivity.
  Qed.

  Lemma bisim_Cs_tm_assert': forall Pa Pb,
    Pa == Pb ->
    term_bisimilar_Cs (tm_assert Pa) (tm_assert Pb).
  Proof.
    intros. intro w.
    apply bisim_Cs_tm_assert.
    setoid_rewrite H; reflexivity.
  Qed.

  Lemma bisim_Cs_context: forall C ba bb,
    bisimilar_Cs ba bb ->
    bisimilar_Cs (set_context C ba) (set_context C bb).
  Proof.
    intros.
    induction C; simpl.
    auto.
    apply bisim_Cs_seq1; auto.
    apply bisim_Cs_par1; auto.
    apply bisim_Cs_par2; auto.
  Qed.

  Lemma bisim_Cs_tm_assign: forall x e1 e2,
    e1 == e2 ->
    term_bisimilar_Cs (tm_assign x e1) (tm_assign x e2).
  Proof.
    intros.
    intro w.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H0.
    eexists; split.
    step_cons 1. eapply st_assign; eauto.
    setoid_rewrite<- H; eauto.
    reflexivity.
    intros.
    inverts H0.
    eexists; split.
    step_cons 1. eapply st_assign; eauto.
    setoid_rewrite H; eauto.
    reflexivity.
  Qed.

  Lemma bisim_Cs_tm_load: forall x e1 e2,
    e1 == e2 ->
    term_bisimilar_Cs (tm_load x e1) (tm_load x e2).
  Proof.
    intros.
    intro w.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H0.
    eexists; split.
    step_cons 1. eapply st_load; eauto.
    setoid_rewrite<- H; eauto.
    reflexivity.
    intros.
    inverts H0.
    eexists; split.
    step_cons 1. eapply st_load; eauto.
    setoid_rewrite H; eauto.
    reflexivity.
  Qed.

  Lemma bisim_Cs_tm_store: forall e1 e2 e3 e4,
    e1 == e3 -> e2==e4 ->
    term_bisimilar_Cs (tm_store e1 e2) (tm_store e3 e4).
  Proof.
    intros.
    intro w.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H1.
    eexists; split.
    step_cons 1. eapply st_store; eauto.
    setoid_rewrite<- H0; eauto.
    setoid_rewrite<- H; eauto.
    reflexivity.
    intros.
    inverts H1.
    eexists; split.
    step_cons 1. eapply st_store; eauto.
    setoid_rewrite H0; eauto.
    setoid_rewrite H; eauto.
    reflexivity.
  Qed.

  Lemma bisim_Cs_tm_local: forall e1 e2 f1 f2,
    e1 == e2 ->
    (forall x, term_bisimilar_Cs (f1 x) (f2 x)) ->
    term_bisimilar_Cs (tm_local e1 f1) (tm_local e2 f2).
  Proof.
    intros; intro w.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H1.
    eexists; split.
    step_cons 1. eapply st_local; eauto.
    setoid_rewrite<- H; eauto.
    apply H0.
    intros.
    inverts H1.
    eexists; split.
    step_cons 1. eapply st_local; eauto.
    setoid_rewrite H; eauto.
    apply H0.
  Qed.

  Lemma bisim_Cs_tm_local': forall e1 e2 f1 f2,
    e1 == e2 ->
    (forall x, term_bisimilar_Cs (f1 x) (f2 x)) ->
    term_bisimilar_Cs (local x:=e1 in f1 x) (local x:=e2 in f2 x).
  Proof.
    intros.
    apply bisim_Cs_tm_local; auto.
    intros.
    apply bisim_Cs_tm_seq1'; auto.
  Qed.


End Congruence.

(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)

Module Morphisms.
  Require Import Setoid.
  Require Export Morphisms.
  Export SetoidTactics.
  Require Export SetoidClass.
  Import Machine.Notation.
  Import Notation.
  Open Scope wterm_scope.

  Add Parametric Relation : wterm bisimilar_Cs
    reflexivity proved by bisim_Cs_refl
    symmetry proved by bisim_Cs_symm
    transitivity proved by bisim_Cs_trans
    as bisimilar_Cs_rel.

  Add Parametric Relation : term term_bisimilar_Cs
    reflexivity proved by term_bisim_Cs_refl
    symmetry proved by term_bisim_Cs_symm
    transitivity proved by term_bisim_Cs_trans
    as term_bisimilar_Cs_rel.

  Program Instance BisimCsSetoid : Setoid wterm :=
      { equiv:= bisimilar_Cs }.

  Program Instance BisimTermSetoid : Setoid term :=
      { equiv:= term_bisimilar_Cs }.

  Add Parametric Morphism: (bisimilar_Cs)
  with signature equiv ==> equiv ==> equiv
  as bisim_Cs_m.
  Proof.
    split; intros;
      cut (x==x0); intros; eauto;
      cut (y==y0); intros; eauto.
    rewrite<- H; rewrite<- H0; auto.
    rewrite H; rewrite H0; auto.
  Qed.

  Add Parametric Morphism: (term_bisimilar_Cs)
  with signature equiv ==> equiv ==> equiv
  as term_bisim_Cs_m.
  Proof.
    split; intros;
      cut (x==x0); intros; eauto;
      cut (y==y0); intros; eauto.
    rewrite<- H; rewrite<- H0; auto.
    rewrite H; rewrite H0; auto.
  Qed.

  Add Parametric Morphism: (wtm_seq)
  with signature equiv ==> equiv ==> equiv
  as bisim_Cs_seq_m.
  Proof.
    intros.
    apply bisim_Cs_seq; auto.
  Qed.

  Add Parametric Morphism: (wtm_par)
  with signature equiv ==> equiv ==> equiv
  as bisim_Cs_par_m.
  Proof.
    intros.
    apply bisim_Cs_par; auto.
  Qed.

  Add Parametric Morphism: (wtm_term)
  with signature eq ==> equiv ==> equiv
  as bisim_Cs_term_m.
  Proof. intros w t1 t2 ?. apply H. Qed.

  Add Parametric Morphism: (tm_seq)
  with signature equiv ==> equiv ==> equiv
  as bisim_Cs_tm_seq_m.
  Proof.
    intros. intro.
    apply bisim_Cs_tm_seq; auto.
  Qed.

  Add Parametric Morphism: (tm_par)
  with signature equiv ==> equiv ==> equiv ==> equiv ==> equiv
  as bisim_Cs_tm_par_m.
  Proof.
    intros Pa1 Pb1 ? ta1 tb1 ? Pa2 Pb2 ? ta2 tb2 ? w.
    apply bisim_Cs_tm_par; auto.
  Qed.

  Add Parametric Morphism: (tm_while_max)
  with signature expr_bool_equiv ==> eq ==> equiv ==> equiv
  as bisim_Cs_tm_while_max_m.
  Proof.
    intros ea eb ?  z ta tb ? w.
    apply bisim_Cs_while; auto.
  Qed.

  Add Parametric Morphism: (tm_while)
  with signature expr_bool_equiv ==> equiv ==> equiv
  as bisim_Cs_tm_while_m.
  Proof.
    intros ea eb ?  ta tb ? w.
    apply bisim_Cs_while; auto.
  Qed. 

  Add Parametric Morphism: (tm_if)
  with signature expr_bool_equiv ==> equiv ==> equiv ==> equiv
  as bisim_Cs_tm_if_m.
  Proof.
    intros e1 e2 ?  ta1 tb1 ? ta2 tb2 ? w.
    apply bisim_Cs_if; auto.
  Qed.

  Add Parametric Morphism: (tm_ift)
  with signature expr_bool_equiv ==> equiv ==> equiv
  as bisim_Cs_tm_ift_m.
  Proof.
    intros e1 e2 ?  ta1 tb1 ? w.
    apply bisim_Cs_if; auto.
    reflexivity.
  Qed.
 
  Add Parametric Morphism: (tm_assert)
  with signature equiv ==> equiv
  as bisim_Cs_tm_assert_m.
  Proof.
    intros Pa Pb ? w.
    apply bisim_Cs_tm_assert'; auto.
  Qed.

  Add Parametric Morphism: (tm_assign)
  with signature eq ==> equiv ==> equiv
  as bisim_Cs_tm_assign_m.
  Proof. intros; apply bisim_Cs_tm_assign; auto. Qed.

  Add Parametric Morphism: (tm_load)
  with signature eq ==> equiv ==> equiv
  as bisim_Cs_tm_load_m.
  Proof. intros; apply bisim_Cs_tm_load; auto. Qed.

  Add Parametric Morphism: (tm_store)
  with signature equiv ==> equiv ==> equiv
  as bisim_Cs_tm_store_m.
  Proof. intros; apply bisim_Cs_tm_store; auto. Qed.

  Program Instance bisim_Cs_tm_local_m :
    Proper (equiv ==> pointwise_relation var equiv ==> equiv)%signature (tm_local).
  Next Obligation.
    intros x y H f1 f2 H0 w.
    apply bisim_Cs_tm_local; auto.
  Qed.

  Add Parametric Morphism: (set_context)
  with signature eq ==> equiv ==> equiv
  as bisim_Cs_context_m.
  Proof.
    intros C ba bb ?.
    apply bisim_Cs_context; auto.
  Qed.

End Morphisms.


(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)

Section SimpleBisimulations.
  Import Machine.Notation.
  Import Notation.
  Import Notation.U.
(*  Import ContextSemantics.Notation. *)
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import Morphisms.
  Open Scope forces_pred.
(*  Open Scope context_scope. *)
  Open Scope wterm_scope.

  Lemma bisim_Cs_seq_assoc: forall b1 t2 t3,
    bisimilar_Cs (b1;; (t2;;; t3)) ((b1;; t2);; t3).
  Proof.
    intros.
    apply bisim_Cs_symm.
    exists (fun ba bb => (exists b1, ba=(b1;;t2);;t3 /\ bb=(b1;;tm_seq t2 t3)) \/ bisimilar_Cs ba bb).
    split; jauto.
    clear b1.
    intros ba bb [[b1[??]]|?]; subst.
    splits.

    intros [?]; discriminate.
    intros [?]; discriminate.

    intros s s''' b''' ?.
    inverts H. inverts H4.
    exists (b1'0;; tm_seq t2 t3); splits; auto.
      apply step_star_seq_app; eauto.
      apply step_single; auto.
      left; exists b1'0; auto.
    exists (w: t2;; t3); splits; auto.
      eapply step_cons.
      apply st_seq_skip.
      apply step_single.
      apply st_seq.
      right; apply bisim_Cs_refl.

    intros s s'' b'' ?.
    inverts H.
    exists (b1';; t2;; t3); splits; auto.
      do 2 apply step_star_seq_app.
      apply step_single; auto.
      left; exists b1'; auto.
    exists (w: t2;; t3); splits; auto.
      eapply step_single.
      apply st_seq1; apply st_seq_skip.
      right. apply bisim_Cs_symm. apply bisim_Cs_tm_seq_seq.

    splits; intros; auto.
    split; intros.
    apply bisim_Cs_halted1; auto.
    right; apply bisim_Cs_refl.
    split; intros.
    apply bisim_Cs_halted2; auto.
    right; apply bisim_Cs_refl.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    edestruct step_bisim_Cs2 as [ba'[??]]; eauto.
  Qed.
    


  Lemma bisim_Cs_tm_par_par_precise: forall w w1 w2 P1 P2 t1 t2,
       gprecise P1 -> gprecise P2 ->
       gjoin w1 w2 w ->
      w1 |= P1 -> w2|=P2 ->
       bisimilar_Cs (w : tm_par P1 t1 P2 t2) (w1 : t1 || w2 : t2).
  Proof.
    introv Hprec1 Hprec2.
    intros.
    eapply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H2.
    edestruct gprecise_TT_eq with (P1:=P1) (P2:=P2); eauto.
    subst w0 w3.
    clear H11 H12.
    exists (w1: t1 || w2:t2); split.
      apply step_nil.
      reflexivity.
    intros.
    inverts H2.
    exists (b1' || w2:t2); split.
      eapply step_cons.
      eapply st_par; eauto.
      apply TT_gstar; auto.
      apply TT_gstar; auto.
      apply step_star_par1_step.
      apply step_single; auto.
      reflexivity.
    exists (w1:t1 || b2'); split.
      eapply step_cons.
      eapply st_par; eauto.
      apply TT_gstar; auto.
      apply TT_gstar; auto.
      apply step_star_par2_step.
      apply step_single; auto.
      reflexivity.
    exists (w4: tm_skip); split.
      eapply step_cons.
      apply st_par; eauto.
      apply TT_gstar; auto.
      apply TT_gstar; auto.
      apply step_single.
      apply st_par_join; auto.
      reflexivity.
  Qed.

  Lemma bisim_Cs_tm_par_par: forall w w1 w2 t1 t2,
    gjoin w1 w2 w ->
    bisimilar_Cs (w: tm_par (exactly w1) t1 (exactly w2) t2) (w1:t1 || w2:t2).
  Proof.
    intros.
    apply bisim_Cs_tm_par_par_precise; auto.
    apply exactly_gprecise.
    apply exactly_gprecise.
    reflexivity.
    reflexivity.
  Qed.

  Lemma bisim_Cs_tm_par_swap: forall P1 P2 t1 t2,
    term_bisimilar_Cs (tm_par P1 t1 P2 t2) (tm_par P2 t2 P1 t1).
  Proof.
    intros. intro.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H.
    eexists; split.
    apply step_single; apply st_par; eauto.
    apply join_com; auto.
    apply bisim_Cs_par_swap.
    intros. 
    inverts H.
    eexists; split.
    apply step_single; apply st_par; eauto.
    apply join_com; auto.
    apply bisim_Cs_par_swap.
  Qed.


  Lemma bisim_Cs_seq_skip1': forall w t,
   bisimilar_Cs (w: tm_skip;; t) (w: t).
  Proof.
    intros.
    apply bisim_Cs_tao_step.
    intros. apply st_seq_skip; auto.
    intros.
    inverts H; split; auto; try solve [false; intuition].
    inverts H5. false; intuition.
    inverts H5.
  Qed.    

  Lemma bisim_Cs_seq_skip1: forall b w t,
   bisimilar_Cs b (w:tm_skip) ->
   bisimilar_Cs (b;; t) (w: t).
  Proof.
    intros.
    eapply bisim_Cs_trans.
    apply bisim_Cs_seq1; eauto.
    apply bisim_Cs_seq_skip1'.
  Qed.

  Lemma bisim_Cs_tm_seq_skip1': forall t,
     term_bisimilar_Cs (tm_seq tm_skip t) t.
  Proof.
    intros. intro.
    eapply bisim_Cs_trans.
    apply bisim_Cs_tm_seq_seq.
    apply bisim_Cs_seq_skip1'.
  Qed.

  Lemma bisim_Cs_tm_seq_skip1: forall t1 t2,
     term_bisimilar_Cs t1 tm_skip ->
     term_bisimilar_Cs (tm_seq t1 t2) t2.
  Proof.
    intros. intro.
    eapply bisim_Cs_trans.
    apply bisim_Cs_tm_seq1; eauto.
    apply bisim_Cs_tm_seq_skip1'.
  Qed.

  Lemma bisim_Cs_seq_skip2': forall b,
   bisimilar_Cs (b;; tm_skip) b.
  Proof.
    intros.
    exists (fun ba bb => ba=bb;;tm_skip \/ ba=bb).
    split; auto.
    intros ba bb [?|?]; subst.
    splits.
    intros [?]; discriminate.

    intros [w ?]; split; auto.
    subst bb.
    intros; eapply step_cons. apply st_seq_skip. apply step_nil.

    intros s s'' b'' ?.
    inverts H.
    exists b1'; splits; auto.
      apply step_single; auto.
    exists (w: tm_skip); split; auto.
      apply step_nil.

    intros s s' bb' ?.
    exists (bb';; tm_skip); splits; auto.
      apply step_star_seq_app; auto.
      apply step_single; auto.

    splits; intros; auto.
    split; intros; auto; apply step_nil.
    split; intros; auto; apply step_nil.
    exists b1'; split; auto; apply step_single; auto.
    exists b2'; split; auto; apply step_single; auto.
  Qed.

  Lemma bisim_Cs_seq_skip2: forall b t,
   (forall w', bisimilar_Cs (w':t) (w':tm_skip)) ->
   bisimilar_Cs (b;; t) b.
  Proof.
    intros.
    eapply bisim_Cs_trans.
    eapply bisim_Cs_seq2; eauto.
    apply bisim_Cs_seq_skip2'.
  Qed.

  Lemma bisim_Cs_tm_seq_skip2': forall t,
   term_bisimilar_Cs (tm_seq t tm_skip) t.
  Proof.
    intros. intro.
    eapply bisim_Cs_trans.
    apply bisim_Cs_tm_seq_seq.
    apply bisim_Cs_seq_skip2'.
  Qed.

  Lemma bisim_Cs_tm_seq_skip2: forall t1 t2,
   term_bisimilar_Cs t2 tm_skip ->
   term_bisimilar_Cs (tm_seq t1 t2) t1.
  Proof.
    intros. intro.
    eapply bisim_Cs_trans.
    eapply bisim_Cs_tm_seq2; eauto.
    apply bisim_Cs_tm_seq_skip2'.
  Qed.

  Lemma bisim_Cs_assert_skip: forall P F w,
    w|=gsep_con P F ->
    bisimilar_Cs (w: tm_assert P) (w: tm_skip).
  Proof.
    intros.
    apply bisim_Cs_tao_step.
    intros.
    apply st_assert; auto.
    eapply gstar_intro with (x:=P) (y:=F); eauto.
    apply TT_intro.
    intros. inverts H0; auto.
  Qed.


  Lemma bisim_Cs_assert_evalFalse_skip: forall w e,
    evalFalse e w ->
    bisimilar_Cs (w: tm_assert (pIsFalse e)) (w: tm_skip).
  Proof.
    intros.
    apply bisim_Cs_assert_skip with TT.
    apply TT_gstar; auto.
  Qed.

  Lemma bisim_Cs_while_inf_post_false: forall w e t,
    bisimilar_Cs (w: tm_while_max e i_inf t) (w: tm_while_max e i_inf t;; tm_assert (pIsFalse e)).
  Proof.
    intros.
    setoid_rewrite<- bisim_Cs_seq_skip2' at 1.
    apply bisim_Cs_seq2.
    intros.
    symmetry.
    apply bisim_Cs_assert_evalFalse_skip.
    eapply step_star_Cs_while_inf_skip_resolve; eauto.
  Qed.

  Lemma bisim_Cs_while_inf_post_false': forall e t,
    term_bisimilar_Cs (tm_while_max e i_inf t) (tm_while_max e i_inf t;;; tm_assert (pIsFalse e)).
  Proof.
    intros. intro.
    eapply bisim_Cs_trans.
    apply bisim_Cs_while_inf_post_false.
    apply bisim_Cs_symm.
    apply bisim_Cs_tm_seq_seq.
  Qed.

  Lemma bisim_Cs_while_false_skip': forall (z:iterations) w e t,
   evalFalse e w ->
   bisimilar_Cs (w: tm_while_max e z t) (w: tm_skip).
  Proof.
    intros.
    eapply bisim_Cs_tao_step.
    intros.
    apply st_while_false; auto.
    intros.
    inverts H0; auto.
    false; eapply evalTrue_evalFalse; eauto.
  Qed.

  Lemma bisim_Cs_false_while_skip: forall z w e t,
   evalFalse e w ->
   bisimilar_Cs (w: tm_while_max e z t) (w: tm_skip).
  Proof.
    intros.
    setoid_rewrite<- bisim_Cs_seq_skip1' at 2.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H0.
    false; eapply evalTrue_evalFalse; eauto.
    exists (w: tm_skip); intuition. apply step_single; apply st_seq_skip.
    exists (w: tm_skip); split. apply step_single; apply st_seq_skip. apply bisim_Cs_while_false_skip'; auto.
    exists (w: tm_skip); intuition. apply step_single; apply st_seq_skip.
    intros.
    inverts H0. inverts H6.
    exists (w: tm_skip); intuition. apply step_single; apply st_while_false; auto.
  Qed.

  Lemma bisim_Cs_while_zero_skip: forall e t,
   term_bisimilar_Cs (tm_while_max e 0 t) tm_skip.
  Proof.
    intros. intro w.
    apply bisim_Cs_tao_step.
    intros.
    apply st_while_limit; auto.
    intros.
    inverts H; auto.
    false. inverts H8. 
  Qed.

  Lemma bisim_Cs_assert_intro': forall w P ta tb,
    (w|=gsep_con P TT -> bisimilar_Cs (w:ta) (w:tb)) ->
    bisimilar_Cs (w: tm_assert P;; ta) (w: tm_assert P;; tb).
  Proof.
    intros.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H0. inverts H6.
    eexists; split.
    step_cons 2. apply st_seq1; apply st_assert; eauto. apply st_seq_skip.
    setoid_rewrite bisim_Cs_seq_skip1'.
    apply H; auto.
    intros.
    inverts H0. inverts H6.
    eexists; split.
    step_cons 2. apply st_seq1; apply st_assert; eauto. apply st_seq_skip.
    setoid_rewrite bisim_Cs_seq_skip1'.
    apply H; auto.
  Qed.

  Lemma bisim_Cs_assert_intro: forall P ta tb,
    (forall w, w|=gsep_con P TT -> bisimilar_Cs (w:ta) (w:tb)) ->
    term_bisimilar_Cs (tm_assert P;;; ta) (tm_assert P;;; tb).
  Proof.
    intros. intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_assert_intro'; auto.
  Qed.

  Lemma bisim_Cs_assert_false_while_nop: forall e t z,
    term_bisimilar_Cs (tm_assert (pIsFalse e);;; tm_while_max e z t) (tm_assert (pIsFalse e)).
  Proof.
    intros.
    setoid_rewrite<- bisim_Cs_tm_seq_skip2' at 5.
    apply bisim_Cs_assert_intro; intros.
    apply bisim_Cs_false_while_skip.
    eapply pIsFalse_gstar; eauto.
  Qed.

  Open Scope entails_pred.
  Lemma bisim_Cs_while_split_inf: forall (u:iterations) e1 e2 t,
    pIsTrue e1 |-- pIsTrue e2 ->
    pIsTrue e2 |-- pIsTrue e1 || pIsFalse e1 ->
    pIsFalse e2 |-- pIsFalse e1 ->
    term_bisimilar_Cs (tm_while_max e1 u t;;; tm_while_max e2 i_inf t) (tm_while_max e2 i_inf t).
  Proof.
    introv Hexpr1 Hexpr2 Hexpr3.
    intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    exists (fun ba bb =>
      (exists (u:iterations) w,
        ba=w:tm_while_max e1 u t;; tm_while_max e2 i_inf t /\
        bb=w:tm_while_max e2 i_inf t)
      \/
      (exists (u:iterations) bba bbb,
        ba=bba;; tm_while_max e1 u t;; tm_while_max e2 i_inf t /\
        bb=bbb;; tm_while_max e2 i_inf t /\
        bisimilar_Cs bba bbb)
      \/ bisimilar_Cs ba bb).
    split.
    2: left; eauto.
    clear u w.
    intros ba bb [ [u[w[??]]] |[ [u[bba[bbb[?[??]]]]] | ? ]]; subst.

    splits.
    intros [?]; discriminate.
    intros [?]; discriminate.

    intros s s' ba' ?.
    inverts H. inverts H4.
    exists (w: t;; tm_while_max e2 i_inf t); splits; auto.
      apply step_single.
      apply st_while_true; auto.
      apply Hexpr1; auto.
      constructor.
      inverts H7.
      right; left; do 3 eexists; splits; eauto. reflexivity.
      right; left; do 3 eexists; splits; eauto. reflexivity.
    exists (w: tm_while_max e2 i_inf t); splits; auto.
      apply step_nil.
      right; right; apply bisim_Cs_seq_skip1'.
    exists (w: tm_while_max e2 i_inf t); splits; auto.
      apply step_nil.
      right; right; apply bisim_Cs_seq_skip1'.

    intros s s' bb' ?.
    inverts H.
    inverts H7.
    destruct u.
    destruct n.
    exists (w: t;; tm_while_max e2 i_inf t); splits; auto.
      eapply step_cons.
      apply st_seq1; apply st_while_limit.
      eapply step_cons.
      apply st_seq_skip.
      apply step_single.
      apply st_while_true; auto.
      constructor.
      right; right; reflexivity.
    edestruct Hexpr2; eauto.
    exists (w: t;; tm_while_max e1 n t;; tm_while_max e2 i_inf t); splits; auto.
      apply step_single.
      apply st_seq1; apply st_while_true; auto.
      constructor.
      right; left; do 3 eexists; splits; eauto. reflexivity.
    exists (w: t;; tm_while_max e2 i_inf t); splits; auto.
      eapply step_cons.
      apply st_seq1; apply st_while_false; auto.
      eapply step_cons.
      apply st_seq_skip.
      apply step_single.
      apply st_while_true; auto.
      constructor.
      right; right; reflexivity.
    edestruct Hexpr2; eauto.
    exists (w: t;; tm_while_max e1 i_inf t;; tm_while_max e2 i_inf t); splits; auto.
      apply step_single.
      apply st_seq1; apply st_while_true; auto.
      constructor.
      right; left; do 3 eexists; splits; eauto. reflexivity.
    exists (w: t;; tm_while_max e2 i_inf t); splits; auto.
      eapply step_cons.
      apply st_seq1; apply st_while_false; auto.
      eapply step_cons.
      apply st_seq_skip.
      apply step_single.
      apply st_while_true; auto.
      constructor.
      right; right; reflexivity.
    exists (w: tm_skip); splits; auto.
      eapply step_cons.
      apply st_seq1; apply st_while_false.
      apply Hexpr3; auto.
      eapply step_cons.
      apply st_seq_skip.
      apply step_single.
      apply st_while_false; auto.
      do 2 right; reflexivity.

    splits.
    intros [?]; discriminate.
    intros [?]; discriminate.

    intros s s' ba' ?.
    inverts H.
    inverts H5.
    edestruct step_bisim_Cs1 as [bbb'[??]]; eauto.
    exists (bbb';; tm_while_max e2 i_inf t); splits; auto.
      apply step_star_seq_app; auto.
      right; left; do 3 eexists; splits; eauto.
    exists (w: tm_while_max e2 i_inf t); splits; auto.
      eapply step_star_seq_join.
      eapply bisim_Cs_halted1; eauto. eexists; eauto.
      apply step_nil.
      left; eauto.

    intros s s' bb' ?.
    inverts H.
    edestruct step_bisim_Cs2 as [bba'[??]]; eauto.
    exists (bba';; tm_while_max e1 u t;; tm_while_max e2 i_inf t); splits.
      do 2 eapply step_star_seq_app; eauto.
      right; left; do 3 eexists; eauto.
    exists (w: tm_while_max e1 u t;; tm_while_max e2 i_inf t); splits; auto.
      apply step_star_seq_app.
      eapply step_star_seq_join.
      eapply bisim_Cs_halted2; eauto. eexists; eauto.
      apply step_nil.
      left; eauto.

    splits; intros.
    split; intros.
    eapply bisim_Cs_halted1; eauto.
    right; right; apply bisim_Cs_refl.
    split; intros.
    eapply bisim_Cs_halted2; eauto.
    right; right; apply bisim_Cs_refl.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    edestruct step_bisim_Cs2 as [ba'[??]]; eauto.
  Qed.


  Lemma bisim_Cs_while_split_iter_plus2: forall z1 (z2:iterations) e t,
    term_bisimilar_Cs (tm_while_max e z1 t;;; tm_while_max e z2 t) (tm_while_max e (iter_plus2 z1 z2) t).
  Proof.
    intros.
    intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    exists (fun ba bb =>
      (exists z1, exists (z2:iterations) w,
        ba=w:tm_while_max e z1 t;; tm_while_max e z2 t /\
        bb=w:tm_while_max e (iter_plus2 z1 z2) t)
      \/
      (exists z1, exists (z2: iterations) bba bbb,
        ba=bba;; tm_while_max e z1 t;; tm_while_max e z2 t /\
        bb=bbb;; tm_while_max e (iter_plus2 z1 z2) t /\
        bisimilar_Cs bba bbb)
      \/ bisimilar_Cs ba bb).
    split.
    2: left; eauto.
    clear z1 z2 w.
    intros ba bb [[z1[z2[w[??]]]]|[[z1[z2[bba[bbb[?[??]]]]]]|?]]; subst.

    splits.
    intros [?]; discriminate.
    intros [?]; discriminate.

    intros s s' ba' ?.
    inverts H. inverts H4.
    exists (w: t;; tm_while_max e (iter_plus2 z' z2) t); splits; auto.
      apply step_single.
      apply st_while_true; auto.
      apply iteration_decr_iter_plus2; auto.
      right; left; do 4 eexists; splits; eauto. reflexivity.
    exists (w: tm_skip); splits; auto.
      apply step_single.
      apply st_while_false; auto.
      do 2 right.
      setoid_rewrite bisim_Cs_seq_skip1'.
      apply bisim_Cs_false_while_skip; auto.
    exists (w: tm_while_max e (iter_plus2 z z2) t); splits; auto.
      apply step_single.
      rewrite iter_plus2_iterations.
      simpl.
      apply st_while_go; auto.
      left; do 3 eexists; split; eauto.
    exists (w: tm_while_max e (iter_plus2 0 z2) t); splits; auto.
      apply step_nil.
      do 2 right.
      rewrite iter_plus2_iterations.
      rewrite i_plus_zero1.
      apply bisim_Cs_seq_skip1'.

    intros s s' bb' ?.
    inverts H.
    edestruct iteration_decr_iter_plus2_resolve as [[??] | [z1'[??]]]; eauto.
    subst z1.
    exists (w: t;; tm_while_max e z' t); splits; auto.
      eapply step_cons.
      apply st_seq1; apply st_while_limit.
      eapply step_cons.
      apply st_seq_skip.
      apply step_single.
      apply st_while_true; auto.
      do 2 right; reflexivity.
    subst z'.
    exists (w: t;; tm_while_max e z1' t;; tm_while_max e z2 t); splits; auto.
      apply step_single.
      apply st_seq1; apply st_while_true; auto.
      right; left; do 4 eexists; splits; eauto. reflexivity.
    exists (w: tm_skip); splits; auto.
      eapply step_cons.
      apply st_seq1; apply st_while_false; auto.
      eapply step_cons.
      apply st_seq_skip.
      apply step_single.
      apply st_while_false; auto.
      do 2 right; reflexivity.
    edestruct iter_plus2_i_stall as [z2a[??]]; eauto.
    subst z1 z.
    exists (w: tm_while_max e z2a t;; tm_while_max e z2 t); splits; auto.
      apply step_single.
      apply st_seq1; apply st_while_go; auto.
      left; do 3 eexists; split; eauto.
      rewrite iter_plus2_iterations; reflexivity.
    edestruct iter_plus2_nat as [n1[n2[?[??]]]]; eauto.
    subst z1 z2.
    assert (n1 = 0); [omega | subst].
    assert (n2 = 0); [omega | subst].
    exists (w: tm_skip); splits; auto.
      eapply step_cons.
      apply st_seq1; apply st_while_limit.
      eapply step_cons.
      apply st_seq_skip.
      apply step_single.
      apply st_while_limit.
      do 2 right; reflexivity.

    splits.
    intros [?]; discriminate.
    intros [?]; discriminate.

    intros s s' ba' ?.
    inverts H.
    inverts H5.
    edestruct step_bisim_Cs1 as [bbb'[??]]; eauto.
    exists (bbb';; tm_while_max e (iter_plus2 z1 z2) t); splits; auto.
      apply step_star_seq_app; auto.
      right; left; do 4 eexists; splits; eauto.
    exists (w: tm_while_max e (iter_plus2 z1 z2) t); splits; auto.
      eapply step_star_seq_join.
      eapply bisim_Cs_halted1; eauto. eexists; eauto.
      apply step_nil.
      left; eauto.

    intros s s' bb' ?.
    inverts H.
    edestruct step_bisim_Cs2 as [bba'[??]]; eauto.
    exists (bba';; tm_while_max e z1 t;; tm_while_max e z2 t); splits.
      do 2 eapply step_star_seq_app; eauto.
      right; left; do 4 eexists; eauto.
    exists (w: tm_while_max e z1 t;; tm_while_max e z2 t); splits; auto.
      apply step_star_seq_app.
      eapply step_star_seq_join.
      eapply bisim_Cs_halted2; eauto. eexists; eauto.
      apply step_nil.
      left; eauto.

    splits; intros.
    split; intros.
    eapply bisim_Cs_halted1; eauto.
    right; right; apply bisim_Cs_refl.
    split; intros.
    eapply bisim_Cs_halted2; eauto.
    right; right; apply bisim_Cs_refl.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    edestruct step_bisim_Cs2 as [ba'[??]]; eauto.
  Qed.

  Lemma bisim_Cs_while_split_iter_plus2': forall w z1 (z2:iterations) e t,
    bisimilar_Cs (w: tm_while_max e z1 t;; tm_while_max e z2 t) (w: tm_while_max e (iter_plus2 z1 z2) t).
  Proof.
    intros.
    setoid_rewrite<- bisim_Cs_tm_seq_seq.
    apply bisim_Cs_while_split_iter_plus2.
  Qed.

  Lemma bisim_Cs_while_split_nat: forall (n m: nat) e t,
    term_bisimilar_Cs (tm_seq (tm_while_max e n t) (tm_while_max e m t)) (tm_while_max e (n+m) t).
  Proof.
    intros.
    apply bisim_Cs_while_split_iter_plus2.
  Qed.

  Lemma bisim_Cs_while_split_inf1: forall (z:iterations) e t,
    term_bisimilar_Cs (tm_seq (tm_while_max e i_inf t) (tm_while_max e z t)) (tm_while_max e i_inf t).
  Proof.
    intros.
    change (iter_const i_inf) with (iter_plus2 i_inf z) at 2.
    apply bisim_Cs_while_split_iter_plus2.
  Qed.

  Lemma bisim_Cs_while_split_inf2: forall (z:iterations) e t,
    term_bisimilar_Cs (tm_seq (tm_while_max e z t) (tm_while_max e i_inf t)) (tm_while_max e i_inf t).
  Proof.
    intros.
    apply bisim_Cs_while_split_inf; auto.
    apply or_intro1; auto.
  Qed.

  Lemma bisim_Cs_par_skip: forall w1 w2 w,
    gjoin w1 w2 w ->
    bisimilar_Cs (w1:tm_skip || w2:tm_skip) (w: tm_skip).
  Proof.
    intros.
    apply bisim_Cs_tao_step.
    intros.
    apply st_par_join; auto.
    intros.
    inverts H0.
    false; intuition.
    inverts H6.
    false; intuition.
    inverts H6.
    split; auto. f_equal.
    eapply @join_eq with (H:=gworld_sa); eauto.
  Qed.


  Lemma bisim_Cs_tm_seq_assoc: forall t1 t2 t3,
    term_bisimilar_Cs (t1;;; (t2;;; t3)) ((t1;;; t2);;; t3).
  Proof.
    intros. intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_seq_assoc.
  Qed.


  Lemma bisim_Cs_assert_skip': forall w P F,
    w |= P*F ->
    bisimilar_Cs (w: tm_assert P) (w: tm_skip).
  Proof.
    intros.
    apply bisim_Cs_tao_step.
    intros.
    apply st_assert; auto.
    apply gstar_intro with (x:=P) (y:=rm_vars TT); auto.
    repeat intro; simpl; auto.
    apply star_gstar.
    apply star_intro with (x:=P) (y:=F); auto.
    repeat intro; simpl; auto.
    intros.
    inverts H0.
    auto.
  Qed.

  Lemma bisim_Cs_blocked: forall ba bb,
    ~halted ba -> ~halted bb ->
    (forall s s' b', step s ba s' b' -> b'=ba) ->
    (forall s s' b', step s bb s' b' -> b'=bb) ->
    bisimilar_Cs ba bb.
  Proof.
    intros.
    rename ba into ba0; rename bb into bb0.
    exists (fun ba bb => ba = ba0 /\ bb = bb0).
    split; auto.
    intros ba bb [??]; subst.
    splits; try contradiction.

    intros s s' ba' ?.
    spec H1 s s' ba' H3. subst ba0.
    apply step_spin_no_change in H3. subst.
    exists bb0; splits; auto.
      apply step_nil.

    intros s s' bb' ?.
    spec H2 s s' bb' H3. subst bb0.
    apply step_spin_no_change in H3. subst.
    exists ba0; splits; auto.
      apply step_nil.
  Qed.

  Lemma bisim_Cs_while_one_if: forall e t,
    term_bisimilar_Cs (tm_while_max e 1 t) (tm_if e t tm_skip).
  Proof.
    intros.
    intro w.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H.
    inverts H8.
    exists (w: t)%wterm; split.
      apply step_single; apply st_if_true; auto.
      setoid_rewrite bisim_Cs_while_zero_skip.
      apply bisim_Cs_seq_skip2'.
    exists (w: tm_skip)%wterm; split.
      apply step_single; apply st_if_false; auto.
      reflexivity.
    intros.
    inverts H.
    exists (w: t;; tm_while_max e 0 t)%wterm; split.
      apply step_single; apply st_while_true; auto.
      constructor.
      setoid_rewrite bisim_Cs_while_zero_skip.
      apply bisim_Cs_seq_skip2'.
    exists (w: tm_skip)%wterm; split.
      apply step_single; apply st_while_false; auto.
      reflexivity.
  Qed.


  Lemma bisim_Cs_assert_assert_and: forall P1 P2,
    term_bisimilar_Cs (tm_assert P1;;; tm_assert P2) (tm_assert (gsep_overlap P1 P2)).
  Proof.
    intros.
    intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.

    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H. inverts H5.
    exists (w: tm_assert (gsep_overlap P1 P2)%pred); split.
      apply step_nil.
    setoid_rewrite bisim_Cs_seq_skip1'.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H.
    exists (w:tm_skip); split.
      apply step_single. apply st_assert; auto.
      apply TT_gstar.
      setoid_rewrite gsep_overlap_con.
      split; auto.
    reflexivity.
    intros.
    inverts H.
    exists (w:tm_skip); split.
      apply step_single. apply st_assert; auto.
      setoid_rewrite gsep_overlap_con in H7.
      eapply gstar_intro with (x:=P2) (y:=gsep_con TT TT); eauto.
      repeat intro; simpl; auto.
      setoid_rewrite gstar_assoc.
      eapply gstar_intro. 3: apply H7.
      apply and_elim4; auto.
      auto.
      reflexivity.

    intros.
    inverts H.
    exists (w:tm_skip); split.
      eapply step_cons.
      apply st_seq1. apply st_assert; auto. 
      setoid_rewrite gsep_overlap_con in H5.
      eapply gstar_intro with (y:=(gsep_con TT TT)%pred); eauto.
      repeat intro; simpl; auto.
      setoid_rewrite gstar_assoc.
      eapply gstar_intro. 3: apply H5.
      apply and_elim3; auto.
      auto.
      eapply step_cons.
      apply st_seq_skip.
      apply step_single.
      apply st_assert.
      setoid_rewrite gsep_overlap_con in H5.
      eapply gstar_intro with (y:=(gsep_con TT TT)%pred); eauto.
      repeat intro; simpl; auto.
      setoid_rewrite gstar_assoc.
      eapply gstar_intro. 3: apply H5.
      apply and_elim4; auto.
      auto.
      reflexivity.
  Qed.

  Lemma bisim_Cs_assert_assert_shadow: forall P1 P2,
    (P1 |-- P2)%entails_pred ->
    term_bisimilar_Cs (tm_assert P1;;; tm_assert P2) (tm_assert P1).
  Proof.
    intros.
    intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H0.
    inverts H6.
    exists (w: tm_skip)%wterm; split.
      apply step_single; apply st_assert; auto.
      setoid_rewrite bisim_Cs_seq_skip1'.
      apply bisim_Cs_assert_skip with TT; auto.
      eapply split_gcon; eauto.
    intros.
    inverts H0.
    exists (w: tm_skip;; tm_assert P2)%wterm; split; auto.
      apply step_single; apply st_seq1; apply st_assert; auto.
      setoid_rewrite bisim_Cs_seq_skip1'.
      apply bisim_Cs_assert_skip with TT; auto.
      eapply split_gcon; eauto.
  Qed.

  Lemma bisim_Cs_tm_par_skip: forall w1 w2 w P1 P2 F1 F2,
    w1|=gsep_con P1 F1 -> w2|=gsep_con P2 F2 ->
    gjoin w1 w2 w ->
    bisimilar_Cs (w: tm_par P1 tm_skip P2 tm_skip) (w: tm_skip).
  Proof.
    intros.
    setoid_rewrite<- bisim_Cs_seq_skip1' at 2.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H2.
    exists (w: tm_skip)%wterm; split.
      apply step_single; apply st_seq_skip.
      apply bisim_Cs_par_skip; auto.
    intros.
    inverts H2. inverts H8.
    exists (w1:tm_skip || w2:tm_skip)%wterm; split.
      apply step_single; apply st_par; auto.
      apply split_gcon with P1 F1; auto.
      eapply TT_intro.
      apply split_gcon with P2 F2; auto.
      eapply TT_intro.
      apply bisim_Cs_par_skip; auto.
  Qed.


  Lemma bisim_Cs_assert_false_if_nop: forall e t1 t2,
    term_bisimilar_Cs (tm_assert (pIsFalse e);;; tm_if e t1 t2) (tm_assert (pIsFalse e);;; t2).
  Proof.
    intros.
    apply bisim_Cs_assert_intro; intros.
    apply bisim_Cs_if_false.
    eapply pIsFalse_gstar; eauto.
  Qed.

  Lemma bisim_Cs_ift_split: forall e t1 t2,
    term_bisimilar_Cs (tm_ift e (t1;;; tm_ift e t2)) ((tm_ift e t1);;; tm_ift e t2).
  Proof.
    intros.
    intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    edestruct (evalTrueFalse_dec e w) as [[?|?] | [??]].

    setoid_rewrite bisim_Cs_if_true; auto.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    reflexivity.

    setoid_rewrite bisim_Cs_if_false; auto.
    setoid_rewrite bisim_Cs_seq_skip1'.
    setoid_rewrite bisim_Cs_if_false; auto.
    reflexivity.

    apply bisim_Cs_blocked.
    intros [?]; discriminate.
    intros [?]; discriminate.
    introv H1; inverts H1; contradiction.

    intros. inverts H1; try contradiction. inverts H6; contradiction.
  Qed.

(*
Import world_model.Morphisms.
Import Setoid.
*)

  Lemma bisim_Cs_while_assert_true: forall e z t,
    term_bisimilar_Cs (tm_while_max e z (tm_assert (pIsTrue e);;; t)) (tm_while_max e z t).
  Proof.
    intros.
    apply bisim_Cs_while.
    reflexivity.
    intros.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    setoid_rewrite bisim_Cs_assert_skip.
    setoid_rewrite bisim_Cs_seq_skip1'.
    reflexivity.
    apply TT_gstar; auto.
  Qed.

  Lemma bisim_Cs_if_assert_true: forall e t1 t2,
    term_bisimilar_Cs (tm_if e t1 t2) (tm_if e (tm_assert (pIsTrue e);;; t1) t2).
  Proof.
    intros. intro w.
    eapply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H.
    exists (w: t1)%wterm; split.
      eapply step_cons.
      apply st_if_true; auto.
      eapply step_cons.
      apply st_seq.
      eapply step_cons.
      apply st_seq1; apply st_assert; apply TT_gstar; auto.
      apply step_single.
      apply st_seq_skip.
      reflexivity.
    exists (w: t2)%wterm; split.
      apply step_single.
      apply st_if_false; auto.
      reflexivity.
    intros.
    inverts H.
    exists (w: t1)%wterm; split.
      apply step_single.
      apply st_if_true; auto.
      symmetry.
      setoid_rewrite bisim_Cs_tm_seq_seq.
      setoid_rewrite<- bisim_Cs_seq_skip1' at 2.
      apply bisim_Cs_seq1.
      apply bisim_Cs_assert_skip with TT.
      apply TT_gstar; auto.
    exists (w: t2)%wterm; split.
      apply step_single.
      apply st_if_false; auto.
      reflexivity.
  Qed.

  Lemma bisim_Cs_assert_true_if: forall e t1 t2,
    term_bisimilar_Cs (tm_assert (pIsTrue e);;; tm_if e t1 t2) (tm_assert (pIsTrue e);;; t1).
  Proof.
    intros.
    apply bisim_Cs_assert_intro; intros.
    apply bisim_Cs_if_true.
    eapply pIsTrue_gstar; eauto.
  Qed.

  Lemma bisim_Cs_assert_false_if: forall e t1 t2,
    term_bisimilar_Cs (tm_assert (pIsFalse e);;; tm_if e t1 t2) (tm_assert (pIsFalse e);;; t2).
  Proof.
    intros.
    apply bisim_Cs_assert_intro; intros.
    apply bisim_Cs_if_false.
    eapply pIsFalse_gstar; eauto.
  Qed.

  Lemma bisim_Cs_ift_assert_false_nop: forall e t1,
    term_bisimilar_Cs (tm_assert (pIsFalse e);;; tm_if e t1 tm_skip) (tm_assert (pIsFalse e)).
  Proof.
    intros.
    setoid_rewrite<- bisim_Cs_tm_seq_skip2' at 6.
    setoid_rewrite<- bisim_Cs_assert_false_if_nop with (t1:=t1) (t2:=tm_skip).
    reflexivity.
  Qed.

  Lemma bisim_Cs_while_peel: forall w t e z z',
    evalTrue e w ->
    iteration_decr z' z ->
    bisimilar_Cs (w: t;; tm_while_max e z t) (w: tm_while_max e z' t).
  Proof.
    intros.
    symmetry.
    apply bisim_Cs_tao_step.
    intros.
    apply st_while_true; auto.
    intros.
    inverts H1; auto.
    inverts H0; inverts H10; auto.
    false; eapply evalTrue_evalFalse; eauto.
    inverts H0.
    inverts H0.
  Qed.

  Lemma bisim_Cs_par_skip_assoc: forall w1 w2 w3 w12 w23,
    gjoin w1 w2 w12 -> gjoin w2 w3 w23 ->
    bisimilar_Cs (w1: tm_skip || w23: tm_skip) (w12: tm_skip || w3: tm_skip).
  Proof.
    intros.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H1; try solve [false; inverts H7].
    clear H2.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w3) (b:=w2) as [w12'[??]]; eauto.
    assert (w12' = w12); subst.
      eapply @join_eq with (H:=gworld_sa); eauto.
    exists (w: tm_skip); split.
      apply step_single; apply st_par_join.
      apply join_com; auto.
      reflexivity.
    intros.
    inverts H1; try solve [false; inverts H7].
    clear H2.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w1) (b:=w2) as [w23'[??]]; eauto.
    assert (w23' = w23); subst.
      eapply @join_eq with (H:=gworld_sa); eauto.
    exists (w: tm_skip); split.
      apply step_single; apply st_par_join; auto.
      reflexivity.
  Qed.

  Lemma bisim_Cs_par_assoc: forall b1 b2 b3,
    bisimilar_Cs (b1 || (b2 || b3)) (b1 || b2 || b3).
  Proof.
    intros.
    exists (fun ba bb =>
      (exists b1 b2 b3, ba = b1 || (b2 || b3) /\ bb = b1 || b2 || b3) \/
      (exists b1 w2 w3 w23, ba = b1 || w23: tm_skip /\ bb = b1 || w2: tm_skip || w3: tm_skip /\ gjoin w2 w3 w23) \/
      (exists b3 w1 w2 w12, ba = w1:tm_skip || (w2: tm_skip || b3) /\ bb = w12: tm_skip || b3 /\ gjoin w1 w2 w12) \/
      bisimilar_Cs ba bb).
    split; iauto.
    clear b1 b2 b3.
    intros ba bb [ [b1[b2[b3[??]]]] | [ [b1[w2[w3[w23[?[??]]]]]] | [ [b3[w1[w2[w12[?[??]]]]]] | ? ] ]]; subst.

    splits.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H.
    rename b1'0 into b1'.
    exists (b1' || b2 || b3); split; iauto.
      apply step_single; apply st_par1; apply st_par1; auto.
    inverts H4.
    rename b1' into b2'.
    exists (b1 || b2' || b3); split; iauto.
      apply step_single; apply st_par1; apply st_par2; auto.
    rename b2'0 into b3'.
    exists (b1 || b2 || b3'); split; iauto.
      apply step_single; apply st_par2; auto.
    eexists; split.
      apply step_nil. right; left. exists b1 w1 w2 w; splits; auto.
    intros.
    inverts H.
    inverts H4.
    rename b1'0 into b1'.
    exists (b1' || (b2 || b3)); split; iauto.
      apply step_single; apply st_par1; auto.
    exists (b1 || (b2' || b3)); split; iauto.
      apply step_single; apply st_par2; apply st_par1; auto.
    eexists; split.
      apply step_nil.
      right; right; left; do 4 eexists; splits; eauto.
    rename b2'0 into b3'.
    exists (b1 || (b2 || b3')); split; iauto.
      apply step_single; apply st_par2; apply st_par2; auto.

    splits.    
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H.
    rename b1'0 into b1'.
    exists (b1' || w2:tm_skip || w3:tm_skip); split.
      apply step_single; apply st_par1; apply st_par1; auto.
      right; left; do 4 eexists; eauto.
    inverts H5.
    exists (w: tm_skip); split.
      edestruct @join_assoc with (H:=gworld_sa) (a:=w3) (b:=w2) as [w12[??]]; eauto.
      eapply step_cons.
      apply st_par1; apply st_par_join; auto.
      eapply join_com; eauto.
      eapply step_single.
      apply st_par_join; auto.
      apply join_com; auto.
      right; right; right; reflexivity.
    intros.
    inverts H.
    inverts H5.
    rename b1'0 into b1'.
    exists (b1' || w23:tm_skip); split.
      apply step_single; apply st_par1; auto.
      right; left; do 4 eexists; eauto.
    inverts H4.
    rename w into w12.
    exists (w1: tm_skip || w23: tm_skip); split.
      apply step_nil.
      right; right; right.
      eapply bisim_Cs_par_skip_assoc; eauto.
    inverts H5.

    splits.    
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H.
    inverts H5.
    inverts H5.
    inverts H4.
    rename b2'0 into b3'.
    exists (w12:tm_skip || b3'); split.
      apply step_single; apply st_par2; auto.
      right; right; left; do 4 eexists; eauto.
    exists (w12: tm_skip || w3: tm_skip); split.
      apply step_nil.
      right; right; right.
      eapply bisim_Cs_par_skip_assoc; eauto.
    intros.
    inverts H.
    inverts H5.
    rename b2'0 into b3'.
    exists (w1: tm_skip || (w2: tm_skip || b3')); split.
      apply step_single; apply st_par2; apply st_par2; auto.
      right; right; left; do 4 eexists; eauto.
    exists (w: tm_skip); split.
      edestruct @join_assoc with (H:=gworld_sa) (a:=w1) (b:=w2) as [w23[??]]; eauto.
      eapply step_cons.
      apply st_par2; apply st_par_join; eauto.
      eapply step_single.
      apply st_par_join; auto.
      right; right; right; reflexivity.

    splits; intros.
    split; intros.
    apply bisim_Cs_halted1; eauto.
    repeat right; reflexivity.
    split; intros.
    apply bisim_Cs_halted2; eauto.
    repeat right; reflexivity.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.
    exists bb'; split; auto.
    edestruct step_bisim_Cs2 as [ba'[??]]; eauto.
    exists ba'; split; auto.
  Qed.

  Lemma bisim_Cs_tm_par_par_precise_frame: forall w w1 w2 P1 P2 t1 t2 w12 wF,
    gprecise P1 -> gprecise P2 ->
    gjoin w1 w2 w12 ->
    P1 w1 -> P2 w2 ->
    gjoin w12 wF w ->
    bisimilar_Cs (w : tm_par P1 t1 P2 t2) (w : tm_par (gsep_con (exactly w1) TT) t1 (gsep_con (exactly w2) TT) t2).
  Proof.
    intros.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    clear H6.
    inverts H5.
    rename w0 into w1F1; rename w3 into w2F2.
    edestruct gprecise_TT_sub with (P1:=P1) (P2:=P2) as [ [wF1 ?] [wF2 ?] ]; eauto.
    exists (w1F1: t1 || w2F2: t2); split.
      apply step_single. apply st_par; auto.
      setoid_rewrite<- gstar_assoc; setoid_rewrite TT_gsplit; exists w1 wF1; splits; simpl; auto.
      setoid_rewrite<- gstar_assoc; setoid_rewrite TT_gsplit; exists w2 wF2; splits; simpl; auto.
      reflexivity.
    (**)
    intros.
    clear H6.
    inverts H5.
    setoid_rewrite<- gstar_assoc in H13; setoid_rewrite TT_gsplit in H13.
    setoid_rewrite<- gstar_assoc in H14; setoid_rewrite TT_gsplit in H14.
    rename w0 into w1F1; rename w3 into w2F2.
    exists (w1F1: t1 || w2F2: t2); split.
      apply step_single. apply st_par; auto.
      eapply gstar_intro with (x:=exactly w1) (y:=TT); auto. apply derives_exactly; auto.
      eapply gstar_intro with (x:=exactly w2) (y:=TT); auto. apply derives_exactly; auto.
      reflexivity.
  Qed.

  Lemma bisim_Cs_if_true_const: forall n t1 t2,
    n <> 0 ->
    term_bisimilar_Cs (tm_if (nat_to_expr n) t1 t2) t1.
  Proof.
    intros. intro w.
    eapply bisim_Cs_if_true.
    apply evalTrue_const; auto.
  Qed.

  Lemma bisim_Cs_if_false_const: forall t1 t2,
    term_bisimilar_Cs (tm_if (nat_to_expr 0) t1 t2) t2.
  Proof.
    intros. intro w.
    eapply bisim_Cs_if_false.
    apply evalFalse_const.
  Qed.

  Lemma w_env_share_new_local1: forall w1 x v w,
    gjoin w1 (new_local x v) w ->
    w_env_share w x = Share.top.
  Proof.
    intros.
    destruct H as [_[? _]].
    unfold fst,snd in H.
    eapply LocalsMap.join_map_singleton; eauto.
    apply H.
  Qed.

  Lemma bisim_Cs_assign_freelocal: forall w x e v,
    eval e v w ->
    bisimilar_Cs (w: tm_assign x e;; tm_freelocal x) (w: tm_freelocal x).
  Proof.
    introv He.
    apply bisim_Cs_step.
    - introv [?]; discriminate.
    - introv [?]; discriminate.
    - intros.
    inverts H.
    inverts H5.
    assert (v0 = v); try subst. eapply eval_eq; eauto.
    exists (w: tm_freelocal x); split.
    + apply step_nil.
    + setoid_rewrite bisim_Cs_seq_skip1'.
    apply bisim_Cs_step.
    * introv [?]; discriminate.
    * introv [?]; discriminate.
    * intros.
    inverts H.
    exists (w': tm_skip); split.
    step_cons 1.
    destruct w_env_get_not_bot with (w:=w) (x:=x) as [v' ?].
    rewrite H7.
    apply Share.nontrivial.
    apply st_freelocal with v'.
    apply w_env_set_get_same in H.
    rewrite<- H.
    rewrite<- w_env_set_same with (x:=x) (v1:=v) (v2:=v').
    eapply gjoin_new_local_set_value; eauto.
    reflexivity.
    * intros.
    inverts H.
    exists (w': tm_skip); split.
    step_cons 1.
    destruct w_env_get_not_bot with (w:=w) (x:=x) as [v' ?].
    rewrite H7.
    apply Share.nontrivial.
    apply st_freelocal with v.
    apply w_env_set_get_same in H.    
    eapply gjoin_new_local_set_value; eauto.
    reflexivity.
    - intros.
    inverts H.
    exists (w': tm_skip); split.
    step_cons 3.
    apply st_seq1.
    apply st_assign with v; auto.
    eapply w_env_share_new_local1; eauto.
    apply st_seq_skip.
    apply st_freelocal with v.
    eapply gjoin_new_local_set_value; eauto.
    reflexivity.
  Qed.

  Lemma bisim_Cs_assign_assert_evaluable: forall x e,
    term_bisimilar_Cs (tm_assert (evaluable e);;; tm_assign x e) (tm_assign x e).
  Proof.
    intros. intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_step.
    - introv [?]; discriminate.
    - introv [?]; discriminate.
    - intros.
    inverts H.
    inverts H5.
    exists (w: tm_assign x e); split.
    apply step_nil.
    setoid_rewrite bisim_Cs_seq_skip1'.
    reflexivity.
    - intros.
    inverts H.
    exists (w_env_set w x v: tm_skip); split.
    step_cons 3.
    apply st_seq1.
    apply st_assert.
    apply TT_gstar.
    exists v; auto.
    apply st_seq_skip.
    apply st_assign with v; auto.
    reflexivity.
  Qed.

  Lemma bisim_Cs_assign_freelocal': forall x e,
    term_bisimilar_Cs (tm_assign x e;;; tm_freelocal x) (tm_assert (evaluable e);;; tm_freelocal x).
  Proof.
    intros.
    setoid_rewrite<- bisim_Cs_assign_assert_evaluable.
    setoid_rewrite<- bisim_Cs_tm_seq_assoc.
    apply bisim_Cs_assert_intro.
    intros.
    destruct H as [w1[w2[?[??]]]].
    destruct H0 as [v ?].
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_assign_freelocal with v.
    eapply eval_gframe; eauto.
  Qed.

  Lemma bisim_Cs_local_assign: forall f e e',
    term_bisimilar_Cs (local x:= e in (f x;;; tm_assign x e')) (local x:= e in (f x;;; tm_assert (evaluable e'))).
  Proof.
    intros. intro w.
    apply bisim_Cs_step.
    - introv [?]; discriminate.
    - introv [?]; discriminate.
    - intros.
    inverts H.
    exists (w': f x;; tm_assert (evaluable e');; tm_freelocal x); split.
    step_cons 3.
    eapply st_local; eauto.
    apply st_seq. apply st_seq1. apply st_seq.
    repeat setoid_rewrite bisim_Cs_tm_seq_seq.
    setoid_rewrite<- bisim_Cs_seq_assoc.
    apply bisim_Cs_seq2. intros.
    apply bisim_Cs_assign_freelocal'.
    - intros.
    inverts H.
    exists (w': f x;; tm_assign x e';; tm_freelocal x); split.
    step_cons 3.
    eapply st_local; eauto.
    apply st_seq. apply st_seq1. apply st_seq.
    repeat setoid_rewrite bisim_Cs_tm_seq_seq.
    setoid_rewrite<- bisim_Cs_seq_assoc.
    apply bisim_Cs_seq2. intros.
    apply bisim_Cs_assign_freelocal'.
  Qed.


  Lemma bisim_Cs_par_freevar: forall f w w' x v,
    (forall y, f x = f y) ->
    gjoin w (new_local x v) w' ->
    bisimilar_Cs (w: f x || new_local x v: tm_skip) (w': f x).
Admitted.

  Lemma bisim_Cs_freelocal_freevar: forall f w w' x v,
    (forall y, f x = f y) ->
    gjoin w (new_local x v) w' ->
    bisimilar_Cs (w': f x;; tm_freelocal x) (w: f x).
  Proof.
Admitted.

  Lemma bisim_Cs_local_freevar: forall w e v f x,
    eval e v w ->
    (forall y, f x = f y) ->
    bisimilar_Cs (w: local x:=e in f x) (w: f x).
  Proof.
    intros.
    apply bisim_Cs_step.
    - introv [?]; discriminate.
    - introv [w_ ?].
    inverts H1; rename w_ into w.
    rewrite H4 in *. clear H4.
    remember (fresh_var w v) as fresh_y.
    step_cons' 2.
    eapply st_local with (x:=FreshLocals.fl_x fresh_y) (w':=FreshLocals.fl_w fresh_y); eauto.
    apply fresh_var_local_props; auto.
    apply st_seq.
    eapply step_star_app with (b':= FreshLocals.fl_w fresh_y: tm_skip;; tm_freelocal (FreshLocals.fl_x fresh_y)).
    apply step_star_seq_app.
    eapply bisim_Cs_halted2.
    eexists; eauto.
    symmetry.
    erewrite H0; reflexivity.
    step_cons 2.
    apply st_seq_skip.
    eapply st_freelocal.
    apply fresh_var_local_props; eauto.
    - intros.
    inverts H1. rename x0 into y.
    exists (w: f x); split.
    apply step_nil.
    rewrite (H0 y).
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_freelocal_freevar with v0; auto.
    intros.
    rewrite<- H0.
    apply H0.
    - intros.
    remember (fresh_var w v) as fresh_y.
    rewrite H0 with (FreshLocals.fl_x fresh_y) in H1.
    edestruct fresh_var_local_props with (fresh_x:= fresh_y); eauto.
    edestruct step_bisim_Cs2 as [ba'[??]]. eauto.
    apply bisim_Cs_freelocal_freevar with v; eauto.
    intros; congruence.
    exists ba'; split; auto.
    step_cons' 2; eauto.
    eapply st_local; eauto.
    apply st_seq.
  Qed.

  Lemma bisim_Cs_local_assert_evaluable: forall e f,
    term_bisimilar_Cs (tm_assert (evaluable e);;; tm_local e f) (tm_local e f).
  Proof.
    intros. intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_step.
    - introv [?]; discriminate.
    - introv [?]; discriminate.
    - intros.
    inverts H. inverts H5.
    exists (w: tm_local e f); split.
    apply step_nil.
    apply bisim_Cs_seq_skip1'.
    - intros.
    inverts H.
    exists (w': f x); split.
    step_cons 3.
    apply st_seq1. apply st_assert. apply TT_gstar. exists v; auto.
    apply st_seq_skip.
    apply st_local with v; auto.
    reflexivity.
  Qed.

  Lemma bisim_Cs_local_freevar': forall e f x,
    (forall y, f x = f y) ->
    term_bisimilar_Cs (local x:= e in f x) (tm_assert (evaluable e);;; f x).
  Proof.
    intros. intro w.
    setoid_rewrite<- bisim_Cs_local_assert_evaluable.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_assert_intro'. intros.
    apply evaluable_gstar in H0.
    destruct H0 as [v ?].
    apply bisim_Cs_local_freevar with v; auto.
  Qed.

(*
NOT QUITE TRUE: the allocation of fresh variables needs to be deterministic for this to be true
  Lemma bisim_Cs_local_split: forall e1 e2 f1 f2,
    term_bisimilar_Cs ((local x:=e1 in f1 x);;; (local y:=e2 in f2 y)) (local x:=e1 in (f1 x;;; x::= e2;;; f2 x)).
  Proof.
    intros; intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H. inverts H5.
    exists (w': f1 x;; x::= e2;; f2 x;; tm_freelocal x); split.
    step_cons 3.
    apply st_local with v; eauto.
    apply st_seq1. apply st_seq.
    do 2 apply st_seq1. apply st_seq.
    do 2 setoid_rewrite<- bisim_Cs_seq_assoc.
    apply bisim_Cs_seq2; intros.
    setoid_rewrite bisim_Cs_tm_seq_assoc.
    repeat setoid_rewrite bisim_Cs_tm_seq_seq.
  Lemma bisim_Cs_freelocal_local: forall w x e f,
    w_env_get w x <> None ->
    bisimilar_Cs (w: tm_freelocal x;; tm_local e f) (w: tm_local e f).
Admitted.
    clear w e1 H0 H6 H7 H.
    case_eq (w_env_get w'0 x); intros.
    setoid_rewrite bisim_Cs_freelocal_local.
    2: congruence.

    apply bisim_Cs_tm_step'.
 
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H. inverts H5.
    exists (w': f1 x;; x::= e2;; f2 x;; tm_freelocal x); split.
    step_cons 3.
    apply st_local with v; eauto.
    apply st_seq1. apply st_seq.
    do 2 apply st_seq1. apply st_seq.
    do 2 setoid_rewrite<- bisim_Cs_seq_assoc.
    apply bisim_Cs_seq2; intros.
    setoid_rewrite bisim_Cs_tm_seq_assoc.
    repeat setoid_rewrite bisim_Cs_tm_seq_seq.
*)

End SimpleBisimulations.

(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)




Section Determinism.
  Import Machine.Notation.
  Open Scope wterm_scope.

  Lemma bisim_Cs_semantic_equivalence: forall ba bb,
    (forall w', step_star_Cs ba (w':tm_skip) <-> step_star_Cs bb (w':tm_skip)) ->
    ignores_global_state ba ->
    ignores_global_state bb ->
    deterministic_Cs_wterm ba ->
    deterministic_Cs_wterm bb ->
    bisimilar_Cs ba bb.
  Proof.
    introv Hequiv Hresp1 Hresp2 Hdet1 Hdet2.
    exists (fun ba bb =>
      (forall w', step_star_Cs ba (w':tm_skip) <-> step_star_Cs bb (w':tm_skip)) /\
      ignores_global_state ba /\ ignores_global_state bb /\
      deterministic_Cs_wterm ba /\ deterministic_Cs_wterm bb).
    split.
    eapply bisimulation_Cs_symm_equiv.
    clear ba bb Hequiv Hresp1 Hresp2 Hdet1 Hdet2.
    intros ba bb [Hequiv[Hresp1[Hresp2[Hdet1 Hdet2]]]].
    splits; eauto.

    introv [w' ?]. subst.
    destruct (Hequiv w') as [? _].
    spec H. apply step_star_Cs_nil.
    splits; intros; auto.
    eapply step_star_Cs_ignores_global_state; eauto.
    reflexivity.

    intros.
    apply Hresp1 in H. destruct H; subst.
    assert (ignores_global_state b1').
      repeat intro. eapply Hresp1; eauto.
      eapply step_star_Cs_app; eauto. eapply step_single_Cs with (s:=s); eauto.
    eexists; split.
    apply step_nil.
    splits; eauto.
    split; intros.
    apply Hequiv. eapply step_star_Cs_app; eauto. eapply step_single_Cs with (s:=s); eauto.
    apply Hequiv in H1.
    apply step_star_Cs_ignores_global_state with (s:=s) in H1; auto.
    eapply step_star_Cs_cons.
    apply step_star_simulates_cstep_star_empty_context.
    eapply bisim_Cs_halted_converge1; eauto.
    eapply Hdet1. eapply step_single_Cs with (s:=s); eauto.
    eexists; eauto.
    apply step_star_Cs_nil.
    repeat intro.
    etransitivity.
    symmetry.
    eapply Hdet1.
    eapply step_single_Cs with (s:=s); eauto.
    eapply Hdet1.
    eapply step_star_Cs_app; eauto.
    eapply step_single_Cs with (s:=s); eauto.
    apply step_star_Cs_nil.

    intro; symmetry; eapply Hequiv.

    splits; auto.
  Qed.

  Lemma deterministic_Cs_wterm_seq: forall b1 t2,
    deterministic_Cs_wterm b1 ->
    (forall w', step_star_Cs b1 (w':tm_skip) -> deterministic_Cs_wterm (w':t2)) ->
    deterministic_Cs_wterm (b1;; t2).
  Proof.
    intros.
    intros b' ?.
    edestruct step_star_Cs_seq_inversion as [ [w'[??]] | [b1''[??]] ]; eauto; clear H1.

    spec H0 w' H2.
    apply H in H2.
    setoid_rewrite H2.
    setoid_rewrite bisim_Cs_seq_skip1'.
    apply H0; auto.

    subst b'.
    apply H in H2.
    setoid_rewrite H2.
    reflexivity.
  Qed.

  Lemma det_simple: forall b,
    simple_step_wterm b ->
    state_independent_step_wterm b ->
    (forall s b'1 b'2, step s b s b'1 -> step s b s b'2 -> b'2=b'1) ->
    deterministic_Cs_wterm b.
  Proof.
    introv Hss Hind Hdet.
    repeat intro.
    edestruct step_star_Cs_step_inversion as [ [s0[s0'[b'0[?[??]]]]] | ?]; eauto.
    edestruct Hss as [?[?|[w' ?]]]; eauto; subst s0' b'0.
    false.
    apply step_star_Cs_skip_resolve in H1. subst b'.
    eapply bisim_Cs_tao_step.
    intros; auto. eapply Hind; eauto.
    intros.
    edestruct Hss with (s:=s) as [?[?|[w'' ?]]]; eauto; subst s'' b''.
    split; auto.
    eapply Hdet; eauto.
    split; auto.
    eapply Hdet; eauto.
    subst.
    reflexivity.
  Qed.

  Lemma det_tm_assign: forall w x e,
    deterministic_Cs_wterm (w: tm_assign x e).
  Proof.
    intros.
    apply det_simple.
    repeat intro; inverts H; eauto.
    repeat intro; inverts H; econstructor; eauto.
    intros.
    inverts H; inverts H0.
    do 2 f_equal.
    eapply eval_eq; eauto.
  Qed.

  Lemma det_tm_load: forall w x e,
    deterministic_Cs_wterm (w: tm_load x e).
  Proof.
    intros.
    apply det_simple.
    repeat intro; inverts H; eauto.
    repeat intro; inverts H; econstructor; eauto.
    intros.
    inverts H; inverts H0.
    do 2 f_equal.
    assert (addr_value addr0 = addr_value addr).
      eapply eval_eq; eauto.
    inverts H.
    rewrite H7 in H9; inverts H9; auto.
  Qed.

  Lemma det_tm_store: forall w e1 e2,
    deterministic_Cs_wterm (w: tm_store e1 e2).
  Proof.
    intros.
    apply det_simple.
    repeat intro; inverts H; eauto.
    repeat intro; inverts H; econstructor; eauto.
    intros.
    inverts H; inverts H0.
    assert (addr_value dest_addr0 = addr_value dest_addr).
      eapply eval_eq; eauto.
    inverts H.
    do 2 f_equal.
    eapply eval_eq; eauto.
  Qed.

  (*
    Semantic equivalence of commuting instructions
  *)
  Lemma semantic_equiv_comm: forall w t1 t2 t1'  t2' w'',
    (forall w, simple_step_wterm (w: t1)) ->
    (forall w, simple_step_wterm (w: t1')) ->
    (forall w, simple_step_wterm (w: t2)) ->
    (forall w, simple_step_wterm (w: t2')) ->
    (forall w, state_independent_step_wterm (w: t1')) ->
    (forall w, state_independent_step_wterm (w: t2)) ->
    t1 <> tm_skip -> t1' <> tm_skip ->
    t2 <> tm_skip -> t2' <> tm_skip ->
    (forall s w w'1 w'',
      step s (w: t1) s (w'1: tm_skip) ->
      step s (w'1: t2) s (w'': tm_skip) ->
      exists w'2, step s (w: t2') s (w'2: tm_skip) /\ step s (w'2: t1') s (w'': tm_skip)) ->
    (forall s w w'2 w'',
      step s (w: t2') s (w'2: tm_skip) ->
      step s (w'2: t1') s (w'': tm_skip) ->
      exists w'1, step s (w: t1) s (w'1: tm_skip) /\ step s (w'1: t2) s (w'': tm_skip)) ->
    (step_star_Cs (w: t1;; t2) (w'': tm_skip) <-> step_star_Cs (w: t2';; t1') (w'': tm_skip)).
  Proof.
    introv Hss1 Hss1' Hss2 Hss2' Hind1' Hind2.
    introv Hskip1 Hskip1' Hskip2 Hskip2' Hcom1 Hcom2.
    split; intros.

    edestruct step_star_Cs_step_inversion as [ [s0[s'0[b'1[?[??]]]]] | ?]; eauto.
    clear H. inverts H0.
    edestruct (Hss1 w) as [?[?|[w'1 ?]]]; eauto; subst s'0 b1'.
    false.
    edestruct step_star_Cs_step_inversion as [ [s1[s'1[b'2[?[??]]]]] | ?]; eauto.
    clear H1.
    inverts H.
    false; inverts H8.
    edestruct step_star_Cs_step_inversion as [ [s2[s'2[b'3[?[??]]]]] | ?]; eauto.
    clear H0.
    edestruct (Hss2 w'1) as [?[?|[w''2 ?]]]; eauto; subst s'2 b'3.
    false.
    eapply step_star_Cs_skip_resolve in H1.
    inversion H1; subst w''2. clear H1.
    edestruct Hcom1 as [w'2[??]]; eauto. eapply Hind2; eauto.
    eapply step_star_Cs_step_star.
    step_cons 3. apply st_seq1; eauto. apply st_seq_skip. eauto.
    (**)
    inverts H.
    false.
    false.
    false.
    false.

    edestruct step_star_Cs_step_inversion as [ [s0[s'0[b'1[?[??]]]]] | ?]; eauto.
    clear H. inverts H0.
    edestruct (Hss2' w) as [?[?|[w'2 ?]]]; eauto; subst s'0 b1'.
    false.
    edestruct step_star_Cs_step_inversion as [ [s1[s'1[b'2[?[??]]]]] | ?]; eauto.
    clear H1.
    inverts H.
    false; inverts H8.
    edestruct step_star_Cs_step_inversion as [ [s2[s'2[b'3[?[??]]]]] | ?]; eauto.
    clear H0.
    edestruct (Hss1' w'2) as [?[?|[w''1 ?]]]; eauto; subst s'2 b'3.
    false.
    eapply step_star_Cs_skip_resolve in H1.
    inversion H1; subst w''1. clear H1.
    edestruct Hcom2 as [w'1[??]]; eauto. eapply Hind1'; eauto.
    eapply step_star_Cs_step_star.
    step_cons 3. apply st_seq1; eauto. apply st_seq_skip. eauto.
    (**)
    inverts H.
    false.
    false.
    false.
    false.
  Qed.


  Lemma bisim_Cs_assign_load_commute: forall (x i:ident) e1 e2,
    x <> i ->
    not_free_in_expr x e1 ->
    term_bisimilar_Cs (tm_assign i e1;;; tm_load x e2) (tm_load x (expr_subst e2 i e1);;; tm_assign i e1).
  Proof.
    intros.
    intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_semantic_equivalence.

    intros.
    eapply semantic_equiv_comm.
    repeat intro. inverts H1; eauto.
    repeat intro. inverts H1; eauto.
    repeat intro. inverts H1; eauto.
    repeat intro. inverts H1; eauto.
    repeat intro. inverts H1; eapply st_assign; eauto.
    repeat intro. inverts H1; eapply st_load; eauto.
    intro; false.
    intro; false.
    intro; false.
    intro; false.
    (**)
    intros.
    inverts H1.
    inverts H2.
    eexists; split.
    eapply st_load; eauto.
    eapply eval_expr_subst1; eauto.
    rewrite w_heap_get_env_set in H11; eauto.
    eapply st_assign; eauto.
    apply eval_not_free_in_expr1; eauto.
    rewrite w_env_share_set; auto.
    rewrite w_env_set_diff; auto.
    (**)
    intros. inverts H1. inverts H2.
    eexists; split.
    eapply st_assign; eauto.
    eapply eval_not_free_in_expr2; eauto.
    rewrite w_env_share_set in H11; auto.
    eapply st_load; eauto.
    eapply eval_expr_subst2; eauto.
    eapply eval_not_free_in_expr2; eauto.
    rewrite w_env_share_set in H11; auto.
    rewrite w_heap_get_env_set; eauto.
    rewrite w_env_set_diff; eauto.

    eapply ignores_global_state_seq; intros.
    apply ignores_global_state_tm_assign.
    apply ignores_global_state_tm_load.

    eapply ignores_global_state_seq; intros.
    apply ignores_global_state_tm_load.
    apply ignores_global_state_tm_assign.

    apply deterministic_Cs_wterm_seq.
    apply det_tm_assign.
    intros. apply det_tm_load.

    apply deterministic_Cs_wterm_seq.
    apply det_tm_load.
    intros. apply det_tm_assign.
  Qed.


  Lemma bisim_Cs_assign_store_commute: forall (i:ident) e1 e2 e3,
    term_bisimilar_Cs (tm_assign i e1;;; tm_store e2 e3) (tm_store (expr_subst e2 i e1) (expr_subst e3 i e1);;; tm_assign i e1).
  Proof.
    intros.
    intro w.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_semantic_equivalence.

    intros.
    eapply semantic_equiv_comm.
    repeat intro. inverts H; eauto.
    repeat intro. inverts H; eauto.
    repeat intro. inverts H; eauto.
    repeat intro. inverts H; eauto.
    repeat intro. inverts H; eapply st_assign; eauto.
    repeat intro. inverts H; eapply st_store; eauto.
    intro; false.
    intro; false.
    intro; false.
    intro; false.
    (**)
    intros.
    inverts H.
    inverts H0.
    eexists; split.
    eapply st_store; eauto.
    eapply eval_expr_subst1; eauto.
    eapply eval_expr_subst1; eauto.
    rewrite w_heap_share_env_set in H10; eauto.
    eapply st_assign; eauto.
    apply eval_w_heap_set; eauto.
    rewrite w_env_share_w_heap_set; auto.
    apply w_heap_set_w_env_set.
    (**)
    intros. inverts H. inverts H0.
    eexists; split.
    eapply st_assign; eauto.
    eapply eval_w_heap_set; eauto.
    rewrite w_env_share_w_heap_set in H10; auto.
    eapply st_store; eauto.
    eapply eval_expr_subst2; eauto.
    eapply eval_w_heap_set; eauto.
    rewrite w_env_share_w_heap_set in H10; auto.
    eapply eval_expr_subst2; eauto.
    eapply eval_w_heap_set; eauto.
    rewrite w_env_share_w_heap_set in H10; auto.
    rewrite w_heap_share_env_set; eauto.
    symmetry; apply w_heap_set_w_env_set.

    eapply ignores_global_state_seq; intros.
    apply ignores_global_state_tm_assign.
    apply ignores_global_state_tm_store.

    eapply ignores_global_state_seq; intros.
    apply ignores_global_state_tm_store.
    apply ignores_global_state_tm_assign.

    apply deterministic_Cs_wterm_seq.
    apply det_tm_assign.
    intros. apply det_tm_store.

    apply deterministic_Cs_wterm_seq.
    apply det_tm_store.
    intros. apply det_tm_assign.
  Qed.


  Lemma bisim_Cs_incr_assign_commute: forall (x y i:ident) (n: nat),
    y <> i ->
    x <> i ->
    term_bisimilar_Cs
      (tm_assign i (expr_plus i 1);;; tm_load y (expr_plus x i);;; tm_store (expr_plus x i) (expr_plus y n))
      (tm_load y (expr_plus x (expr_plus i 1));;; tm_store (expr_plus x  (expr_plus i 1)) (expr_plus y n);;; tm_assign i (expr_plus i 1)).
  Proof.
    intros.
    setoid_rewrite bisim_Cs_assign_load_commute; auto.
    setoid_rewrite<- bisim_Cs_tm_seq_assoc.
    setoid_rewrite bisim_Cs_assign_store_commute.
    intro w.
    setoid_rewrite bisim_Cs_tm_seq_assoc.
    repeat setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_seq1.
    apply bisim_Cs_seq; intros.

    setoid_rewrite expr_subst_plus.
    setoid_rewrite expr_subst_id.
    setoid_rewrite expr_subst_not_free.
    reflexivity.
    apply not_free_in_expr_var; auto.

    setoid_rewrite expr_subst_plus.
    setoid_rewrite expr_subst_id.
    setoid_rewrite expr_subst_not_free.
    reflexivity.
    apply not_free_in_expr_var; auto.
    apply not_free_in_expr_var; auto.
    apply not_free_in_expr_nat.

    apply not_free_in_expr_expr_plus.
    apply not_free_in_expr_var; auto.
    apply not_free_in_expr_nat.
  Qed.


End Determinism.








Module TestMorphisms.
  Import Morphisms.
  Import Machine.Notation.
  Open Scope wterm_scope.
  Open Scope term_scope.

  Lemma test_local: forall e1 e2 a,
    e1 == e2 ->
    (local x := e1 in tm_skip;;; tm_load x a) == local y:= e2 in tm_load y a.
  Proof.
    intros.
    setoid_rewrite H.
    setoid_rewrite bisim_Cs_tm_seq_skip1'.
    reflexivity.
  Qed.

End TestMorphisms.



















