Require Import Machine.
Require Import StepStar.
Require Import IterationBounds.
Require Import world_model.
Require Import msl.msl_standard.
Require Import SfLib.
Require Import LibTactics.
(*Import Machine.PMapFacts.*)

(*Require Import Omega.*)

Section Definitions.
  Import Machine.Notation.
  Import world_model.Notation.
  Open Scope forces_pred.

Inductive context : Type :=
| cx_hole: context
| cx_seq: context->term->context
| cx_par1: context->wterm->context
| cx_par2: wterm->context->context
.

Fixpoint set_context C (b:wterm) := match C with
| cx_hole => b
| cx_seq C1 t2 => wtm_seq (set_context C1 b) t2
| cx_par1 C1 b2 => wtm_par (set_context C1 b) b2
| cx_par2 b1 C2 => wtm_par b1 (set_context C2 b)
end.
  Arguments Scope set_context [_ wterm_scope].
(*  Coercion set_context : context >-> Funclass. *)

  Delimit Scope context_scope with context.
  Notation "C [ b ]" := (set_context C b%wterm)
    (no associativity, at level 9, format "C [ b ]") : context_scope.
 Open Scope context_scope.


Inductive cstep (s:state) : context->wterm->state->context->wterm->Prop :=
| cst_seq: forall C s' C' b1 t2,
    cstep s C b1 s' C' b1 ->
    cstep s (cx_seq C t2) b1 s' (cx_seq C' t2) b1
| cst_par1: forall C s' C' b1 b2,
    cstep s C b1 s' C' b1 ->
    cstep s (cx_par1 C b2) b1 s' (cx_par1 C' b2) b1
| cst_par2: forall C s' C' b1 b2,
    cstep s C b2 s' C' b2 ->
    cstep s (cx_par2 b1 C) b2 s' (cx_par2 b1 C') b2
| cst_par1_step2: forall C s' b1 b2 b2',
    step s b2 s' b2' ->
    cstep s (cx_par1 C b2) b1 s' (cx_par1 C b2') b1
| cst_par2_step1: forall C s' b1 b2 b1',
    step s b1 s' b1' ->
    cstep s (cx_par2 b1 C) b2 s' (cx_par2 b1' C) b2
| cst_hole: forall C b s' b',
    step s b s' b' ->
    cstep s C b s' C b'
.

Inductive cstep_star s C b: state->context->wterm->Prop :=
| cstep_nil: 
    cstep_star s C b s C b
| cstep_cons: forall s' b' C' s'' C'' b'',
    cstep s C b s' C' b' ->
    cstep_star s' C' b' s'' C'' b'' ->
    cstep_star s C b s'' C'' b''
| cstep_iteration: forall s' s'' C'' b'',
    s_iter s s' ->
    cstep_star s' C b s'' C'' b'' ->
    cstep_star s C b s'' C'' b''
.

Inductive cstep_star_sans_s_iter s C b: state->context->wterm->Prop :=
| cstep_nil_si: 
    cstep_star_sans_s_iter s C b s C b
| cstep_cons_si: forall s' b' C' s'' C'' b'',
    cstep s C b s' C' b' ->
    cstep_star_sans_s_iter s' C' b' s'' C'' b'' ->
    cstep_star_sans_s_iter s C b s'' C'' b''
.

  Definition halted_assert s C (b:wterm) (P: pred world) := forall s' C' w',
    cstep_star s C b s' C' (w':tm_skip) -> P w'.

  Definition frame_halted_assert P t Q :=
    forall s C w F, w|= gsep_con P F -> halted_assert s C (w: t) (gsep_con Q F).

Fixpoint compose_contexts C1 C2 := match C1 with
| cx_hole => C2
| cx_seq C' t2 => cx_seq (compose_contexts C' C2) t2
| cx_par1 C' b2 => cx_par1 (compose_contexts C' C2) b2
| cx_par2 b1 C' => cx_par2 b1 (compose_contexts C' C2)
end.

Lemma context_compose_correct: forall C1 C2 b,
  C1[C2[b]] = (compose_contexts C1 C2)[b].
Proof.
  intros.
  induction C1; simpl; auto.
  rewrite IHC1; auto.
  rewrite IHC1; auto.
  rewrite IHC1; auto.
Qed.

Lemma context_compose_par1: forall C b1 b2,
  C[b1 || b2] = (compose_contexts C (cx_par1 cx_hole b2))[b1].
Proof.
  intros.
  replace (b1 || b2)%wterm with ((cx_par1 cx_hole b2)[b1]).
  apply context_compose_correct.
  simpl; auto.
Qed.

Lemma context_compose_par2: forall C b1 b2,
  C[b1 || b2] = (compose_contexts C (cx_par2 b1 cx_hole))[b2].
Proof.
  intros.
  replace (b1 || b2)%wterm with ((cx_par2 b1 cx_hole)[b2]).
  apply context_compose_correct.
  simpl; auto.
Qed.


End Definitions.

 Tactic Notation "cstep_cases" tactic(first) ident(c) :=
  first;
  [ Case_aux c "cst_seq" | Case_aux c "cst_par1"| Case_aux c "cst_par2" |
   Case_aux c "cst_par1_step2" | Case_aux c "cst_par2_step1" |
   Case_aux c "cst_hole" ].

Tactic Notation "cstep_cases" tactic(first) ident(c) ident(H) :=
  first;
  [ Case_aux c "cst_seq"; pose cst_seq as H  |
     Case_aux c "cst_par1"; pose cst_par1 as H | Case_aux c "cst_par2"; pose cst_par2 as H |
   Case_aux c "cst_par1_step2"; pose cst_par1_step2 as H | Case_aux c "cst_par2_step1"; pose cst_par2_step1 as H |
   Case_aux c "cst_hole"; pose cst_hole as H 
 ].

  Arguments Scope set_context [_ wterm_scope].
  Delimit Scope context_scope with context.

Module Notation.
  Export Machine.Notation.
  Notation "C [ b ]" := (set_context C b%wterm)
    (no associativity, at level 9, format "C [ b ]") : context_scope.
End Notation.


(*
Note: cstep used to be defined as:
  Inductive cstep (s:state) : context->term->state->context->term->Prop :=
  | cst_context: forall C t s' C',
      step s C[tm_stuck] s' C'[tm_stuck] ->
      cstep s C t s' C' t
  | cst_hole: forall C t s' t',
      step s C[t] s' C[t'] ->
      cstep s C t s' C t'
  .

But this is no good because par_cstep_star_split in not provable under this
definition. Counter example: let C = while true do skip || []
Thus C can step to C'= skip; while true do skip || [], then
again step to C. In parallel, t steps to t''.

It is not always the case the context will always be C as t
steps to t''.


Lemma par_cstep_star_split: forall s C t s'' C'' t'',
  cstep_star s C t s'' C'' t'' ->
  exists s', exists C', exists t',
    cstep_star s C t s' C' t' /\ cstep_star s' C' t' s'' C'' t''

But with the new def, it's provable!
*)



(*
Notation "s1 : t1 --> s2 : t2" := (step s1 t1 s2 t2) (no associativity, at level 30, t1 at level 0, s2 at level 0, t2 at level 0).
Notation "s1 : t1 -->* s2 : t2" := (step_star s1 t1 s2 t2) (no associativity, at level 30, t1 at level 0, s2 at level 0, t2 at level 0).

Notation "s1 : C1 [ t1 ] -->> s2 : C2 [ t2 ]" := (cstep s1 C1 t1 s2 C2 t2) (no associativity, at level 30, C1 at level 0, t1 at level 0, s2 at level 0, C2 at level 0, t2 at level 0).
Notation "s1 : C1 [ t1 ] -->>* s2 : C2 [ t2 ]" := (cstep_star s1 C1 t1 s2 C2 t2) (no associativity, at level 30, C1 at level 0, t1 at level 0, s2 at level 0, C2 at level 0, t2 at level 0).
*)
(*
Notation "s1 : C1 [ t1 ] -->>* s2 : C2 [ t2 ] -->>* s3 : C3 [ t3 ]" :=
  (cstep_star s1 C1 t1 s2 C2 t2 /\ cstep_star s2 C2 t2 s3 C3 t3) (no associativity, at level 30, C1 at level 0, t1 at level 0, s2 at level 0, C2 at level 0, t2 at level 0, s3 at level 0, C3 at level 0, t3 at level 0).
Notation "s1 : C1 [ t1 ] -->>* s2 : C2 [ t2 ] -->> s3 : C3 [ t3 ]" :=
  (cstep_star s1 C1 t1 s2 C2 t2 /\ cstep s2 C2 t2 s3 C3 t3) (no associativity, at level 30, C1 at level 0, t1 at level 0, s2 at level 0, C2 at level 0, t2 at level 0, s3 at level 0, C3 at level 0, t3 at level 0).
Notation "s1 : C1 [ t1 ] -->> s2 : C2 [ t2 ] -->>* s3 : C3 [ t3 ]" :=
  (cstep s1 C1 t1 s2 C2 t2 /\ cstep_star s2 C2 t2 s3 C3 t3) (no associativity, at level 30, C1 at level 0, t1 at level 0, s2 at level 0, C2 at level 0, t2 at level 0, s3 at level 0, C3 at level 0, t3 at level 0).
*)

Section Lemmas.
  Import Notation.
  Import Machine.Notation.
  Open Scope context_scope.
  Open Scope wterm_scope.


Lemma cstep_inversion: forall s C t s' C' t',
  cstep s C t s' C' t' -> ((step s t s' t' /\ C'=C) \/ (cstep s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) /\ t'=t)).
Proof.
  intros.
  induction H;
    try (destruct IHcstep as [[? ?] | [? ?]]; subst; simpl; auto);
    try (right; split; auto; econstructor; eauto).

  induction C; simpl in *; auto;
    inversion H; try subst;
    try solve [
      edestruct IHC as [[? ?] | [? ?]]; subst; auto;
      right; split; auto;
      try solve [
        econstructor; eauto
      | apply st_par1; auto
      | apply st_par2; auto
      ]
    |  (replace t' with t in *; [| eapply context_equal_terms; eauto]);
          right; split; auto;
          apply cstep_simulates_step;
          eapply st_iteration; eauto
    ].
Qed.

Lemma context_equal_terms: forall (C:context) t t',
  C[t] = C[t'] -> t=t'.
Proof.
  intros.
  induction C; simpl in H; auto; try inversion H; eauto.
Qed.

Lemma cstep_simulates_step: forall s C b s' b',
  step s C[b] s' C[b'] -> cstep s C b s' C b'.
Proof.
  induction C; simpl; intros.
  - constructor; auto.
  - inverts H. apply IHC in H3.
  edestruct cstep_inversion as [[?]|[?]]; eauto.
  apply cst_hole; auto.
  subst b'; apply cst_seq; auto.
  - inverts H. apply IHC in H3.
  edestruct cstep_inversion as [[?]|[?]]; eauto.
  apply cst_hole; auto.
  subst b'; apply cst_par1; auto.
  apply context_equal_terms in H4; subst b'.
  apply cst_par1_step2; auto.
  - inverts H.
  apply context_equal_terms in H5; subst b'.
  apply cst_par2_step1; auto.
  apply IHC in H3.
  edestruct cstep_inversion as [[?]|[?]]; eauto.
  apply cst_hole; auto.
  subst b'; apply cst_par2; auto.
Qed.

(*
Lemma step_steps_sterm_or_nop: forall s t s' t',
  step s t s' t' ->
  (exists C, exists w, exists ts:sterm, exists b', C[w:ts]=t /\ C[b'] = t' /\ step s (w:ts) s' b')
  \/ (s'=s /\ forall s, step s t s t').
Proof.
  intros.
  induction H;
    try solve
      [ right; split; auto; intros; constructor; auto
      | left; exists cx_hole; simpl; do 3 eexists; do 2 split; eauto; econstructor; eauto
      ].

  left; exists cx_hole; simpl; do 3 eexists; do 2 split; eauto; apply st_while_max_limit; auto.



  destruct IHstep as [[C [w [ts [b' [? [? ?]]]]]] | [? ?]]; subst.
  left; exists (cx_seq C t2); simpl; do 3 eexists; do 2 split; eauto; constructor; auto.
  right; split; auto; intros; constructor; auto.

  destruct IHstep as [[C [w [ts [b' [? [? ?]]]]]] | [? ?]]; subst.
  left; exists (cx_par1 C b2); simpl; do 3 eexists; do 2 split; eauto; constructor; auto.
  right; split; auto; intros; constructor; auto.

  destruct IHstep as [[C [w [ts [b' [? [? ?]]]]]] | [? ?]]; subst.
  left; exists (cx_par2 b1 C); simpl; do 3 eexists; do 2 split; eauto; constructor; auto.
  right; split; auto; intros; constructor; auto.
Qed.
*)

  Lemma cstep_commute_front_s_iter: forall s C b s' C' b' s'',
    cstep s C b s' C' b' -> s_iter s' s'' ->
    exists s0, s_iter s s0 /\ cstep s0 C b s'' C' b'.
  Proof.
    intros.
    cstep_cases (induction H) Case HH;
      try solve [edestruct IHcstep as [s0[??]]; eauto | edestruct step_commute_front_s_iter as [s0[??]]; eauto].
  Qed.

  Lemma cstep_commute_front_s_iter_star: forall s C b s' C' b' s'',
    cstep s C b s' C' b' -> s_iter_star s' s'' ->
    exists s0, s_iter_star s s0 /\ cstep s0 C b s'' C' b'.
  Proof.
    intros.
    revert s b b' H.
    induction H0; intros.
    eexists; split; eauto; apply s_iter_nil.
    edestruct cstep_commute_front_s_iter as [s0'[??]]; eauto.
    edestruct IHs_iter_star as [s0''[??]]; eauto.
    eexists; split; eauto.
    eapply s_iter_cons; eauto.
  Qed.

  Lemma cstep_star_commute_front_s_iter_star: forall s C b s' C' b',
    cstep_star s C b s' C' b' ->
    exists s0,
      s_iter_star s s0 /\ cstep_star_sans_s_iter s0 C b s' C' b'.
  Proof.
    intros.
    induction H.
    eexists; split. apply s_iter_nil. apply cstep_nil_si.
    edestruct IHcstep_star as [s0[??]].
    edestruct cstep_commute_front_s_iter_star as [s0'[??]]; eauto.
    eexists; split; eauto.
    eapply cstep_cons_si; eauto.
    edestruct IHcstep_star as [s0[??]].
    eexists; split; eauto.
    eapply s_iter_cons; eauto.
  Qed.


Lemma context_substitution_term_size: forall C t,
  wterm_size C[t] >= wterm_size t.
Proof.
  intros.
  induction C; simpl; intros; auto; omega.
Qed.

Lemma wterm_size_hole: forall C t,
  wterm_size t = wterm_size C[t] -> C = cx_hole.
Proof.
  intros.
  induction C; simpl in *; auto;
    try solve [elimtype False;
    generalize (context_substitution_term_size C t);
    omega].
Qed.  


Lemma context_size_hole: forall C t,
  t = C[t] -> C=cx_hole.
Proof.
  intros.
  apply wterm_size_hole with t.
  rewrite H at 1.
  auto.
Qed.

Lemma context_term_hole: forall C b w t,
  (w:t)%wterm = C[b] -> C=cx_hole /\ b=(w:t)%wterm.
Proof.
  induction C; intros; simpl in *; subst; auto.
  inverts H.
  inverts H.
  inverts H.
Qed.

Lemma step_simulates_cstep: forall s C t s' C' t',
  cstep s C t s' C' t' -> step s C[t] s' C'[t'].
Proof.
  induction C; intros; simpl; inverts H.
  - auto.
  - apply st_seq1; auto.
  - apply st_seq1.
  apply IHC.
  apply cst_hole; auto.
  - apply st_par1; auto.
  - apply st_par2; auto.
  - apply st_par1; auto.
  apply IHC.
  apply cst_hole; auto.
  - apply st_par2; auto.
  - apply st_par1; auto.
  - apply st_par2; auto.
  apply IHC.
  apply cst_hole; auto.
Qed.

Lemma step_simulates_cstep1: forall s C t s' t',
  cstep s C t s' C t' -> step s C[t] s' C[t'].
Proof. intros; apply step_simulates_cstep; auto. Qed.
  
Lemma step_simulates_cstep2: forall C s t s' C',
  cstep s C t s' C' t -> step s C[t] s' C'[t].
Proof. intros; apply step_simulates_cstep; auto. Qed.



Lemma cstep_spin_no_change: forall s b s' C,
  cstep s C b s' C b ->
  s'=s.
Proof.
  intros.
  eapply step_spin_no_change;
  apply step_simulates_cstep; eauto.
Qed.



Lemma step_star_simulates_cstep_star: forall s C t s'' C'' t'',
  cstep_star s C t s'' C'' t'' -> step_star s C[t] s'' C''[t''].
Proof.
  intros.
  induction H.
  apply step_nil.
  apply step_simulates_cstep in H.
  eapply step_cons; eauto.
  eapply st_iteration; eauto.
Qed.

Lemma cstep_star_DOES_NOT_simulate_step_star: forall s,
  exists s'', exists t, exists t', exists C, exists C',
  step_star s C[t] s'' C'[t'] /\ ~cstep_star s C t s'' C' t'.
Proof.
  intros.
  remember (gwemp:tm_skip)%wterm as skip.
  exists s.
  exists skip.
  exists skip.
  exists (cx_par1 cx_hole skip).
  exists (cx_par2 skip cx_hole).
  split.
  simpl; apply step_nil.

  intro.
  remember (cx_par1 cx_hole skip) as t1.
  remember (cx_par2 skip cx_hole) as t2.
  revert Heqt1 Heqt2.
  remember skip as t in H at 1.
  remember skip as t' in H at 1.
  revert Heqt Heqt'.
  induction H; intros; subst; try discriminate.

  inversion H; subst.
  inversion H7; subst; auto.
  inverts H7; eauto.

  apply IHcstep_star; auto.
  simpl in *.
  inverts H1; auto.

  apply IHcstep_star; auto.
Qed.

Lemma step_star_simulates_cstep_star_empty_context: forall s t s'' t'',
   step_star s t s'' t'' -> cstep_star s cx_hole t s'' cx_hole t''.
Proof.
  intros.
  induction H.
  apply cstep_nil.
  eapply cstep_cons; eauto.
  apply cstep_simulates_step; auto.
  eapply cstep_iteration; eauto.
Qed.

Lemma cstep_skip_subst: forall C s s'' C'' w t2,
  cstep s C (w: tm_skip) s'' C'' (w: tm_skip) ->
  cstep s C t2 s'' C'' t2.
Proof.
  induction C; intros.

  inversion H; inverts H0; subst.

  inversion H; subst.
  econstructor; eapply IHC; eauto.
  simpl in *. inverts H0.
  
  inversion H; subst.
  constructor; eapply IHC; eauto.
  simpl in *.
  apply cst_par1_step2; auto.
  inverts H0.  
  
  inversion H; subst.
  constructor; eapply IHC; eauto.
  constructor; auto.
  inverts H0.
Qed.

Lemma step_remove_context: forall s C t s' t',
  step s C[t] s' C[t'] -> (step s t s' t' \/ (step s C[t] s' C[t] /\ t=t')).
(* case: C can step to C in one step: blocked recv *)
Proof.
  intros.

  induction C; auto; simpl in *.

  inverts keep H; auto.
  destruct IHC as [? | [? ?]]; subst; auto.

  inverts keep H; auto.
  destruct IHC as [? | [? ?]]; subst; auto.
  eapply context_equal_terms in H4. subst.
  right; split; auto.

  inverts keep H; auto.
  eapply context_equal_terms in H5. subst.
  right; split; auto.
  destruct IHC as [? | [? ?]]; subst; auto.
Qed.

Lemma step_add_context: forall s t s' t' C,
  step s t s' t' -> step s C[t] s' C[t'].
Proof.
  intros.
  induction C; simpl in *; try constructor; auto.
Qed.



(*
Lemma terminating_run_cstep_inversion: forall s C t s'' C'',
  cstep_star s C t s'' C'' tm_skip ->
  (exists s', exists C', cstep s C tm_stuck s' C' tm_stuck /\ cstep_star s' C' t s'' C'' tm_skip)
  \/ (exists s', exists t', cstep s C t s' C t' /\ cstep_star s' C t' s'' C'' tm_skip /\ t'<>t)
  \/ t=tm_skip.
Proof.
  intros.
  induction H.
  right; right; auto.

  edestruct cstep_inversion as [[? ?] | [? ?]]; eauto.
  subst C'.
  destruct (term_eq_dec t t').
  subst t'.
  replace s' with s in *.
  clear s'.
  apply IHcstep_star.
  apply step_spin_no_change with t; auto.
  right; left.
  exists s'; exists t'; split; auto.
  subst t'.
  left; exists s'; exists C'; split; auto.
Qed.
*)

Lemma step_star_add_context_cstep_star: forall s t s' t' C,
  step_star s t s' t' ->
  cstep_star s C t s' C t'.
Proof.
  intros.
  induction H.
  apply cstep_nil.
  apply cstep_cons with s' b' C; auto.
  apply cstep_simulates_step.
  eapply step_add_context; auto.
  eapply cstep_iteration; eauto.
Qed.


(*
Definition terminates s t:= forall C s' C' t',
  cstep_star s C t s' C' t'  ->
  exists s'', exists w'', step_star s' t' s'' (w'':tm_skip).

Lemma nonblocking_recv_channel_nonempty: forall s w x c,
  nonblocking s (w: tm_recv x c) ->
  s_c s c <> nil.
Proof.
  unfold nonblocking.
  intros.
  destruct H as [[s' [t' [? ?]]] | [w' ?]]; try discriminate.
  inversion H0; subst.
  intro H1; rewrite H1 in H6; discriminate.
  intuition.
Qed.

Lemma step_star_to_skip_nonblocking: forall s t s' w',
  step_star s t s' (w':tm_skip) ->
  nonblocking s t.
Proof.
  intros.
  unfold nonblocking.
  remember (w':tm_skip)%wterm as t'.
  revert w' Heqt'.
  induction H; intros; subst.
  eauto.
  destruct (wterm_eq_dec t t').
  subst t'.
  replace s' with s in *.
  apply IHstep_star with w'; auto.
  apply step_spin_no_change with t; auto.
  left; exists s'; exists t'; split; auto.
Qed.

Lemma terminates_nonblocking: forall s w (t:sterm),
  terminates s (w:t) ->
  nonblocking s (w:t).
Proof.
  unfold terminates, nonblocking; intros.
  destruct (term_eq_dec t tm_skip).
  rewrite e in *; eauto.
  left.
  destruct H with cx_hole s cx_hole (w:t)%wterm as [s'' [w'' ?]].
  apply cstep_nil.
  clear H.
  inversion H0; subst.
  intuition.
  clear H0.
  assert (exists s'0, exists t'0, step s (w:t) s'0 t'0 /\ t'0 <> (w:t)%wterm).
  inversion H; subst; eexists; eexists; split; eauto;
    try solve [
      intro H0; inversion H0; discriminate
    | intro H0; inversion H0; absurd_rec_def
    ].

  assert ((w'':tm_sterm tm_skip)%wterm = (w:tm_recv x c)%wterm).
  apply step_star_blocking_no_resolution with s' s''; auto.
  discriminate.

  destruct H0 as [? [? [? ?]]].
  eauto.
Qed.

Lemma step_terminates_nonblocking: forall s t1 s' t1' t2,
  step s t1 s' t1' ->
  terminates s t2 ->
  nonblocking s' t2.
Proof.
  intros.
  unfold terminates in H0.
  specialize (H0 (cx_par2 t1 cx_hole) s' (cx_par2 t1' cx_hole) t2).
  destruct H0 as [s''' [w'' ?]].
  eapply cstep_cons.
  apply cst_par2_step1; eauto.
  constructor.
  edestruct step_star_to_skip_nonblocking; eauto.
  destruct H1 as [s2' [t2' [? ?]]].
  unfold nonblocking; left; exists s2'; exists t2'; auto.
  unfold nonblocking; auto.
Qed.

*)

Lemma cstep_skip_resolve: forall s C w s' C' t',
  cstep s C (w:tm_skip) s' C' t' ->
  t' = (w:tm_skip)%wterm.
Proof.
  intros.
  inversion H; subst; auto.
  inverts H0.
Qed.

Lemma cstep_star_skip_resolve: forall s C w s' C' t',
  cstep_star s C (w:tm_skip) s' C' t' ->
  t' = (w:tm_skip)%wterm.
Proof.
  intros.
  remember (w:tm_skip)%wterm as t.
  induction H; auto; subst.
  inversion H; subst; auto.
  inverts H1.
Qed.

Lemma cstep_context_or_hole_consistent: forall s C t s' C' t',
  cstep s C t s' C' t' ->
  (C=C' \/ t=t').
Proof.
intros.
induction H; auto.
Qed.


Lemma cstep_star_skip_subst: forall C s w s'' C'' t,
  cstep_star s C (w:tm_skip) s'' C'' (w:tm_skip) ->
  cstep_star s C t s'' C'' t.
Proof.
intros.
remember (w: tm_skip)%wterm as t0 in H.
rewrite Heqt0 in H at 2.
remember tm_skip as t1 in H.
revert Heqt0 Heqt1.
induction H; intros; subst.
apply cstep_nil.
replace b' with (w: tm_skip)%wterm in *.
specialize (IHcstep_star (refl_equal _) (refl_equal _)).
eapply cstep_cons; eauto.
apply cstep_skip_subst with w; eauto.
erewrite cstep_skip_resolve; eauto.
  eapply cstep_iteration; eauto.
Qed.

Lemma cstep_par_resolve: forall s C t1 t2 s' C' t',
  cstep s C (wtm_par t1 t2) s' C' t' ->
     (exists t1', t' = wtm_par t1' t2 /\ C'=C)
  \/ (exists t2', t' = wtm_par t1 t2' /\ C'=C)
  \/ (t' = wtm_par t1 t2)
  \/ ((exists w1, t1=(w1:tm_skip)%wterm) /\ (exists w2, t2=(w2:tm_skip)%wterm) /\ (exists w', t'=(w':tm_skip)%wterm)).
Proof.
  intros.

  edestruct cstep_context_or_hole_consistent; eauto.
  subst C'.
  inversion H; subst;
    try solve [left; eauto].

  inverts H0; subst.
  left; eauto.
  right; eauto.
  do 3 right; split; eauto.
Qed.

Lemma cstep_star_par_resolve: forall s C t1 t2 s' C' t',
  cstep_star s C (wtm_par t1 t2) s' C' t' ->
  (exists t1', exists t2', t' = wtm_par t1' t2')
  \/ (exists w1', exists w2', exists w',
      cstep_star s C (wtm_par t1 t2) s' C' (wtm_par (w1':tm_skip)%wterm (w2':tm_skip)%wterm)
      /\ gjoin w1' w2' w' /\ t' = (w':tm_skip)%wterm).
Proof.
  intros.
  remember (wtm_par t1 t2) as t.
  revert t1 t2 Heqt.
  induction H; intros; auto; subst.
  left; eexists; eexists; auto.
  edestruct cstep_par_resolve; eauto.

  destruct H1 as [t1' [? ?]]; subst b'.
  edestruct IHcstep_star as [? | ?]; eauto.
  destruct H1 as [w1' [w2' [w' [? [? ?]]]]].
  subst b''.
  right; do 3 eexists; split; eauto.
  econstructor; eauto.

  destruct H1 as [[t2' [? ?]] | [? | [[w1 ?] [[w2 ?] [w ?]]]]]; subst.

  edestruct IHcstep_star as [? | H1]; eauto.
  destruct H1 as [w1' [w2' [w' [? [? ?]]]]].
  subst b''.
  right; do 3 eexists; split; eauto.
  econstructor; eauto.

  edestruct IHcstep_star as [? | H1]; eauto.
  destruct H1 as [w1' [w2' [w' [? [? ?]]]]].
  subst b''.
  right; do 3 eexists; split; eauto.
  econstructor; eauto.

  clear IHcstep_star.
  right.
  replace b'' with (w:tm_skip)%wterm in *.
  inversion H; subst.
  inversion H1; subst s'.
  do 3 eexists; split; eauto.
  eapply cstep_star_skip_subst; eauto.
  symmetry; eapply cstep_star_skip_resolve; eauto.

  edestruct IHcstep_star as [? | H1]; eauto.
  destruct H1 as [w1' [w2' [w' [? [? ?]]]]].
  right; do 3 eexists; split; eauto.
  eapply cstep_iteration; eauto.
Qed.


Lemma cstep_star_split: forall s C t s'' C'' t'',
  cstep_star s C t s'' C'' t'' ->
  exists s', exists C', exists t',
    cstep_star s C t s' C' t' /\ cstep_star s' C' t' s'' C'' t''.
Proof.
  intros.
  induction H as [? | s C t s' C' t' s''' C''' t''' | s C b s' s''' C''' b'''].
  eexists; eexists; eexists; split; eapply cstep_nil.
  destruct IHcstep_star as [s'' [C'' [t'' [? ?]]]].
  eexists; eexists; eexists; split; eauto.
  econstructor; eauto.
  destruct IHcstep_star as [s'' [C'' [t'' [? ?]]]].
  eexists; eexists; eexists; split; eauto.
  eapply cstep_iteration; eauto.
Qed.

Lemma cstep_star_app: forall s C t s' C' t' s'' C'' t'',
  cstep_star s C t s' C' t' -> cstep_star s' C' t' s'' C'' t'' ->
  cstep_star s C t s'' C'' t''.
Proof.
  intros.
  induction H; auto.
  econstructor; eauto.
  eapply cstep_iteration; eauto.
Qed.

Lemma cstep_star_split_ind:
       forall P : state->context->wterm->state->context->wterm->Prop,
       (forall s C t, P s C t s C t) ->
       (forall s C t s' C' t' s'' C'' t'',
        cstep_star s C t s' C' t' ->
        cstep_star s' C' t' s'' C'' t'' ->
        P s' C' t' s'' C'' t'' ->
        P s C t s'' C'' t'') ->
       forall s C t s' C' t',
         cstep_star s C t s' C' t' -> P s C t s' C' t'.
Proof.
  intros.
  induction H1.
  apply H.
  edestruct cstep_star_split; eauto.
  destruct H3 as [C'0 [t'0 [H3 H4]]].
  apply H0 with s' C' b'; eauto.
  econstructor; eauto.
  econstructor; eauto.
  edestruct cstep_star_split; eauto.
  destruct H3 as [C'0 [t'0 [H3 H4]]].
  apply H0 with s'' C'' b''; eauto.
  eapply cstep_iteration; eauto.
  econstructor; eauto.
Qed.

Lemma cstep_weaken_context: forall s C1 t s' C1' t' C2,
  cstep s C1 t s' C1' t' ->
  cstep s C1 C2[t] s' C1' C2[t'].
Proof.
  intros.
  edestruct cstep_inversion as [[? ?] | [? ?]]; eauto.
  subst C1'.
  induction C1; simpl in *;
    econstructor; apply step_add_context; auto.
  subst t'.
  eapply cstep_skip_subst; eauto.
Qed.

Lemma cstep_star_weaken_context: forall s C1 t s'' C1'' t'' C2,
  cstep_star s C1 t s'' C1'' t'' ->
  cstep_star s C1 C2[t] s'' C1'' C2[t''].
Proof.
  intros.
  induction H.
  constructor.
  econstructor; eauto.
  eapply cstep_weaken_context; eauto.
  eapply cstep_iteration; eauto.
Qed.

  Lemma step_cstep: forall s C b s' b',
    step s b s' b' ->
    cstep s C b s' C b'.
  Proof.
    intros.
    apply cstep_simulates_step.
    apply step_add_context.
    auto.
  Qed.

  Lemma context_cstep: forall s C b s' C',
    cstep s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) ->
    cstep s C b s' C' b.
  Proof.
    intros.
    eapply cstep_skip_subst; eauto.
  Qed.

  Lemma cstep_remove_par_skip: forall s C t1 w s' C' t1',
    cstep s C (wtm_par t1 (w:tm_skip)%wterm) s' C' (wtm_par t1' (w:tm_skip)%wterm) ->
    cstep s C t1 s' C' t1'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H0.
    apply step_cstep; auto.
    inverts H4.
    inverts H1.
    apply context_cstep; eauto.
  Qed.

  Lemma cstep_star_remove_par_skip: forall s C b1 w s' C' b1',
    cstep_star s C (wtm_par b1 (w:tm_skip)%wterm) s' C' (wtm_par b1' (w:tm_skip)%wterm) ->
    cstep_star s C b1 s' C' b1'.
  Proof.
    intros.
    remember (b1||w:tm_skip) as b.
    remember (b1'||w:tm_skip) as b'.
    revert b1 b1' Heqb Heqb'.
    induction H; intros; subst.

    inverts Heqb'.
    apply cstep_nil.

    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1.
    eapply cstep_cons.
    apply step_cstep; eauto.
    apply IHcstep_star; auto.
    false; inverts H6.
    false. apply cstep_star_skip_resolve in H0. discriminate.
    subst b'.
    eapply cstep_cons; eauto.
    eapply cstep_remove_par_skip; eauto.

    eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_star_par_swap: forall s C b1 b2 s' C' b1' b2',
    cstep_star s C (b1 || b2) s' C' (b1' || b2') ->
    cstep_star s C (b2 || b1) s' C' (b2' || b1').
  Proof.
    intros.
    remember (b1||b2) as b.
    remember (b1'||b2') as b'.
    revert b1 b2 b1' b2' Heqb Heqb'.
    induction H; intros; subst.

    inverts Heqb'.
    apply cstep_nil.

    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1.
    eapply cstep_cons; eauto.
    apply step_cstep; apply st_par2; eauto.
    eapply cstep_cons; eauto.
    apply step_cstep; apply st_par1; eauto.
    false. apply cstep_star_skip_resolve in H0; discriminate.
    subst b'.
    eapply cstep_cons; eauto.
    apply context_cstep; auto.

    eapply cstep_iteration; eauto.
  Qed.

Lemma cstep_star_snoc: forall s C t s' C' t' s'' C'' t'',
  cstep_star s C t s' C' t' ->
  cstep s' C' t' s'' C'' t'' ->
  cstep_star s C t s'' C'' t''.
Proof.
  intros.
  revert s'' C'' t'' H0.
  induction H; intros;
    try solve [repeat (econstructor; eauto)].
  eapply cstep_iteration; eauto.
Qed.
  

  Lemma cstep_context_cstep_star: forall s C s' C' b' s'' C'' b'',
    cstep s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) ->
    cstep_star s' C' b' s'' C'' b'' ->
    cstep_star s C b' s'' C'' b''.
  Proof.
    intros; eapply cstep_cons; eauto.
    apply cstep_skip_subst with gwemp; eauto.    
  Qed.

Lemma cstep_composed_context_step1: forall s C1 w C2 t s' C1',
  cstep s C1 (w:tm_skip) s' C1' (w:tm_skip) ->
  cstep s (compose_contexts C1 C2) t s' (compose_contexts C1' C2) t.
Proof.
  intros.
  revert s s' C1' H.
  induction C1; intros.

  inverts H; simpl in *; inverts H0.

  destruct C1'; simpl in *; inverts H.
  constructor; auto.
  simpl in *.
  inverts H6.

  destruct C1'; simpl in *; inverts H.
  constructor; auto.
  constructor; auto.
  inverts H6.

  destruct C1'; simpl in *; inverts H.
  constructor; auto.
  constructor; auto.
  simpl in *.
  inverts H6.
Qed.

Lemma cstep_composed_context_step2: forall s C1 C2 w t s' C2',
  cstep s C2 (w:tm_skip) s' C2' (w:tm_skip) ->
  cstep s (compose_contexts C1 C2) t s' (compose_contexts C1 C2') t.
Proof.
  intros.
  apply cstep_skip_subst with w.
  induction C1.
  simpl; auto.
  simpl; constructor; auto.
  simpl; constructor; auto.
  simpl; constructor; auto.
Qed.


Lemma step_cstep_star: forall s b s' C' b' s'' C'' b'',
  step s b s' b' ->
  cstep_star s' C' b' s'' C'' b'' ->
  cstep_star s C' b s'' C'' b''.
Proof.
  intros.
  eapply cstep_cons; eauto.
  apply cstep_simulates_step.
  apply step_add_context; eauto.
Qed.

Lemma step_single_cstep_star: forall s C b s' b',
  step s b s' b' -> cstep_star s C b s' C b'.
Proof. intros; eapply step_cstep_star; eauto; constructor. Qed.

Lemma cstep_star_cx_par_swap: forall s C b1 b2 s'' C'' b1'' b2'',
  cstep_star s (cx_par1 C b2) b1 s'' (cx_par1 C'' b2'') b1'' <->
  cstep_star s (cx_par2 b2 C) b1 s'' (cx_par2 b2'' C'') b1''.
Proof.
  split; intros.
  remember (cx_par1 C b2)%wterm as C2.
  remember (cx_par1 C'' b2'')%wterm as C2''.
  revert C C'' b2 b2'' HeqC2 HeqC2''.
  induction H; intros; subst.

  inverts HeqC2''; apply cstep_nil.

  edestruct cstep_inversion as [[?]|[?]]; eauto. subst C'.
  eapply step_cstep_star; eauto.
  subst b'.
  inverts H1.
  eapply cstep_cons; eauto.
  apply cst_par2; eapply cstep_skip_subst; eauto.
  eapply cstep_cons; eauto.
  apply cst_par2_step1; eauto.
  simpl in H2.
  apply cstep_spin_no_change in H.
  subst s'; apply IHcstep_star; auto.

  eapply cstep_iteration; eauto.

  remember (cx_par2 b2 C)%wterm as C2.
  remember (cx_par2 b2'' C'')%wterm as C2''.
  revert C C'' b2 b2'' HeqC2 HeqC2''.
  induction H; intros; subst.
  inverts HeqC2''.
  apply cstep_nil.
  edestruct cstep_inversion as [[?]|[?]]; eauto. subst C'.
  eapply step_cstep_star; eauto.
  subst b'.
  inverts H1.
  eapply cstep_cons; eauto.
  apply cst_par1; eapply cstep_skip_subst; eauto.
  eapply cstep_cons; eauto.
  apply cst_par1_step2; eauto.
  simpl in H2.
  apply cstep_spin_no_change in H.
  subst s'; apply IHcstep_star; auto.
  eapply cstep_iteration; eauto.
Qed.

Lemma cstep_star_cx_par1_par: forall s C b1 b2 s'' C'' b1'' b2'',
  cstep_star s (cx_par1 C b2) b1 s'' (cx_par1 C'' b2'') b1'' ->
  cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'').
Proof.
  intros.
  remember (cx_par1 C b2)%wterm as C2.
  remember (cx_par1 C'' b2'')%wterm as C2''.
  revert C C'' b2 b2'' HeqC2 HeqC2''.
  induction H; intros; subst.
  inverts HeqC2''.
  apply cstep_nil.
  edestruct cstep_inversion as [[?]|[?]]; eauto. subst C'.
  eapply step_cstep_star. apply st_par1; eauto.
  apply IHcstep_star; auto.
  subst.
  inverts H1.
  eapply cstep_cons.
  apply cstep_skip_subst with gwemp; eauto.
  apply IHcstep_star; auto.
  eapply step_cstep_star. apply st_par2; eauto.
  apply IHcstep_star; auto.
  simpl in H2.
  apply cstep_spin_no_change in H.
  subst s'; apply IHcstep_star; auto.
  eapply cstep_iteration; eauto.
Qed.

Lemma cstep_star_par_cx_par1: forall s C b1 b2 s'' C'' b1'' b2'',
  cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'') ->
  cstep_star s (cx_par1 C b2) b1 s'' (cx_par1 C'' b2'') b1''.
Proof.
  intros.
  remember (b1 || b2)%wterm as b.
  remember (b1'' || b2'')%wterm as b''.
  revert b1 b2 b1'' b2'' Heqb Heqb''.
  induction H; intros; subst.
  inversion Heqb''; subst b1'' b2''.
  apply cstep_nil.
  edestruct cstep_inversion as [[?]|[?]]; eauto.
  subst C'.
  inverts H1.

  eapply step_cstep_star; eauto.

  eapply cstep_cons; eauto; apply cst_par1_step2; eauto.
  elimtype False. clear -H0.
  eapply cstep_star_skip_resolve in H0.
  inversion H0.
  subst b'.
  eapply cstep_cons; eauto.
  apply cst_par1; eauto.
  apply cstep_skip_subst with gwemp; eauto.
  eapply cstep_iteration; eauto.
Qed.

  Lemma cstep_star_cx_par2_par: forall s C b1 b2 s'' C'' b1'' b2'',
    cstep_star s (cx_par2 b1 C) b2 s'' (cx_par2 b1'' C'') b2'' ->
    cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'').
  Proof.
    intros.
    apply cstep_star_par_swap.
    apply cstep_star_cx_par1_par.
    apply cstep_star_cx_par_swap.
    auto.
  Qed.

  Lemma cstep_star_par_cx_par2: forall s C b1 b2 s'' C'' b1'' b2'',
    cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'') ->
    cstep_star s (cx_par2 b1 C) b2 s'' (cx_par2 b1'' C'') b2''.
  Proof.
    intros.
    apply cstep_star_cx_par_swap.
    apply cstep_star_par_cx_par1.
    apply cstep_star_par_swap.
    auto.
  Qed.


Lemma cstep_star_par_compose_contexts1: forall s C b1 b2 s'' C'' b1'' b2'',
  cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'') ->
  cstep_star s (compose_contexts C (cx_par1 cx_hole b2)) b1 s'' (compose_contexts C'' (cx_par1 cx_hole b2'')) b1''.
Proof.
  intros.
  remember (b1 || b2)%wterm as b.
  remember (b1'' || b2'')%wterm as b''.
  revert b1 b2 b1'' b2'' Heqb Heqb''.
  induction H; intros; subst.
  inversion Heqb''; subst b1'' b2''.
  apply cstep_nil.
  edestruct cstep_par_resolve as [? | [? | [? | ?]]]; eauto.

  destruct H1 as [b1' [? ?]]; subst b' C'.
  specialize (IHcstep_star _ _ _ _ (refl_equal _) (refl_equal _)).
  eapply cstep_cons; eauto.
  apply cstep_simulates_step.
  rewrite<- context_compose_par1.
  rewrite<- context_compose_par1.
  apply step_simulates_cstep.
  auto.

  destruct H1 as [b2' [? ?]]; subst b' C'.
  specialize (IHcstep_star _ _ _ _ (refl_equal _) (refl_equal _)).
  eapply cstep_cons; eauto.
  edestruct cstep_inversion; eauto.
  destruct H1 as [? _].
  inverts H1.
  apply cstep_simulates_step.
  rewrite<- context_compose_par1.
  apply step_simulates_cstep.
  auto.
  apply cstep_composed_context_step2 with gwemp.
  constructor; auto.
  destruct H1.
  inversion H2; subst b2; clear H2.
  apply cstep_simulates_step.
  rewrite<- context_compose_par1.
  apply step_simulates_cstep.
  auto.
  
  subst b'.
  specialize (IHcstep_star _ _ _ _ (refl_equal _) (refl_equal _)).
  eapply cstep_cons; eauto.
  edestruct cstep_inversion as [[?]|[?]]; eauto.
  subst C'.
  inverts H1.
  constructor; auto.
  apply cstep_composed_context_step2 with gwemp; auto.
  apply cst_par1_step2; auto.
  apply cstep_composed_context_step1 with gwemp; auto.

  destruct H1 as [[? ?] [[? ?] [? ?]]]; subst.
  clear IHcstep_star.
  eapply cstep_star_skip_resolve in H0.
  discriminate.
  eapply cstep_iteration; eauto.
Qed.  


(*
Lemma compose_contexts_eq1: forall C1 C2 C3,
  compose_contexts C1 C2 = compose_contexts C3 C2 ->
  C1=C3.
Proof.
  induction C1; intros.
  destruct C3; simpl in *; auto; try discriminate.
admit.
admit.
admit.

  destruct C3; simpl in *; auto; try discriminate.
admit.
  inversion H; subst.
  erewrite IHC1; eauto.
  
  destruct C3; simpl in *; auto; try discriminate.
admit.
  inversion H; subst.
  erewrite IHC1; eauto.

  destruct C3; simpl in *; auto; try discriminate.
admit.
  inversion H; subst.
  erewrite IHC1; eauto.
Qed.

Lemma compose_contexts_eq2: forall C1 C2 C3,
  compose_contexts C1 C2 = compose_contexts C1 C3 ->
  C2=C3.
Proof.
  induction C1; intros; simpl in *;
    auto;
    inversion H; auto.
Qed.
*)

Lemma context_par1_resolution: forall s C1 C2 b s'' C'' b'',
  cstep_star s (cx_par1 C1 C2) b s'' C'' b'' ->
  exists C1'', exists C2'', C'' = cx_par1 C1'' C2''.
Proof.
  intros.
  remember (cx_par1 C1 C2) as C.
  revert C1 C2 HeqC.
  induction H; intros; subst.

  eauto.

  edestruct cstep_inversion; eauto.
  destruct H1; subst.
  eapply IHcstep_star; eauto.
  destruct H1; subst.
  inversion H1; subst.
  eapply IHcstep_star; eauto.
  eapply IHcstep_star; eauto.
  eapply IHcstep_star; eauto.
  eapply IHcstep_star; eauto.
Qed.  

Lemma context_par2_resolution: forall s C1 C2 b s'' C'' b'',
  cstep_star s (cx_par2 C1 C2) b s'' C'' b'' ->
  exists C1'', exists C2'', C'' = cx_par2 C1'' C2''.
Proof.
  intros.
  remember (cx_par2 C1 C2) as C.
  revert C1 C2 HeqC.
  induction H; intros; subst.

  eauto.

  edestruct cstep_inversion; eauto.
  destruct H1; subst.
  eapply IHcstep_star; eauto.
  destruct H1; subst.
  inversion H1; subst.
  eapply IHcstep_star; eauto.
  eapply IHcstep_star; eauto.
  eapply IHcstep_star; eauto.
  eapply IHcstep_star; eauto.
Qed.  

Lemma context_seq_resolution: forall s C1 t2 b s'' C'' b'',
  cstep_star s (cx_seq C1 t2) b s'' C'' b'' ->
  exists C1'', exists t2'', C'' = cx_seq C1'' t2''.
Proof.
  intros.
  remember (cx_seq C1 t2) as C.
  revert C1 t2 HeqC.
  induction H; intros; subst.

  eauto.

  edestruct cstep_inversion; eauto.
  destruct H1; subst.
  eapply IHcstep_star; eauto.
  destruct H1; subst.
  inversion H1; subst.
  eapply IHcstep_star; eauto.
  eapply IHcstep_star; eauto.
  eapply IHcstep_star; eauto.
Qed.  

Lemma context_hole_resolution: forall s b s'' C'' b'',
  cstep_star s cx_hole b s'' C'' b'' ->
  C'' = cx_hole.
Proof.
  intros.
  remember cx_hole as C.
  revert HeqC.
  induction H; intros; subst.

  eauto.

  edestruct cstep_inversion; eauto.
  destruct H1; subst.
  eapply IHcstep_star; eauto.
  destruct H1; subst.
  inversion H1; subst.
  eapply IHcstep_star; eauto.
  eapply IHcstep_star; eauto.
Qed.  

Lemma context_hole_origin: forall s C b s'' b'',
  cstep_star s C b s'' cx_hole b'' ->
  C = cx_hole.
Proof.
  intros.
  remember cx_hole as C''.
  revert HeqC''.
  induction H; intros; subst.

  auto.

  rewrite IHcstep_star in *; auto.
  edestruct cstep_inversion; eauto.
  destruct H1; auto.
  destruct H1; subst.
  inversion H1; auto.
  eapply IHcstep_star; eauto.
Qed.

Lemma compose_hole1: forall C, compose_contexts C cx_hole = C.
Proof. intros. induction C; simpl; auto; rewrite IHC; auto. Qed.
Lemma compose_hole2: forall C, compose_contexts cx_hole C = C.
Proof. auto. Qed.






Lemma cstep_star_seq_inversion: forall s C b1 t2 s'' C'' b'',
  cstep_star s C (b1;; t2) s'' C'' b'' ->
  (exists s', exists C', exists w',
  cstep_star s C b1 s' C' (w':tm_skip) /\ cstep_star s' C' (w':t2) s'' C'' b'')
  \/ (exists b1'', cstep_star s C b1 s'' C'' b1'' /\ b'' = (b1'';; t2)%wterm).
Proof.
  intros.
  remember (b1;; t2)%wterm as b.
  revert b1 t2 Heqb.
  induction H; intros; subst.
  right; exists b1; splits; auto; apply cstep_nil.
  edestruct cstep_inversion as [[??]|[??]]; eauto. subst.
  inverts H1.

  edestruct IHcstep_star as [[s'1[C'[w'[??]]]] | [b1''[??]]]; eauto.
  left; exists s'1 C' w'; split; auto; eapply step_cstep_star; eauto.
  subst; right; exists b1''; split; auto; eapply step_cstep_star; eauto.

  clear IHcstep_star; left; exists s' C w; split; auto; apply cstep_nil.

  edestruct IHcstep_star as [[s'1[C'1[w'[??]]]] | [b1''[??]]]; eauto.
  left; exists s'1 C'1 w'; split; auto; eapply cstep_cons; eauto; eapply cstep_skip_subst; eauto.
  subst; right; exists b1''; split; auto; eapply cstep_cons; eauto; eapply cstep_skip_subst; eauto.

  edestruct IHcstep_star as [[s'1[C'1[w'[??]]]] | [b1''[??]]]; eauto.
  left; exists s'1 C'1 w'; split; auto; eapply cstep_iteration; eauto.
  subst; right; exists b1''; split; auto; eapply cstep_iteration; eauto.
Qed.

  Lemma cstep_star_seq_append: forall s C b s'' C'' b'' t2,
    cstep_star s C b s'' C'' b'' ->
    cstep_star s C (b;; t2) s'' C'' (b'';; t2).
  Proof.
    intros.
    induction H.

    apply cstep_nil.

    eapply cstep_cons; eauto.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    apply step_cstep; apply st_seq1; auto.
    subst b'.
    apply context_cstep; auto.

    eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_star_seq_join: forall s C b1 t2 s' C' w' s'' C'' b'',
    cstep_star s C b1 s' C' (w':tm_skip) -> cstep_star s' C' (w':t2) s'' C'' b'' ->
    cstep_star s C (b1;; t2) s'' C'' b''.
  Proof.
    intros.
    remember (w':tm_skip) as b'.
    revert w' s'' C'' b'' Heqb' H0.
    induction H; intros; subst.

    eapply step_cstep_star; eauto. apply st_seq_skip.

    eapply cstep_star_app; eauto.
    eapply cstep_star_app.
    eapply cstep_star_seq_append.
    eapply cstep_cons; eauto.
    apply step_single_cstep_star. apply st_seq_skip.

    eapply cstep_iteration; eauto.
  Qed.

Lemma cstep_single_cstep_star: forall s C b s' C' b',
  cstep s C b s' C' b' -> cstep_star s C b s' C' b'.
Proof. intros; eapply cstep_cons; eauto; constructor. Qed.

Lemma cstep_uncompose_context2: forall s C1 C2 b s' C' b',
  cstep s (compose_contexts C1 C2) b s' C' b' ->
  exists C1', exists C2', cstep s C1 C2[b] s' C1' C2'[b'] /\ (C' = compose_contexts C1' C2').
Proof.
  intros.
  remember (compose_contexts C1 C2) as C.
  revert C1 C2 HeqC.
  induction H; intros; subst.

  destruct C1; simpl in *; try discriminate.
  destruct C2; simpl in *; try discriminate.
  inversion HeqC; subst; clear HeqC.
  edestruct (IHcstep cx_hole C2) as [C1' [C2' [? ?]]]; auto.
  apply cstep_single_cstep_star in H0.
  apply context_hole_resolution in H0.
  subst C1'.
  simpl in H1; subst.
  exists cx_hole; exists (cx_seq C2' t); split; auto.
  apply cstep_simulates_step.
  apply step_add_context.
  simpl.
  apply st_seq1.
  apply step_simulates_cstep; auto.

  inversion HeqC; subst; clear HeqC.
  edestruct IHcstep as [C1' [C2' [? ?]]]; eauto.
  edestruct cstep_inversion.
  apply H0.
  destruct H2; subst C1'.
  exists (cx_seq C1 t); exists C2'; split; auto.
  apply cstep_simulates_step.
  apply step_add_context; auto.
  simpl.
  rewrite H1; auto.
  destruct H2.
  exists (cx_seq C1' t); exists C2'; split; auto.
  rewrite H3.
  apply cst_seq; auto.
  rewrite H3 in H0; auto.
  simpl.
  rewrite H1; auto.


  destruct C1; simpl in *; try discriminate.
  destruct C2; simpl in *; try discriminate.
  rename w into t2'.
  inversion HeqC; subst t2' C2; clear HeqC.
  edestruct (IHcstep cx_hole C) as [C1' [C2' [? ?]]]; auto.
  apply cstep_single_cstep_star in H0.
  apply context_hole_resolution in H0.
  subst C1'.
  simpl in H1; subst.
  exists cx_hole; exists (cx_par1 C2' b2); split; auto.
  apply cstep_simulates_step.
  apply step_add_context.
  simpl.
  apply st_par1.
  apply step_simulates_cstep; auto.
  
  rename w into t2'.
  inversion HeqC; subst t2'; subst; clear HeqC.
  edestruct IHcstep as [C1' [C2' [? ?]]]; eauto.
  edestruct cstep_inversion.
  apply H0.
  destruct H2; subst C1'.
  exists (cx_par1 C1 b2); exists C2'; split; auto.
  apply cstep_simulates_step.
  apply step_add_context; auto.
  simpl.
  rewrite H1; auto.
  destruct H2.
  exists (cx_par1 C1' b2); exists C2'; split; auto.
  rewrite H3.
  apply cst_par1; auto.
  rewrite H3 in H0; auto.
  simpl.
  rewrite H1; auto.


  destruct C1; simpl in *; try discriminate.
  destruct C2; simpl in *; try discriminate.
  rename w into t1'.
  inversion HeqC; subst t1' C2; clear HeqC.
  edestruct (IHcstep cx_hole C) as [C1' [C2' [? ?]]]; auto.
  apply cstep_single_cstep_star in H0.
  apply context_hole_resolution in H0.
  subst C1'.
  simpl in H1; subst.
  exists cx_hole; exists (cx_par2 b1 C2'); split; auto.
  apply cstep_simulates_step.
  apply step_add_context.
  simpl.
  apply st_par2.
  apply step_simulates_cstep; auto.
  
  rename w into t1'.
  inversion HeqC; subst t1'; subst; clear HeqC.
  edestruct IHcstep as [C1' [C2' [? ?]]]; eauto.
  edestruct cstep_inversion.
  apply H0.
  destruct H2; subst C1'.
  exists (cx_par2 b1 C1); exists C2'; split; auto.
  apply cstep_simulates_step.
  apply step_add_context; auto.
  simpl.
  rewrite H1; auto.
  destruct H2.
  exists (cx_par2 b1 C1'); exists C2'; split; auto.
  rewrite H3.
  apply cst_par2; auto.
  rewrite H3 in H0; auto.
  simpl.
  rewrite H1; auto.


  destruct C1; simpl in *; try discriminate.
  subst C2.
  exists cx_hole; exists (cx_par1 C b2'); split; auto.
  apply cstep_simulates_step.
  simpl.
  apply st_par2; auto.
  inversion HeqC; subst C w; clear HeqC.
  exists (cx_par1 C1 b2'); exists C2; split; auto.
  apply cst_par1_step2; auto.


  destruct C1; simpl in *; try discriminate.
  subst C2.
  exists cx_hole; exists (cx_par2 b1' C); split; auto.
  apply cstep_simulates_step.
  simpl.
  apply st_par1; auto.
  inversion HeqC; subst C w; clear HeqC.
  exists (cx_par2 b1' C1); exists C2; split; auto.
  apply cst_par2_step1; auto.

  exists C1; exists C2; split; auto.
  apply cstep_simulates_step.
  do 2 rewrite context_compose_correct; auto.
  apply step_add_context; auto.
Qed.  

Lemma cstep_star_uncompose_context2: forall s C1 C2 b s'' C'' b'',
  cstep_star s (compose_contexts C1 C2) b s'' C'' b'' ->
  exists C1'', exists C2'', cstep_star s C1 C2[b] s'' C1'' C2''[b''] /\ (C'' = compose_contexts C1'' C2'').
Proof.
  intros.
  remember (compose_contexts C1 C2) as C.
  revert C1 C2 HeqC.
  induction H; intros; subst.

  eexists; eexists; split; eauto.
  apply cstep_nil.

  edestruct cstep_uncompose_context2 as [C1' [C2' [? ?]]]; eauto.
  edestruct IHcstep_star as [C1'' [C2'' [? ?]]]; eauto.
  eexists; eexists; split; eauto.
  eapply cstep_cons; eauto.

  edestruct IHcstep_star as [C1'' [C2'' [? ?]]]; eauto.
  eexists; eexists; split; eauto.
  eapply cstep_iteration; eauto.
Qed.

(*
  Lemma cstep_composed_context_skip: forall s C1 C2 w s' C1' C2',
    cstep s (compose_contexts C1 C2) (w:tm_skip) s' (compose_contexts C1' C2') (w:tm_skip) ->
    cstep s C1 C2[w:tm_skip] s' C1' C2'[w:tm_skip].
  Proof.
Admitted.
*)

Lemma cstep_composed_context_inversion: forall s C1 C2 b s' C' b',
  cstep s (compose_contexts C1 C2) b s' C' b' ->
  (exists C1', cstep s C1 (gwemp:tm_skip) s' C1' (gwemp:tm_skip) /\ C'=compose_contexts C1' C2 /\ b'=b)
  \/ (exists C2', cstep s C2 (gwemp:tm_skip) s' C2' (gwemp:tm_skip) /\ C'=compose_contexts C1 C2' /\ b'=b)
  \/ (step s b s' b' /\ C'=compose_contexts C1 C2).
Proof.
  induction C1; intros.

  right.
  simpl in H.
  edestruct cstep_inversion; eauto.
  destruct H0; subst b'.
  left; exists C'; split; auto.

  simpl in H.
  inverts H.
  edestruct IHC1 as [? | [? | ?]]; eauto.
  destruct H as [C1'0 [? [? ?]]].
  left.  exists (cx_seq C1'0 t); splits; eauto.
  apply cst_seq; auto.
  subst; auto.
  destruct H as [C0 [? [? ?]]]; subst.
  right;left. exists C0; splits; eauto.
  destruct H; subst.
  right; right; auto.
  simpl in *.
  edestruct IHC1 as [? | [? | ?]].
  apply step_cstep with (C:=compose_contexts C1 C2); eauto.
  destruct H as [C1'0 [? [? ?]]].
  left.  exists (cx_seq C1'0 t); splits; eauto.
  apply cst_seq; auto.
  subst b'.
  simpl; f_equal.
  rewrite H1; auto.
  destruct H as [C0 [? [? ?]]]; subst.
  right;left. exists C0; splits; eauto.
  rewrite H1; auto.
  destruct H.
  right; right; auto.
  
  simpl in H.
  inverts H.
  edestruct IHC1 as [? | [? | ?]]; eauto.
  destruct H as [C1'0 [? [? ?]]].
  left; exists (cx_par1 C1'0 w); splits; eauto.
  apply cst_par1; auto.
  subst; auto.
  destruct H as [C0 [? [? ?]]]; subst.
  right;left. exists C0; splits; eauto.
  destruct H; subst.
  right; right; auto.
  simpl in *.
  left; exists (cx_par1 C1 b2'); splits; auto.
  apply cst_par1_step2; auto.
  simpl in *.
  edestruct IHC1 as [? | [? | ?]]; eauto.
  apply step_cstep with (C:=compose_contexts C1 C2); eauto.

  simpl in H.
  inverts H.
  edestruct IHC1 as [? | [? | ?]]; eauto.
  destruct H as [C1'0 [? [? ?]]].
  left; exists (cx_par2 w C1'0); splits; eauto.
  apply cst_par2; auto.
  subst; auto.
  destruct H as [C0 [? [? ?]]]; subst.
  right;left. exists C0; splits; eauto.
  destruct H; subst.
  right; right; auto.
  simpl in *.
  left; exists (cx_par2 b1' C1); splits; auto.
  apply cst_par2_step1; auto.
  simpl in *.
  right; right; auto.
Qed.


  (*
Lemma compose_cstep_star_context: forall s C1 C2 b s'' C'' b'',
  cstep_star s C1 C2[b] s'' C'' b'' ->
  exists C1'', exists C2'',
    cstep_star s (compose_contexts C1 C2) b s'' (compose_contexts C1'' C2'') b''
    /\ C'' = compose_contexts C1'' C2''.
Admitted.
*)
(*
Lemma simulation_weaken_context': forall C b1 b2,
  sims b1 b2 -> sims C[b1] C[b2].
Proof.
  intros.
  unfold sims.
  intros.
  assert ((forall s1' C1' b1', ~cstep_star s C0 C[b1] s1' C1' b1' \/ ~eqprog s1' C1' b1' s' C' b2') -> False).
  intros.
  unfold sims in H.
  edestruct compose_cstep_star_context as [C1'' [C2'' [? ?]]]; eauto.
  apply H in H2.
  destruct H2 as [s1' [C1' [b1' [? ?]]]].
  specialize (H1 s1' C1' b1').  
  

cstep_star C[b2] .. C'[b2'] -> forall cstep_star C'[b2'] C''[b2''], ~cstep_star C[b1] C''[b2'']

then C[b2] can do more than C[b1]; but then, any other context could use C to do
more with b2 than with b1 such that b1 doesn't simulate b2. contradiction!


Lemma aaaa: forall s C b1 b2 s'' C'' b1'' b2'',
  cstep_star s (compose_contexts C (cx_par2 b1 cx_hole)) b2 s'' (compose_contexts C'' (cx_par2 b1'' cx_hole)) b2'' ->
  cstep_star s (compose_contexts C (cx_par1 cx_hole b2)) b1 s'' (compose_contexts C'' (cx_par1 cx_hole b2'')) b1'' ->
  cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'').
Admitted.  
*)
(*
Lemma cstep_star_par_seq_resolve: forall s C b1 t3 b2 s'' C'' b'',
  cstep_star s C (b1; t3 || b2) s'' C'' b'' ->
  (exists s', exists C', exists w1', exists b2',
    cstep_star s C (b1 || b2) s' C' (w1':tm_skip || b2')
    /\ cstep_star s' C' (w1':t3 || b2') s'' C'' b'')
  \/ (exists b1'', exists b2'',
    cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'') /\ b'' = (b1'';t3 || b2'')%wterm).
Proof.
  intros.

  assert (exists b1'', exists b2'', b'' = (b1'' || b2'')%wterm).
  admit.
  destruct H0 as [b1'' [b2'' ?]].
  subst.

  generalize H; intro HH.
  apply cstep_star_par_compose_contexts1 in HH.
  edestruct cstep_star_seq_resolve; eauto.
  destruct H0 as [s' [C' [w1' [? ?]]]].
  
  rename C' into C1'.
  assert (exists C', exists b2', C1' = compose_contexts C' (cx_par1 cx_hole b2')).
  admit.
  destruct H2 as [C' [b2' ?]].
  subst C1'.

  left; exists s'; exists C'; exists w1'; exists b2'; split; auto.
  apply aaaa; auto.
  apply cstep_star_par_compose_contexts2.
  
  






  replace (b1; t3 || b2)%wterm with ((cx_par1 cx_hole b2)[b1; t3]); auto.
  Check cstep_star_weaken_context.
  

Focus 2.
auto.

  remember (b1; t3 || b2)%wterm as b.
  revert b1 t3 b2 Heqb.
  induction H; intros; subst.

  right; eexists; eexists; split; eauto; constructor.

  

 apply cstep_nil.



*)

  Lemma cstep_star_par1_seq_inversion: forall s C b1 t2 b3 s'' C'' b'',
    cstep_star s C ((b1;;t2) || b3) s'' C'' b'' ->
    (exists b1', exists b3',
      cstep_star s C (b1 || b3) s'' C'' (b1' || b3')
      /\ b'' = ((b1';;t2) || b3')%wterm)
    \/ (exists s', exists C', exists w2, exists b3',
      cstep_star s C (b1 || b3) s' C' (w2:tm_skip || b3')
      /\ cstep_star s' C' (w2:t2 || b3') s'' C'' b'').
  Proof.
    intros.
    remember (b1;;t2 || b3) as b.
    revert b1 t2 b3 Heqb.
    induction H; intros; subst.

    left; do 2 eexists; split; eauto. apply cstep_nil.

    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1. inverts H6.
    edestruct IHcstep_star as [ [b1'[b3'[??]]] | [s'0[C'[w2[b3'[??]]]]] ]; clear IHcstep_star; eauto.
    subst b''.
    left; do 2 eexists; split; eauto.
    eapply step_cstep_star; eauto; apply st_par1; auto.
    right; do 4 eexists; split; eauto.
    eapply step_cstep_star; eauto; apply st_par1; auto.
    clear IHcstep_star.
    right; do 4 eexists; split; eauto.
    apply cstep_nil.
    edestruct IHcstep_star as [ [b1'[b3'[??]]] | [s'0[C'[w2[b3'[??]]]]] ]; clear IHcstep_star; eauto.
    subst b''.
    left; do 2 eexists; split; eauto.
    eapply step_cstep_star; eauto; apply st_par2; auto.
    right; do 4 eexists; split; eauto.
    eapply step_cstep_star; eauto; apply st_par2; auto.
    subst b'.
    edestruct IHcstep_star as [ [b1'[b3'[??]]] | [s'0[C'''[w2[b3'[??]]]]] ]; clear IHcstep_star; eauto.
    subst b''.
    left; do 2 eexists; split; eauto.
    eapply cstep_cons; eauto.
    apply context_cstep; eauto.
    right; do 4 eexists; split; eauto.
    eapply cstep_cons; eauto.
    apply context_cstep; eauto.

    edestruct IHcstep_star as [ [b1'[b3'[??]]] | [s'0[C'''[w2[b3'[??]]]]] ]; clear IHcstep_star; eauto.
    subst b''.
    left; do 2 eexists; split; eauto.
    eapply cstep_iteration; eauto.
    right; do 4 eexists; split; eauto.
    eapply cstep_iteration; eauto.
  Qed.


Lemma cstep_star_par_seq_inversion: forall s C b1 b2 t3 s'' C'' b'',
  cstep_star s C (b1 || (b2;; t3)) s'' C'' b'' ->
  (exists b1', exists b2',
    cstep_star s C (b1 || b2) s'' C'' (b1' || b2')
    /\ b'' = (b1' || (b2';; t3))%wterm)
  \/ (exists s', exists C', exists b1', exists w3,
    cstep_star s C (b1 || b2) s' C' (b1' || w3:tm_skip)
    /\ cstep_star s' C' (b1' || w3:t3) s'' C'' b'').
Proof.
  intros.
  remember (b1 || (b2;; t3))%wterm as b.
  revert b1 b2 t3 Heqb.
  induction H; intros; subst.

  left; exists b1; exists b2.
  split; auto.
  apply cstep_nil.

  edestruct cstep_inversion; eauto.
  destruct H1; subst C'.
  inversion H1; try subst b0 b3 s'0 t'; substs.

  edestruct IHcstep_star as [? | ?]; eauto;
    clear IHcstep_star.
  destruct H2 as [b1'' [b2'' [? ?]]].
  left; exists b1''; exists b2''.
  split; auto.
  eapply cstep_cons; eauto.
  apply cstep_simulates_step.
  apply step_add_context.
  apply st_par1; auto.
  destruct H2 as [s'0 [C'0 [b1'0 [w2'0 [? ?]]]]].
  right; exists s'0; exists C'0; exists b1'0; exists w2'0.
  split; auto.
  eapply cstep_cons; eauto.
  apply cstep_simulates_step.
  apply step_add_context.
  apply st_par1; auto.

  inversion H6; subst.

  edestruct IHcstep_star as [? | ?]; eauto;
    clear IHcstep_star.
  destruct H2 as [b1'' [b2'' [? ?]]].
  left; exists b1''; exists b2''.
  split; auto.
  eapply cstep_cons; eauto.
  apply cstep_simulates_step.
  apply step_add_context.
  apply st_par2; auto.
  destruct H2 as [s'0 [C'0 [b1'0 [w2'0 [? ?]]]]].
  right; exists s'0; exists C'0; exists b1'0; exists w2'0.
  split; auto.
  eapply cstep_cons; eauto.
  apply cstep_simulates_step.
  apply step_add_context.
  apply st_par2; auto.
  clear IHcstep_star.
  right; exists s'; exists C; exists b1; exists w.
  split; auto.
  apply cstep_nil.

  destruct H1; subst b'.
  edestruct IHcstep_star as [? | ?]; eauto;
    clear IHcstep_star.
  destruct H2 as [b1'' [b2'' [? ?]]].
  subst b''.
  left; exists b1''; exists b2''.
  split; auto.
  eapply cstep_cons; eauto.
  apply cstep_skip_subst with gwemp; auto.
  destruct H2 as [s'0 [C'0 [b1'0 [w2'0 [? ?]]]]].
  right; exists s'0; exists C'0; exists b1'0; exists w2'0.
  split; auto.
  eapply cstep_cons; eauto.
  apply cstep_skip_subst with gwemp; auto.

  edestruct IHcstep_star as [? | ?]; eauto;
    clear IHcstep_star.
  destruct H1 as [b1'' [b2'' [? ?]]].
  subst b''.
  left; exists b1''; exists b2''.
  split; auto.
  eapply cstep_iteration; eauto.
  destruct H1 as [s'0 [C'0 [b1'0 [w2'0 [? ?]]]]].
  right; exists s'0; exists C'0; exists b1'0; exists w2'0.
  split; auto.
  eapply cstep_iteration; eauto.
Qed.  

(*
Lemma cstep_star_par_seq_inversion': forall s C b1 t2 b3 t4 s'' C'' b'',
  cstep_star s C ((b1;; t2) || (b3;; t4)) s'' C'' b'' ->
  (exists b1', exists b3',
    cstep_star s C (b1 || b3) s'' C'' (b1' || b3')
    /\ b'' = ((b1';; t2) || (b3';; t4))%wterm)
  \/ (exists s', exists C', exists b1', exists w4,
    cstep_star s C (b1 || b3) s' C' (b1' || w4:tm_skip)
    /\ cstep_star s' C' (b1';;t2 || w4:t4) s'' C'' b'')
  \/ (exists s', exists C', exists b3', exists w2,
    cstep_star s C (b1 || b3) s' C' (w2:tm_skip || b3')
    /\ cstep_star s' C' (w2:t2 || (b3';;t4)) s'' C'' b'').
Proof.
Admitted.
*)

Lemma cstep_star_term_par_seq_inversion: forall s C b1 w2 t2 t3 s'' C'' b'',
  cstep_star s C (b1 || w2:(tm_seq t2 t3)) s'' C'' b'' ->
  (exists b1', exists b2',
    cstep_star s C (b1 || w2:t2) s'' C'' (b1' || b2')
    /\ b'' = (b1' || (b2';; t3))%wterm)
  \/ (exists s', exists C', exists b1', exists w2',
    cstep_star s C (b1 || w2:t2) s' C' (b1' || w2':tm_skip)
    /\ cstep_star s' C' (b1' || w2':t3) s'' C'' b'')
  \/ (s''=s /\ C''=C /\ b''=(b1 || w2:(tm_seq t2 t3))%wterm).
Proof.
Admitted.

Lemma cstep_star_term_par1_seq_inversion: forall s C w1 t1 t2 b3 s'' C'' b'',
  cstep_star s C ((w1:tm_seq t1 t2) || b3) s'' C'' b'' ->
  (exists b1', exists b3',
    cstep_star s C (w1:t1 || b3) s'' C'' (b1' || b3')
    /\ b'' = ((b1';;t2) || b3')%wterm)
  \/ (exists s', exists C', exists w2, exists b3',
    cstep_star s C (w1:t1 || b3) s' C' (w2:tm_skip || b3')
    /\ cstep_star s' C' (w2:t2 || b3') s'' C'' b'')
  \/ (s''=s /\ C''=C /\ b''=((w1:tm_seq t1 t2) || b3)%wterm).
Proof.
Admitted.

Lemma cstep_star_extend_world: forall s C w1 w2 w b s' C' b' b2 b2',
  cstep_star s C b s' C' b' ->
  join_wterms b w1 ->
  join w1 w2 w ->
  b2 = extend_world b w2 ->
  b2' = extend_world b' w2 ->
  cstep_star s C (b2) s' C' (b2').
Admitted.

(*
  Lemma cstep_star_par2_seq_append: forall s C b1 b2 t3 s'' C'' b'',
    cstep_star s C (b1 || b2) s'' C'' b'' ->
    cstep_star s C (b1 || (b2;; t3)) s'' C'' (b'';; t3).
Admitted.
*)

(*
Lemma cstep_star_par2_append: forall s C b1 b2 t3 s' C' b1' b2',
  cstep_star s C (b1 || b2) s' C' (b1'||b2') ->
  cstep_star s C (b1 || (b2;; t3)) s' C' (b1' || (b2';; t3)).
Admitted.
*)

(*
Lemma cstep_star_par1_seq_append: forall s C b1 b2 t3 s'' C'' b'',
  cstep_star s C (b1 || b2) s'' C'' b'' ->
  cstep_star s C ((b1;; t3) || b2) s'' C'' (b'';; t3).
Admitted.
*)

(*
Lemma cstep_star_preserves_wjoins: forall s C b s' C' b',
  cstep_star s C b s' C' b' -> exists w', join_wterms b' w'.
Admitted.
*)

  Lemma cstep_star_par1_inner_seq_append: forall s C b1 b2 s' C' b1' b2' t3,
    cstep_star s C (b1||b2) s' C' (b1'||b2') ->
    cstep_star s C (b1;;t3||b2) s' C' (b1';;t3||b2').
  Proof.
    intros.
    remember (b1||b2) as b.
    remember (b1'||b2') as b'.
    revert b1 b2 b1' b2' Heqb Heqb'.
    induction H; intros; subst.

    inverts Heqb'. apply cstep_nil.

    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1.
    eapply step_cstep_star; eauto. apply st_par1; apply st_seq1; auto.
    eapply step_cstep_star; eauto. apply st_par2; auto.
    false; eapply cstep_star_skip_resolve in H0; discriminate.
    subst b'.
    eapply cstep_context_cstep_star; eauto.

    eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_star_par2_inner_seq_append: forall s C b1 b2 s' C' b1' b2' t3,
    cstep_star s C (b1||b2) s' C' (b1'||b2') ->
    cstep_star s C (b1||(b2;;t3)) s' C' (b1'||(b2';;t3)).
  Proof.
    intros.
    remember (b1||b2) as b.
    remember (b1'||b2') as b'.
    revert b1 b2 b1' b2' Heqb Heqb'.
    induction H; intros; subst.

    inverts Heqb'. apply cstep_nil.

    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1.
    eapply step_cstep_star; eauto. apply st_par1; auto.
    eapply step_cstep_star; eauto. apply st_par2; apply st_seq1; auto.
    false; eapply cstep_star_skip_resolve in H0; discriminate.
    subst b'.
    eapply cstep_context_cstep_star; eauto.

    eapply cstep_iteration; eauto.
  Qed.

Lemma cstep_star_par1_inner_tm_seq_append: forall s C w1 t1 b2 s' C' b1' b2' t3,
  cstep_star s C (w1:t1||b2) s' C' (b1'||b2') ->
  cstep_star s C ((w1:tm_seq t1 t3)||b2) s' C' ((b1';;t3)||b2').
  Proof.
    intros.
    remember (w1:t1||b2) as b.
    remember (b1'||b2') as b'.
    revert t1 b2 b1' b2' Heqb Heqb'.
    induction H; intros; subst.

    inverts Heqb'.
    apply step_single_cstep_star.
    apply st_par1; apply st_seq.

    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1.
    eapply step_cstep_star; eauto. apply st_par1; apply st_seq; auto.
    eapply step_cstep_star; eauto. apply st_par1; apply st_seq1; eauto.
    eapply cstep_star_par1_inner_seq_append; eauto.
    eapply step_cstep_star; eauto. apply st_par2; auto.
    false; eapply cstep_star_skip_resolve in H0; discriminate.
    subst b'.
    eapply cstep_context_cstep_star; eauto.

    eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_star_par2_inner_tm_seq_append: forall s C b1 w2 t2 s' C' b1' b2' t3,
    cstep_star s C (b1||w2:t2) s' C' (b1'||b2') ->
    cstep_star s C (b1||(w2:tm_seq t2 t3)) s' C' (b1'||(b2';;t3)).
  Proof.
    intros.
    remember (b1||w2:t2) as b.
    remember (b1'||b2') as b'.
    revert b1 t2 b1' b2' Heqb Heqb'.
    induction H; intros; subst.

    inverts Heqb'.
    apply step_single_cstep_star.
    apply st_par2; apply st_seq.

    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1.
    eapply step_cstep_star; eauto. apply st_par1; auto.
    eapply step_cstep_star; eauto. apply st_par2; apply st_seq; auto.
    eapply step_cstep_star; eauto. apply st_par2; apply st_seq1; eauto.
    eapply cstep_star_par2_inner_seq_append; eauto.
    false; eapply cstep_star_skip_resolve in H0; discriminate.
    subst b'.
    eapply cstep_context_cstep_star; eauto.

    eapply cstep_iteration; eauto.
  Qed.


Lemma cstep_star_par1_step: forall s C b1 b2 s' C' b1',
  cstep_star s C b1 s' C' b1' ->
  cstep_star s C (b1||b2) s' C' (b1'||b2).
Proof.
  intros.
  replace (b1 || b2)%wterm with ((cx_par1 cx_hole b2)[b1]); auto.
  replace (b1' || b2)%wterm with ((cx_par1 cx_hole b2)[b1']); auto.
  apply cstep_star_weaken_context.
  auto.
Qed.

Lemma cstep_star_par2_step: forall s C b1 b2 s' C' b2',
  cstep_star s C b2 s' C' b2' ->
  cstep_star s C (b1||b2) s' C' (b1||b2').
Proof.
  intros.
  replace (b1 || b2)%wterm with ((cx_par2 b1 cx_hole)[b2]); auto.
  replace (b1 || b2')%wterm with ((cx_par2 b1 cx_hole)[b2']); auto.
  apply cstep_star_weaken_context.
  auto.
Qed.

Lemma seq_tm_seq_equiv2: forall s C1 C2 w2 t2 t3 s' C1' C2' b23',
  cstep_star s C1 C2[w2: t2;; t3] s' C1' C2'[b23'] ->
  cstep_star s C1 C2[w2: tm_seq t2 t3] s' C1' C2'[b23'].
Proof.
  intros.
  eapply cstep_cons; eauto.
  apply step_cstep.
  apply step_add_context.
  econstructor.
Qed.

Lemma seq_tm_seq_equiv2': forall s C w2 t2 t3 s' C' b23',
  cstep_star s C (w2: t2;; t3) s' C' b23' ->
  cstep_star s C (w2: tm_seq t2 t3) s' C' b23'.
Proof.
  intros.
  eapply cstep_cons; eauto.
  apply step_cstep.
  econstructor.
Qed.


Lemma cstep_star_composed_context_step2: forall s C1 C2 w b s' C2',
  cstep_star s C2 (w : tm_skip) s' C2' (w : tm_skip) ->
  cstep_star s (compose_contexts C1 C2) b s' (compose_contexts C1 C2') b.
Proof.
  intros.
  remember (w : tm_skip)%wterm as b2 in H at 1.
  remember (w : tm_skip)%wterm as b2' in H.
  induction H; subst.
  apply cstep_nil.
  replace b' with (w:tm_skip)%wterm in *.
  eapply cstep_cons; eauto.
  eapply cstep_composed_context_step2; eauto.
  symmetry; eapply cstep_skip_resolve; eauto.
  eapply cstep_iteration; eauto.
Qed.

Lemma cstep_star_composed_context_step1: forall s C1 C2 w b s' C1',
  cstep_star s C1 (w : tm_skip) s' C1' (w : tm_skip) ->
  cstep_star s (compose_contexts C1 C2) b s' (compose_contexts C1' C2) b.
Proof.
  intros.
  remember (w : tm_skip)%wterm as b1 in H at 1.
  remember (w : tm_skip)%wterm as b1' in H.
  induction H; subst.
  apply cstep_nil.
  replace b' with (w:tm_skip)%wterm in *.
  eapply cstep_cons; eauto.
  eapply cstep_composed_context_step1; eauto.
  symmetry; eapply cstep_skip_resolve; eauto.
  eapply cstep_iteration; eauto.
Qed.


Lemma cstep_star_composed_context_step2': forall s C1 C2 b s' C2' b',
  cstep_star s C2 b s' C2' b' ->
  cstep_star s (compose_contexts C1 C2) b s' (compose_contexts C1 C2') b'.
Proof.
  intros.
  induction H; subst.
  apply cstep_nil.
  edestruct cstep_inversion as [[??]|[??]]; eauto. subst.
  eapply step_cstep_star; eauto.
  subst b'.
  eapply cstep_cons; eauto.
  eapply cstep_composed_context_step2; eauto.
  eapply cstep_iteration; eauto.
Qed.

Lemma cstep_star_composed_context_step1': forall s C1 C2 b s' C1' b',
  cstep_star s C1 b s' C1' b' ->
  cstep_star s (compose_contexts C1 C2) b s' (compose_contexts C1' C2) b'.
Proof.
  intros.
  induction H; subst.
  apply cstep_nil.
  edestruct cstep_inversion as [[??]|[??]]; eauto. subst.
  eapply step_cstep_star; eauto.
  subst b'.
  eapply cstep_cons; eauto.
  eapply cstep_composed_context_step1; eauto.
  eapply cstep_iteration; eauto.
Qed.

Lemma cstep_star_seq_compose_contexts: forall s C b s'' C'' b'' t,
  cstep_star s C b s'' C'' b'' ->
  cstep_star s (compose_contexts C (cx_seq cx_hole t)) b s'' (compose_contexts C'' (cx_seq cx_hole t)) b''.
Proof.
  intros.
  induction H; intros; subst.
  apply cstep_nil.
  edestruct cstep_inversion as [[??]|[??]]; eauto.
  subst.
  eapply step_cstep_star; eauto.
  subst.
  eapply cstep_cons; eauto.
  eapply cstep_composed_context_step1; eauto.
  eapply cstep_iteration; eauto.
Qed.  

Lemma compose_context_assoc:  forall C1 C2 C3,
  compose_contexts C1 (compose_contexts C2 C3) = compose_contexts (compose_contexts C1 C2) C3.
Proof.
  intros.
  induction C1; simpl in *; auto;
    rewrite IHC1; auto.
Qed.


Lemma cstep_star_compose_contexts_skip2: forall s C w b s' C' b',
  cstep_star s (compose_contexts C (cx_par1 cx_hole (w:tm_skip))) b s' C' b' ->
  exists C'0,
    cstep_star s C b s' C'0 b'.
Proof.
  intros. remember (compose_contexts C (cx_par1 cx_hole (w : tm_skip))) as C_.
  revert C w HeqC_.
  induction H; intros; subst.
  exists C0; apply cstep_nil.
  edestruct cstep_composed_context_inversion as [?|[?|?]]; eauto.

  destruct H1 as [C'0[?[??]]]. subst.
  destruct IHcstep_star with C'0 w as [C''0 H2].
  reflexivity.
  exists C''0. eapply cstep_cons; eauto. apply cstep_skip_subst with gwemp; auto.
  destruct H1 as [C'0[?[??]]]. subst.

  inverts H1. inverts H7. simpl in *. inverts H1. inverts H7.
  inverts H2.

  destruct H1 as [??]. subst.
  destruct IHcstep_star with C0 w as [C''0 H2].
  reflexivity.
  exists C''0. eapply step_cstep_star; eauto.

  edestruct IHcstep_star as [C''0 H2]; eauto.
  exists C''0. eapply cstep_iteration; eauto.
Qed.

Lemma cstep_star_context_step_inversion: forall s C b s'''' C'''' b'''',
  cstep_star s C b s'''' C'''' b'''' ->
  (exists s', exists C', exists s'', exists b'',
  cstep_star s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) /\
  step s' b s'' b'' /\
  cstep_star s'' C' b'' s'''' C'''' b'''')
  \/ (cstep_star s C (gwemp:tm_skip) s'''' C'''' (gwemp:tm_skip) /\ b''''=b).
Proof.
  intros.
  induction H.

  right; split; [constructor | auto].

  rename s'' into s''''; rename C'' into C''''; rename b'' into b''''; rename b' into b'; rename b into b.
  destruct IHcstep_star as [[s'' [C'' [s''' [b''' [? [? ?]]]]]]|[? ?]].
  edestruct cstep_inversion as [[??]|[??]]; eauto.
  subst C'.
  left; do 4 eexists; splits; eauto.
  apply cstep_nil.
  subst.
  left; do 4 eexists; splits; eauto.
  eapply cstep_cons; eauto.
  subst.
  edestruct cstep_inversion as [[??]|[??]]; eauto.
  subst.
  left; do 4 eexists; splits; eauto.
  apply cstep_nil.
  subst.
  right; split; auto.
  econstructor; eauto.

  rename s'' into s''''; rename C'' into C''''; rename b'' into b''''; rename b into b.
  destruct IHcstep_star as [[s'' [C'' [s''' [b''' [? [? ?]]]]]]|[? ?]].
  left; do 4 eexists; splits; eauto.
  eapply cstep_iteration; eauto.
  right; split; auto.
  eapply cstep_iteration; eauto.
Qed.

Lemma cstep_star_par_compose_contexts1': forall s C b1 b2 s'' C'' b1'' b2'',
  cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'') ->
  cstep_star s (compose_contexts (cx_par1 cx_hole b2) C) b1 s''
    (compose_contexts (cx_par1 cx_hole b2'') C'') b1''.
Proof.
  intros.
  remember (b1 || b2)%wterm as b.
  remember (b1'' || b2'')%wterm as b''.
  revert b1 b2 b1'' b2'' Heqb Heqb''.
  induction H; intros; subst.
  inversion Heqb''; subst b1'' b2''.
  apply cstep_nil.
  edestruct cstep_inversion as [[?]|[?]]; eauto.
  subst.
  inverts H1.
  eapply step_cstep_star; eauto.
  eapply cstep_cons; eauto.
  constructor. auto.
  apply cstep_star_skip_resolve in H0. discriminate.
  subst.
  eapply cstep_cons; eauto.
  eapply cstep_composed_context_step2; eauto.
  eapply cstep_iteration; eauto.
Qed.
  

Lemma cstep_star_par_compose_contexts2': forall s C b1 b2 s'' C'' b1'' b2'',
  cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'') ->
  cstep_star s (compose_contexts (cx_par2 b1 cx_hole) C) b2 s''
    (compose_contexts (cx_par2 b1'' cx_hole) C'') b2''.
Proof.
  intros.
  remember (b1 || b2)%wterm as b.
  remember (b1'' || b2'')%wterm as b''.
  revert b1 b2 b1'' b2'' Heqb Heqb''.
  induction H; intros; subst.
  inversion Heqb''; subst b1'' b2''.
  apply cstep_nil.
  edestruct cstep_inversion as [[?]|[?]]; eauto.
  subst.
  inverts H1.
  eapply cstep_cons; eauto.
  constructor; auto.
  eapply step_cstep_star; eauto.
  apply cstep_star_skip_resolve in H0. discriminate.
  subst.
  eapply cstep_cons; eauto.
  eapply cstep_composed_context_step2; eauto.
  eapply cstep_iteration; eauto.
Qed.



  Lemma cstep_star_tm_seq_inversion: forall s C w t1 t2 s'' C'' b'',
    cstep_star s C (w: tm_seq t1 t2) s'' C'' b'' ->
    exists s' C',
    cstep_star s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) /\ 
      ((s''=s' /\ C''=C' /\ b''=(w: tm_seq t1 t2)%wterm) \/ cstep_star s' C' (w:t1;; t2) s'' C'' b'').
  Proof.
    intros.
    remember (w:tm_seq t1 t2)%wterm as b.
    induction H; subst.
    exists s C; splits; auto; apply cstep_nil.
    edestruct cstep_inversion as [[?]|[?]]; eauto; try subst C'.
    inverts H1.

    exists s' C; splits; subst; eauto; apply cstep_nil.
    subst.
    edestruct IHcstep_star as [s'1[C'1[?[[?[??]]|?]]]]; auto; subst; clear IHcstep_star;
      exists s'1 C'1; splits; auto; eapply cstep_cons; eauto.

    edestruct IHcstep_star as [s'1[C'1[?[[?[??]]|?]]]]; auto; subst; clear IHcstep_star;
      exists s'1 C'1; splits; auto; eapply cstep_iteration; eauto.
  Qed.


Import world_model.Notation.
Open Scope forces_pred.

Lemma cstep_star_tm_par_resolve: forall s C w P1 t1 P2 t2 s' C' b',
  cstep_star s C (w: tm_par P1 t1 P2 t2) s' C' b' ->
  (cstep_star s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) /\ b'=(w: tm_par P1 t1 P2 t2)%wterm)
  \/ (exists w1 w2 b1' b2', gjoin w1 w2 w /\ w1|=gsep_con P1 TT /\ w2|=gsep_con P2 TT /\ cstep_star s C (w1:t1||w2:t2) s' C' b' /\ b' = (b1'||b2')%wterm)
  \/ (exists w1 w2 w1' w2' w',
      gjoin w1 w2 w /\ w1|=gsep_con P1 TT /\ w2|=gsep_con P2 TT /\
      cstep_star s C (w1:t1 || w2:t2) s' C' (wtm_par (w1':tm_skip)%wterm (w2':tm_skip)%wterm)
      /\ gjoin w1' w2' w' /\ b' = (w':tm_skip)%wterm).
Proof.
  intros.
  remember (w : tm_par P1 t1 P2 t2)%wterm as b.
  revert P1 t1 P2 t2 Heqb.
  induction H; intros; subst.
  left; split; auto. apply cstep_nil.
  edestruct cstep_inversion as [[?]|[?]]; eauto.
  subst C'.
  inverts H1.

  right.
  clear IHcstep_star.
  edestruct cstep_star_par_resolve as [[b1'[b2' ?]] | [w1'[w2'[w'[?[??]]]]]]; eauto.
  subst b''.
  left. exists w1 w2 b1' b2'; splits; auto.
  right.
  subst b''.
  exists w1 w2 w1' w2' w'; splits; auto.
  subst b'.
  edestruct IHcstep_star as [[??] | [[w1[w2[b1'[b2'[?[?[?[??]]]]]]]] | [w1[w2[w1'[w2'[w'[?[?[?[?[??]]]]]]]]]]]]; eauto; clear IHcstep_star.
  subst b''.
  left; split; auto. eapply cstep_cons; eauto.
  subst b''.
  right; left. exists w1 w2 b1' b2'; splits; auto.
  eapply cstep_cons; eauto. apply cstep_skip_subst with gwemp. auto.
  subst b''.
  right; right. exists w1 w2 w1' w2' w'; splits; auto.
  eapply cstep_cons; eauto. apply cstep_skip_subst with gwemp. auto.

  edestruct IHcstep_star as [[??] | [[w1[w2[b1'[b2'[?[?[?[??]]]]]]]] | [w1[w2[w1'[w2'[w'[?[?[?[?[??]]]]]]]]]]]]; eauto; clear IHcstep_star.
  subst b''.
  left; split; auto. eapply cstep_iteration; eauto.
  subst b''.
  right; left. exists w1 w2 b1' b2'; splits; auto.
  eapply cstep_iteration; eauto.
  subst b''.
  right; right. exists w1 w2 w1' w2' w'; splits; auto.
  eapply cstep_iteration; eauto.
Qed.

  Lemma cstep_star_while_max_skip_resolve': forall s C w e (z:iterations) t s' C' w',
    cstep_star s C (w: tm_while_max e z t) s' C' (w': tm_skip) ->
    evalFalse e w' \/
    (exists (n:nat), z=n /\ cstep_star s C (w: tm_while_max e z t) s' C' (w': tm_while_max e 0 t)).
  Proof.
    intros.
    remember (w: tm_while_max e z t)%wterm as b.
    assert (exists z', iteration_decrs z z' /\ ((exists w, b = w: tm_while_max e z' t) \/ (exists bt, b = bt;; tm_while_max e z' t)))%wterm.
      subst; exists z; split; jauto; apply ids_nil.
    clear Heqb w.
    destruct H0 as [z'[??]].
    revert z' H1 H0.
    remember (w': tm_skip)%wterm as b'.
    revert w' Heqb'.
    induction H; intros; subst.

    destruct H1 as [[w ?] | [b ?]]; discriminate.

    edestruct cstep_inversion as [[?]|[?]]; eauto; clear H.
    subst.
    destruct H1 as [[w ?] | [bt ?]]; subst b.
    inverts H3.
    edestruct IHcstep_star as [? | [n[??]]]; eauto.
    eapply ids_decr; eauto.
    subst z.
    right; exists n; split; eauto.
    eapply step_cstep_star; eauto.
    apply st_while_true; eauto.
    apply cstep_star_skip_resolve in H0.
    inverts H0; auto.
    false.
    destruct z.
    edestruct iter_decrs_nat1 as [n'[??]]; eauto; discriminate.
    apply iter_decrs_inf1 in H2; discriminate.
    edestruct iter_decrs_nat2 as [n[??]]; eauto.
    inverts H.
    replace w' with w in *.
    2: apply cstep_star_skip_resolve in H0; inverts H0; auto.
    right; exists n; splits; auto.
    apply cstep_star_skip_subst with w; auto.
    inverts H3.
    edestruct IHcstep_star as [? | [n[??]]]; eauto.
    inverts H.
    right; exists n; splits; auto.
    eapply step_cstep_star; eauto.
    apply st_seq1; auto.
    edestruct IHcstep_star as [? | [n[??]]]; eauto.
    inverts H.
    right; exists n; splits; auto.
    eapply step_cstep_star; eauto.
    apply st_seq_skip; auto.
    subst.
    edestruct IHcstep_star as [? | [n[??]]]; eauto.
    right; exists n; splits; auto.
    eapply cstep_context_cstep_star; eauto.
    edestruct IHcstep_star as [? | [n[??]]]; eauto.
    right; exists n; splits; auto.
    eapply cstep_iteration; eauto.
  Qed.

(*
  Lemma cstep_star_while_max_skip_resolve_stalled: forall s C w e i n z1 t s' C' w' z,
    cstep_star s C (w: tm_while_max e (i_stall i n z1) t) s' C' (w': tm_skip) ->
    s_iter_get s' i <> Some iter_go ->
    evalFalse e w' /\ cstep_star s C (w: tm_while_max e (i_plus n z) t) s' C' (w': tm_skip).
  Proof.
    intros.
    remember (w: tm_while_max e (i_stall i n z1) t)%wterm as b.
    remember (w: tm_while_max e (i_plus n z) t)%wterm as bp.
    assert (exists z' n', z'=i_stall i n' z1 /\ n'<=n /\ ((exists w, b = w: tm_while_max e z' t /\ bp = w: tm_while_max e (i_plus n' z) t) \/ (exists bt, b = bt; tm_while_max e z' t /\ bp = bt; tm_while_max e (i_plus n' z) t)))%wterm.
      subst; exists (i_stall i n z1) n; split; jauto; apply ids_nil.
    clear w Heqb Heqbp.
    destruct H1 as [z'[n'[?[??]]]].
    revert z' n' bp H0 H1 H2 H3.
    remember (w': tm_skip)%wterm as b'.
    revert w' Heqb'.
    induction H; intros; subst.

    destruct H3 as [[w[??]] | [b [??]]]; discriminate.

    edestruct cstep_inversion as [[?]|[?]]; eauto; clear H.
    subst.
    destruct H4 as [[w[??]] | [bt[??]]]; subst.
    inverts H2.
    inverts H10.
    edestruct IHcstep_star with (n':=n0) (bp:= (w:t;tm_while_max e (i_plus n0 z) t)%wterm) as [??]; clear IHcstep_star; eauto.
    omega.
    split; auto.
    eapply step_cstep_star; eauto.
    apply st_while_true; eauto.
    simpl. destruct z; constructor.
    assert (w' = w); subst.
      apply cstep_star_skip_resolve in H0.
      inverts H0; auto.
    split; auto.
    eapply cstep_star_app.
    apply cstep_star_skip_subst with w; eauto.
    apply step_single_cstep_star. apply st_while_false; auto.

clear IHcstep_star.
    edestruct cstep_star_while_max_skip_resolve' as [? | [n'[??]]]; eauto.
    cstep_star s C (w: tm_while_max e z t) s' C' (w': tm_skip) ->
    evalFalse e w' \/
    (exists (n:nat), z=n /\ cstep_star s C (w: tm_while_max e z t) s' C' (w': tm_while_max e 0 t)).
    
    edestruct IHcstep_star with (n':=0) (w'0:=w') (bp:= (w: tm_while_max e (i_plus 0 z) t)%wterm) as [??]; clear IHcstep_star.
eauto. auto.

; eauto.
    left; eauto.


    false; clear -H12 H1 H0.
    apply cstep_star_s_iter_preservation with (i:=i) in H0; auto.
    inverts H2.
    edestruct IHcstep_star with (n'0:=n') (bp:= (b1';tm_while_max e (i_plus n' z) t)%wterm) as [??]; clear IHcstep_star; eauto.
    split; auto.
    eapply step_cstep_star; eauto.
    apply st_seq1; auto.
    edestruct IHcstep_star with (n'0:=n') (bp:= (w: tm_while_max e (i_plus n' z) t)%wterm) as [??]; clear IHcstep_star; eauto.
    split; auto.
    eapply step_cstep_star; eauto.
    apply st_seq_skip; auto.

    subst b'.
    edestruct IHcstep_star as [??]; clear IHcstep_star; eauto.
    split; auto.
    eapply cstep_context_cstep_star; eauto.

    edestruct IHcstep_star as [??]; clear IHcstep_star; eauto.
    split; auto.
    eapply cstep_iteration; eauto.
  Qed.
*)
  Lemma cstep_s_iter: forall s C b s' C' b',
    cstep s C b s' C' b' -> s_i s = s_i s'.
  Proof.
    intros.
    apply step_simulates_cstep in H.
    eapply step_s_iter; eauto.
  Qed.


  Lemma ss_iter_cstep_snoc: forall s s' s'' C b C' b',
    ss_iter_star s s' ->
    cstep s' C b s'' C' b' ->
    ss_iter_star s s''.
  Proof.
    intros.
    unfold ss_iter_star in *.
    rewrite<- (cstep_s_iter s' C b s'' C' b'); eauto.
  Qed.

  Lemma cstep_go_preservation: forall s C b s' C' b' i,
    s_iter_get s i = Some iter_go ->
    cstep s C b s' C' b' ->
    s_iter_get s' i = Some iter_go.
  Proof.
    intros.
    unfold s_iter_get in *.
    erewrite<- cstep_s_iter; eauto.
  Qed.

  Lemma ss_iter_cstep_cons : forall s C b s' C' b' s'',
    cstep s C b s' C' b' ->
    ss_iter_star s' s'' ->
    ss_iter_star s s''.
  Proof.
    intros.
    unfold ss_iter_star in *.
    rewrite (cstep_s_iter s C b s' C' b') in *; eauto.
  Qed.

  Lemma ss_iter_star_cstep_star: forall s C b s' C' b',
    cstep_star s C b s' C' b' ->
    ss_iter_star s s'.
  Proof.
    intros.
    induction H.
    apply s_iter_nil.
    eapply ss_iter_cstep_cons; eauto.
    eapply ss_iter_cons; eauto.
  Qed.

  Lemma cstep_star_s_iter_set: forall s C b s',
    s_iter_star s s' ->
    cstep_star s C b s' C b.
  Proof.
    intros.
    induction H.
    apply cstep_nil.
    eapply cstep_iteration; eauto.
  Qed.


  Lemma cstep_star_s_iter_preservation: forall i s C b s'' C'' b'',
    s_iter_get s i = Some iter_go ->
    cstep_star s C b s'' C'' b'' ->
    s_iter_get s'' i = Some iter_go.
  Proof.
    intros.
    induction H0; auto.
    apply IHcstep_star.
    unfold s_iter_get in *.
    erewrite<- cstep_s_iter; eauto.
    apply IHcstep_star.
    inverts H0.
    destruct (Pos.eq_dec i0 i); subst.
    apply s_iter_get_set_eq.
    rewrite s_iter_get_set_neq; auto.
  Qed.

  Lemma cstep_star_while_max_skip_resolve_stall: forall s C w e i n z1 t s' C' w',
    cstep_star s C (w: tm_while_max e (i_stall i n z1) t) s' C' (w': tm_skip) ->
    evalFalse e w' \/
    (exists (n1:nat), s_iter_get s' i = Some iter_go /\ z1=n1 /\ cstep_star s C (w: tm_while_max e (i_stall i n z1) t) s' C' (w': tm_while_max e 0 t)).
  Proof.
    intros.
    remember (w: tm_while_max e (i_stall i n z1) t)%wterm as b.
    assert (exists z', ((exists n', z'=i_stall i n' z1 /\ n'<=n) \/ (s_iter_get s i = Some iter_go /\ iteration_decrs z1 z')) /\ ((exists w, b = w: tm_while_max e z' t) \/ (exists bt, b = bt;; tm_while_max e z' t)))%wterm.
      subst; exists (i_stall i n z1); split; jauto; apply ids_nil.
    clear w Heqb.
    destruct H0 as [z'[??]].
    revert z' H1 H0.
    remember (w': tm_skip)%wterm as b'.
    revert w' Heqb'.
    induction H; intros; subst.

    destruct H1 as [[w ?] | [b ?]]; discriminate.

    edestruct cstep_inversion as [[?]|[?]]; eauto; clear H.
    subst.
    destruct H1 as [[w ?] | [bt ?]]; subst b.
    inverts H3.
    edestruct IHcstep_star as [? | [n'[?[??]]]]; clear IHcstep_star; eauto.
    destruct H2 as [[n'[??]] | [??]].
      subst z'. inverts H9. left; exists n0; split; auto. omega. 
      right; split; auto. eapply ids_decr; eauto.
    right; exists n'; splits; auto.
    eapply cstep_star_app; eauto.
    eapply step_single_cstep_star; eauto.
    apply st_while_true; eauto.
    clear IHcstep_star.
    assert (w' = w); subst.
      apply cstep_star_skip_resolve in H0.
      inverts H0; auto.
    left; auto.
    edestruct IHcstep_star as [? | [n'[?[??]]]]; clear IHcstep_star; eauto.
    destruct H2 as [[n'[??]] | [??]].
    inverts H. inverts H1.
    right; split; auto; apply ids_nil.
    right; split; auto; apply ids_nil.
    false; clear -H1.
      destruct z1.
      edestruct iter_decrs_nat1 as [n'[??]]; eauto; discriminate.
      apply iter_decrs_inf1 in H1; discriminate.
    subst z1.
    right; exists n'; splits; auto.
    eapply step_cstep_star; eauto.
    apply st_while_go; eauto.
    destruct H2 as [[n'[??]] | [??]].
    discriminate.
    edestruct iter_decrs_nat2 as [n1[??]]; eauto.
    inverts H1.
    replace w' with w in *.
    2: apply cstep_star_skip_resolve in H0; inverts H0; auto.
    inverts H2.
    right; exists 0; splits; auto.
    eapply cstep_star_s_iter_preservation; eauto.
    apply cstep_star_skip_subst with w; auto.
    inverts H2.
    clear IHcstep_star.
    edestruct iter_decrs_nat1 as [n'[??]]; eauto.
    subst z'.
    right; exists n1; splits; auto.
    eapply cstep_star_s_iter_preservation; eauto.
    replace w' with w in *.
    2: apply cstep_star_skip_resolve in H0; inverts H0; auto.
    apply cstep_star_skip_subst with w; auto.
    inverts H3.
    edestruct IHcstep_star as [? | [n'[?[??]]]]; clear IHcstep_star; eauto.
    destruct H2 as [[n'[??]] | [??]].
    subst z'.
    left; eauto.
    right; split; auto.
    unfold s_iter_get in *.
    erewrite<- step_s_iter; eauto.
    
    subst z1.
    right; exists n'; splits; auto.
    eapply step_cstep_star; eauto.
    apply st_seq1; auto.
    edestruct IHcstep_star as [? | [n'[?[??]]]]; clear IHcstep_star; eauto.
    subst z1.
    right; exists n'; splits; auto.
    eapply step_cstep_star; eauto.
    apply st_seq_skip; auto.
    subst b'.
    edestruct IHcstep_star as [? | [n'[?[??]]]]; clear IHcstep_star; eauto.
    destruct H2 as [[n'[??]] | [??]].
    subst z'.
    left; eauto.
    right; split; auto.
    unfold s_iter_get in *.
    erewrite<- cstep_s_iter; eauto.
    subst z1.
    right; exists n'; splits; auto.
    eapply cstep_context_cstep_star; eauto.
    edestruct IHcstep_star as [? | [n'[?[??]]]]; clear IHcstep_star; eauto.
    destruct H2 as [[n'[??]] | [??]].
    subst z'.
    left; eauto.
    right; split; auto.
    inverts H.
    destruct (Pos.eq_dec i0 i); subst.
    rewrite s_iter_get_set_eq; auto.
    rewrite s_iter_get_set_neq; auto.
    subst z1.
    right; exists n'; splits; auto.
    eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_star_while_max_skip_resolve: forall s C w e z t s' C' w',
    cstep_star s C (w: tm_while_max e z t) s' C' (w': tm_skip) ->
    evalFalse e w' \/
    (exists (n:nat), z=n /\ cstep_star s C (w: tm_while_max e z t) s' C' (w': tm_while_max e 0 t)) \/
    (exists i, exists (n n':nat), s_iter_get s' i = Some iter_go /\ z=(i_stall i n n') /\ cstep_star s C (w: tm_while_max e z t) s' C' (w': tm_while_max e 0 t)).
  Proof.
    intros.
    destruct z.
    edestruct cstep_star_while_max_skip_resolve_stall as [? | [n'[?[??]]]]; eauto.
    subst i.
    right; right; jauto.
    edestruct cstep_star_while_max_skip_resolve' as [? | [n'[??]]]; eauto.
    subst i.
    right; left; jauto.
  Qed.


  Lemma cstep_star_while_resolve': forall s C w e z t s'' C'' b'',
    cstep_star s C (w: tm_while_max e (iter_const z) t) s'' C'' b'' ->
    (cstep_star s C (gwemp:tm_skip) s'' C'' (gwemp:tm_skip) /\ b''=w:tm_while_max e z t)%wterm
    \/
    (exists (z':iterations),
      iteration_decrs z z' /\
      ((exists w'', b'' = w'': tm_while_max e z' t) \/ exists bt'', b'' = bt'';; tm_while_max e z' t))%wterm
    \/
    (exists w'', b'' = w'':tm_skip)%wterm.
  Proof.
    intros.
    remember (w:tm_while_max e z t)%wterm as bb.
    assert (exists z', iteration_decrs z z' /\ ((exists b, bb = b;; tm_while_max e z' t) \/ (exists w, bb=w:tm_while_max e z' t)))%wterm.
      exists z; split; jauto. apply ids_nil.
    clear Heqbb.
    clear w.
    revert H0.
    induction H; intros; subst.

    left; split; auto. apply cstep_nil.

    rename s'' into s'''; rename C'' into C'''; rename b'' into b'''.
    edestruct cstep_inversion as [[?]|[?]]; eauto. subst.
    destruct H1 as [z' [? [[bb ?] | [w ?]]]]; subst.
    rename bb into b.
    inverts H2.

    edestruct IHcstep_star as [[??]| [[z''[??]] | [w'' ?]]]; clear IHcstep_star; eauto; subst.
    right; left.
      destruct z.
      edestruct iter_decrs_nat1 as [n'[??]]; eauto.
      subst z'. exists n'; split; eauto.
      apply iter_decrs_inf1 in H1.
      subst z'. exists i_inf; split; eauto. apply ids_nil.

    edestruct IHcstep_star as [[??]| [[z''[??]] | [w'' ?]]]; clear IHcstep_star; eauto; subst.
    right; left.
      destruct z.
      edestruct iter_decrs_nat1 as [n'[??]]; eauto.
      subst z'. exists n'; split; eauto.
      apply iter_decrs_inf1 in H1.
      subst z'. exists i_inf; split; eauto. apply ids_nil.

    inverts H2.
    edestruct IHcstep_star as [[??]| [[z''[??]] | [w'' ?]]]; clear IHcstep_star; eauto; subst.
    exists z'0; split; eauto; eapply ids_decr; eauto.
    right; left.
    inverts H10.
    exists n; split; eauto. eapply ids_decr; eauto. constructor.
    exists i_inf; split; eauto.
    false; clear -H1.
      destruct z.
      edestruct iter_decrs_nat1 as [n'[??]]; eauto; discriminate.
      apply iter_decrs_inf1 in H1; discriminate.
    clear IHcstep_star.
    right; right; exists w.
    eapply cstep_star_skip_resolve; eauto.
    false; clear -H1.
      destruct z.
      edestruct iter_decrs_nat1 as [n'[??]]; eauto; discriminate.
      apply iter_decrs_inf1 in H1; discriminate.
    clear IHcstep_star.
    right; right; exists w.
    eapply cstep_star_skip_resolve; eauto.

    subst b'.
    edestruct IHcstep_star as [[??]| [[z''[??]] | [w'' ?]]]; clear IHcstep_star; eauto; subst.
    left; split; auto. eapply cstep_cons; eauto.

    edestruct IHcstep_star as [[??]| [[z''[??]] | [w'' ?]]]; clear IHcstep_star; eauto; subst.
    left; split; auto. eapply cstep_iteration; eauto.
  Qed.



  Lemma cstep_star_while_resolve_stall: forall s C w e i n z1 t s'' C'' b'',
    cstep_star s C (w: tm_while_max e (i_stall i n z1) t) s'' C'' b'' ->
    (cstep_star s C (gwemp:tm_skip) s'' C'' (gwemp:tm_skip) /\ b''=w:tm_while_max e (i_stall i n z1) t)%wterm
    \/
    (exists z',
      (iteration_decrs (i_stall i n z1) z' \/ (s_iter_get s'' i = Some iter_go /\ iteration_decrs z1 z')) /\
      ((exists w'', b'' = w'': tm_while_max e z' t) \/ exists bt'', b'' = bt'';; tm_while_max e z' t))%wterm
    \/
    (exists w'', b'' = w'':tm_skip)%wterm.
  Proof.
    intros.
    remember (w:tm_while_max e (i_stall i n z1) t)%wterm as bb.
    assert (exists z', iteration_decrs (i_stall i n z1) z' /\ ((exists b, bb = b;; tm_while_max e z' t) \/ (exists w, bb=w:tm_while_max e z' t)))%wterm.
      exists (i_stall i n z1); split; jauto. left; apply ids_nil.
    clear Heqbb.
    clear w.
    revert H0.
    induction H; intros; subst.

    left; split; auto. apply cstep_nil.

    rename s'' into s'''; rename C'' into C'''; rename b'' into b'''.
    edestruct cstep_inversion as [[?]|[?]]; eauto. subst.
    destruct H1 as [z'[?[[bb ?]|[w ?]]]]; subst.
    rename bb into b.
    inverts H2.

    edestruct IHcstep_star as [[??]| [[z''[??]] | [w'' ?]]]; clear IHcstep_star; eauto; subst.
    right; left.
    exists z'; split; eauto.

    edestruct IHcstep_star as [[??]| [[z''[??]] | [w'' ?]]]; clear IHcstep_star; eauto; subst.
    right; left.
    exists z'; split; eauto.

    inverts H2.
    edestruct IHcstep_star as [[??]| [[z''[??]] | [w'' ?]]]; clear IHcstep_star; eauto; subst.
    exists z'0; split; eauto; eapply ids_decr; eauto.
    right; left.
    exists z'0; split; eauto. left; eapply ids_decr; eauto.
    right; right; exists w.
    eapply cstep_star_skip_resolve; eauto.
    clear IHcstep_star.
    edestruct iter_decrs_stall2 as [n'[??]]; eauto.
    symmetry in H2; inverts H2.
    edestruct cstep_star_while_resolve' as [[??] | [[z'[??]] | [w'' ?]]]; eauto.
    subst.
    right; left; exists z1. split; eauto.
    right; split; auto.
    eapply cstep_star_s_iter_preservation; eauto.
    apply ids_nil.
    right; left; exists z'; split; eauto.
    right; split; auto.
    eapply cstep_star_s_iter_preservation; eauto.
    false; clear -H1.
      edestruct iter_decrs_stall1 as [n'[??]]; eauto; discriminate.

    subst b'.
    edestruct IHcstep_star as [[??]| [[z''[??]] | [w'' ?]]]; clear IHcstep_star; eauto; subst.
    left; split; auto. eapply cstep_cons; eauto.

    edestruct IHcstep_star as [[??]| [[z''[??]] | [w'' ?]]]; clear IHcstep_star; eauto; subst.
    left; split; auto. eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_star_while_resolve: forall s C w e z t s' C' b',
    cstep_star s C (w: tm_while_max e z t) s' C' b' ->
    (cstep_star s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) /\ b'=w:tm_while_max e z t)%wterm
    \/
    (exists z',
      (iteration_decrs z z' \/ (exists i n z1, z=i_stall i n z1 /\ s_iter_get s' i = Some iter_go /\ iteration_decrs z1 z')) /\
      ((exists w', b' = w': tm_while_max e z' t) \/ exists bt', b' = bt';; tm_while_max e z' t))%wterm
    \/
    (exists w', b' = w':tm_skip)%wterm.
  Proof.
    intros.
    destruct z.    
    edestruct cstep_star_while_resolve_stall as [[??]| [[z'[??]] | [w' ?]]]; eauto.
    right; left; exists z'; split; auto.
    destruct H0 as [? | [??]]; auto.
    right.
    exists p n i; split; auto.
    edestruct cstep_star_while_resolve' as [[??]| [[z'[??]] | [w' ?]]]; eauto.
    right; left; exists z'; split; auto.
  Qed.



  Lemma cstep_star_assert_resolve: forall s C s' C' P w b',
    cstep_star s C (w : tm_assert P) s' C' b' ->
    (b'=w:tm_assert P \/ (b'=w:tm_skip /\ w|=gsep_con P TT))%wterm.
  Proof.
    intros.
    remember (w: tm_assert P)%wterm as b.
    revert Heqb.
    induction H; intros; subst; eauto.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst.
    inverts H1; auto.
    apply cstep_star_skip_resolve in H0; subst b''; eauto.
    subst.
    eauto.
  Qed.

  Lemma cstep_star_assert_skip_resolve: forall s C b P s' C' w',
    cstep_star s C (b;; tm_assert P) s' C' (w':tm_skip) ->
    w' |= gsep_con P TT.
  Proof.
    intros.
    edestruct cstep_star_seq_inversion as [[s'0[C'0[w'0[??]]]] | [bx'[??]]]; eauto.
    edestruct cstep_star_assert_resolve as [?|[??]]; eauto.
    discriminate.
    inversion H2; subst; auto.
    discriminate.
  Qed.












  Lemma cstep_star_if_resolve: forall s C w e t1 t2 s' C' b',
    cstep_star s C (w: tm_if e t1 t2)%wterm s' C' b' ->
    (cstep_star s C (gwemp: tm_skip) s' C' (gwemp:tm_skip) /\ b'=(w:tm_if e t1 t2))%wterm \/
    (evalTrue e w /\ cstep_star s C (w: t1) s' C' b')%wterm \/
    (evalFalse e w /\ cstep_star s C (w: t2) s' C' b')%wterm.
  Proof.
    intros.
    remember (w: tm_if e t1 t2)%wterm as b.
    revert Heqb.
    induction H; intros; subst.
    left; split; auto; apply cstep_nil.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1; eauto.
    subst b'.
    destruct IHcstep_star as [[??] | [[??] | [??]]]; auto.
    left; split; auto. eapply cstep_cons; eauto.
    right; left; split; auto. eapply cstep_cons; eauto. eapply cstep_skip_subst; eauto.
    right; right; split; auto. eapply cstep_cons; eauto. eapply cstep_skip_subst; eauto.
    destruct IHcstep_star as [[??] | [[??] | [??]]]; auto.
    left; split; auto. eapply cstep_iteration; eauto.
    right; left; split; auto. eapply cstep_iteration; eauto.
    right; right; split; auto. eapply cstep_iteration; eauto.
  Qed.


  Lemma cstep_s_iter_conflict: forall s C b s' C' b',
    s_iter s s' -> cstep s C b s' C' b' -> False.
  Proof.
    intros.
    eapply step_s_iter_conflict; eauto.
    apply step_simulates_cstep; eauto.
  Qed.

  Lemma cstep_s_iter_resolve: forall s C b s' C' b',
    s_iter s s' ->
    cstep s C b s' C' b' ->
    cstep s' C b s' C' b'.
  Proof.
    intros.
    revert H.
    cstep_cases (induction H0) Case HH; intros;
      try solve [eapply HH; eauto].
    false; eapply step_s_iter_conflict; eauto.
    false; eapply step_s_iter_conflict; eauto.
    false; eapply step_s_iter_conflict; eauto.
  Qed.

  Lemma cstep_respecting_s_i: forall s C b s' C' b',
    cstep s C b s' C' b' -> s_i s = s_i s'.
  Proof.
    intros.
    eapply step_respecting_s_i; eauto.
    apply step_simulates_cstep; eauto.
  Qed.

  Lemma cstep_star_si_respecting_s_i: forall s C b s' C' b',
    cstep_star_sans_s_iter s C b s' C' b' -> s_i s = s_i s'.
  Proof.
    intros.
    induction H; auto.
    rewrite<- IHcstep_star_sans_s_iter.
    eapply cstep_respecting_s_i; eauto.
  Qed.

  Lemma cstep_star_si_cstep_star: forall s C b s' C' b',
    cstep_star_sans_s_iter s C b s' C' b' ->
    cstep_star s C b s' C' b'.
  Proof.
    intros.
    induction H.
    apply cstep_nil.
    eapply cstep_cons; eauto.
  Qed.

  Lemma cstep_star_s_iter_resolve: forall s C b s' C' b',
    s_iter s s' ->
    cstep_star s C b s' C' b' ->
    cstep_star s' C b s' C' b'.
  Proof.
    intros.
    edestruct cstep_star_commute_front_s_iter_star as [s0[??]]; eauto.
    clear H0.
    cut (s0 = s'); intros; subst.
    apply cstep_star_si_cstep_star; auto.
    destruct s0; destruct s'.
    simpl in *; f_equal.
    apply s_iter_respects_s_c in H.
    apply s_iter_star_respects_s_c in H1.
    simpl in *; congruence.
    apply cstep_star_si_respecting_s_i in H2.
    simpl in *; congruence.
    apply s_iter_respects_s_all_res in H.
    apply s_iter_star_respects_s_all_res in H1.
    simpl in *; congruence.
  Qed.

  Lemma cstep_par_inversion1: forall s C b1 b2 s' C' b12',
    cstep s C (b1||b2) s' C' b12' ->
    (exists b1', cstep s C b1 s' C' b1' /\ b12' = b1'||b2) \/ (exists b2', step s b2 s' b2' /\ b12' = b1 || b2' /\ C'=C) \/ (exists w1 w2 w, gjoin w1 w2 w /\ b1=w1:tm_skip /\ b2=w2:tm_skip /\ b12'=w:tm_skip /\ C'=C /\ s'=s).
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto; clear H.
    subst C'.
    inverts H0.
    left; exists b1'; split; auto.
    apply cstep_simulates_step; apply step_add_context; auto.
    right; left; exists b2'; split; auto.
    right; right; exists w1 w2 w; splits; auto.
    subst b12'.
    left; exists b1; split; auto.
    apply cstep_skip_subst with gwemp; auto.
  Qed.

  Lemma cstep_par_resolve': forall s C b1 b2 s' C' b12',
    cstep s C (b1||b2) s' C' b12' ->
    (exists b1' b2', b12' = b1'||b2') \/ (exists w1 w2 w, gjoin w1 w2 w /\ b1=w1:tm_skip /\ b2=w2:tm_skip /\ b12'=w:tm_skip).
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H0; eauto.
    right; exists w1 w2 w; splits; auto.
  Qed.

  Lemma cstep_star_iter_snoc: forall s C b s' C' b' s'',
    cstep_star s C b s' C' b' -> s_iter s' s'' -> cstep_star s C b s'' C' b'.
  Proof.
    intros.
    induction H.
    eapply cstep_iteration; eauto. apply cstep_nil.
    eapply cstep_cons; eauto.
    eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_composed_context_step1': forall s C1 C2 b s' C1' b',
    cstep s C1 b s' C1' b' ->
    cstep s (compose_contexts C1 C2) b s' (compose_contexts C1' C2) b'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C1'.
    apply cstep_simulates_step; apply step_add_context; auto.
    subst b'.
    apply cstep_composed_context_step1 with gwemp; auto.
  Qed.

  Lemma cstep_composed_context_step2': forall s C1 C2 b s' C2' b',
    cstep s C2 b s' C2' b' ->
    cstep s (compose_contexts C1 C2) b s' (compose_contexts C1 C2') b'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C2'.
    apply cstep_simulates_step; apply step_add_context; auto.
    subst b'.
    apply cstep_composed_context_step2 with gwemp; auto.
  Qed.

  Lemma cstep_seq_inversion: forall s C b t s' C' bt',
    cstep s C (b;; t) s' C' bt' ->
    ((exists b', cstep s C b s' C' b' /\ bt'=b';;t) \/ (exists w, b=w:tm_skip /\ C'=C /\ s'=s /\ bt'=w:t)).
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H0.
    left; exists b1'; split; auto.
      apply cstep_simulates_step; apply step_add_context; auto.
    right; exists w; split; auto.
    left; exists b; split; auto.
      apply cstep_skip_subst with gwemp; auto.
  Qed.

  Lemma cstep_star_par_skip_resolve: forall s C b1 w2 s' C' b',
    cstep_star s C (b1||w2:tm_skip) s' C' b' ->
    exists b1',
      cstep_star s C b1 s' C' b1' /\ (b' = b1'||w2:tm_skip \/ (exists w1' w', b1'=w1':tm_skip /\ gjoin w1' w2 w' /\ b'=w':tm_skip)).
  Proof.
    intros.
    remember (b1||w2:tm_skip) as b.
    revert b1 Heqb.
    induction H; intros; subst.
    exists b1; split; auto.
      apply cstep_nil.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1.
    edestruct IHcstep_star as [b1''[?[?| [w1''[w''[?[??]]]] ]]]; eauto.
    subst b''.
    exists b1''; split; auto.
      eapply step_cstep_star; eauto.
    subst b1'' b''.
    exists (w1'':tm_skip); split.
      eapply step_cstep_star; eauto.
      right; exists w1'' w''; splits; auto.
    false; inverts H6.
    assert (b''=w:tm_skip).
      eapply cstep_star_skip_resolve; eauto.
    subst b''.
    exists (w1:tm_skip); split.
      apply cstep_star_skip_subst with w; auto.
      right; exists w1 w; splits; auto.
    subst b'.
    edestruct IHcstep_star as [b1''[?[?| [w1''[w''[?[??]]]] ]]]; clear IHcstep_star; eauto.
    subst b''.
    exists b1''; split; auto.
      eapply cstep_context_cstep_star; eauto.
    subst b1'' b''.
    exists (w1'':tm_skip); split.
      eapply cstep_context_cstep_star; eauto.
      right; exists w1'' w''; splits; auto.

    edestruct IHcstep_star as [b1''[?[?| [w1''[w''[?[??]]]] ]]]; clear IHcstep_star; eauto.
    subst b''.
    exists b1''; split; auto.
      eapply cstep_iteration; eauto.
    subst b1'' b''.
    exists (w1'':tm_skip); split.
      eapply cstep_iteration; eauto.
      right; exists w1'' w''; splits; auto.
  Qed.

  Lemma cstep_star_step_inversion: forall s C b s'' C'' b'',
    cstep_star s C b s'' C'' b'' ->
    (exists s' C' s'0 b'0,
      cstep_star s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) /\ step s' b s'0 b'0 /\ cstep_star s'0 C' b'0 s'' C'' b'' /\ b'0<>b) \/
    (b'' = b /\ cstep_star s C (gwemp:tm_skip) s'' C'' (gwemp:tm_skip)).
  Proof.
    intros.
    induction H.

    right; split; auto; apply cstep_nil.

    edestruct cstep_inversion as [[?]|[?]]; eauto; clear H.
    subst C'.
    destruct (wterm_eq_dec b' b). subst b'.
    apply step_spin_no_change in H1. subst s'.
    destruct IHcstep_star as [ [s'0[C'0[s'1[b'1[?[?[??]]]]]]] | [??] ].
    left; do 4 eexists; splits; eauto.
    subst b''.
    right; split; auto; eapply cstep_cons; eauto.
    left; do 4 eexists; splits; eauto; apply cstep_nil.
    subst b'.
    destruct IHcstep_star as [ [s'0[C'0[s'1[b'1[?[?[??]]]]]]] | [??] ].
    left; do 4 eexists; splits; eauto.
      eapply cstep_cons; eauto.
    subst b''.
    right; split; auto; eapply cstep_cons; eauto.

    destruct IHcstep_star as [ [s'0[C'0[s'1[b'1[?[?[??]]]]]]] | [??] ].
    left; do 4 eexists; splits; eauto.
      eapply cstep_iteration; eauto.
    subst b''.
    right; split; auto; eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_star_tm_seq_resolve: forall s C w t1 t2 s' C' b',
    cstep_star s C (w: tm_seq t1 t2) s' C' b' ->
    (cstep_star s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) /\ b'=(w: tm_seq t1 t2)%wterm) \/ cstep_star s C (w: t1;; t2) s' C' b'.
  Proof.
    intros.
    remember (w: (t1;;;t2)%term)%wterm as b.
    revert Heqb.
    induction H; intros; subst.
    left; split; auto; apply cstep_nil.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1.
    right; auto.
    subst b'.
    destruct IHcstep_star as [[??] | ?]; auto; subst.
    left; split; auto. eapply cstep_cons; eauto.
    right; eapply cstep_context_cstep_star; eauto.
    destruct IHcstep_star as [[??] | ?]; auto; subst.
    left; split; auto. eapply cstep_iteration; eauto.
    right; eapply cstep_iteration; eauto.
  Qed.


  Lemma cstep_par_cx_par1: forall s C b1 b2 s' C' b1' b2',
    cstep s C (b1||b2) s' C' (b1'||b2') ->
    cstep s (cx_par1 C b2) b1 s' (cx_par1 C' b2') b1'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H0.
    apply step_cstep; auto.
    apply cst_par1_step2; auto.
    inverts H1.
    apply cst_par1; auto.
    apply cstep_skip_subst with gwemp; auto.
  Qed.

  Lemma cstep_cx_par1_par: forall s C b1 b2 s' C' b1' b2',
    cstep s (cx_par1 C b2) b1 s' (cx_par1 C' b2') b1' ->
    cstep s C (b1||b2) s' C' (b1'||b2').
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    inverts H1.
    apply step_cstep; apply st_par1; auto.
    subst b1'.
    clear H.
    inverts H0.
    apply cstep_skip_subst with gwemp; auto.
    apply step_cstep; apply st_par2; auto.
    simpl in H6.
    inverts H6.
  Qed.

  Lemma cstep_par1_step: forall s C b1 b2 s' C' b1',
    cstep s C b1 s' C' b1' ->
    cstep s C (b1||b2) s' C' (b1'||b2).
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    apply step_cstep; apply st_par1; auto.
    subst b1'.
    apply cstep_skip_subst with gwemp; auto.
  Qed.

  Lemma cstep_par2_step: forall s C b1 b2 s' C' b2',
    cstep s C b2 s' C' b2' ->
    cstep s C (b1||b2) s' C' (b1||b2').
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    apply step_cstep; apply st_par2; auto.
    subst b2'.
    apply cstep_skip_subst with gwemp; auto.
  Qed.

  Lemma cstep_par_swap: forall s C b1 b2 s' C' b1' b2',
    cstep s C (b1||b2) s' C' (b1'||b2') ->
    cstep s C (b2||b1) s' C' (b2'||b1').
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H0; apply step_cstep; [apply st_par2 | apply st_par1]; auto.
    inverts H1.
    apply cstep_skip_subst with gwemp; auto.
  Qed.

  Lemma cstep_star_par_assoc: forall s C b1 b2 b3 s' C' b1' b2' b3',
    cstep_star s C (b1 || (b2 || b3)) s' C' (b1' || (b2' || b3')) <->
    cstep_star s C (b1 || b2 || b3) s' C' (b1' || b2' || b3').
  Proof.
    split; intros.
    remember (b1 || (_ || _))%wterm as b.
    remember (b1' || (_ || _))%wterm as b'.
    revert b1 b2 b3 b1' b2' b3' Heqb Heqb'.
    induction H; intros; subst.
    inverts Heqb'. apply cstep_nil.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1.
    eapply step_cstep_star; eauto. apply st_par1; apply st_par1; auto.
    inverts H6.
    eapply step_cstep_star; eauto. apply st_par1; apply st_par2; auto.
    eapply step_cstep_star; eauto. apply st_par2; auto.
    false.
    apply cstep_star_par_cx_par2 in H0.
    apply cstep_star_skip_resolve in H0.
    discriminate.
    subst b'.
    eapply cstep_context_cstep_star; eauto.
    eapply cstep_iteration; eauto.

    remember (b1 || _ || _)%wterm as b.
    remember (b1' || _ || _)%wterm as b'.
    revert b1 b2 b3 b1' b2' b3' Heqb Heqb'.
    induction H; intros; subst.
    inverts Heqb'. apply cstep_nil.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H1.
    inverts H6.
    eapply step_cstep_star; eauto. apply st_par1; auto.
    eapply step_cstep_star; eauto. apply st_par2; apply st_par1; auto.
    false.
    apply cstep_star_par_cx_par1 in H0.
    apply cstep_star_skip_resolve in H0.
    discriminate.
    eapply step_cstep_star; eauto. apply st_par2; apply st_par2; auto.
    subst b'.
    eapply cstep_context_cstep_star; eauto.
    eapply cstep_iteration; eauto.
  Qed.

  Lemma set_context_eq_context: forall s C s' C' b,
    cstep s C b s' C' b ->
    C[b] = C'[b] -> C=C'.
  Proof.
    intros.
    revert H0.
    induction H; intros; simpl in *; subst; auto; inverts H0; try rewrite IHcstep; auto.
  Qed.

  Lemma cstep_par_cx_par2: forall s C b1 b2 s' C' b1' b2',
    cstep s C (b1||b2) s' C' (b1'||b2') ->
    cstep s (cx_par2 b1 C) b2 s' (cx_par2 b1' C') b2'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H0.
    apply cst_par2_step1; auto.
    apply step_cstep; auto.
    inverts H1.
    apply cst_par2; auto.
    apply cstep_skip_subst with gwemp; auto.
  Qed.

  Lemma cstep_cx_par2_par: forall s C b1 b2 s' C' b1' b2',
    cstep s (cx_par2 b1 C) b2 s' (cx_par2 b1' C') b2' ->
    cstep s C (b1||b2) s' C' (b1'||b2').
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    inverts H1.
    apply step_cstep; apply st_par2; auto.
    subst b2'.
    clear H.
    inverts H0.
    apply cstep_skip_subst with gwemp; auto.
    apply step_cstep; apply st_par1; auto.
    simpl in H6.
    inverts H6.
  Qed.



  Lemma cstep_simulates_cstep_step: forall s1 s2 C b s1' s2' C' b',
    cstep s1 C b s1' C' b' ->
    step s2 C[b] s2' C'[b'] ->
    cstep s2 C b s2' C' b'.
  Proof.
    intros.
    revert s2 s2' H0.
    cstep_cases (induction H) Case HH; intros.
    - eapply HH; inverts H0; auto.
    - inverts H0; auto.
    apply set_context_eq_context in H; auto; subst C'.
    apply cst_par1_step2; auto.
    - inverts H0; auto.
    apply set_context_eq_context in H; auto; subst C'.
    apply cst_par2_step1; auto.
    - inverts H0; auto.
    apply cstep_par_cx_par1.
    apply cstep_par1_step.
    apply cstep_simulates_step; auto.
    - inverts H0; auto.
    apply cstep_par_cx_par2.
    apply cstep_par2_step.
    apply cstep_simulates_step; auto.
    - apply cstep_simulates_step; auto.
  Qed.


End Lemmas.