Require Import Machine.
Require Import ContextSemantics.
Require Import Bisimulation.
Require Import ParBisimulation.
Require Import world_model.
Require Import msl.msl_standard.
Require Import SfLib.
Require Import LibTactics.
Require Import IterationBounds.
Require Import StepStar.

Import world_model.Notation.
Import Bisimulation.Notation.
Import Bisimulation.Morphisms.
Import world_model.Notation.

Section Definitions.
  Inductive iter_schema : Type :=
  | istm_term: term->iter_schema
  | istm_loop_var: expr->nat->iter_schema->iter_schema
  | istm_while: expr->iteration_expr->iter_schema->iter_schema
  | istm_if: expr->iter_schema->iter_schema->iter_schema
  | istm_seq: iter_schema->iter_schema->iter_schema
  | istm_par: pred world->iter_schema->pred world->iter_schema->iter_schema
  | istm_local: expr->(var->iter_schema)->iter_schema.

  Inductive witer_schema : Type :=
  | iswtm_term: world->iter_schema->witer_schema
  | iswtm_seq: witer_schema->iter_schema->witer_schema
  | iswtm_par: witer_schema->witer_schema->witer_schema
  .


  Fixpoint iswtm_wterm (b:wterm) :=
  match b with
  | wtm_term w t => iswtm_term w (istm_term t)
  | wtm_seq b t => iswtm_seq (iswtm_wterm b) (istm_term t)
  | wtm_par b1 b2 => iswtm_par (iswtm_wterm b1) (iswtm_wterm b2)
  end.

  Definition iter_expr_minus (z:iteration_expr) (n:nat) :=
  match z with
  | i_nat n' => n'-n
  | i_inf => i_inf
  | i_stall i n' z => i_stall i (n'-n) (i_minus z (n-n'))
  end.

  Fixpoint term_subst_iter (t:iter_schema) (zs: iteration_expr) : term :=
    match t with
    | istm_term t => t
    | istm_loop_var e n t => tm_while_max e (iter_expr_minus zs n) (term_subst_iter t zs)
    | istm_while e z t => tm_while_max e z (term_subst_iter t zs)
    | istm_if e t1 t2 => tm_if e (term_subst_iter t1 zs) (term_subst_iter t2 zs)
    | istm_seq t1 t2 => tm_seq (term_subst_iter t1 zs) (term_subst_iter t2 zs)
    | istm_par P1 t1 P2 t2 => tm_par P1 (term_subst_iter t1 zs) P2 (term_subst_iter t2 zs)
    | istm_local e f => tm_local e (fun x => term_subst_iter (f x) zs)
    end.

  Definition iter_schema_app f zs:= term_subst_iter f zs.



(*
  Definition expr_equiv_dec: forall (e1 e2: expr), {e1=e2}+{e1<>e2}.
Admitted.
*)

  Open Scope forces_pred.

  Definition iter_expr_eq_dec: forall (z1 z2:iteration_expr), {z1=z2}+{z1<>z2}.
    intros.
    destruct z1; destruct z2.
    destruct (Pos.eq_dec p p0); subst.
    destruct (eq_nat_dec n n0); subst.
    destruct i; destruct i0; auto.
    destruct (eq_nat_dec n n1); subst; auto.
    right; intros H; inverts H; intuition.
    right; intros H; inverts H. 
    right; intros H; inverts H.
    right; intros H; inverts H; intuition.
    right; intros H; inverts H; intuition.
    right; intros H; inverts H.
    right; intros H; inverts H.
    destruct i; destruct i0; auto.
    destruct (eq_nat_dec n n0); subst; auto.
    right; intro H; inverts H; auto.
    right; intros H; inverts H. 
    right; intros H; inverts H.
  Qed.

  Import Machine.Notation.

  Fixpoint subst_iter (b:witer_schema) (z: iteration_expr) : wterm :=
    match b with
    | iswtm_term w t => w: term_subst_iter t z
    | iswtm_seq b1 t2 => subst_iter b1 z;; term_subst_iter t2 z
    | iswtm_par b1 b2 => subst_iter b1 z || subst_iter b2 z
    end.

  Lemma eq_dec_refl: forall p, Pos.eq_dec p p = left (eq_refl p).
  Proof.
    intros; auto. destruct (Pos.eq_dec p p); intuition. simpl.
    f_equal.
    apply proof_irr.
  Qed.

  Lemma eq_dec_neq_: forall p p' (H: p <> p'), Pos.eq_dec p p' = right H.
  Proof.
    intros; auto. destruct (Pos.eq_dec p p'); intuition.
    f_equal.
    apply proof_irr.
  Qed.
  Tactic Notation "eq_dec_neq" hyp(H):=
    repeat progress (first [erewrite (eq_dec_neq_ _ _ H) | erewrite (eq_dec_neq_ _ _ (not_eq_sym H)) ]).

  Fixpoint free_iter_var (f:iter_schema) (i:positive) : Prop :=
  match f with
    | istm_loop_var _ _ t1 => free_iter_var t1 i
    | istm_while _ (i_stall j _ _) t1 => i <> j /\ free_iter_var t1 i
    | istm_while e z t1 => free_iter_var t1 i
    | istm_if e t1 t2 => free_iter_var t1 i /\ free_iter_var t2 i
    | istm_seq t1 t2 =>free_iter_var t1 i /\ free_iter_var t2 i
    | istm_par P1 t1 P2 t2 => free_iter_var t1 i /\ free_iter_var t2 i
    | istm_term t1 => term_free_iter_var t1 i
    | istm_local e f => forall x, free_iter_var (f x) i
  end.

  Fixpoint wfree_iter_var (f:witer_schema) (i:positive) : Prop :=
  match f with
    | iswtm_term _ t => free_iter_var t i
    | iswtm_seq b1 t2 => wfree_iter_var b1 i /\ free_iter_var t2 i
    | iswtm_par b1 b2 => wfree_iter_var b1 i /\ wfree_iter_var b2 i
  end.

  Definition iteration_expr_gt (z:iteration_expr) n := match z with
  | i_nat m => m > n
  | i_stall i m z => m > n
  | i_inf => True
  end.

End Definitions.


Import Machine.Notation.

Module Notation.

  Coercion term_subst_iter : iter_schema >-> Funclass.
  Coercion subst_iter : witer_schema >-> Funclass.
  Coercion istm_term : term >-> iter_schema.
  Coercion iswtm_wterm : wterm >-> witer_schema.

End Notation.

Section SubstChoiceLemmas.
  Import Notation.

  Definition s_iter_schema_exists_fresh (s:state) (f:iter_schema): {i | s_iter_get s i = None /\ free_iter_var f i}. admit. Qed. 
  Definition s_iter_schema_context_exists_fresh (s:state) (f:iter_schema) (C:context) : {i | s_iter_get s i = None /\ free_iter_var f i /\ wterm_free_iter_var (set_context C (gwemp:tm_skip)) i}. admit. Qed. 


  Lemma iteration_decr_minus: forall z n,
    iteration_expr_gt z n ->
    iteration_decr (iter_expr_minus z n) (iter_expr_minus z (S n)).
  Proof.
    intros.
    destruct z; simpl in *.
    replace (n0-n) with (S (n0- S n)); [ | omega].
    destruct n0. false. omega.
    replace (n - S n0) with (n - n0).
    constructor.
    omega.
    destruct i.
    replace (n0-n) with (S (n0- S n)); [constructor | omega].
    constructor.
  Qed.
  Lemma iteration_expr_gt_nat : forall m n (z:iteration_expr),
    m > n ->
    iteration_expr_gt (iter_plus1 m z) n.
  Proof.
    intros.
    unfold iteration_expr_gt. simpl.
    destruct z. omega.
    destruct i; auto. omega.    
  Qed.

  Lemma i_minus_zero: forall z,
    i_minus z 0 = z.
  Proof.
    intros.
    destruct z; auto.
    rewrite i_minus_nat.
    f_equal; omega.
  Qed.

  Lemma iter_expr_minus_zero: forall z,
    iter_expr_minus z 0 = z.
  Proof.
    intros.
    destruct z; simpl; auto.
    f_equal; auto. omega. apply i_minus_zero.
    destruct i; auto. f_equal. f_equal. omega.
  Qed.

  Lemma term_subst_iter_skip: forall t zs,
    tm_skip = term_subst_iter t zs -> t = tm_skip.
  Proof.
    intros.
    destruct t; simpl in *; subst; auto; inverts H.
  Qed.

  Lemma subst_iter_skip: forall w b zs,
    w: tm_skip = subst_iter b zs -> b = w:tm_skip.
  Proof.
    intros.
    destruct b.
    rename i into t.
    simpl in *.
    inverts H.
    apply term_subst_iter_skip in H2; subst; auto.
    simpl in *.
    inverts H.
    inverts H.
  Qed.

  Lemma wterm_free_iter_var_step_preservation: forall i s b s' b',
    step s b s' b' ->
    wterm_free_iter_var b i ->
    wterm_free_iter_var b' i.
  Proof.
    intros.
    step_cases (induction H) Case; intros; simpl in *; try solve [inverts H1; intuition | intuition].
  Qed.

  Lemma wterm_free_iter_var_cstep_preservation: forall i s C b s' C' b',
    cstep s C b s' C' b' ->
    wterm_free_iter_var b i ->
    wterm_free_iter_var b' i.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    clear H.
    induction H1; simpl in *; try solve [inverts H1; intuition | intuition].
    subst b'.
    auto.
  Qed.

  Lemma wterm_free_iter_var_step_star_preservation: forall i s b s' b',
    step_star s b s' b' ->
    wterm_free_iter_var b i ->
    wterm_free_iter_var b' i.
  Proof.
    intros.
    revert H0.
    induction H; intros; auto.
    apply IHstep_star.
    eapply wterm_free_iter_var_step_preservation; eauto.
  Qed.

  Lemma wterm_free_iter_var_preservation: forall i s C b s' C' b',
    cstep_star s C b s' C' b' ->
    wterm_free_iter_var b i ->
    wterm_free_iter_var b' i.
  Proof.
    intros.
    revert H0.
    induction H; intros; auto.
    apply IHcstep_star.
    eapply wterm_free_iter_var_cstep_preservation; eauto.
  Qed.

  Lemma wfree_iter_var_wterm: forall (b:wterm) i,
    wfree_iter_var b i <-> wterm_free_iter_var b i.
  Proof. intros; induction b; simpl in *; intuition. Qed.

  Lemma term_subst_iter_id: forall (t:term) zs,
    term_subst_iter t zs = t.
  Proof. intros; reflexivity. Qed.

  Lemma subst_iter_id: forall (b:wterm) zs,
    subst_iter b zs = b.
  Proof.
    intros.
    induction b; simpl in *; auto.
    rewrite IHb; reflexivity.
    rewrite IHb1; rewrite IHb2; reflexivity.
  Qed.

  Lemma iter_expr_minus_iterations: forall (z1:iterations) n,
    iter_expr_minus z1 n = i_minus z1 n.
  Proof. intros; destruct z1; simpl; auto. Qed.

  Lemma step_term_inf_nat_iter: forall w t s s' ba',
    step s (w: term_subst_iter t i_inf) s' ba' ->
    (exists (n:nat) (b':witer_schema),
      ba' = subst_iter b' i_inf /\
      (forall j:positive, free_iter_var t j -> wfree_iter_var b' j) /\
      forall (z:iteration_expr) m, m>=n -> step s (w: term_subst_iter t (iter_plus1 m z)) s' (subst_iter b' (iter_plus1 m z))).
  Proof.
    intros.
    destruct t.

    exists 0 ba'; splits; intros; auto.
      rewrite subst_iter_id; reflexivity.
      apply wfree_iter_var_wterm.
      eapply wterm_free_iter_var_step_preservation; eauto.
      simpl. auto.
      rewrite subst_iter_id; auto.

    inverts H.
    inverts H7.
    exists (S n) (iswtm_seq (iswtm_term w t) (istm_loop_var e (S n) t)); splits; intros; auto.
      simpl; intuition.
      apply st_while_true; auto.
      apply iteration_decr_minus.
      apply iteration_expr_gt_nat. omega.
    exists 0 (w: tm_skip); splits; intros; auto.
      simpl; intuition.
      apply st_while_false; auto.

    simpl in *; inverts H.
    exists 0 (iswtm_seq (iswtm_term w t) (istm_while e z' t)); splits; intros; auto.
      simpl; destruct i; inverts H7; intuition.
      apply st_while_true; auto.
    exists 0 (w: tm_skip); splits; intros; auto.
      simpl; intuition.
      apply st_while_false; auto.
    exists 0 (iswtm_term w (istm_while e z t)); splits; intros; auto.
      simpl; intuition.
      apply st_while_go; auto.
    exists 0 (w: tm_skip); splits; intros; auto.
      simpl; intuition.
      apply st_while_limit; auto.

    simpl in *; inverts H.
    exists 0 (iswtm_term w t1); splits; intros; auto.
      simpl; intuition.
      apply st_if_true; auto.
    exists 0 (iswtm_term w t2); splits; intros; auto.
      simpl; intuition.
      apply st_if_false; auto.

    inverts H.
    exists 0 (iswtm_seq (iswtm_term w t1) t2); splits; intros; auto.
      simpl; intuition.
      apply st_seq; auto.

    inverts H.
    exists 0 (iswtm_par (iswtm_term w1 t1) (iswtm_term w2 t2)); splits; intros; auto.
      simpl; intuition.
      apply st_par; auto.

    rename i into f.
    inverts H.
    exists 0 (iswtm_term w' (f x)); splits; intros; auto.
      simpl; intuition.
    change ((iswtm_term w' (f x)) (iter_plus1 m z))
      with (w': (fun x => (f x) (iter_plus1 m z)) x).
    eapply st_local; eauto.
  Qed.

  Lemma step_inf_nat_iter: forall b s s' ba',
    step s (subst_iter b i_inf) s' ba' ->
    (exists (n:nat) (b':witer_schema),
      ba' = subst_iter b' i_inf /\
      (forall j:positive, wfree_iter_var b j -> wfree_iter_var b' j) /\
      forall (z:iteration_expr) m, m>=n -> step s (subst_iter b (iter_plus1 m z)) s' (subst_iter b' (iter_plus1 m z))).
  Proof.
    intros.
    revert ba' H.
    induction b; intros; simpl in *; try rename i into t.

    edestruct step_term_inf_nat_iter as [n[b'[??]]]; subst; eauto.

    inverts H.
    edestruct IHb as [n'[b'[?[??]]]]; eauto.
    subst b1'.
    exists n' (iswtm_seq b' t); splits; intros; auto.
      simpl; intuition.
      apply st_seq1; auto.

    clear IHb.
    apply subst_iter_skip in H1.
    subst b.
    exists 0 (iswtm_term w t); splits; intros; auto.
      simpl; intuition.
      apply st_seq_skip.

    inverts H.
    edestruct IHb1 as [n'[b'[?[??]]]]; clear IHb1 IHb2; eauto.
    subst b1'.
    exists n' (iswtm_par b' b2); splits; intros; auto.
      simpl; intuition.
      apply st_par1; auto.
    edestruct IHb2 as [n'[b'[?[??]]]]; clear IHb1 IHb2; eauto.
    subst b2'.
    exists n' (iswtm_par b1 b'); splits; intros; auto.
      simpl; intuition.
      apply st_par2; auto.
    clear IHb1 IHb2.
    apply subst_iter_skip in H0.
    apply subst_iter_skip in H2.
    subst b1 b2.
    exists 0 (w:tm_skip); splits; intros; auto.
      simpl; intuition.
      apply st_par_join; auto.
  Qed.

  Lemma cstep_inf_nat_iter: forall b s s' ba' C C',
    cstep s C (subst_iter b i_inf) s' C' ba' ->
    (exists (n:nat) (b':witer_schema),
      ba' = subst_iter b' i_inf /\
      (forall j:positive, wfree_iter_var b j -> wfree_iter_var b' j) /\
      forall (z:iteration_expr) m, m>=n -> cstep s C (subst_iter b (iter_plus1 m z)) s' C' (subst_iter b' (iter_plus1 m z))).
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    edestruct step_inf_nat_iter as [n'[b'[?[??]]]]; eauto.
    exists n' b'; splits; simpl; intros; auto.
      apply cstep_simulates_step.
      apply step_add_context; eauto.
    subst ba'.
    exists 0 b; splits; simpl; intros; auto.
      apply cstep_skip_subst with gwemp; auto.
  Qed.

  Lemma step_star_inf_nat_iter: forall s b s' ba',
    step_star s (subst_iter b i_inf) s' ba' ->
    (exists (n:nat) (b':witer_schema),
      ba' = subst_iter b' i_inf /\
      (forall j:positive, wfree_iter_var b j -> wfree_iter_var b' j) /\
      forall z m, m>=n -> step_star s (subst_iter b (iter_plus1 m z)) s' (subst_iter b' (iter_plus1 m z))).
  Proof.
    intros.
    remember (subst_iter b i_inf) as ba.
    revert b Heqba.
    induction H; intros; subst.

    exists 0 b0; splits; simpl; intros; auto.
      apply step_nil.

    rename b' into ba'; rename b'' into ba''.
    edestruct step_inf_nat_iter as [n'[b'[?[??]]]]; eauto.
    subst.
    edestruct IHstep_star as [n''[b''[?[??]]]]; clear IHstep_star; eauto.
    exists (n'+n'') b''; splits; simpl; intros; auto.
      eapply step_cons.
      apply H3. omega.
      apply H5. omega.
    edestruct IHstep_star as [n'[b'[?[??]]]]; clear IHstep_star; eauto.
    exists n' b'; splits; simpl; intros; auto.
      eapply st_iteration; eauto.
  Qed.

  Lemma step_star_schema_inf_nat_iter: forall s w (f:iter_schema) s' ba',
    step_star s (w: f i_inf) s' ba' ->
    (exists (n:nat) b',
      ba' = subst_iter b' i_inf /\
      (forall j:positive, free_iter_var f j -> wfree_iter_var b' j) /\
      forall z m, m>=n -> step_star s (w: f (iter_plus1 m z)) s' (subst_iter b' (iter_plus1 m z))).
  Proof.
    intros.
    change (w: f i_inf) with (subst_iter (iswtm_term w f) i_inf) in H.
    edestruct step_star_inf_nat_iter as [n[b'[??]]]; eauto.
  Qed.

  Lemma step_star_schema_inf_nat: forall s w (f:iter_schema) s' w',
    step_star s (w: f i_inf) s' (w': tm_skip) ->
    exists (n:nat), forall m,
      m>=n -> step_star s (w: f m) s' (w': tm_skip).
  Proof.
    intros.
    edestruct step_star_schema_inf_nat_iter as [n[b'[?[??]]]]; eauto.
    exists n; intros.
    apply subst_iter_skip in H0.
    subst b'.
    spec H2 0 m H3.
    simpl in H2.
    replace (m+0) with m in H2.
    apply H2; auto.
    omega.
  Qed.

  (* says that we can take a run that assumes infinite iterations and get the same immediate effect by using
       at most n variables, and that we can tack on arbitrarily more iterations that won't be run (note that
       ndi is equal in both the updated s and s').
  *)
  Lemma cstep_star_inf_nat_iter: forall s C b s' C' ba',
    cstep_star s C (subst_iter b i_inf) s' C' ba' ->
    (exists (n:nat) (b':witer_schema),
      ba' = subst_iter b' i_inf /\
      (forall j:positive, wfree_iter_var b j -> wfree_iter_var b' j) /\
      forall z m, m>=n -> cstep_star s C (subst_iter b (iter_plus1 m z)) s' C' (subst_iter b' (iter_plus1 m z))).
  Proof.
    intros.
    remember (subst_iter b i_inf) as ba.
    revert b Heqba.
    induction H; intros; subst.

    exists 0 b0; splits; simpl; intros; auto.
      apply cstep_nil.

    rename b' into ba'; rename b'' into ba''.
    edestruct cstep_inf_nat_iter as [n'[b'[?[??]]]]; eauto.
    subst.
    edestruct IHcstep_star as [n''[b''[?[??]]]]; clear IHcstep_star; eauto.
    exists (n'+n'') b''; splits; simpl; intros; auto.
      eapply cstep_cons.
      apply H3. omega.
      apply H5. omega.
    edestruct IHcstep_star as [n'[b'[?[??]]]]; clear IHcstep_star; eauto.
    exists n' b'; splits; simpl; intros; auto.
      eapply cstep_iteration; eauto.
  Qed.

  Lemma cstep_star_schema_inf_nat_iter: forall s C w (f:iter_schema) s' C' ba',
    cstep_star s C (w: f i_inf) s' C' ba' ->
    (exists (n:nat) b',
      ba' = subst_iter b' i_inf /\
      (forall j:positive, free_iter_var f j -> wfree_iter_var b' j) /\
      forall z m, m>=n -> cstep_star s C (w: f (iter_plus1 m z)) s' C' (subst_iter b' (iter_plus1 m z))).
  Proof.
    intros.
    change (w: f i_inf) with (subst_iter (iswtm_term w f) i_inf) in H.
    edestruct cstep_star_inf_nat_iter as [n[b'[??]]]; eauto.
  Qed.

  Lemma cstep_star_schema_inf_nat: forall s C w (f:iter_schema) s' C' w',
    cstep_star s C (w: f i_inf) s' C' (w': tm_skip) ->
    exists (n:nat), forall m,
      m>=n -> cstep_star s C (w: f m) s' C' (w': tm_skip).
  Proof.
    intros.
    edestruct cstep_star_schema_inf_nat_iter as [n[b'[?[??]]]]; eauto.
    exists n; intros.
    apply subst_iter_skip in H0.
    subst b'.
    spec H2 0 m H3.
    simpl in H2.
    replace (m+0) with m in H2.
    apply H2; auto.
    omega.
  Qed.


End SubstChoiceLemmas.
(******************************************************************)



Section Lemmas.
  Import Notation.
  Open Scope forces_pred.


  Lemma step_term_stall_iter: forall w t i n z s s' ba',
    s_iter_get s' i <> Some iter_go ->
    step s (w: term_subst_iter t (i_stall i n z)) s' ba' ->
    exists (b':witer_schema),
      ba' = subst_iter b' (i_stall i n z) /\
      (forall j:positive, free_iter_var t j -> wfree_iter_var b' j) /\
      forall (z1:iteration_expr), step s (w: term_subst_iter t (iter_plus1 n z1)) s' (subst_iter b' (iter_plus1 n z1)).
  Proof.
    intros.
    destruct t.

    exists ba'; splits; intros; auto.
      rewrite subst_iter_id; reflexivity.
      apply wfree_iter_var_wterm.
      eapply wterm_free_iter_var_step_preservation; eauto.
      simpl. auto.
      rewrite subst_iter_id; auto.

    inverts H0.
    inverts H8.
    simpl; replace n1 with (n- S n0); [| omega].
    destruct n. false; omega.
    replace (n0-S n) with (n0-n); [| omega].
    exists (iswtm_seq (iswtm_term w t) (istm_loop_var e (S n0) t)); splits; intros; auto.
      simpl; intuition.
      apply st_while_true; auto.
      apply iteration_decr_minus.
      apply iteration_expr_gt_nat. omega.
    exists (w: tm_skip); splits; intros; auto.
      simpl; intuition.
      apply st_while_false; auto.
    false; intuition.

    simpl in *; inverts H0.
    exists (iswtm_seq (iswtm_term w t) (istm_while e z' t)); splits; intros; auto.
      simpl; destruct i; inverts H8; intuition.
      apply st_while_true; auto.
    exists (w: tm_skip); splits; intros; auto.
      simpl; intuition.
      apply st_while_false; auto.
    destruct (Pos.eq_dec i i1); subst.
    false; intuition.
    exists (iswtm_term w (istm_while e z0 t)); splits; intros; auto.
      simpl; intuition.
      apply st_while_go; auto.
    exists (w: tm_skip); splits; intros; auto.
      simpl; intuition.
      apply st_while_limit; auto.

    simpl in *; inverts H0.
    exists (iswtm_term w t1); splits; intros; auto.
      simpl; intuition.
      apply st_if_true; auto.
    exists (iswtm_term w t2); splits; intros; auto.
      simpl; intuition.
      apply st_if_false; auto.

    inverts H0.
    exists (iswtm_seq (iswtm_term w t1) t2); splits; intros; auto.
      simpl; intuition.
      apply st_seq; auto.

    inverts H0.
    exists (iswtm_par (iswtm_term w1 t1) (iswtm_term w2 t2)); splits; intros; auto.
      simpl; intuition.
      apply st_par; auto.

    rename i0 into f.
    inverts H0.
    exists (iswtm_term w' (f x)); splits; intros; auto.
      simpl; intuition.
    change ((iswtm_term w' (f x)) (iter_plus1 n z1))
      with (w': (fun x => (f x) (iter_plus1 n z1)) x).
    eapply st_local; eauto.
  Qed.

  Lemma step_stall_iter: forall i n z b s s' ba',
    s_iter_get s' i <> Some iter_go ->
    step s (subst_iter b (i_stall i n z)) s' ba' ->
    exists (b':witer_schema),
      ba' = subst_iter b' (i_stall i n z) /\
      (forall j:positive, wfree_iter_var b j -> wfree_iter_var b' j) /\
      forall (z1:iteration_expr), step s (subst_iter b (iter_plus1 n z1)) s' (subst_iter b' (iter_plus1 n z1)).
  Proof.
    intros.
    revert ba' H0.
    induction b; intros; simpl in *; try rename i0 into t.

    edestruct step_term_stall_iter as [b'[??]]; subst; eauto.

    inverts H0.
    edestruct IHb as [b'[?[??]]]; clear IHb; eauto.
    subst b1'.
    exists (iswtm_seq b' t); splits; intros; auto.
      simpl; intuition.
      apply st_seq1; auto.

    clear IHb.
    apply subst_iter_skip in H2.
    subst b.
    exists (iswtm_term w t); splits; intros; auto.
      simpl; intuition.
      apply st_seq_skip.

    inverts H0.
    edestruct IHb1 as [b'[?[??]]]; clear IHb1 IHb2; eauto.
    subst b1'.
    exists (iswtm_par b' b2); splits; intros; auto.
      simpl; intuition.
      apply st_par1; auto.
    edestruct IHb2 as [b'[?[??]]]; clear IHb1 IHb2; eauto.
    subst b2'.
    exists (iswtm_par b1 b'); splits; intros; auto.
      simpl; intuition.
      apply st_par2; auto.
    clear IHb1 IHb2.
    apply subst_iter_skip in H1.
    apply subst_iter_skip in H3.
    subst b1 b2.
    exists (w:tm_skip); splits; intros; auto.
      simpl; intuition.
      apply st_par_join; auto.
  Qed.

  Lemma cstep_stall_iter: forall i n z b s s' ba' C C',
    s_iter_get s' i <> Some iter_go ->
    cstep s C (subst_iter b (i_stall i n z)) s' C' ba' ->
    exists (b':witer_schema),
      ba' = subst_iter b' (i_stall i n z) /\
      (forall j:positive, wfree_iter_var b j -> wfree_iter_var b' j) /\
      forall (z1:iteration_expr), cstep s C (subst_iter b (iter_plus1 n z1)) s' C' (subst_iter b' (iter_plus1 n z1)).
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    edestruct step_stall_iter as [b'[?[??]]]; eauto.
    exists b'; splits; simpl; intros; auto.
      apply cstep_simulates_step.
      apply step_add_context; eauto.
    subst ba'.
    exists b; splits; simpl; intros; auto.
      apply cstep_skip_subst with gwemp; auto.
  Qed.

  Lemma cstep_star_stall_iter: forall i n z s C b s' C' ba',
    s_iter_get s' i <> Some iter_go ->
    cstep_star s C (subst_iter b (i_stall i n z)) s' C' ba' ->
    exists (b':witer_schema),
      ba' = subst_iter b' (i_stall i n z) /\
      (forall j:positive, wfree_iter_var b j -> wfree_iter_var b' j) /\
      forall (z1:iteration_expr), cstep_star s C (subst_iter b (iter_plus1 n z1)) s' C' (subst_iter b' (iter_plus1 n z1)).
  Proof.
    intros.
    remember (subst_iter b (i_stall i n z)) as ba.
    revert b Heqba.
    induction H0; intros; subst.

    exists b0; splits; simpl; intros; auto.
      apply cstep_nil.

    rename b' into ba'; rename b'' into ba''.
    edestruct cstep_stall_iter as [b'[?[??]]].
    2: eauto.
    intro; eapply H.
    eapply cstep_star_s_iter_preservation; eauto.
    subst.
    edestruct IHcstep_star as [b''[?[??]]]; clear IHcstep_star; eauto.
    exists b''; splits; simpl; intros; auto.
      eapply cstep_cons; eauto.
    edestruct IHcstep_star as [b'[?[??]]]; clear IHcstep_star; eauto.
    exists b'; splits; simpl; intros; auto.
      eapply cstep_iteration; eauto.
  Qed.


  Lemma step_star_stall_iter: forall i n z s b s' ba',
    s_iter_get s' i <> Some iter_go ->
    step_star s (subst_iter b (i_stall i n z)) s' ba' ->
    exists (b':witer_schema),
      ba' = subst_iter b' (i_stall i n z) /\
      (forall j:positive, wfree_iter_var b j -> wfree_iter_var b' j) /\
      forall (z1:iteration_expr), step_star s (subst_iter b (iter_plus1 n z1)) s' (subst_iter b' (iter_plus1 n z1)).
  Proof.
    intros.
    remember (subst_iter b (i_stall i n z)) as ba.
    revert b Heqba.
    induction H0; intros; subst.

    exists b0; splits; simpl; intros; auto.
      apply step_nil.

    rename b' into ba'; rename b'' into ba''.
    edestruct step_stall_iter as [b'[?[??]]].
    2: eauto.
    intro; eapply H.
    eapply step_star_s_iter_preservation; eauto.
    subst.
    edestruct IHstep_star as [b''[?[??]]]; clear IHstep_star; eauto.
    exists b''; splits; simpl; intros; auto.
      eapply step_cons; eauto.
    edestruct IHstep_star as [b'[?[??]]]; clear IHstep_star; eauto.
    exists b'; splits; simpl; intros; auto.
      eapply st_iteration; eauto.
  Qed.

  Lemma step_star_stalled_skip: forall s w (f:iter_schema) i x z s' w',
    s_iter_get s' i <> Some iter_go ->
    step_star s (w: f (i_stall i x z)) s' (w':tm_skip) ->
    step_star s (w: f (i_plus x z)) s' (w': tm_skip).
  Proof.
    intros.
    change (w: f (i_stall i x z)) with (subst_iter (iswtm_term w f) (i_stall i x z)) in H0.
    edestruct step_star_stall_iter as [b'[?[??]]]; eauto.
    spec H3 z.
    rewrite iter_plus1_iterations in H3.
    apply subst_iter_skip in H1. subst b'.
    apply H3.
  Qed.

  Lemma step_star_stalled_skip_inf: forall s w (f:iter_schema) i x z s' w',
    s_iter_get s' i <> Some iter_go ->
    step_star s (w: f (i_stall i x z)) s' (w':tm_skip) ->
    step_star s (w: f i_inf) s' (w': tm_skip).
  Proof.
    intros.
    change (w: f (i_stall i x z)) with (subst_iter (iswtm_term w f) (i_stall i x z)) in H0.
    edestruct step_star_stall_iter as [b'[?[??]]]; eauto.
    spec H3 i_inf.
    apply subst_iter_skip in H1. subst b'.
    simpl in H3.
    auto.
  Qed.

  Lemma cstep_star_stalled_skip_inf: forall s C w (f:iter_schema) i x z s' C' w',
    s_iter_get s' i <> Some iter_go ->
    cstep_star s C (w: f (i_stall i x z)) s' C' (w':tm_skip) ->
    cstep_star s C (w: f i_inf) s' C' (w': tm_skip).
  Proof.
    intros.
    change (w: f (i_stall i x z)) with (subst_iter (iswtm_term w f) (i_stall i x z)) in H0.
    edestruct cstep_star_stall_iter as [b'[?[??]]]; eauto.
    spec H3 i_inf.
    apply subst_iter_skip in H1. subst b'.
    simpl in H3.
    auto.
  Qed.

(*
  Lemma bisim_Cs_while_plus1: forall (z1:iterations) (z2:iteration_expr) e t,
    term_bisimilar_Cs (tm_seq (tm_while_max e z1 t) (tm_while_max e z2 t)) (tm_while_max e (iter_plus1 z1 z2) t).
  Proof.
Admitted.
*)

  Lemma bisim_while_go_stall: forall s w e i n z t,
   s_iter_get s i = Some iter_go ->
   bisimilar s (w: tm_while_max e (i_stall i n z) t) (w: tm_while_max e (i_plus n z) t).
  Proof.
    intros.
    exists (fun s ba bb =>
      s_iter_get s i = Some iter_go /\
      ((exists n w, ba = w: tm_while_max e (i_stall i n z) t /\ bb = w: tm_while_max e (i_plus n z) t) \/
        (exists n bt, ba = bt;; tm_while_max e (i_stall i n z) t /\ bb = bt;; tm_while_max e (i_plus n z) t) \/
        ba=bb)).
    split.
    2: split; auto; left; eauto.
    clear s w n H.
    intros s ba bb [? [ [n[w[??]]] | [ [n[bt[??]]] |?]]]; subst.
    splits.

    intros [?]; discriminate.
    intros [?]; discriminate.
    intros s' ?.
    split; eauto.
    eapply s_iter_go_preservation; eauto.
    intros.
    split; eauto.
    eapply step_go_preservation; eauto.

    intros s' ba' ?.
    inverts H0.
    inverts H8.
    exists (w: t;; tm_while_max e (i_plus n0 z) t); splits; auto.
      apply step_single; apply st_while_true; auto.
      apply iteration_decr_i_plus. constructor.
      right; left; eauto.
    exists (w: tm_skip); splits; auto.
      apply step_single; apply st_while_false; auto.
    exists (w: tm_while_max e z t); splits; auto.
      rewrite i_plus_zero1.
      apply step_nil.

    intros s' ba' ?.
    inverts H0.
    assert ((n=0 /\ iteration_decr z z') \/ (exists n0, n=S n0 /\ z'=i_plus n0 z)).
      inverts H8.
      destruct z; try discriminate.
      inverts H1.
      destruct n; simpl in H2. subst n1. left; split; constructor.
      right; exists n; split; auto. simpl. inverts H2; auto.
      destruct z; try discriminate.
      destruct n. left; split; constructor.
      right; exists n; split; auto.
    clear H8.
    destruct H0 as [[??] | [n0[??]]].
    subst n.
    exists (w: t;; tm_while_max e z' t); splits; auto.
      eapply step_cons.
      apply st_while_go; auto.
      apply step_single.
      apply st_while_true; auto.
    subst n z'.
    exists (w: t;; tm_while_max e (i_stall i n0 z) t); splits; auto.
      apply step_single; apply st_while_true; auto.
      constructor; auto.
      right; left; eauto.
    exists (w: tm_skip); splits; auto.
      apply step_single; apply st_while_false; auto.
    assert (n=0 /\ z=0).
      clear -H4. destruct z; destruct n; try discriminate. simpl in *; subst; auto.
    clear H4.
    destruct H0; subst n z.
    exists (w: tm_skip); splits; auto.
      eapply step_cons.
      apply st_while_go; auto.
      apply step_single.
      apply st_while_limit; auto.

    splits.
    intros [?]; discriminate.
    intros [?]; discriminate.
    intros s' ?.
    split; eauto.
    eapply s_iter_go_preservation; eauto.
    right; left; eauto.
    intros.
    split; eauto.
    eapply step_go_preservation; eauto.
    right; left; eauto.

    intros s' ba' ?.
    inverts H0.
    exists (b1';; tm_while_max e (i_plus n z) t); splits; auto.
      apply step_single; apply st_seq1; auto.
      eapply step_go_preservation; eauto.
      right; left; eauto.
    exists (w: tm_while_max e (i_plus n z) t); splits; auto.
      apply step_single; apply st_seq_skip.
      left; eauto.

    intros s' bb' ?.
    remember (i_plus n z) as zz.
    inverts H0; subst zz.
    exists (b1';; tm_while_max e (i_stall i n z) t); splits; auto.
      apply step_single; apply st_seq1; auto.
      eapply step_go_preservation; eauto.
      right; left; eauto.
    exists (w: tm_while_max e (i_stall i n z) t); splits; auto.
      apply step_single; apply st_seq_skip.
      left; eauto.

    splits.
    intro; splits; eauto; apply step_nil.
    intro; splits; eauto; apply step_nil.
    intros; split; eauto. eapply s_iter_go_preservation; eauto.
    intros; split; eauto. eapply step_go_preservation; eauto.
    intros; exists b1'; splits; eauto.
      apply step_single; eauto.
      eapply step_go_preservation; eauto.
    intros; exists b2'; splits; eauto.
      apply step_single; eauto.
      eapply step_go_preservation; eauto.
  Qed.




  Lemma wterm_free_iter_var_set_context1: forall C b i,
    wterm_free_iter_var (set_context C (gwemp : tm_skip)) i ->
    wterm_free_iter_var b i ->
    wterm_free_iter_var (set_context C b) i.
  Proof.
    intros.
    induction C; simpl in *; intuition.
  Qed.

  Lemma wterm_free_iter_var_set_context2: forall C b i,
    wterm_free_iter_var (set_context C b) i ->
    wterm_free_iter_var (set_context C (gwemp : tm_skip)) i /\ wterm_free_iter_var b i.
  Proof.
    intros.
    induction C; simpl in *; intuition.
  Qed.


  Lemma cstep_elim_free_iter_var: forall s i n C b ss' C' b',
    cstep (s_iter_set s i n) C b ss' C' b' ->
    wterm_free_iter_var (set_context C (gwemp:tm_skip)) i ->
    wterm_free_iter_var b i ->
    exists s', ss' = s_iter_set s' i n /\ cstep s C b s' C' b'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    edestruct step_elim_free_iter_var as [s'[??]]; eauto.
    exists s'; split; auto. apply cstep_simulates_step; apply step_add_context; auto.
    subst b'.
    clear H H1.
    remember (s_iter_set s i n) as ss.
    remember (gwemp: tm_skip) as bb in H2 at 1.
    remember (gwemp: tm_skip) as bb' in H2 at 1.
    revert Heqss Heqbb Heqbb' H0.
    induction H2; intros; subst.
    edestruct IHcstep as [ss'[??]]; eauto. apply H0.
    exists ss'; split; auto. constructor; auto.
    edestruct IHcstep as [ss'[??]]; eauto. apply H0.
    exists ss'; split; auto. constructor; auto.
    edestruct IHcstep as [ss'[??]]; eauto. apply H0.
    exists ss'; split; auto. constructor; auto.
    edestruct step_elim_free_iter_var as [ss'[??]]; eauto. apply H0.
    exists ss'; split; auto. constructor; auto.
    edestruct step_elim_free_iter_var as [ss'[??]]; eauto. apply H0.
    exists ss'; split; auto. constructor; auto.
    inverts H.
  Qed.

  Lemma step_spin_elim_free_iter_var : forall s i n b,
    step (s_iter_set s i n) b (s_iter_set s i n) b ->
    step s b s b.
  Proof.
    intros.
    remember (s_iter_set s i n) as ss in H at 1.
    remember (s_iter_set s i n) as ss' in H at 1.
    remember b as b' in H at 2.
    revert Heqss Heqss' Heqb'.
    step_cases (induction H) Case; intros; try inverts Heqb'; try absurd_rec_def; subst.
    - Case "st_recv_blocked".
    rewrite s_get_ch_iter_set in H2.
    eapply st_recv_blocked; eauto.
    - Case "st_local".
    rewrite H3 at 1.
    eapply st_local; eauto.
    - Case "st_seq1".
    apply st_seq1; eauto.
    - Case "st_par1".
    apply st_par1; eauto.
    - Case "st_par2".
    apply st_par2; eauto.
  Qed.

  Lemma cstep_spin_elim_free_iter_var : forall s i n C b,
    cstep (s_iter_set s i n) C b (s_iter_set s i n) C b ->
    cstep s C b s C b.
  Proof.
    intros.
    apply cstep_simulates_step.
    eapply step_spin_elim_free_iter_var.
    apply step_simulates_cstep; eauto.
  Qed.

  Lemma step_elim_free_iter_var': forall s0 s i n b s0' ss' b',
    step (s_iter_set s i n) b ss' b' ->
    step s0 b s0' b' ->
    s_iter_get s0 i = None ->
    exists s', ss' = s_iter_set s' i n /\ step s b s' b'.
  Proof.
    introv H.
    step_cases (induction H) Case HH; intros; eauto.
    - Case "st_while_go".
    destruct (Pos.eq_dec i0 i).
    subst i0.
    rewrite s_iter_get_set_eq in H.
    inverts H0.
    false.
    rewrite s_iter_get_set_neq in H; eauto.
    - Case "st_send".
    rewrite s_get_ch_iter_set in *; eauto.
    rewrite s_push_ch_iter_set in *; eauto.
    - Case "st_recv".
    rewrite s_get_ch_iter_set in *; eauto.
    rewrite s_pop_ch_iter_set in *; eauto.
    - Case "st_recv_blocked".
    rewrite s_get_ch_iter_set in *; eauto.
    - Case "st_newch".
admit. (* fresh_channel ignores s_i *)
    - Case "st_endch".
    rewrite s_get_ch_iter_set in *; eauto.
    rewrite s_rm_ch_iter_set in *; eauto.
    - Case "st_seq1".
    inverts H0.
    destruct IHstep as [s'0[??]]; eauto.
    - Case "st_par1".
    inverts H0.
    destruct IHstep as [s'0[??]]; eauto.
    assert (s' = s_iter_set s i n); subst.
      eapply step_spin_no_change; eauto.
    exists s; split; auto.
    apply st_par1.
    eapply step_spin_elim_free_iter_var; eauto.
    - Case "st_par2".
    inverts H0.
    assert (s' = s_iter_set s i n); subst.
      eapply step_spin_no_change; eauto.
    exists s; split; auto.
    apply st_par2.
    eapply step_spin_elim_free_iter_var; eauto.
    destruct IHstep as [s'0[??]]; eauto.
  Qed.

  (* instead of requiring C to not access i, we demonstrate that C can execute w/o i *)
  Lemma cstep_elim_free_iter_var': forall s0 s i n C b0 b s0' ss' C' b0' b',
    cstep (s_iter_set s i n) C b ss' C' b' ->
    cstep s0 C b0 s0' C' b0' ->
    s_iter_get s0 i = None ->
    wterm_free_iter_var b i ->
    exists s', ss' = s_iter_set s' i n /\ cstep s C b s' C' b'.
  Proof.
    intros.
    edestruct cstep_inversion with (t:=b) as [[?]|[?]]; eauto.
    subst C'.
    edestruct step_elim_free_iter_var as [s'[??]]; eauto.
    exists s'; split; auto. apply cstep_simulates_step; apply step_add_context; auto.
    subst b'.
    clear H H2.
    edestruct cstep_inversion with (t:=b0) as [[?]|[?]]; eauto.
    subst C'.
    assert (ss'=s_iter_set s i n); subst.
      eapply cstep_spin_no_change; eauto.
    exists s; split; auto.
    eapply cstep_skip_subst.
    eapply cstep_spin_elim_free_iter_var; eauto.
    subst b0'.
    edestruct step_elim_free_iter_var' with (s:=s) (s0:=s0) (i:=i) (n:=n) as [s'[??]]; eauto.
    apply step_simulates_cstep; eauto.
    apply step_simulates_cstep; eauto.
    exists s'; split; auto.
    eapply cstep_skip_subst; eauto.
    eapply cstep_simulates_cstep_step; eauto.
  Qed.

(*
  Lemma cstep_star_elim_free_iter_var': forall s0 s i n C b0 b s0' ss' C' b0' b',
    cstep_star (s_iter_set s i n) C b ss' C' b' ->
    cstep_star s0 C b0 s0' C' b0' ->
    s_iter_get s0 i = None ->
    wterm_free_iter_var b i ->
    exists s', ss' = s_iter_set s' i n /\ cstep_star s C b s' C' b'.
  Proof.
    intros.
    remember (s_iter_set s i n) as ss.
    revert s s0 s0' b0 b0' Heqss H0 H1 H2.
    induction H; intros; eauto.
    - subst; exists s0; split; auto. apply cstep_nil.
    - subst.
    
    edestruct cstep_elim_free_iter_var' with (s:=s0) (s0:=s1) as [s'0[??]]; eauto.
    edestruct IHcstep_star as [s''0[??]]; eauto.
  Qed.
*)

  Lemma cstep_star_elim_free_iter_var: forall s i n C b s' C' b',
    ss_iter_star s s' ->
    cstep_star (s_iter_set s i n) C b (s_iter_set s' i n) C' b' ->
    wterm_free_iter_var (set_context C (gwemp:tm_skip)) i ->
    wterm_free_iter_var b i ->
    cstep_star s C b s' C' b'.
  Proof.
    intros.
    remember (s_iter_set s i n) as ss.
    remember (s_iter_set s' i n) as ss'.
    revert s s' Heqss Heqss' H H1 H2.
    induction H0; intros; subst.

    eapply cstep_star_s_iter_set.
    eapply s_iter_set_eq; eauto.

    rename s' into ss'.
    edestruct cstep_elim_free_iter_var as [s'[??]]; eauto.
    subst ss'.
    eapply cstep_cons; eauto.
    edestruct wterm_free_iter_var_set_context2 with (b:=b'); eauto.
    eapply wterm_free_iter_var_step_preservation.
    apply step_simulates_cstep. eauto.
    apply wterm_free_iter_var_set_context1; eauto.
    apply IHcstep_star; auto.
    clear -H1 H5.
    apply cstep_s_iter in H5.
    unfold ss_iter_star in *; rewrite<- H5; simpl in *; auto.

    rename s' into ss'.
    apply ss_iter_star_cstep_star in H0.
    inverts H.
    destruct (Pos.eq_dec i0 i); subst.
    false. clear IHcstep_star.
    rewrite s_iter_set_set_eq' in H0.
    rewrite s_iter_get_set_eq in H4.
    inverts H4.
    unfold ss_iter_star in H0.
    eapply s_iter_star_go_preservation with (i:=i) in H0.
    unfold s_iter_get in H0; simpl in H0.
    rewrite s_i_iter_set in H0.
    rewrite State.PMapFacts.add_eq_o in H0; auto.
    discriminate.
    unfold s_iter_get; simpl.
    rewrite s_i_iter_set.
    rewrite State.PMapFacts.add_eq_o; auto.
    (**)
    rewrite s_iter_get_set_neq in H4; auto.
    rewrite s_iter_set_set_neq' in H0; auto.
    eapply cstep_iteration.
    constructor; eauto.
    apply IHcstep_star; auto.
    rewrite s_iter_set_set_neq'; auto.
    clear IHcstep_star H2 H3.
    unfold ss_iter_star in *.
    cut (s_iter_star (s_iter_set (supd_iter empty_state (s_i s0)) i0 iter_go) (supd_iter empty_state (s_i s'0))); intros.
    replace (supd_iter empty_state (s_i (s_iter_set s0 i0 iter_go)))
      with (s_iter_set (supd_iter empty_state (s_i s0)) i0 iter_go).
    apply H.
    rewrite s_i_iter_set.
    reflexivity.
    eapply s_iter_normalize; eauto.
    eapply s_iter_star_go_preservation with (i:=i0) in H0; eauto.
    rewrite s_i_iter_set in H0.
    unfold s_iter_get in *.
    rewrite s_i_supd_iter in H0.
    rewrite State.PMapFacts.add_neq_o in H0; auto.
    unfold s_iter_get in *.
    repeat rewrite s_i_iter_set.
    rewrite s_i_supd_iter.
    rewrite State.PMapFacts.add_neq_o, State.PMapFacts.add_eq_o; auto.
  Qed.

  Lemma step_star_elim_free_iter_var: forall s i n b s' b',
    ss_iter_star s s' ->
    step_star (s_iter_set s i n) b (s_iter_set s' i n) b' ->
    wterm_free_iter_var b i ->
    step_star s b s' b'.
  Proof.
    intros.
    change b with (set_context cx_hole b).
    change b' with (set_context cx_hole b').
    apply step_star_simulates_cstep_star.
    eapply cstep_star_elim_free_iter_var; eauto.
    apply step_star_simulates_cstep_star_empty_context; eauto.
    simpl. auto.
  Qed.


  Lemma s_iter_cstep_None_preservation: forall i s C b s' C' b',
    s_iter_get s i = None ->
    cstep s C b s' C' b' ->
    s_iter_get s' i = None.
  Proof.
    intros.
    unfold s_iter_get in *.
    erewrite<- cstep_s_iter; eauto.
  Qed.

  Lemma s_iter_step_star_None_preservation: forall i s b s' b',
    s_iter_get s i = None ->
    step_star s b s' b' ->
    s_iter_get s' i = None.
  Proof.
    intros.
    revert H.
    induction H0; intros; auto; apply IHstep_star.
    eapply s_iter_step_None_preservation; eauto.
    eapply s_iter_None_preservation; eauto.
  Qed.

  Lemma s_iter_cstep_star_None_preservation: forall i s C b s' C' b',
    s_iter_get s i = None ->
    cstep_star s C b s' C' b' ->
    s_iter_get s' i = None.
  Proof.
    intros.
    revert H.
    induction H0; intros; auto; apply IHcstep_star.
    eapply s_iter_cstep_None_preservation; eauto.
    eapply s_iter_None_preservation; eauto.
  Qed.

  Lemma step_intro_free_iter_var: forall i ndi s b s' b',
    step s b s' b' ->
    s_iter_get s i = None ->
    step (s_iter_set s i ndi) b (s_iter_set s' i ndi) b'.
  Proof.
    intros.
    step_cases (induction H) Case HH; eapply HH; subst;
      try rewrite s_get_ch_iter_set;
      try solve [symmetry; apply s_ch_apply_iter_set];
      first[Case "st_newch" | eauto].

    destruct (Pos.eq_dec i i0); subst.
    false.
    rewrite s_iter_get_set_neq; auto.

    rewrite fresh_channel_iter_set; reflexivity.
    rewrite fresh_channel_iter_set; reflexivity.
    rewrite fresh_channel_iter_set, s_new_ch_iter_set; eauto.
  Qed.

  Lemma cstep_intro_free_iter_var: forall i ndi s C b s' C' b',
    cstep s C b s' C' b' ->
    s_iter_get s i = None ->
    cstep (s_iter_set s i ndi) C b (s_iter_set s' i ndi) C' b'.
  Proof. intros; induction H; constructor; auto; apply step_intro_free_iter_var; auto. Qed.

  Lemma cstep_star_intro_free_iter_var: forall i ndi s C b s' C' b',
    cstep_star s C b s' C' b' ->
    s_iter_get s i = None ->
    cstep_star (s_iter_set s i ndi) C b (s_iter_set s' i ndi) C' b'.
  Proof.
    intros.
    induction H.
    - apply cstep_nil.
    - eapply cstep_cons.
    eapply cstep_intro_free_iter_var; eauto.
    apply IHcstep_star.
    eapply s_iter_cstep_None_preservation; eauto.
    - eapply cstep_iteration.
    eapply s_iter_set_None; eauto.
    apply IHcstep_star.
    eapply s_iter_None_preservation; eauto.
  Qed.

  Lemma bisim_Cs_term_wterm_subst_iter: forall w t z,
    bisimilar_Cs (w: term_subst_iter t z) (subst_iter (iswtm_term w t) z).
  Proof. reflexivity. Qed.

  Lemma wterm_free_iter_var_cstep_context_preservation: forall i s C b s' C' b',
    cstep s C b s' C' b' ->
    wterm_free_iter_var (set_context C b) i ->
    wterm_free_iter_var (set_context C' b') i.
  Proof.
    intros.
    eapply wterm_free_iter_var_step_preservation; eauto.
    apply step_simulates_cstep; eauto.
  Qed.

  Lemma wterm_free_iter_var_context_preservation: forall i s C b s' C' b',
    cstep_star s C b s' C' b' ->
    wterm_free_iter_var (set_context C b) i ->
    wterm_free_iter_var (set_context C' b') i.
  Proof.
    intros.
    induction H; auto.
    apply IHcstep_star.
    eapply wterm_free_iter_var_cstep_context_preservation; eauto.
  Qed.

  Lemma bisim_intro_free_iter_var: forall i ii s ba bb,
    s_iter_get s i = None ->
    wterm_free_iter_var ba i ->
    wterm_free_iter_var bb i ->
    bisimilar (s_iter_set s i ii) ba bb ->
    bisimilar s ba bb.
  Proof.
    intros.
    exists (fun s ba bb => s_iter_get s i = None /\ wterm_free_iter_var ba i /\ wterm_free_iter_var bb i /\ bisimilar (s_iter_set s i ii) ba bb).
    split; auto.
    apply bisimulation_symm_equiv.
    clear s ba bb H H0 H1 H2.
    intros s ba bb [?[?[??]]].
    splits; auto.

    intro; splits; auto.
    eapply step_star_elim_free_iter_var; eauto.
    apply s_iter_nil.
    eapply bisim_halted1; eauto.
    apply bisim_refl.

    intros.
    splits; auto.
    eapply s_iter_None_preservation; eauto.
    eapply bisim_s_iter; eauto.
    apply s_iter_set_None; eauto.  

    intros.
    splits; auto.
    eapply s_iter_step_None_preservation; eauto.
    eapply bisim_context; eauto.
    apply step_intro_free_iter_var; eauto.

    intros s' ba' ?.
    edestruct step_bisim1 as [bb'[??]].
    eapply step_intro_free_iter_var; eauto. eauto.
    exists bb'; splits; auto.
      eapply step_star_elim_free_iter_var; eauto.
      eapply ss_iter_step_cons; eauto. apply s_iter_nil.
      eapply s_iter_step_None_preservation; eauto.
      eapply wterm_free_iter_var_step_preservation; eauto.
      eapply wterm_free_iter_var_step_star_preservation; eauto.

    apply bisim_symm; auto.
  Qed.

  Lemma s_i_s_iter_set_eq: forall s1 s2 i z,
    s_iter_get s1 i = s_iter_get s2 i ->
    s_i (s_iter_set s1 i z) = s_i (s_iter_set s2 i z) ->
    s_i s1 = s_i s2.
  Proof.
    unfold s_iter_set; intros.
    do 2 rewrite s_i_supd_iter in H0.
    cut ({|s_c:=empty_ch_state;s_i:=s_i s1;s_all_res:=ResSet.empty|} = {|s_c:=empty_ch_state;s_i:=s_i s2;s_all_res:=ResSet.empty|}); intros.
    inverts H1; auto.
    apply Morphisms.state_equiv_eq.
    split.
    reflexivity.
    split.
    simpl.
    intro j.
    destruct (Pos.eq_dec j i). subst j.
    apply H.
    rewrite<- PMapFacts.add_neq_o with (m:=s_i s1) (x:=i) (e:=z); auto.
    rewrite<- PMapFacts.add_neq_o with (m:=s_i s2) (x:=i) (e:=z); auto.
    congruence.
    reflexivity.
  Qed.

  Lemma s_i_s_iter_set_eq': forall s1 s2 i z,
    s_iter_get s1 i = s_iter_get s2 i ->
    s_iter_set s1 i z = s_iter_set s2 i z ->
    s1 = s2.
  Proof.
    intros.
    apply Morphisms.state_equiv_eq.
    split.
    rewrite<- (s_iter_set_respects_s_c s1 i z).
    rewrite<- (s_iter_set_respects_s_c s2 i z).
    rewrite H0; reflexivity.
    split.
    rewrite (s_i_s_iter_set_eq s1 s2 i z); auto.
    reflexivity.
    congruence.
    rewrite<- (s_iter_set_respects_s_all_res s1 i z).
    rewrite<- (s_iter_set_respects_s_all_res s2 i z).
    rewrite H0; reflexivity.
  Qed.

  Lemma s_iter_set_canc: forall s1 s2 i z1 z2,
    s_i (s_iter_set s1 i z1) = s_i (s_iter_set s2 i z2) ->
    z1 = z2.
  Proof.
    intros.
    cut (s_iter_get (s_iter_set s1 i z1) i = s_iter_get (s_iter_set s2 i z2) i); intros.
    do 2 rewrite s_iter_get_set_eq in H0.
    congruence.
    apply s_iter_get_eq_s_i; auto.
  Qed.

  Lemma s_iter_set_s_iter_inv: forall s i n s',
    s_iter_get s i = s_iter_get s' i ->
    s_iter (s_iter_set s i n) (s_iter_set s' i n) ->
    s_iter s s'.
  Proof.
    intros.
    inverts H0.
    subst.
    destruct (Pos.eq_dec i0 i). subst i0.
    rewrite s_iter_get_set_eq in H3.
    rewrite s_iter_set_set_eq' in H1.
    inverts H3.
    assert (s_i (s_iter_set s i iter_go) = s_i (s_iter_set s' i iter_stall)) by congruence.
    eapply s_iter_set_canc in H0.
    discriminate.
    (**)
    rewrite s_iter_get_set_neq in H3; auto.
    eapply s_iter_go'; eauto.
    rewrite s_iter_set_set_neq' in H1; auto.
    apply s_i_s_iter_set_eq' in H1; auto.
    rewrite<- H.
    rewrite s_iter_get_set_neq; auto.
  Qed.

  Lemma s_iter_s_iter_set_inv: forall s i z s'',
    s_iter_get s i = None ->
    s_iter_get s'' i = Some z ->
    s_iter (s_iter_set s i z) s'' ->
    exists s',
      s_iter s s' /\
      s'' = s_iter_set s' i z.
  Proof.
    intros.
    inverts H1.
    destruct (Pos.eq_dec i0 i). subst i0.
    false.
    rewrite s_iter_get_set_eq in H0.
    rewrite s_iter_get_set_eq in H2.
    congruence.
    rewrite s_iter_get_set_neq in H2; auto.
    exists (s_iter_set s i0 iter_go); split.
    apply s_iter_go; auto.
    rewrite s_iter_set_set_neq'; auto.
  Qed.

  Definition s_iter_rm s i := supd_iter s (FMapPositive.PositiveMap.remove i (s_i s)).

  Lemma s_iter_get_rm_eq: forall s i,
    s_iter_get (s_iter_rm s i) i = None.
  Proof.
    unfold s_iter_get, s_iter_rm; intros.
    rewrite s_i_supd_iter.
    apply PMapFacts.remove_eq_o; auto.
  Qed.

  Lemma s_iter_get_rm_neq: forall s i1 i2,
    i1 <> i2 ->
    s_iter_get (s_iter_rm s i1) i2 = s_iter_get s i2.
  Proof.
    unfold s_iter_get, s_iter_rm; intros.
    rewrite s_i_supd_iter.
    apply PMapFacts.remove_neq_o; auto.
  Qed.

  Lemma supd_iter_respects_s_all_res: forall s z,
    s_all_res (supd_iter s z) = s_all_res s.
  Proof. intros; destruct s; reflexivity. Qed.

  Lemma s_iter_set_rm_eq: forall s i z,
    s_iter_set (s_iter_rm s i) i z == s_iter_set s i z.
  Proof.
    unfold s_iter_set, s_iter_rm; intros.
    rewrite supd_iter_redundant.
    rewrite s_i_supd_iter.
    split.
    do 2 rewrite supd_iter_respects_s_c; reflexivity.
    split.
    repeat rewrite s_i_supd_iter.
    intro j.
    destruct (Pos.eq_dec j i). subst j.
    repeat rewrite PMapFacts.add_eq_o; auto.
    repeat rewrite PMapFacts.add_neq_o; auto.
    repeat rewrite PMapFacts.remove_neq_o; auto.
    do 2 rewrite supd_iter_respects_s_all_res; reflexivity.
  Qed.

  Lemma s_iter_set_rm_eq': forall s i z,
    s_iter_set (s_iter_rm s i) i z = s_iter_set s i z.
  Proof. intros; apply Morphisms.state_equiv_eq; apply s_iter_set_rm_eq; auto. Qed.


(*

  Lemma abisim_intro_free_iter_var: forall i ii wO sa ba sb bb,
    s_iter_get sa i = None ->
    s_iter_get sb i = None ->
    wterm_free_iter_var ba i ->
    wterm_free_iter_var bb i ->
    abisimilar wO (s_iter_set sa i ii) ba (s_iter_set sb i ii) bb ->
    abisimilar wO sa ba sb bb.
  Proof.
    intros.
    exists (fun wO sa ba sb bb => s_iter_get sa i = None /\ s_iter_get sb i = None /\ wterm_free_iter_var ba i /\ wterm_free_iter_var bb i /\ abisimilar wO (s_iter_set sa i ii) ba (s_iter_set sb i ii) bb).
    split; auto.
    clear wO sa sb ba bb H H0 H1 H2 H3.
    intros wC sa ba sb bb [?[?[?[??]]]].
    splits; auto.
    - intro; splits; auto.
    eapply step_star_elim_free_iter_var; eauto.
    unfold ss_iter_star.
    erewrite s_i_s_iter_set_eq with (i:=i) (s1:=sa) (s2:=sb); eauto.
    apply s_iter_nil.
    congruence.
    eapply abisim_s_i_eq; eauto.
    eapply abisim_halted1; eauto.
    apply abisim_refl.
    - intros.
    edestruct abisim_s_iter with (sp:=s_iter_set sa i ii) as [sb'[??]]; eauto.
    apply s_iter_set_None; eauto.
    destruct (s_iter_s_iter_set_inv sb i ii sb') as [sb''[??]]; auto.
    rewrite (s_iter_get_eq_s_i sb' (s_iter_set sp' i ii)).
    rewrite s_iter_get_set_eq; auto.
    symmetry; eapply abisim_s_i_eq; eauto.
    subst sb'; rename sb'' into sb'.
    exists sb'; splits; auto.
    eapply s_iter_None_preservation with (s:=sa); eauto.
    eapply s_iter_None_preservation with (s:=sb); eauto.
    - eapply s_i_s_iter_set_eq with (i:=i).
    congruence.
    eapply abisim_s_i_eq; eauto.
    - intros.
    edestruct cstep_star_abisim with (C:=C) (C':=C') (sp:=s_iter_set sa i ii) as [sb'[bb'[??]]]; eauto.
    apply cstep_star_intro_free_iter_var; eauto.
    assert (sb' = s_iter_set sb' i ii).
      symmetry; eapply s_iter_set_nop.
      rewrite (s_iter_get_eq_s_i sb' (s_iter_set sp' i ii)).
      rewrite s_iter_get_set_eq; auto.
      eapply abisim_s_i_eq; eauto.
    rewrite H9 in H7.
    exists (s_iter_rm sb' i) bb'; splits; auto.
    eapply cstep_star_elim_free_iter_var; eauto.
    unfold ss_iter_star.
    rewrite (s_i_s_iter_set_eq sb sa i ii).
    rewrite (s_i_s_iter_set_eq (s_iter_rm sb' i) sp' i ii).
    eapply ss_iter_star_cstep_star; eauto.
    rewrite (s_iter_cstep_star_None_preservation i sa C ba sp' C' bp'); auto.
    apply s_iter_get_rm_eq.
    apply s_i_s_iter_set_eq with i ii; auto.

    rewrite s_iter_set_rm_eq'.
    do 2 rewrite s_iter_get_set_eq; reflexivity.
    rewrite s_iter_set_rm_eq'.
    do 2 rewrite s_iter_set_set_eq'.
    rewrite<- H9.
    eapply abisim_s_i_eq; eauto.
    congruence.
    symmetry; eapply abisim_s_i_eq; eauto.
    rewrite s_iter_set_rm_eq'; eauto.
Check wterm_free_iter_var_step_star_preservation.
      eapply wterm_free_iter_var_step_star_preservation; eauto.
      eapply wterm_free_iter_var_step_star_preservation; eauto.


    edestruct abisim_context as [sq'[bq'[??]]]; eauto.
    splits; auto.
    eapply s_iter_step_None_preservation; eauto.
    eapply bisim_context; eauto.
    apply step_intro_free_iter_var; eauto.

    intros s' ba' ?.
    edestruct step_bisim1 as [bb'[??]].
    eapply step_intro_free_iter_var; eauto. eauto.
    exists bb'; splits; auto.
      eapply step_star_elim_free_iter_var; eauto.
      eapply ss_iter_step_cons; eauto. apply s_iter_nil.
      eapply s_iter_step_None_preservation; eauto.
      eapply wterm_free_iter_var_step_preservation; eauto.
      eapply wterm_free_iter_var_step_star_preservation; eauto.

    apply bisim_symm; auto.
  Qed.

*)


  Lemma term_free_iter_var_subst: forall (t:iter_schema) (z:iteration_expr) i,
    free_iter_var t i ->
    (forall j n z1, z = i_stall j n z1 -> i<>j) ->
    term_free_iter_var (term_subst_iter t z) i.
  Proof.
    intros.
    induction t; simpl in *; try destruct H; auto.

    destruct z. simpl.
    split; eauto.
    destruct i0; simpl; auto.
    destruct i0; simpl; auto.
    destruct H.
    split; auto.
  Qed.

  Lemma wterm_free_iter_var_subst: forall b z i,
    wfree_iter_var b i ->
    (forall j n z1, z = i_stall j n z1 -> i<>j) ->
    wterm_free_iter_var (subst_iter b z) i.
  Proof.
    intros.
    induction b; simpl in *; try destruct H.
    apply term_free_iter_var_subst; auto.
    split; auto; apply term_free_iter_var_subst; auto.
    split; auto.
  Qed.


  Lemma step_star_s_iter_derived_paused: forall s b s' b' i,
    step_star s b s' b' ->
    s_iter_get s' i = Some iter_stall ->
    s_iter_get s i = Some iter_stall.
  Proof.
    intros.
    revert H0.
    induction H; intros; auto.
    unfold s_iter_get in *.
    erewrite step_s_iter. 2: eauto.
    eauto.
    inverts H.
    destruct (Pos.eq_dec i0 i); subst.
    rewrite s_iter_get_set_eq in IHstep_star.
    spec IHstep_star; auto; discriminate.
    rewrite s_iter_get_set_neq in IHstep_star; auto.
  Qed.


  Lemma cstep_star_s_iter_derived_paused: forall s C b s' C' b' i,
    cstep_star s C b s' C' b' ->
    s_iter_get s' i = Some iter_stall ->
    s_iter_get s i = Some iter_stall.
  Proof.
    intros.
    revert H0.
    induction H; intros; auto.
    unfold s_iter_get in *.
    erewrite cstep_s_iter. 2: eauto.
    eauto.
    inverts H.
    destruct (Pos.eq_dec i0 i); subst.
    rewrite s_iter_get_set_eq in IHcstep_star.
    spec IHcstep_star; auto; discriminate.
    rewrite s_iter_get_set_neq in IHcstep_star; auto.
  Qed.


(*
  Lemma cstep_star_i_stall_resolve: forall i n z1 z' s C b s' C' b',
    s_iter_get s' i = Some iter_stall ->
    cstep_star s C (subst_iter b (i_stall i n z1)) s' C' (subst_iter b' z') ->
    exists n', z' = i_stall i n' z1 /\ n>=n'.
Admitted.
*)

  Lemma step_add_iter: forall i n s b s' b',
    step s b s' b' ->
    s_iter_get s i = None ->
    step (s_iter_set s i n) b (s_iter_set s' i n) b'.
  Proof.
    intros.
    step_cases (induction H) Case HH; subst; eapply HH;
      try rewrite s_get_ch_iter_set;
      try solve [symmetry; apply s_ch_apply_iter_set];
      first [Case "st_newch" | eauto].
    Case "st_while_go".
      destruct (Pos.eq_dec i i0); subst.
      false.
      rewrite s_iter_get_set_neq; auto.
    Case "st_newch".
      rewrite fresh_channel_iter_set; reflexivity.
      rewrite fresh_channel_iter_set; reflexivity.
      rewrite fresh_channel_iter_set, s_new_ch_iter_set; reflexivity.
  Qed.

  Lemma cstep_add_iter: forall i n s C b s' C' b',
    cstep s C b s' C' b' ->
    s_iter_get s i = None ->
    cstep (s_iter_set s i n) C b (s_iter_set s' i n) C' b'.
  Proof.
    intros.
    cstep_cases (induction H) Case HH; subst; eapply HH; eauto.
    apply step_add_iter; auto.
    apply step_add_iter; auto.
    apply step_add_iter; auto.
  Qed.

  Lemma step_star_add_iter: forall i n s b s' b',
    step_star s b s' b' ->
    s_iter_get s i = None ->
    step_star (s_iter_set s i n) b (s_iter_set s' i n) b'.
  Proof.
    intros.
    induction H.
    apply step_nil.
    eapply step_cons.
    apply step_add_iter; eauto.
    apply IHstep_star.
    eapply s_iter_step_star_None_preservation; eauto; eapply step_single; eauto.
    inverts H.
    destruct (Pos.eq_dec i i0); subst.
    false.
    eapply st_iteration with (s_iter_set (s_iter_set s i n) i0 iter_go).
    constructor.
    rewrite s_iter_get_set_neq; auto.
    rewrite s_iter_set_set_neq'; auto.
    apply IHstep_star.
    rewrite s_iter_get_set_neq; auto.
  Qed.

  Lemma cstep_star_add_iter: forall i n s C b s' C' b',
    cstep_star s C b s' C' b' ->
    s_iter_get s i = None ->
    cstep_star (s_iter_set s i n) C b (s_iter_set s' i n) C' b'.
  Proof.
    intros.
    induction H.
    apply cstep_nil.
    eapply cstep_cons.
    apply cstep_add_iter; eauto.
    apply IHcstep_star.
    eapply s_iter_cstep_star_None_preservation; eauto; eapply cstep_single_cstep_star; eauto.
    inverts H.
    destruct (Pos.eq_dec i i0); subst.
    false.
    eapply cstep_iteration with (s_iter_set (s_iter_set s i n) i0 iter_go).
    constructor.
    rewrite s_iter_get_set_neq; auto.
    rewrite s_iter_set_set_neq'; auto.
    apply IHcstep_star.
    rewrite s_iter_get_set_neq; auto.
  Qed.

  Lemma step_add_iter_go: forall i s b s' b',
    step s b s' b' ->
    step (s_iter_set s i iter_go) b (s_iter_set s' i iter_go) b'.
  Proof.
    intros.
    step_cases (induction H) Case HH; subst; eapply HH;
      try rewrite s_get_ch_iter_set;
      try solve [symmetry; apply s_ch_apply_iter_set];
      first [Case "st_newch" | eauto].
    Case "st_while_go".
      destruct (Pos.eq_dec i i0); subst.
      rewrite s_iter_get_set_eq; auto.
      rewrite s_iter_get_set_neq; auto.
    Case "st_newch".
      rewrite fresh_channel_iter_set; reflexivity.
      rewrite fresh_channel_iter_set; reflexivity.
      rewrite fresh_channel_iter_set, s_new_ch_iter_set; reflexivity.
  Qed.

  Lemma cstep_add_iter_go: forall i s C b s' C' b',
    cstep s C b s' C' b' ->
    cstep (s_iter_set s i iter_go) C b (s_iter_set s' i iter_go) C' b'.
  Proof.
    intros.
    cstep_cases (induction H) Case HH; subst; eapply HH; eauto.
    apply step_add_iter_go; auto.
    apply step_add_iter_go; auto.
    apply step_add_iter_go; auto.
  Qed.

  Lemma step_star_add_iter_go: forall i s b s' b',
    step_star s b s' b' ->
    step_star (s_iter_set s i iter_go) b (s_iter_set s' i iter_go) b'.
  Proof.
    intros.
    induction H.
    apply step_nil.
    eapply step_cons.
    apply step_add_iter_go; eauto.
    apply IHstep_star.
    inverts H.
    destruct (Pos.eq_dec i i0); subst.
    rewrite s_iter_set_set_eq' in IHstep_star; auto.
    eapply st_iteration with (s_iter_set (s_iter_set s i iter_go) i0 iter_go).
    constructor.
    rewrite s_iter_get_set_neq; auto.
    rewrite s_iter_set_set_neq'; auto.
  Qed.


  Lemma cstep_star_add_iter_go: forall i s C b s' C' b',
    cstep_star s C b s' C' b' ->
    cstep_star (s_iter_set s i iter_go) C b (s_iter_set s' i iter_go) C' b'.
  Proof.
    intros.
    induction H.
    apply cstep_nil.
    eapply cstep_cons.
    apply cstep_add_iter_go; eauto.
    apply IHcstep_star.
    inverts H.
    destruct (Pos.eq_dec i i0); subst.
    rewrite s_iter_set_set_eq' in IHcstep_star; auto.
    eapply cstep_iteration with (s_iter_set (s_iter_set s i iter_go) i0 iter_go).
    constructor.
    rewrite s_iter_get_set_neq; auto.
    rewrite s_iter_set_set_neq'; auto.
  Qed.


  Lemma term_free_iter_var_subst_not_in_list: forall t i z,
    term_free_iter_var (term_subst_iter t z) i -> (forall n z1, z<>i_stall i n z1) \/ free_iter_var t i.
  Proof.
    intros.
    revert H.
    induction t; simpl; intros; auto.

    destruct IHt; auto; destruct z; simpl in *; try destruct H; destruct i0; auto.
    destruct IHt; auto; destruct z; simpl in *; destruct i0; try destruct H; auto.
    destruct IHt1; destruct IHt2; destruct H; auto.
    destruct IHt1; destruct IHt2; destruct H; auto.
    destruct IHt1; destruct IHt2; destruct H; auto.

    rename i0 into f.
(* need to add restrictions on f: that it doesn't mess with iteration variables *)
admit.
  Qed.

  Lemma step_star_halted_bisim2: forall s ba bb s' bb',
    step_star s bb s' bb' ->
    halted ba ->
    bisimilar s ba bb ->
    bisimilar s' bb bb'.
  Proof.
    intros.
    destruct H0 as [w ?]; subst ba.
    edestruct step_star_bisim2 as [ba'[??]]; eauto.
    edestruct step_star_skip_resolve as [??]; eauto.
    subst.
    eapply bisim_trans; eauto.
    apply bisim_symm.
    eapply bisim_s_iter_star; eauto.
  Qed.

  Lemma step_star_halted_bisim1: forall s ba bb s' ba',
    step_star s ba s' ba' ->
    halted bb ->
    bisimilar s ba bb ->
    bisimilar s' ba ba'.
  Proof.
    intros.
    destruct H0 as [w ?]; subst bb.
    edestruct step_star_bisim1 as [bb'[??]]; eauto.
    edestruct step_star_skip_resolve as [??]; eauto.
    subst.
    apply bisim_symm.
    eapply bisim_trans; eauto.
    apply bisim_symm.
    eapply bisim_s_iter_star; eauto.
  Qed.

  Lemma ss_iter_star_step_star: forall s b s' b',
    step_star s b s' b' ->
    ss_iter_star s s'.
  Proof.
    intros.
    induction H.
    apply s_iter_nil.
    eapply ss_iter_step_cons; eauto.
    eapply ss_iter_cons; eauto.
  Qed.

  Lemma s_iter_star_bisim_halted_step_star: forall s ba bb s' ba',
    step_star s ba s' ba' ->
    halted bb ->
    bisimilar s ba bb ->
    s_iter_star s s'.
  Proof.
    intros.
    edestruct step_star_bisim1 as [bb'[??]]; eauto.
    destruct H0 as [w ?]; subst bb.
    edestruct step_star_skip_resolve as [??]; eauto.
  Qed.


End Lemmas.





