Require Import Omega.
Require Import List.
Require Import FMapPositive.
Require Import FMapFacts.
Require Import world_model.
Require Import msl.msl_standard.
Require Import LibTactics.
Require Import SfLib.
Require Import MSetPositive.
Require Import MSetList.
Require Import MSetFacts.
Require Import MSetProperties.

Module PMapFacts := FMapFacts.WFacts_fun PositiveMap.E PositiveMap.

Section Iterations.

  CoInductive iterations : Set :=
  | i_nat : nat->iterations
  | i_inf : iterations
  .
  Coercion i_nat : nat >-> iterations.
  Definition i_minus (i:iterations) (n:nat) : iterations := match i, n with
  | i_inf, n => i_inf
  | i_nat m, n => m-n
  end.
  Infix "-" := i_minus : iter_scope.

  Inductive iteration_expr : Type :=
  |  i_stall : positive->nat->iterations->iteration_expr
  |  iter_const : iterations->iteration_expr.

Coercion iter_const : iterations >-> iteration_expr.
  Inductive iteration_decr : iteration_expr->iteration_expr->Prop :=
  | id_nat: forall n,
      iteration_decr (S n) n
  | id_inf:
      iteration_decr i_inf i_inf
  | id_stall: forall i n z,
      iteration_decr (i_stall i (S n) z) (i_stall i n z).

  Inductive iteration_decrs: iteration_expr->iteration_expr->Prop :=
  | ids_nil: forall z, iteration_decrs z z
  | ids_decr: forall z z' z'',
      iteration_decrs z z' ->
      iteration_decr z' z'' ->
      iteration_decrs z z''.

  Lemma iter_decrs_nat1: forall (n:nat) z',
    iteration_decrs n z' -> exists (n':nat), z'=n' /\ n>=n'.
  Proof.
    intros.
    remember (iter_const (i_nat n)) as z.
    revert Heqz.
    induction H; intros; subst; eauto.
    destruct IHiteration_decrs as [n'[??]]; auto.
    subst z'.
    inverts H0.
    exists n0; split; auto. omega.
  Qed.

  Lemma iter_decrs_nat2: forall z (n':nat),
    iteration_decrs z n' -> exists (n:nat), z=n /\ n>=n'.
  Proof.
    intros.
    remember (iter_const (i_nat n')) as z'.
    revert n' Heqz'.
    induction H; intros; subst; eauto.
    inverts H0.
    edestruct IHiteration_decrs as [n[??]]; eauto.
    subst z.
    exists n; split; auto. omega.
  Qed.

  Lemma iter_decrs_inf1: forall z',
    iteration_decrs i_inf z' -> z'=i_inf.
  Proof.
    intros.
    remember (iter_const (i_inf)) as z.
    revert Heqz.
    induction H; intros; subst; eauto.
    spec IHiteration_decrs; auto; subst z'.
    inverts H0; reflexivity.
  Qed.

  Lemma iter_decrs_inf2: forall z,
    iteration_decrs z i_inf -> z=i_inf.
  Proof.
    intros.
    remember (iter_const (i_inf)) as z'.
    revert Heqz'.
    induction H; intros; subst; eauto.
    inverts H0.
    spec IHiteration_decrs; auto.
  Qed.

    Definition i_plus (z1 z2:iterations) : iterations:=
    match z1, z2 with
    | i_inf, _ => i_inf
    | _, i_inf => i_inf
    | i_nat n1, i_nat n2 => n1+n2
    end.

    Definition iter_plus1 (z1:iterations) (z2:iteration_expr):=
    match z1, z2 with
    | i_inf, _ => i_inf
    | _, i_inf => i_inf
    | i_nat n1, i_nat n2 => n1+n2
    | i_nat n1, i_stall i n2 z => i_stall i (n1+n2) z
    end.

    Definition iter_plus2 (z1:iteration_expr) (z2:iterations):=
    match z1, z2 with
    | i_inf, _ => i_inf
    | i_nat n1, i_nat n2 => n1+n2
    | i_nat n1, i_inf => i_inf
    | i_stall i n z1, z2 => i_stall i n (i_plus z1 z2)
    end.


  Lemma i_minus_inf: forall n, i_minus i_inf n = i_inf.
  Proof. auto. Qed.

  Lemma i_minus_nat: forall (n m:nat), i_minus n m = n-m.
  Proof. auto. Qed.

  Fixpoint gt_iterations_nat (i:iterations) (n:nat) : Prop :=
    match i with
    | i_inf => True
    | i_nat m => m > n
    end.

  Fixpoint ge_iterations_nat (i:iterations) (n:nat) : Prop :=
    gt_iterations_nat i n \/ i=n.

End Iterations.






  Definition channel_state := ((value->pred world)*(list world)*(list value))%type.
  Definition ch_state := PositiveMap.t channel_state.

  Definition empty_ch_state : ch_state := PositiveMap.empty channel_state.
  Definition ch_eq (ch1 ch2: channel_state) := match ch1, ch2 with
    | (R1,w1,q1), (R2,w2,q2) => R1=R2 /\ w1=w2 /\ q1=q2
    end.

  Definition ch_state_eq := @PositiveMap.Equal channel_state.

  Inductive iter_state : Type := iter_stall | iter_go.
  Definition iters_state := PositiveMap.t iter_state.
  Definition empty_iters_state : iters_state := PositiveMap.empty iter_state.
  Definition iters_state_eq := @PositiveMap.Equal iter_state.

  Inductive Resource : Set :=
    | cname_res: cname->Resource | heap_res: addr->Resource.

  Module Resource_OWL <: OrderedTypeWithLeibniz.
    Definition t:= Resource.
    Definition eq : t->t->Prop.
      intros [c1|a1] [c2|a2].
      apply (eq c1 c2).
      apply False.
      apply False.
      apply (eq a1 a2).
    Defined.
    Instance eq_equiv : Equivalence eq.
    Proof.
      constructor.
      intros [c|a]; reflexivity.
      intros [c1|a1] [c2|a2] ?; simpl in *; auto.
      intros [c1|a1] [c2|a2] [c3|c3] ? ?; simpl in *; subst; auto; false.
    Qed.
    Definition lt: t->t->Prop.
      intros [c1|a1] [c2|a2].
      apply (Pos.lt c1 c2).
      apply True.
      apply False.
      apply (lt a1 a2).
    Defined.
    Instance lt_strorder: StrictOrder lt.
    Proof.
      constructor.
      red; intros [c|a]; simpl in *.
      apply Pos.lt_strorder; eauto.
      apply NPeano.Nat.lt_strorder; eauto.
      red; intros [c1|a1] [c2|a2] [c3|a3] ? ?; simpl in *; auto.
      eapply Pos.lt_strorder; eauto.
      false. false.
      eapply NPeano.Nat.lt_strorder; eauto.
    Qed.
    Instance lt_compat : Proper (eq==>eq==>iff) lt.
    Proof.
      intros [c1|a1] [c2|c2] ? [c3|a3] [c4|a4] ?; simpl in *; subst; intuition.
    Qed.
    Definition compare: t->t->comparison.
      intros [c1|a1] [c2|a2].
      apply (Pos.compare c1 c2).
      apply Lt.
      apply Gt.
      apply (Compare_dec.nat_compare a1 a2).
    Defined.
    Lemma compare_spec: forall (r1 r2: t), CompSpec eq lt r1 r2 (compare r1 r2).
    Proof.
      intros [c1|a1] [c2|a2].
      simpl.
      red; simpl; apply Pos.compare_spec.
      apply CompLt; simpl; auto.
      apply CompGt; simpl; auto.
      apply NPeano.Nat.compare_spec.
    Qed.
    Lemma eq_dec: forall (r1 r2:Resource), {eq r1 r2}+{~eq r1 r2}.
    Proof.
      intros [c1|a1] [c2|a2]; simpl; auto; do 2 decide equality.
    Qed.
    Definition eq_leibniz: forall x y, eq x y -> x = y.
      intros [c1|a1] [c2|a2]; simpl in *; intros; subst; auto; false.
    Qed.
  End Resource_OWL.

(*
  Module Nat_OWL <: OrderedTypeWithLeibniz.
    Include NPeano.Nat.
    Definition eq_leibniz: forall x y, eq x y -> x = y. introv H; exact H. Qed.
  End Nat_OWL.

  Module NatSet := MSetList.MakeWithLeibniz Nat_OWL.
*)
  Module ResSet := MSetList.MakeWithLeibniz Resource_OWL.
  Module ResSetFacts := MSetFacts.WFacts ResSet.
  Module ResSetProps := MSetProperties.WProperties ResSet.

  Record state : Type := makeState {
   s_c : ch_state;
   s_i : iters_state;
   (* set of all allocated resources *)
   s_all_res: ResSet.t
  }.

  Definition empty_state := makeState empty_ch_state empty_iters_state ResSet.empty.

  Definition gjoins' w1 w2 := {w | gjoin w1 w2 w}.

  Definition supd_all_res s all := match s with
  | {|s_c:=ch; s_i:=i; s_all_res:=_|} => {|s_c:=ch; s_i:=i; s_all_res:=all|}
  end.

  Definition supd_ch s ch := match s with
  | {|s_c:=_; s_i:=i; s_all_res:=all|} => {|s_c:=ch; s_i:=i; s_all_res:=all|}
  end.

  Definition supd_iter s i := match s with
  | {|s_c:=ch; s_i:=_; s_all_res:=all|} => {|s_c:=ch; s_i:=i; s_all_res:=all|}
  end.

  Definition s_get_ch s c := PositiveMap.find c (s_c s).

  Definition s_alloc_heap s (a:addr):=
    supd_all_res s (ResSet.add (heap_res a) (s_all_res s)).

  Definition s_alloc_ch s (c:cname):=
    supd_all_res s (ResSet.add (cname_res c) (s_all_res s)).

  Definition s_is_alloced_heap s a:=
    ResSet.In (heap_res a) (s_all_res s).

  Definition s_is_alloced_ch s (c:cname):=
    ResSet.In (cname_res c) (s_all_res s).

Module FreshChannels.

  Record FreshChannel : Type :=
  { fc_c: cname;
    fc_w: world
  }.

End FreshChannels.

Module FreshLocals.

  Record FreshLocal : Type :=
  { fl_x: var;
    fl_w: world
  }.

End FreshLocals.

Section Definitions.
  Import FreshChannels.
  Import FreshLocals.

  Definition max_cname_world (w: world) :
    {c:cname | forall c', (c<=c')%positive -> w_prod w c' = None /\ w_cons w c' = None}.
Admitted.

  Definition add_cname_world (c:cname) (w: world) :
    w_prod w c = None -> w_cons w c = None -> {w' | gjoin w (new_channel c) w'}.
Admitted.

  Definition max_cname_alloced (all: ResSet.t) :
    {c:cname | forall c', (c<=c')%positive -> ~ResSet.In (cname_res c') all}.
Admitted.

  Definition fresh_channel (s:state) (w:world) : FreshChannel.
    destruct (max_cname_world w) as [c1 ?].
    destruct (max_cname_alloced (s_all_res s)) as [c2 ?].
    remember (Pos.max c1 c2) as c.
    destruct (a c).
    subst c; apply Pos.le_max_l.
    destruct (add_cname_world c w) as [w' ?]; auto.
    apply (Build_FreshChannel c w').
  Defined.

  Lemma fresh_channel_global_props: forall s w fresh_ch,
    fresh_ch = fresh_channel s w ->
    let c:= fc_c fresh_ch in
      ~s_is_alloced_ch s c /\
      s_get_ch s c = None.
  Proof.
Admitted.

  Lemma fresh_channel_local_props: forall s w fresh_ch,
    fresh_ch = fresh_channel s w ->
    let c:= fc_c fresh_ch in
    let w':= fc_w fresh_ch in
      w_prod w c = None /\ w_cons w c = None /\ gjoin w (new_channel c) w'.
Admitted.


  Definition max_var_world (w: world) :
    {x:var | forall y, (x<=y)%positive -> w_env_get w y = None}.
Admitted.

  Definition add_var_world (x:var) (v:value) (w: world) :
    w_env_get w x = None -> {w' | gjoin w (new_local x v) w'}.
Admitted.

  Definition fresh_var (w:world) (v:value) : FreshLocal.
    destruct (max_var_world w) as [x ?].
    destruct (add_var_world x v w) as [w' ?]; auto.
    apply e. reflexivity.
    apply (Build_FreshLocal x w').
  Defined.

  Lemma fresh_var_local_props: forall w v fresh_x,
    fresh_x = fresh_var w v ->
    let x:= fl_x fresh_x in
    let w':= fl_w fresh_x in
      w_env_get w x = None /\ gjoin w (new_local x v) w'.
Admitted.


  Definition s_ch_apply s c f := supd_ch s
  match f (s_get_ch s c) with
  | Some (R, wl,vl) => PositiveMap.add c (R, wl, vl) (s_c s)
  | None => PositiveMap.remove c (s_c s)
  end.

  Definition s_set_ch s c w q := s_ch_apply s c (fun ch => match ch with
  | Some (R,_,_) => Some (R,w,q)
  | _ => None
  end).

  Definition s_push_ch s c w v := s_ch_apply s c (fun ch => match ch with
  | Some (R,wl,vl) => Some (R,wl++w::nil,vl++v::nil)
  | _ => None
  end).

  Definition s_pop_ch s c := s_ch_apply s c (fun ch => match ch with
  | Some (R,w::wl,v::vl) => Some (R,wl,vl)
  | Some cc => Some cc
  | _ => None
  end).

  Definition s_add_ch s c R w q := s_ch_apply s c (fun ch => Some (R,w,q)).

  Definition s_rm_ch s c := s_ch_apply s c (fun ch => None).

  Definition s_new_ch s (F:FreshChannel) R := match F with
  | {|fc_c:=c;fc_w:=_|} => s_add_ch (s_alloc_ch s c) c R nil nil
  end.

  Definition s_iter_set s z i := supd_iter s (PositiveMap.add z i (s_i s)).

  Definition s_iter_get s z := PositiveMap.find z (s_i s).

  Inductive s_iter : state->state->Prop :=
  | s_iter_go: forall s i,
      s_iter_get s i = Some iter_stall ->
      s_iter s (s_iter_set s i iter_go).

  Lemma s_iter_go': forall s s' i,
      s_iter_get s i = Some iter_stall ->
      s' = s_iter_set s i iter_go ->
      s_iter s s'.
  Proof. intros; subst; constructor; auto. Qed.

  Inductive s_iter_star : state->state->Prop :=
  | s_iter_nil : forall s, s_iter_star s s
  | s_iter_cons : forall s s' s'', s_iter s s' -> s_iter_star s' s'' -> s_iter_star s s''.


  Definition vname_eq : forall (x y : var), {x=y}+{x<>y} := Pos.eq_dec.

  Lemma values_eq : forall (v1 v2: value), {v1=v2}+{v1<>v2}.
  Proof.
    intros; induction v1; destruct v2; repeat decide equality.
  Qed.

Lemma state_eq: forall s1 s2,
  (ch_state_eq (s_c s1) (s_c s2)) ->
  (s_i s1) = (s_i s2) ->
  (s_all_res s1) = (s_all_res s2) ->
  s1 = s2.
Admitted.


Fixpoint join_world_list (wl: list world) (ww: world) : Prop := 
 match wl with
  | w::wl' => exists w', join_world_list wl' w' /\ join w w' ww
  | nil => ww=gwemp
  end.

Lemma join_world_list_functional: forall wl w1 w2,
  join_world_list wl w1 -> join_world_list wl w2 -> w1 = w2.
Admitted.

Lemma join_world_list_dec wl : {w | join_world_list wl w}+{forall w, ~join_world_list wl w}.
Admitted.


Fixpoint join_state_list (sl: list (cname * channel_state)) (ww:world) : Prop := match sl with
  | (_, (_,wl,_))::sl' => exists w1, exists w2, join_state_list sl' w1 /\ join_world_list wl w2 /\ join w1 w2 ww
  | nil => ww=gwemp
  end.

Definition join_state s w := join_state_list (PositiveMap.elements (s_c s)) w.

Lemma join_state_functional: forall s w1 w2,
  join_state s w1 -> join_state s w2 -> w1 = w2.
Admitted.

Lemma join_state_dec s : {w | join_state s w}+{forall w, ~join_state s w}.
Admitted.


End Definitions.





Module Morphisms.
  Require Export Setoid.
  Require Export Morphisms.
  Export SetoidTactics.
  Require Export SetoidClass.

  Definition state_equiv s1 s2 :=
    ch_state_eq (s_c s1) (s_c s2) /\
    PositiveMap.Equal (s_i s1) (s_i s2) /\
    ResSet.Equal (s_all_res s1) (s_all_res s2).

  Program Instance ChStateSetoid : Setoid ch_state :=
      { equiv:= ch_state_eq }.

  Program Instance IterStateSetoid : Setoid (@PositiveMap.t iter_state) :=
      { equiv:= PositiveMap.Equal }.

  Program Instance ResSetSetoid : Setoid ResSet.t :=
      { equiv:= ResSet.Equal }.

  Program Instance StateSetoid : Setoid state :=
      { equiv:= state_equiv }.
  Next Obligation.
    constructor.
    intros s; splits; reflexivity.
    intros s1 s2 [?[??]]; splits; symmetry; auto.
    intros s1 s2 s3 [?[??]] [?[??]]; splits; etransitivity; eauto.
  Qed.



  Add Parametric Morphism: makeState
  with signature equiv ==> equiv ==> equiv ==> equiv
  as makeState_m.
  Proof. intros; split; auto. Qed.

  Add Parametric Morphism: supd_ch
  with signature equiv ==> equiv ==> equiv
  as supd_ch_m.
  Proof.
    intros; auto.
    destruct x; destruct y.
    apply makeState_m; auto; apply H.
  Qed.

  Add Parametric Morphism: supd_iter
  with signature equiv ==> equiv ==> equiv
  as supd_iter_m.
  Proof.
    intros; auto.
    destruct x; destruct y.
    apply makeState_m; auto; apply H.
  Qed.

  Add Parametric Morphism: supd_all_res
  with signature equiv ==> equiv ==> equiv
  as supd_all_res_m.
  Proof.
    intros; auto.
    destruct x; destruct y.
    apply makeState_m; auto; apply H.
  Qed.

  Add Parametric Morphism: s_c
  with signature equiv ==> equiv
  as s_c_m.
  Proof.
    intros; auto; apply H.
  Qed.

  Add Parametric Morphism: s_i
  with signature equiv ==> equiv
  as s_i_m.
  Proof.
    intros; auto; apply H.
  Qed.

  Add Parametric Morphism: s_all_res
  with signature equiv ==> equiv
  as s_all_res_m.
  Proof.
    intros; auto; apply H.
  Qed.

  Add Parametric Morphism: s_get_ch
  with signature equiv ==> eq ==> eq
  as s_get_ch_m.
  Proof.
    intros; auto; apply H.
  Qed.

  Add Parametric Morphism: s_ch_apply
  with signature equiv ==> eq ==> eq ==> equiv
  as s_ch_apply_m.
  Proof.
    intros; auto.
    unfold s_ch_apply.
    case_eq (y1 (s_get_ch x y0)); intros.
    setoid_rewrite H in H0.
    rewrite H0.
    destruct p as [[??]?].
    setoid_rewrite H.
    reflexivity.
    setoid_rewrite H in H0.
    rewrite H0.
    setoid_rewrite H.
    reflexivity.
  Qed.

  Add Parametric Morphism: s_pop_ch
  with signature equiv ==> eq ==> equiv
  as s_pop_ch_m.
  Proof. intros; apply s_ch_apply_m; auto. Qed.

  Add Parametric Morphism: s_push_ch
  with signature equiv ==> eq ==> eq ==> eq ==> equiv
  as s_push_ch_m.
  Proof. intros; apply s_ch_apply_m; auto. Qed.

  Add Parametric Morphism: s_add_ch
  with signature equiv ==> eq ==> eq ==> eq ==> eq ==> equiv
  as s_add_ch_m.
  Proof. intros; apply s_ch_apply_m; auto. Qed.

  Add Parametric Morphism: s_rm_ch
  with signature equiv ==> eq ==> equiv
  as s_rm_ch_m.
  Proof. intros; apply s_ch_apply_m; auto. Qed.

  Add Parametric Morphism: s_alloc_ch
  with signature equiv ==> eq ==> equiv
  as s_alloc_ch_m.
  Proof.
    intros; unfold s_alloc_ch.
    setoid_rewrite H; reflexivity.
  Qed.

  Add Parametric Morphism: s_new_ch
  with signature equiv ==> eq ==> eq ==> equiv
  as s_new_ch_m.
  Proof.
    intros; unfold s_new_ch.
    destruct y0.
    setoid_rewrite H; reflexivity.
  Qed.

  Lemma state_equiv_eq: forall (s1 s2: state),
    s1 == s2 -> s1=s2.
Admitted.

End Morphisms.











Section BasicLemmas.
  Import Morphisms.

  Ltac destruct_state := match reverse goal with
    s : state |- _ => destruct s
  end.

  Ltac basic := intros; repeat destruct_state; reflexivity.

  Lemma s_iter_respects_all_res: forall s s',
    s_iter s s' -> s_all_res s' = s_all_res s.
  Proof. intros; inverts H; basic. Qed.

  Lemma s_iter_respects_ch: forall s s',
    s_iter s s' -> s_c s' = s_c s.
  Proof. intros; inverts H; basic. Qed.

  Lemma supd_ch_respects_s_i: forall s ch,
    s_i (supd_ch s ch) = s_i s.
  Proof. basic. Qed.

  Lemma supd_ch_respects_s_all_res: forall s ch,
    s_all_res (supd_ch s ch) = s_all_res s.
  Proof. basic. Qed.


  Lemma supd_all_res_respects_s_c: forall s H,
    s_c (supd_all_res s H) = s_c s.
  Proof. basic. Qed.

  Lemma supd_iter_respects_s_c: forall s ii,
    s_c (supd_iter s ii) = s_c s.
  Proof. basic. Qed.

  Lemma s_get_ch_supd_ch: forall s c ch,
    s_get_ch (supd_ch s ch) c = PositiveMap.find c ch.
  Proof. basic. Qed.

  Lemma supd_ch_redundant: forall s ch1 ch2,
    supd_ch (supd_ch s ch1) ch2 = supd_ch s ch2.
  Proof. basic. Qed.

  Lemma s_c_supd_ch: forall s ch,
    s_c (supd_ch s ch) = ch.
  Proof. basic. Qed.

  Lemma supd_ch_eq: forall s ch1 ch2,
    ch_state_eq ch1 ch2 ->
    supd_ch s ch1 = supd_ch s ch2.
  Proof.
    intros; apply state_eq; auto.
    repeat rewrite s_c_supd_ch; auto.
    repeat rewrite supd_ch_respects_s_i; reflexivity.
    repeat rewrite supd_ch_respects_s_all_res; reflexivity.
  Qed.

  Lemma s_iter_get_supd: forall s ch i,
    s_iter_get (supd_ch s ch) i = s_iter_get s i.
  Proof. basic. Qed.

  Lemma s_iter_get_ch_apply: forall s c f i,
    s_iter_get (s_ch_apply s c f) i = s_iter_get s i.
  Proof. basic. Qed.

  Lemma s_iter_set_ch_apply: forall s c f i z,
    s_iter_set (s_ch_apply s c f) i z = s_ch_apply (s_iter_set s i z) c f.
  Proof. basic. Qed.

  Lemma s_get_ch_s_iter_noninterference: forall s s' c,
    s_iter s s' ->
    s_get_ch s' c = s_get_ch s c.
  Proof. intros; inverts H; basic. Qed.

  Lemma s_i_supd_iter: forall s ii,
    s_i (supd_iter s ii) = ii.
  Proof. basic. Qed.

  Lemma supd_iter_redundant: forall s ii1 ii2,
    supd_iter (supd_iter s ii1) ii2 = supd_iter s ii2.
  Proof. basic. Qed.

  Lemma s_iter_s_get_ch: forall s s' c,
    s_iter s s' -> s_get_ch s c = s_get_ch s' c.
  Proof. intros; inverts H; basic. Qed.

  Lemma supd_ch_respects_all_res: forall s s' ch,
    s' = supd_ch s ch ->
    s_all_res s' = s_all_res s.
  Proof. intros; inverts H; basic. Qed.

  Lemma s_ch_apply_respects_all_res: forall s s' c f,
    s' = s_ch_apply s c f ->
    s_all_res s' = s_all_res s.
  Proof. intros; inverts H; basic. Qed.

  Lemma s_pop_ch_respects_all_res: forall s s' c,
    s' = s_pop_ch s c ->
    s_all_res s' = s_all_res s.
  Proof. intros; eapply s_ch_apply_respects_all_res; eauto. Qed.

  Lemma s_ch_apply_supd_all_res: forall s all c f,
    s_ch_apply (supd_all_res s all) c f = supd_all_res (s_ch_apply s c f) all.
  Proof. basic. Qed.

  Lemma s_pop_ch_supd_all_res: forall s all c,
    s_pop_ch (supd_all_res s all) c = supd_all_res (s_pop_ch s c) all.
  Proof. intros; apply s_ch_apply_supd_all_res; auto. Qed.

  Lemma s_all_res_s_ch_apply: forall s c f,
    s_all_res (s_ch_apply s c f) = s_all_res s.
  Proof. basic. Qed.

  Lemma s_all_res_s_pop_ch: forall s c,
    s_all_res (s_pop_ch s c) = s_all_res s.
  Proof. intros; apply s_all_res_s_ch_apply; auto. Qed.

  Lemma s_get_ch_supd_all_res: forall s all' c,
    s_get_ch (supd_all_res s all') c = s_get_ch s c.
  Proof.
    unfold s_get_ch; intros.
    f_equal; apply supd_all_res_respects_s_c.
  Qed.

  Lemma supd_all_res_respects_s_i: forall s all,
    s_i (supd_all_res s all) = s_i s.
  Proof. basic. Qed.

  Lemma s_ch_apply_respects_s_i: forall s c f,
    s_i (s_ch_apply s c f) = s_i s.
  Proof. intros; apply supd_ch_respects_s_i. Qed.

  Lemma s_push_ch_respects_s_i: forall s c w v,
    s_i (s_push_ch s c w v) = s_i s.
  Proof. intros; apply s_ch_apply_respects_s_i. Qed.

  Lemma s_pop_ch_respects_s_i: forall s c,
    s_i (s_pop_ch s c) = s_i s.
  Proof. intros; apply s_ch_apply_respects_s_i. Qed.

  Lemma s_add_ch_respects_s_i: forall s c R wl vl,
    s_i (s_add_ch s c R wl vl) = s_i s.
  Proof. intros; apply s_ch_apply_respects_s_i. Qed.

  Lemma s_new_ch_respects_s_i: forall s fch R,
    s_i (s_new_ch s fch R) = s_i s.
  Proof.
    unfold s_new_ch; intros.
    destruct fch.
    rewrite s_add_ch_respects_s_i.
    apply supd_all_res_respects_s_i.
  Qed.

  Lemma s_rm_ch_respects_s_i: forall s c,
    s_i (s_rm_ch s c) = s_i s.
  Proof. intros; apply s_ch_apply_respects_s_i. Qed.

  Lemma s_iter_get_push_ch: forall s c w v,
    s_iter_get (s_push_ch s c w v) = s_iter_get s.
  Proof. basic. Qed.
  Lemma s_iter_get_pop_ch: forall s c,
    s_iter_get (s_pop_ch s c) = s_iter_get s.
  Proof. basic. Qed.
  Lemma s_iter_get_add_ch: forall s c R wl vl,
    s_iter_get (s_add_ch s c R wl vl) = s_iter_get s.
  Proof. basic. Qed.
  Lemma s_iter_get_rm_ch: forall s c,
    s_iter_get (s_rm_ch s c) = s_iter_get s.
  Proof. basic. Qed.
  Lemma s_iter_get_new_ch: forall s fch R,
    s_iter_get (s_new_ch s fch R) = s_iter_get s.
  Proof. destruct fch; basic. Qed.

  Lemma s_iter_set_respects_s_c: forall s i z,
    s_c (s_iter_set s i z) = s_c s.
  Proof. basic. Qed.

  Lemma s_iter_set_respects_s_all_res: forall s i z,
    s_all_res (s_iter_set s i z) = s_all_res s.
  Proof. basic. Qed.

  Lemma s_iter_respects_s_c: forall s s',
    s_iter s s' -> s_c s = s_c s'.
  Proof.
    intros. inverts H. symmetry; apply supd_iter_respects_s_c.
  Qed.

  Lemma s_iter_respects_s_all_res: forall s s',
    s_iter s s' -> s_all_res s = s_all_res s'.
  Proof.
    intros. inverts H. symmetry; apply s_iter_set_respects_s_all_res.
  Qed.

  Lemma s_iter_star_respects_s_c: forall s s',
    s_iter_star s s' -> s_c s = s_c s'.
  Proof.
    intros. induction H; auto. rewrite<- IHs_iter_star. apply s_iter_respects_s_c; auto.
  Qed.

  Lemma s_iter_star_respects_s_all_res: forall s s',
    s_iter_star s s' -> s_all_res s = s_all_res s'.
  Proof.
    intros. induction H; auto. rewrite<- IHs_iter_star. apply s_iter_respects_s_all_res; auto.
  Qed.


  Lemma s_get_ch_iter_set: forall s i n c,
    s_get_ch (s_iter_set s i n) c = s_get_ch s c.
  Proof. basic. Qed.

  Lemma s_c_get_ch: forall s s' c,
    s_c s = s_c s' ->
    s_get_ch s c = s_get_ch s' c.
  Proof. intros; unfold s_get_ch; congruence. Qed.

  Lemma s_ch_apply_supd_iter: forall s ii c f,
    s_ch_apply (supd_iter s ii) c f = supd_iter (s_ch_apply s c f) ii.
  Proof. basic. Qed.

  Lemma s_push_ch_supd_iter: forall s ii c w v,
    s_push_ch (supd_iter s ii) c w v = supd_iter (s_push_ch s c w v) ii.
  Proof. basic. Qed.

  Lemma s_pop_ch_supd_iter: forall s ii c,
    s_pop_ch (supd_iter s ii) c = supd_iter (s_pop_ch s c) ii.
  Proof. basic. Qed.

  Lemma s_add_ch_supd_iter: forall s ii c R wl vl,
    s_add_ch (supd_iter s ii) c R wl vl = supd_iter (s_add_ch s c R wl vl) ii.
  Proof. basic. Qed.

  Lemma s_rm_ch_supd_iter: forall s ii c,
    s_rm_ch (supd_iter s ii) c = supd_iter (s_rm_ch s c) ii.
  Proof. basic. Qed.

  Lemma s_new_ch_supd_iter: forall s ii fch R,
    s_new_ch (supd_iter s ii) fch R = supd_iter (s_new_ch s fch R) ii.
  Proof. intros; destruct fch; basic. Qed.

  Lemma supd_all_res_iter_set: forall s all i z,
    supd_all_res (s_iter_set s i z) all = s_iter_set (supd_all_res s all) i z.
  Proof. basic. Qed.

  Lemma s_all_res_supd_all_res: forall s r,
    s_all_res (supd_all_res s r) = r.
  Proof. basic. Qed.

  Lemma supd_all_res_redundant: forall s all all',
    supd_all_res (supd_all_res s all') all = supd_all_res s all.
  Proof. basic. Qed.

  Lemma s_alloc_ch_neq: forall s c1 c2,
    s_alloc_ch (s_alloc_ch s c1) c2 == s_alloc_ch (s_alloc_ch s c2) c1.
  Proof.
    intros.
    unfold s_alloc_ch.
    repeat rewrite supd_all_res_redundant.
    repeat rewrite s_all_res_supd_all_res.
    rewrite ResSetProps.add_add.
    reflexivity.
  Qed.

  Lemma s_pop_ch_alloc_ch: forall s c c',
    s_pop_ch (s_alloc_ch s c') c = s_alloc_ch (s_pop_ch s c) c'.
  Proof. basic. Qed.

  Lemma s_push_ch_alloc_ch: forall s c w v c',
    s_push_ch (s_alloc_ch s c') c w v = s_alloc_ch (s_push_ch s c w v) c'.
  Proof. basic. Qed.

  Lemma s_add_ch_alloc_ch: forall s c R wl vl c',
    s_add_ch (s_alloc_ch s c') c R wl vl = s_alloc_ch (s_add_ch s c R wl vl) c'.
  Proof. basic. Qed.

  Lemma s_rm_ch_alloc_ch: forall s c c',
    s_rm_ch (s_alloc_ch s c') c = s_alloc_ch (s_rm_ch s c) c'.
  Proof. basic. Qed.

  Lemma s_new_ch_alloc_ch: forall s fch R c',
    s_new_ch (s_alloc_ch s c') fch R == s_alloc_ch (s_new_ch s fch R) c'.
  Proof.
    intros; destruct fch.
    unfold s_new_ch.
    setoid_rewrite s_alloc_ch_neq.
    rewrite s_add_ch_alloc_ch.
    reflexivity.
  Qed.

  Lemma s_i_iter_set: forall s i z,
    s_i (s_iter_set s i z) = PositiveMap.add i z (s_i s).
  Proof. basic. Qed.

  Lemma s_iter_is_alloced_ch: forall s s',
    s_iter s s' ->
    forall c, s_is_alloced_ch s c <-> s_is_alloced_ch s' c.
  Proof.
    intros.
    cut (s_all_res s == s_all_res s'); intros.
    apply H0.
    symmetry; erewrite s_iter_respects_all_res; eauto.
    reflexivity.
  Qed.

  Lemma s_push_ch_respects_s_all_res: forall s c w v,
    s_all_res (s_push_ch s c w v) = s_all_res s.
  Proof. intros; apply s_all_res_s_ch_apply. Qed.
  Lemma s_pop_ch_respects_s_all_res: forall s c,
    s_all_res (s_pop_ch s c) = s_all_res s.
  Proof. intros; apply s_all_res_s_ch_apply. Qed.
  Lemma s_add_ch_respects_s_all_res: forall s c R wl vl,
    s_all_res (s_add_ch s c R wl vl) = s_all_res s.
  Proof. intros; apply s_all_res_s_ch_apply. Qed.
  Lemma s_rm_ch_respects_s_all_res: forall s c,
    s_all_res (s_rm_ch s c) = s_all_res s.
  Proof. intros; apply s_all_res_s_ch_apply. Qed.
  Lemma s_is_alloced_ch_ch_apply: forall s c1 f c2,
    s_is_alloced_ch (s_ch_apply s c1 f) c2 <-> s_is_alloced_ch s c2.
  Proof. intros; unfold s_is_alloced_ch; rewrite s_all_res_s_ch_apply; reflexivity. Qed.

  Lemma s_is_alloced_ch_push_ch: forall s c1 w v c2,
    s_is_alloced_ch (s_push_ch s c1 w v) c2 <-> s_is_alloced_ch s c2.
  Proof. intros; apply s_is_alloced_ch_ch_apply. Qed.
  Lemma s_is_alloced_ch_pop_ch: forall s c1 c2,
    s_is_alloced_ch (s_pop_ch s c1) c2 <-> s_is_alloced_ch s c2.
  Proof. intros; apply s_is_alloced_ch_ch_apply. Qed.
  Lemma s_is_alloced_ch_add_ch: forall s c1 R wl vl c2,
    s_is_alloced_ch (s_add_ch s c1 R wl vl) c2 <-> s_is_alloced_ch s c2.
  Proof. intros; apply s_is_alloced_ch_ch_apply. Qed.
  Lemma s_is_alloced_ch_rm_ch: forall s c1 c2,
    s_is_alloced_ch (s_rm_ch s c1) c2 <-> s_is_alloced_ch s c2.
  Proof. intros; apply s_is_alloced_ch_ch_apply. Qed.

  Lemma s_is_alloced_heap_ch_apply: forall s c1 f c2,
    s_is_alloced_heap (s_ch_apply s c1 f) c2 <-> s_is_alloced_heap s c2.
  Proof. intros; unfold s_is_alloced_heap; rewrite s_all_res_s_ch_apply; reflexivity. Qed.
  Lemma s_is_alloced_heap_push_ch: forall s c1 w v c2,
    s_is_alloced_heap (s_push_ch s c1 w v) c2 <-> s_is_alloced_heap s c2.
  Proof. intros; apply s_is_alloced_heap_ch_apply. Qed.
  Lemma s_is_alloced_heap_pop_ch: forall s c1 c2,
    s_is_alloced_heap (s_pop_ch s c1) c2 <-> s_is_alloced_heap s c2.
  Proof. intros; apply s_is_alloced_heap_ch_apply. Qed.
  Lemma s_is_alloced_heap_add_ch: forall s c1 R wl vl c2,
    s_is_alloced_heap (s_add_ch s c1 R wl vl) c2 <-> s_is_alloced_heap s c2.
  Proof. intros; apply s_is_alloced_heap_ch_apply. Qed.
  Lemma s_is_alloced_heap_rm_ch: forall s c1 c2,
    s_is_alloced_heap (s_rm_ch s c1) c2 <-> s_is_alloced_heap s c2.
  Proof. intros; apply s_is_alloced_heap_ch_apply. Qed.

  Lemma s_is_alloced_ch_new_ch: forall s fch R c,
    s_is_alloced_ch s c -> s_is_alloced_ch (s_new_ch s fch R) c.
  Proof.
    unfold s_is_alloced_ch; intros.
    destruct fch.
    simpl. unfold s_add_ch.
    setoid_rewrite s_all_res_s_ch_apply.
    destruct s; simpl in *.
    apply ResSetProps.FM.add_2; auto.
  Qed.

  Lemma s_is_alloced_heap_new_ch: forall s fch R c,
    s_is_alloced_heap (s_new_ch s fch R) c <-> s_is_alloced_heap s c.
  Proof.
    intros; unfold s_is_alloced_heap; destruct fch; destruct s; red.
    simpl.
    split; intros.
    apply ResSetProps.FM.add_3 in H; auto.
    apply ResSetProps.FM.add_2; auto.    
 Qed.


End BasicLemmas.













Section IterLemmas.
  Import PMapFacts.
  Import Morphisms.

  Lemma s_iter_get_set_neq: forall i j s n,
    i <> j -> s_iter_get (s_iter_set s i n) j = s_iter_get s j.
  Proof.
    intros. unfold s_iter_get,s_iter_set.
    rewrite s_i_supd_iter.
    rewrite add_neq_o; auto.
  Qed.

  Lemma s_iter_get_set_eq: forall i s n,
    s_iter_get (s_iter_set s i n) i = Some n.
  Proof.
    intros. unfold s_iter_get,s_iter_set.
    rewrite s_i_supd_iter.
    rewrite add_eq_o; auto.
  Qed.

  Lemma s_iter_get_set_None_neq: forall i j s n ,
    s_iter_get (s_iter_set s i n) j = None -> i<>j.
  Proof.
    unfold s_iter_get,s_iter_set; intros.
    rewrite s_i_supd_iter in H.
    destruct (Pos.eq_dec i j); subst; auto.
    rewrite add_eq_o in H; auto.
    false.
  Qed.

  Lemma s_iter_set_set_eq: forall i s n m,
    s_iter_set (s_iter_set s i n) i m == s_iter_set s i m.
  Proof.
    intros.
    split.
    repeat rewrite s_iter_set_respects_s_c; reflexivity.
    split.
    repeat rewrite s_i_iter_set.
    intro. repeat rewrite add_o.
    destruct (eq_dec i y); reflexivity.
    repeat rewrite s_iter_set_respects_s_all_res; reflexivity.
  Qed.

  Lemma s_iter_set_set_eq': forall i s n m,
    s_iter_set (s_iter_set s i n) i m = s_iter_set s i m.
  Proof. intros; apply state_equiv_eq; apply s_iter_set_set_eq. Qed.

  Lemma s_iter_get_eq_s_i: forall s s' i,
    s_i s = s_i s' -> s_iter_get s i = s_iter_get s' i.
  Proof.
    intros. destruct s; destruct s'; simpl in *; subst; auto.
  Qed.

  Lemma s_iter_change_s_i: forall s s', s_iter s s' -> s_i s <> s_i s'.
  Proof.
    intros.
    inverts H.
    intro.
    apply s_iter_get_eq_s_i with (i:=i) in H.
    rewrite s_iter_get_set_eq in H.
    false.
  Qed.

  Lemma s_iter_change: forall s s', s_iter s s' -> s <> s'.
  Proof.
    intros.
    inverts H.
    intro.
    rewrite H in H0.
    rewrite s_iter_get_set_eq in H0.
    discriminate.
  Qed.

  Lemma s_iter_set_set_neq: forall s i j n1 n2,
    i <> j ->
    s_iter_set (s_iter_set s i n1) j n2 == s_iter_set (s_iter_set s j n2) i n1.
  Proof.
    intros.
    split.
    repeat rewrite s_iter_set_respects_s_c; reflexivity.
    split.
    repeat rewrite s_i_iter_set.
    intro. repeat rewrite add_o.
    destruct (eq_dec i y); destruct (eq_dec j y); subst; auto.
    false.
    repeat rewrite s_iter_set_respects_s_all_res; reflexivity.
  Qed.

  Lemma s_iter_set_set_neq': forall s i j n1 n2,
    i <> j ->
    s_iter_set (s_iter_set s i n1) j n2 = s_iter_set (s_iter_set s j n2) i n1.
  Proof. intros; apply state_equiv_eq; apply s_iter_set_set_neq; auto. Qed.

  Lemma s_iter_noninterfering: forall s s' i ndi,
    s_iter s s' ->
    s_iter_get s i = None ->
    s_iter (s_iter_set s i ndi) (s_iter_set s' i ndi).
  Proof.
    intros.
    inverts H.
    destruct (Pos.eq_dec i i0); subst.
    false.
    rewrite s_iter_set_set_neq'; auto.
    econstructor.
    rewrite s_iter_get_set_neq; auto.
  Qed.



End IterLemmas.











Section ChannelLemmas.
  Import FreshChannels.
  Import PMapFacts.

  Lemma fresh_channel_cname_total_world_eq: forall s1 s2 w,
    (forall c, s_get_ch s1 c = None <-> s_get_ch s2 c=None) ->
    (forall c, s_is_alloced_ch s1 c <-> s_is_alloced_ch s2 c) ->
    fc_c (fresh_channel s1 w) = fc_c (fresh_channel s2 w).
  Proof.
(* depends on implementation of fresh_channel *)
Admitted.

  Lemma fresh_channel_world_total_world_eq: forall s1 s2 w,
   (forall c, s_get_ch s1 c = None <-> s_get_ch s2 c = None) ->
    (forall c, s_is_alloced_ch s1 c <-> s_is_alloced_ch s2 c) ->
    fc_w (fresh_channel s1 w) = fc_w (fresh_channel s2 w).
  Proof.
    intros.
    generalize (fresh_channel_cname_total_world_eq s1 s2 w H H0); intro.
    edestruct fresh_channel_local_props with (s:=s1) (w:=w) as [?[??]]; eauto.
    edestruct fresh_channel_local_props with (s:=s2) (w:=w) as [?[??]]; eauto.
    destruct (fresh_channel s1 w) as [c wS' w' [?[?[??]]] [?[??]]].
    destruct (fresh_channel s2 w) as [c_ wS'_ w'_ [?[?[??]]] [?[??]]].
    simpl in *.
    subst c_.
    eapply gjoin_eq; eauto.
  Qed.

  Lemma fresh_channel_s_ch_apply: forall s c w f,
    (s_get_ch s c <> None -> f (s_get_ch s c) <> None) ->
    fresh_channel (s_ch_apply s c f) w = fresh_channel s w.
  Proof.
Admitted.

  Lemma fresh_channel_iter_set: forall s i z w,
    fresh_channel (s_iter_set s i z) w = fresh_channel s w.
  Proof.
Admitted.

  Lemma s_ch_apply_eq: forall s c f1 f2,
    (forall ch, f1 ch = f2 ch) ->
    s_ch_apply s c f1 = s_ch_apply s c f2.
  Proof.
    intros.
    apply supd_ch_eq.
    unfold ch_state_eq, PositiveMap.Equal.
    intros.
    rewrite H.
    reflexivity.
  Qed.


  Lemma s_ch_get_apply_diff: forall s c1 c2 f,
    c1<>c2 ->
    s_get_ch (s_ch_apply s c1 f) c2 = s_get_ch s c2.
  Proof.
    unfold s_ch_apply, s_get_ch; intros; simpl.
    rewrite s_c_supd_ch.
    destruct (f (PositiveMap.find c1 (s_c s))) as [ [[R wl]vl] | ].
    rewrite add_neq_o; auto.
    rewrite remove_neq_o; auto.
  Qed.

  Lemma s_ch_get_apply_same: forall s c ch f,
    s_get_ch s c = ch ->
    s_get_ch (s_ch_apply s c f) c = f ch.
  Proof.
    unfold s_ch_apply,s_get_ch; intros.
    rewrite s_c_supd_ch.
    rewrite H.
    destruct (f ch) as [ [[R wl]vl] | ].
    rewrite add_eq_o; auto.
    rewrite remove_eq_o; auto.
  Qed.

  Lemma s_ch_apply_diff: forall s c1 f1 c2 f2,
    c1<>c2 ->
    s_ch_apply (s_ch_apply s c1 f1) c2 f2 = s_ch_apply (s_ch_apply s c2 f2) c1 f1.
  Proof. 
    unfold s_ch_apply, s_get_ch; intros.
    apply state_eq.
    unfold ch_state_eq, supd_ch, PositiveMap.Equal.
    intros.
    destruct s as [sc si wS]. simpl.
    clear si wS.
    case_eq (f1 (PositiveMap.find c1 sc)); intros.
    destruct p as [[R1 wl1]vl1].
    case_eq (f2 (PositiveMap.find c2 sc)); intros.
    destruct p as [[R2 wl2]vl2].
    do 2 (rewrite add_neq_o; auto).
    setoid_rewrite H0; setoid_rewrite H1.
    destruct (Pos.eq_dec y c1); subst.
    rewrite add_neq_o; auto.
    repeat rewrite add_eq_o; auto.
    destruct (Pos.eq_dec y c2); subst.
    rewrite add_eq_o; auto.
    rewrite add_neq_o; auto.
    rewrite add_eq_o; auto.
    repeat rewrite add_neq_o; auto.
    rewrite add_neq_o; auto.
    rewrite remove_neq_o; auto.
    setoid_rewrite H0; setoid_rewrite H1.
    destruct (Pos.eq_dec y c1); subst.
    rewrite remove_neq_o; auto.
    rewrite add_eq_o; auto.
    rewrite add_eq_o; auto.
    rewrite add_neq_o; auto.
    destruct (Pos.eq_dec y c2); subst.
    repeat rewrite remove_eq_o; auto.
    repeat rewrite remove_neq_o; auto.
    rewrite add_neq_o; auto.
    case_eq (f2 (PositiveMap.find c2 sc)); intros.
    destruct p as [[R2 wl2]vl2].
    rewrite add_neq_o with (x:=c2) (y:=c1); auto.
    setoid_rewrite H0.
    rewrite remove_neq_o; auto.
    rewrite H1.
    destruct (Pos.eq_dec y c1); subst.
    rewrite add_neq_o; auto.
    repeat rewrite remove_eq_o; auto.
    destruct (Pos.eq_dec y c2); subst.
    rewrite remove_neq_o; auto.
    repeat rewrite add_eq_o; auto.
    rewrite add_neq_o; auto.
    repeat rewrite remove_neq_o; auto.
    rewrite add_neq_o; auto.
    repeat rewrite remove_neq_o; auto.
    rewrite H0; rewrite H1.
    destruct (Pos.eq_dec y c1); subst.
    rewrite remove_neq_o; auto.
    repeat rewrite remove_eq_o; auto.
    destruct (Pos.eq_dec y c2); subst.
    rewrite remove_eq_o; auto.
    rewrite remove_neq_o; auto.
    rewrite remove_eq_o; auto.
    repeat rewrite remove_neq_o; auto.

    repeat rewrite supd_ch_respects_s_i.
    reflexivity.

    repeat rewrite supd_ch_respects_s_all_res.
    reflexivity.
  Qed.

  Lemma s_pop_ch_add_ch: forall c1 c2 s R wl vl,
    c1<>c2 ->
    s_pop_ch (s_add_ch s c1 R wl vl) c2 = s_add_ch (s_pop_ch s c2) c1 R wl vl.
  Proof. intros; apply s_ch_apply_diff; auto. Qed.

  Lemma s_ch_pop_pop_diff: forall s c1 c2,
    c1<>c2 ->
    s_pop_ch (s_pop_ch s c1) c2 = s_pop_ch (s_pop_ch s c2) c1.
  Proof.
    intros; eapply s_ch_apply_diff; eauto.
  Qed.

  Lemma s_ch_apply_compose: forall s c f1 f2,
    s_ch_apply (s_ch_apply s c f1) c f2 = s_ch_apply s c (fun ch => f2 (f1 ch)).
  Proof.
    intros; unfold s_ch_apply.
    case_eq (f1 (s_get_ch s c)); intros.
    destruct p as [[R wl]vl].
    rewrite s_get_ch_supd_ch.
    rewrite add_eq_o; auto.
    rewrite supd_ch_redundant.
    case_eq (f2 (Some (R,wl,vl))); intros.
    destruct p as [[R' wl']vl'].
    rewrite s_c_supd_ch.
    apply supd_ch_eq; unfold ch_state_eq, PositiveMap.Equal.
    intros.
    destruct (Pos.eq_dec y c); subst.
    rewrite add_eq_o; auto.
    rewrite add_eq_o; auto.
    repeat rewrite add_neq_o; auto.
    (**)
    apply supd_ch_eq; unfold ch_state_eq, PositiveMap.Equal.
    intros.
    rewrite s_c_supd_ch.
    destruct (Pos.eq_dec y c); subst.
    repeat rewrite remove_eq_o; auto.
    repeat rewrite remove_neq_o; auto.
    rewrite add_neq_o; auto.
    (**)
    rewrite supd_ch_redundant.
    rewrite s_get_ch_supd_ch.
    repeat rewrite remove_eq_o; auto.
    case_eq (f2 None); intros.
    destruct p as [[R' wl']vl'].
    setoid_rewrite H0.
    rewrite s_c_supd_ch.
    apply supd_ch_eq; unfold ch_state_eq, PositiveMap.Equal.
    intros.
    destruct s as [sc si wS]. simpl.
    destruct (Pos.eq_dec y c); subst.
    repeat rewrite add_eq_o; auto.
    repeat rewrite add_neq_o; auto.
    rewrite remove_neq_o; auto.
    (**)
    setoid_rewrite H0.
    rewrite s_c_supd_ch.
    apply supd_ch_eq; unfold ch_state_eq, PositiveMap.Equal.
    intros.
    destruct (Pos.eq_dec y c); subst.
    repeat rewrite remove_eq_o; auto.
    rewrite remove_neq_o; auto.
  Qed.

  Lemma s_ch_push_pop_same_comm: forall R wc wcl v q w' v' s c,
    s_get_ch s c = Some (R, wc :: wcl, v :: q) ->
    s_push_ch (s_pop_ch s c) c w' v' = s_pop_ch (s_push_ch s c w' v') c.
  Proof.
    intros.
    unfold s_pop_ch, s_push_ch; intros.
    do 2 rewrite s_ch_apply_compose.
    apply supd_ch_eq; unfold ch_state_eq, PositiveMap.Equal.
    intros.
    rewrite H.
    reflexivity.
  Qed.

  Lemma s_ch_push_pop_diff_comm: forall w' v' s c1 c2,
    c1 <> c2 ->
    s_push_ch (s_pop_ch s c1) c2 w' v' = s_pop_ch (s_push_ch s c2 w' v') c1.
  Proof. intros; apply s_ch_apply_diff; auto. Qed.

  Lemma s_ch_get_pop_diff: forall s c1 c2,
    c1 <> c2 -> s_get_ch (s_pop_ch s c1) c2 = s_get_ch s c2.
  Proof. intros; apply s_ch_get_apply_diff; auto. Qed.

  Lemma s_ch_get_pop_same: forall R wc wl v q s c,
    s_get_ch s c = Some (R, wc :: wl, v :: q) ->
    s_get_ch (s_pop_ch s c) c = Some (R, wl, q).
  Proof.
    intros.
    unfold s_pop_ch.
    erewrite s_ch_get_apply_same; eauto.
    reflexivity.
  Qed.

  Lemma s_ch_get_push_same: forall R w wl v q s c,
    s_get_ch s c = Some (R, wl, q) ->
    s_get_ch (s_push_ch s c w v) c = Some (R, wl++w::nil, q++v::nil).
  Proof.
    intros.
    unfold s_push_ch.
    erewrite s_ch_get_apply_same; eauto.
    reflexivity.
  Qed.

  Lemma s_ch_get_pop_neq_None: forall s c c',
    s_get_ch s c <> None <->
    s_get_ch (s_pop_ch s c') c <> None.
  Proof.
    intros.
    unfold s_pop_ch. unfold s_ch_apply.
    rewrite s_get_ch_supd_ch.
    case_eq (s_get_ch s c'); intros.
    destruct c0 as [[R wl] ?].
    destruct wl.
    destruct (Pos.eq_dec c' c); subst.
    rewrite add_eq_o; auto.
    split; intros; intro; congruence.
    rewrite add_neq_o; auto. split; auto.
    destruct l.
    destruct (Pos.eq_dec c' c); subst.
    rewrite add_eq_o; auto. split; intros; intro; congruence.
    rewrite add_neq_o; auto. split; auto.
    destruct (Pos.eq_dec c' c); subst.
    rewrite add_eq_o; auto. split; intros; intro; congruence.
    rewrite add_neq_o; auto. split; auto.
    destruct (Pos.eq_dec c' c); subst.
    rewrite remove_eq_o; auto.
    split; intros; congruence.
    rewrite remove_neq_o; auto. split; auto.
  Qed.

  Lemma s_ch_get_pop_eq_None: forall s c c',
    s_get_ch s c = None <->
    s_get_ch (s_pop_ch s c') c = None.
  Proof.
    intros.
    destruct s_ch_get_pop_neq_None with s c c' as [? ?].
    split; intro.
    destruct (s_get_ch (s_pop_ch s c') c); auto.
    spec H0; congruence; auto.
    destruct (s_get_ch s c); auto.
    spec H; congruence; auto.
  Qed.

  Lemma s_ch_get_pop_eq_None1: forall s c c',
    s_get_ch s c = None ->
    s_get_ch (s_pop_ch s c') c = None.
  Proof. intros; apply s_ch_get_pop_eq_None; auto. Qed.

  Lemma fresh_channel_unique_cname: forall s w c,
    c = fc_c (fresh_channel s w) ->
    s_get_ch s c = None /\ w_prod w c = None /\ w_cons w c = None.
  Proof.
    intros. subst c.
    edestruct fresh_channel_local_props with (s:=s) (w:=w) as [?[??]].
    reflexivity.
    splits; auto.
    edestruct fresh_channel_global_props with (s:=s) (w:=w) as [??]; eauto.
  Qed.

  Lemma fresh_channel_unique_cname1: forall s w,
    s_get_ch s (fc_c (fresh_channel s w)) = None.
  Proof. intros; eapply fresh_channel_unique_cname; eauto. Qed.

  Lemma fresh_channel_s_pop_ch: forall s c w,
    fresh_channel (s_pop_ch s c) w = fresh_channel s w.
  Proof.
    intros; apply fresh_channel_s_ch_apply.
    intros.
    destruct (s_get_ch s c) as [[[R wl]vl] |]; auto.
    destruct wl.
    congruence.
    destruct vl; congruence.
  Qed.

  Lemma s_ch_pop_newch_comm: forall s w c R fresh_ch,
    s_get_ch s c <> None ->
    fresh_ch = fresh_channel s w ->
    s_pop_ch (s_new_ch s fresh_ch R) c = s_new_ch (s_pop_ch s c) (fresh_channel (s_pop_ch s c) w) R.
  Proof.
    intros.
    unfold s_new_ch.
    rewrite fresh_channel_s_pop_ch.
    edestruct fresh_channel_global_props with (s:=s) (w:=w) as [??]; eauto.
    subst fresh_ch.
    destruct (fresh_channel s w) as [c' wS' w'].
    rewrite s_pop_ch_add_ch; auto.
    rewrite s_pop_ch_alloc_ch; auto.
    intro; subst c'; simpl in *; false.
  Qed.

  Lemma s_ch_pop_rm_diff_comm: forall s c1 c2,
    c1 <> c2 ->
    s_pop_ch (s_rm_ch s c1) c2 = s_rm_ch (s_pop_ch s c2) c1.
  Proof. intros; eapply s_ch_apply_diff; eauto. Qed.

  Lemma s_ch_get_push_diff: forall w v s c c',
    c <> c' ->
    s_get_ch (s_push_ch s c' w v) c = s_get_ch s c.
  Proof. intros; eapply s_ch_get_apply_diff; eauto. Qed.

  Lemma s_ch_get_add_same: forall s c R wl vl,
    s_get_ch (s_add_ch s c R wl vl) c = Some (R,wl,vl).
  Proof. intros; unfold s_add_ch; erewrite s_ch_get_apply_same; eauto. Qed.

  Lemma s_ch_get_add_diff: forall s c c' R wl vl,
    c' <> c ->
    s_get_ch (s_add_ch s c R wl vl) c' = s_get_ch s c'.
  Proof. intros; unfold s_add_ch; erewrite s_ch_get_apply_diff; eauto. Qed.

  Lemma s_ch_get_newch_same: forall R s w c fresh_ch,
    fresh_ch = fresh_channel s w ->
    c = fc_c fresh_ch ->
    s_get_ch (s_new_ch s fresh_ch R) c = Some (R,nil,nil).
  Proof.
    intros.
    unfold s_new_ch.
    subst.
    destruct (fresh_channel s w) as [c' wS' w'].
    simpl in *.
    apply s_ch_get_add_same.
  Qed.

  Lemma s_ch_get_newch_diff: forall R s w c c' fresh_ch,
    c' <> c ->
    fresh_ch = fresh_channel s w ->
    c = fc_c fresh_ch ->
    s_get_ch (s_new_ch s fresh_ch R) c' = s_get_ch s c'.
  Proof.
    intros.
    unfold s_new_ch.
    destruct (fresh_channel s w) as [c_ wS' w'].
    subst.
    simpl in *.
    rewrite s_ch_get_add_diff; auto.
    apply s_get_ch_supd_all_res.
  Qed.

  Lemma s_ch_push_push_diff: forall s c1 c2 w1 w2 v1 v2,
    c1 <> c2 -> s_push_ch (s_push_ch s c1 w1 v1) c2 w2 v2 = s_push_ch (s_push_ch s c2 w2 v2) c1 w1 v1.
  Proof. intros; apply s_ch_apply_diff; auto. Qed.

  Lemma s_ch_get_rm_same: forall s c,
    s_get_ch (s_rm_ch s c) c = None.
  Proof. intros; unfold s_rm_ch; erewrite s_ch_get_apply_same; auto. Qed.


End ChannelLemmas.







Section Lemmas.
  Import FreshChannels.
  Import PMapFacts.

  Lemma supd_ch_s_iter_commute1: forall s ch s'',
    s_iter (supd_ch s ch) s'' ->
    exists s',
      s_iter s s' /\ s'' = supd_ch s' ch.
  Proof.
    intros.
    destruct s.
    inverts H.
    eexists; split.
    constructor; eauto.
    reflexivity.
  Qed.

  Lemma s_ch_apply_s_iter_commute1: forall s s'' c f,
    s_iter (s_ch_apply s c f) s'' ->
    exists s',
      s_iter s s' /\
      s'' = s_ch_apply s' c f.
  Proof.
    intros.
    inverts H.
    rewrite s_iter_get_ch_apply in H0.
    exists (s_iter_set s i iter_go).
    split.
    constructor; auto.
    apply s_iter_set_ch_apply.
  Qed.

  Lemma s_push_ch_s_iter_commute1: forall s c w v s'',
    s_iter (s_push_ch s c w v) s'' ->
    exists s',
      s_iter s s' /\ s'' = s_push_ch s' c w v.
  Proof. intros; eapply s_ch_apply_s_iter_commute1; eauto. Qed.

  Lemma s_pop_ch_s_iter_commute1: forall s c s'',
    s_iter (s_pop_ch s c) s'' ->
    exists s'0,
      s_iter s s'0 /\ s'' = s_pop_ch s'0 c.
  Proof. intros; eapply s_ch_apply_s_iter_commute1; eauto. Qed.

  Lemma supd_all_res_s_iter_commute1: forall s all s'',
    s_iter (supd_all_res s all) s'' ->
    exists s'0,
      s_iter s s'0 /\ s'' = supd_all_res s'0 all.
  Proof.
    intros.
    destruct s.
    inverts H.
    eexists; split.
    constructor; eauto.
    reflexivity.
  Qed.

  Lemma s_alloc_ch_s_iter_commute1: forall s c s'',
    s_iter (s_alloc_ch s c) s'' ->
    exists s'0,
      s_iter s s'0 /\ s'' = s_alloc_ch s'0 c.
  Proof.
    intros.
    destruct s.
    inverts H.
    eexists; split.
    constructor; eauto.
    reflexivity.
  Qed.

  Lemma fresh_channel_eq: forall s1 s2 w,
    s_c s1 = s_c s2 ->
    s_all_res s1 = s_all_res s2 ->
    fresh_channel s1 w = fresh_channel s2 w.
Admitted.

  Lemma s_new_ch_s_iter_commute1: forall s w R s'',
    s_iter (s_new_ch s (fresh_channel s w) R) s'' ->
    exists s'0,
      s_iter s s'0 /\ s'' = s_new_ch s'0 (fresh_channel s'0 w) R.
  Proof.
    unfold s_new_ch; intros.
    case_eq (fresh_channel s w). intros c w' ?.
    rewrite H0 in H.
    edestruct s_ch_apply_s_iter_commute1 as [s'0[??]]; eauto.
    edestruct s_alloc_ch_s_iter_commute1 as [s'1[??]]; eauto.
    exists s'1; split; auto.
    subst.
    rewrite (fresh_channel_eq s'1 s w).
    rewrite H0.
    reflexivity.
    eapply s_iter_respects_ch; eauto.
    eapply s_iter_respects_all_res; eauto.
  Qed.

  Lemma s_rm_ch_s_iter_commute1: forall s c s'',
    s_iter (s_rm_ch s c) s'' ->
    exists s',
      s_iter s s' /\ s'' = s_rm_ch s' c.
  Proof. intros; eapply s_ch_apply_s_iter_commute1; eauto. Qed.

  Lemma s_ch_apply_iter_set: forall s i z c f,
    s_ch_apply (s_iter_set s i z) c f = s_iter_set (s_ch_apply s c f) i z.
  Proof.
    unfold s_iter_set; intros.
    rewrite s_ch_apply_supd_iter.
    erewrite s_ch_apply_respects_s_i; reflexivity.
  Qed.

  Lemma s_push_ch_iter_set: forall s i z c w v,
    s_push_ch (s_iter_set s i z) c w v = s_iter_set (s_push_ch s c w v) i z.
  Proof. intros; apply s_ch_apply_iter_set. Qed.

  Lemma s_pop_ch_iter_set: forall s i z c,
    s_pop_ch (s_iter_set s i z) c = s_iter_set (s_pop_ch s c) i z.
  Proof. intros; apply s_ch_apply_iter_set. Qed.

  Lemma s_add_ch_iter_set: forall s i z c R wl vl,
    s_add_ch (s_iter_set s i z) c R wl vl = s_iter_set (s_add_ch s c R wl vl) i z.
  Proof. intros; apply s_ch_apply_iter_set. Qed.

  Lemma s_rm_ch_iter_set: forall s i z c,
    s_rm_ch (s_iter_set s i z) c = s_iter_set (s_rm_ch s c) i z.
  Proof. intros; apply s_ch_apply_iter_set. Qed.

  Lemma s_alloc_ch_iter_set: forall s i z c,
    s_alloc_ch (s_iter_set s i z) c = s_iter_set (s_alloc_ch s c) i z.
  Proof.
    intros; unfold s_alloc_ch.
    rewrite s_iter_set_respects_s_all_res, supd_all_res_iter_set.
    reflexivity.
  Qed.

  Lemma s_new_ch_iter_set: forall s i z fch R,
    s_new_ch (s_iter_set s i z) fch R = s_iter_set (s_new_ch s fch R) i z.
  Proof.
    intros; destruct fch; simpl.
    rewrite s_alloc_ch_iter_set.
    apply s_add_ch_iter_set.
  Qed.

End Lemmas.







Section IterationLemmas.

  Lemma iter_plus2_iterations: forall (z1 z2: iterations),
    iter_plus2 z1 z2 = i_plus z1 z2.
  Proof.
    intros; simpl.
    destruct z1; auto.
    destruct z2; reflexivity.
  Qed.

  Lemma iter_plus1_iterations: forall (z1 z2: iterations),
    iter_plus1 z1 z2 = i_plus z1 z2.
  Proof.
    intros; simpl.
    destruct z1; auto.
    destruct z2; reflexivity.
  Qed.

  Lemma i_plus_assoc: forall z1 z2 z3,
    i_plus z1 (i_plus z2 z3) = i_plus (i_plus z1 z2) z3.
  Proof.
    intros.
    destruct z1; auto.
    destruct z2; auto.
    destruct z3; auto.
    simpl. rewrite plus_assoc. auto.
  Qed.

  Lemma i_plus_comm: forall z1 z2,
    i_plus z1 z2 = i_plus z2 z1.
  Proof.
    intros.
    destruct z1; destruct z2; auto.
    simpl; rewrite plus_comm; auto.
  Qed.

  Lemma iter_plus_assoc: forall z1 z2 z3,
    iter_plus1 z1 (iter_plus2 z2 z3) = iter_plus2 (iter_plus1 z1 z2) z3.
  Proof.
    intros; simpl.
    destruct z1; auto.
    destruct z2; auto.
    destruct i; auto.
    destruct z3; auto.
    simpl. rewrite plus_assoc. reflexivity.
  Qed.

  Lemma i_plus_inf: forall z,
    i_plus z i_inf = i_inf.
  Proof.
    intros.
    destruct z; auto.
  Qed.

  Lemma i_plus_zero1: forall z,
    i_plus 0 z = z.
  Proof.
    intros. destruct z; reflexivity.
  Qed.

  Lemma iter_plus2_i_stall: forall i z1 z2 zz1 z2b,
    i_stall i z1 z2 = iter_plus2 zz1 z2b ->
    exists z2a, zz1 = i_stall i z1 z2a /\ z2 = i_plus z2a z2b.
  Proof.
    intros.
    destruct zz1; try solve [inverts H].
    simpl in H. inverts H.
    exists i0; split; auto.
    rewrite iter_plus2_iterations in H. inverts H.
  Qed.

  Lemma iter_plus2_nat: forall (n:nat) z1 z2,
    iter_const n = iter_plus2 z1 z2 ->
    exists (n1 n2: nat), z1=n1 /\ z2=n2 /\ n=n1+n2.
  Proof.
    intros.
    destruct z1; destruct z2; simpl in *; try discriminate.
    destruct i; try discriminate.
    inverts H. eauto.
    destruct i; try discriminate.
  Qed.

  Lemma iteration_decr_iter_plus2_resolve: forall z1 z2 z',
    iteration_decr (iter_plus2 z1 z2) z' ->
    (z1=0 /\ iteration_decr z2 z') \/
    exists z1', iteration_decr z1 z1' /\ z' = (iter_plus2 z1' z2).
  Proof.
    intros.
    inverts H; simpl in *.
    destruct z1; destruct z2; simpl H1; try discriminate.
    destruct i; simpl in *; try discriminate.
    destruct n1; inverts H1.
    left; split; auto; constructor.
    right; exists n1; split; simpl; auto; constructor.
    destruct i; inverts H1.
    destruct z1; destruct z2; simpl H1; try discriminate.
    destruct i; simpl in *; try discriminate.
    right; exists i_inf; split; auto; constructor.
    destruct i; simpl in *; try discriminate.
    destruct n.
    left; split; constructor.
    right; exists n; split; auto; constructor.
    right; exists i_inf; split; auto; constructor.
    destruct z1; simpl H1; try discriminate.
    inverts H1.
    right; exists (i_stall p n i0); split; constructor.
    rewrite iter_plus2_iterations in H1.
    inverts H1.
  Qed.


End IterationLemmas.


