(*
 * Copyright (c) 2009-2010, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Unset Automatic Introduction.
Generalizable All Variables.

Require Import base.
Require Import ageable.
Require Import sepalg.

Class ASA (A:Type) {as_alg : sepalg A} {as_age : ageable A} :=
mkASA
{ unage_identity:
    forall x z : A, age x z -> identity z -> identity x
; age1_join : forall x y z x', join x y z -> age x x' ->
    exists y':A, exists z':A, join x' y' z' /\ age y y' /\ age z z'
; age1_join2 : forall x y z z', join x y z -> age z z' ->
    exists x':A, exists y':A, join x' y' z' /\ age x x' /\ age y y'
; unage_join : forall x x' y' z', join x' y' z' -> age x x' ->
    exists y:A, exists z:A, join x y z /\ age y y' /\ age z z'
; unage_join2 : forall z x' y' z', join x' y' z' -> age z z' ->
    exists x:A, exists y:A, join x y z /\ age x x' /\ age y y'
}.

Implicit Arguments age1_join [[A] [as_alg] [as_age] x y z x'].
Implicit Arguments age1_join2 [[A] [as_alg] [as_age] x y z z'].
Implicit Arguments unage_join [[A] [as_alg] [as_age] x x' y' z'].
Implicit Arguments unage_join2 [[A] [as_alg] [as_age] z x' y' z'].

Lemma age1_None_joins `{ASA A}: forall phi1 phi2, joins phi1 phi2 -> age1 phi1 = None -> age1 phi2 = None.
Proof.
  intros A HA XA.
  intros.
  destruct H0.
  case_eq (age1 phi2); intros; auto.
  destruct (age1_join _ (join_com H0) H2)  as [phi1' [x' [? [? ?]]]].
  unfold age in *; rewrite H1 in H4; inv H4.
Qed.

Lemma age1_joins_eq `{ASA A}: forall phi1 phi2,
        joins phi1 phi2 -> 
        forall phi1', age1 phi1 = Some phi1' -> 
        forall phi2', age1 phi2 = Some phi2' ->
        joins phi1' phi2'.
Proof.
  intros A HA XA XA'.
  intros.
  destruct H.
  destruct (age1_join _ H H0) as [phi7 [x' [? [? ?]]]].
  unfold age in *; rewrite H1 in H3; inv H3.
  exists x'; auto.
Qed.

Require Import sepalg_generators.

Section BIJECTION.
  Variable A B : Type.
  Variable sa: sepalg A.
  Variable ag: ageable A.
  Variable bijAB: bijection A B.
  Variable asa : ASA A.

  Definition agB := (bij_ageable _ _ ag bijAB).
  Definition saB := (sa_bijection _ _ bijAB sa).

  Existing Instance agB.
  Existing Instance saB.

  Lemma asa_bij_unage_identity:
    forall x z : B, age x z -> identity z -> identity x.
  Proof.
    unfold age; simpl.
    intros x z.
    repeat rewrite identity_unit_equiv.
    simpl; unfold join_bijection.
    destruct bijAB as [f g fg gf]; simpl.
    case_eq (age1 (g x)); intros; try discriminate.
    inv H0.
    simpl.
    cut (identity (g x)).
    rewrite identity_unit_equiv; auto.
    intros.
    apply unage_identity with a; auto.
    rewrite gf in H1.
    rewrite identity_unit_equiv; auto.
  Qed.

  Theorem asa_bijection : @ASA B saB agB.
  Proof.
    constructor.
    exact asa_bij_unage_identity.

    (* commute1 *)
    unfold age; simpl.
    destruct bijAB as [f g fg gf].
    simpl in *.
    intros.
    revert H0; case_eq (age1 (g x)); intros; try discriminate.
    inv H1.
    rename a into gx'.
    destruct (age1_join _ H H0) as  [gy' [gz' [? [? ?]]]].
    exists (f gy'); exists (f gz').
    repeat rewrite gf.
    split; auto.
    rewrite H2; rewrite H3.
    auto.

    (* commute2 *)
    unfold age; simpl.
    destruct bijAB as [f g fg gf].
    simpl in *.
    intros.
    revert H0; case_eq (age1 (g z)); intros; try discriminate.
    inv H1.
    rename a into gz'.
    destruct (age1_join2 _ H H0) as  [gx' [gy' [? [? ?]]]].
    exists (f gx'); exists (f gy').
    repeat rewrite gf.
    split; auto.
    rewrite H2; rewrite H3.
    auto.
    
    (* commute3 *)
    unfold age; simpl.
    destruct bijAB as [f g fg gf].
    simpl in *.
    intros.
    revert H0; case_eq (age1 (g x)); intros; try discriminate.
    inv H1.
    rename a into gx'.
    rewrite gf in *.
    destruct (unage_join _ H H0) as  [gy' [gz' [? [? ?]]]].
    exists (f gy'); exists (f gz').
    repeat rewrite gf.
    split; auto.
    rewrite H2; rewrite H3.
    repeat rewrite fg; split; auto.
    
    (* commute4 *)
    unfold age; simpl.
    destruct bijAB as [f g fg gf].
    simpl in *.
    intros.
    revert H0; case_eq (age1 (g z)); intros; try discriminate.
    inv H1.
    rename a into gz'.
    rewrite gf in *.
    destruct (unage_join2 _ H H0) as  [gx' [gy' [? [? ?]]]].
    exists (f gx'); exists (f gy').
    repeat rewrite gf.
    split; auto.
    rewrite H2; rewrite H3.
    repeat rewrite fg; split; auto.
  Qed.
End BIJECTION.

Section PROD.
  Variable A B : Type.
  Variable saA : sepalg A.
  Variable agA : ageable A.
  Variable saB : sepalg B.
  Variable asa : ASA A.

  Theorem asa_prod : @ASA (prod A B) (sa_prod saA saB) (prod_ageable A B agA).
  Proof.
    constructor; unfold age; simpl; unfold join_prod.

    intros.
    destruct x; destruct z; simpl in *.
    case_eq (age1 a); intros; rewrite H1 in H; inv H.
    cut (identity a0); intros.
    assert (identity a).
    eapply unage_identity; eauto.
    rewrite identity_unit_equiv; split; simpl; auto.
    rewrite identity_unit_equiv in H2; auto.
    rewrite identity_unit_equiv in H0; destruct H0; auto.
    rewrite identity_unit_equiv.
    rewrite identity_unit_equiv in H0; destruct H0; auto.

    (* commute1 *)
    intros [xa xb] [ya yb] [za zb] [xa' xb'] [? ?].
    simpl in *.
    case_eq (age1 xa); intros; inv H2.
    destruct (age1_join _ H H1) as [ya' [za' [? [? ?]]]].
    exists (ya',yb); exists (za',zb); 
      rewrite H3; rewrite H4; repeat split; auto.

    (* commute2 *)
    intros [xa xb] [ya yb] [za zb] [za' zb'] [? ?].
    simpl in *.
    case_eq (age1 za); intros; inv H2.
    destruct (age1_join2 _ H H1) as [xa' [ya' [? [? ?]]]].
    exists (xa',xb); exists (ya',yb); 
      rewrite H3; rewrite H4; repeat split; auto.

    (* commute3 *)
    intros [xa xb] [xa' xb'] [ya' yb'] [za' zb'] [? ?].
    simpl in *.
    case_eq (age1 xa); intros; inv H2.
    destruct (unage_join _ H H1) as [ya [za [? [? ?]]]].
    exists (ya,yb'); exists (za,zb'); simpl.
    rewrite H3; rewrite H4; repeat split; auto.

    (* commute4 *)
    intros [za zb] [xa' xb'] [ya' yb'] [za' zb'] [? ?].
    simpl in *.
    case_eq (age1 za); intros; inv H2.
    destruct (unage_join2 _ H H1) as [xa [ya [? [? ?]]]].
    exists (xa,xb'); exists (ya,yb'); simpl.
    rewrite H3; rewrite H4; repeat split; auto.
  Qed.
End PROD.


Lemma joins_fashionR `{ASA A} : forall x y,
  joins x y -> fashionR x y.
Proof.  
  intros.
  unfold fashionR.
  destruct H0 as [z ?].
  revert y z H0; induction x using age_induction; intros.
  case_eq (age1 x); intros.
  destruct (age1_join _ H1 H2) as [p [q [? [? ?]]]].
  assert (level a = level p).
  apply H0 with q; auto.
  replace (level x) with (S (level a)).
  replace (level y) with (S (level p)).
  f_equal; auto.
  symmetry; apply age_level; auto.
  symmetry; apply age_level; auto.
  case_eq (age1 y); intros.
  apply join_com in H1.
  destruct (age1_join _ H1 H3) as [p [q [? [? ?]]]].
  hnf in H5; rewrite H5 in H2; discriminate.
  rewrite age1_level0 in H2.
  rewrite age1_level0 in H3.
  congruence.
Qed.

Lemma comparable_fashionR `{ASA A} : forall x y,
  comparable x y -> fashionR x y.
Proof.
  intros.
  apply comparable_common_unit in H0.
  destruct H0 as [u [H1 H2]].
  hnf; transitivity (level u).
  apply joins_fashionR; eauto.
  symmetry.
  apply joins_fashionR; eauto.
Qed.

Lemma age_identity `{ASA A}: forall phi phi', age phi phi'->
    identity phi -> identity phi'.
Proof.
intros A saA agA asaA.
intros.
unfold identity in *.
intros.
destruct (unage_join _ H1 H) as [y [? [? [? ?]]]].
spec H0 y x H2.
subst.
unfold age in *. congruence.
Qed.

Lemma age_comparable `{ASA A}:
    forall phi1 phi2 phi1' phi2', age phi1 phi1' -> age phi2 phi2' ->
      comparable phi1 phi2 -> comparable phi1' phi2'.
Proof.
  intros A saA agA asaA.
  intros.
  destruct (comparable_common_unit H1) as [e [? ?]].
  destruct (age1_join _ (join_com H2) H) as [a [b [? [? ?]]]].
  destruct (age1_join _ (join_com H3) H0) as [c [d [? [? ?]]]].
  assert (c=a) by (unfold age in *; congruence); subst c.
  assert (b=phi1') by (unfold age in *; congruence). subst b.
  assert (d=phi2') by (unfold age in *; congruence). subst d.
  apply common_unit_comparable.
  exists a.
  split; apply join_com; auto.
Qed.

Section NAT.

  Lemma nat_ASA : @ASA nat (sa_equiv nat) nat_ageable.
  Proof.
    constructor.
    simpl; intros. hnf in H.
    rewrite identity_unit_equiv. hnf; auto.
    intros.
    destruct H; subst.
    exists x'. exists x'.
    intuition. split; auto.
    intros.
    destruct H; subst.
    exists z'; exists z'; intuition.
    split; auto.
    intros. destruct H; subst.
    exists x; exists x. intuition. split; auto.
    intros. destruct H; subst.
    exists z; exists z; intuition. split; auto.
  Qed.
End NAT.

Section FLAT.

Variable A : Type.

Lemma flat_ageable_facts: ageable_facts A (fun _ => 0) (fun _ => None).
  intros.
  constructor.
  intro; intros. constructor. intros; discriminate.
  intros; discriminate.
  intuition.
  intros; discriminate.
Qed.

Definition flat_ageable : ageable A :=
   mkAgeable _ _ _ flat_ageable_facts.

Lemma flat_asa (H: sepalg A) : @ASA A H flat_ageable.
  intros.
  constructor; intros. inv H0.
  inv H1.
  inv H1.
  inv H1.
  inv H1.
Qed.

End FLAT.
