(*
 * Copyright (c) 2009-2010, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import sepalg.
Require Import sepalg_generators.

(* This file defines a series of lemmas for proving functions from the carrier set
   of one SA to the carrier set of another are join homomorphisms, and then shows
   that some simple properties of SA elems, like identity, comparability, etc.,
   are preserved by join homs.  The idea is due to Rob.
*)

(* Two-argument join homomorphisms *)
Section join_hom2.
  Variables (A B C: Type)
            (saA: sepalg A)
            (saB: sepalg B)
            (saC: sepalg C)
            (f: A -> B -> C).

  Definition join_hom2 := forall (a1 a2 a3: A)(b1 b2 b3: B),
      join a1 a2 a3
      -> join b1 b2 b3
      -> join (f a1 b1) (f a2 b2) (f a3 b3).

  (* Two-argument join homomorphisms with a dummy argument *)
  Definition join_hom2' := forall (a1 a2 a3: A)(b1 b2 b3: B),
      join a1 a2 a3
      -> join (f a1 b1) (f a2 b2) (f a3 b3).
End join_hom2.
Implicit Arguments join_hom2 [A B C saA saB saC].
Implicit Arguments join_hom2' [A B C saA saC].

(* [id] is join hom *)
Section join_hom_id.
  Variables (A: Type) (saA: sepalg A).

  Lemma join_hom_id : join_hom (fun x => x).
  Proof. unfold join_hom; auto. Qed.
End join_hom_id.

(* Product SA
   Join hom functions on products *)
Section join_hom_prod.
  Variables (A A' B B': Type)
            (saA: sepalg A) (saA': sepalg A') 
            (saB: sepalg B) (saB': sepalg B')
            (f: A -> A') (g: B -> B')
            (join_hom_f: join_hom f)
            (join_hom_g: join_hom g).

  Definition saAB := sa_prod saA saB.
  Definition saA'B' := sa_prod saA' saB'.

  Lemma join_hom2_pair
    : @join_hom2 _ _ _ saA saB saA'B' (fun a b => (f a, g b)).
  Proof.
    firstorder.
  Qed.

  Lemma join_hom2_pair'
    : @join_hom2 _ _ _ saA saB saAB (fun a b => (a, b)).
  Proof.
    firstorder.
  Qed.

  (* The function from [(a,b)] to [(a',b')] *)  
  Lemma join_hom_prod 
    : @join_hom _ saAB _ saA'B' (fun p => (f (fst p), g (snd p))).
  Proof. unfold join_hom; firstorder. Qed.

  (* The function from [a] to [(a', e)] *)
  Lemma join_hom_prodA 
    : forall e: B', identity e -> @join_hom _ saA _ saA'B' (fun a => (f a, e)).
  Proof.
    unfold join_hom in *; firstorder; simpl.
    rewrite identity_unit_equiv in H.
    auto.
  Qed.

  Lemma join_hom_prodA'
    : forall e: B, identity e -> @join_hom _ saA _ saAB (fun a => (a, e)).
  Proof.
    unfold join_hom in *; firstorder; simpl.
    rewrite identity_unit_equiv in H.
    auto.
  Qed.

  (* The function from [b] to [(e, b')] *)
  Lemma join_hom_prodB
    : forall e: A', identity e -> @join_hom _ saB _ saA'B' (fun b => (e, g b)).
  Proof.
    unfold join_hom in *; firstorder; simpl.
    rewrite identity_unit_equiv in H.
    auto.
  Qed.

  Lemma join_hom_prodB'
    : forall e: A, identity e -> @join_hom _ saB _ saAB (fun b => (e, b)).
  Proof.
    unfold join_hom in *; firstorder; simpl.
    rewrite identity_unit_equiv in H.
    auto.
  Qed.

  (* Projections are join hom. *)
  Lemma join_hom_proj1
    : @join_hom _ saAB _ saA (fun p => fst p).
  Proof. unfold join_hom; firstorder. Qed.

  Lemma join_hom_proj2
    : @join_hom _ saAB _ saB (fun p => snd p).
  Proof. unfold join_hom; firstorder. Qed.
End join_hom_prod.
Implicit Arguments join_hom2_pair [A A' B B' saA saA' saB saB'].
Implicit Arguments join_hom2_pair' [A B saA saB].
Implicit Arguments join_hom_prodA [A B' saA saB'].
Implicit Arguments join_hom_prodA' [A B saA saB].
Implicit Arguments join_hom_prodB [A' B saA' saB].
Implicit Arguments join_hom_prodB' [A B saA saB].
Implicit Arguments join_hom_proj1 [A B saA saB].
Implicit Arguments join_hom_proj2 [A B saA saB].


(* Disjoint Sum SA *)
Section join_hom_disjoint_sum.
  Variables (A A' B B': Type)
            (saA: sepalg A) (saA': sepalg A') 
            (saB: sepalg B) (saB': sepalg B')
            (f: A -> A') (g: B -> B')
            (join_hom_f: join_hom f)
            (join_hom_g: join_hom g).

  Definition saAorB := sa_sum saA saB.
  Definition saA'orB' := sa_sum saA' saB'.

  Lemma join_hom_sum
    : @join_hom _ saAorB _ saA'orB' (fun s => 
        match s with
          | inl x => inl _ (f x)
          | inr y => inr _ (g y)
        end).
  Proof. 
    unfold join_hom.
    destruct x; destruct y; destruct z; firstorder.
  Qed.

  Lemma join_hom_sum_l
    : @join_hom _ saAorB _ (sa_sum saA' saB) (fun s => 
        match s with
          | inl x => inl _ (f x)
          | inr y => inr _ y
        end).
  Proof. 
    unfold join_hom.
    destruct x; destruct y; destruct z; firstorder.
  Qed.

  Lemma join_hom_sum_r
    : @join_hom _ saAorB _ (sa_sum saA saB') (fun s => 
        match s with
          | inl x => inl _ x
          | inr y => inr _ (g y)
        end).
  Proof. 
    unfold join_hom.
    destruct x; destruct y; destruct z; firstorder.
  Qed.

  Lemma join_hom_inj_l
    : @join_hom _ saA _ (sa_sum saA' sa_void) (fun a => inl _ (f a)).
  Proof. firstorder. Qed.

  Lemma join_hom_inj_r
    : @join_hom _ saB _ (sa_sum sa_void saB') (fun b => inr _ (g b)).
  Proof. firstorder. Qed.

  (* Bijection between [A+unit] and [option A], for convenience later on *)
  Definition sa_sum_option (s: A+unit): option A :=
    match s with 
      | inl s' => Some s'
      | inr _ => None
    end.

  Definition option_sa_sum (s: option A): A+unit := 
    match s with
      | Some s' => inl _ s'
      | None => inr _ tt
    end.

  Lemma sa_sum__option: forall s, sa_sum_option (option_sa_sum s) = s.
  Proof. destruct s; firstorder. Qed.

  Lemma option__sa_sum: forall s, option_sa_sum (sa_sum_option s) = s.
  Proof. destruct s; firstorder; destruct u; firstorder. Qed.

  Definition bij_sa_sum_option : bijection (A+unit) (option A) :=
    Bijection _ _ sa_sum_option option_sa_sum sa_sum__option option__sa_sum.    
End join_hom_disjoint_sum.

(* List SA *)
Section join_hom_list.
  Variables (A: Type)
            (saA: sepalg A).

  Lemma join_hom_list_nil
    : @join_hom _ saA _ (list_sa saA) (fun a => a :: nil).
  Proof. 
    unfold join_hom; 
      solve [constructor; auto || constructor].
  Qed.

  Lemma join_hom2_list_cons
    : @join_hom2 _ _ _ saA (list_sa saA) (list_sa saA) (fun a l => a :: l).
  Proof.
    unfold join_hom2; constructor; auto.
  Qed.
End join_hom_list.

(* FPMs
   This section proves a join hom lemma specialized to finite partial maps
   producing [option]s.
*)
Section join_hom_fun.
  Variables (Key A: Type)
            (Key_dec_eq: forall k1 k2: Key, {k1=k2}+{~k1=k2})
            (saA: sepalg A).
  
  Definition saKey := sa_equiv Key.
  Definition saKeyA := sa_prod saKey saA.
  Definition saKeyAList := list_sa saKeyA.
  Definition saSum := sa_sum saA sa_unit.
  Definition saRange := sa_bijection _ _ (bij_sa_sum_option A) saSum.

  Fixpoint lookup k (rho: list (Key*A)) :=
    match rho with
      | nil => None
      | (k', a) :: rho' => 
          if Key_dec_eq k k' then Some a else lookup k rho'
    end.

  Lemma join_hom_fun 
    : @join_hom _ saKeyAList _ (sa_fun Key saRange) (fun env k =>
        lookup k env).
  Proof.
    unfold join_hom; intros x y z H. 
    induction H.

    (* env is nil *)
    simpl; unfold m_join; firstorder. 

    (* env is cons *)
    simpl; unfold m_join, join; intro x0.
    destruct x as [k1 a1]; destruct y as [k2 a2]; destruct z as [k3 a3].
    assert (Hk1_eq_k2: k1 = k2) by firstorder.
    assert (Hk2_eq_k3: k2 = k3) by firstorder.
    rewrite <- Hk2_eq_k3; rewrite <- Hk1_eq_k2. 
    destruct (Key_dec_eq x0 k1).
      inversion H; simpl in *; auto.
      apply IHlist_join.
  Qed.
End join_hom_fun.
Implicit Arguments lookup [Key A].


(* Bijection SA
   Bijections are join hom
*)
Section join_hom_bijection.
  Variables (A B: Type)
            (saA: sepalg A).

  Definition saA_bij bij := @sa_bijection A B bij saA.

  Lemma join_hom_bij 
    : forall (bij: bijection A B), match bij with Bijection f _ _ _ => 
          @join_hom _ saA _ (saA_bij bij) f 
      end. 
  Proof.
    unfold join_hom. destruct bij. intros.
    unfold join. simpl. repeat rewrite bij_gf. 
    auto.
  Qed.
End join_hom_bijection.


(* Some simple properties preserved by join homs *)
Section join_hom_properties.
  Variables (A B C: Type)
            (saA: sepalg A)
            (saB: sepalg B)
            (saC: sepalg C)
            (a1 a2: A)
            (b1 b2: B)
            (f: A -> B)
            (g: A -> B -> C).

  Lemma join_hom_join_sub : join_sub a1 a2 -> join_hom f -> join_sub (f a1) (f a2).
  Proof.
    intros H1 H2.
    unfold join_sub in *.
    destruct H1 as [b H1].
    exists (f b). 
    unfold join_hom in H2. apply H2. auto.
  Qed.

  Lemma join_hom_identity : identity a1 -> join_hom f -> identity (f a1).
  Proof.
    intros H1 H2.
    unfold join_hom in *.
    rewrite identity_unit_equiv in H1.
    rewrite identity_unit_equiv.
    unfold unit_for in *. auto.
  Qed.

  Lemma join_hom2_identity : identity a1 -> identity b1 -> join_hom2 g -> identity (g a1 b1).
  Proof.
    intros H1 H2 H3.
    unfold join_hom2 in *.
    rewrite identity_unit_equiv in H1.
    rewrite identity_unit_equiv in H2.
    rewrite identity_unit_equiv.
    unfold unit_for in *. auto.
  Qed.

  Lemma join_hom_comparable : comparable a1 a2 -> join_hom f -> comparable (f a1) (f a2).
  Proof.
    intros H1 H2.
    unfold join_hom in *.
    unfold comparable in H1.
    destruct (join_ex_identities a1) as [ea1 [Hea1 H3]].
    destruct (join_ex_identities a2) as [ea2 [Hea2 H4]]. simpl in *. subst.
 
    apply common_unit_comparable.
    exists (f ea2).
    split; apply H2;
      (eapply identity_unit in Hea1; eauto
       || eapply identity_unit in Heb; eauto).
  Qed.

  Lemma join_hom2_comparable 
    : comparable a1 a2 
      -> comparable b1 b2 
      -> join_hom2 g
      -> comparable (g a1 b1) (g a2 b2).
  Proof.
    intros H1 H2 H3.
    unfold join_hom2 in *.
    unfold comparable in *.
    destruct (join_ex_identities a1) as [ea1 [Hea1 H4]].
    destruct (join_ex_identities a2) as [ea2 [Hea2 H5]]. simpl in *. subst.
    destruct (join_ex_identities b1) as [eb1 [Heb1 H6]].
    destruct (join_ex_identities b2) as [eb2 [Heb2 H7]]. simpl in *. subst.
 
    apply common_unit_comparable.
    exists (g ea2 eb2).
    split; apply H3;
      (eapply identity_unit in Hea2; eauto
       || eapply identity_unit in Heb2; eauto).
  Qed.

  (* Should probably move this into "cross_split.v" *)
  Definition cross_split_raw {A} `{sepalg A} a b c d z := 
    join a b z
    -> join c d z 
    ->{x : A * A * A * A & 
       let (p, bd) := x in
       let (p1, bc) := p in
       let (ac, ad) := p1 in
       join ac ad a /\ join bc bd b /\ join ac bc c /\ join ad bd d}.

  Lemma join_hom_cross_split_raw : forall a b c d z: A,
    join a b z
    -> join c d z
    -> cross_split_raw a b c d z
    -> join_hom f
    -> @cross_split_raw _ saB (f a) (f b) (f c) (f d) (f z).
  Proof.
    intros a b c d z H1 H2 H3 H4.
    unfold cross_split_raw in *.
    destruct (H3 H1 H2) as [(((ac, ad), bc), bd) [H5 [H6 [H7 H8]]]].
    intros H9 H10.
    exists ((f ac), (f ad), (f bc), (f bd)).
    clear - H4 H5 H6 H7 H8.
    unfold join_hom in *. 
    auto.
  Qed.
End join_hom_properties.
Implicit Arguments join_hom_identity [A B saA saB].

(* Examples: *)
(* Finite Partial Maps from [nat]s to [option]s *)
Section fpm_ex.
  Variables (A: Type)
            (saA: sepalg A)
            (a b: A)
            (l: list (nat*A)).

  Definition mkEnv := fun p:nat*A => p :: l.
  Definition env1 := mkEnv (1%nat, a).
  Definition env2 := mkEnv (1%nat, b).
  Definition fpm (env: list (nat*A)) := fun k:nat => lookup eq_nat_dec k env.
  Definition saListA := list_sa (sa_prod (sa_equiv nat) saA).
  Definition saOption := sa_bijection _ _ (bij_sa_sum_option A) (saSum A saA).
  Definition saB := sa_fun nat saOption.

  Lemma fpm_comparable_ex
    : comparable a b 
      -> @comparable _ saB (fpm env1) (fpm env2).
  Proof.
    intros H.
    eapply join_hom_comparable with (f := fpm); try (apply join_hom_fun).
    eapply join_hom2_comparable with (g := (fun p l => p :: l)); try (apply join_hom2_list_cons).
    eapply join_hom2_comparable with (g := fun a b => (a,b)); try (apply join_hom2_pair'); 
      apply comparable_refl || auto.
    apply comparable_refl.
  Qed.
End fpm_ex.