(*
 * Copyright (c) 2009-2010, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Unset Automatic Introduction.
Generalizable All Variables.

Require Import base.
Require Import ageable.

Delimit Scope pred with pred.
Local Open Scope pred.

(* A "pre-predicate" is hereditary iff whenever it is
   true at world a, it is also true at all worlds
   accessable from a via R.
 *)
Definition hereditary {A} (R:A->A->Prop) (p:A->Prop) :=
  forall a a':A, R a a' -> p a -> p a'.

(* A predicate is a hereditary pre-predicate *)
Definition pred (A:Type) {AG:ageable A} :=
  { p:A -> Prop | hereditary age p }.

(* Here is some junk that makes the definition of "pred" opaque
   to most tactics but sill allows the "Program" extension to
   see it is a subset type.  The coercion is sugar that allows us to use
   predicates easily.
 *)
Definition app_pred `{ageable A} (p:pred A) : A -> Prop := proj1_sig p.
Definition pred_hereditary `{ageable} (p:pred A) := proj2_sig p.
Coercion app_pred : pred >-> Funclass.
Global Opaque pred.

Hint Resolve @pred_hereditary.

Lemma pred_nec_hereditary `{ageable A} (p:pred A) :
  forall a a':A, necR a a' -> p a -> p a'.
Proof.
  intros.
  induction H0; auto.
  apply pred_hereditary with x; auto.
Qed.

Program Definition mkPred `{ageable A} (p:A -> Prop) : pred A :=
  fun x => forall x', necR x x' -> p x'.
Next Obligation.
  repeat intro.
  apply H1.
  apply rt_trans with a'; auto.
  apply rt_step; auto.
Qed.

(* The semantic notion of entailment.
 *)
Definition derives `{ageable A} (P Q:pred A) := forall a:A, P a -> Q a.
Implicit Arguments derives.
Arguments Scope derives [pred pred].

(* A utility lemma that applies functional extensionality
   and proof irrelevance.
 *)
Lemma equiv_eq `{ageable A} : forall (P Q:pred A),
  derives P Q -> derives Q P -> P = Q.
Proof.
  intros.
  destruct P; destruct Q; unfold derives in *; simpl in *.
  revert h0; replace x0 with x; intros.
  replace h0 with h by apply proof_irr; auto.
  extensionality a.
  apply prop_ext; intuition.
Qed.

Lemma derives_cut `{ageable A} : forall Q P R,
  derives P Q ->
  derives Q R ->
  derives P R.
Proof.
  repeat intro.
  apply H1.
  apply H0.
  auto.
Qed.

(* "valid" relations are those that commute with aging.  These
   relations are the ones that can be turned into modalities.
 *)
Definition valid_rel `{ageable A} (R:relation A) : Prop :=
  commut A age R /\ commut A R age.

(* A modaility is a valid relation *)
Definition modality `{ageable A} := { R:relation A | valid_rel R }.

(* More black magic to make the definition of modaility mostly opaque. *)
Definition app_mode `{ageable A} (m:modality) : A -> A -> Prop := proj1_sig m.
Definition mode_valid `{ageable A} (m:modality) := proj2_sig m.
Global Opaque modality.
Coercion app_mode : modality >-> Funclass.

(* commutivity facts for the basic relations *)

Lemma valid_rel_commut_later1 `{ageable A} : forall R,
  valid_rel R ->
  commut A laterR R.
Proof.
  intros; hnf; intros.
  revert z H2.
  induction H1; intros.
  destruct H0.
  destruct (H0 _ _ H1 _ H2).
  exists x0; auto.
  apply t_step; auto.
  destruct (IHclos_trans1 _ H2).
  destruct (IHclos_trans2 _ H1).
  exists x1; auto.
  eapply t_trans; eauto.
Qed.

Lemma valid_rel_commut_later2 `{ageable A} : forall R,
  valid_rel R ->
  commut A R laterR.
Proof.
  intros; hnf; intros.
  revert x H1.
  induction H2; intros.
  destruct H0.
  destruct (H3 _ _ H2 _ H1).
  exists x1; auto.
  apply t_step; auto.
  destruct (IHclos_trans2 _ H1).
  destruct (IHclos_trans1 _ H3).
  exists x2; auto.
  eapply t_trans; eauto.
Qed.

Lemma valid_rel_commut_nec1 `{ageable A} : forall R,
  valid_rel R ->
  commut A necR R.
Proof.
  intros; hnf; intros.
  apply nec_refl_or_later in H1; destruct H1; subst.
  exists z; auto.
  apply rt_refl.
  destruct (valid_rel_commut_later1 R H0 x y H1 z H2).
  exists x0; auto.
  apply Rt_Rft; auto.
Qed.

Lemma valid_rel_commut_nec2 `{ageable A} : forall R,
  valid_rel R ->
  commut A R necR.
Proof.
  intros; hnf; intros.
  apply nec_refl_or_later in H2; destruct H2; subst.
  exists x; auto.
  apply rt_refl.
  destruct (valid_rel_commut_later2 R H0 x y H1 z H2).
  exists x0; auto.
  apply Rt_Rft; auto.
Qed.

Lemma valid_rel_age `{ageable A} : valid_rel age.
Proof.
  intros; split; hnf; intros; firstorder.
Qed.

Lemma valid_rel_later `{ageable A} : valid_rel laterR.
Proof.
  intros; split; hnf; intros.
  revert x H0.
  induction H1; intros.
  exists y; auto.
  apply t_step; auto.
  destruct (IHclos_trans2 _ H0).
  destruct (IHclos_trans1 _ H2).
  exists x2; auto.
  eapply t_trans; eauto.

  revert z H1.
  induction H0; intros.
  exists x; auto.
  apply t_step; auto.
  destruct (IHclos_trans1 _ H1).
  destruct (IHclos_trans2 _ H0).
  exists x1; auto.
  eapply t_trans; eauto.
Qed.

Lemma valid_rel_nec `{ageable A} : valid_rel necR.
Proof.
  intros; split; hnf; intros.
  revert x H0.
  induction H1; intros.
  exists y; auto.
  apply rt_step; auto.
  exists x0; auto.
  apply rt_refl.
  destruct (IHclos_refl_trans2 _ H0).
  destruct (IHclos_refl_trans1 _ H2).
  exists x2; auto.
  eapply rt_trans; eauto.

  revert z H1.
  induction H0; intros.
  exists x; auto.
  apply rt_step; auto.
  exists z; auto.
  apply rt_refl.

  destruct (IHclos_refl_trans1 _ H1).
  destruct (IHclos_refl_trans2 _ H0).
  exists x1; auto.
  eapply rt_trans; eauto.
Qed.

Lemma valid_rel_fashion `{ageable A} : valid_rel fashionR.
Proof.
  intros A AG.
  split; hnf; intros.

  unfold fashionR in *.
  hnf in H.
  case_eq (age1 z); intros.
  exists a.
  rewrite (af_level2 age_facts y x) in H0; auto.
  rewrite (af_level2 age_facts z a) in H0; auto.
  auto.
  rewrite (af_level1 age_facts) in H1; auto.
  rewrite H1 in H0.
  symmetry in H0.
  rewrite <- (af_level1 age_facts) in H0.
  rewrite H in H0; discriminate.

  hnf in H.
  destruct (af_unage age_facts _ _ _ H H0) as [x0 ?].
  exists x0; auto. 
  hnf.
  rewrite (af_level2 age_facts x0 x); auto.
  rewrite (af_level2 age_facts z y); auto.
Qed.

(* Definitions of the basic modalities.
 *)
Definition ageM `{ageable A} : modality
  := exist _ age valid_rel_age.
Definition laterM `{ageable A} : modality
  := exist _ laterR valid_rel_later.
Definition necM `{ageable A} : modality
  := exist _ necR valid_rel_nec.
Definition fashionM `{ageable A} : modality
  := exist _ fashionR valid_rel_fashion.

Hint Resolve rt_refl rt_trans t_trans.
Hint Unfold necR.
Obligation Tactic := unfold hereditary; intuition;
    first [eapply pred_hereditary; eauto; fail | eauto ].

(* Definitions of the basic propositional conectives.
 *)

Program Definition TT `{ageable A}: pred A := fun a:A => True.
Program Definition FF  `{ageable A}: pred A := fun a:A => False.
Program Definition implies `{ageable A} (P Q:pred A) : pred A :=
   fun a:A => forall a':A, necR a a' -> P a' -> Q a'.
Next Obligation.
  apply H1; auto.
  apply rt_trans with a'; auto.
  apply rt_step; auto.
Qed.
Program Definition union `{ageable A} (P Q:pred A) : pred A :=
   fun a:A => P a \/ Q a.
Next Obligation.
  left; eapply pred_hereditary; eauto.
  right; eapply pred_hereditary; eauto.
Qed.

Program Definition intersection `{ageable A} (P Q:pred A) : pred A :=
   fun a:A => P a /\ Q a.

(* Universal and existential quantification
 *)

Program Definition universal `{ageable A} {B: Type} (f: B -> pred A) : pred A
  := fun a => forall b, f b a.
Next Obligation.
  apply pred_hereditary with a; auto.
  apply H1.
Qed.

Program Definition existential `{ageable A} {B: Type} (f: B -> pred A) : pred A
  := fun a => exists b, f b a.
Next Obligation.
  destruct H1; exists x; eapply pred_hereditary; eauto.
Qed.


(* Definition of the "box" modal operator.  This operator turns
   modalities (relations) into a "necessarily" type operator.
 *)

Program Definition box `{ageable A} (M:modality) (P:pred A) : pred A :=
  fun a:A => forall a', M a a' -> P a'.
Next Obligation.
  destruct M as [M [H3 H4]]; simpl in *.
  destruct (H4 _ _ H2 _ H0).
  apply pred_hereditary with x; auto.
  apply H1; auto.
Qed.

(* Definition of the "diamond" modal operator.  This operator
   turns modalities into a "possibly" type operator.  _However_,
   note that this is NOT the boolean dual to "box", as usually
   found in accounts of modal logic. Instead, this is the
   "proof-theoretic" dual as found in Restall's "A Introduction
   to Substructural Logic" (2000).
 *)

Program Definition diamond `{ageable A} (M:modality) (P:pred A) : pred A :=
  fun a:A => exists a', M a' a /\ P a'.
Next Obligation.
  destruct M as [M [H3 H4]]; simpl in *.
  destruct H1 as [x [? ?]].
  destruct (H3 _ _ H0 _ H1).
  exists x0; split; auto.
  apply pred_hereditary with x; auto.
Qed.

Definition boxy `{ageable A} (m: modality) (p: pred A): Prop :=  box m p = p.

(* Lifting pure mathematical facts to predicates *)
Program Definition prop `{ageable A}  (P: Prop) : pred A := (fun _  => P).

(* A pile of notations for the operators we have defined *)
Module Notation.
  Delimit Scope entails_pred with entails_pred.

  Notation "P '|--' Q" := (derives P%pred Q%pred) (at level 80, no associativity) : entails_pred.
  Notation "'Ex_'  x ':' T ',' P " := (existential (fun x:T => P%pred)) (at level 65, x at level 99) : pred.
  Notation "'All_'  x ':' T  ',' P " := (universal (fun x:T => P%pred)) (at level 65, x at level 99) : pred.
  Infix "||" := union (at level 50, left associativity) : pred.
  Infix "&&" := intersection (at level 40, left associativity) : pred.
  Notation "P '-->' Q" := (implies P%pred Q%pred) (at level 55, right associativity) : pred.
  Notation "P '<-->' Q" := (intersection (implies P%pred Q%pred) (implies Q%pred P%pred)) (at level 57, no associativity) : pred.
  Notation "'[]' e" := (box necM e) (at level 30, right associativity): pred.
  Notation "'|>' e" := (box laterM e) (at level 30, right associativity): pred.
  Notation "'!!' e" := (prop e) (at level 25) : pred.
  Notation "P '>=>' Q" := (box fashionM (implies P%pred Q%pred)) (at level 55, right associativity) : pred.
  Notation "P '<=>' Q" := (box fashionM (intersection (implies P%pred Q%pred) (implies Q%pred P%pred))) (at level 57, no associativity) : pred.
End Notation.

Import Notation.
Local Open Scope entails_pred.

(* Rules for the propositional connectives *)
Lemma modus_ponens `{ageable A} : forall (X P Q:pred A),
  X |-- P ->
  X |-- (P --> Q) ->
  X |-- Q.
Proof.
  unfold derives, implies; simpl; intuition eauto.
Qed.

Lemma and_intro `{ageable A} : forall (X P Q:pred A),
  X |-- P ->
  X |-- Q ->
  X |-- P && Q.
Proof.
  unfold derives, implies, intersection; simpl; intuition.
Qed.

Lemma and1 `{ageable A} : forall (X P Q:pred A),
  X |-- P && Q --> P.
Proof.
  unfold derives, implies, intersection; simpl; intuition eauto.
Qed.

Lemma and2 `{ageable A} : forall (X P Q:pred A),
  X |-- P && Q --> Q.
Proof.
  unfold derives, implies, intersection; simpl; intuition eauto.
Qed.

Lemma and3 `{ageable A} : forall (X P Q R:pred A),
  X |-- (P --> Q) --> (P --> R) --> (P --> Q && R).
Proof.
  unfold derives, implies, intersection; simpl; intuition eauto.
Qed.

Lemma or1 `{ageable A} : forall (X P Q:pred A),
  X |-- P --> P || Q.
Proof.
  unfold derives, implies, union; simpl; intuition.
Qed.

Lemma or2 `{ageable A} : forall (X P Q:pred A),
  X |-- Q --> P || Q.
Proof.
  unfold derives, implies, union; simpl; intuition.
Qed.

Lemma or3 `{ageable A} : forall (X P Q R:pred A),
  X |-- (P --> R) --> (Q --> R) --> (P || Q --> R).
Proof.
  unfold derives, implies, union; simpl; intuition eauto.
Qed.

Lemma TTrule `{ageable A} : forall X P,
  X |-- P --> TT.
Proof.
  unfold derives, implies, TT; simpl; intuition.
Qed.

Lemma FFrule `{ageable A} : forall X P,
  X |-- FF --> P.
Proof.
  unfold derives, implies, FF; simpl; intuition.
Qed.

Lemma distribution `{ageable A} : forall (X P Q R:pred A),
  X |-- P && (Q || R) --> (P && Q) || (P && R).
Proof.
  unfold derives, implies, union, intersection; simpl; intuition.
Qed.

(* Characterize the relation between conjunction and implication *)
Lemma impl_and_adjoint `{ageable A} : forall (P Q R:pred A),
  (P && Q) |-- R <-> P |-- (Q --> R).
Proof.
  split; intros.
  hnf; intros; simpl; intros.
  apply H0.
  split; auto.
  apply pred_nec_hereditary with a; auto.
  hnf; intros.
  hnf in H0.
  unfold implies in H0; simpl in H0.
  destruct H1.
  apply H0 with a; auto.
Qed.

(* Some facts about modalities *)

Lemma box_e0 `{ageable A}: forall (M: modality) Q, 
            reflexive _ M -> box M Q  |-- Q.
Proof.
intros.
intro; intros.
apply H1; simpl.
apply H0.
Qed.
Implicit Arguments box_e0.

Lemma boxy_i `{ageable A}: 
  forall (Q: pred A) (M: modality),
    reflexive _ M -> 
    (forall w w', M w w' -> Q w -> Q w') ->
    boxy M Q.
Proof.
intros.
unfold boxy.
apply equiv_eq; hnf; intros.
eapply box_e0; eauto.
hnf; intros.
eapply H1; eauto.
Qed.

Lemma necM_refl `{ageable A}: reflexive _ necM.
Proof.
intros; intro; simpl.
unfold necR.
constructor 2.
Qed.

Hint Resolve @necM_refl.

(* relationship between box and diamond *)
Lemma box_diamond `{ageable A} : forall M (P Q:pred A),
  (diamond M P) |-- Q <-> P |-- (box M Q).
Proof.
  unfold derives; intuition.
  hnf; intros.
  apply H0.
  hnf; eauto.
  destruct H1 as [a' [? ?]].
  apply H0 with a'; auto.
Qed.

(* Box is a normal modal operator *)

Lemma ruleNec `{ageable A} : forall M (P:pred A),
  derives TT P ->
  derives TT (box M P).
Proof.
  intros.
  rewrite <- box_diamond.
  hnf; intros.
  apply H0; hnf; auto.
Qed.

Lemma axiomK `{ageable A}: forall M (P Q:pred A),
  (box M (P --> Q)) |-- (box M P --> box M Q).
Proof.
  intros; do 3 (hnf; intros).
  destruct M as [R HR]; simpl in *.
  destruct (valid_rel_commut_nec2 R HR _ _ H3 _ H1).
  apply H0 with x; auto.
Qed.

(* Box and diamond are positive modal operators *)

Lemma box_positive `{ageable A} : forall M (P Q:pred A),
  P |-- Q ->
  box M P |-- box M Q.
Proof.
  unfold derives, box; simpl; intuition.
Qed.

Lemma diamond_positive `{ageable A} : forall M (P Q:pred A),
  P |-- Q ->
  diamond M P |-- diamond M Q.
Proof.
  unfold derives, diamond; simpl; firstorder.
Qed.

Lemma box_refl_trans `{ageable A}: forall (m:modality) p,
  reflexive _ m ->
  transitive _ m ->
  box m (box m p) = box m p.
Proof.
  intros.
  apply equiv_eq.
  repeat intro.
  assert (box m p a').
  apply H2; auto.
  apply H4.
  apply H0.
  repeat intro.
  apply H2.
  eapply H1; eauto.
Qed.

(* Disribuitivity of box over various connectives *)

Lemma box_and `{ageable A}: forall R (P Q:pred A),
  box R (P && Q) = box R P && box R Q.
Proof.
  intros; apply equiv_eq; hnf; intuition;
    unfold intersection, box in *; simpl in *; firstorder.
Qed.

Lemma box_all `{ageable A} : forall B R (F:B -> pred A),
  box R (universal F) = All_ x:B, box R (F x).
Proof.
  intros; apply equiv_eq; hnf; intuition;
    unfold universal, box in *; simpl in *; firstorder.
Qed.

Lemma box_ex `{ageable A} : forall B R (F:B->pred A),
  Ex_ x:B, box R (F x) |-- box R (existential F).
Proof.
  unfold derives, existential, box; simpl; firstorder.
Qed.

Lemma box_or `{ageable A} : forall R (P Q:pred A),
   box R P || box R Q |-- box R (P || Q).
Proof.
  unfold derives, union, box; simpl; firstorder.
Qed.

(* Distributivity of diamond over various operators *)

Lemma diamond_or `{ageable A} : forall R (P Q:pred A),
  diamond R (P || Q) = diamond R P || diamond R Q.
Proof.
  intros; apply equiv_eq; hnf; intuition;
    unfold diamond, union in *; simpl in *; firstorder.
Qed.

Lemma diamond_ex `{ageable A} : forall B R (F:B -> pred A),
  diamond R (existential F) = Ex_ x:B, diamond R (F x).
Proof.
  intros; apply equiv_eq; hnf; intuition;
    unfold diamond, existential in *; simpl in *; firstorder.
Qed.

Lemma diamond_and `{ageable A} : forall R (P Q:pred A),
  diamond R (P && Q) |-- diamond R P && diamond R Q.
Proof.
  unfold derives, intersection, diamond; simpl; firstorder.
Qed.

Lemma diamond_all `{ageable A} : forall B R (F:B->pred A),
  diamond R (universal F) |-- All_ x:B, diamond R (F x).
Proof.
  unfold derives, universal, diamond; simpl; firstorder.
Qed.


(* Lemmas about aging and the later operator *)

Lemma nec_useless `{ageable A} :
  forall P, []P = P.
intros.
  apply equiv_eq; intros.
  hnf; intros; apply H0.
  simpl; apply necM_refl.
  hnf; intros.
  hnf; intros.
  apply pred_nec_hereditary with a; auto.
Qed.

Lemma later_age `{ageable A} : forall P,
  |>P = box ageM P.
Proof.
  intros; apply equiv_eq; do 2 (hnf; intros).
  simpl in H0.
  apply H0.
  apply t_step; auto.
  revert H0; induction H1; intros.
  apply H1; auto.
  apply pred_nec_hereditary with y.
  apply Rt_Rft; auto.
  apply IHclos_trans1; auto.
Qed.

Lemma now_later `{ageable A} : forall P,
  P |-- |>P.
Proof.
  repeat intro.
  apply pred_nec_hereditary with a; auto.
  apply Rt_Rft; auto.
Qed.

Lemma now_later2 `{ageable A} : forall G P,
  G |-- P ->
  G |-- |>P.
Proof.
  intros; apply derives_cut with P; auto.
  apply now_later.
Qed.

(* The "induction" rule for later *)

Lemma goedel_loeb `{ageable A} : forall (P Q:pred A),
  Q && |>P |-- P ->
  Q |-- P.
Proof.
  intros; hnf; intro a.
  induction a using age_induction.
  intros; simpl in H0.
  eapply H0; auto.
  split; auto.
  rewrite later_age.
  simpl; intros.
  apply H1; auto.
  apply pred_hereditary with x; auto.
Qed.

(* Later distributes over almost everything! *)

Lemma later_commute_dia `{ageable A} : forall M (P:pred A),
  diamond M (|> P) |-- |> (diamond M P).
Proof.
  intros.
  repeat rewrite later_age.
  do 3 (hnf; intros).
  simpl in H0.
  firstorder.
  destruct M as [R HR].
  simpl in *.
  destruct HR.
  destruct (H3 _ _ H1 _ H0).
  exists x0; split; auto.
Qed.

Lemma later_commute `{ageable A} : forall M (P:pred A),
  box M (|>P) = |>(box M P).
Proof.
  intros.
  apply equiv_eq; do 3 (hnf; intros).
  destruct M as [R HR].
  destruct (valid_rel_commut_later2 R HR _ _ H2 _ H1).
  apply H0 with x; simpl; auto.
  destruct M as [R HR].
  destruct (valid_rel_commut_later1 R HR _ _ H2 _ H1).
  apply H0 with x; auto.
Qed.

Lemma later_and `{ageable A} : forall P Q,
  |>(P && Q) = |>P && |> Q.
Proof.
  intros; apply box_and.
Qed.

Lemma later_or `{ageable A} : forall (P Q:pred A),
  |>(P || Q) = |>P || |>Q.
Proof.
  intros.
  repeat rewrite later_age.
  apply equiv_eq.
  2: apply box_or.
  hnf; intros.
  simpl in H0.
  case_eq (age1 a); intros.
  destruct (H0 a0); auto.
  left; simpl; intros.
  replace a' with a0; auto.
  congruence.
  right; simpl; intros.
  replace a' with a0; auto.
  congruence.
  left.
  hnf; simpl; intros.
  hnf in H2.
  rewrite H1 in H2; discriminate.
Qed.

Lemma later_ex `{ageable A} : forall B (F:B->pred A),
  B ->
  |>(existential F) = Ex_ x:B, |>(F x).
Proof.
  intros.
  apply equiv_eq.
  2: apply box_ex.
  hnf; intros.
  rewrite later_age in H0.
  case_eq (age1 a); intros.
  destruct (H0 a0); auto.
  exists x.
  rewrite later_age; simpl; intros.
  replace a' with a0; auto.
  congruence.
  exists X.
  rewrite later_age.
  hnf; simpl; intros.
  unfold age in H2.
  rewrite H1 in H2; discriminate.
Qed.

Lemma later_impl `{ageable A} : forall P Q,
  |>(P --> Q) = |>P --> |>Q.
Proof.
  intros; repeat rewrite later_age.
  apply equiv_eq.
  apply axiomK.
  hnf; intros.
  simpl; intros.
  simpl in H0.
  destruct valid_rel_nec.
  destruct (H5 _ _ H2 _ H1).
  apply H0 with x; auto.
  intros.
  replace a'1 with a'0; auto.
  congruence.
Qed.

Lemma TT_boxy `{ageable A} : forall M,
  boxy M TT.
Proof.
  intros; hnf.
  apply equiv_eq; repeat intro; simpl; auto.
Qed.

Lemma positive_boxy `{ageable A} : forall P Q M,
  boxy M P ->
  P |-- Q ->
  P |-- box M Q.
Proof.
  intros.
  rewrite <- H0.
  apply box_positive.
  auto.
Qed.

Lemma forallI `{ageable A} : forall A G X,
  (forall x:A, G |-- X x) ->
  G |-- universal X.
Proof.
  repeat intro.
  eapply H0; auto.
Qed.

Lemma TT_and `{ageable A} : forall P,
  TT && P = P.
Proof.
  intros; apply equiv_eq; repeat intro.
  destruct H0; auto.
  split; simpl; auto.
Qed.

Lemma and_comm `{ageable A} : forall P Q,
  P && Q = Q && P.
Proof.
  intros; apply equiv_eq; unfold intersection; repeat intro; simpl in *; intuition.
Qed.

Lemma and_assoc `{ageable A} : forall P Q R,
  P && (Q && R) = (P && Q) && R.
Proof.
  intros; apply equiv_eq; auto; unfold derives, intersection; simpl; intuition.
Qed.

Lemma ex_and : forall `{ageable A} B (P:B->pred A) Q,
  (existential P) && Q = Ex_ x:B, P x && Q.
Proof.
  intros. apply equiv_eq.
  repeat intro. destruct H0. destruct H0.
  exists x. split; auto.
  repeat intro.
  destruct H0. destruct H0.
  split; auto. exists x; auto.
Qed.

Lemma FF_and : forall `{ageable A} (P:pred A),
  FF && P = FF.
Proof.
  intros. apply equiv_eq; repeat intro.
  destruct H0; auto.
  elim H0.
Qed.
