(*
 * Copyright 2011, Christian J. Bell
 *
 * Parts of this file are taken from the "imp" example
 * in the Mechanized Semantic Library and are copyrighted
 * by A. Appel, R. Dockins, and A. Hobor.
 * TODO: include their copyright and disclaimers
 *)

Require Import msl.msl_standard.
Require Import channel_sa.
Require Import SfLib.
Require Import LibTactics.

Set Implicit Arguments.

Section Definitions.
  Definition ident := positive.
  Definition addr  := nat.
  Definition cname := positive.
  Inductive value : Set :=
    | nat_value : nat -> value
    | addr_value : addr -> value
    | cname_value : cname -> value.

  Definition cname_eq : forall (c1 c2 : cname), {c1=c2}+{c1<>c2} := positive_eq_dec.

End Definitions.
Let var:= ident.

  Lemma pair_is_identity: forall A B sa_a sa_b (a:A) (b:B),
    @identity _ sa_a a -> @identity _ sa_b b -> @identity _ (sa_prod sa_a sa_b) (a,b).
  Proof.
    intros.
    unfold identity. intros ab0 ab1 ?.
    destruct ab0 as [a0 b0].
    destruct ab1 as [a1 b1].
    destruct H1; simpl in *.
    rewrite H; eauto.
    rewrite H0; eauto.
  Qed.

Module ChannelsModel.


  Definition chanpred := (channels cname) -> Prop.

  Definition endpoint (c:cname) (p:Share.t) (count:nat) : chanpred :=
    fun ch =>
        (forall c', c<>c' -> ch c' = None) /\
        match ch c with
        | Some (exist pn H) => proj1_sig pn=(p,count)
        | None => identity p
        end.

  Definition has_endpoint (c:cname) (p:Share.t) (count:nat) : chanpred :=
    fun ch =>
        match ch c with
        | Some (exist pn H) => proj1_sig pn=(p,count)
        | None => False
        end.

  Definition has_access (c:cname) : chanpred :=
    fun ch => ch c <> None.

  Definition has_total_endpoint (c:cname) : chanpred :=
    fun ch =>
        match ch c with
        | Some (exist ce H) => ce_share ce = Share.top
        | None => False
        end.

  

  Program Definition endpoint_incr (ce : ch_endpoint) : ch_endpoint:= 
    let (x, e) := ce in
    let (p,n) := x in
    if dec_share_identity p then ce else (p,n+1).

  Definition channels_unit : channels cname := @channels_unit cname.
  Let sa_channels := sa_channels cname.
  Existing Instance sa_channels.

  Program Definition just_endpoint (c:cname) (ce: ch_endpoint) : channels cname :=
    fun c' => if positive_eq_dec c c'
      then if dec_share_identity (ce_share ce) then None else Some ce
      else None.
  Next Obligation.
    destruct ce; simpl in *; firstorder.
  Qed.
  Next Obligation.
    apply H0.
    rewrite identity_unit_equiv.
    rewrite identity_unit_equiv in H1.
    destruct ce as [x]; destruct x.
    firstorder.
  Qed.

  Lemma endpoint_share_identity: forall ce,
    @identity _ Share.sa (ce_share ce) <-> @identity _ sa_ce ce.
  Proof.
    destruct ce as [[p n] Hce].
    simpl in *.
    split; intros.
    assert (n=0); auto.
    subst n.
    rewrite identity_unit_equiv in H.
    rewrite identity_unit_equiv.
    simpl in *.
    unfold join_ce, join_ce'.
    auto.

    rewrite identity_unit_equiv in H.
    rewrite identity_unit_equiv.
    apply H.
  Qed.


  Program Definition cons_channel (c:cname) (ce: ch_endpoint) (ch: channels cname) (H: ch c = None) : channels cname :=
    fun c' => if positive_eq_dec c c'
      then if dec_share_identity (ce_share ce) then None else Some ce
      else ch c'.
  Next Obligation.
    destruct ce; simpl in *; firstorder.
  Qed.
  Next Obligation.
    subst.
    apply H1.
    apply endpoint_share_identity in H2.
    simpl in *; destruct ce; auto. 
  Qed.


  Definition channels_incr' {A:Type} (x:option (pjoinable sa_ce)) (f: forall (ce:ch_endpoint), ~identity ce -> A) (d:A) : A :=
    match x with
    | Some (exist ce H) => f ce H
    | None => d
    end.

  Program Definition channels_incr (c:cname) (ch: channels cname) : channels cname :=
    channels_incr' (ch c) (fun ce H c'=> if cname_eq c c' then Some (endpoint_incr ce) else ch c') ch.
  Next Obligation.
    subst c'. destruct ce as [[p n] Hce].
    simpl in *.
    destruct (dec_share_identity p).
    auto. simpl in *; contradiction.
  Qed.
  Next Obligation.
    subst c'.
    apply endpoint_share_identity in H1.
    simpl in H1.
    destruct ce as [[p n] Hce].
    simpl in *.
    destruct (dec_share_identity p).
    apply H.
    apply endpoint_share_identity.
    auto.
    intuition.
  Qed.


  Definition pendpoint (pce: option (pjoinable sa_ce)) p n :=
    match pce with
    | Some (exist (exist (p',m) B) nonident) => p'=p /\ m=n
    | None => False
    end.
  Lemma pendpoint_Some: forall pce p n,
    pendpoint pce p n <->
    exists P, pce = Some P /\ fst (proj1_sig (proj1_sig P)) = p /\ snd (proj1_sig (proj1_sig P))=n /\ nonidentity p.
  Proof.
    split; intros.
    unfold pendpoint in H.
    destruct pce; intuition.
    exists p0; destruct p0 as [[[?]]]; intuition; subst; intro; apply n1; apply endpoint_share_identity; auto.
    unfold pendpoint.
    destruct H as [P[?[?[??]]]].
    rewrite H. destruct P as [[[?]]]. simpl in *.  subst. auto.
  Qed. 
  


Lemma pendpoint_incr_decr: forall ch c n,
pendpoint (channels_incr c ch c) Share.top n <->
(pendpoint (ch c) Share.top (n-1) /\ n>0).
Proof.
  unfold channels_incr, pendpoint.
  intros.
  case_eq (ch c).
  intros P H0.
  destruct P as [[[p n']]].
  simpl.
  destruct (cname_eq c c); [| intuition].
  destruct (dec_share_identity p); simpl.
  elimtype False. apply n0. apply endpoint_share_identity. auto.
  intuition.
  intros H. simpl. rewrite H.
  intuition.
Qed.



(*
  Program Definition channels_incr (c:cname) (ch: channels cname) : channels cname :=
    match ch c with 
    | Some (exist ce H) => ch c (*fun c'=> if cname_eq c c' then Some (endpoint_incr ce) else ch c'*)
    | None => ch
    end.
  Next Obligation.
    subst c'. destruct ce as [[p n] Hce].
    simpl in *.
    destruct (dec_share_identity p).
    auto. simpl in *; contradiction.
  Qed.
  Next Obligation.
    subst c'.
    apply endpoint_share_identity in H1.
    simpl in H1.
    destruct ce as [[p n] Hce].
    simpl in *.
    destruct (dec_share_identity p).
    apply H.
    apply endpoint_share_identity.
    auto.
    intuition.
  Qed.
*)

End ChannelsModel.

Module LocalsModel.
  (* Get an implemenation of share maps for the local environments*)
  Module LocalsMapInput <: ShareMapInput.
    Definition A := ident.
    Definition A_dec_eq := positive_eq_dec.
    Definition B := value.
  End LocalsMapInput.

  Module LocalsMap := ShareMap(LocalsMapInput).

  Definition locals := LocalsMap.map.
  Definition locals_sa := LocalsMap.map_sa.
  Definition locals_eqv_sa := sa_equiv locals.

  Existing Instance locals_eqv_sa. 

  Definition locals_In (x:ident) (l:locals) :=
    LocalsMap.map_share x l <> Share.bot /\ LocalsMap.map_val x l <> None.

  Definition locpred := locals -> Prop.

  Definition hasval (i:ident) (v:value) : locpred :=
    fun rho =>
      LocalsMap.map_val i rho = Some v /\
      LocalsMap.map_share i rho = Share.top /\
      forall i', i <> i' -> LocalsMap.map_val i' rho = None.

  Definition expr:=
    { f: (ident -> option value) -> option value |
      forall l l' v, (forall x v, l x = Some v -> l' x = Some v) -> f l = Some v -> f l' = Some v }.

(*  Definition expr:=
    { f: locals -> option value |
      forall l l' v, @join_sub locals locals_sa l l' -> f l = Some v -> f l' = Some v }.
*)

  Program Definition nat_to_expr (n:nat) : expr :=
    exist _ (fun _=>Some (nat_value n)) _.

  Coercion nat_to_expr : nat >-> expr.

  Lemma locals_eqv_identity: forall (l:locals), @identity _ locals_eqv_sa l.
  Proof. firstorder. Qed.

End LocalsModel.

Module HeapModel.
  (* Get an implemenation of share maps for memories *)
  Module MemMapInput <: ShareMapInput.
    Definition A := addr.
    Definition A_dec_eq := eq_nat_dec.
    Definition B := value.
  End MemMapInput.

  Module MemMap := ShareMap(MemMapInput).

  Definition heap := MemMap.map.
  Definition heap_sa := MemMap.map_sa.

  Definition heappred := heap -> Prop.
  (* Address a has value v with share sh. *)
  Definition mapsto (a:addr) (sh:Share.t) (v:value) : heappred :=
    fun m =>
      MemMap.map_val a m = Some v /\
      MemMap.map_share a m = sh /\
      forall a', a <> a' -> MemMap.map_val a' m  = None.
End HeapModel.

Export LocalsModel.
Export HeapModel.
Export ChannelsModel.

Section WorldModel.
  Import predicates_sl.Notation.
  Open Scope entails_pred.
  Open Scope pred.

  Definition world' := ((channels cname * channels cname) * (locals * heap))%type.

  Record world := MakeWorld
    { w_prod: channels cname;
      w_cons: channels cname;
      w_env: locals;
      w_heap: heap }.

  Definition wupd_heap w h' := match w with
    | {|w_prod:=p; w_cons:=c; w_env:=e; w_heap:=h|} => {|w_prod:=p;w_cons:=c;w_env:=e;w_heap:=h'|}
  end.
  Definition wupd_env w e' := match w with
    | {|w_prod:=p; w_cons:=c; w_env:=e; w_heap:=h|} => {|w_prod:=p;w_cons:=c;w_env:=e';w_heap:=h|}
  end.
  Definition wupd_prod w p' := match w with
    | {|w_prod:=p; w_cons:=c; w_env:=e; w_heap:=h|} => {|w_prod:=p';w_cons:=c;w_env:=e;w_heap:=h|}
  end.
  Definition wupd_cons w c' := match w with
    | {|w_prod:=p; w_cons:=c; w_env:=e; w_heap:=h|} => {|w_prod:=p;w_cons:=c';w_env:=e;w_heap:=h|}
  end.
  Lemma w_prod_wupd_prod: forall w p,
    w_prod (wupd_prod w p) = p.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_cons_wupd_prod: forall w p,
    w_cons (wupd_prod w p) = w_cons w.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_env_wupd_prod: forall w p,
    w_env (wupd_prod w p) = w_env w.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_heap_wupd_prod: forall w p,
    w_heap (wupd_prod w p) = w_heap w.
  Proof. destruct w; reflexivity. Qed.

  Lemma w_prod_wupd_cons: forall w c,
    w_prod (wupd_cons w c) = w_prod w.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_cons_wupd_cons: forall w c,
    w_cons (wupd_cons w c) = c.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_env_wupd_cons: forall w c,
    w_env (wupd_cons w c) = w_env w.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_heap_wupd_cons: forall w c,
    w_heap (wupd_cons w c) = w_heap w.
  Proof. destruct w; reflexivity. Qed.

  Lemma w_prod_wupd_env: forall w e,
    w_prod (wupd_env w e) = w_prod w.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_cons_wupd_env: forall w e,
    w_cons (wupd_env w e) = w_cons w.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_env_wupd_env: forall w e,
    w_env (wupd_env w e) = e.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_heap_wupd_env: forall w e,
    w_heap (wupd_env w e) = w_heap w.
  Proof. destruct w; reflexivity. Qed.

  Lemma w_prod_wupd_heap: forall w h,
    w_prod (wupd_heap w h) = w_prod w.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_cons_wupd_heap: forall w h,
    w_cons (wupd_heap w h) = w_cons w.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_env_wupd_heap: forall w h,
    w_env (wupd_heap w h) = w_env w.
  Proof. destruct w; reflexivity. Qed.
  Lemma w_heap_wupd_heap: forall w h,
    w_heap (wupd_heap w h) = h.
  Proof. destruct w; reflexivity. Qed.


  Program Definition world_tuple_bijection: bijection world' world :=
    Bijection world' world (fun w=> match w with ((p,c),(l,h)) => {|w_prod:=p; w_cons:= c; w_env:= l; w_heap:= h|} end)
      (fun w=> ((w_prod w,w_cons w),(w_env w,w_heap w))) _ _.
  Next Obligation. destruct x; auto. Qed.
  Next Obligation. destruct x as [[??][??]]; auto. Qed.

  Definition world'_sa : sepalg world' :=
    sa_prod (sa_prod sa_channels sa_channels) (sa_prod locals_eqv_sa heap_sa).
  Definition world_sa : sepalg world :=
    sa_bijection world' world world_tuple_bijection world'_sa.
  Existing Instance world_sa.

  Lemma world_ageable_facts:
    ageable_facts world (fun _=>0) (fun _=>None).
  Proof. constructor; try discriminate; intuition. Qed.
  Definition world_ageable : ageable world :=
    mkAgeable world (fun _=>0) (fun _=>None) world_ageable_facts.
  Existing Instance world_ageable.


  Definition gworld'_sa : sepalg world' :=
    sa_prod (sa_prod sa_channels sa_channels) (sa_prod locals_sa heap_sa).
  Definition gworld_sa : sepalg world :=
    sa_bijection world' world world_tuple_bijection gworld'_sa.
  Definition gjoin := @join world gworld_sa.
  Definition gjoin_sub := @join_sub world gworld_sa.
  Definition gjoins := @joins world gworld_sa.

  Lemma world_asa : @ASA world world_sa world_ageable.
  Proof. constructor; intros; discriminate. Qed.
  Existing Instance world_asa.

  Lemma gworld_asa : @ASA world gworld_sa world_ageable.
  Proof. constructor; intros; discriminate. Qed.

  Definition locals_unit : locals := LocalsMap.empty_map.
  Definition heap_unit : heap := MemMap.empty_map.
  Definition prod_unit := @channels_unit.
  Definition cons_unit := @channels_unit.
  Definition world_unit l := MakeWorld prod_unit cons_unit l heap_unit.
  Definition gworld_unit := MakeWorld prod_unit cons_unit locals_unit heap_unit.

  Lemma world_unit_identity: forall l, identity (world_unit l).
  Proof.
    intros.
    assert (@identity _ world'_sa (bij_g _ _ world_tuple_bijection (world_unit l))).
    simpl.
    do 2 (apply pair_is_identity).
    apply channels_unit_identity.
    apply channels_unit_identity.
    apply locals_eqv_identity.
    apply MemMap.empty_map_identity.
    repeat intro.
    assert ((bij_g _ _ world_tuple_bijection a) = (bij_g _ _ world_tuple_bijection b)).
    eapply H.
    auto.
    destruct a; destruct b.
    inversion H1; subst.
    auto.
  Qed.

  Lemma gworld_unit_identity: @identity _ gworld_sa gworld_unit.
  Proof.
    intros.
    assert (@identity _ gworld'_sa (bij_g _ _ world_tuple_bijection gworld_unit)).
    simpl.
    do 2 (apply pair_is_identity).
    apply channels_unit_identity.
    apply channels_unit_identity.
    apply LocalsMap.empty_map_identity.
    apply MemMap.empty_map_identity.
    repeat intro.
    assert ((bij_g _ _ world_tuple_bijection a) = (bij_g _ _ world_tuple_bijection b)).
    eapply H.
    auto.
    destruct a; destruct b.
    inversion H1; subst.
    auto.
  Qed.

  Lemma value_is_addr_dec: forall (v:value), {a|v=addr_value a}+{forall a, v<>addr_value a}.
  Proof.
    intros.
    destruct v; eauto.
    right; intros; intro; discriminate.
    right; intros; intro; discriminate.
  Qed.

  Definition hexpr:=
    { f: (addr -> option value) -> (ident -> option value) -> option value |
      forall h h' l l' v,
        (forall a v, h a = Some v -> h' a = Some v) ->
        (forall x v, l x = Some v -> l' x = Some v) ->
        f h l = Some v ->
        f h' l' = Some v }.

  Definition w_env_share w x:= LocalsMap.map_share x (w_env w).


  Definition w_prod_incr c w : world :=
    wupd_prod w (channels_incr c (w_prod w)).
  Definition w_cons_incr c w : world :=
    wupd_cons w (channels_incr c (w_cons w)).
  Program Definition w_add_new_channel (c:cname) (w: world) (H1: w_prod w c = None) (H2: w_cons w c = None) : world :=
    wupd_prod (wupd_cons w (cons_channel c (ce_make_top 0) (w_cons w) H2)) (cons_channel c (ce_make_top 0) (w_prod w) H1).

  Program Definition new_channel (c:cname): world :=
        {|w_prod:=(just_endpoint c (ce_make_top 0)); w_cons:= (just_endpoint c (ce_make_top 0)); w_env:= locals_unit; w_heap:= heap_unit|}.

  Program Definition new_local (x:var) (v:value): world :=
        {|w_prod:=channels_unit; w_cons:= channels_unit; w_env:= LocalsMap.singleton x v; w_heap:= heap_unit|}.

  Lemma w_prod_incr_eq_dec: forall c w, ({w'| w = w_prod_incr c w'} + {forall w', w <> w_prod_incr c w'}).
  Admitted.

  Lemma channels_incr_eq: forall c ch1 ch2, channels_incr c ch1 = channels_incr c ch2 -> ch1 = ch2.
  Proof.
    intros.
    extensionality c'.
    assert (channels_incr c ch1 c' = channels_incr c ch2 c').
      rewrite H; auto.
    clear H. rename H0 into H.
    unfold channels_incr in H.
    case_eq (ch1 c); [intros p1 Hch1; destruct p1 as [[[p1 n1]]]|intros Hch1];
      (case_eq (ch2 c); [intros p2 Hch2; destruct p2 as [[[p2 n2]]] | intros Hch2]);
      rewrite Hch1 in H; rewrite Hch2 in H; simpl in H; auto.
    destruct (cname_eq c c'); auto. subst c'.
    rewrite Hch1; rewrite Hch2.
    inversion H; clear H.
    unfold proj1_sig in H1.
    destruct (dec_share_identity p1); destruct (dec_share_identity p2).
    inversion H1; subst.
    f_equal; do 2 apply exist_ext; auto.
    elimtype False; apply n; apply endpoint_share_identity; simpl; auto.
    elimtype False; apply n0; apply endpoint_share_identity; simpl; auto. 
    f_equal; do 2 apply exist_ext; auto.
    inversion H1; subst.
    f_equal; omega.
    destruct (cname_eq c c'); auto. subst c'; rewrite Hch2 in H; discriminate.
    destruct (cname_eq c c'); auto. subst c'; rewrite Hch1 in H; discriminate.
  Qed.

  Lemma w_prod_incr_eq: forall c w1 w2, w_prod_incr c w1 = w_prod_incr c w2 -> w1 = w2.
  Proof.
    intros. destruct w1 as [w1_p w1_c w1_e w1_h]; destruct w2 as [w2_p w2_c w2_e w2_h].
    unfold w_prod_incr, w_prod, w_cons, w_env, w_heap in *.
    inversion H; subst; auto; clear H.
    f_equal; auto.
    apply channels_incr_eq with c; auto.
  Qed.

  Lemma w_cons_incr_eq: forall c w1 w2, w_cons_incr c w1 = w_cons_incr c w2 -> w1 = w2.
  Proof.
    intros. destruct w1 as [w1_p w1_c w1_e w1_h]; destruct w2 as [w2_p w2_c w2_e w2_h].
    unfold w_prod_incr, w_prod, w_cons, w_env, w_heap in *.
    inversion H; subst; auto; clear H.
    f_equal; auto.
    apply channels_incr_eq with c; auto.
  Qed.

  (* Either updates the memory address mapping in the heap or fails (returning the same world) *)
  Definition w_heap_set w (a:addr) (v:value) :=
    match MemMap.map_upd a v (w_heap w) with
    | Some h' => wupd_heap w h'
    | None => w
    end.

  Definition w_heap_get (w:world) (a:addr) :=
    MemMap.map_val a (w_heap w).

  Definition w_heap_share (w:world) (a:addr) :=
    MemMap.map_share a (w_heap w).

  (* Either updates the variable mapping in the environment or fails (returning the same world) *)
  Definition w_env_set w (x: ident) (v:value) :=
    match LocalsMap.map_upd x v (w_env w) with
    | Some e' => wupd_env w e'
    | None => w
    end.

  Definition w_env_rm w (x: ident) :=
    match LocalsMap.map_rm x (w_env w) with
    | Some e' => wupd_env w e'
    | None => w
    end.

  Definition w_env_add w x v :=
    match LocalsMap.map_add x v (w_env w) with
    | Some e' => wupd_env w e'
    | None => w
    end.

  Definition w_env_get w (x: ident) :=
    LocalsMap.map_val x (w_env w).

  Lemma w_env_set_rm: forall w x v,
    w_env_rm (w_env_set w x v) x = w_env_rm w x.
  Proof.
    intros.
    unfold w_env_rm, w_env_set.
    destruct w as [w_p w_c w_e w_h].
    simpl.
    case_eq (LocalsMap.map_upd x v w_e); intros; simpl; auto.
    erewrite<- LocalsMap.map_upd_rm_eq; eauto.
    edestruct LocalsMap.map_rm_success as [m' ?]; eauto.
    eapply LocalsMap.map_set_share1; eauto.
    rewrite H0.
    reflexivity.
  Qed.

  Lemma w_env_add_rm: forall w x v,
    w_env_rm (w_env_add w x v) x = w_env_rm w x.
  Proof.
    intros.
    unfold w_env_rm, w_env_add.
    destruct w as [w_p w_c w_e w_h].
    simpl.
    case_eq (LocalsMap.map_add x v w_e); intros; simpl; auto.
    erewrite LocalsMap.map_add_rm; eauto.
    rewrite LocalsMap.map_rm_failure.
    reflexivity.
    intro.
    elim Share.nontrivial.
    apply LocalsMap.map_add_share1 in H.
    congruence.
  Qed.

  Lemma w_prod_env_add: forall w x v,
    w_prod (w_env_add w x v) = w_prod w.
  Proof.
    intros; destruct w. unfold w_env_add. simpl.
    destruct (LocalsMap.map_add x v w_env0); reflexivity.
  Qed.

  Lemma w_cons_env_add: forall w x v,
    w_cons (w_env_add w x v) = w_cons w.
  Proof.
    intros; destruct w. unfold w_env_add. simpl.
    destruct (LocalsMap.map_add x v w_env0); reflexivity.
  Qed.

  Lemma w_heap_env_add: forall w x v,
    w_heap (w_env_add w x v) = w_heap w.
  Proof.
    intros; destruct w. unfold w_env_add. simpl.
    destruct (LocalsMap.map_add x v w_env0); reflexivity.
  Qed.

  Lemma channels_unit_for: forall x, join channels_unit x x.
Admitted.

  Lemma heap_unit_for: forall x, join heap_unit x x.
Admitted.

  Lemma w_env_add_join: forall w x v,
    w_env_get w x = None ->
    gjoin w (new_local x v) (w_env_add w x v).
  Proof.
    intros.
    destruct w as [w_p w_c w_e w_h].
    split; split; unfold fst, snd.
    rewrite w_prod_env_add.
    apply join_com; apply channels_unit_for.
    rewrite w_cons_env_add.
    apply join_com; apply channels_unit_for.
    unfold w_env_get in H.
    unfold w_env_add.
    unfold w_env in *.
    edestruct LocalsMap.map_add_singleton with (b:=v) as [m'[??]].
    apply LocalsMap.map_val_bot; eauto.
    rewrite H0.
    apply H1.
    rewrite w_heap_env_add.
    apply join_com; apply heap_unit_for.
  Qed.

  Lemma w_prod_env_set: forall w x v,
    w_prod (w_env_set w x v) = w_prod w.
  Proof. intros; unfold w_env_set. case (LocalsMap.map_upd x v (w_env w)); auto; apply w_prod_wupd_env. Qed.
  Lemma w_cons_env_set: forall w x v,
    w_cons (w_env_set w x v) = w_cons w.
  Proof. intros; unfold w_env_set. case (LocalsMap.map_upd x v (w_env w)); auto; apply w_cons_wupd_env. Qed.
  Lemma w_heap_env_set: forall w x v,
    w_heap (w_env_set w x v) = w_heap w.
  Proof. intros; unfold w_env_set. case (LocalsMap.map_upd x v (w_env w)); auto; apply w_heap_wupd_env. Qed.

  Lemma w_env_share_new_local1: forall w1 x v w,
    gjoin w1 (new_local x v) w ->
    w_env_share w x = Share.top.
  Proof.
    intros.
    destruct H as [_[? _]].
    unfold fst,snd in H.
    eapply LocalsMap.join_map_singleton; eauto.
    apply H.
  Qed.

  Lemma w_env_set_join: forall w x v,
    w_env_get w x = None ->
    gjoin w (new_local x v) (w_env_add w x v).
  Proof.
    intros.
    destruct w as [w_p w_c w_e w_h].
    split; split; unfold fst, snd.
    rewrite w_prod_env_add.
    apply join_com; apply channels_unit_for.
    rewrite w_cons_env_add.
    apply join_com; apply channels_unit_for.
    unfold w_env_get in H.
    unfold w_env_add.
    unfold w_env in *.
    edestruct LocalsMap.map_add_singleton with (b:=v) as [m'[??]].
    apply LocalsMap.map_val_bot; eauto.
    rewrite H0.
    apply H1.
    rewrite w_heap_env_add.
    apply join_com; apply heap_unit_for.
  Qed.

  Lemma w_env_set_same_share: forall w x v,
    LocalsMap.map_share x (w_env w) = LocalsMap.map_share x (w_env (w_env_set w x v)).
  Proof.
    intros. unfold w_env_set.
    destruct w as [w_p w_c w_e w_h].
    simpl.
    case_eq (LocalsMap.map_upd x v w_e); intros; simpl; auto.
    apply LocalsMap.map_set_share3 with x v; auto.
  Qed.

  Lemma gjoin_w_env_set': forall w1 w2 w x v,
    LocalsMap.map_share x (w_env w1) = Share.top ->
    gjoin w1 w2 w ->
    gjoin (w_env_set w1 x v) w2 (w_env_set w x v).
  Proof.
    intros.
    unfold w_env_set.
    destruct w as [w_p w_c w_e w_h]. destruct w1 as [w1_p w1_c w1_e w1_h]; destruct w2 as [w2_p w2_c w2_e w2_h].
    unfold w_prod, w_cons, w_env, w_heap in *.

    assert (LocalsMap.map_share x w_e = Share.top).
      destruct H0 as [_ [? _]].
      unfold w_env in *.
      unfold fst,snd in H0.
      apply LocalsMap.map_join_char with (a:=x) in H0.
      destruct H0 as [? _].
      rewrite H in H0.
      simpl in H0.
      destruct H0.
      rewrite<- H1.
      rewrite Share.lub_commute.
      apply Share.lub_top.
    edestruct LocalsMap.map_upd_success with (v:=v) as [l'].
      apply H.

    edestruct (LocalsMap.map_upd_join w1_e w2_e w_e x) as [l'''];
      eauto.
    apply H0.
    destruct H3.
    rewrite H2.
    rewrite H3.
      destruct H0 as [[?][?]].
      unfold fst, snd in *.
      split; split; unfold fst, snd in *; intuition.
  Qed.
  Lemma gjoin_w_env_set2: forall w1 w2 w3 x v v0,
    gjoin (w_env_set w1 x v) w2 w3 ->
    w_env_get w1 x = Some v0 ->
    gjoin w1 (w_env_set w2 x v0) (w_env_set w3 x v0).
  Proof.
admit.
  Qed.

  Lemma gjoin_w_env_set_sub: forall w w1 x v,
    gjoin_sub (w_env_set w1 x v) w ->
    LocalsMap.map_val x (w_env w) = Some v.
  Proof.
  Admitted.

  Lemma gjoin_w_env_set'': forall w1 w2 w x v,
    LocalsMap.map_share x (w_env w1) = Share.top ->
    gjoin (w_env_set w1 x v) w2 (w_env_set w x v) ->
    gjoin w1 w2 w.
  Proof.
  Admitted.

Lemma gjoins_w_env_set'': forall w1 w2 x v,
  gjoins w1 w2 <-> gjoins (w_env_set w1 x v) w2.
Proof.
  intros.
  destruct (Share.eq_dec (LocalsMap.map_share x (w_env w1)) Share.top).
  split; introv H.
  destruct H as [w12' H].
  apply gjoin_w_env_set' with (x:=x) (v:=v) in H; auto.
  eexists; eauto.

  destruct H as [w12 H].
  replace w12 with (w_env_set w12 x v) in H.
  apply gjoin_w_env_set'' with (x:=x) (v:=v) in H; auto.
  eexists; eauto.
  rewrite w_env_set_same_share with (v:=v) in e.
  unfold w_env_set.
  destruct LocalsMap.map_upd_success with x v (w_env w12); auto.
  destruct H as [_ [H _]]. unfold fst,snd in *.
  apply LocalsMap.map_join_char with (a:=x) in H.
  destruct H as [[_ H] _].
  rewrite e in H.
  rewrite Share.lub_commute in H.
  rewrite Share.lub_top in H.
  rewrite H; auto.
  rewrite H0.
  apply join_join_sub in H.
  apply gjoin_w_env_set_sub in H; auto.
admit.

  unfold w_env_set.
  case_eq (LocalsMap.map_upd x v (w_env w1)); intros.
  apply LocalsMap.map_set_share1 in H.
  contradiction.
  split; auto.
Qed.

  Lemma join_sub_locals_map_upd_None_Some: forall m1 m3 m3' x v,
    @join_sub _ locals_sa m1 m3 ->
    LocalsMap.map_upd x v m1 = None ->
    LocalsMap.map_upd x v m3 = Some m3' ->
    @join_sub _ locals_sa m1 m3'.
Admitted.


(*
  Program Definition glift_loc (P:locpred) : pred gworld :=
    fun w => P (gw_env w) /\ identity (gw_heap w) /\ identity (gw_prod w) /\ identity (gw_cons w).
  Solve Obligations using discriminate.
*)
  Program Definition lift_loc (P:locpred) : pred world :=
    fun w => P (w_env w).
  Solve Obligations using discriminate.

Program Definition ganylocals : pred world :=
    fun w => identity (w_heap w) /\ identity (w_prod w) /\ identity (w_cons w).
  Solve Obligations using discriminate.


  Program Definition lift_heap (P:heappred) : pred world :=
    fun w => P (w_heap w) /\ identity (w_env w) /\ identity (w_prod w) /\ identity (w_cons w).
  Solve Obligations using discriminate.

  Coercion lift_loc : locpred >-> pred.
  Coercion lift_heap : heappred >-> pred.

  Program Definition prod_endpoint (c:cname) (p:Share.t) (count:nat) : pred world :=
    fun w =>
      (endpoint c p count (w_prod w)) /\ identity (w_env w) /\ identity (w_heap w) /\ identity (w_cons w).
  Solve Obligations using discriminate.

  Program Definition cons_endpoint (c:cname) (p:Share.t) (count:nat) : pred world :=
    fun w =>
      (endpoint c p count (w_cons w)) /\ identity (w_env w) /\ identity (w_heap w) /\ identity (w_prod w).
  Solve Obligations using discriminate.

  Definition locals_to_map (l:locals) := fun x => LocalsMap.map_val x l.
  Definition heap_to_map (h:heap) := fun a => MemMap.map_val a h.

  Program Definition eval (e:expr) (v:value) : pred world :=
    fun w => e (locals_to_map (w_env w)) = Some v.
  Solve Obligations using discriminate.

  Program Definition heval (e:hexpr) (v:value) : pred world :=
    fun w => e (heap_to_map (w_heap w)) (locals_to_map (w_env w)) = Some v.
  Solve Obligations using discriminate.

  Definition eval_expr (e:expr) l := proj1_sig e l.
  Coercion eval_expr: expr >-> Funclass.

  Definition eval_hexpr (e:hexpr) h l := proj1_sig e h l.
  Coercion eval_hexpr: hexpr >-> Funclass.

  Program Definition expr_to_hexpr (e:expr) : hexpr := fun h l => e l.
  Next Obligation. destruct e; eauto. Qed.

  Program Definition hexpr_to_expr (e:hexpr) : expr := e (fun _ => None).
  Next Obligation. destruct e; simpl in *. eapply e; eauto. intros; discriminate. Qed.

  Coercion expr_to_hexpr : expr >-> hexpr.

  Definition evalTrue e w :=
   exists (n:nat), eval e (nat_value n) w /\ n<>0.
  Definition evalFalse e w := eval e (nat_value 0) w.

  Lemma locals_equiv_eq: forall (l1 l2:locals),
    (forall x, LocalsMap.map_val x l1 = LocalsMap.map_val x l2) -> l1 = l2.
Admitted.

  Lemma wupd_env2: forall w l1 l2,
    wupd_env (wupd_env w l1) l2 = wupd_env w l2.
  Proof.
    intros.
    destruct w; simpl; f_equal.
  Qed.

  Lemma w_env_get_set_same: forall w x v,
    w_env_share w x = Share.top ->
    w_env_get (w_env_set w x v) x = Some v.
  Proof.
    unfold w_env_get, w_env_set.
    intros.
    remember (LocalsMap.map_upd x v (w_env w)).
    destruct o; auto.
    rewrite w_env_wupd_env.
    erewrite LocalsMap.map_gss_val; eauto.
    false.
    unfold w_env_share in H.
    edestruct LocalsMap.map_upd_success with (v:=v); eauto.
    rewrite<- Heqo in H0; discriminate.
  Qed.

  Lemma w_env_set_get_same: forall w x v,
    w_env_get w x = Some v ->
    w_env_set w x v = w.
  Proof.
    unfold w_env_get, w_env_set.
    intros.
    remember (LocalsMap.map_upd x v (w_env w)).
    destruct o; auto.
    destruct w; simpl in *; f_equal.
    apply locals_equiv_eq.
    intro y.
    destruct (Pos.eq_dec y x).
    subst y.
    erewrite LocalsMap.map_gss_val; eauto.
    rewrite<- (LocalsMap.map_gso_val y x v w_env0 m); eauto.
  Qed.

  Lemma w_env_set_same: forall w x v1 v2,
    w_env_set (w_env_set w x v1) x v2 = w_env_set w x v2.
  Proof.
    intros.
    unfold w_env_set.
    remember (LocalsMap.map_upd x v1 (w_env w)).
    destruct o; auto.
    rewrite w_env_wupd_env.
    edestruct LocalsMap.map_upd_success with (v:=v2).
    eapply LocalsMap.map_set_share1; eauto.
    rewrite H.
    edestruct LocalsMap.map_upd_success with (m:=m) (v:=v2).
    eapply LocalsMap.map_set_share2; eauto.
    rewrite H0.
    rewrite wupd_env2.
    f_equal.
    apply locals_equiv_eq.
    intros.
    destruct (Pos.eq_dec x2 x).
    subst x2.
    do 2 (erewrite LocalsMap.map_gss_val; eauto).
    rewrite<- (LocalsMap.map_gso_val x2 x v2 (w_env w) x0); eauto.
    rewrite<- (LocalsMap.map_gso_val x2 x v2 m x1); eauto.
    rewrite<- (LocalsMap.map_gso_val x2 x v1 (w_env w) m); eauto.
  Qed.

  Definition expr_fun:= value->expr.
  Program Definition expr_fun_app (f:expr_fun) (e:expr) : expr :=
    exist _ (fun (l:ident->option value) =>
    match e l with
    | Some v => f v l
    | None => None
    end) _.
  Next Obligation with auto.
    intros.
    destruct e as [ff H1].
    simpl in *.
    spec H1 l l'.
    destruct (ff l).
    spec H1 v0 H. spec H1. reflexivity.
    rewrite H1.
    rewrite<- H0.
    destruct (f v0) as [f' H2].
    simpl in *.
    rewrite H0.
    eauto.
    discriminate.
  Qed.
  Definition hexpr_fun:= value->hexpr.
  Program Definition hexpr_fun_app (f:hexpr_fun) (e:hexpr) : hexpr :=
    exist _ (fun (h:addr->option value) (l:ident->option value) =>
    match e h l with
    | Some v => f v h l
    | None => None
    end) _.
  Next Obligation with auto.
    intros.
    destruct e as [ff H2].
    simpl in *.
    spec H2 h h' l l'.
    destruct (ff h l).
    spec H2 v0 H H0. spec H2. reflexivity.
    rewrite H2.
    rewrite<- H1.
    destruct (f v0) as [f' H3].
    simpl in *.
    rewrite H1.
    eauto.
    discriminate.
  Qed.

  Definition expr_fun2:= value->value->expr.
  Program Definition expr_fun_app2 (f: expr_fun2) (e1:expr) (e2:expr) : expr:=
    expr_fun_app (fun v1 => expr_fun_app (f v1) e2) e1.

  Definition hexpr_fun2:= value->value->hexpr.
  Program Definition hexpr_fun_app2 (f: hexpr_fun2) (e1:hexpr) (e2:hexpr) : hexpr:=
    hexpr_fun_app (fun v1 => hexpr_fun_app (f v1) e2) e1.

  Program Definition expr_value (v:value) : expr := fun _=> Some v.
  Program Definition expr_value' (v:option value) : expr := fun _=> v.
  Program Definition expr_wrong : expr := fun _=> None.
  Definition expr_nat (n:nat) : expr := expr_value (nat_value n).
  Coercion expr_nat : nat >-> expr.

  Program Definition expr_var (x: ident) : expr:=
    exist _ (fun l:ident->option value => l x) _.
  Coercion expr_var : ident >-> expr.

  Program Definition expr_fetch (a: addr) : hexpr:=
    exist _ (fun (h:addr->option value) _ => h a) _.

  Lemma w_heap_get_env_set: forall w x v a,
    w_heap_get (w_env_set w x v) a = w_heap_get w a.
Admitted.

  Lemma eval_dec: forall e w,
    {v | eval e v w} + {forall v, ~eval e v w}.
Admitted.

  Lemma heval_dec: forall e w,
    {v | heval e v w} + {forall v, ~heval e v w}.
Admitted.

  Definition value_plus v1 v2:= match v1, v2 with
    | Some (nat_value n1), Some (nat_value n2) => Some (nat_value (n1+n2))
    | Some (addr_value a1), Some (nat_value n2) => Some (addr_value (a1+n2))
    | Some (nat_value n1), Some (addr_value a2) => Some (addr_value (n1+a2))
    | Some (cname_value c1), Some (nat_value n2) => Some (cname_value (Pos.of_nat (Pos.to_nat c1 + n2)))
    | Some (nat_value n1), Some (cname_value c2) => Some (cname_value (Pos.of_nat (n1 + Pos.to_nat c2)))
    | _, _ => None
    end.

  Definition expr_plus :=
    expr_fun_app2 (fun v1 v2 => expr_value' (value_plus (Some v1) (Some v2))).

  Definition hexpr_plus :=
    hexpr_fun_app2 (fun v1 v2 => expr_value' (value_plus (Some v1) (Some v2))).

  Definition expr_equiv (e1 e2: expr):= (forall (l:ident->option value), e1 l = e2 l).
  Definition hexpr_equiv (e1 e2: hexpr):= (forall (h:addr->option value) (l:ident->option value), e1 h l = e2 h l).

  Lemma expr_fun_equiv: forall f1 f2 e1 e2,
    (forall ve, expr_equiv (f1 ve) (f2 ve)) ->
    (expr_equiv e1 e2) ->
    (expr_equiv (expr_fun_app f1 e1) (expr_fun_app f2 e2)).
  Proof.
    unfold expr_equiv.
    intros.
    unfold expr_fun_app.
    simpl.
    destruct e1; destruct e2.
    simpl in *.
    remember (x l).
    destruct o.
    rewrite<- H0.
    rewrite<- Heqo.
    auto.
    remember (x0 l).
    destruct o.
    false.
    reflexivity.
  Qed.

  Lemma hexpr_fun_equiv: forall f1 f2 e1 e2,
    (forall ve, hexpr_equiv (f1 ve) (f2 ve)) ->
    (hexpr_equiv e1 e2) ->
    (hexpr_equiv (hexpr_fun_app f1 e1) (hexpr_fun_app f2 e2)).
  Proof.
    unfold hexpr_equiv.
    intros.
    unfold hexpr_fun_app.
    simpl.
    destruct e1; destruct e2.
    simpl in *.
    remember (x h l).
    destruct o.
    rewrite<- H0.
    rewrite<- Heqo.
    auto.
    remember (x0 h l).
    destruct o.
    false.
    reflexivity.
  Qed.

  Lemma expr_fun2_equiv: forall f1 f2 e1 e2 e3 e4,
    (forall ve1 ve2, expr_equiv (f1 ve1 ve2) (f2 ve1 ve2)) ->
    (expr_equiv e1 e3) ->
    (expr_equiv e2 e4) ->
    (expr_equiv (expr_fun_app2 f1 e1 e2) (expr_fun_app2 f2 e3 e4)).
  Proof.
    intros.
    unfold expr_fun_app2.
    simpl.
    apply expr_fun_equiv; intros; auto.
    apply expr_fun_equiv; intros; auto.
  Qed.

  Lemma hexpr_fun2_equiv: forall f1 f2 e1 e2 e3 e4,
    (forall ve1 ve2, hexpr_equiv (f1 ve1 ve2) (f2 ve1 ve2)) ->
    (hexpr_equiv e1 e3) ->
    (hexpr_equiv e2 e4) ->
    (hexpr_equiv (hexpr_fun_app2 f1 e1 e2) (hexpr_fun_app2 f2 e3 e4)).
  Proof.
    intros.
    unfold hexpr_fun_app2.
    simpl.
    apply hexpr_fun_equiv; intros; auto.
    apply hexpr_fun_equiv; intros; auto.
  Qed.

  Lemma eval_expr_fun_app: forall (f:expr_fun) (e:expr) (v v':value) l,
    e l = Some v ->
    (f v) l = Some v' ->
    (expr_fun_app f e) l = Some v'.
  Proof.
    intros.
    simpl.
    rewrite H; auto.
  Qed.

  Lemma eval_hexpr_fun_app: forall (f:hexpr_fun) (e:hexpr) (v v':value) h l,
    e h l = Some v ->
    (f v) h l = Some v' ->
    (hexpr_fun_app f e) h l = Some v'.
  Proof.
    intros.
    simpl.
    rewrite H; auto.
  Qed.

  Lemma expr_fun_app_eval: forall f v,
    expr_equiv (expr_fun_app f (expr_value v)) (f v).
  Proof.
    unfold expr_equiv, expr_fun_app.
    intros.
    simpl.
    reflexivity.
  Qed.

  Lemma hexpr_fun_app_eval: forall f v,
    hexpr_equiv (hexpr_fun_app f (expr_value v)) (f v).
  Proof.
    unfold hexpr_equiv, hexpr_fun_app.
    intros.
    simpl.
    reflexivity.
  Qed.

  Definition locals_map_upd (l:ident->option value) (x:ident) (v:option value) :=
    fun y => if Pos.eq_dec y x then v else l y.

  Definition not_free_in_expr (x:ident) (e:expr) :=
    forall l v, e l = e (locals_map_upd l x v).


  Program Definition expr_subst (e:expr) (x:ident) (ex:expr) : expr :=
    exist _ (fun (l:ident->option value) => e (locals_map_upd l x (ex l))) _.
  Next Obligation with auto.
    destruct e as [fe H1].
    simpl in *.
    eapply H1. intros y v' ?.
    2: apply H0.
    unfold locals_map_upd in *.
    destruct (Pos.eq_dec y x); subst; auto.
    destruct ex as [fex H3].
    simpl in *.
    eapply H3; eauto.
  Qed.

  Lemma expr_subst_equiv: forall x e e' ex ex',
    (expr_equiv e e') ->
    (expr_equiv ex ex') ->
    (expr_equiv (expr_subst e x ex) (expr_subst e' x ex')).
  Proof.
    intros. intro l.
    simpl.
    rewrite H.
    rewrite H0.
    reflexivity.
  Qed.

  Lemma w_heap_share_env_set: forall w x v a,
    w_heap_share (w_env_set w x v) a = w_heap_share w a.
  Proof.
    intros.
    unfold w_heap_share, w_env_set.
    case_eq (LocalsMap.map_upd x v (w_env w)); intros; auto.
    rewrite w_heap_wupd_env; eauto.
  Qed.

  Lemma eval_wupd_heap: forall e v w h',
    eval e v w <-> eval e v (wupd_heap w h').
  Proof.
    intros. simpl. rewrite w_env_wupd_heap; reflexivity.
  Qed.

  Lemma eval_w_heap_set: forall e v w a v',
    eval e v w <-> eval e v (w_heap_set w a v').
  Proof.
    intros. simpl.
    unfold w_heap_set.
    destruct MemMap.map_upd; intros.
    rewrite w_env_wupd_heap; reflexivity.
    reflexivity.
  Qed.

  Lemma w_env_share_w_heap_set: forall w a v x,
    w_env_share (w_heap_set w a v) x = w_env_share w x.
  Proof.
    intros.
    unfold w_heap_set.
    destruct MemMap.map_upd; auto.
    unfold w_env_share.
    rewrite w_env_wupd_heap; eauto.
  Qed.

  Lemma wupd_heap_env: forall w e' h',
    wupd_heap (wupd_env w e') h' = wupd_env (wupd_heap w h') e'.
  Proof.
    destruct w; simpl; reflexivity.
  Qed.

  Lemma w_heap_set_w_env_set: forall w x v a v',
    w_heap_set (w_env_set w x v) a v' = w_env_set (w_heap_set w a v') x v.
  Proof.
    intros.
    unfold w_heap_set, w_env_set.
    case_eq (LocalsMap.map_upd x v (w_env w)); intros.
    rewrite w_heap_wupd_env.
    case_eq (MemMap.map_upd a v' (w_heap w)); intros.
    rewrite w_env_wupd_heap.
    rewrite H.
    apply wupd_heap_env.
    rewrite H; auto.
    case_eq (MemMap.map_upd a v' (w_heap w)); intros.
    rewrite w_env_wupd_heap; rewrite H; auto.
    rewrite H; auto.
  Qed.


  Lemma gjoin_new_local: forall w1 w x v,
    gjoin w1 (new_local x v) w ->
    w_env_get w x = Some v /\ w_env_share w1 x = Share.bot.
  Proof.
    intros.
    destruct w1 as [w1_p w1_c w1_e w1_h].
    destruct w as [w_p w_c w_e w_h].
    unfold new_local in H.
    destruct H as [[??][??]].
    unfold fst,snd,w_prod,w_cons,w_env,w_heap in *.
    split.
    eapply LocalsMap.join_map_singleton; eauto.
    eapply LocalsMap.join_map_singleton; eauto.
  Qed.

  Lemma gjoin_new_local_set_value: forall w1 w x v1 v2,
    gjoin w1 (new_local x v1) w ->
    gjoin w1 (new_local x v2) (w_env_set w x v2).
  Proof.
    intros.
    destruct w1 as [w1_p w1_c w1_e w1_h].
    destruct w as [w_p w_c w_e w_h].
    unfold new_local in *.
    destruct H as [[??][??]].
    split; split;
      try rewrite w_prod_env_set;
      try rewrite w_cons_env_set;
      try rewrite w_heap_env_set;
      unfold fst,snd, w_prod,w_cons,w_env,w_heap in *.
    apply join_com in H; apply channel_sa.channels_unit_identity in H;
      subst; apply join_com; apply channels_unit_for.
    apply join_com in H0; apply channel_sa.channels_unit_identity in H0;
      subst; apply join_com; apply channels_unit_for.

    edestruct LocalsMap.join_map_singleton as [?[?[??]]]; eauto.
    edestruct LocalsMap.map_add_singleton with (b:=v2) as [m3[??]]; eauto.
    unfold w_env_set, w_env.
    edestruct LocalsMap.map_upd_singleton with (m3:=w_e) (v2:=v2) as [m'[??]]; eauto.
    rewrite H9.
    unfold wupd_env.
    apply H10.

    apply join_com in H2; apply MemMap.empty_map_identity in H2.
      subst; apply join_com; apply heap_unit_for.
  Qed.


  (* Local i has some unspecified value *)
  Definition defined (i:ident) : pred world :=
    Ex_ v:value, hasval i v.

  Lemma split_endpoint: forall c p q pq n m w,
    join p q pq ->
    endpoint c pq (n+m) w -> exists w1, exists w2, join w1 w2 w /\ endpoint c p n w1 /\ endpoint c q m w2.
  Proof.
  Admitted.

  Lemma pjoinable_join_unit: forall {A:Type} {sa:sepalg A} (a:option (pjoinable sa)),
    @join _ (sa_lift sa) a None a.
  Proof. intros; simpl; constructor. Qed.

  Lemma pfun_join_unit: forall {A B:Type} {sa:sepalg A} (a:B->option (pjoinable sa)),
    @join _ (sa_fun B (sa_lift sa)) a (fun _=>None) a.
  Proof. intros; simpl; constructor. Qed.

  Lemma map_join_unit: forall {A B:Type} {sa:sepalg A} (a:B->option (pjoinable sa)),
    @join _ (sa_fun B (sa_lift sa)) a (fun _=>None) a.
  Proof. intros; simpl; constructor. Qed.

  Lemma split_prod: forall c (p q pq: Share.t) n m,
    join p q pq ->
    prod_endpoint c (pq) (n+m) |-- prod_endpoint c p n * prod_endpoint c q m.
  Proof.
    repeat intro.
    destruct H0 as [?[?[??]]].
    edestruct split_endpoint as [ce1[ce2[?[??]]]]; eauto.
    exists (MakeWorld ce1 (w_cons a) (w_env a) (w_heap a)).
    exists (MakeWorld ce2 cons_unit (w_env a) heap_unit).
    split.
    split; simpl.
    split; auto. apply pfun_join_unit.
    split; simpl. split; auto. apply pfun_join_unit.
    split; simpl.
    intuition.
    intuition.
    apply MemMap.empty_map_identity.
    apply channels_unit_identity.
  Qed.

  Lemma split_cons: forall c p q pq n m,
    join p q pq ->
    cons_endpoint c (pq) (n+m) |-- cons_endpoint c p n * cons_endpoint c q m.
  Proof.
    repeat intro.
    destruct H0 as [?[?[??]]].
    edestruct split_endpoint as [ce1[ce2[?[??]]]]; eauto.
    exists (MakeWorld (w_prod a) ce1 (w_env a) (w_heap a)).
    exists (MakeWorld prod_unit ce2 (w_env a) heap_unit).
    split.
    split; simpl.
    split; auto. apply pfun_join_unit.
    split; simpl. split; auto. apply pfun_join_unit.
    split; simpl.
    intuition.
    intuition.
    apply MemMap.empty_map_identity.
    apply channels_unit_identity.
  Qed.

  Program Definition exactly w : pred world := fun w' => w'=w.
  Solve Obligations using discriminate.
  Lemma world_eq_dec (w1 w2: world) : {w1=w2}+{w1<>w2}. Admitted.

  Definition wemp:=world_unit.
  Definition gwemp:=gworld_unit.
  Lemma join_bot1: forall w l, join w (wemp l) w. Admitted.
  Lemma join_bot2: forall w l, join (wemp l) w w. Admitted.
  Lemma gjoin_bot1: forall w, gjoin w gwemp w. Admitted.
  Lemma gjoin_bot2: forall w, gjoin gwemp w w. Admitted.
  Lemma join_dec (w1 w2: world) : {w | join w1 w2 w}+{forall w, ~join w1 w2 w}. Admitted.
  Lemma gjoin_dec (w1 w2: world) : {w | gjoin w1 w2 w}+{forall w, ~gjoin w1 w2 w}. Admitted.

  Definition can_send c (w: world) : Prop :=
    has_access c (w_prod w).

  Definition can_recv c (w: world) : Prop :=
    has_access c (w_cons w).

  Definition total_send_recv (c:cname) (w: world) : Prop :=
    has_total_endpoint c (w_prod w) /\ has_total_endpoint c (w_cons w).

(*
  Definition pCanSend (c:cname) : pred world :=
    Ex_ p:Share.t, Ex_ count:nat, !!(~identity p) && prod_endpoint c p count * TT.

  Definition pCanRecv (c:cname) : pred world :=
    Ex_ p:Share.t, Ex_ count:nat, !!(~identity p) && cons_endpoint c p count * TT.

  Definition pTotalSendRecv (c:cname) : pred world :=
    Ex_ count:nat, prod_endpoint c Share.top count * cons_endpoint c Share.top count.
*)

  Definition pSend ec p e : pred world :=
    Ex_ c:cname, Ex_ v:nat, eval ec (cname_value c) && eval e (nat_value v) && !!(~identity p) && prod_endpoint c p v.

  Definition pRecv ec p e : pred world :=
    Ex_ c:cname, Ex_ v:nat, eval ec (cname_value c) && eval e (nat_value v) && !!(~identity p) && cons_endpoint c p v.

  Definition pIsTrue (e: expr) : pred world :=
    Ex_ n:nat, eval e (nat_value n) && !!(n <> 0).

  Definition pIsFalse (e: expr) : pred world := eval e (nat_value 0).

  Program Definition pNeg (P: pred world) : pred world :=
    exist _ (fun w => ~P w) _.
  Solve Obligations using discriminate.


  Definition disjoint (P1 P2: pred world) := forall w1 w2,
    P1 w1 -> P2 w2 -> joins w1 w2.

  Program Definition no_channels : pred world :=
    fun w => w_prod w = channels_unit /\ w_cons w = channels_unit.
  Solve Obligations using discriminate.

  Program Definition no_heap : pred world :=
    fun w => w_heap w = heap_unit.
  Solve Obligations using discriminate.


  Lemma eval_star: forall e v x,
    eval e v * x |-- eval e v.
  Proof.
    unfold eval; repeat intro.
    simpl in *.
    destruct H as [y[z[?[??]]]].
    destruct H as [_ [? _]].
    simpl in H. destruct H.
    rewrite<- H2. rewrite<- H. auto.
  Qed.

  Lemma disjoint_send_recv: forall c1 p1 n1 c2 p2 n2,
    disjoint (pSend c1 p1 n1) (pRecv c2 p2 n2).
  Proof.
  Admitted.

  Definition locals_pred (P:pred world) := forall w w',
    P w -> @join_sub locals locals_sa (w_env w) (w_env w') -> P w'.


  Lemma locals_pred_up: forall (A P Q R : pred world),
    locals_pred A ->
    P && R && A * Q |-- P && R * Q && A.
  Proof.
    repeat intro.
    destruct H0 as [wp [wq [?[[[??]?]?]]]].
    split.
    exists wp; exists wq.
    intuition. split; auto.
    apply H with wp; auto.
    destruct H0 as [_ [? _]].
    unfold fst, snd in *.
    inversion H0. rewrite H5 in *; rewrite H6 in *.
    exists locals_unit.
    apply join_com. apply LocalsMap.empty_map_join.
  Qed.

  Lemma locals_pred_up': forall (A P Q : pred world),
    locals_pred A ->
    P && A * Q |-- P * Q && A.
  Proof.
    intros.
    rewrite <- (TT_and A) at 1.
    rewrite <- (TT_and P) at 2.
    rewrite (and_comm TT P).
    rewrite and_assoc.
    apply locals_pred_up; auto.
  Qed.

  Lemma join_sub_locals_to_map: forall m1 m3,
    @join_sub _ locals_sa m1 m3 ->
    (forall x v, (locals_to_map m1) x = Some v -> (locals_to_map m3)  x = Some v).
  Proof.
    unfold locals_to_map.
    intros.
    destruct H as [m2 ?].
    eapply LocalsMap.map_join_char with (a:=x) in H.
    intuition.
    rewrite<- H3. rewrite<- H2. auto.
    false.
    rewrite<- H. auto.
  Qed.

  Lemma join_sub_locals_to_map': forall m1 m3,
    @join_sub _ locals_eqv_sa m1 m3 ->
    (forall x v, (locals_to_map m3) x = Some v -> (locals_to_map m1)  x = Some v).
  Proof.
    unfold locals_to_map.
    intros.
    destruct H as [m2 ?].
    destruct H.
    subst; auto.
  Qed.

  Lemma locals_pred_true_expr: forall (e:expr),
    locals_pred (pIsTrue e).
  Proof.
    unfold locals_pred; intros.
    destruct H as [n].
    exists n.
    destruct H; split; auto.
    destruct e as [fe H2]. simpl in *.
    apply H2 with (l:= locals_to_map (w_env w)); intros; auto.
    eapply join_sub_locals_to_map; eauto.
  Qed.

  Lemma locals_pred_eval: forall (e:expr) v,
    locals_pred (eval e v).
  Proof.
    unfold locals_pred; intros.
    destruct e as [fe H2]. simpl in *.
    apply H2 with (l:= locals_to_map (w_env w)); intros; auto.
    eapply join_sub_locals_to_map; eauto.
  Qed.

  Lemma locals_pred_false_expr: forall (e:expr),
    locals_pred (pIsFalse e).
  Proof. intros; apply locals_pred_eval. Qed.

  Lemma eval_up : forall e v P Q R,
    (P && R && eval e v) * Q |--
    ((P && R) * Q) && eval e v.
  Proof.
    intros.
    apply locals_pred_up.
    apply locals_pred_eval.
  Qed.

  Lemma eval_up' : forall e v P Q,
    (P && eval e v) * Q |--
    (P * Q) && eval e v.
  Proof.
    intros.
    apply locals_pred_up'.
    apply locals_pred_eval.
  Qed.

  Definition pEvalEq {A:Type} {age:ageable A} e1 e2 : pred world := existential (fun v => eval e1 v && eval e2 v).

Program Definition env_wand (x y:pred world): pred world :=
  (fun w1 => forall w2 w, gjoin w1 w2 w -> x (MakeWorld (w_prod w2) (w_cons w2) (w_env w) (w_heap w2)) -> y w).
Solve Obligations using discriminate.
Definition locals_imprecise (x:pred world):= forall w l' w',
  x w ->
  @join_sub locals locals_sa (w_env w) l' ->
  w' = (MakeWorld (w_prod w) (w_cons w) l' (w_heap w)) ->
  x w'.

  Program Definition gsep_con (x y: pred world) : pred world:=
    (fun w => exists w1, exists w2, gjoin w1 w2 w /\ x w1 /\ y w2).
  Solve Obligations using discriminate.

  Program Definition sep_overlap {A:Type} `{ASA A} (x y: pred A) : pred A := fun w:A => 
    exists (w1 w2 w3 w12 w23: A), join w1 w2 w12 /\ join w2 w3 w23 /\ join w12 w3 w /\ x w12 /\ y w23.
  Next Obligation.
    repeat intro.
    destruct H1 as [w1[w2[w3[w12[w23[?[?[?[??]]]]]]]]].
    destruct (age1_join2 _ H3 H0) as [w12' [w3' [? [? ?]]]].
    destruct (age1_join2 _ H1 H7) as [w1' [w2' [? [? ?]]]].
    destruct (age1_join _ H2 H11) as [w3''[w23'[?[??]]]].
    assert (w3' = w3'').
      clear -H13 H8.
      unfold age in *.
      rewrite H8 in H13. inverts H13; auto.
    subst w3''.
    exists w1' w2' w3' w12' w23'; splits; auto.
    apply pred_hereditary with w12; auto.
    apply pred_hereditary with w23; auto.
  Qed.

  Program Definition gsep_overlap := @sep_overlap world gworld_sa world_ageable gworld_asa.


  Definition expr_bool_equiv (e1 e2: expr):=
    (forall w, evalFalse e1 w <-> evalFalse e2 w) /\ (forall w, evalTrue e1 w <-> evalTrue e2 w).

  Lemma expr_bool_equiv_refl: forall e, expr_bool_equiv e e.
  Proof. split; split; intros; auto. Qed.
  Lemma expr_bool_equiv_symm: forall e1 e2, expr_bool_equiv e1 e2 -> expr_bool_equiv e2 e1.
  Proof. split; split; intros; apply H; auto. Qed.
  Lemma expr_bool_equiv_trans: forall e1 e2 e3, expr_bool_equiv e1 e2 -> expr_bool_equiv e2 e3 -> expr_bool_equiv e1 e3.
  Proof. split; (split; intros; [apply H0; apply H | apply H; apply H0]; auto). Qed.

  Definition world_set_vars (w: world) (l':locals) :=
    match w with
    | MakeWorld p c l h => MakeWorld p c l' h
    end.

  Lemma gjoin_join2: forall w1 w2 w,
    w_env w2 = locals_unit ->
    gjoin w1 w2 w -> join w1 (world_set_vars w2 (w_env w1)) w.
  Proof.
    intros.
    unfold gjoin in H0.
    destruct w1; destruct w2; destruct w.
    destruct H0 as [[??][??]]; simpl in *.
    split; split; auto.
    simpl.
    clear H0 H1 H3.
    constructor; auto.
    apply locals_equiv_eq.
    intro x.
    spec H2 x.
    simpl in H2.
    rewrite H in H2.
    simpl in H2.
    unfold LocalsMap.map_val, lookup_fpm.
    inverts H2; auto.
  Qed.

  Lemma evalTrueFalse_dec: forall e w, {evalTrue e w}+{evalFalse e w}+{~evalTrue e w /\ ~evalFalse e w}.
  Proof.
    unfold evalTrue, evalFalse; intros.
    destruct e. simpl in *.
    destruct (x (locals_to_map (w_env w))) as [[?|?|?]|].
    destruct (eq_nat_dec n 0). subst.
    left; right; auto.
    left; left; exists n; split; auto.
    right; split. intros [n[??]]; discriminate.
    discriminate.
    right; split. intros [n[??]]; discriminate.
    discriminate.
    right; split. intros [n[??]]; discriminate.
    discriminate.
  Qed.
  Lemma evalTrue_evalFalse: forall e w, evalTrue e w -> ~evalFalse e w.
  Proof.
    unfold evalTrue, evalFalse; intros.
    destruct H as [n[??]].
    intro.
    destruct e. simpl in *.
    rewrite H in H1. inverts H1. intuition.
  Qed.
  Lemma evalFalse_evalTrue: forall e w, evalFalse e w -> ~evalTrue e w.
  Proof.
    unfold evalTrue, evalFalse; intros.
    intros [n[??]].
    destruct e. simpl in *.
    rewrite H in H0. inverts H0. intuition.
  Qed.
  Lemma evalTrue_dec: forall e w, {evalTrue e w}+{~evalTrue e w}.
  Proof.
    intros.
    destruct (evalTrueFalse_dec e w) as [[?|?]|[??]]; auto.
    right; apply evalFalse_evalTrue; auto.
  Qed.
  Lemma evalFalse_dec: forall e w, {evalFalse e w}+{~evalFalse e w}.
  Proof.
    intros.
    destruct (evalTrueFalse_dec e w) as [[?|?]|[??]]; auto.
    right; apply evalTrue_evalFalse; auto.
  Qed.

  Lemma eval_eq: forall e v v' w,
    eval e v w ->
    eval e v' w ->
    v' = v.
  Proof.
    simpl; intros. rewrite H in H0; inverts H0; reflexivity.
  Qed.

  

End WorldModel.


Existing Instance world_sa.
Existing Instance world_ageable.
Existing Instance world_asa.


Module Notation.
  Export predicates_sl.Notation.

  Delimit Scope forces_pred with forces_pred.
  Notation "w '|=' P" := (app_pred P%pred w)
   (no associativity, at level 0, P at level 40) : forces_pred.
  Notation "~ P" := (pNeg P%pred) : pred.

  Notation "P '(*)' Q" := (gsep_con P%pred Q%pred) (at level 40, left associativity) : pred.

  Delimit Scope expr_scope with expr.
  Notation " e1 + e2" := (expr_plus e1%expr e2%expr) : expr_scope.
  Notation "e1 [ e2 / x ]" := (expr_subst e1%expr x e2%expr)
     (no associativity, at level 9, format "e1 [ e2 / x ]") : expr_scope.


  Module U.
    Notation "P '⊢' Q" := (derives P%pred Q%pred) (at level 80, no associativity) : entails_pred.
    Notation "e '⇓' v" := (eval e%expr v) (at level 50, left associativity) : pred.
    Notation "e1 '≎' e2" := (eval_eq e1%expr e2%expr) (at level 30) : pred.
    Notation "a '↦' v" := (mapsto a Share.top v) (at level 50, left associativity) : pred.
    Notation "'∃'  x ':' T ',' P " := (existential (fun x:T => P%pred)) (at level 65, x at level 99) : pred.
    Notation "'∀'  x ':' T  ',' P " := (universal (fun x:T => P%pred)) (at level 65, x at level 99) : pred.
    Infix "⋁" := union (at level 50, left associativity) : pred.
    Infix "⋀" := intersection (at level 40, left associativity) : pred.
    Notation "P '⇒' Q" := (implies P%pred Q%pred) (at level 55, right associativity) : pred.
    Notation "P '⇔' Q" := (intersection (implies P%pred Q%pred) (implies Q%pred P%pred)) (at level 57, no associativity) : pred.
    Notation "'⎕' e" := (box necM e) (at level 30, right associativity): pred.
    Notation "'⊳' e" := (box laterM e) (at level 30, right associativity): pred.
    Notation "'‼' e" := (prop e) (at level 25) : pred.
    Notation "P '⊑' Q" := (box fashionM (implies P%pred Q%pred)) (at level 55, right associativity) : pred.
    Notation "P '⊒⊑' Q" := (box fashionM (intersection (implies P%pred Q%pred) (implies Q%pred P%pred))) (at level 57, no associativity) : pred.
    Notation "P '⋆' Q" := (sep_con P%pred Q%pred) (at level 40, left associativity) : pred.
    Notation "P '⍟' Q" := (gsep_con P%pred Q%pred) (at level 40, left associativity) : pred.
    Notation "P '⋓' Q" := (sep_overlap P%pred Q%pred) (at level 40, left associativity) : pred.
    Notation "P '⤙' Q" := (wand P%pred Q%pred) (at level 60, right associativity) : pred.
    Notation "'%' e"  := (box extendM e) (at level 30, right associativity): pred.
    Notation "'ε'"  := (emp) (at level 30, no associativity): pred.

    Delimit Scope forces_pred with forces_pred.
    Notation "w ⊨ P" := (app_pred P%pred w)
     (no associativity, at level 0, P at level 40) : forces_pred.
    Notation "¬ P" := (pNeg P%pred) (at level 75, right associativity) : pred.


(*
⊗⊘⊙⊚⊛⊜⊝⊞⊟⊠⊡⊕⊖⌾
        ҈       ҉‗₌≐≑≒≔≕≖≗≘
‒⋆ –⋆ —⋆―⋆‖⎻⎼⎽⌌▁⟝⟞⟘⟙◌⦀⦓⦔⧿
⌗⎕⊡⌷⍞
①②⏏❶❷▒◌♠♡⟘⟙⟡ᴥᴛ
⍟⊺≣≡≎≏≈∧∨∩∪∗
◊○█├⤙⤞⤳⥇⧠⨁⨅⨆⨟⨹⨯⫏⫢⫿⬄⦂⦃⦄⦅⦆ε
¬├⊢∀⇒⇐⇔‼
↓↔↕∆∏∑∩≈←↑→
∩∪⋂⋃⦇⦈⦉⦊〔〕
↓↔↕∆∏∑∩≈←↑→↪⇊⇓∃∇∈∉∏∏≤≥≲≾≿⊂⊆⊎⊏⊐⊑⊒⊕⨁⊕⊗⊖⊚⊛⊦⊣⊤⊥
⊧⊨⊩⊪⊫⊳⨞⊲⋈⋉⋊⋀⋁⋂⋃
⊌⊍⊎⋒⋓⍝⨃⨄⩎⩏⩁⩂⩃⩄⩅⩆⩇⩈⩉
┑┐⊤⊥∧∨⇨⇀↦
⊞⊠⊴⊵┒┓Ø
Γ
Ξᴾᵖ∎℗⎈⥤
ΛΦΩΨϕ┐
⬌⬇⬍⬆⇿⇾⇽
ㆪ↺↷↠⇝⇢∗⋆⊸⊷□○●◐◑◒◓◔◕
⇛⇚⇲⇱⇰⇶⟾⤋⤊⇈⇊⇅↕↕
ʰʱͪͭᴴᵀᵗᴾᵖ〖〗⪻⪼⩩⩨⩥⨝⨞
͓ͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͣͥͥͥ-ͣ
*)

  End U.

End Notation.


Section EntailmentLemmas.
  Import Notation.
  Open Scope forces_pred.
  Open Scope entails_pred.

  Hint Unfold  derives intersection union sep_con implies wand exactly prop.
  Hint Resolve @join_com.
  Hint Resolve @join_eq.

  Lemma derives_forces {A:Type} {as_age: ageable A}: forall (x y : pred A) w, w |= x -> x |-- y -> w |= y.
  Proof. eauto. Qed.

  Lemma derives_refl {A:Type} {as_age: ageable A}: forall x, x |-- x.
  Proof. compute; auto. Qed.

  Lemma FF_elim {A:Type} {as_age: ageable A}: forall P, FF |-- P.
  Proof. compute; intuition. Qed.

  Lemma TT_intro {A:Type} {as_age: ageable A}: forall P, P |-- TT.
  Proof. compute; intuition. Qed.

  Lemma TT_star {A:Type} {as_age: ageable A}: forall P, P |-- P * TT.
  Proof.
    repeat intro. destruct (join_ex_units a) as [x H1].
    exists a; exists x; intuition. simpl; auto.
  Qed.

  Lemma TT_gstar: forall P : pred world, (P |-- gsep_con P TT)%entails_pred.
  Proof. repeat intro. exists a gwemp; splits; auto. apply gjoin_bot1. simpl; auto. Qed.

  Lemma and_intro {A:Type} {as_age: ageable A}: forall x y z,
    x |-- y ->
    x |-- z ->
    x |-- y && z.
  Proof. split; firstorder. Qed.

  Lemma and_elim1 {A:Type} {as_age: ageable A}: forall x y z,
    x |-- y && z -> x |-- y.
  Proof. repeat intro; apply H; auto. Qed.

  Lemma and_elim3 {A:Type} {as_age: ageable A}: forall A B C,
    A |-- C -> A&&B |-- C.
  Proof. repeat intro; apply H; apply H0. Qed.

  Lemma and_elim4 {A:Type} {as_age: ageable A}: forall A B C,
    B |-- C -> A&&B |-- C.
  Proof. repeat intro; apply H; apply H0. Qed.

  Lemma and_elim2 {A:Type} {as_age: ageable A}: forall x y z,
    x |-- y && z -> x |-- z.
  Proof. repeat intro; apply H; apply H0. Qed.

  Lemma and_com_entails {A:Type} {as_age: ageable A}: forall x y,
    x && y |-- y && x.
  Proof. repeat intro; split; apply H. Qed.

  Lemma and_assoc1 {A:Type} {as_age: ageable A}: forall x y z,
    (x && (y && z)) |-- ((x && y) && z).
  Proof. repeat intro; simpl in *; intuition. Qed.

  Lemma and_assoc2 {A:Type} {as_age: ageable A}: forall x y z,
    ((x && y) && z) |-- (x && (y && z)).
  Proof. repeat intro; simpl in *; intuition. Qed.

  Lemma or_elim {A:Type} {as_age: ageable A}: forall x y z,
    x |-- z ->
    y |-- z ->
    x || y |-- z.
  Proof. repeat intro; simpl in *; firstorder. Qed.

  Lemma or_intro1 {A:Type} {as_age: ageable A}: forall x y z,
    x |-- y ->
    x |--  y || z.
  Proof. repeat intro; simpl in *; firstorder. Qed.

  Lemma or_intro2 {A:Type} {as_age: ageable A}: forall x y z,
    x |-- z ->
    x |-- y || z.
  Proof. repeat intro; simpl in *; firstorder. Qed.

  Lemma impl_intro {A:Type} {as_age: ageable A}: forall x y z,
    x && y |-- z ->
    x |-- y --> z.
  Proof. apply impl_and_adjoint. Qed.

  Lemma impl_elim {A:Type} {as_age: ageable A}: forall x y z w,
    x |-- y --> z ->
    w |-- y ->
    x && w |-- z.
  Proof. intros. apply modus_ponens with y. apply and_elim4; auto. apply and_elim3; auto. Qed.

  Lemma impl_apply {A:Type} {as_age: ageable A}: forall x y,
    (x --> y) && x |-- y.
  Proof.
    repeat intro. apply H; auto. apply H.
  Qed.

  Lemma star_assoc1 {A:Type} `{H_ASA: ASA A}: forall x y z,
    (x * (y * z)) |-- ((x * y) * z).
  Proof.
    repeat intro.
    destruct H as [a1[a23[?[??]]]].
    destruct H1 as [a2[a3[?[??]]]].
    apply join_com in H.
    apply join_com in H1.
    destruct (join_assoc H1 H) as [a21[H4 H5]]; auto.
    exists a21; exists a3. intuition.
    exists a1; exists a2; intuition.
  Qed.

  Lemma star_assoc2 {A:Type} `{H_ASA: ASA A}: forall x y z,
    ((x * y) * z) |-- (x * (y * z)).
  Proof.
    repeat intro.
    destruct H as [a12[a3[?[??]]]].
    destruct H0 as [a1[a2[?[??]]]].
    destruct (join_assoc H0 H) as [a23[H4 H5]]; auto.
    exists a1; exists a23. intuition.
    exists a2; exists a3; intuition.
  Qed.

  Lemma star_com_entails {A:Type} `{H_ASA: ASA A}: forall x y,
    x * y |-- y * x.
  Proof. repeat intro; destruct H as [a1[a2[?[??]]]]; exists a2; exists a1; intuition. Qed.

  Lemma star_unit1_entails {A:Type} `{H_ASA: ASA A}: forall x,
    x |-- emp * x.
  Proof.
    repeat intro.
    destruct (join_ex_units a) as [unit H0].
    exists unit; exists a; intuition.
    simpl; apply unit_identity with a; auto.
  Qed.

  Lemma star_unit2_entails {A:Type} `{H_ASA: ASA A}: forall x,
    x |-- x * emp.
  Proof.
    intros. eapply derives_cut.
    apply star_unit1_entails.
    apply star_com_entails.
  Qed.

  Lemma star_unit3_entails {A:Type} `{H_ASA: ASA A}: forall x,
    emp * x |-- x.
  Proof.
    repeat intro. destruct H as [a1[a2[?[??]]]].
    replace a with a2; auto.
  Qed.

  Lemma star_unit4_entails {A:Type}   `{H_ASA: ASA A}: forall x,
    x * emp |-- x.
  Proof.
    intros; eapply derives_cut.
    apply star_com_entails.
    apply star_unit3_entails.
  Qed.

  Lemma star_intro {A:Type} `{H_ASA: ASA A}: forall x y z w,
    x |-- z ->
    y |-- w ->
    x * y |-- z * w.
  Proof. repeat intros; apply split_con; auto. Qed.

  Lemma gstar_assoc1 {A:Type} `{H_ASA: ASA A}: forall x y z,
    (gsep_con x (gsep_con y z)) |-- (gsep_con (gsep_con x y) z).
  Proof.
    repeat intro.
    destruct H as [a1[a23[?[??]]]].
    destruct H1 as [a2[a3[?[??]]]].
    apply join_com in H.
    apply join_com in H1.
    destruct (join_assoc H1 H) as [a21[H4 H5]]; auto.
    exists a21; exists a3; unfold gjoin in *. intuition.
    exists a1; exists a2; unfold gjoin; intuition.
  Qed.

  Lemma gstar_assoc2 {A:Type} `{H_ASA: ASA A}: forall x y z,
    (gsep_con (gsep_con x y) z) |-- (gsep_con x (gsep_con y z)).
  Proof.
    repeat intro.
    destruct H as [a12[a3[?[??]]]].
    destruct H0 as [a1[a2[?[??]]]].
    unfold gjoin in *.
    destruct (join_assoc H0 H) as [a23[H4 H5]]; auto.
    exists a1; exists a23. intuition.
    exists a2; exists a3; intuition.
  Qed.

  Lemma gstar_com_entails {A:Type} `{H_ASA: ASA A}: forall x y,
    gsep_con x y |--gsep_con y x.
  Proof. repeat intro; destruct H as [a1[a2[?[??]]]]; exists a2; exists a1; unfold gjoin in *; intuition. Qed.

  Lemma gstar_unit1_entails {A:Type} `{H_ASA: ASA A}: forall x,
    x |-- gsep_con (exactly gwemp) x.
  Proof.
    repeat intro.
    destruct (join_ex_units a) as [unit H0].
    exists gwemp a; intuition.
    apply gjoin_bot2.
    reflexivity.
  Qed.

  Lemma gstar_unit2_entails {A:Type} `{H_ASA: ASA A}: forall x,
    x |-- gsep_con x (exactly gwemp).
  Proof.
    intros. eapply derives_cut.
    apply gstar_unit1_entails.
    apply gstar_com_entails.
  Qed.

  Lemma gstar_unit3_entails {A:Type} `{H_ASA: ASA A}: forall x,
    gsep_con (exactly gwemp) x |-- x.
  Proof.
    repeat intro. destruct H as [a1[a2[?[??]]]].
    simpl in H0; subst.
    replace a with a2; auto.
    eapply gworld_unit_identity; apply H.
  Qed.

  Lemma gstar_unit4_entails {A:Type}   `{H_ASA: ASA A}: forall x,
    gsep_con x (exactly gwemp) |-- x.
  Proof.
    intros; eapply derives_cut.
    apply gstar_com_entails.
    apply gstar_unit3_entails.
  Qed.

  Lemma split_gcon: forall P Q R S : pred world,
    P |-- Q -> R |-- S -> gsep_con P R |-- gsep_con Q S.
  Proof.
    intros. repeat intro. destruct H1 as [a1[a2[?[??]]]].
    exists a1 a2; splits; auto.
  Qed.
  Lemma gstar_intro {A:Type} `{H_ASA: ASA A}: forall x y z w,
    x |-- z ->
    y |-- w ->
    gsep_con x y |-- gsep_con z w.
  Proof. repeat intros; apply split_gcon; auto. Qed.

  Lemma wand_intro {A:Type} `{H_ASA: ASA A}: forall x y z,
    x * y |-- z ->
    x |-- y -* z.
  Proof.
  Admitted.

  Lemma wand_elim {A:Type} `{H_ASA: ASA A}: forall x y z w,
    x |-- y -* z ->
    w |-- y ->
    x * w |-- z.
  Proof.
  Admitted.

  Lemma wand_elim2 {A:Type} `{H_ASA: ASA A}: forall A B,
    A * (A -* B) |-- B.
  Proof.
  Admitted.

  Lemma wand_elim1 {A:Type} `{H_ASA: ASA A}: forall A B C, A |-- B * (B -* C) -> A |-- C.
  Proof.
  Admitted.

  Lemma star_dist1 {A:Type} `{H_ASA: ASA A}: forall A B C D,
    A |-- (B && C) * D ->
    A |-- (B * D) && (C * D).
  Proof.
    repeat intro.
    apply H in H0.
    destruct H0 as [wbc[wd[?[??]]]].
    split; exists wbc; exists wd; intuition; apply H1.
  Qed.

  Lemma gstar_dist1 {A:Type} `{H_ASA: ASA A}: forall A B C D,
    A |-- gsep_con (B && C) D ->
    A |-- (gsep_con B D) && (gsep_con C D).
  Proof.
    repeat intro.
    apply H in H0.
    destruct H0 as [wbc[wd[?[??]]]].
    split; exists wbc; exists wd; intuition; apply H1.
  Qed.

(*
  Lemma can_send {A:Type} `{H_ASA: ASA A}: forall c p e,
    pSend c p e * TT |-- pCanSend c.
  Proof.
    repeat intro.
    destruct H as [w1[w2[?[[n] ?]]]].
    exists p; exists n.
    exists w1; exists w2.
    split; auto.
    simpl in *.
    intuition.
  Qed.

  Lemma can_recv {A:Type} `{H_ASA: ASA A}: forall c p e,
    pRecv c p e * TT |-- pCanRecv c.
  Proof.
    repeat intro.
    destruct H as [w1[w2[?[[n] ?]]]].
    exists p; exists n.
    exists w1; exists w2.
    split; auto.
    simpl in *.
    intuition.
  Qed.
*)

  Lemma eval_exist_consistent {A:Type} `{H_ASA: ASA A}: forall e v1 v2 P Q,
    P |-- eval e v1 ->
    Q |-- eval e v2 ->
    P |-- Q ->
    v1=v2.
  Proof.
    admit.
  Qed.

(*
  Lemma eval_exist_consistent: forall A e v P Q (Q':A->pred world),
    P |-- eval e v ->
    Q v |-- eval e v ->
    existential Q' |-- Q v ->
    P * existential Q' |-- P * Q v.
  Proof.
    admit.
  Qed.
*)
  Lemma exists_elim {A:Type} {as_age: ageable A}: forall B (P Q: B->pred A),
    (forall v, P v |-- Q v) ->
    existential P |-- existential Q.
  Proof.
    repeat intro.
    destruct H0.
    exists x.
    apply H; auto.
  Qed.

  Lemma forall_elim {A:Type} {as_age: ageable A}: forall B (P Q: B->pred A),
    (forall v, P v |-- Q v) ->
    universal P |-- universal Q.
  Proof.
    repeat intro.
    apply H.
    apply H0.
  Qed.

Lemma exists_intro {A:Type} {as_age: ageable A}: forall {B:Type} P (Q:B->pred A),
  (exists x, P |-- Q x) ->
  P |-- existential Q.
Proof. repeat intro; simpl; firstorder. Qed.

Lemma forall_intro {A:Type} {as_age: ageable A}: forall {B:Type} P (Q:B->pred A),
  (forall x, P |-- Q x) ->
  P |-- universal Q.
Proof. repeat intro; simpl; firstorder. Qed.



(*  Definition precise A := forall B C, (equ (A * (B && C)) ((A * B) && (A * C)))%predicate.*)

End EntailmentLemmas.

Module Morphisms.
  Require Import Setoid.
  Require Export Morphisms.
  Export SetoidTactics.
  Require Export SetoidClass.
  Import Notation.
  Open Scope forces_pred.
  Open Scope entails_pred.

  Lemma derives_trans {A:Type} {as_age: ageable A}: forall x y z, x |-- y -> y |-- z -> x |-- z.
    Proof. intros; apply derives_cut with y; auto. Qed.

  Add Parametric Relation (A:Type) (as_age: ageable A): (pred A) (@derives A as_age)
    reflexivity proved by derives_refl
    transitivity proved by derives_trans
    as derives_rel.

  Add Parametric Relation : expr expr_bool_equiv
    reflexivity proved by expr_bool_equiv_refl
    symmetry proved by expr_bool_equiv_symm
    transitivity proved by expr_bool_equiv_trans
    as expr_bool_equiv_rel.

  Program Instance BoolExprSetoid : Setoid expr :=
      { equiv:= expr_bool_equiv }.

  Program Instance ExprSetoid : Setoid expr :=
      { equiv:= expr_equiv }.
  Next Obligation.
    unfold expr_equiv.
    constructor; auto.
    repeat intro.
    rewrite H. rewrite H0.
    reflexivity.
  Qed.

  Program Instance PredSetoid : forall {A:Type} {H: ageable A}, Setoid (pred A) :=
      { equiv:= (fun P Q => (P |-- Q) /\ (Q |-- P)) }.
  Solve Obligations using firstorder.

  Add Parametric Morphism A (H: ageable A): (@derives A H)
  with signature equiv ==> equiv ==> equiv
  as derives_m.
    firstorder.
  Qed.

  Add Parametric Morphism A (H: ageable A): (@app_pred A H)
  with signature equiv ==> eq ==> equiv
  as forces_m.
    firstorder.
  Qed.

  Hint Unfold  derives intersection union sep_con implies wand exactly prop.
  Hint Resolve @join_com.
  Hint Resolve @join_eq.


  Add Parametric Morphism: evalTrue
  with signature expr_bool_equiv ==> eq ==> equiv
  as evalTrue_m.
    intros e1 e2 ? w.
    split; intros; apply H; auto.
  Qed.
  Add Parametric Morphism: pIsTrue
  with signature expr_bool_equiv ==> equiv
  as pIsTrue_m.
    intros.
    split; intros.
    intros w; apply evalTrue_m; symmetry; auto.
    intros w; apply evalTrue_m; auto.
  Qed.
  Add Parametric Morphism: evalFalse
  with signature expr_bool_equiv ==> eq ==> equiv
  as evalFalse_m.
    intros e1 e2 ? w.
    split; intros; apply H; auto.
  Qed.
  Add Parametric Morphism: pIsFalse
  with signature expr_bool_equiv ==> equiv
  as pIsFalse_m.
    split; intros.
    intro w; apply evalFalse_m; symmetry; auto.
    intro w; apply evalFalse_m; auto.
  Qed.

  Add Parametric Morphism: eval
  with signature (@equiv expr ExprSetoid) ==> eq ==> equiv
  as eval_m.
    split; intros.
    intros w ?. simpl in *; rewrite<- H; auto.
    intros w ?. simpl in *; rewrite H; auto.
  Qed.

  Add Parametric Morphism : expr_bool_equiv
  with signature (@equiv expr ExprSetoid) ==> (@equiv expr ExprSetoid) ==> iff
  as expr_bool_m.
    intros. unfold expr_bool_equiv.
    unfold evalFalse,evalTrue in *.
    setoid_rewrite H.
    setoid_rewrite H0.
    intuition.
  Qed.

  Program Instance ExprFunSetoid : Setoid (value->expr) :=
      { equiv:= fun e1 e2 => forall v, e1 v == e2 v }.
  Next Obligation.
    constructor.
    constructor; auto.
    repeat intro.
    symmetry; apply H.
    repeat intro.
    rewrite H.
    rewrite H0.
    reflexivity.
  Qed.

  Add Parametric Morphism: evalTrue
  with signature equiv ==> eq ==> equiv
  as evalTrue_m'.
    intros e1 e2 ? w.
    apply evalTrue_m; auto.
    setoid_rewrite H.
    reflexivity.
  Qed.
  Add Parametric Morphism: pIsTrue
  with signature equiv ==> equiv
  as pIsTrue_m'.
    intros.
    apply pIsTrue_m; auto.
    setoid_rewrite H.
    reflexivity.
  Qed.
  Add Parametric Morphism: evalFalse
  with signature equiv ==> eq ==> equiv
  as evalFalse_m'.
    intros e1 e2 ? w.
    apply evalFalse_m; auto.
    setoid_rewrite H.
    reflexivity.
  Qed.
  Add Parametric Morphism: pIsFalse
  with signature equiv ==> equiv
  as pIsFalse_m'.
    intros.
    apply pIsFalse_m; auto.
    setoid_rewrite H.
    reflexivity.
  Qed.

  Add Parametric Morphism: expr_fun_app
  with signature equiv ==> equiv ==> equiv
  as expr_fun_m.
    intros.
    apply expr_fun_equiv; auto.
  Qed.

  Add Parametric Morphism: expr_subst
  with signature equiv ==> eq ==> equiv ==> equiv
  as expr_subst_m.
    intros.
    apply expr_subst_equiv; auto.
  Qed.

  Add Parametric Morphism: expr_plus
  with signature equiv ==> equiv ==> equiv
  as expr_plus_m.
    intros.
    unfold expr_plus.
    apply expr_fun2_equiv; intros; auto.
    apply ExprSetoid.
  Qed.

  (* * *)
  Add Parametric Morphism A (as_alg : sepalg A) (as_age : ageable A) (H: ASA A): (@sep_con A as_alg as_age H)
  with signature equiv ==> equiv ==> equiv
  as sep_con_m.
    split; repeat intro; destruct H2 as [a1[a2[?[??]]]];
      exists a1; exists a2;
      split; auto;
      (split; [apply H0; auto | apply H1; auto]).
  Qed.

  (* (x) *)
  Add Parametric Morphism: gsep_con
  with signature equiv ==> equiv ==> equiv
  as gsep_con_m.
    split; repeat intro; destruct H1 as [a1[a2[?[??]]]];
      exists a1; exists a2;
      split; auto;
      (split; [apply H; auto | apply H0; auto]).
  Qed.

  (* && *)
  Add Parametric Morphism A (as_age : ageable A): (@intersection A as_age)
  with signature equiv ==> equiv ==> equiv
  as intersection_m.
    split; repeat intro; destruct H1;
      (split; [apply H; auto | apply H0; auto]).
  Qed.

  (* || *)
  Add Parametric Morphism A (as_age : ageable A): (@union A as_age)
  with signature equiv ==> equiv ==> equiv
  as union_m.
    intros; destruct H; destruct H0; split. 
    apply or_elim.
    apply or_intro1; auto.
    apply or_intro2; auto.
    apply or_elim.
    apply or_intro1; auto.
    apply or_intro2; auto.
  Qed.

  (* --> *)
  Add Parametric Morphism A (as_age : ageable A): (@implies A as_age)
  with signature equiv ==> equiv ==> equiv
  as implies_m.
    split; intros.
    apply impl_intro.
    apply impl_elim with x.
    apply impl_intro.
    eapply derives_cut.
    apply impl_apply.
    apply H0.
    apply H.
    apply impl_intro.
    apply impl_elim with y.
    apply impl_intro.
    eapply derives_cut.
    apply impl_apply.
    apply H0.
    apply H.
  Qed.

  (* -* *)
  Add Parametric Morphism A (as_alg : sepalg A) (as_age : ageable A) (H: ASA A): (@wand A as_alg as_age H)
  with signature equiv ==> equiv ==> equiv
  as wand_m.
    split; intros; apply wand_intro.
    setoid_rewrite<- H0.
    apply derives_cut with x0.
    apply con_cut with x; auto.
    apply H1.
    setoid_rewrite H0.
    apply derives_cut with y0.
    apply con_cut with y; auto.
    apply H1.
  Qed.


  Program Instance FunPredSetoid : forall {A:Type} {H: ageable A} {B:Type}, Setoid (B -> pred A) :=
      { equiv:= (fun P Q => forall (b:B), P b == Q b) }.
  Solve Obligations using firstorder.

  Definition fun_equiv {A:Type} {H: ageable A} {B:Type} := (fun (P Q: B->@pred A H) => (fun (b:B) => P b) == (fun (b:B) => Q b)).

  (* exists *)
  Program Instance exists_m {A : Type} (as_age : ageable A) B :
    Proper (pointwise_relation B equiv ==> equiv)%signature (@existential A as_age B).
  Next Obligation.
    split; intros; apply exists_elim; repeat intro; apply H; auto.
  Qed.

  (* forall *)
  Program Instance forall_m {A : Type} (as_age : ageable A) B :
    Proper (pointwise_relation B equiv ==> equiv)%signature (@universal A as_age B).
  Next Obligation.
    split; intros; apply forall_elim; repeat intro; apply H; auto.
  Qed.

  (* !! *)
  Add Parametric Morphism A (as_age : ageable A): (@prop A as_age)
  with signature equiv ==> equiv
  as prop_m.
    split; repeat intro; simpl in *; apply H; auto.
  Qed.

  (* [], |>, etc. *)
  Add Parametric Morphism A (as_age : ageable A): (@box A as_age)
  with signature eq ==> equiv ==> equiv
  as box_m.
    split; repeat intro; simpl in *; apply H; auto.
  Qed.

  Add Parametric Morphism A (as_age : ageable A): (@diamond A as_age)
  with signature eq ==> equiv ==> equiv
  as diamond_m.
    split; repeat intro; simpl in *; destruct H0 as [a'[??]]; exists a'; intuition.
  Qed.

  Module Tests.
    Import Notation.
    Import Notation.U.
    Open Scope pred.

    Lemma test_exists_morphism: forall (A:Type) (as_age:ageable A) (B: Type) (C D: B->pred A),
      (forall x, C x == D x) -> (Ex_ x:B, C x == Ex_ y:B, D y).
    Proof. intros; setoid_rewrite H; reflexivity. Qed.
    Lemma test_exists_morphism2: forall (A:Type) (as_age:ageable A) (B: Type) (C: B->pred A) (D E:pred A),
      D==E -> (Ex_ x:B, (C x && D) == Ex_ y:B, (C y && E)).
    Proof. intros. setoid_rewrite H; reflexivity. Qed.
    Lemma test_forall_morphism: forall (A:Type) (as_age:ageable A) (B: Type) (C D: B->pred A),
      (forall x, C x == D x) -> (All_ x:B, C x == All_ y:B, D y).
    Proof. intros. setoid_rewrite H. reflexivity. Qed.
    Lemma test_forall_morphism2: forall (A:Type) (as_age:ageable A) (B: Type) (C: B->pred A) (D E:pred A),
      D==E -> (All_ x:B, (C x && D) == All_ y:B, (C y && E)).
    Proof. intros. setoid_rewrite H; reflexivity. Qed.
    Lemma test_sub_morphism: forall A B C D,
      A == B -> C == D -> (A >=> C == B >=> D).
    Proof. intros; rewrite  H; rewrite H0; reflexivity. Qed.
    Lemma test_sub_equ_morphism: forall A B C D,
      A == B -> C == D -> (A <=> C == B <=> D).
    Proof. intros; rewrite  H; rewrite H0; reflexivity. Qed.
    Lemma test_later_morphism: forall A B,
      A == B -> (|> A == |> B).
    Proof. intros; rewrite  H; reflexivity. Qed.
    Lemma test_nec_morphism: forall A B,
      A == B -> ([] A == [] B).
    Proof. intros; rewrite  H; reflexivity. Qed.
    Lemma test_extend_morphism: forall A B,
      A == B -> (% A == % B).
    Proof. intros; rewrite  H; reflexivity. Qed.
    Lemma test_prop_morphism: forall A B,
      A == B -> (!! A == !! B).
    Proof. intros; rewrite  H; reflexivity. Qed.
    Lemma test_derives_morphism: forall A B,
      A == B -> ( A|-- B).
    Proof. intros; rewrite  H; auto. Qed.
    Lemma test_forces_morphism: forall A B a,
      A == B -> a |= A -> a |= B.
    Proof. intros; rewrite<-  H; auto. Qed.
    Lemma test_and_morphism: forall A B C D,
      A == B -> C == D -> (A && C == B && D).
    Proof. intros; rewrite H; rewrite H0; reflexivity. Qed.
    Lemma test_or_morphism: forall A B C D,
      A == B -> C == D -> (A || C == B || D).
    Proof. intros; rewrite H; rewrite H0; reflexivity. Qed.
    Lemma test_star_morphism: forall A B C D,
      A == B -> C == D -> (A * C == B * D).
    Proof. intros; rewrite H; rewrite H0; reflexivity. Qed.
    Lemma test_wand_morphism: forall A B C D,
      A == B -> C == D -> (A -* C == B -* D).
    Proof. intros; rewrite H; rewrite H0; reflexivity. Qed.
    Lemma test_impl_morphism: forall A B C D,
      A == B -> C == D -> (A --> C == B --> D).
    Proof. intros; rewrite H; rewrite H0; reflexivity. Qed.
 
  End Tests.

End Morphisms.


Hint Unfold derives : seplog_hints.
Hint Resolve @derives_refl @derives_cut @and_intro @and_elim1 : seplog_hints.
Hint Resolve @and_elim3 @and_elim4 @and_elim2 @or_elim @or_intro1 : seplog_hints.
Hint Resolve @or_intro2 @impl_intro @impl_elim @star_assoc1 @gstar_assoc1  : seplog_hints.
Hint Resolve @star_assoc2 @star_com_entails @star_unit1_entails @star_unit2_entails  : seplog_hints.
Hint Resolve @gstar_assoc2 @gstar_com_entails @gstar_unit1_entails @gstar_unit2_entails @gstar_intro : seplog_hints.
Hint Resolve @star_intro @wand_intro @wand_elim @wand_elim2  : seplog_hints.
Hint Resolve @wand_elim1 @star_dist1 @star_unit3_entails @star_unit4_entails : seplog_hints.
Hint Resolve @gstar_dist1 @gstar_unit3_entails @gstar_unit4_entails : seplog_hints.

Section ExpressionLemmas.
  Import Morphisms.

  Lemma expr_fun_norm_inv: forall f (e:expr) l,
    (exists v, e l = Some v /\ (expr_fun_app f e) l = (f v) l)
    \/ (e l = None /\ (expr_fun_app f e) l = None).
  Proof.
    intros. simpl.
    destruct (e l); intros; eauto.
  Qed.

  Lemma expr_plus_norm: forall (e1 e2: expr) l,
    expr_plus e1 e2 l = value_plus (e1 l) (e2 l).
  Proof.
    intros.
    simpl.
    destruct (e1 l) as [v1 | ].
    destruct (e2 l) as [v2 | ].
    destruct v1; destruct v2; reflexivity.
    destruct v1; reflexivity.
    reflexivity.
  Qed.

  Lemma value_plus_com: forall v1 v2,
    value_plus v1 v2 = value_plus v2 v1.
  Proof.
    intros.
    destruct v1 as [[]|]; destruct v2 as [[]|]; simpl; do 2 f_equal; try omega.
    f_equal; omega.
    f_equal; omega.
  Qed.

  Lemma value_plus_assoc: forall v1 v2 v3,
    value_plus v1 (value_plus v2 v3) = value_plus (value_plus v1 v2) v3.
  Proof.
    intros.
    destruct v1 as [[]|]; destruct v2 as [[]|]; destruct v3 as [[]|]; simpl; do 2 f_equal; try omega.
    generalize (Pos2Nat.is_pos c); intro.
    rewrite Nat2Pos.id; f_equal; omega.
    generalize (Pos2Nat.is_pos c); intro.
    rewrite Nat2Pos.id.
    rewrite Nat2Pos.id; f_equal; omega.
    omega.
    generalize (Pos2Nat.is_pos c); intro.
    rewrite Nat2Pos.id; f_equal; omega.
  Qed.

  Lemma expr_plus_com: forall e1 e2,
    expr_plus e1 e2 == expr_plus e2 e1.
  Proof.
    intros.
    intro l.
    do 2 rewrite expr_plus_norm.
    apply value_plus_com.
  Qed.

  Lemma expr_plus_assoc: forall e1 e2 e3,
    expr_plus e1 (expr_plus e2 e3) == expr_plus (expr_plus e1 e2) e3.
  Proof.
    intros.
    intro l.
    do 4 rewrite expr_plus_norm.
    apply value_plus_assoc.
  Qed.

  Lemma value_plus_zero1: forall v1,
    value_plus v1 (Some (nat_value 0)) = v1.
  Proof.
    intros.
    destruct v1 as [[]|]; simpl; try (do 2 f_equal; try omega).
    rewrite plus_0_r.
    apply Pos2Nat.id.
  Qed.

  Lemma value_plus_zero2: forall v2,
    value_plus (Some (nat_value 0)) v2 = v2.
  Proof.
    intros.
    destruct v2 as [[]|]; simpl; try (do 2 f_equal; omega).
    do 2 f_equal.
    apply Pos2Nat.id.
  Qed.

  Lemma expr_plus_zero1: forall e1,
    expr_plus e1 (expr_value (nat_value 0)) == e1.
  Proof.
    intros.
    intro l.
    repeat rewrite expr_plus_norm.
    apply value_plus_zero1.
  Qed.

  Lemma expr_plus_zero2: forall e2,
    expr_plus (expr_value (nat_value 0)) e2 == e2.
  Proof.
    intros.
    setoid_rewrite expr_plus_com.
    apply expr_plus_zero1.
  Qed.

  Lemma expr_plus_const: forall v1 v2,
    expr_plus (expr_value' v1) (expr_value' v2)  == expr_value' (value_plus v1 v2).
  Proof.
    intros.
    intro l.
    setoid_rewrite expr_plus_norm. reflexivity. 
  Qed.

  Lemma not_free_in_expr_expr_fun_app: forall f e x,
    not_free_in_expr x e ->
    (forall v, not_free_in_expr x (f v)) ->
    not_free_in_expr x (expr_fun_app f e).
  Proof.
    unfold not_free_in_expr.
    intros.
    simpl in *.
    rewrite<- H.
    destruct (e l); auto.
  Qed.

  Lemma w_env_share_set: forall w x y v,
    w_env_share (w_env_set w y v) x = w_env_share w x.
  Proof.
    intros.
    unfold w_env_set, w_env_share.
    remember (LocalsMap.map_upd y v (w_env w)).
    destruct o; auto.
    rewrite w_env_wupd_env.
    symmetry.
    eapply LocalsMap.map_set_share3; eauto.
  Qed.

  Lemma w_env_set_diff: forall w x y v1 v2,
    x<>y ->
    w_env_set (w_env_set w x v1) y v2 = w_env_set (w_env_set w y v2) x v1.
  Proof.
    intros.
    unfold w_env_set.
    remember (LocalsMap.map_upd x v1 (w_env w)).
    destruct o.
    rewrite w_env_wupd_env.
    remember (LocalsMap.map_upd y v2 (w_env w)).
    destruct o.
    rewrite w_env_wupd_env.
    edestruct LocalsMap.map_upd_success with (a:=x)(v:=v1) (m:=m0); eauto.
    erewrite<- LocalsMap.map_set_share3; eauto.
    erewrite LocalsMap.map_set_share1; eauto.
    edestruct LocalsMap.map_upd_success with (a:=y)(v:=v2) (m:=m); eauto.
    erewrite<- LocalsMap.map_set_share3; eauto.
    erewrite LocalsMap.map_set_share1; eauto.
    rewrite H0.
    rewrite H1.
    do 2 rewrite wupd_env2.
    f_equal.
    apply locals_equiv_eq.
    intro z.
    destruct (Pos.eq_dec z x).
    subst z.
    rewrite<- (LocalsMap.map_gso_val x y v2 m x1); eauto.
    erewrite (LocalsMap.map_gss_val x v1 m0 x0); eauto.
    erewrite (LocalsMap.map_gss_val x v1 (w_env w) m); eauto.
    destruct (Pos.eq_dec z y).
    subst z.
    erewrite (LocalsMap.map_gss_val y v2 m x1); eauto.
    erewrite<- (LocalsMap.map_gss_val y v2 (w_env w) m0); eauto.
    eapply LocalsMap.map_gso_val; eauto.
    rewrite<- (LocalsMap.map_gso_val z y v2 m x1); eauto.
    rewrite<- (LocalsMap.map_gso_val z x v1 (w_env w) m); eauto.
    rewrite<- (LocalsMap.map_gso_val z x v1 m0 x0); eauto.
    rewrite<- (LocalsMap.map_gso_val z y v2 (w_env w) m0); eauto.

    case_eq (LocalsMap.map_upd y v2 m); intros.
    false.
    destruct (LocalsMap.map_upd_success y v2 (w_env w)); auto.
    erewrite LocalsMap.map_gso_share with (j:=x); auto.
    eapply LocalsMap.map_set_share1; eauto.
    symmetry; eauto.
    rewrite<- Heqo0 in H1.
    discriminate.
    destruct (LocalsMap.map_upd_success x v1 (w_env w)); auto.
    eapply LocalsMap.map_set_share1; eauto.
    rewrite H1.
    rewrite<- Heqo in H1.
    inverts H1.
    reflexivity.

    case_eq (LocalsMap.map_upd y v2 (w_env w)); intros.
    rewrite w_env_wupd_env.
    case_eq (LocalsMap.map_upd x v1 m); intros; auto.
    false.
    destruct (LocalsMap.map_upd_success x v1 (w_env w)); auto.
    erewrite LocalsMap.map_gso_share with (j:=y); auto.
    eapply LocalsMap.map_set_share1; eauto.
    symmetry; eauto.
    rewrite<- Heqo in H2.
    discriminate.
    rewrite<- Heqo; auto.
  Qed.

  Lemma locals_map_upd_diff: forall l x y vx vy,
    x <> y ->
    forall z, locals_map_upd (locals_map_upd l x vx) y vy z = locals_map_upd (locals_map_upd l y vy) x vx z.
  Proof.
    intros. unfold locals_map_upd.
    destruct (Pos.eq_dec z y); auto.
    destruct (Pos.eq_dec z x); auto.
    false.
  Qed.

  Lemma locals_map_upd_same: forall l x v v',
    forall z, locals_map_upd (locals_map_upd l x v) x v' z = locals_map_upd l x v' z.
  Proof.
    intros. unfold locals_map_upd.
    destruct (Pos.eq_dec z x); auto.
  Qed.

  Lemma expr_locals_equiv: forall (e:expr) l1 l2,
    (forall z, l1 z = l2 z) ->
    e l1 = e l2.
  Proof.
    intros.
    destruct e as [fe H0].
    simpl.
    case_eq (fe l1); intros.
    symmetry; eapply (H0 l1); eauto.
    intros. rewrite<- H2. symmetry; apply H.
    case_eq (fe l2); intros.
    apply H0 with (l':=l1) in H2.
    false.
    intros. rewrite<- H3. apply H.
    auto.
  Qed.

  Lemma not_free_in_expr_expr_subst1: forall e x y ex,
    not_free_in_expr x e ->
    not_free_in_expr x ex ->
    not_free_in_expr x (expr_subst e y ex).
  Proof.
    intros.
    repeat intro.
    simpl.
    rewrite<- H0.
    destruct (Pos.eq_dec y x).
    subst y.
    apply expr_locals_equiv.
    symmetry; apply locals_map_upd_same.
    rewrite H with (v:=v).
    apply expr_locals_equiv.
    apply locals_map_upd_diff; auto.
  Qed.

  Lemma not_free_in_expr_expr_value: forall x v,
    not_free_in_expr x (expr_value v).
  Proof.
    intros.
    unfold not_free_in_expr; intros. reflexivity.
  Qed.

  Lemma not_free_in_expr_expr_value': forall x v,
    not_free_in_expr x (expr_value' v).
  Proof.
    intros.
    unfold not_free_in_expr; intros. reflexivity.
  Qed.

  Lemma not_free_in_expr_expr_plus: forall x e1 e2,
    not_free_in_expr x e1 ->
    not_free_in_expr x e2 ->
    not_free_in_expr x (expr_plus e1 e2).
  Proof.
    intros.
    apply not_free_in_expr_expr_fun_app; intros; auto.
    apply not_free_in_expr_expr_fun_app; intros; auto.
    apply not_free_in_expr_expr_value'.
  Qed.

  Lemma w_env_set_not_top: forall w x v,
    w_env_share w x <> Share.top ->
    w_env_set w x v = w.
  Proof.
    intros.
    unfold w_env_set.
    case_eq (LocalsMap.map_upd x v (w_env w)); intros; auto.
    false.
    apply H.
    eapply LocalsMap.map_set_share1; eauto.
  Qed.

  Lemma w_env_get_not_bot: forall w x,
    w_env_share w x <> Share.bot ->
    exists v, w_env_get w x = Some v.
  Proof.
    intros.
    unfold w_env_get.
    case_eq (LocalsMap.map_val x (w_env w)); intros; eauto.
    false. apply H.
    apply LocalsMap.map_val_bot; auto.
  Qed.

  Lemma expr_subst_norm: forall e x ex l,
    expr_subst e x ex l = e (locals_map_upd l x (ex l)).
  Proof.
    intros. auto.
  Qed.

  Lemma expr_subst_plus: forall e1 e2 x ex,
    expr_subst (expr_plus e1 e2) x ex == expr_plus (expr_subst e1 x ex) (expr_subst e2 x ex).
  Proof.
    intros.
    intro l.
    rewrite expr_subst_norm.
    repeat rewrite expr_plus_norm.
    repeat rewrite expr_subst_norm.
    reflexivity.
  Qed.

  Lemma expr_subst_not_free: forall x (e ex: expr),
    not_free_in_expr x e ->
    expr_subst e x ex == e.
  Proof.
    intros. intro.
    rewrite expr_subst_norm.
    symmetry; apply H.
  Qed.

  Lemma locals_map_upd_get_diff: forall x y l v,
    y <> x ->
    (locals_map_upd l x v) y = l y.
  Proof.
    intros. unfold locals_map_upd.
    destruct Pos.eq_dec; eauto. false.
  Qed.

  Lemma locals_map_upd_get_same: forall x l v,
    (locals_map_upd l x v) x = v.
  Proof.
    intros. unfold locals_map_upd.
    destruct Pos.eq_dec; eauto. false.
  Qed.

  Lemma not_free_in_expr_var: forall x y,
    x <> y ->
    not_free_in_expr x (expr_var y).
  Proof.
    intros.
    repeat intro.
    simpl.
    rewrite locals_map_upd_get_diff; auto.
  Qed.

  Lemma expr_subst_id: forall x ex,
    expr_subst (expr_var x) x ex == ex.
  Proof.
    intros. intro l.
    rewrite expr_subst_norm.
    simpl.
    apply locals_map_upd_get_same.
  Qed.

  Lemma locals_to_map_upd: forall (e:expr) x vx w,
    w_env_share w x = Share.top ->
    e (locals_to_map (w_env (w_env_set w x vx))) = e (locals_map_upd (locals_to_map (w_env w)) x (Some vx)).
  Proof.
    simpl; intros.
    apply expr_locals_equiv. intro z.
    unfold locals_map_upd,locals_to_map,w_env_set.
    case_eq (LocalsMap.map_upd x vx (w_env w)); intros.
    rewrite w_env_wupd_env.
    destruct (Pos.eq_dec z x).
    subst z.
    eapply LocalsMap.map_gss_val; eauto.
    symmetry.
    eapply LocalsMap.map_gso_val; eauto.
    destruct (LocalsMap.map_upd_success x vx (w_env w)) as [m' ?]; eauto.
    false.
  Qed.

  Lemma eval_eval: forall e v w,
    eval e v w <-> e (locals_to_map (w_env w)) = Some v.
  Proof. reflexivity. Qed.

  Lemma eval_expr_subst1: forall e x ex v vx w,
    eval ex vx w ->
    w_env_share w x = Share.top ->
    eval e v (w_env_set w x vx) ->
    eval (expr_subst e x ex) v w.
  Proof.
    intros.
    apply eval_eval. apply eval_eval in H. apply eval_eval in H1.
    rewrite expr_subst_norm; eauto.
    erewrite locals_to_map_upd in H1; eauto.
    rewrite H.
    auto.
  Qed.

  Lemma eval_expr_subst2: forall e x ex v vx w,
    eval ex vx w ->
    w_env_share w x = Share.top ->
    eval (expr_subst e x ex) v w ->
    eval e v (w_env_set w x vx).
  Proof.
    intros.
    apply eval_eval. apply eval_eval in H. apply eval_eval in H1.
    erewrite expr_subst_norm in H1; eauto.
    erewrite locals_to_map_upd; eauto.
    rewrite<- H; auto.
  Qed.

  Lemma eval_not_free_in_expr1: forall e v w x vx,
    not_free_in_expr x e ->
    eval e v w -> eval e v (w_env_set w x vx).
  Proof.
    intros.
    apply eval_eval. apply eval_eval in H0.
    destruct (Share.eq_dec (w_env_share w x) Share.top).
    erewrite locals_to_map_upd; auto.
    rewrite<- H; auto.
    rewrite w_env_set_not_top; auto.
  Qed.

  Lemma eval_not_free_in_expr2: forall e v w x vx,
    not_free_in_expr x e ->
    eval e v (w_env_set w x vx) -> eval e v w.
  Proof.
    intros.
    apply eval_eval. apply eval_eval in H0.
    destruct (Share.eq_dec (w_env_share w x) Share.top).
    erewrite locals_to_map_upd in H0; auto.
    rewrite<- H in H0; auto.
    rewrite w_env_set_not_top in H0; auto.
  Qed.

  Lemma w_env_w_heap_set: forall w a v,
    w_env (w_heap_set w a v) = w_env w.
  Proof.
    unfold w_heap_set; intros.
    destruct (MemMap.map_upd); auto.
    rewrite w_env_wupd_heap; auto.
  Qed.

  Lemma not_free_in_expr_nat: forall x n,
    not_free_in_expr x (nat_to_expr n).
  Proof.
    repeat intro. reflexivity.
  Qed.


End ExpressionLemmas.

Section MorphismLemmas.
  Import Notation.
  Open Scope forces_pred.
  Open Scope entails_pred.
  Open Scope pred.
  Import Morphisms.

  Lemma derives_equiv: forall {A:Type} {H:ageable A} (P Q: pred A),
    P |-- Q -> Q |-- P -> P == Q.
  Proof. firstorder. Qed.

  Hint Resolve @derives_equiv : seplog_hints.

  Open Scope pred.
  Lemma box_and {A:Type} {as_age: ageable A}: forall R (P Q:pred A),
    box R (P && Q) == box R P && box R Q.
  Proof.
    split; repeat intro; unfold intersection, box in *; simpl in *; firstorder.
  Qed.

  Lemma box_all {A:Type}  {as_age: ageable A} : forall B R (F:B -> pred A),
    box R (universal F) == All_ x:B, box R (F x).
  Proof.
    split; repeat intro; unfold universal, box in *; simpl in *; firstorder.
  Qed.

  Lemma diamond_or {A:Type}  {as_age: ageable A} : forall R (P Q:pred A),
    diamond R (P || Q) == diamond R P || diamond R Q.
  Proof.
    split; repeat intro; unfold diamond, union in *; simpl in *; firstorder.
  Qed.

  Lemma diamond_ex {A:Type}  {as_age: ageable A} : forall B R (F:B -> pred A),
    diamond R (existential F) == Ex_ x:B, diamond R (F x).
  Proof.
    split; repeat intro; unfold diamond, existential in *; simpl in *; firstorder.
  Qed.

  Lemma nec_useless {A:Type}  {as_age: ageable A} : forall P, []P == P.
  Proof.
    split; repeat intro.
    apply H.
    simpl; apply necM_refl.
    apply pred_nec_hereditary with a; auto.
  Qed.

  Lemma later_age {A:Type}  {as_age: ageable A} : forall P,
    |>P == box ageM P.
  Proof.
    split; repeat intro.
    simpl in H.
    apply H.
    apply t_step; auto.
    revert H; induction H0; intros.
    apply H0; auto.
    apply pred_nec_hereditary with y.
    apply Rt_Rft; auto.
    apply IHclos_trans1; auto.
  Qed.

  Lemma later_commute {A:Type}  {as_age: ageable A} : forall M (P:pred A),
    box M (|>P) == |>(box M P).
  Proof.
    split; repeat intro.
    destruct M as [R HR].
    destruct (valid_rel_commut_later2 R HR _ _ H1 _ H0).
    apply H with x; simpl; auto.
    destruct M as [R HR].
    destruct (valid_rel_commut_later1 R HR _ _ H1 _ H0).
    apply H with x; auto.
  Qed.

  Lemma later_and {A:Type}  {as_age: ageable A} : forall P Q,
    |>(P && Q) == |>P && |> Q.
  Proof.
    intros; apply box_and.
  Qed.

  Lemma later_or {A:Type}  {as_age: ageable A} : forall (P Q:pred A),
    |>(P || Q) == |>P || |>Q.
  Proof.
    setoid_rewrite later_age.
    split; repeat intro.
    simpl in H.
    case_eq (age1 a); intros.
    destruct (H a0); auto.
    left; simpl; intros; replace a' with a0; auto; congruence.
    right; simpl; intros; replace a' with a0; auto; congruence.
    left; simpl; unfold age; intros; rewrite H0 in H1; discriminate.
    destruct H.
    left; apply H. auto.
    right; apply H. auto.
  Qed.

  Lemma later_ex {A:Type} {as_age: ageable A} : forall B (F:B->pred A),
    B -> |>(existential F) == Ex_ x:B, |>(F x).
  Proof.
    intros.
    rewrite later_age.
    split; repeat intro.
    case_eq (age1 a); intros.
    destruct (H a0); auto.
    exists x.
    apply later_age; simpl; intros.
    replace a' with a0; auto; congruence.
    exists X.
    apply later_age.
    hnf; simpl; intros.
    unfold age in H1.
    rewrite H0 in H1; discriminate.
    eapply box_ex; eauto. simpl in *.
    unfold laterR; constructor; auto.
  Qed.

  Lemma later_impl {A:Type}  {as_age: ageable A} : forall P Q,
    |>(P --> Q) == |>P --> |>Q.
  Proof.
    intros; repeat rewrite later_age.
    split; intros.
    apply axiomK.
    repeat intro.
    simpl in H.
    destruct valid_rel_nec.
    destruct (H4 _ _ H1 _ H0).
    apply H with x; auto.
    intros; replace a'1 with a'0; auto; congruence.
  Qed.

  Lemma TT_and {A:Type} {as_age: ageable A}: forall P,
    TT && P == P.
  Proof. split; auto with seplog_hints; apply and_intro; auto; apply TT_intro. Qed.

  Lemma ex_and : forall {A:Type} {as_age: ageable A} B (P:B->pred A) Q,
    (existential P) && Q == Ex_ x:B, P x && Q.
  Proof.
    split; repeat intro.
    destruct H as [[x]?].
    exists x. split; auto.
    repeat intro.
    destruct H as [b[??]].
    split; auto. exists b; auto.
  Qed.

  Lemma FF_and : forall {A:Type} {as_age: ageable A} (P:pred A),
    FF && P == FF.
  Proof. split; auto with seplog_hints; apply FF_elim. Qed.

  Lemma star_assoc {A:Type} `{H_ASA: ASA A}: forall (x y z: pred A),
    (x * (y * z)) == ((x * y) * z).
  Proof. auto with seplog_hints. Qed.

  Lemma gstar_assoc: forall (x y z: pred world),
    (gsep_con x (gsep_con y z)) == (gsep_con (gsep_con x y) z).
  Proof. intros. split; intros. apply gstar_assoc1. apply gstar_assoc2. Qed.

  Lemma gstar_com: forall (x y: pred world),  gsep_con x y == gsep_con y x.
  Proof. split; intros; apply gstar_com_entails. Qed.

  Lemma star_com {A:Type} `{H_ASA: ASA A}: forall (x y: pred A), x * y == y * x.
  Proof. auto with seplog_hints. Qed.

  Lemma gstar_unit1 {A:Type} `{H_ASA: ASA A}: forall x, gsep_con (exactly gwemp) x == x.
  Proof. eauto with seplog_hints. Qed.

  Lemma gstar_unit2 {A:Type} `{H_ASA: ASA A}: forall x, gsep_con x (exactly gwemp) == x.
  Proof. eauto with seplog_hints. Qed.

  Lemma star_unit1 {A:Type} `{H_ASA: ASA A}: forall x, emp * x == x.
  Proof. auto with seplog_hints. Qed.

  Lemma star_unit2 {A:Type} `{H_ASA: ASA A}: forall x, x * emp == x.
  Proof. auto with seplog_hints. Qed.

  Lemma and_com {A:Type} {as_age: ageable A}: forall (x y: pred A), x && y == y && x.
  Proof. auto with seplog_hints. Qed.

  Lemma and_assoc {A:Type} {as_age: ageable A}: forall (x y z: pred A), x && (y && z) == ((x && y) && z).
  Proof. repeat intro; simpl in *; intuition. Qed.

  Lemma or_assoc {A:Type} {as_age: ageable A}: forall (x y z: pred A), x || (y || z) == ((x || y) || z).
  Proof. repeat intro; simpl in *; intuition. Qed.

End MorphismLemmas.

Hint Rewrite @star_assoc @star_com @star_unit1 @star_unit2 @and_assoc @or_assoc @and_com: seplog_hints.



Section Lemmas.
  Import Notation.
  Import Morphisms.
  Open Scope forces_pred.
  Open Scope entails_pred.
  Open Scope pred.

(*
  Lemma disjoint_recv_send: forall c1 p1 n1 c2 p2 n2,
    disjoint (pRecv c1 p1 n1) (pSend c2 p2 n2).
  Admitted.
*)

  Definition precise B {A:Type} `{H_ASA: ASA A}:= forall C D, B * (C && D) == (B * C) && (B * D).

(*
  Lemma new_channel_equiv: forall c l,
    pId (new_channel c l) == pSend c Share.top 0 * pRecv c Share.top 0.
  Proof.
  Admitted.
*)

  Lemma eval_exist_consistent_nat: forall e n P Q,
    P |-- eval e (nat_value n) ->
    (forall n, Q n |-- eval e (nat_value n)) ->
    P * existential Q |-- P * Q n.
  Proof.
    repeat intro.
    destruct H1 as [a1[a2[?[??]]]].
    exists a1; exists a2; intuition.
    destruct H3 as [n'].
    replace n with n'; auto.
    apply H in H2.
    apply H0 in H3.
    destruct e. simpl in H2; simpl in H3.
    destruct H1 as [_ [? _]].
    simpl in H1. destruct H1. rewrite H1 in *; rewrite H4 in *.
    rewrite H2 in H3. inversion H3; auto.
  Qed.

(*
  Lemma TotalSendRecvElim: forall c e,
    pSend c Share.top e * pRecv c Share.top e |-- pTotalSendRecv c.
  Proof.
    unfold pSend, pRecv, pTotalSendRecv.
    intros.
    rewrite ex_con.
    apply exists_elim; intro n.
    remember (eval e (nat_value n) && !!(~ identity Share.top) &&
      prod_endpoint c Share.top n)%pred as P.
    remember (fun n:nat => eval e (nat_value n) && !!(~ identity Share.top) &&
      cons_endpoint c Share.top n)%pred as Q.
    eapply derives_cut.
    apply eval_exist_consistent_nat with e.
    subst. do 2 apply and_elim3; eauto.
    subst; simpl; intros. do 2 apply and_elim3; auto.
    subst; simpl.
    apply star_intro; apply and_elim4; auto.
  Qed.

  Lemma TotalSendRecvElim': forall c,
    pTotalSendRecv c == Ex_ n:nat, pSend c Share.top n * pRecv c Share.top n.
  Proof.
    split; intros.

    unfold pSend, pRecv, pTotalSendRecv.
    apply exists_elim; intro n.
    apply star_intro.
    exists n.
    split; auto. split; simpl; auto.
    apply top_share_nonidentity.
    exists n.
    split; auto. split; simpl; auto.
    apply top_share_nonidentity.

    repeat intro.
    destruct H as [n].
    eapply TotalSendRecvElim; eauto.
  Qed.    
  *)
  Lemma eval_join_sub: forall e v w1 w,
    eval e v w ->
    join_sub w1 w ->
    eval e v w1.
  Proof.
    intros.
    destruct H0 as [w2 ?].
    destruct e.
    simpl in *.
    apply e with (l:=locals_to_map (w_env w)); eauto.
    intros.
    destruct H0 as [_ [? _]].
    unfold fst,snd in H0.
    eapply join_sub_locals_to_map'; eauto.
  Qed.

  Lemma eval_join_super: forall e v w1 w,
    eval e v w1 ->
    join_sub w1 w ->
    eval e v w.
  Proof.
    intros.
    destruct H0 as [w2 ?].
    destruct e.
    simpl in *.
    destruct H0 as [_ [? _]].
    unfold fst,snd in H0.
    destruct H0.
    rewrite H0 in H; rewrite H1 in H; auto.
  Qed.

  Lemma eval_joins_eq: forall e v1 v2 w1 w2,
    eval e v1 w1 ->
    joins w1 w2 ->
    eval e v2 w2 ->
    v2 = v1.
  Proof.
    intros.
    destruct H0 as [w ?].
    destruct e.
    simpl in *.
    destruct H0 as [_ [? _]].
    unfold fst,snd in H0.
    destruct H0.
    rewrite H0 in H; rewrite H2 in H.
    rewrite H2 in H1.
    rewrite H in H1.
    inverts H1; auto.
  Qed.

  Lemma pSend_split_combine: forall ec pq (n m: nat),
    let (p,q) := Share.split pq in
    pSend ec (pq) (n+m) == pSend ec p n * pSend ec q m.
  Proof.
    intros.
    case_eq (Share.split pq); intros p q Hpq.
    split; repeat intro.

    destruct H as [c[nm [[[??]?]?]]].
    simpl in H0; inversion H0; subst nm; clear H0.
    simpl in H1.
    destruct (dec_share_identity p).
      elimtype False; apply H1; eapply split_nontrivial'; eauto.
    destruct (dec_share_identity q).
      elimtype False; apply H1; eapply split_nontrivial'; eauto.
    apply split_prod with (p:=p) (q:=q) in H2; eauto.
    destruct H2 as [a1[a2[?[??]]]].
    exists a1; exists a2.
    split; auto; split.
    exists c n; do 3 (split; auto).
    eapply eval_join_sub with (w1:=a1) (w:=a); eauto.
    reflexivity.
    exists c m; do 3 (split; auto).
    eapply eval_join_sub with (w1:=a2) (w:=a); eauto.
    reflexivity.
    apply split_join; auto.

    destruct H as [a1[a2[?[??]]]].
    destruct H0 as [c1[e1[[[??]?]?]]].
    destruct H1 as [c2[e2[[[??]?]?]]].
    assert (cname_value c2=cname_value c1).
      eapply eval_joins_eq; eauto.
    inverts H8.
    exists c1 (n+m).
    split.
    split.
    split.
    eapply eval_join_super; eauto.    
     simpl; auto.
    pose (split_join _ _ _ Hpq).
    intro. apply split_identity in j; auto.

admit.
  Qed.

  Lemma endpoint_channels_incr: forall c p n ch,
    ~identity p ->
    endpoint c p n ch ->
    endpoint c p (n + 1) (channels_incr c ch).
  Proof.
    intros.
    destruct H0.
    split.
    intros.
    specialize (H0 c' H2).
    unfold channels_incr.
    destruct (ch c); auto.
    simpl.
    destruct p0.
    destruct (cname_eq c c'); try contradiction; auto.

    unfold channels_incr.
    destruct (ch c); intros; auto.
    destruct p0.
    simpl.
    destruct (cname_eq c c); intuition.
    simpl.
    destruct x as [[p' m]].
    simpl in *.
    inversion H1; subst p' m.
    destruct (dec_share_identity p); try contradiction.
    auto.
    contradiction.
  Qed.

  Lemma endpoint_channels_decr: forall c p n ch,
    endpoint c p (n + 1) (channels_incr c ch) ->
    endpoint c p n ch.
  Proof.
    intros.
    destruct H.
    split.
    intros.
    specialize (H c' H1).
    unfold channels_incr in H.
    destruct (ch c); auto.
    simpl in H.
    destruct p0.
    destruct (cname_eq c c'); try contradiction; auto.

    unfold channels_incr in H0.
    case_eq (ch c); intros; auto.
    rewrite H1 in H0.
    destruct p0.
    simpl in H0.
    destruct (cname_eq c c); intuition.
    simpl in H0.
    destruct x as [[p' m]].
    simpl in *.
    destruct (dec_share_identity p'); try contradiction.
    elimtype False.
    apply e0 in i. subst m.
    simpl in *.
    inversion H0; subst p'.
    intuition.
    simpl in *.
    inversion H0; subst p'.
    replace m with n; auto.
    omega.
    rewrite H1 in H0.
    simpl in H0.
    rewrite H1 in H0.
    auto.
  Qed.

(*
  Lemma join_prod_incr1: forall c w1 w2 w,
    join w1 w2 w -> join (w_prod_incr c w) w2 (w_prod_incr c w).
  Proof.
*)    

  Lemma join_channels_incr1: forall c ch1 ch2 ch3,
    has_access c ch1 ->
    join ch1 ch2 ch3 ->
    join (channels_incr c ch1) ch2 (channels_incr c ch3).
  Proof.
  Admitted.

  Lemma join_channels_incr2: forall c ch1 ch2 ch3,
    has_access c ch2 ->
    join ch1 ch2 ch3 ->
    join ch1 (channels_incr c ch2) (channels_incr c ch3).
  Proof.
  Admitted.

  Lemma join_channels_decr1: forall c ch1 ch2 ch3,
    join (channels_incr c ch1) ch2 (channels_incr c ch3) ->
    join ch1 ch2 ch3.
  Proof.
  Admitted.

  Lemma join_channels_decr2: forall c ch1 ch2 ch3,
    join ch1 (channels_incr c ch2) (channels_incr c ch3) ->
    join ch1 ch2 ch3.
  Proof.
  Admitted.

  Lemma eval_join_sub_eq: forall e v1 v w1 w,
    eval e v1 w1 ->
    join_sub w1 w ->
    eval e v w ->
    v1=v.
  Proof.
    intros.
    destruct H0 as [w2 ?].
    destruct H0 as [_ [? _]].
    unfold fst,snd in *.
    destruct H0.
    destruct e.
    simpl in *.
    rewrite H0,H2,H1 in H.
    inverts H; auto.
  Qed.

  Lemma pSend_prod_incr: forall w ec c p (n:nat) P,
    w |= eval ec (cname_value c) ->
    w |= pSend ec p n * P ->
    (w_prod_incr c w) |= pSend ec p (n+1) * P.
  Proof.
    introv Hc.
    unfold pSend; intros.
    destruct H as [w1[w2[?[??]]]].
    destruct H0 as [c'[m[[[??]?]?]]].
    assert (cname_value c' = cname_value c).
      eapply eval_join_sub_eq; eauto.
    inverts H5.
    simpl in H2. inversion H2. subst m. clear H2.
    exists (w_prod_incr c w1) w2.
    split.
    destruct w1; destruct w2; destruct w.
    unfold w_prod,w_cons,w_env,w_heap,fst,snd in *.
    unfold w_prod,w_cons,w_env,w_heap in *.
    destruct H as [[?][?]].
    split; split; auto.
    unfold w_prod_incr,wupd_prod.
    unfold w_prod,w_cons,w_env,w_heap,fst,snd in *.
    simpl in H3.
    simpl in H4.
    intuition.
    apply join_channels_incr1; auto.
    destruct H7; intro; rewrite H11 in H9; contradiction.

    split; auto.
    exists c (n+1).
    split.
    split.
    split.
    simpl.
    unfold w_prod_incr.
    rewrite w_env_wupd_prod.
    apply H0. reflexivity.
    simpl; auto.
    simpl.
    unfold w_prod_incr.
    rewrite w_prod_wupd_prod.
    rewrite w_env_wupd_prod.
    rewrite w_heap_wupd_prod.
    rewrite w_cons_wupd_prod.
    destruct w.
    simpl in *.
    intuition.
    apply endpoint_channels_incr; auto.
  Qed.

  Lemma pSend_prod_decr': forall w ec c p (n:nat),
    (w_prod_incr c w) |= eval ec (cname_value c) && pSend ec p (n+1) ->
    w |= pSend ec p n.
  Proof.
    unfold pSend, w_prod_incr; intros.
    destruct H.
    destruct H0 as [c'[m[[[??]?]?]]].
    simpl in H1. inversion H1. subst m. clear H1.
    assert (c' = c).
      congruence.
    subst c'.
    exists c n.
    split.
    split; auto.
    split.
    destruct w.
    simpl in *.
    intuition.
    destruct w.
    simpl in *.
    intuition.
    destruct w.
    simpl in *.
    intuition.
    apply endpoint_channels_decr; auto.
  Qed.

  Lemma pSend_prod_decr: forall w ec c p (n:nat) P,
    (w_prod_incr c w) |= eval ec (cname_value c) && pSend ec p (n+1) * P ->
    w |= pSend ec p n * P.
  Proof.
  Admitted.

  Lemma pRecv_cons_incr: forall w ec c p (n:nat) P,
    w |= (eval ec (cname_value c) && pRecv ec p n) * P ->
    (w_cons_incr c w) |= pRecv ec p (n+1) * P.
  Proof.
    unfold pRecv, w_cons_incr; intros.
    destruct H as [w1[w2[?[??]]]].
    destruct H0 as [Hc H0].
    destruct H0 as [c'[m[[[Hc' ?]?]?]]].
    assert (c'=c). congruence. subst.
    simpl in H0. inversion H0. subst m. clear H0.
    exists (w_cons_incr c w1) w2.
    split; auto.
    destruct w1; destruct w2; destruct w.
    unfold w_prod,w_cons,w_env,w_heap,fst,snd in *; fold w_prod in *; fold w_cons in *; fold w_env in *; fold w_heap in *.
    destruct H as [[?][?]].
    split; split; auto.
    unfold w_prod,w_cons,w_env,w_heap,fst,snd in *; fold w_prod in *; fold w_cons in *; fold w_env in *; fold w_heap in *.
    unfold w_cons_incr.
    unfold w_cons.
    simpl in H2.
    simpl in H3.
    intuition.
    apply join_channels_incr1; auto.
    destruct H6; intro; rewrite H10 in H8; contradiction.

    split; auto.
    exists c (n+1).
    split.
    split; auto.
    split.
    unfold w_cons_incr.
    simpl.
    rewrite w_env_wupd_cons.
    auto.
    simpl; auto.
    simpl.
    unfold w_cons_incr.
    rewrite w_cons_wupd_cons.
    rewrite w_env_wupd_cons.
    rewrite w_heap_wupd_cons.
    rewrite w_prod_wupd_cons.
    destruct w.
    simpl in *.
    intuition.
    apply endpoint_channels_incr; auto.
  Qed.

  Lemma pRecv_cons_decr': forall w ec c p (n:nat),
    (w_cons_incr c w) |= eval ec (cname_value c) && pRecv ec p (n+1) ->
    w |= eval ec (cname_value c) && pRecv ec p n.
  Proof.
    unfold pRecv, w_cons_incr; intros.
    destruct H as [Hc H].
    destruct H as [c'[m[[[Hc' ?]?]?]]].
    simpl in H. inversion H. subst m. clear H.
    assert (c'=c). congruence. subst c'.
    split.
    simpl in Hc. rewrite w_env_wupd_cons in Hc.
    apply Hc.
    exists c n.
    split.
    split; auto.
    split.
    simpl in Hc. rewrite w_env_wupd_cons in Hc.
    apply Hc.
    simpl; auto.
    simpl; auto.
    destruct w.
    simpl in *.
    intuition.
    apply endpoint_channels_decr; auto.
  Qed.

  Lemma pRecv_cons_decr: forall w ec c p (n:nat) P,
    (w_cons_incr c w) |= (eval ec (cname_value c) && pRecv ec p (n+1)) * P ->
    w |= (eval ec (cname_value c) && pRecv ec p n) * P.
  Proof.
  Admitted.

  Lemma pSend_prod_top_count: forall w c (n:nat) P,
    (w_prod_incr c w) |= pSend (expr_value (cname_value c)) Share.top n * P -> n > 0.
  Proof.
  Admitted.

  Lemma pRecv_cons_top_count: forall w c (n:nat) P,
    (w_cons_incr c w) |= pRecv (expr_value (cname_value c)) Share.top n * P -> n > 0.
  Proof.
  Admitted.

  Lemma pRecv_prod: forall w c c' p (n:nat),
    (w_prod_incr c' w) |= pRecv c p n * TT <-> w |= pRecv c p n * TT.
  Proof.
  Admitted.

  Lemma pSend_cons: forall w c c' p (n:nat),
    (w_cons_incr c' w) |= pSend c p n * TT <-> w |= pSend c p n * TT.
  Proof.
  Admitted.

  Lemma pSend_prod_neq: forall w c c' p (n:nat),
    c <> c' ->
   ((w_prod_incr c' w) |= pSend (expr_value (cname_value c)) p n * TT <-> w |= pSend (expr_value (cname_value c)) p n * TT).
  Admitted.

  Lemma pRecv_cons_neq: forall w c c' p (n:nat),
    c <> c' ->
    ((w_cons_incr c' w) |= pRecv (expr_value (cname_value c)) p n * TT <-> w |= pRecv (expr_value (cname_value c)) p n * TT).
  Admitted.


Lemma cons_prod_incr: forall c c' w,
  w_cons (w_prod_incr c w) c' = w_cons w c'.
Proof. intros; destruct w; auto. Qed.

Lemma prod_cons_incr: forall c c' w,
  w_prod (w_cons_incr c w) c' = w_prod w c'.
Proof. intros; destruct w; auto. Qed.

Lemma channels_incr_diff: forall c c' ch,
  c <> c' -> channels_incr c ch c' = ch c'.
Proof.
  intros; unfold channels_incr.
  destruct (ch c) as [[[[p n]]] |]; auto.
  simpl; destruct (cname_eq c c'); intuition.
Qed.

Lemma channels_incr_count: forall c ch C,
  channels_incr c ch c = Some C ->
  let (p,n) := proj1_sig (proj1_sig C) in
  ~identity p ->
  n > 0.
Proof.
  unfold channels_incr; intros.
  case_eq (ch c); intros.
  destruct p as [[[p' n']]].
  rewrite H0 in H.
  simpl in H.
  destruct (cname_eq c c); intuition.
  inversion H. clear H.
  simpl. clear H2.
  destruct (dec_share_identity p').
  simpl; intuition.
  simpl; intuition.
  rewrite H0 in H. simpl in H. rewrite H in H0. discriminate.
Qed.

Lemma prod_incr_count: forall c w C,
  w_prod (w_prod_incr c w) c = Some C ->
  let (p,n) := proj1_sig (proj1_sig C) in
  ~identity p ->
  n > 0.
Proof.
  intros; destruct w; simpl in *.
  apply channels_incr_count in H; auto.
Qed.

Lemma cons_incr_count: forall c w C,
  w_cons (w_cons_incr c w) c = Some C ->
  let (p,n) := proj1_sig (proj1_sig C) in
  ~identity p ->
  n > 0.
Proof.
  intros; destruct w; simpl in *.
  apply channels_incr_count in H; auto.
Qed.

Lemma prod_incr_count': forall c w C,
  w_prod (w_prod_incr c w) c = Some C ->
  let (p,n) := proj1_sig (proj1_sig C) in
  ~identity p ->
  n > 0.
Proof.
  intros; destruct w; simpl in *.
  apply channels_incr_count in H; auto.
Qed.

Lemma cons_incr_count': forall c w C,
  w_cons (w_cons_incr c w) c = Some C ->
  let (p,n) := proj1_sig (proj1_sig C) in
  ~identity p ->
  n > 0.
Proof.
  intros; destruct w; simpl in *.
  apply channels_incr_count in H; auto.
Qed.

Lemma channels_incr_decr: forall c ch C,
  channels_incr c ch c = Some C ->
  let (p,n) := proj1_sig (proj1_sig C) in
  ~identity p ->
  match ch c with
  | Some C' => proj1_sig (proj1_sig C') = (p,n-1) /\ n>0
  | None => False
  end.
Proof.
  intros.
  unfold channels_incr in *.
  case_eq (ch c); intros.
  destruct p as [[[p' n']]].
  rewrite H0 in H.
  simpl in H.
  destruct (cname_eq c c); intuition.
  inversion H. clear H H2.
  simpl.
  destruct (dec_share_identity p');
    simpl; intuition.
  rewrite H0 in H. simpl in H. rewrite H in H0. discriminate.
Qed.



Lemma prod_incr_decr: forall c w C,
  w_prod (w_prod_incr c w) c = Some C ->
  let (p,n) := proj1_sig (proj1_sig C) in
  ~identity p ->
  match w_prod w c with
  | Some C' => proj1_sig (proj1_sig C') = (p,n-1) /\ n>0
  | None => False
  end.
Proof.
  intros; destruct w; simpl in *.
  eapply channels_incr_decr; eauto.
Qed.

Lemma cons_incr_decr: forall c w C,
  w_cons (w_cons_incr c w) c = Some C ->
  let (p,n) := proj1_sig (proj1_sig C) in
  ~identity p ->
  match w_cons w c with
  | Some C' => proj1_sig (proj1_sig C') = (p,n-1) /\ n>0
  | None => False
  end.
Proof.
  intros; destruct w; simpl in *.
  eapply channels_incr_decr; eauto.
Qed.

Lemma pendpoint_prod_decr: forall w c n,
  pendpoint (w_prod (w_prod_incr c w) c) Share.top n <->
  (pendpoint (w_prod w c) Share.top (n-1) /\ n>0).
Proof.
  intros; destruct w; simpl in *.
  eapply pendpoint_incr_decr; eauto.
Qed.
Lemma pendpoint_cons_decr: forall w c n,
  pendpoint (w_cons (w_cons_incr c w) c) Share.top n <->
  (pendpoint (w_cons w c) Share.top (n-1) /\ n>0).
Proof.
  intros; destruct w; simpl in *.
  eapply pendpoint_incr_decr; eauto.
Qed.


Lemma has_endpoint_pendpoint: forall ch c p n,
  has_endpoint c p n ch <-> pendpoint (ch c) p n.
Proof.
  intros. unfold has_endpoint, pendpoint.
  destruct (ch c) as [[[[? ?]]] |]; simpl in *; intuition.
  inversion H; subst; auto. 
  inversion H; subst; auto. 
  subst; auto.
Qed.

Lemma endpoint_pendpoint: forall ch c n,
  endpoint c Share.top n ch ->
  pendpoint (ch c) Share.top n.
Proof.
  intros. destruct H.
  destruct (ch c) as [[[[p n']]] |].
  simpl in *. inversion H0; auto.
  elimtype False. apply top_share_nonidentity. auto.
Qed.

Lemma pendpoint_top_sub: forall a1 a2 a c n,
  join a1 a2 a ->
  pendpoint (a1 c) Share.top n ->
  pendpoint (a c) Share.top n.
Proof.
  intros.
  specialize (H c).
  case_eq (a1 c).
  intros P H1. destruct P as [[[p n']]].
  rewrite H1 in *. simpl in *.
  destruct H0; subst p n'.
  inversion H.
  simpl. auto.
destruct y. destruct z. destruct x1. destruct x1.
subst x.
unfold pjoin in H4.
apply join_com in H4.
simpl in *.
destruct H4. simpl in *.
destruct H0.
      rewrite Share.glb_top in H0.
      rewrite Share.lub_top in H5.
  subst. simpl in *.
  destruct x0. destruct x.
  simpl in *. subst t.
  elimtype False. apply n1. apply endpoint_share_identity. simpl. apply bot_identity.
  intros. rewrite H1 in *.
  clear -H0. simpl in H0. contradiction.
Qed.

  Lemma pSend_total_endpoint: forall w ec c n,
    w |= eval ec (cname_value c) && pSend ec Share.top n ->
    has_total_endpoint c (w_prod w).
  Proof.
  Admitted.

  Lemma pRecv_total_endpoint: forall w ec c n,
    w |= eval ec (cname_value c) && pRecv ec Share.top n ->
    has_total_endpoint c (w_cons w).
  Proof.
  Admitted.


  Lemma pSend_has_endpoint: forall w ec c p (n:nat),
    w |= eval ec (cname_value c) && pSend ec p n ->
    has_endpoint c p n (w_prod w).
  Proof.
  Admitted.

  Lemma pRecv_has_endpoint: forall w ec c p (n:nat),
    w |= eval ec (cname_value c) && pRecv ec p n ->
    has_endpoint c p n (w_cons w).
  Proof.
  Admitted.

  Lemma pRecv_can_recv: forall w ec c n,
    w |= eval ec (cname_value c) && pRecv ec Share.top n ->
    can_recv c w.
  Admitted.
  Lemma pSend_can_send: forall w ec c n,
    w |= eval ec (cname_value c) && pSend ec Share.top n ->
    can_send c w.
  Admitted.


  Lemma pSend_pRecv_has_endpoint_prod: forall w1 w2 w12 ec c ec' p p' (n n':nat),
    w1 |= eval ec (cname_value c) && pSend ec p n ->
    w2 |= pRecv ec' p' n' ->
    join w1 w2  w12 ->
    has_endpoint c p n (w_prod w12).
  Admitted.

  Lemma pSend_pRecv_has_endpoint_prod': forall w1 w2 w12 c ec ec' p p' (n n':nat),
    w1 |= eval ec (cname_value c) && pSend ec p n ->
    w2 |= pRecv ec' p' n' ->
    gjoin w1 w2  w12 ->
    has_endpoint c p n (w_prod w12).
  Admitted.


  Lemma pSend_pRecv_has_endpoint_cons: forall w1 w2 w12 c' ec ec' p p' (n n':nat),
    w1 |= pSend ec p n ->
    w2 |= eval ec (cname_value c') && pRecv ec' p' n' ->
    join w1 w2  w12 ->
    has_endpoint c' p' n' (w_cons w12).
  Admitted.
  Lemma pSend_pRecv_has_endpoint_cons': forall w1 w2 w12 c' ec ec' p p' (n n':nat),
    w1 |= pSend ec p n ->
    w2 |= eval ec (cname_value c') && pRecv ec' p' n' ->
    gjoin w1 w2  w12 ->
    has_endpoint c' p' n' (w_cons w12).
  Admitted.

  Lemma pRecv_top_pendpoint: forall w c ec (n:nat),
    w |= eval ec (cname_value c) && pRecv ec Share.top n ->
    pendpoint (w_cons w c) Share.top n.
  Proof.
    intros.
    apply endpoint_pendpoint.
    destruct H as [? [c_[m [[[??]?] [?[?[??]]] ] ]]]; simpl in *.
    assert (c_ = c). congruence. subst c_.
    congruence.
  Qed.
  
  Lemma pSend_top_pendpoint: forall w c ec (n:nat),
    w |= eval ec (cname_value c) && pSend ec Share.top n ->
    pendpoint (w_prod w c) Share.top n.
  Proof.
    intros.
    apply endpoint_pendpoint.
    destruct H as [? [c_[m [[[??]?] [?[?[??]]] ] ]]]; simpl in *.
    assert (c_ = c). congruence. subst c_.
    congruence.
  Qed.

  Lemma pjoin_top_None: forall (x y z: pjoinable sa_ce),
    ce_share (proj1_sig x) = Share.top ->
    pjoin x y z ->
    False.
  Proof.
    intros.
    destruct x as [[[p1 n1] H1] H2].
    simpl in *.
    subst p1.
    red in H0.
    destruct y as [[[p2 n2]H3]H4].
    destruct z as [[[p n]H5]H6].
    apply join_com in H0.
    simpl in H0.
    unfold join_ce,join_ce' in H0.
    simpl in H0.
    destruct H0.
    destruct H.
    rewrite Share.glb_top in H.
    rewrite Share.lub_top in H7.
    subst p2 p.
    unfold fst,snd in *.
    clear -H4.
    apply H4.
    unfold identity.
    intros [[p3 n3] H5] [[p4 n4] H6].
    intro.
    apply join_com in H.
    simpl in *.
    unfold join_ce,join_ce' in H.
    unfold fst,snd in H.
    simpl in H.
    destruct H.
    subst.
    apply exist_ext; auto.
    replace n2 with 0 in *.
    f_equal.
    destruct H.
    rewrite Share.lub_bot in H0.
    auto.
    omega.
    symmetry; apply H3.
    apply bot_identity.
  Qed.

  Lemma pRecv_cons_Some: forall w c ec (n:nat) E C,
    w |= (eval ec (cname_value c) && pRecv ec Share.top n) * TT ->
    w_cons w c = Some (exist _ (exist _ (Share.top,n) E) C).
  Proof.
    intros.
    simpl in E. simpl in C.
    destruct H as [w1[w2[?[??]]]].
    destruct H0 as [? [c_[m[[[??]?] [?[?[??]]] ] ]]].
    assert (c_ = c) by congruence; subst c_.
    simpl in *.
    assert (m = n) by congruence; subst m.
    destruct H as [[_ ?] _].
    spec H c.
    clear H1 H3 H2 H4.
    unfold fst,snd in *.
    simpl in *.
    destruct H5 as [??].
    inverts H.

    rewrite<- H4 in H2.
    false; apply top_share_nonidentity; auto.
    rewrite H9 in *.
    destruct (w_cons w c) as [[[p m]?] | ].
    f_equal.
    do 2 (apply exist_ext; auto).
    false; apply top_share_nonidentity; auto.

    false.
    eapply pjoin_top_None; eauto.
    clear -H2 H3.
    rewrite<- H3 in H2.
    destruct x as [[[p1 n1] ? ] ?].
    simpl in *.
    inverts H2.
    reflexivity.
  Qed.

  Lemma pSend_prod_Some: forall w c ec (n:nat) E C,
    w |= (eval ec (cname_value c) && pSend ec Share.top n) * TT ->
    w_prod w c = Some (exist _ (exist _ (Share.top,n) E) C).
  Proof.
    intros.
    simpl in E. simpl in C.
    destruct H as [w1[w2[?[??]]]].
    destruct H0 as [? [c_[m[[[??]?] [?[?[??]]] ] ]]].
    assert (c_ = c) by congruence; subst c_.
    simpl in *.
    assert (m = n) by congruence; subst m.
    destruct H as [[? _] _].
    spec H c.
    clear H1 H3 H2 H4.
    unfold fst,snd in *.
    simpl in *.
    destruct H5 as [??].
    inverts H.

    rewrite<- H4 in H2.
    false; apply top_share_nonidentity; auto.
    rewrite H9 in *.
    destruct (w_prod w c) as [[[p m]?] | ].
    f_equal.
    do 2 (apply exist_ext; auto).
    false; apply top_share_nonidentity; auto.

    false.
    eapply pjoin_top_None; eauto.
    clear -H2 H3.
    rewrite<- H3 in H2.
    destruct x as [[[p1 n1] ? ] ?].
    simpl in *.
    inverts H2.
    reflexivity.
  Qed.

  Lemma has_access_sub: forall ch1 ch12 c,
    join_sub ch1 ch12 ->
    has_access c ch1 -> has_access c ch12.
  Admitted.


  Lemma has_endpoint_top_sub: forall ch ch' c n,
    has_endpoint c Share.top n ch ->
    join_sub ch ch' ->
    has_endpoint c Share.top n ch'.
  Admitted.

  Lemma can_recv_sub: forall (w1 w12: world) c,
    gjoin_sub w1 w12 ->
    can_recv c w1 -> can_recv c w12.
  Proof.
    intros.
    eapply has_access_sub; eauto.
    destruct H as [w2 [[H1 H2] _]].
    eexists; eauto.
  Qed.
  Lemma can_recv_sub': forall (w1 w12: world) c,
    gjoin_sub w1 w12 ->
    can_recv c w1 -> can_recv c w12.
  Proof.
    intros. eapply has_access_sub; eauto.
    destruct H as [w2 [[H1 H2] _]].
    eexists; eauto.
  Qed.

  Lemma can_send_sub: forall (w1 w12: world) c,
    gjoin_sub w1 w12 ->
    can_send c w1 -> can_send c w12.
  Proof.
    intros. eapply has_access_sub; eauto.
    destruct H as [w2 [[H1 H2] _]].
    eexists; eauto.
  Qed.
  Lemma can_send_sub': forall (w1 w12: world) c,
    gjoin_sub w1 w12 ->
    can_send c w1 -> can_send c w12.
  Proof.
    intros. eapply has_access_sub; eauto.
    destruct H as [w2 [[H1 H2] _]].
    eexists; eauto.
  Qed.


Lemma total_has_endpoint: forall c ch,
  has_total_endpoint c ch -> exists n, has_endpoint c Share.top n ch.
Admitted.
Lemma has_endpoint_total: forall c ch n,
  has_endpoint c Share.top n ch -> has_total_endpoint c ch.
Admitted.

Lemma pRecv_eval_expr: forall w ec p e,
  w |= pRecv ec p e ->
  exists (n:nat), w |= pRecv ec p n.
Proof.
  intros. destruct H as [c[n ?]]. exists n c n. split. split. split. apply H. reflexivity. apply H. apply H.
Qed.



Lemma locals_imprecise_star1: forall x y,
  locals_imprecise x ->
  locals_imprecise (x*y)%pred.
Admitted.

Lemma join_env: forall P1 P2 w1 w2 w,
  locals_imprecise P1 ->
  locals_imprecise P2 ->
  gjoin w1 w2 w ->
  w1 |= P1 ->
  w2 |= P2 ->
  w |= P1 * P2.
Admitted.


Lemma env_wand_elim: forall x y,
  x * (env_wand x y) |-- y.
Proof.
  repeat intro. destruct H as [a1[a2[?[??]]]].
unfold app_pred, proj1_sig, env_wand in H1.
destruct H as [[??][??]].
apply H1 with (MakeWorld (w_prod a1) (w_cons a1) locals_unit (w_heap a1)).
split; split; auto. 
unfold fst,snd,w_heap,w_env,w_prod,w_cons.
unfold w_env.
clear H H2 H4.
unfold w_env in *.
destruct a2 as [a2p a2c a2e a2h]. destruct a as [ap ac ae ah].
simpl in H3.
inversion H3.
rewrite H in *; rewrite H2 in *.
apply join_com.

apply LocalsMap.empty_map_join.
simpl. 
destruct a2 as [a2p a2c a2e a2h]. destruct a as [ap ac ae ah].
destruct a1 as [a1p a1c a1e a1h].
simpl in *.
clear H1.
inversion H3.
subst.
auto.
Qed.

Lemma and_locals_pred: forall x y,
  locals_pred x -> locals_pred y ->
  locals_pred (x && y)%pred.
Proof.
  unfold locals_pred; intros.
  destruct H1.
  split.
  apply H with w; auto.
  apply H0 with w; auto.
Qed.

Lemma exists_locals_pred: forall B x,
  (forall v:B, locals_pred (x v)) ->
  locals_pred (existential x).
Proof.
  unfold locals_pred; intros.
  destruct H0 as [v].
  exists v.
  apply H with w; auto.
Qed.

Lemma pEvalEq_locals_pred: forall e1 e2,
  locals_pred (pEvalEq e1 e2).
Proof.
  intros.
  unfold pEvalEq.
  apply exists_locals_pred. intros.
  apply and_locals_pred.
  apply locals_pred_eval.
  apply locals_pred_eval.
Qed.


Lemma env_wand_intro: forall  y z x,
  locals_imprecise x ->
  x * y |-- z ->
  x |-- env_wand y z.
Proof.
  intros.
  repeat intro.
  apply H0.
  eexists (MakeWorld (w_prod a) (w_cons a) (w_env w) (w_heap a)); eexists. intuition.  
  Focus 3. apply H3.
  split; split; try apply H2.
  unfold fst,snd,w_prod,w_cons,w_env,w_heap.
  simpl. constructor; auto.
  eapply (H a); eauto.
  exists (w_env w2).
  apply H2.
Qed.
  
Lemma locals_imprecise_emp: locals_imprecise emp.
Admitted.

Lemma env_wand_trivial: forall  x,
  emp |-- env_wand x x.
Proof.
  intros.
  apply env_wand_intro. apply locals_imprecise_emp.
  rewrite star_unit1. auto.
Qed.

Lemma env_wand_equiv: forall x y,
  locals_pred y ->
  (x * env_wand x (x && y) == x && y)%pred.
Proof.
  split; intros.
  apply env_wand_elim.
  repeat intro.
  exists a; exists (wemp (w_env a)). intuition.
  apply join_bot1.
  apply H0.
  repeat intro.
  destruct H0.
  destruct a as [ap ac ae ah].
  destruct w2 as [w2p w2c w2e w2h].
  destruct w as [wp wc we wh].
  simpl in *.
  destruct H1 as [[??][??]].
  unfold fst,snd in *.
  apply channel_sa.channels_unit_identity in H1.
  apply channel_sa.channels_unit_identity in H4.
  apply MemMap.empty_map_identity in H6.
  unfold w_prod in H1.
  unfold w_cons in H4.
  unfold w_heap in H6.
  subst.
  split; auto.
  eapply H; eauto.
Qed.

Lemma locals_pred_imprecise: forall x y,
  locals_pred x ->
  x * y |-- x.
Proof.
  repeat intro.
  destruct H0 as [a1[a2[?[??]]]].
  eapply H; eauto.
  destruct H0 as [_[? _]].
  simpl in H0; inversion H0.
  rewrite H3 in *; rewrite H4 in *. 
  exists locals_unit.
  apply join_com. apply LocalsMap.empty_map_join.
Qed.

Lemma locals_pred_and_dist: forall x y z,
  locals_pred x ->
  (y * z && x == (y&&x) * (z&&x))%pred.
Proof.
  split; intros.
  repeat intro.
  destruct H0.
  destruct H0 as [a1[a2[?[??]]]].
  exists a1; exists a2; intuition.
  split; auto.
  eapply H; eauto.
  destruct H0 as [_[? _]].
  simpl in H0; inversion H0.
  rewrite H4 in *; rewrite H5 in *.
  exists locals_unit.
  apply join_com. apply LocalsMap.empty_map_join.
  split; auto.
  eapply H; eauto.
  destruct H0 as [_[? _]].
  simpl in H0; inversion H0.
  rewrite H4 in *; rewrite H5 in *.
  exists locals_unit.
  apply join_com. apply LocalsMap.empty_map_join.
  apply and_intro.
  auto with seplog_hints.
  etransitivity.
  apply star_dist1.
  eauto.
  apply and_elim4.
  apply locals_pred_imprecise; auto.
Qed.

Lemma decouple_expr_equiv: forall P1 P2 e1 e2,
  locals_imprecise P2 ->
  (P1 * P2 && pEvalEq e1 e2 == P1 * P2 * (env_wand P2 (P2 && pEvalEq e1 e2)))%pred.
Proof.
  pose pEvalEq_locals_pred.
  split; intros.
  setoid_rewrite<- star_assoc.
  setoid_rewrite locals_pred_and_dist; auto.
  apply star_intro. auto with seplog_hints.
  setoid_rewrite env_wand_equiv; auto.
  setoid_rewrite<- star_assoc.
  transitivity (P1 * (P2 && pEvalEq e1 e2))%pred.
  apply star_intro; auto.
  apply env_wand_elim.
  setoid_rewrite star_com.
  apply locals_pred_up'; auto.
Qed.

Lemma combine_pred_decouple_expr_equiv: forall P1 P2 e1 e2 w1 w2 w,
  locals_imprecise P1 ->
  locals_imprecise P2 ->
  w1 |= P1 * env_wand P2 (P2 && pEvalEq e1 e2) ->
  w2 |= P2 ->
  gjoin w1 w2 w ->
  w |= P1 * P2 && pEvalEq e1 e2.
Proof.
  intros.
  apply decouple_expr_equiv; auto.
  setoid_rewrite<- star_assoc.
  setoid_rewrite star_com at 2.
  setoid_rewrite star_assoc.
  eapply join_env; auto.
  apply locals_imprecise_star1; auto.
  apply H3.
  auto. auto.
Qed.  

(*
Lemma split_pred_decouple_expr_equiv: forall P1 P2 e1 e2 w,
  locals_imprecise P1 ->
  locals_imprecise P2 ->
  (lworld w) |= P1 * P2 && eval_eq e1 e2 ->
  exists w1 w2,
  join w1 w2 w /\
  (lworld w1) |= P1 * env_wand P2 (P2 && eval_eq e1 e2) /\
  (lworld w2) |= P2.
Proof.
  intros.
  destruct H1.
  destruct H1 as [w1[w2[?[??]]]].
  exists
  apply decouple_expr_equiv; auto.
  setoid_rewrite<- star_assoc.
  setoid_rewrite star_com at 2.
  setoid_rewrite star_assoc.
  eapply join_env; auto.
  apply locals_imprecise_star1; auto.
  apply H3.
  auto. auto.
Qed.  
*)


Definition gprecise (A: pred world) := forall C D : pred world, (gsep_con A (C && D))%pred == (gsep_con A C && (gsep_con A D))%pred.
Lemma gprecise_star_eq: forall (P1: pred world) (w w1 w1' w2 w2': world),
  gprecise P1 -> w1|=P1 -> w1'|=P1->
  gjoin w1 w2 w ->
  gjoin w1' w2' w ->
  w2 = w2'.
Proof.
  intros.
  spec H (exactly w2) (exactly w2').
  destruct H as [_ ?].
  spec H w. spec H.
  split.
  exists w1 w2; splits; auto.  reflexivity.
  exists w1' w2'; splits; auto.  reflexivity.
  destruct H as [w1_[w2_[?[??]]]].
  simpl in H5; destruct H5. subst w2_ w2'.
  auto.
Qed.
Lemma exactly_gprecise: forall w,
  gprecise (exactly w).
Proof.
  unfold gprecise; intros.
  split; intro; intros.
  destruct H as [a1[a2[?[?[??]]]]].
  simpl in H0; subst a1.
  split; exists w a2; splits; auto; reflexivity.
  destruct H.
  destruct H as [a1[a2[?[??]]]].
  simpl in H1; subst a1.
  destruct H0 as [a1[a3[?[??]]]].
  simpl in H1; subst a1.
  assert (a3=a2).
  eapply @join_canc with (H:=gworld_sa);
    apply @join_com; eauto.
  subst a3.
  exists w a2; splits; auto. reflexivity.
  split; auto.
Qed.


  Definition world_rm_vars (w: world) : world :=
    MakeWorld (w_prod w) (w_cons w) (locals_unit) (w_heap w).
  Program Definition no_vars (x: pred world) : pred world :=
    fun w => (world_rm_vars w) |= x.
  Program Definition rm_vars (x: pred world) : pred world :=
    fun w => w_env w = locals_unit /\ exists (e:locals), (MakeWorld (w_prod w) (w_cons w) e (w_heap w)) |= x.

  Lemma join_gjoin1: forall w1 w2 w,
    join w1 w2 w -> gjoin w1 (world_rm_vars w2) w.
  Proof.
    intros.
    split; split; try apply H.
    destruct H as [_[? _]].
    unfold fst,snd in *.
    simpl in H. inverts H.
    destruct w2.
    rewrite H0 in *. rewrite H1 in *.
    unfold world_rm_vars, w_env.
    apply join_com.
    apply LocalsMap.empty_map_join.
  Qed.

  Lemma join_gjoin2: forall w1 w2 w,
    join w1 w2 w -> gjoin (world_rm_vars w1) w2 w.
  Proof. intros; apply join_com; apply join_gjoin1; apply join_com; auto. Qed.

  Lemma star_gstar_rm_vars: forall w (x y : pred world),
    w |= x * y -> w |= gsep_con x (rm_vars y).
  Proof.
    intros.
    destruct H as [w1[w2[?[??]]]].
    exists w1 (world_rm_vars w2); splits; auto.
    apply join_gjoin1; auto.
    split; auto.
    exists (w_env w2).
    destruct w2. simpl. apply H1.
  Qed.

  Lemma star_gstar_no_vars: forall w x y,
    w |= x * (no_vars y) -> w |= gsep_con x y.
  Proof.
    intros.
    destruct H as [w1[w2[?[??]]]].
    exists w1 (world_rm_vars w2); splits; auto.
    apply join_gjoin1; auto.
  Qed.

  Lemma star_gstar_assoc1: forall x y z,
    gsep_con (x * y)%pred z |-- gsep_con x (gsep_con (rm_vars y) z).
  Proof.
    repeat intro.
    destruct H as [wxy[wz[?[??]]]].
    destruct H0 as [wx[wy[?[??]]]].
    edestruct @join_assoc with (H:=gworld_sa) (d:=wxy) (c:=wz) as [wyz[??]]; eauto.
    apply join_gjoin1; eauto.
    exists wx wyz; splits; auto.
    exists (world_rm_vars wy) wz; splits; auto.
    split; auto. exists (w_env wy).
    destruct wy; simpl; auto.
  Qed.



  Definition is_bool (e:expr) := existential (fun n:nat => eval e (nat_value n)).

  Definition evaluable (e: expr) : pred world := existential (fun v => eval e v).



  Lemma evaluable_dec: forall (e:expr) (w:world), {w|=evaluable e} + {~w|=evaluable e}.
Admitted.


  Lemma evaluable_gstar_up: forall A B C e v,
    A |-- gsep_con (B && evaluable e) C ->
    A && eval e v |-- gsep_con (B && eval e v) C.
  Proof.
Admitted.

  Lemma ex_con: forall (A : Type) (as_alg : sepalg A) (as_age : ageable A) (H : ASA A) (B : Type) (P : B -> pred A) (Q : pred A),
    (sep_con (existential P) Q)%pred == (existential (fun x:B=> sep_con (P x) Q))%pred.
  Proof.
    split.
    repeat intro.
    destruct H0 as [w1 [w2 [? [? ?]]]].
    destruct H1 as [x ?]. exists x.
    exists w1. exists w2. intuition.
    repeat intro.
    destruct H0. destruct H0 as [w1 [w2 [? [? ?]]]].
    exists w1. exists w2; intuition.
    exists x; auto.
  Qed.

  Lemma ex_gcon: forall (B : Type) (P : B -> pred world) (Q : pred world),
    (gsep_con (existential P) Q)%pred == (existential (fun x:B=> gsep_con (P x) Q))%pred.
  Proof.
    split.
    repeat intro.
    destruct H as [w1 [w2 [? [[x ?] ?]]]].
    exists x w1 w2; intuition.
    repeat intro.
    destruct H as [x [w1 [w2 [? [? ?]]]]].
    exists w1 w2; intuition.
    exists x; auto.
  Qed.

  Lemma prop_sep_con: forall A B C,
    ((A && prop B) * C == (A * C) && prop B)%pred.
  Proof.
    intros.
    split.
    apply and_intro.
    apply split_con; eauto.
    apply and_elim3; auto.
    repeat intro.
    destruct H as [w1[w2[?[[??]?]]]].
    apply H1.
    repeat intro.
    destruct H.
    eapply split_con with A C; auto.
    apply and_intro; auto.
  Qed.

  Lemma prop_gsep_con: forall A B C,
    (gsep_con (A && prop B) C == (gsep_con A C) && prop B)%pred.
  Proof.
    intros.
    split.
    apply and_intro.
    apply split_gcon; eauto.
    apply and_elim3; auto.
    repeat intro.
    destruct H as [w1[w2[?[[??]?]]]].
    apply H1.
    repeat intro.
    destruct H.
    eapply split_gcon with A C; auto.
    apply and_intro; auto.
  Qed.

  Lemma pIsTrue_gstar_up: forall e A B C,
    A |-- gsep_con (B && evaluable e) C ->
    A && pIsTrue e  |-- gsep_con (B && pIsTrue e) C.
  Proof.
    intros.
    unfold pIsTrue.
    setoid_rewrite and_comm.
    setoid_rewrite ex_and.
    setoid_rewrite ex_gcon.
    apply exists_elim.
    intro v.
    setoid_rewrite<- and_assoc at 2.
    setoid_rewrite and_comm at 4.
    setoid_rewrite and_assoc at 1.
    setoid_rewrite prop_gsep_con.
    apply and_intro.
    setoid_rewrite<- and_assoc.
    setoid_rewrite and_comm at 2.
    setoid_rewrite and_assoc.
    apply and_elim3.
    setoid_rewrite and_comm at 2.
    eapply derives_cut.
    2: apply evaluable_gstar_up; eauto.
    setoid_rewrite and_comm at 1. auto.

    apply and_elim3. apply and_elim4; auto.
  Qed.

  Lemma pIsFalse_gstar_up: forall e A B C,
    A |-- gsep_con (B && evaluable e) C ->
    A && pIsFalse e  |-- gsep_con (B && pIsFalse e) C.
  Proof.
    intros.
    unfold pIsFalse.
    apply evaluable_gstar_up; auto.
  Qed.

  Lemma eval_gframe: forall e v w1 w2 w, 
    eval e v w1 -> gjoin w1 w2 w -> eval e v w.
Admitted.

  Lemma pIsTrue_gframe: forall e w1 w2 w, 
    pIsTrue e w1 -> gjoin w1 w2 w -> pIsTrue e w.
  Proof.
    intros.
    destruct H as [n[??]].
    exists n; split; auto.
    eapply eval_gframe; eauto.
  Qed.
  Lemma pIsFalse_gframe: forall e w1 w2 w, 
    pIsFalse e w1 -> gjoin w1 w2 w -> pIsFalse e w.
  Proof.
    intros.
    eapply eval_gframe; eauto.
  Qed.
  Lemma pIsTrue_gstar: forall e F, 
    gsep_con (pIsTrue e) F |-- pIsTrue e.
  Proof.
    intros. intro w. intros.
    destruct H as [w1[w2[?[??]]]].
    eapply pIsTrue_gframe; eauto.
  Qed.
  Lemma pIsFalse_gstar: forall e F, 
    gsep_con (pIsFalse e) F |-- pIsFalse e.
  Proof.
    intros. intro w. intros.
    destruct H as [w1[w2[?[??]]]].
    eapply pIsFalse_gframe; eauto.
  Qed.


  Lemma star_gstar: forall (A B: pred world),
    (A * B |-- gsep_con A (rm_vars B))%entails_pred.
  Proof.
    intros.
    repeat intro.
    destruct H as [wA[wB[?[??]]]].
    exists wA (world_rm_vars wB); splits; auto.
    apply join_gjoin1; auto.
    simpl; split; auto.
    exists (w_env wB).
    destruct wB; auto.
  Qed.

  Lemma sep_overlap_con: forall {A:Type} `{ASA A} B C, (sep_overlap B C == (B * TT) && (C * TT))%pred.
  Proof.
    intros.
    split; intros.
    repeat intro.
    destruct H0 as [w1[w2[w3[w12[w23[?[?[?[??]]]]]]]]].
    split.
    exists w12 w3; splits; auto.
    simpl; auto.
    exists w23 w1; splits; auto.
    apply join_com; auto.
    edestruct (join_assoc H0 H2) as [w23'[??]].
    assert (w23 = w23').
    eapply join_eq; eauto.
    subst w23'.
    auto.
    simpl; auto.

    repeat intro.
Admitted.

  Lemma gsep_overlap_con: forall B C, (@sep_overlap world gworld_sa world_ageable gworld_asa B C == (gsep_con B TT) && (gsep_con C TT))%pred.
  Proof.
    intros.
    apply sep_overlap_con.
  Qed.

  Lemma gjoin_com: forall w1 w2 w, gjoin w1 w2 w -> gjoin w2 w1 w.
  Proof. eapply @join_com. Qed.
  Lemma gjoin_assoc: forall a b c d e,
    gjoin a b d -> gjoin d c e -> {f | gjoin b c f /\ gjoin a f e}.
  Proof. intros. edestruct @join_assoc with (a:=a) (b:=b) (H:=gworld_sa); eauto. Qed.
  Lemma gjoin_eq: forall a b ab ab',
    gjoin a b ab -> gjoin a b ab' -> ab = ab'.
  Proof. eapply @join_eq. Qed.
  Lemma gjoin_canc: forall a1 a2 b c,
    gjoin a1 b c -> gjoin a2 b c -> a1 = a2.
  Proof. eapply @join_canc. Qed.


  Lemma gjoin_bot3: forall a b,
    gjoin a b gwemp -> a=gwemp /\ b=gwemp.
  Proof.
    intros.
    split.
    eapply @join_sub_antisym with (H:=gworld_sa).
    eexists; eauto.
    exists a. apply gjoin_bot2.
    eapply @join_sub_antisym with (H:=gworld_sa).
    apply gjoin_com in H.
    eexists; eauto.
    exists b. apply gjoin_bot2.
  Qed.

  Lemma gprecise_TT_eq: forall P1 P2 w1 w2 w1' w2' w,
    gprecise P1 ->
    gprecise P2 ->
    w1 |= P1 -> w2 |= P2 ->
    w1' |= (gsep_con P1 TT) -> w2' |= (gsep_con P2 TT) ->
    gjoin w1 w2 w -> gjoin w1' w2' w ->
    w1'=w1 /\ w2'=w2.
  Proof.
    introv Hprec1 Hprec2.
    intros.
    rename w1' into w1F1; rename w2' into w2F2.
    destruct H1 as [w1'[wF1[?[? _]]]].
    destruct H2 as [w2'[wF2[?[? _]]]].
    edestruct gjoin_assoc with (a:=w1') (b:=wF1) as [w2F[??]]; eauto.
    apply gjoin_com in H7.
    edestruct gjoin_assoc with (a:=w2') (b:=wF2) (c:=wF1) as [wF[??]]; eauto.
    apply gjoin_com in H9.
    assert (w2F = w2); try subst w2F.
      eapply gprecise_star_eq with (P1:=P1) (w1:=w1') (w1':=w1); eauto.
    assert (w1'=w1); try subst w1'.
      eapply gjoin_canc; eauto.
    assert (gjoin w2 gwemp w2).
      apply gjoin_bot1.
    assert (wF = gwemp); try subst wF.
      eapply gprecise_star_eq with (P1:=P2) (w1:=w2') (w1':=w2); eauto.
    assert (w2'=w2); try subst w2'.
      eapply gjoin_canc; eauto.
    destruct gjoin_bot3 with (a:=wF1) (b:=wF2); auto.
    subst wF1 wF2.
    clear H9.
    split.
    symmetry.
    eapply gworld_unit_identity.
    apply gjoin_com. auto.
    symmetry.
    eapply gworld_unit_identity.
    apply gjoin_com. auto.
  Qed.

  Lemma gjoin_sub_framing: forall P1 P2 w1 w2 w1F w2F w12 wF wF1 wF2 w,
    w1 |= P1 -> w2 |= P2 ->
    w1F |= gsep_con P1 TT -> w2F |= gsep_con P2 TT ->
    gjoin w1 w2 w12 ->
    gjoin w12 wF w ->
    gjoin w1F w2F w ->
    gjoin w1 wF1 w1F ->
    gjoin w2 wF2 w2F ->
    gjoin wF1 wF2 wF.
  Proof.
    intros.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w1) (b:=w2) as [w20[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w1) (b:=wF1) as [wF12F[??]]; eauto.
    assert (wF12F = w20); try subst wF12F.
      eapply @join_canc with (H:=gworld_sa); eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w2) (b:=wF2) as [wF'[??]]; eauto.
    assert (wF' = wF); try subst wF'.
      eapply @join_canc with (H:=gworld_sa); eauto.
    apply join_com; auto.
  Qed.

  Lemma TT_gsplit: gsep_con TT TT == TT.
    split; intros.
    apply TT_intro.
    apply TT_gstar.
  Qed.

  Lemma gprecise_TT_sub: forall P1 P2 w1 w2 w3 w4 w12 wF w,
    gprecise P1 -> gprecise P2 ->
    P1 w1 -> P2 w2 -> gjoin w1 w2 w12 ->
    w3 |= gsep_con P1 TT -> w4|= gsep_con P2 TT -> gjoin w3 w4 w ->
    gjoin w12 wF w ->
    gjoin_sub w1 w3 /\ gjoin_sub w2 w4.
  Proof.
    intros.
    destruct H4 as [w1'[wF1[?[??]]]].
    destruct H5 as [w2'[wF2[?[??]]]].
    edestruct @join_assoc with (H:=gworld_sa) (a:=w1) (b:=w2) as [w2F[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w1') (b:=wF1) as [w2F'[??]]; eauto.
    assert (w2F = w2F'); try subst w2F'.
      eapply gprecise_star_eq with (P1:=P1) (w1:=w1) (w2:=w2F); eauto.
    assert (w1 = w1'); try subst w1'.
      eapply @join_canc with (H:=gworld_sa); eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w2') (b:=wF2) as [wF'[??]]; eauto.
    assert (wF = wF'); try subst wF'.
      eapply gprecise_star_eq with (P1:=P2) (w1:=w2) (w2:=wF); eauto.
    assert (w2 = w2'); try subst w2'.
      eapply @join_canc with (H:=gworld_sa); eauto.
    split; eexists; eauto.
  Qed.

  Lemma derives_exactly: forall w P,
    w|= P -> exactly w |-- P.
  Proof.
    repeat intro. simpl in H0; subst a. auto.
  Qed.

  Lemma is_bool_evaluable: forall e,
    is_bool e |-- evaluable e.
  Proof.
    repeat intro.
    destruct H as [n].
    exists (nat_value n); auto.
  Qed.

  Lemma and_dist: forall x y z,
    (x && (y || z) == (x&&y) || (x&&z))%pred.
  Proof.
    split; repeat intro.
    destruct H.
    destruct H0.
    left; split; auto.
    right; split; auto.
    destruct H as [[??]|[??]]; split; auto.
    left; auto.
    right; auto.
  Qed.


  Definition expr_true := nat_to_expr 1.
  Definition expr_false := nat_to_expr 0.


  Lemma evalTrue_const: forall n w,
    n <> 0 ->
    evalTrue n w.
  Proof. intros. eexists. split; eauto. reflexivity. Qed.

  Lemma evalFalse_const: forall w,
    evalFalse 0 w.
  Proof. intros. reflexivity. Qed.

  Lemma expr_plus_nat_const: forall n1 n2,
    expr_plus (nat_to_expr n1) (nat_to_expr n2) == nat_to_expr (n1+n2).
  Proof.
    intros. intro l. reflexivity.
  Qed.

  Lemma pIsTrue_evaluable: forall e,
    pIsTrue e |-- evaluable e.
  Proof. repeat intro. destruct H as [?[??]]; eexists; eauto. Qed.
  Lemma pIsFalse_evaluable: forall e,
    pIsFalse e |-- evaluable e.
  Proof. repeat intro. eexists; eauto. Qed.

  Lemma evaluable_gstar: forall e F,
    gsep_con (evaluable e) F |-- evaluable e.
  Proof.
    intros. repeat intro.
    destruct H as [w1[w2[?[??]]]].
    destruct H0 as [v ?].
    exists v.
    eapply eval_gframe; eauto.
  Qed.

End Lemmas.












