
Require Import Machine.
Require Import ContextSemantics.
Require Import Bisimulation.
Require Import ParBisimulation.
Require Import HoareLogic.
Require Import WorldLogic.
Require Import Termination.
Require Import world_model.
Require Import msl.msl_standard.
Require Import LibTactics.
Require Import SfLib.


Section Definitions.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import Machine.Notation.
  Import ContextSemantics.Notation.

  Open Scope context_scope.
  Open Scope wterm_scope.

  (* declares that channel c has resource invariant function R *)
  Definition s_has_ch_ri s c R:= exists wl vl,
    s_get_ch s c = Some (R, wl, vl).

  Definition cannot_communicate_channels_world s w :=
    forall c R v w',
      can_recv c w ->
      s_has_ch_ri s c R ->
      R v w' ->
      (forall c', ~can_recv c' w' /\ ~can_send c' w').

  Fixpoint cannot_communicate_channels_term t := match t with
    (* no new resource invariants may hold on a resource htat may be used to receive permissions on channel c *)
  | tm_newch _ R => forall v w', R v w' -> (forall c', ~can_recv c' w' /\ ~can_send c' w')
  | tm_while_max _ _ t1 => cannot_communicate_channels_term t1
  | tm_if _ t1 t2 => cannot_communicate_channels_term t1 /\ cannot_communicate_channels_term t2
  | tm_local _ f => forall x, cannot_communicate_channels_term (f x)
  | tm_seq t1 t2 => cannot_communicate_channels_term t1 /\ cannot_communicate_channels_term t2
  | tm_par _ t1 _ t2 => cannot_communicate_channels_term t1 /\ cannot_communicate_channels_term t2
  | _ => True
  end.

  Fixpoint cannot_communicate_channels_wterm b := match b with
  | wtm_term w t => cannot_communicate_channels_term t
  | wtm_seq b t => cannot_communicate_channels_wterm b /\ cannot_communicate_channels_term t
  | wtm_par b1 b2 => cannot_communicate_channels_wterm b1 /\ cannot_communicate_channels_wterm b2
  end.

  Definition cannot_communicate_channels s b :=
    exists w, join_wterms b w /\ cannot_communicate_channels_world s w /\ cannot_communicate_channels_wterm b.


  Definition noninterfering w1 w2 := forall c,
    (can_send c w1 -> ~can_send c w2) /\
    (can_send c w2 -> ~can_send c w1) /\
    (can_recv c w1 -> ~can_recv c w2) /\
    (can_recv c w2 -> ~can_recv c w1).


  Definition noninterfering_wterm b1 b2 := forall w1 w2,
    join_wterms b1 w1 -> join_wterms b2 w2 -> noninterfering w1 w2.

  Definition no_send_conflict wC wb := forall c,
    can_send c wb -> (~can_send c wC /\ ~can_recv c wC).

  (*
    if no_send_to_context C b1, b1 must terminate, and b1 and b2 don't conflict,
    then C[b1]; b2 simulates C[b1]; b2
  *)
  Definition no_send_to_context C b := forall wC wb,
      join_wterms C[gwemp:tm_skip] wC ->
      join_wterms b wb ->
      no_send_conflict wC wb.

End Definitions.

Module Refactor.
  Import Bisimulation.Notation.
  Import Machine.Notation.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import HoareLogic.Notation.
  Import HoareLogic.Notation.U.
  Import WorldLogic.Notation.
  Import WorldLogic.Notation.U.
  Import WorldLogic.Tactics.
  Import ContextSemantics.Notation.
  Import WorldLogic.Morphisms.

  Open Scope forces_pred.
  Open Scope context_scope.
  Open Scope bsim_scope.
  Open Scope htriple_scope.
  Open Scope term_scope.
  Open Scope join_scope.


  Lemma terminates_sub: forall wC wC1 s b,
    terminates s wC b ->
    gjoin_sub wC1 wC ->
    valid_machine_state wC s b ->
    terminates s wC1 b.
  Proof.
    unfold terminates; intros.
    destruct H0 as [wC2 ?].
    rename wC' into wC1'.
    edestruct valid_machine_state_cstep_star with (C:=cx_par1 C (wC2:tm_skip)) (C':=cx_par1 C' (wC2:tm_skip)) as [wC'[??]]; eauto.
    apply cstep_star_par_cx_par1; apply cstep_star_par1_step; eauto.
    exists wC1 wC2; splits; auto. reflexivity.
    eapply H with (C:=cx_par1 C (wC2:tm_skip)) (C':=cx_par1 C' (wC2:tm_skip)); eauto.
    exists wC1 wC2; splits; auto. reflexivity.
    apply cstep_star_par_cx_par1; apply cstep_star_par1_step; eauto.
  Qed.




Import PermuteTactics.

Import ParBisimulation.PermuteTactics.
Import JoinList.





  Lemma terminates_s_iter: forall wC s b s',
    terminates s wC b ->
    s_iter s s' ->
    terminates s' wC b.
  Proof.
    unfold terminates; intros.
    edestruct H with (C:=C) as [s''[w'' ?]]; eauto.
    eapply cstep_iteration; eauto.
  Qed.


  Lemma valid_machine_state_subO: forall wC1 wC s b,
    valid_machine_state wC s b ->
    gjoin_sub wC1 wC ->
    valid_machine_state wC1 s b.
  Proof.
    introv [wS[wb[ww[?[??]]]]] [wC2 ?].
    apply gjoin_join_list in H2.
    jrewrite<- H2 in H1.
    apply join_list_perm with (wl2:=[wC2]++[wC1,wS,wb]) in H1.
    edestruct join_list_app_inv as [wC2_[ww'[?[??]]]]; eauto.
    apply join_list_single in H3. subst wC2_.
    exists wS wb ww'; splits; auto.
    eapply join_list_perm; eauto.
    permut4.
    permut4.
  Qed.

  Lemma s_ch_get_rm_diff: forall s c1 c2,
    c1 <> c2 ->
    s_get_ch (s_rm_ch s c1) c2 = s_get_ch s c2.
  Proof.
    unfold s_get_ch;intros.
    destruct s; simpl.
    apply POrdProps.P.F.remove_neq_o; auto.
  Qed.

  Lemma no_recv_None: forall w c, ~can_recv c w <-> w_cons w c = None.
  Proof.
    unfold can_recv, has_access; split; intros.
    case_eq (w_cons w c); intros; auto.
    false; apply H. intro. congruence.
    intro; apply H0; auto.
  Qed.

  Lemma no_send_None: forall w c, ~can_send c w <-> w_prod w c = None.
  Proof.
    unfold can_send, has_access; split; intros.
    case_eq (w_prod w c); intros; auto.
    false; apply H. intro. congruence.
    intro; apply H0; auto.
  Qed.


End Refactor.

Section Lemmas.
  Import Bisimulation.Notation.
  Import Machine.Notation.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import HoareLogic.Notation.
  Import HoareLogic.Notation.U.
  Import WorldLogic.Notation.
  Import WorldLogic.Notation.U.
  Import WorldLogic.Tactics.
  Import ContextSemantics.Notation.
  Import WorldLogic.Morphisms.
  Import JoinList.

Import Refactor.

  Open Scope forces_pred.
  Open Scope context_scope.
  Open Scope bsim_scope.
  Open Scope htriple_scope.
  Open Scope term_scope.
  Open Scope join_scope.


(*
  Lemma no_send_to_context_compose: forall C1 C2 b,
    no_send_to_context (compose_contexts C2 C1) b ->
    no_send_to_context C1 b.
  Admitted.
*)

(*
  Lemma cstep_star_channel_noninterference: forall (s:state) (C:context) (b:wterm) s' C' b' ww (w':world) c,
    ww ||= s (x) C (x) b (x) w' ->
    has_total_endpoint c (w_cons w') ->
    cstep_star s C b s' C' b' ->
    (s_get_ch s c = None /\ s_get_ch s' c = None) \/
    (exists q, exists q', exists R, exists wcl, exists wcl',
      s_get_ch s c = Some (R,wcl,q) /\ s_get_ch s' c = Some (R,wcl++wcl',q++q')).
  Admitted.
*)

  Lemma cannot_communicate_channels_wterm_step: forall s b s' b',
    cannot_communicate_channels_wterm b ->
    step s b s' b' ->
    cannot_communicate_channels_wterm b'.
  Proof.
    intros.
    induction H0; simpl in *; eauto; intuition.
  Qed.

  Lemma cannot_communicate_channels_world_step: forall s w b s' w' b',
    cannot_communicate_channels_world s w ->
    join_wterms b w ->
    join_wterms b' w' ->
    step s b s' b' ->
    cannot_communicate_channels_world s' w'.
  Proof.
    intros.
    revert w w' H H0 H1.
    induction H2; intros; simpl in *; subst; subst; eauto.
Admitted.

  Lemma cannot_communicate_channels_step: forall wC s b s' b',
    cannot_communicate_channels s b ->
    step s b s' b' ->
    valid_machine_state wC s b ->
    cannot_communicate_channels s' b'.
  Proof.
    intros.
    destruct H as [w[?[??]]].
    edestruct valid_machine_state_join_wterms with (s:=s') as [w' ?].
    eapply valid_machine_state_step; eauto.
    exists w'; splits; auto.
    eapply cannot_communicate_channels_world_step; eauto.
    eapply cannot_communicate_channels_wterm_step; eauto.
  Qed.


  Lemma cannot_communicate_channels_stepO: forall wO s bO b s' bO',
    cannot_communicate_channels s b ->
    step s bO s' bO' ->
    join_wterms bO wO ->
    valid_machine_state wO s b ->
    cannot_communicate_channels s' b.
  Proof.
    intros.
    revert wO H1 H2 H.
    step_cases (induction H0) Case; intros; subst; eauto.
    - Case "st_send".
    simpl in H6; subst wO.
    destruct H8 as [wb[?[??]]].
    exists wb; splits; auto.
    red; intros.
    destruct H10 as [wl'[vl' ?]].
    destruct (Pos.eq_dec c0 c). subst c0.
    erewrite s_ch_get_push_same in H10; eauto.
    inverts H10.
    eapply H6; eauto. do 2 eexists; eauto.
    rewrite s_ch_get_push_diff in H10; auto.
    eapply H6; eauto. do 2 eexists; eauto.
    - Case "st_recv".
    simpl in H6; subst wO.
    destruct H8 as [wb[?[??]]].
    exists wb; splits; auto.
    red; intros.
    eapply H6; eauto.
    destruct H10 as [wl'[vl' ?]].
    destruct (Pos.eq_dec c0 c). subst c0.
    erewrite s_ch_get_pop_same in H10; eauto.
    inverts H10.
    do 2 eexists; eauto.
    rewrite s_ch_get_pop_diff in H10; auto.
    do 2 eexists; eauto.
    - Case "st_newch".
(* TODO: add condition to valid_machine_state that says that every channel permission must refer to a valid channel in the global state; thus when the context allocates a channel that is not in the current global state, it cannot be possible for b to have had permission to the channel. *)
admit.
    - Case "st_endch".
    simpl in H4; subst wO.
    destruct H6 as [wb[?[??]]].
    exists wb; splits; auto.
    red; intros.
    eapply H4; eauto.
    destruct H8 as [wl'[vl' ?]].
    destruct (Pos.eq_dec c0 c). subst c0.
    erewrite s_ch_get_rm_same in H8; eauto.
    false.
    rewrite s_ch_get_rm_diff in H8; auto.
    do 2 eexists; eauto.
    - Case "st_par1".
    destruct H1 as [wO1[wO2[?[??]]]].
    apply IHstep with wO1; auto.
    apply valid_machine_state_subO with wO; auto. exists wO2; auto.
    - Case "st_par2".
    destruct H1 as [wO1[wO2[?[??]]]].
    apply IHstep with wO2; auto.
    apply valid_machine_state_subO with wO; auto. exists wO1; auto.
  Qed.


  Lemma cannot_communicate_channels_cstep: forall wC C s b s' C' b',
    cannot_communicate_channels s b ->
    cstep s C b s' C' b' ->
    context_world C wC ->
    valid_machine_state wC s b ->
    cannot_communicate_channels s' b'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    eapply cannot_communicate_channels_step; eauto.
    subst b'; auto. clear H0.
    eapply cannot_communicate_channels_stepO; eauto.
    eapply step_simulates_cstep; eauto.
    apply context_world_join_wterms; auto.    
  Qed.

  Lemma cannot_communicate_channels_s_iter: forall s b s',
    cannot_communicate_channels s b ->
    s_iter s s' ->
    cannot_communicate_channels s' b.
  Proof.
    intros.
    destruct H as [wb[?[??]]].
    exists wb; splits; auto.
    red; intros.
    eapply H1; eauto.
    destruct H4 as [wl[vl ?]]; exists wl vl; auto.
    erewrite s_iter_s_get_ch; eauto.
  Qed.

  Lemma cannot_communicate_channels_step_star: forall wC s b s' b',
    cannot_communicate_channels s b ->
    step_star s b s' b' ->
    valid_machine_state wC s b ->
    cannot_communicate_channels s' b'.
  Proof.
    intros.
    induction H0; auto.
    eapply IHstep_star; eauto.
    eapply cannot_communicate_channels_step; eauto.
    eapply valid_machine_state_step; eauto.
    eapply IHstep_star; eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma cannot_communicate_channels_cstep_star: forall wC C s b s' C' b',
    cannot_communicate_channels s b ->
    cstep_star s C b s' C' b' ->
    context_world C wC ->
    valid_machine_state wC s b ->
    cannot_communicate_channels s' b'.
  Proof.
    intros.
    revert wC H1 H2.
    induction H0; intros; auto.
    edestruct valid_machine_state_cstep as [wC'[??]]; eauto.
    eapply IHcstep_star; eauto.
    eapply cannot_communicate_channels_cstep; eauto.
    eapply IHcstep_star; eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma cannot_communicate_channels_seq: forall s b t,
    cannot_communicate_channels s (b;; t) ->
    cannot_communicate_channels s b.
  Proof.
    intros.
    destruct H as [wb[?[??]]]; simpl in *; exists wb; splits; intuition.
  Qed.

  Lemma cannot_communicate_channels_world_sub: forall s w w',
    cannot_communicate_channels_world s w ->
    gjoin_sub w' w ->
    cannot_communicate_channels_world s w'.
  Proof.
    unfold cannot_communicate_channels_world; intros.
    eapply H; eauto.
    eapply can_recv_sub; eauto.
  Qed.

  Lemma cannot_communicate_channels_par1: forall s b1 b2,
    cannot_communicate_channels s (b1 || b2) ->
    cannot_communicate_channels s b1.
  Proof.
    intros.
    destruct H as [wb[?[?[??]]]]; simpl in *.
    destruct H as [w1[w2[?[??]]]].
    exists w1; splits; auto.
    eapply cannot_communicate_channels_world_sub; eauto.
    exists w2; auto.
  Qed.

  Lemma cannot_communicate_channels_par2: forall s b1 b2,
    cannot_communicate_channels s (b1 || b2) ->
    cannot_communicate_channels s b2.
  Proof.
    intros.
    destruct H as [wb[?[?[??]]]]; simpl in *.
    destruct H as [w1[w2[?[??]]]].
    exists w2; splits; auto.
    eapply cannot_communicate_channels_world_sub; eauto.
    exists w1; auto.
  Qed.

  Lemma noninterfering_symm: forall w1 w2,
    noninterfering w1 w2 -> noninterfering w2 w1.
  Proof.
     unfold noninterfering; intros.
     destruct (H c) as [?[?[??]]].
     auto.
  Qed.

  Lemma noninterfering_sub1: forall w1 w1' w2,
    noninterfering w1 w2 ->
    gjoin_sub w1' w1 ->
    noninterfering w1' w2.
  Proof.
    unfold noninterfering; intros.
    destruct (H c) as [?[?[??]]]; clear H.
    splits; intros.
    apply H1. eapply can_send_sub; eauto.
    intro; apply H2; auto. eapply can_send_sub; eauto.
    apply H3. eapply can_recv_sub; eauto.
    intro; apply H4; auto. eapply can_recv_sub; eauto.
  Qed.

  Lemma noninterfering_sub2: forall w1 w2 w2',
    noninterfering w1 w2 ->
    gjoin_sub w2' w2 ->
    noninterfering w1 w2'.
  Proof.
    intros.
    apply noninterfering_symm in H.
    apply noninterfering_symm.
    eapply noninterfering_sub1; eauto.
  Qed.


  Lemma can_send_w_env_set: forall w c x v,
    can_send c w <-> can_send c (w_env_set w x v).
  Proof.
    unfold can_send; intros; rewrite w_prod_env_set; reflexivity.
  Qed.

  Lemma w_prod_heap_set: forall w x v,
    w_prod (w_heap_set w x v) = w_prod w.
  Proof. unfold w_heap_set; intros; destruct w; destruct (MemMap.map_upd x v); reflexivity. Qed.
  Lemma w_cons_heap_set: forall w x v,
    w_cons (w_heap_set w x v) = w_cons w.
  Proof. unfold w_heap_set; intros; destruct w; destruct (MemMap.map_upd x v); reflexivity. Qed.

  Lemma can_send_w_heap_set: forall w c a v,
    can_send c w <-> can_send c (w_heap_set w a v).
  Proof.
    unfold can_send; intros; rewrite w_prod_heap_set; reflexivity.
  Qed.

  Lemma can_recv_w_heap_set: forall w c a v,
    can_recv c w <-> can_recv c (w_heap_set w a v).
  Proof.
    unfold can_recv; intros; rewrite w_cons_heap_set; reflexivity.
  Qed.

  Lemma noninterfering_step1: forall wC b1 b2 s s' b1',
    noninterfering_wterm b1 b2 ->
    step s b1 s' b1' ->
    cannot_communicate_channels s b1 ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1' b2.
  Proof.
    unfold noninterfering_wterm; introv Hni Hs Hcc1 Hvms.
    revert wC b2 Hni Hcc1 Hvms.
    step_cases (induction Hs) Case; intros; auto.
    - Case "st_send".
(* by Hcc1, w' cannot contain any channel permissions, so the effective
channel permissions of w1 will equal the permissions of w_w' *)
admit.
    - Case "st_recv".
(* by Hcc1, w' cannot contain any channel permissions, so the effective
channel permissions of w1 will equal the permissions of w_w' *)
admit.
    - Case "st_newch".
(* by syntactic restriction, R cannot accept worlds that contain permissions to a channel *)
admit.
    - Case "st_endch".
    simpl in H4; subst w.
    intro c'.
    destruct (Hni w_w' w2) with c' as [?[?[??]]]; auto; try reflexivity.
    destruct (Pos.eq_dec c' c).
    subst c'.
    splits; intros; auto.
    apply H4.
    eapply can_send_sub; eauto. exists w'; auto.
    intro.
    eapply has_total_endpoint_send_conflict; eauto. apply H0.
    exists w_w'; auto.
    apply H7.
    eapply can_recv_sub; eauto. exists w'; auto.
    intro.
    eapply has_total_endpoint_recv_conflict; eauto. apply H0.
    exists w_w'; auto.
(* need to tweak 'total_send_recv' so that it can contain NO MORE that endpoints for channel c *)
admit.
    - Case "st_assign".
    simpl in H2; subst w'; subst.
    intro c.
    edestruct (Hni w w2) with c as [?[?[??]]]; auto; try reflexivity.
    splits; intros; auto.
    apply H1. eapply can_send_w_env_set; eauto.
    intro; apply H1; auto. eapply can_send_w_env_set; eauto.
    apply H4. eapply can_recv_w_env_set; eauto.
    intro; apply H4; auto. eapply can_recv_w_env_set; eauto.
    - Case "st_load".
    simpl in H2; subst w'; subst.
    intro c.
    edestruct (Hni w w2) with c as [?[?[??]]]; auto; try reflexivity.
    splits; intros; auto.
    apply H1. eapply can_send_w_env_set; eauto.
    intro; apply H1; auto. eapply can_send_w_env_set; eauto.
    apply H4. eapply can_recv_w_env_set; eauto.
    intro; apply H4; auto. eapply can_recv_w_env_set; eauto.
    - Case "st_store".
    simpl in H3; subst w'; subst.
    intro c.
    edestruct (Hni w w2) with c as [?[?[??]]]; auto; try reflexivity.
    splits; intros; auto.

    apply H2. eapply can_send_w_heap_set; eauto.
    intro; apply H2; auto. eapply can_send_w_heap_set; eauto.
    apply H5. eapply can_recv_w_heap_set; eauto.
    intro; apply H5; auto. eapply can_recv_w_heap_set; eauto.
    - Case "st_local".
(* allocating a new local variable does not change the channel permissions *)
    admit.
    - Case "st_freelocal".
(* deallocating a local variable does not change the channel permissions *)
    admit.
    - Case "st_par".
    intro c.
    destruct H2 as [w1_[w2_[?[??]]]]; simpl in *; subst w1_ w2_.
    replace w0 with w in *.
    edestruct (Hni w w3) with c as [?[?[??]]]; auto; try reflexivity.
    eapply gjoin_eq; eauto.
    - Case "st_seq1".
    simpl in H.
    eapply IHHs; eauto.
    eapply cannot_communicate_channels_seq; eauto.
    - Case "st_par1".
    renames b1 to b1a, b1' to b1a', b2 to b1b, b0 to b2, w1 to w1'.
admit.
    - Case "st_par2".
admit.
    - Case "st_par_join".
    simpl in H0; subst w0.
    apply Hni; auto.
    exists w1 w2; splits; auto; reflexivity.
  Qed.

  Lemma noninterfering_wterm_symm: forall b1 b2,
    noninterfering_wterm b1 b2 -> noninterfering_wterm b2 b1.
  Proof.
    unfold noninterfering_wterm; intros; apply noninterfering_symm; auto.
  Qed.

  Lemma noninterfering_step2: forall wC b1 b2 s s' b2',
    noninterfering_wterm b1 b2 ->
    step s b2 s' b2' ->
    cannot_communicate_channels s b2 ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1 b2'.
  Proof.
    intros.
    apply noninterfering_wterm_symm; eapply noninterfering_step1; eauto.
    apply noninterfering_wterm_symm; auto.
    apply valid_machine_state_par_swap; eauto.
  Qed.

  Lemma noninterfering_step: forall wC b1 b2 s s' b1' b2',
    noninterfering_wterm b1 b2 ->
    step s (b1||b2) s' (b1'||b2') ->
    cannot_communicate_channels s b1 ->
    cannot_communicate_channels s b2 ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1' b2'.
  Proof.
    intros.
    inverts H0.
    eapply noninterfering_step1; eauto.
    eapply noninterfering_step2; eauto.
  Qed.

  Lemma noninterfering_cstep1: forall wC C b1 b2 s s' C' b1',
    noninterfering_wterm b1 b2 ->
    cstep s C b1 s' C' b1' ->
    cannot_communicate_channels s b1 ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1' b2.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    eapply noninterfering_step1; eauto.
    subst b1'; auto.
  Qed.

  Lemma noninterfering_cstep2: forall wC C b1 b2 s s' C' b2',
    noninterfering_wterm b1 b2 ->
    cstep s C b2 s' C' b2' ->
    cannot_communicate_channels s b2 ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1 b2'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    eapply noninterfering_step2; eauto.
    subst b2'; auto.
  Qed.

  Lemma noninterfering_cstep: forall wC C b1 b2 s s' C' b1' b2',
    noninterfering_wterm b1 b2 ->
    cstep s C (b1||b2) s' C' (b1'||b2') ->
    cannot_communicate_channels s b1 ->
    cannot_communicate_channels s b2 ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1' b2'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    eapply noninterfering_step; eauto.
    inverts H5; auto.
  Qed.

Import ParBisimulation.PermuteTactics.

  Lemma valid_machine_state_par1_sub: forall wC s b1 b2,
    valid_machine_state wC s (b1||b2) ->
    valid_machine_state wC s b1.
  Proof.
    introv [wS[wb[ww[?[??]]]]].
    destruct H0 as [w1[w2[?[??]]]].
    apply gjoin_join_list in H3.
    jrewrite<- H3 in H1.
    edestruct join_list_app_inv with (wl1:=[wS,w1,wC]) (wl2:=[w2]) (w:=ww)
      as [ww1[w2_[?[??]]]].
    eapply join_list_perm; eauto.
    permut3.
    exists wS w1 ww1; splits; auto.
  Qed.
  Lemma valid_machine_state_par2_sub: forall wC s b1 b2,
    valid_machine_state wC s (b1||b2) ->
    valid_machine_state wC s b2.
  Proof.
    intros.
    eapply valid_machine_state_par1_sub.
    apply valid_machine_state_par_swap; eauto.
  Qed.

  Lemma noninterfering_step_star1: forall wC b1 b2 s s' b1',
    noninterfering_wterm b1 b2 ->
    step_star s b1 s' b1' ->
    cannot_communicate_channels s b1 ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1' b2.
  Proof.
    intros.
    induction H0; auto.
    apply IHstep_star; auto.
    eapply noninterfering_step1; eauto.
    eapply cannot_communicate_channels_step; eauto.
    eapply valid_machine_state_par1_sub; eauto.
    eapply valid_machine_state_step; eauto. apply st_par1; auto.
    eapply IHstep_star; auto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma noninterfering_step_star2: forall wC b1 b2 s s' b2',
    noninterfering_wterm b1 b2 ->
    step_star s b2 s' b2' ->
    cannot_communicate_channels s b2 ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1 b2'.
  Proof.
    intros.
    apply noninterfering_wterm_symm.
    eapply noninterfering_step_star1; eauto.
    apply noninterfering_wterm_symm; auto.
    eapply valid_machine_state_par_swap; eauto.
  Qed.

  Lemma noninterfering_cstep_star1: forall wC C b1 b2 s s' C' b1',
    noninterfering_wterm b1 b2 ->
    cstep_star s C b1 s' C' b1' ->
    cannot_communicate_channels s b1 ->
    context_world C wC ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1' b2.
  Proof.
    intros.
    revert wC H H1 H2 H3.
    induction H0; intros; auto.
    edestruct valid_machine_state_cstep with (b:=(b||b2)%wterm) as [wC'[??]]; eauto.
    apply cstep_par1_step; eauto.
    eapply IHcstep_star; eauto.
    eapply noninterfering_cstep1; eauto.
    eapply cannot_communicate_channels_cstep; eauto.
    eapply valid_machine_state_par1_sub; eauto.
    eapply IHcstep_star; eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma noninterfering_cstep_star2: forall wC C b1 b2 s s' C' b2',
    noninterfering_wterm b1 b2 ->
    cstep_star s C b2 s' C' b2' ->
    cannot_communicate_channels s b2 ->
    context_world C wC ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1 b2'.
  Proof.
    intros.
    apply noninterfering_wterm_symm.
    eapply noninterfering_cstep_star1; eauto.
    apply noninterfering_wterm_symm; auto.
    eapply valid_machine_state_par_swap; eauto.
  Qed.

  Lemma cannot_communicate_channels_world_join: forall s w1 w2 w,
    cannot_communicate_channels_world s w1 ->
    cannot_communicate_channels_world s w2 ->
    gjoin w1 w2 w ->
    cannot_communicate_channels_world s w.
  Proof.
    unfold cannot_communicate_channels_world; intros.
    edestruct can_recv_super as [?|?]; eauto.
  Qed.

  Lemma cannot_communicate_channels_par: forall s b1 b2 w,
    cannot_communicate_channels s b1 ->
    cannot_communicate_channels s b2 ->
    join_wterms (b1||b2)%wterm w ->
    cannot_communicate_channels s (b1||b2).
  Proof.
    intros.
    destruct H as [w1[?[??]]].
    destruct H0 as [w2[?[??]]].
    destruct H1 as [w1_[w2_[?[??]]]].
    replace w1_ with w1 in *. replace w2_ with w2 in *.
    exists w; splits; auto.
    exists w1 w2; splits; auto.
    apply cannot_communicate_channels_world_join with w1 w2; eauto.
    split; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma noninterfering_cstep_star: forall wC C b1 b2 s s' C' b1' b2',
    noninterfering_wterm b1 b2 ->
    cstep_star s C (b1||b2) s' C' (b1'||b2') ->
    cannot_communicate_channels s b1 ->
    cannot_communicate_channels s b2 ->
    context_world C wC ->
    valid_machine_state wC s (b1||b2) ->
    noninterfering_wterm b1' b2'.
  Proof.
    introv Hni Hcs Hcc1 Hcc2 HwC Hvms.
    edestruct valid_machine_state_join_wterms as [w ?]; eauto.
    assert (Hcc: cannot_communicate_channels s (b1||b2)).
      eapply cannot_communicate_channels_par; eauto.
    clear Hcc1 Hcc2 H.
    remember (b1||b2)%wterm as b.
    remember (b1'||b2')%wterm as b'.
    revert wC b1 b2 b1' b2' Heqb Heqb' Hni HwC Hcc Hvms.
    induction Hcs; intros; subst; auto.
    - inverts Heqb'; auto.
    - edestruct valid_machine_state_cstep with (b:=(b1||b2)%wterm) as [wC'[??]]; eauto.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    inverts H2.
    eapply IHHcs; eauto.
    eapply noninterfering_step1; eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply cannot_communicate_channels_step; eauto. apply st_par1; auto.
    eapply IHHcs; eauto.
    eapply noninterfering_step2; eauto.
    eapply cannot_communicate_channels_par2; eauto.
    eapply cannot_communicate_channels_step; eauto. apply st_par2; auto.
    false; eapply cstep_star_skip_resolve in Hcs; discriminate.
    subst b'.
    eapply IHHcs; eauto.
    eapply cannot_communicate_channels_stepO with (bO:=C[gwemp:tm_skip]); eauto.
    apply step_simulates_cstep; eauto.
    apply context_world_join_wterms; auto.
    - eapply IHHcs; eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
  Qed.




(*
  Lemma noninterfering_schedule_step: forall s b1 b2 s' b2' s'' b1'',
    noninterfering_wterm b1 b2 ->
    b2 <> b2' ->
    step s b2 s' b2' ->
    step s' b1 s'' b1'' ->
    exists s'0,
         step s b1 s'0 b1''
    /\ step s'0 b2 s'' b2'.
  Proof.
    introv H Hnospin2 H0 H1.
    step_cases (induction H0; intros; subst) Case;
      try solve [exists s''; split; auto; constructor; auto].

    Case "st_while_max_limit".
      exists s''; split; auto; apply st_while_max_limit; auto.
    Case "st_send".
admit.
    Case "st_recv".
admit.
    Case "st_recv_blocked".
      contradiction Hnospin2; reflexivity.
    Case "st_newch".
admit.
    Case "st_endch".
admit.
    Case "st_seq1".
      destruct IHstep as [s'0[??]]; auto.
      intro; apply Hnospin2; subst; auto.
      exists s'0; split; auto; constructor; auto.
    Case "st_par1".
      destruct IHstep as [s'0[??]]; auto.
      eapply noninterfering_par1; eauto.
      intro; apply Hnospin2; subst; auto.
      exists s'0; split; auto; constructor; auto.
    Case "st_par2".
      destruct IHstep as [s'0[??]]; auto.
      eapply noninterfering_par2; eauto.
      intro; apply Hnospin2; subst; auto.
      exists s'0; split; auto; constructor; auto.
  Qed.
*)


  Lemma s_ch_fresh_channel_push: forall s c wc v w,
    c <> FreshChannels.fc_c (fresh_channel s w) ->
    fresh_channel (s_push_ch s c wc v) w = fresh_channel s w.
  Proof.
Admitted.

  Lemma s_ch_new_push: forall s c wc v w R,
    s_new_ch (s_push_ch s c wc v) (fresh_channel (s_push_ch s c wc v) w) R
    = s_push_ch (s_new_ch s (fresh_channel s w) R) c wc v.
Admitted.
  Lemma s_ch_get_new_diff: forall s w c R,
    c <> FreshChannels.fc_c (fresh_channel s w) ->
    s_get_ch (s_new_ch s (fresh_channel s w) R) c = s_get_ch s c.
Admitted.
  Lemma s_alloced_ch_new_diff: forall s w c,
    s_is_alloced_ch s c ->
    c <> FreshChannels.fc_c (fresh_channel s w).
Admitted.
  Lemma s_ch_get_new_cname_diff: forall s w c,
    s_get_ch s c <> None ->
    c <> FreshChannels.fc_c (fresh_channel s w).
  Proof.
    intros.
    destruct (fresh_channel_global_props s w (fresh_channel s w)) as [??]; auto.
    intro.
    subst.
    congruence.
  Qed.

  Lemma s_ch_rm_push: forall s c wc v c',
    s_rm_ch (s_push_ch s c wc v) c' = s_push_ch (s_rm_ch s c') c wc v.
  Proof.
    intros.
    unfold s_rm_ch,s_push_ch.
    destruct (Pos.eq_dec c c'). subst c'.
    do 2 rewrite s_ch_apply_compose; reflexivity.
    rewrite s_ch_apply_diff; auto.
  Qed.


  Lemma s_ch_get_rm_diff: forall s c1 c2,
    c1 <> c2 ->
    s_get_ch (s_rm_ch s c1) c2 = s_get_ch s c2.
  Proof. intros; apply s_ch_get_apply_diff; auto. Qed.

  Lemma step_star_sans_s_iter_seq_inversion: forall s b1 t2 s'' b'',
    step_star_sans_s_iter s (b1;; t2) s'' b'' ->
    (exists s' w',
     step_star_sans_s_iter s b1 s' (w': tm_skip) /\ step_star_sans_s_iter s' (w': t2) s'' b'') \/
    (exists b1'', step_star_sans_s_iter s b1 s'' b1'' /\ b'' = (b1'';; t2)%wterm).
  Proof.
    intros.
    remember (b1;;t2)%wterm as b.
    revert b1 t2 Heqb.
    induction H; intros; subst.
    right; exists b1; splits; auto.
    apply step_nil_.
    inverts H.
    edestruct IHstep_star_sans_s_iter as [ [s'0[w'0[??]]] | [b1'0[??]] ]; eauto.
    left; do 2 eexists; split; eauto.
    eapply step_cons_; eauto.
    subst.
    right; eexists; split; eauto.
    eapply step_cons_; eauto.
    left; do 2 eexists; split; eauto.
    apply step_nil_.
  Qed.


  Lemma gjoin_sub_w_prod_None: forall w1 w c,
    w_prod w c = None ->
    gjoin_sub w1 w ->
    w_prod w1 c = None.
  Proof.
    intros.
    destruct H0 as [w2 ?].
    destruct w1 as [w1_p w1_c w1_e w1_m].
    destruct w2 as [w2_p w2_c w2_e w2_m].
    destruct w as [w_p w_c w_e w_m].
    destruct H0 as [[? _]_].
    simpl in *.
    spec H0 c; inverts H0; congruence.
  Qed.

  Lemma gjoin_sub_w_prod_Some: forall w1 w c,
    w_prod w1 c <> None ->
    gjoin_sub w1 w ->
    w_prod w c <> None.
  Proof.
    intros.
    destruct H0 as [w2 ?].
    destruct w1 as [w1_p w1_c w1_e w1_m].
    destruct w2 as [w2_p w2_c w2_e w2_m].
    destruct w as [w_p w_c w_e w_m].
    destruct H0 as [[? _]_].
    simpl in *.
    spec H0 c; inverts H0; congruence.
  Qed.

  Lemma gjoin_sub_w_cons_None: forall w1 w c,
    w_cons w c = None ->
    gjoin_sub w1 w ->
    w_cons w1 c = None.
  Proof.
    intros.
    destruct H0 as [w2 ?].
    destruct w1 as [w1_p w1_c w1_e w1_m].
    destruct w2 as [w2_p w2_c w2_e w2_m].
    destruct w as [w_p w_c w_e w_m].
    destruct H0 as [[_ ?]_].
    simpl in *.
    spec H0 c; inverts H0; congruence.
  Qed.

  Lemma gjoin_sub_w_cons_Some: forall w1 w c,
    w_cons w1 c <> None ->
    gjoin_sub w1 w ->
    w_cons w c <> None.
  Proof.
    intros.
    destruct H0 as [w2 ?].
    destruct w1 as [w1_p w1_c w1_e w1_m].
    destruct w2 as [w2_p w2_c w2_e w2_m].
    destruct w as [w_p w_c w_e w_m].
    destruct H0 as [[_ ?]_].
    simpl in *.
    spec H0 c; inverts H0; congruence.
  Qed.


  Lemma no_send_conflict_sub: forall w1 w2 w1' w2',
    gjoin_sub w1 w1' ->
    gjoin_sub w2 w2' ->
    no_send_conflict w1' w2' ->
    no_send_conflict w1 w2.
  Proof.
    unfold no_send_conflict; intros.
    destruct (H1 c).
    eapply can_send_sub; eauto.
    split.
    intro; apply H3; eapply can_send_sub; eauto.
    intro; apply H4; eapply can_recv_sub; eauto.
  Qed.


  Lemma join_wterms_context: forall C b wCb,
    join_wterms (set_context C b) wCb ->
    exists wC wb,
      context_world C wC /\ join_wterms b wb /\ gjoin wC wb wCb.
  Proof.
    intro C; induction C; simpl; intros; auto.
    - exists gwemp wCb; splits; auto. apply gjoin_bot2.
    - rename w into b2.
    destruct H as [wCb1[wb2[?[??]]]].
    edestruct IHC as [wC[wb[?[??]]]]; eauto.
    edestruct gjoin_assoc with (a:=wC) (b:=wb) as [?[??]]; eauto.
    edestruct gjoin_assoc with (a:=wb) (b:=wb2) as [?[??]]; eauto.
    apply gjoin_com; eauto.
    exists x0 wb; splits; auto.
    exists wC wb2; splits; auto.
    apply gjoin_com; eauto.
    apply gjoin_com; eauto.
    - rename w into b1.
    destruct H as [wb1[wCb2[?[??]]]].
    edestruct IHC as [wC[wb[?[??]]]]; eauto.
    edestruct gjoin_assoc with (a:=wC) (b:=wb) as [?[??]]; eauto.
    apply gjoin_com; eauto.
    edestruct gjoin_assoc with (a:=wb) (b:=wb1) as [?[??]]; eauto.
    apply gjoin_com; eauto.
    exists x0 wb; splits; auto.
    exists wb1 wC; splits; auto.
    apply gjoin_com; eauto.
  Qed.

  Lemma join_wterms_context2: forall C wC wb b wCb,
    context_world C wC ->
    join_wterms b wb ->
    gjoin wC wb wCb ->
    join_wterms C[b] wCb.
  Proof.
    induction C; simpl; intros; subst; eauto.
    apply gworld_unit_identity in H1. subst wCb; auto.
    rename w into b'.
    destruct H as [wC1[wb'[?[??]]]].
    edestruct gjoin_assoc with (a:=wb') (b:=wC1) as [?[??]]; eauto.
    apply gjoin_com; auto.
    exists x wb'; splits; eauto. apply gjoin_com; auto.
    rename w into b'.
    destruct H as [wb'[wC1[?[??]]]].
    edestruct gjoin_assoc with (a:=wb') (b:=wC1) as [?[??]]; eauto.
    exists wb' x; splits; eauto.
  Qed.


  Lemma gjoins_sub1: forall w1 w1' w2,
    gjoins w1' w2 ->
    gjoin_sub w1 w1' ->
    gjoins w1 w2.
  Proof.
    introv [w1'2 ?] [w3 ?].
    apply gjoin_com in H0.
    edestruct gjoin_assoc with (a:=w3) (b:=w1) as [?[??]]; eauto.
    exists x; auto.
  Qed.

  Lemma gjoins_sub2: forall w1 w2 w2',
    gjoins w1 w2' ->
    gjoin_sub w2 w2' ->
    gjoins w1 w2.
  Proof.
    introv [w12' ?] [w3 ?].
    apply gjoin_com in H.
    apply gjoin_com in H0.
    edestruct gjoin_assoc with (a:=w3) (b:=w2) as [?[??]]; eauto.
    exists x; auto.
  Qed.


  Lemma terminate_context_tm_recv_inv: forall s wO w C w1 x ec c,
    terminates s wO (set_context C (w1: tm_recv x ec))%wterm ->
    cannot_communicate_channels s C[w1: tm_recv x ec] ->
    eval ec (cname_value c) w1 ->
    join_wterms (set_context C (w1: tm_recv x ec)) w ->
    ~can_send c w ->
    exists R wc wl v vl,
      s_get_ch s c = Some (R, wc::wl, v::vl).
  Proof.
(* Justification:
-- we know that C[w1:...] cannot send on c and does not have the power to
 eventually receive sending permission on c. Thus in order for it to
 terminate, it must be the case that the channel queue is not empty.
*)
Admitted.
(*
  Lemma terminate_tm_recv_inv: forall s w x ec wC,
    terminates s wC (w: tm_recv x ec) ->
    exists c R wc wl v vl,
      eval ec (cname_value c) w /\ s_get_ch s c = Some(R,wc::wl,v::vl).
  Proof.
    intros.
    edestruct terminates_now as [s'[w' ?]]; eauto.
    clear -H0.
    remember (w: tm_recv x ec)%wterm as b.
    remember (w':tm_skip)%wterm as b'.
    revert Heqb Heqb'.
    induction H0; intros; subst.
    discriminate.
    inverts H; eauto.
    exists c R w'0 wl v q; split; auto.
  Qed.
*)

  Lemma empty_channel_queue': forall s c w x R wl vl,
    s_get_ch s c = Some (R, wl, vl) ->
    can_recv c w ->
    w_env_share w x = Share.top ->
    exists t s' w' wl' vl',
      step_star s (w: t) s' (w': tm_skip) /\
      s_get_ch s' c = Some (R, wl', vl') /\
      (wl'=nil\/vl'=nil).
  Proof.
    intros.
    revert s w vl H H0 H1.
    induction wl; intros.
    exists tm_skip s w (@nil world) vl; split; eauto.
    apply step_nil.
    destruct vl.
    exists tm_skip s w (a::wl) (@nil value); split; eauto.
    apply step_nil.
    assert (gjoins w a).
admit.
    destruct H2 as [w_w' ?].
    destruct (IHwl (s_pop_ch s c) (w_cons_incr c (w_env_set w_w' x v)) vl)
      as [t[s''[w''[wl''[vl''[?[??]]]]]]]; auto.
    eapply s_ch_get_pop_same; eauto.
    apply-> can_recv_w_cons_incr.
    apply-> can_recv_upd_var.
    eapply can_recv_sub; eauto.
    exists a; auto.
    unfold w_cons_incr.
    rewrite w_env_share_wupd_cons.
    rewrite w_env_share_set.
    eapply w_env_share_sub_top; eauto.
    exists (tm_recv x (expr_value (cname_value c));;; t)%term.
    do 4 eexists; splits.
    eapply step_cons.
    apply st_seq.
    eapply step_cons.
    apply st_seq1. eapply st_recv; eauto.
    reflexivity.
admit.
    eapply step_cons; eauto.
    apply st_seq_skip.
    eauto.
    auto.
  Qed.


  Lemma empty_channel_queue: forall s c w R wl vl,
    s_get_ch s c = Some (R, wl, vl) ->
    can_recv c w ->
    exists t s' w' wl' vl',
      step_star s (w: t) s' (w': tm_skip) /\
      s_get_ch s' c = Some (R, wl', vl') /\
      (wl'=nil\/vl'=nil).
  Proof.
    intros.
    remember (fresh_var w (nat_value 0)) as fresh_x.
    destruct (fresh_var_local_props w (nat_value 0) fresh_x) as [??]; auto.
    edestruct empty_channel_queue' with (w:=FreshLocals.fl_w fresh_x) (x:=FreshLocals.fl_x fresh_x) as [t[s'[w'[wl'[vl'[?[??]]]]]]]; eauto.
    eapply can_recv_sub; eauto. eexists; eauto.
    eapply w_env_share_sub_top.
    2: apply gjoin_com; eauto.
    apply w_env_share_new_local_same.
    exists (tm_local (expr_value (nat_value 0)) (fun x => t)).
    exists s' w' wl' vl'; splits; auto.
    eapply step_cons.
    eapply st_local; eauto.
    reflexivity.
    auto.
  Qed.

  Lemma cannot_communicate_channels_step_starO: forall wO s bO b s' bO',
    cannot_communicate_channels s b ->
    step_star s bO s' bO' ->
    join_wterms bO wO ->
    valid_machine_state wO s b ->
    cannot_communicate_channels s' b.
  Proof.
    introv Hcc Hss HwO Hvms.
    revert wO HwO Hcc Hvms.
    induction Hss; intros; auto.
    - edestruct valid_machine_state_stepO as [wO'[??]]; eauto.
    eapply IHHss; eauto.
    eapply cannot_communicate_channels_stepO; eauto.
    - eapply IHHss; eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma terminates_can_recv_conflict: forall s wO C w w1 x ec c,
    terminates s wO C[w1: tm_recv x ec] ->
    join_wterms C[w1: tm_recv x ec] w ->
    cannot_communicate_channels s C[w1: tm_recv x ec] ->
    valid_machine_state wO s C[w1: tm_recv x ec] ->
    eval ec (cname_value c) w1 ->
    can_recv c wO ->
    ~can_send c w ->
    False.
  Proof.
    intros.
    case_eq (s_get_ch s c); intros.
    destruct c0 as [[R wl]vl].
    edestruct empty_channel_queue as [t[s'[w'[wl'[vl'[?[??]]]]]]]; eauto.
    edestruct terminate_context_tm_recv_inv with (s:=s') as [R_[wc_[wl_[v_[vl_ ?]]]]]; eauto.
    eapply terminates_step_starO_skip; eauto. reflexivity.
    eapply cannot_communicate_channels_step_starO; eauto.
    reflexivity.
    rewrite H8 in H10.
    inverts H10.
    destruct H9; discriminate.
    edestruct terminate_context_tm_recv_inv as [R_[wc_[wl_[v_[vl_ ?]]]]]; eauto.
    rewrite H6 in H7; discriminate.
  Qed.



  Lemma no_send_conflict_step_s_push_ch: forall wO s w b s' b' c wc v,
    step (s_push_ch s c wc v) b s' b' ->
    can_send c wO ->
    join_wterms b w ->
    ~can_send c w ->
(* and assuming and b can never receive send permissions for channel c in the future *)
    terminates s wO b ->
    cannot_communicate_channels s b ->
    s_get_ch s c <> None ->
    exists s'0,
      s' = s_push_ch s'0 c wc v /\
      (s_get_ch s'0 c = s_get_ch s c
      \/  (exists R wc1 wl1 wl2 wc2 v1 vl1 vl2 v2,
          wc1::wl1 = wl2 ++ [wc2] /\
          v1::vl1 = vl2 ++ [v2] /\
          s_get_ch s c = Some (R, wl2++[wc2], vl2++[v2]) /\
          s_get_ch s'0 c = Some (R, wl1, vl1))) /\
      step s b s'0 b'.
  Proof.
    introv H Hcansend Hw Hnosend Hterm Hcc Hallc.
    remember (s_push_ch _ _ _ _) as s0.
    assert (Hrecv_inv: forall x ec w' wC C,
      gjoin_sub w' w ->
      join_wterms C[w':tm_recv x ec] wC ->
      b=C[w':tm_recv x ec] ->
      eval ec (cname_value c) w' ->
      ~can_send c wC ->
      exists R wc wl v vl,
        s_get_ch s c = Some (R, wc::wl, v::vl)).
      intros. subst b.
      eapply terminate_context_tm_recv_inv; eauto.
    clear Hterm Hcc.
    revert s w Heqs0 Hcansend Hw Hnosend Hrecv_inv Hallc.
    step_cases (induction H) Case constructor; intros;
      first [Case "st_newch" | try solve
        [ eauto
        | eexists; splits; eauto]].
    - Case "st_while_go".
    subst; eexists; splits; eauto.
    eapply st_while_go; eauto.
    rewrite s_iter_get_push_ch in H; auto.
    - Case "st_send".
    destruct (Pos.eq_dec c c0).
    + false; subst c0; subst.
    apply Hnosend; auto.
    simpl in Hw; subst w0; subst. auto.
    + subst.
    rewrite s_ch_get_push_diff in H1; auto.
    eexists (s_push_ch s c0 w' v0); splits; eauto.
    apply s_ch_push_push_diff; auto.
    rewrite s_ch_get_push_diff; auto.
    - Case "st_recv".
    subst.
    destruct (Pos.eq_dec c c0).
    + subst c0. simpl in Hw; subst w0.
    edestruct Hrecv_inv with (C:=cx_hole) as [R_[wc'[wl'[v'[vl' ?]]]]]; clear Hrecv_inv; eauto.
    apply join_sub_refl.
    reflexivity.
    reflexivity.
    edestruct s_ch_get_push_same' as [wl''[vl''[?[??]]]]; eauto.
    rewrite H4 in H6.
    inverts H6.
    inverts H7; inverts H8.
    destruct (@exists_last world (wc'::wl')) as [wl''[wc'' ?]]. congruence.
    destruct (@exists_last value (v'::vl')) as [vl''[v'' ?]]. congruence.
    exists (s_pop_ch s c); splits; auto.
    symmetry; eapply s_ch_push_pop_same_comm; eauto.
    right.
    rewrite H4.
    erewrite s_ch_get_pop_same; eauto.
    do 9 eexists; splits; eauto. congruence.
    eapply st_recv; eauto.
    + exists (s_pop_ch s c0); splits.
    symmetry; apply s_ch_push_pop_diff_comm; auto.
    left; apply s_ch_get_pop_diff; auto.
    rewrite s_ch_get_push_diff in H5; auto.
    eapply constructor; eauto.

    - Case "st_recv_blocked".
    subst.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false.
    edestruct s_ch_get_push_same' as [wl''[vl''[?[??]]]]; eauto.
    eapply app_cons_not_nil; eauto.
    + rewrite s_ch_get_push_diff in H2; auto.
    eexists; splits; eauto.

    - Case "st_newch".
    subst.
    eexists; splits.
    apply s_ch_new_push.
    left; apply s_ch_get_new_diff.
    apply s_ch_get_new_cname_diff; auto.
    eapply constructor.
    reflexivity.
    rewrite s_ch_fresh_channel_push; auto.
    apply s_ch_get_new_cname_diff; auto.
    reflexivity.

    - Case "st_endch".
    subst.
    destruct (Pos.eq_dec c0 c).
    + false; subst c0.
    edestruct s_ch_get_push_same' as [wl[vl[?[??]]]]; eauto.
    eapply app_cons_not_nil; eauto.
    + eexists; splits.
    apply s_ch_rm_push.
    left; apply s_ch_get_rm_diff; auto.
    rewrite s_ch_get_push_diff in H2; auto.
    eapply constructor; eauto.

    - Case "st_seq1".
    edestruct IHstep  as [s'0[?[??]]]; clear IHstep; eauto.
    intros.
    subst.
    eapply Hrecv_inv with (C0:=cx_seq C t2); eauto.

    - Case "st_par1".
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    intro; apply Hnosend; eapply can_send_sub; eauto. exists w2; auto.
    intros.
    subst.
    replace w1 with wC in *.
    eapply Hrecv_inv with (C0:=cx_par1 C b2) (wC:=w); eauto.
    eapply join_sub_trans; eauto.
    simpl.
    exists wC w2; splits; eauto.
    reflexivity.
    eapply join_wterms_eq; eauto.

    - Case "st_par2".
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    intro; apply Hnosend; eapply can_send_sub; eauto. exists w1; auto.
    intros.
    subst.
    replace w2 with wC in *.
    eapply Hrecv_inv with (C0:=cx_par2 b1 C) (wC:=w); eauto.
    eapply join_sub_trans; eauto.
    simpl.
    exists w1 wC; splits; eauto.
    reflexivity.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma no_send_conflict_step_s_push_ch': forall s w b s' b' c wc v,
    step (s_push_ch s c wc v) b s' b' ->
    join_wterms b w ->
    ~can_send c w ->
    ~can_recv c w ->
    s_get_ch s c <> None ->
    exists s'0,
      s' = s_push_ch s'0 c wc v /\
      s_get_ch s'0 c = s_get_ch s c /\
      step s b s'0 b'.
  Proof.
    introv H Hw Hnosend Hnorecv Hallc.
    remember (s_push_ch _ _ _ _) as s0.
    revert s w Heqs0 Hw Hnosend Hnorecv Hallc.
    step_cases (induction H) Case constructor; intros;
      first [Case "st_newch" | try solve
        [ eauto
        | eexists; splits; eauto]].
    - Case "st_while_go".
    subst; eexists; splits; eauto.
    eapply st_while_go; eauto.
    rewrite s_iter_get_push_ch in H; auto.
    - Case "st_send".
    destruct (Pos.eq_dec c c0).
    + false; subst c0; subst.
    simpl in Hw; subst w_w'.
    apply Hnosend; auto.
    + subst.
    rewrite s_ch_get_push_diff in H1; auto.
    eexists (s_push_ch s c0 w' v0); splits; eauto.
    apply s_ch_push_push_diff; auto.
    rewrite s_ch_get_push_diff; auto.
    - Case "st_recv".
    subst.
    destruct (Pos.eq_dec c c0).
    + false; subst c0. simpl in Hw; subst w0.
    apply Hnorecv; auto.
    + exists (s_pop_ch s c0); splits.
    symmetry; apply s_ch_push_pop_diff_comm; auto.
    apply s_ch_get_pop_diff; auto.
    rewrite s_ch_get_push_diff in H5; auto.
    eapply constructor; eauto.

    - Case "st_recv_blocked".
    subst.
    destruct (Pos.eq_dec c0 c).
    + false; simpl in Hw; subst c0 w0.
    apply Hnorecv; auto.
    + rewrite s_ch_get_push_diff in H2; auto.
    eexists; splits; eauto.

    - Case "st_newch".
    subst.
    eexists; splits.
    apply s_ch_new_push.
    apply s_ch_get_new_diff.
    apply s_ch_get_new_cname_diff; auto.
    eapply constructor.
    reflexivity.
    rewrite s_ch_fresh_channel_push; auto.
    apply s_ch_get_new_cname_diff; auto.
    reflexivity.

    - Case "st_endch".
    subst.
    destruct (Pos.eq_dec c0 c).
    + false; simpl in Hw; subst c0 w0.
    apply Hnosend.
    eapply total_send_recv_can_send_recv; eauto.
    eapply total_send_recv_sub; eauto. exists w; auto.
    + eexists; splits.
    apply s_ch_rm_push.
    apply s_ch_get_rm_diff; auto.
    rewrite s_ch_get_push_diff in H2; auto.
    eapply constructor; eauto.

    - Case "st_seq1".
    edestruct IHstep  as [s'0[?[??]]]; clear IHstep; eauto.

    - Case "st_par1".
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    intro; apply Hnosend; eapply can_send_sub; eauto. exists w2; auto.
    intro; apply Hnorecv; eapply can_recv_sub; eauto. exists w2; auto.

    - Case "st_par2".
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    intro; apply Hnosend; eapply can_send_sub; eauto. exists w1; auto.
    intro; apply Hnorecv; eapply can_recv_sub; eauto. exists w1; auto.
  Qed.

  Lemma s_ch_new_pop: forall s c w R,
    s_new_ch (s_pop_ch s c) (fresh_channel (s_pop_ch s c) w) R =
    s_pop_ch (s_new_ch s (fresh_channel s w) R) c.
Admitted.

  Lemma s_ch_fresh_channel_pop: forall s c w,
    c <> FreshChannels.fc_c (fresh_channel s w) ->
    fresh_channel (s_pop_ch s c) w = fresh_channel s w.
  Proof.
Admitted.

  Lemma no_send_conflict_step_s_pop_ch': forall R s w b s' b' c wc wl v vl,
    step (s_pop_ch s c) b s' b' ->
    ~can_send c w -> ~can_recv c w ->
    join_wterms b w ->
    s_get_ch s c = Some (R, wc::wl, v::vl) ->
    exists s'0,
      s' = s_pop_ch s'0 c /\
      s_get_ch s'0 c = s_get_ch s c /\
      step s b s'0 b'.
  Proof.
    introv H Hnosend Hnorecv Hw Hgetch.
    remember (s_pop_ch _ _) as s0.

    revert s w Heqs0 Hnosend Hnorecv Hw Hgetch.
    step_cases (induction H) Case constructor; intros;
      first [Case "st_newch" | try solve
        [ eauto
        | eexists; splits; try left; eauto]].
    - Case "st_while_go".
    subst; eexists; splits; eauto.
    eapply st_while_go; eauto.
    rewrite s_iter_get_pop_ch in H; auto.
    - Case "st_send". clear constructor.
    subst; destruct (Pos.eq_dec c c0).
    + simpl in Hw; subst w0 c0.
    false; apply Hnosend; auto.
    + exists (s_push_ch s c0 w' v0); splits; auto.
    apply s_ch_push_pop_diff_comm; auto.
    rewrite s_ch_get_push_diff; auto; reflexivity.
    rewrite s_ch_get_pop_diff in H1; auto.
    eapply st_send; eauto.
    - Case "st_recv".
    subst. clear constructor.
    destruct (Pos.eq_dec c c0).
    + simpl in Hw; subst c0 w0.
    false; apply Hnorecv; auto.
    + rewrite s_ch_get_pop_diff in H5; auto.
    rewrite s_ch_pop_pop_diff; auto.
    eexists; splits; eauto.
    apply s_ch_get_pop_diff; auto.
    eapply st_recv; eauto.

    - Case "st_recv_blocked".
    subst. clear constructor.
    destruct (Pos.eq_dec c0 c).
    + simpl in Hw; subst c0 w0.
    false; apply Hnorecv; auto.
    + rewrite s_ch_get_pop_diff in H2; auto.
    eexists; splits; eauto.
    eapply st_recv_blocked; eauto.

    - Case "st_newch".
    subst.
    eexists; splits.
    apply s_ch_new_pop.
    apply s_ch_get_new_diff.
    apply s_ch_get_new_cname_diff; auto.
    congruence.
    eapply st_newch.
    reflexivity.
    rewrite s_ch_fresh_channel_pop; auto.
    apply s_ch_get_new_cname_diff; auto.
    congruence.
    reflexivity.

    - Case "st_endch".
    subst.
    simpl in Hw; subst w0.
    destruct (Pos.eq_dec c0 c).
    + false; subst c0.
    apply Hnorecv.
    apply total_send_recv_can_send_recv.
    eapply total_send_recv_sub; eauto. exists w; auto.
    + rewrite s_ch_get_pop_diff in H2; auto.
    exists (s_rm_ch s c0); splits; eauto.
    symmetry; apply s_ch_pop_rm_diff_comm; auto.
    apply s_ch_get_rm_diff; auto.

    - Case "st_seq1".
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.

    - Case "st_par1".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep with (w:=w1) as [s'0[?[??]]]; clear IHstep; jauto.
    intro; apply Hnosend. eapply can_send_sub; eauto. exists w2; auto.
    intro; apply Hnorecv. eapply can_recv_sub; eauto. exists w2; auto.

    - Case "st_par2".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep with (w:=w2) as [s'0[?[??]]]; clear IHstep; jauto.
    intro; apply Hnosend. eapply can_send_sub; eauto. exists w1; auto.
    intro; apply Hnorecv. eapply can_recv_sub; eauto. exists w1; auto.
  Qed.


  Lemma no_send_conflict_step_s_pop_ch: forall R wO s w b s' b' c wc wl v vl,
    step (s_pop_ch s c) b s' b' ->
    can_recv c wO ->
    (~can_recv c w \/ ~can_send c w) ->
    join_wterms b w ->
    gjoins wO w ->
    terminates s wO b ->
    cannot_communicate_channels s b ->
    valid_machine_state wO s b ->
    s_get_ch s c = Some (R, wc::wl, v::vl) ->
    exists s'0,
      s' = s_pop_ch s'0 c /\
      (s_get_ch s'0 c = s_get_ch s c \/
      (can_send c w /\ exists wc2 v2,
        s_get_ch s'0 c = s_get_ch (s_push_ch s c wc2 v2) c)) /\
      step s b s'0 b'.
  Proof.
    introv H Hcanrecv Hnosendrecv Hw HwOw Hterm Hcc Hvms Hgetch.
    remember (s_pop_ch _ _) as s0.

    assert (Hterm': forall C x ec w',
        gjoin_sub w' w ->
        (b=C[w':tm_recv x ec])%wterm ->
        eval ec (cname_value c) w' ->
        can_recv c wO ->
        can_recv c w ->
        False).
      intros. destruct H0. subst.
      eapply terminates_can_recv_conflict; eauto.
      destruct Hnosendrecv; auto.
    clear Hterm. rename Hterm' into Hterm.
    revert s w Heqs0 Hcanrecv Hnosendrecv Hw HwOw Hterm Hcc Hvms Hgetch.
    step_cases (induction H) Case constructor; intros;
      first [Case "st_newch" | try solve
        [ eauto
        | eexists; splits; try left; eauto]].
    - Case "st_while_go".
    subst; eexists; splits; eauto.
    eapply st_while_go; eauto.
    rewrite s_iter_get_pop_ch in H; auto.
    - Case "st_send". clear constructor.
    subst; destruct (Pos.eq_dec c c0).
    + subst c0.
    simpl in Hw; subst w0.
    destruct Hnosendrecv. Focus 2. false. apply H4; auto.
    exists (s_push_ch s c w' v0); splits; auto.
    eapply s_ch_push_pop_same_comm; eauto.
    right; split; auto; exists w' v0; reflexivity.
    replace R0 with R in *.
    eapply st_send; eauto.
    erewrite s_ch_get_pop_same in H1; eauto. inverts H1; auto.
    + exists (s_push_ch s c0 w' v0); splits; auto.
    apply s_ch_push_pop_diff_comm; auto.
    rewrite s_ch_get_push_diff; auto; reflexivity.
    rewrite s_ch_get_pop_diff in H1; auto.
    eapply st_send; eauto.
    - Case "st_recv".
    subst. clear constructor.
    destruct (Pos.eq_dec c c0).
    + subst c0. simpl in Hw; subst w0.
    false; eapply Hterm with (C:=cx_hole); eauto.
    apply join_sub_refl. reflexivity.
    + rewrite s_ch_get_pop_diff in H5; auto.
    rewrite s_ch_pop_pop_diff; auto.
    eexists; splits; eauto.
    left.
    apply s_ch_get_pop_diff; auto.
    eapply st_recv; eauto.

    - Case "st_recv_blocked".
    subst. clear constructor.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false.
    simpl in Hw; subst w0; auto.
    eapply Hterm with (C:=cx_hole); eauto.
    apply join_sub_refl. reflexivity.
    + rewrite s_ch_get_pop_diff in H2; auto.
    eexists; splits; eauto.
    eapply st_recv_blocked; eauto.

    - Case "st_newch".
    subst.
    eexists; splits.
    apply s_ch_new_pop.
    left.
    apply s_ch_get_new_diff.
    apply s_ch_get_new_cname_diff; auto.
    congruence.
    eapply st_newch.
    reflexivity.
    rewrite s_ch_fresh_channel_pop; auto.
    apply s_ch_get_new_cname_diff; auto.
    congruence.
    reflexivity.

    - Case "st_endch".
    subst.
    simpl in Hw; subst w0.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false. clear Hterm constructor.
    eapply has_total_endpoint_recv_conflict with (w1:=w_w'); eauto.
    eapply total_send_recv_sub; eauto. exists w; auto.
    destruct HwOw as [wO' ?]. exists wO'; auto.
    + rewrite s_ch_get_pop_diff in H2; auto.
    exists (s_rm_ch s c0); splits; eauto.
    symmetry; apply s_ch_pop_rm_diff_comm; auto.
    left.
    apply s_ch_get_rm_diff; auto.

    - Case "st_seq1".
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    intros; subst.
    eapply Hterm with (C0:= cx_seq C t2); eauto.
    reflexivity.
    eapply cannot_communicate_channels_seq; eauto.

    - Case "st_par1".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep with (w:=w1) as [s'0[?[[?|[??]]?]]]; clear IHstep; jauto.
    destruct Hnosendrecv; [left|right].
    intro; apply H3. eapply can_recv_sub; eauto. exists w2; auto.
    intro; apply H3. eapply can_send_sub; eauto. exists w2; auto.
    eapply gjoins_sub2; eauto. exists w2; auto.
    intros; subst.
    eapply Hterm with (C0:= cx_par1 C b2); eauto.
    eapply join_sub_trans; eauto.
    reflexivity.
    subst.
    eapply can_recv_sub; eauto. exists w2; auto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply valid_machine_state_par1_sub; eauto.
    exists s'0; splits; auto.
    right; split; eauto.
    eapply can_send_sub; eauto.
    exists w2; auto.


    - Case "st_par2".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep with (w:=w2) as [s'0[?[[?|[??]]?]]]; clear IHstep; jauto.
    destruct Hnosendrecv; [left|right].
    intro; apply H3. eapply can_recv_sub; eauto. exists w1; auto.
    intro; apply H3. eapply can_send_sub; eauto. exists w1; auto.
    eapply gjoins_sub2; eauto. exists w1; auto.
    intros; subst.
    eapply Hterm with (C0:= cx_par2 b1 C); eauto.
    eapply join_sub_trans; eauto.
    reflexivity.
    subst.
    eapply can_recv_sub; eauto. exists w1; auto.
    eapply cannot_communicate_channels_par2; eauto.
    eapply valid_machine_state_par2_sub; eauto.
    exists s'0; splits; auto.
    right; split; eauto.
    eapply can_send_sub; eauto.
    exists w1; auto.
  Qed.


(*
  Lemma no_send_conflict_step_s_pop_ch: forall R wO s w b s' b' c wc wl v vl,
    step (s_pop_ch s c) b s' b' ->
    can_recv c wO ->
    join_wterms b w ->
    gjoins wO w ->
    terminates s wO b ->
    cannot_communicate_channels s b ->
    valid_machine_state wO s b ->
    s_get_ch s c = Some (R, wc::wl, v::vl) ->
    exists s'0,
      s' = s_pop_ch s'0 c /\
      (s_get_ch s'0 c = s_get_ch s c \/
      (can_send c w /\ exists wc2 v2,
        s_get_ch s'0 c = s_get_ch (s_push_ch s c wc2 v2) c)) /\
      step s b s'0 b'.
  Proof.
    introv H Hcanrecv Hw HwOw Hterm Hcc Hvms Hgetch.
    remember (s_pop_ch _ _) as s0.

    assert (Hterm': forall C x ec w',
        gjoin_sub w' w ->
        (b=C[w':tm_recv x ec])%wterm ->
        eval ec (cname_value c) w' ->
        can_recv c wO ->
        False).
      intros. destruct H0. subst.
      eapply terminates_can_recv_conflict; eauto.

    clear Hterm. rename Hterm' into Hterm.
    revert s w Heqs0 Hcanrecv Hw HwOw Hterm Hgetch.
    step_cases (induction H) Case constructor; intros;
      first [Case "st_newch" | try solve
        [ eauto
        | eexists; splits; try left; eauto]].
    - Case "st_while_go".
    subst; eexists; splits; eauto.
    eapply st_while_go; eauto.
    rewrite s_iter_get_pop_ch in H; auto.
    - Case "st_send". clear constructor.
    subst; destruct (Pos.eq_dec c c0).
    + subst c0.
    simpl in Hw; subst w0.
    exists (s_push_ch s c w' v0); splits; auto.
    eapply s_ch_push_pop_same_comm; eauto.
    right; split; auto; exists w' v0; reflexivity.
    replace R0 with R in *.
    eapply st_send; eauto.
    erewrite s_ch_get_pop_same in H1; eauto. inverts H1; auto.
    + exists (s_push_ch s c0 w' v0); splits; auto.
    apply s_ch_push_pop_diff_comm; auto.
    rewrite s_ch_get_push_diff; auto; reflexivity.
    rewrite s_ch_get_pop_diff in H1; auto.
    eapply st_send; eauto.
    - Case "st_recv".
    subst. clear constructor.
    destruct (Pos.eq_dec c c0).
    + subst c0. simpl in Hw; subst w0.
    false; eapply Hterm with (C:=cx_hole); eauto.
    apply join_sub_refl. reflexivity.
    + rewrite s_ch_get_pop_diff in H5; auto.
    rewrite s_ch_pop_pop_diff; auto.
    eexists; splits; eauto.
    left.
    apply s_ch_get_pop_diff; auto.
    eapply st_recv; eauto.

    - Case "st_recv_blocked".
    subst. clear constructor.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false.
    simpl in Hw; subst w0; auto.
    eapply Hterm with (C:=cx_hole); eauto.
    apply join_sub_refl. reflexivity.
    + rewrite s_ch_get_pop_diff in H2; auto.
    eexists; splits; eauto.
    eapply st_recv_blocked; eauto.

    - Case "st_newch".
    subst.
    eexists; splits.
    apply s_ch_new_pop.
    left.
    apply s_ch_get_new_diff.
    apply s_ch_get_new_cname_diff; auto.
    congruence.
    eapply st_newch.
    reflexivity.
    rewrite s_ch_fresh_channel_pop; auto.
    apply s_ch_get_new_cname_diff; auto.
    congruence.
    reflexivity.

    - Case "st_endch".
    subst.
    simpl in Hw; subst w0.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false. clear Hterm constructor.
    eapply has_total_endpoint_recv_conflict with (w1:=w_w'); eauto.
    eapply total_send_recv_sub; eauto. exists w; auto.
    destruct HwOw as [wO' ?]. exists wO'; auto.
    + rewrite s_ch_get_pop_diff in H2; auto.
    exists (s_rm_ch s c0); splits; eauto.
    symmetry; apply s_ch_pop_rm_diff_comm; auto.
    left.
    apply s_ch_get_rm_diff; auto.

    - Case "st_seq1".
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    intros; subst.
    eapply Hterm with (C0:= cx_seq C t2); eauto.
    reflexivity.

    - Case "st_par1".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [s'0[?[[?|[??]]?]]]; clear IHstep; jauto.
    eapply gjoins_sub2; eauto. exists w2; auto.
    intros; subst.
    eapply Hterm with (C0:= cx_par1 C b2); eauto.
    eapply join_sub_trans; eauto.
    reflexivity.
    subst.
    exists s'0; splits; auto.
    right; split; eauto.
    eapply can_send_sub; eauto.
    exists w2; auto.


    - Case "st_par2".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [s'0[?[[?|[??]]?]]]; clear IHstep; jauto.
    eapply gjoins_sub2; eauto. exists w1; auto.
    intros; subst.
    eapply Hterm with (C0:= cx_par2 b1 C); eauto.
    eapply join_sub_trans; eauto.
    reflexivity.
    subst.
    exists s'0; splits; auto.
    right; split; eauto.
    eapply can_send_sub; eauto.
    exists w1; auto.
  Qed.
*)


(*
  Lemma no_send_conflict_step_s_pop_ch: forall R wO s w b s' b' c wc wl v vl,
    step (s_pop_ch s c) b s' b' ->
    can_recv c wO ->
    join_wterms b w ->
    ~can_send c w ->
    terminates s wO b ->
    s_get_ch s c = Some (R, wc::wl, v::vl) ->
    exists s'0,
      s' = s_pop_ch s'0 c /\
      s_get_ch s'0 c = s_get_ch s c /\
      step s b s'0 b'.
  Proof.
    introv H Hcanrecv Hw Hnosend Hterm Hgetch.
    remember (s_pop_ch _ _) as s0.

    assert (Hterm': forall C x ec w',
        gjoin_sub w' w ->
        (b=C[w':tm_recv x ec])%wterm ->
        eval ec (cname_value c) w' ->
        can_recv c wO ->
        False).
      intros. destruct H0. subst.
      eapply terminates_can_recv_conflict; eauto.
admit.
    clear Hterm. rename Hterm' into Hterm.
    revert s w Heqs0 Hcanrecv Hw Hnosend Hterm Hgetch.
    step_cases (induction H) Case constructor; intros;
      first [Case "st_newch" | try solve
        [ eauto
        | eexists; splits; eauto]].
    - Case "st_while_go".
    subst; eexists; splits; eauto.
    eapply st_while_go; eauto.
    rewrite s_iter_get_pop_ch in H; auto.
    - Case "st_send". clear constructor.
    subst; destruct (Pos.eq_dec c c0).
    + subst c0.
    false. apply Hnosend.
    simpl in Hw; subst w0; auto.
    + exists (s_push_ch s c0 w' v0); splits; auto.
    apply s_ch_push_pop_diff_comm; auto.
    rewrite s_ch_get_push_diff; auto; reflexivity.
    rewrite s_ch_get_pop_diff in H1; auto.
    eapply st_send; eauto.
    - Case "st_recv".
    subst. clear constructor.
    destruct (Pos.eq_dec c c0).
    + subst c0. simpl in Hw; subst w0.
    false; eapply Hterm with (C:=cx_hole); eauto.
    apply join_sub_refl. reflexivity.
    + rewrite s_ch_get_pop_diff in H5; auto.
    rewrite s_ch_pop_pop_diff; auto.
    eexists; splits; eauto.
    apply s_ch_get_pop_diff; auto.
    eapply st_recv; eauto.

    - Case "st_recv_blocked".
    subst. clear constructor.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false.
    simpl in Hw; subst w0; auto.
    eapply Hterm with (C:=cx_hole); eauto.
    apply join_sub_refl. reflexivity.
    + rewrite s_ch_get_pop_diff in H2; auto.
    eexists; splits; eauto.
    eapply st_recv_blocked; eauto.

    - Case "st_newch".
    subst.
    eexists; splits.
    apply s_ch_new_pop.
    apply s_ch_get_new_diff.
    apply s_ch_get_new_cname_diff; auto.
    congruence.
    eapply st_newch.
    reflexivity.
    rewrite s_ch_fresh_channel_pop; auto.
    apply s_ch_get_new_cname_diff; auto.
    congruence.
    reflexivity.

    - Case "st_endch".
    subst.
    simpl in Hw; subst w0.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false.
    destruct H0 as [? _].
    unfold has_total_endpoint in H0.
    apply Hnosend.
    eapply can_send_sub; eauto.
    exists w; eauto.
    intro. rewrite H3 in H0; auto.
    + rewrite s_ch_get_pop_diff in H2; auto.
    exists (s_rm_ch s c0); splits; eauto.
    symmetry; apply s_ch_pop_rm_diff_comm; auto.
    apply s_ch_get_rm_diff; auto.

    - Case "st_seq1".
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    intros; subst.
    eapply Hterm with (C0:= cx_seq C t2); eauto.
    reflexivity.

    - Case "st_par1".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    intro; apply Hnosend.
    eapply can_send_sub; eauto.
    exists w2; auto.
    intros; subst.
    eapply Hterm with (C0:= cx_par1 C b2); eauto.
    eapply join_sub_trans; eauto.
    reflexivity.

    - Case "st_par2".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    intro; apply Hnosend.
    eapply can_send_sub; eauto.
    exists w1; auto.
    intros; subst.
    eapply Hterm with (C0:= cx_par2 b1 C); eauto.
    eapply join_sub_trans; eauto.
    reflexivity.
  Qed.
*)



(*  Lemma no_send_conflict_step_s_pop_ch: forall R wC s w b s' b' c wc wl v vl,
    step (s_pop_ch s c) b s' b' ->
    can_recv c wC ->
    join_wterms b w ->
    terminates s wC b ->
    ~can_send c w ->
    s_get_ch s c = Some (R, wc::wl, v::vl) ->
    exists s'0,
      s' = s_pop_ch s'0 c /\
      (s_get_ch s'0 c = s_get_ch s c
      \/  exists wc2 v2,
          s_get_ch s'0 c = Some (R, wc::wl++[wc2], v::vl++[v2])) /\
      step s b s'0 b'.
  Proof.
    introv H Hcanrecv Hw Hterm Hnosend Hgetch.
    remember (s_pop_ch _ _) as s0.

    assert (Hterm': forall C x ec w', gjoin_sub w' w -> (b=C[w':tm_recv x ec] \/ b=C[w':tm_endch ec])%wterm -> eval ec (cname_value c) w' -> ~can_recv c wC).
      intros. destruct H1. subst.
      intro.
      eapply terminates_can_recv_conflict; eauto.
      intro.
      subst.
admit.
(*
      edestruct terminates_tm_endch_inv as [c_[R_[?[??]]]]; eauto.
      apply terminate_move_context; eauto.
      assert (cname_value c_ = cname_value c).
        eapply eval_eq; eauto.
      inverts H6.
      congruence.
*)

(*    assert (Hterm': forall x ec w', gjoin_sub w' w -> (b=w':tm_recv x ec] \/ b=C[w':tm_endch ec])%wterm -> eval ec (cname_value c) w' -> ~can_recv c wC).
      intros. destruct H1. subst.
      intro.
      eapply terminates_can_recv_conflict; eauto.
      apply terminate_move_context; eauto.
      intro.
      subst.
      edestruct terminates_tm_endch_inv as [c_[R_[?[??]]]]; eauto.
      apply terminate_move_context; eauto.
      assert (cname_value c_ = cname_value c).
        eapply eval_eq; eauto.
      inverts H6.
      congruence. *)
    clear Hterm; rename Hterm' into Hterm.
    revert s w Heqs0 Hcanrecv Hw Hterm Hnosend Hgetch.
    step_cases (induction H) Case constructor; intros;
      first [Case "st_newch" | try solve
        [ eauto
        | eexists; splits; eauto]].
    - Case "st_while_go".
    subst; eexists; splits; eauto.
    eapply st_while_go; eauto.
    rewrite s_iter_get_pop_ch in H; auto.
    - Case "st_send". clear constructor.
    subst; destruct (Pos.eq_dec c c0).
    + subst c0.
    generalize (s_ch_push_pop_same_comm _ _ _ _ _ w' v0 _ _ Hgetch); intros.
    exists (s_push_ch s c w' v0); splits; eauto.
    right.
    erewrite s_ch_get_pop_same in H1; eauto.
    inversion H1; subst R0 wl0 q; clear H1.
    exists w' v0. erewrite s_ch_get_push_same; eauto.
    reflexivity.
    erewrite s_ch_get_pop_same in H1; eauto.
    inversion H1; subst R0 wl0 q; clear H1.
    eapply st_send; eauto.
    + exists (s_push_ch s c0 w' v0); splits; auto.
    apply s_ch_push_pop_diff_comm; auto.
    left; rewrite s_ch_get_push_diff; auto; reflexivity.
    rewrite s_ch_get_pop_diff in H1; auto.
    eapply st_send; eauto.
    - Case "st_recv".
    subst. clear constructor.
    destruct (Pos.eq_dec c c0).
    + subst c0.
    false.
    simpl in *; subst.
    eapply Hterm with (x0:=x) (C:=cx_hole) (w':=w0); simpl; eauto.
    apply join_sub_refl.
    + rewrite s_ch_get_pop_diff in H5; auto.
    rewrite s_ch_pop_pop_diff; auto.
    eexists; splits; eauto.
    left; apply s_ch_get_pop_diff; auto.
    eapply st_recv; eauto.

    - Case "st_recv_blocked".
    subst. clear constructor.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false.
    simpl in *; subst.
    eapply Hterm with (x0:=x) (C:=cx_hole) (w':=w0); simpl; eauto.
    apply join_sub_refl.
    + rewrite s_ch_get_pop_diff in H2; auto.
    eexists; splits; eauto.
    eapply st_recv_blocked; eauto.

    - Case "st_newch".
    subst.
    eexists; splits.
    apply s_ch_new_pop.
    left; apply s_ch_get_new_diff.
    apply s_ch_get_new_cname_diff; auto.
    congruence.
    eapply st_newch.
    reflexivity.
    rewrite s_ch_fresh_channel_pop; auto.
    apply s_ch_get_new_cname_diff; auto.
    congruence.
    reflexivity.

    - Case "st_endch".
    subst.
    simpl in Hw; subst w0.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false.
    (* tm_endch terminates under state s, which means that the queue for channel c
       should already be empty; contradiction *)
    simpl in *; subst.
    eapply Hterm with (x:=1%positive) (C:=cx_hole) (w':=w_w'); simpl; eauto.
    apply join_sub_refl.
    eapply eval_gframe with (w:=w_w'); eauto.
    apply H.
(*    edestruct terminates_tm_endch_inv as [c_[R_[?[??]]]]; eauto.
    assert (cname_value c_ = cname_value c).
      eapply eval_eq with (w:=w_w'); eauto.
      eapply eval_gframe; eauto. 
    inverts H6.
    congruence. *)
    + rewrite s_ch_get_pop_diff in H2; auto.
    exists (s_rm_ch s c0); splits; eauto.
    symmetry; apply s_ch_pop_rm_diff_comm; auto.
    left; apply s_ch_get_rm_diff; auto.

    - Case "st_seq1".
    edestruct IHstep as [s'0[?[??]]]; eauto.
    clear -Hterm.
    intros.
    destruct H0 as [?|?]; subst.
    eapply Hterm with (x0:=x) (C0:=cx_seq C t2); simpl; eauto.
    eapply Hterm with (x:=x) (C0:=cx_seq C t2); simpl; eauto.
(*    red; intros.
    edestruct Hterm as [s''[w1'' ?]].
    eapply cstep_star_par1_inner_seq_append; eauto.
    edestruct step_star_sans_s_iter_seq_inversion as [ [s'0[w'0[??]]] | [b1''[??]] ]; eauto.
    discriminate.*)

    - Case "st_par1".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    clear -Hterm H2 Hnosend.
    intros.
    destruct H0 as [?|?]; subst.
    eapply Hterm with (x0:=x) (C0:=cx_par1 C b2) (w'0:=w'); simpl; eauto.
    eapply join_sub_trans; eauto.
    eapply Hterm with (x:=x) (C0:=cx_par1 C b2) (w'0:=w'); simpl; eauto.
    eapply join_sub_trans; eauto.
    intro; apply Hnosend.
    eapply can_send_sub; eauto.
    exists w2; auto.

    - Case "st_par2".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [s'0[?[??]]]; clear IHstep; eauto.
    clear -Hterm H2 Hnosend.
    intros.
    destruct H0 as [?|?]; subst.
    eapply Hterm with (x0:=x) (C0:=cx_par2 b1 C) (w'0:=w'); simpl; eauto.
    eapply join_sub_trans; eauto.
    eapply Hterm with (x:=x) (C0:=cx_par2 b1 C) (w'0:=w'); simpl; eauto.
    eapply join_sub_trans; eauto.
    intro; apply Hnosend.
    eapply can_send_sub; eauto.
    exists w1; auto.
  Qed.
*)

  Lemma no_send_conflict_step_ch_empty: forall s w b s' b' c,
    step s b s' b' ->
    join_wterms b w ->
    ~can_send c w ->
    s_get_ch s c <> None ->
    (s_get_ch s c = s_get_ch s' c \/
    exists R wc wl v vl,
      s_get_ch s c = Some (R, wc::wl, v::vl) /\ s_get_ch s' c = Some (R, wl, vl)).
  Proof.
    introv H Hw Hno_send.
    revert w Hw Hno_send.
    step_cases (induction H) Case constructor; intros;
      first [Case "st_newch" | try solve
        [ eauto
        ]].
    - Case "st_send".
    simpl in *; subst; destruct (Pos.eq_dec c c0).
    + subst c0.
    false. clear -Hno_send H2. contradiction.
    + left; rewrite s_ch_get_push_diff; auto.
    - Case "st_recv".
    simpl in *; subst; destruct (Pos.eq_dec c c0).
    + subst c0.
    right.
    do 5 eexists; split; eauto.
    eapply s_ch_get_pop_same; eauto.
    + left; rewrite s_ch_get_pop_diff; auto.
    - Case "st_newch".
    subst.
    left; rewrite s_ch_get_new_diff; auto.
    apply s_ch_get_new_cname_diff; auto.
    - Case "st_endch".
    simpl in *; subst; destruct (Pos.eq_dec c c0).
    + subst c0.
    false; clear -Hno_send H0 H1.
    destruct H0 as [? _].
    red in H.
    apply Hno_send.
    eapply can_send_sub.
    exists w; eauto.
    intro.
    rewrite H0 in H; auto.
    + left; rewrite s_ch_get_rm_diff; auto.

    - Case "st_par1".
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [?| [R[wc[wl[v[vl[??]]]]]] ]; clear IHstep; eauto.
    intro.
    eapply Hno_send.
    eapply can_send_sub; eauto.
    exists w2; auto.
    right; do 5 eexists; split; eauto.

    - Case "st_par2".
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep as [?| [R[wc[wl[v[vl[??]]]]]] ]; clear IHstep; eauto.
    intro.
    eapply Hno_send.
    eapply can_send_sub; eauto.
    exists w1; auto.
    right; do 5 eexists; split; eauto.
  Qed.

  Lemma s_ch_new_rm: forall s c w R,
    s_new_ch (s_rm_ch s c) (fresh_channel (s_rm_ch s c) w) R =
      s_rm_ch (s_new_ch s (fresh_channel s w) R) c.
(* this is not quite provable: need to prevent freed channel names from being reused *)
Admitted.

  Lemma s_ch_fresh_channel_rm: forall s c w,
    fresh_channel (s_rm_ch s c) w = fresh_channel s w.
(* not quite provable: need to prevent freed channels from being reallocated. *)
Admitted.

  Lemma s_ch_rm_rm: forall s c1 c2,
    s_rm_ch (s_rm_ch s c1) c2 = s_rm_ch (s_rm_ch s c2) c1.
  Proof.
    unfold s_rm_ch; intros.
    destruct (Pos.eq_dec c1 c2). subst c1.
    reflexivity.
    apply s_ch_apply_diff; auto.
  Qed.

  Lemma no_send_conflict_step_s_rm_ch: forall s w b s' b' c,
    step (s_rm_ch s c) b s' b' ->
    ~can_send c w ->
    join_wterms b w ->
    s_get_ch s c <> None ->
    exists s'0,
      s' = s_rm_ch s'0 c /\
      s_get_ch s c = s_get_ch s'0 c /\
      step s b s'0 b'.
  Proof.
    introv H Hno_send Hw Hgetch.
    remember (s_rm_ch _ _) as s0.
    revert s w Heqs0 Hno_send Hw Hgetch.
    step_cases (induction H) Case constructor; intros;
      first [Case "st_newch" | try solve
        [ eauto
        | eexists; splits; eauto]].
    - Case "st_while_go".
    subst; eexists; splits; eauto.
    eapply st_while_go; eauto.
    rewrite s_iter_get_rm_ch in H; auto.
    - Case "st_send". clear constructor.
    subst; destruct (Pos.eq_dec c c0).
    + subst c0.
    false.
    rewrite s_ch_get_rm_same in H1.
    discriminate.
    + exists (s_push_ch s c0 w' v); splits; auto.
    symmetry; apply s_ch_rm_push.
    rewrite s_ch_get_push_diff; auto; reflexivity.
    rewrite s_ch_get_rm_diff in H1; auto.
    eapply st_send; eauto.
    - Case "st_recv".
    subst. clear constructor.
    destruct (Pos.eq_dec c c0).
    + subst c0.
    false.
    rewrite s_ch_get_rm_same in H5.
    discriminate.
    + rewrite s_ch_get_rm_diff in H5; auto.
    eexists (s_pop_ch s c0); splits; eauto.
    apply s_ch_pop_rm_diff_comm; auto.
    rewrite s_ch_get_pop_diff; auto.
    eapply st_recv; eauto.
    - Case "st_recv_blocked".
    subst. clear constructor.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false.
    rewrite s_ch_get_rm_same in H2.
    discriminate.
    + rewrite s_ch_get_rm_diff in H2; auto.
    eexists; splits; eauto.
    eapply st_recv_blocked; eauto.
    - Case "st_newch".
    simpl in *; subst; clear constructor.
    eexists; splits.
    apply s_ch_new_rm.
    rewrite s_ch_get_new_diff; auto.
    apply s_ch_get_new_cname_diff; auto.
    rewrite s_ch_fresh_channel_rm.
    eapply st_newch; eauto.
    - Case "st_endch".
    simpl in *; subst. clear constructor.
    destruct (Pos.eq_dec c0 c).
    + subst c0.
    false.
    rewrite s_ch_get_rm_same in H2.
    discriminate.
    + rewrite s_ch_get_rm_diff in H2; auto.
    exists (s_rm_ch s c0); splits; eauto.
    apply s_ch_rm_rm.
    rewrite s_ch_get_rm_diff; auto.
    eapply st_endch; eauto.
    apply H.

    - Case "st_seq1".
    edestruct IHstep as [s'0[?[??]]]; eauto.

    - Case "st_par1".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep with (w:=w1) as [s'0[?[??]]]; clear IHstep; eauto.
    intro; apply Hno_send.
    eapply can_send_sub; eauto. exists w2; auto.

    - Case "st_par2".
    subst.
    destruct Hw as [w1[w2[?[??]]]].
    edestruct IHstep with (w:=w2) as [s'0[?[??]]]; clear IHstep; eauto.
    intro; apply Hno_send.
    eapply can_send_sub; eauto. exists w1; auto.
  Qed.

  Lemma noninterfering_schedule_step: forall s w1 b1 w2 b2 s' b2' s'' b1'',
    join_wterms b1 w1 ->
    join_wterms b2 w2 ->
    no_send_conflict w2 w1 ->
    gjoins w1 w2 ->
    terminates s w2 b1 ->
    cannot_communicate_channels s b1 ->
    valid_machine_state w2 s b1 ->
    step s b2 s' b2' ->
    step s' b1 s'' b1'' ->
    exists s'0,
      step s b1 s'0 b1'' /\
      step s'0 b2 s'' b2'.
  Proof.
    introv Hjb1 Hjb2 Hno_send Hw Hterm Hcc Hvms Hs2 Hs1.
    assert (Hjs: gjoin_sub w2 w2). apply join_sub_refl.
    remember w2 as w0 in Hjs at 1.
    rewrite<- Heqw0 in Hjb2.
    clear Heqw0.
    revert w0 Hjs Hjb2 Hw Hterm Hcc Hvms.
    step_cases (induction Hs2; intros; subst) Case constructor;
      try solve [Case "st_newch"; admit
        | exists s''; split; eauto].

    - Case "st_while_go".
    exists s''; split; eauto; apply constructor; clear constructor; auto.
    eapply IterationBounds.step_go_preservation; eauto.
    - Case "st_send". clear constructor.
    simpl in Hjb2; subst w_w'.
    edestruct no_send_conflict_step_s_push_ch
      with (wO:=w2) as [s'0[?[[?| [R_[wc1[wl1[wl2[wc2[c1[vl1[vl2[v2[?[?[??]]]]]]]]]]]] ]?]]]; eauto.
    eapply can_send_sub; eauto.
    intro; edestruct Hno_send; eauto. apply H6.
    eapply can_send_sub; eauto.
    congruence.
    exists s'0; split; auto.
    eapply st_send with (R:=R) (v:=v); eauto.
    rewrite H6; eauto.
    rewrite H1 in H8. inversion H8; subst R_ wl q; clear H8.
    exists s'0; split; auto.
    eapply st_send with (R:=R) (v:=v); eauto.
    - Case "st_recv".
    clear constructor.
    simpl in Hjb2; subst w0.
    subst.
    edestruct no_send_conflict_step_s_pop_ch
      with (wO:=w2) as [s'0[?[??]]]; eauto.
    eapply can_recv_sub; eauto.
    right.
    intro Hsend; apply Hno_send in Hsend. destruct Hsend as [??].
    apply H6. eapply can_recv_sub; eauto.
    destruct Hw; eexists; eauto.
    destruct H6 as [?|[? _]].
    exists s'0; split; auto.
    eapply st_recv with (R:=R); eauto.
    rewrite H6; rewrite H5; reflexivity.
    false.
    apply Hno_send in H6. destruct H6 as [??].
    apply H8. eapply can_recv_sub; eauto.
    - Case "st_recv_blocked".
    clear constructor; subst.
    simpl in Hjb2; subst w0.
    exists s''; split; auto.
    edestruct no_send_conflict_step_ch_empty with (c:=c) as [?| [R_[wc[wl[v[vl[??]]]]]] ]; eauto.
    intro; destruct Hno_send with c as [??]; auto.
    apply H5.
    eapply can_recv_sub; eauto.
    congruence.
    rewrite H3 in H2.
    eapply st_recv_blocked; eauto.
    false; congruence.
    - Case "st_endch".
    clear constructor; simpl in *; subst.
    edestruct no_send_conflict_step_s_rm_ch as [s'0[?[??]]]; eauto.
    intro; destruct Hno_send with c as [??]; auto.
    apply H5.
    eapply can_recv_sub; eauto.
    eapply can_recv_sub; eauto. exists w; eauto.
    apply total_send_recv_can_send_recv; auto.
    congruence.
    rewrite H4 in H2.
    exists s'0; split; auto.
    eapply st_endch; eauto.
    apply H.
    - Case "st_seq1".
    destruct IHHs2 with w0 as [s'0[??]]; clear IHHs2; auto.
    exists s'0; splits; eauto.
    - Case "st_par1".
    unfold join_wterms in Hjb2; fold join_wterms in Hjb2.
    destruct Hjb2 as [w0'[w3[?[? ?]]]].
    destruct IHHs2 with w0' as [s'0[??]]; auto.
    apply join_sub_trans with w0; auto.
    exists w3; auto.
    exists s'0; splits; eauto.
    - Case "st_par2".
    unfold join_wterms in Hjb2; fold join_wterms in Hjb2.
    destruct Hjb2 as [w0'[w3[?[? ?]]]].
    destruct IHHs2 with w3 as [s'0[??]]; auto.
    apply join_sub_trans with w0; auto.
    exists w0'; apply join_com; auto.
    exists s'0; splits; eauto.
  Qed.


  (* any channel that w1 can access is completely owned by w *)
  Definition private_communications w1 w :=
    (forall c, (can_send c w1 \/ can_recv c w1) -> total_send_recv c w).

  Lemma private_communications_sub: forall w1 w ww1 ww,
    private_communications w ww1 ->
    gjoin_sub w1 w ->
    gjoin_sub ww1 ww ->
    private_communications w1 ww.
  Proof.
    unfold private_communications; intros.
    eapply total_send_recv_sub; eauto.
    apply H.
    destruct H2; [left|right].
    eapply can_send_sub; eauto.
    eapply can_recv_sub; eauto.
  Qed.

  Lemma gjoins_com: forall w1 w2,
    gjoins w1 w2 -> gjoins w2 w1.
  Proof. introv [w ?]; exists w; apply gjoin_com; auto. Qed.

  Lemma no_conflict_schedule_stepO: forall s wO bO wb w b s' b' s'' bO'',
    join_wterms b w ->
    join_wterms bO wO ->
    gjoin_sub w wb ->
    private_communications w wb ->
    gjoins wO wb ->
    step s b s' b' ->
    step s' bO s'' bO'' ->
    exists s'0,
      step s bO s'0 bO'' /\
      step s'0 b s'' b'.
  Proof.
    introv Hjb1 Hjb2 Hwb Hprivate HwOb Hs2 Hs1.
    revert wO w Hjb1 Hjb2 Hwb Hprivate HwOb.
    step_cases (induction Hs2; intros; subst) Case constructor;
      try solve [Case "st_newch"; admit
        | exists s''; split; eauto].

    - Case "st_while_go".
    exists s''; split; eauto; apply constructor; clear constructor; auto.
    eapply IterationBounds.step_go_preservation; eauto.
    - Case "st_send". clear constructor.
    simpl in Hjb1; subst w0.
    edestruct no_send_conflict_step_s_push_ch' as [s'0[?[??]]]; eauto.
    intro; eapply has_total_endpoint_send_conflict with (w1:=wb); eauto.
    apply Hprivate; auto. apply gjoins_com; auto.
    intro; eapply has_total_endpoint_recv_conflict with (w1:=wb); eauto.
    apply Hprivate; auto. apply gjoins_com; auto.
    congruence.
    exists s'0; split; auto.
    rewrite<- H6 in H1.
    eapply st_send with (R:=R) (v:=v); eauto.
    - Case "st_recv".
    clear constructor.
    simpl in Hjb1; subst w0.
    subst.
    edestruct no_send_conflict_step_s_pop_ch' as [s'0[?[??]]]; eauto.
    intro; eapply has_total_endpoint_send_conflict with (w1:=wb); eauto.
    apply Hprivate; auto. apply gjoins_com; auto.
    intro; eapply has_total_endpoint_recv_conflict with (w1:=wb); eauto.
    apply Hprivate; auto. apply gjoins_com; auto.
    exists s'0; split; auto.
    eapply st_recv with (R:=R); eauto.
    rewrite H6; rewrite H5; reflexivity.
    - Case "st_recv_blocked".
    clear constructor; subst.
    simpl in Hjb1; subst w0.
    exists s''; split; auto.
    edestruct no_send_conflict_step_ch_empty with (c:=c) as [?| [R_[wc[wl[v[vl[??]]]]]] ]; eauto.
    intro; eapply has_total_endpoint_send_conflict with (w1:=wb); eauto.
    apply Hprivate; auto. apply gjoins_com; auto.
    congruence.
    rewrite H3 in H2.
    eapply st_recv_blocked; eauto.
    false; congruence.
    - Case "st_endch".
    clear constructor; simpl in *; subst.
    edestruct no_send_conflict_step_s_rm_ch as [s'0[?[??]]]; eauto.
    intro; eapply has_total_endpoint_send_conflict with (w1:=wb); eauto.
    eapply total_send_recv_sub; eauto. eapply join_sub_trans; eauto. apply gjoins_com; auto.
    congruence.
    rewrite H4 in H2.
    exists s'0; split; auto.
    eapply st_endch; eauto.
    apply H.
    - Case "st_seq1".
    destruct IHHs2 with wO w as [s'0[??]]; clear IHHs2; auto.
    exists s'0; splits; eauto.
    - Case "st_par1".
    unfold join_wterms in Hjb2; fold join_wterms in Hjb2.
    destruct Hjb1 as [w0'[w3[?[? ?]]]].
    destruct IHHs2 with wO w0' as [s'0[??]]; auto.
    apply join_sub_trans with w; auto.
    exists w3; auto.
    eapply private_communications_sub; eauto.
    exists w3; auto. apply join_sub_refl.
    exists s'0; splits; eauto.
    - Case "st_par2".
    unfold join_wterms in Hjb2; fold join_wterms in Hjb2.
    destruct Hjb1 as [w0'[w3[?[? ?]]]].
    destruct IHHs2 with wO w3 as [s'0[??]]; auto.
    apply join_sub_trans with w; auto.
    exists w0'; apply join_com; auto.
    eapply private_communications_sub; eauto.
    exists w0'; auto. apply join_sub_refl.
    exists s'0; splits; eauto.
  Qed.

Import StepStar.


(*
  Lemma step_s_new_ch_terminates: forall w R s b s' b',
    step (s_new_ch s (fresh_channel s w) R) b s' b' ->
    ~can_send (FreshChannels.fc_c (fresh_channel s w)) w ->
    ~can_recv (FreshChannels.fc_c (fresh_channel s w)) w ->
    join_wterms b w ->
    exists s'0,
      s' = (s_new_ch s'0 (fresh_channel s w) R) /\
      (fresh_channel s'0 w = fresh_channel s w \/
      exists c wc v,
        fresh_channel s'0 w = fresh_channel (s_push_ch s c wc v) w) /\
      s_new_ch s (fresh_channel s w) R = s_new_ch s'0 (fresh_channel s'0 w) R /\
      step s b s'0 b'.
  Proof.
    introv H Hnosend Hnorecv Hw.
    remember (fresh_channel s w) as fresh_ch.
    remember (s_new_ch _ _ _) as s0.
    revert s w Heqs0 Heqfresh_ch Hnosend Hnorecv Hw.
    step_cases (induction H) Case constructor; intros; subst; jauto.
    - Case "st_while_go".
    rewrite s_iter_get_new_ch in H.
    eexists; splits; eauto; eapply constructor; eauto.
    - Case "st_send".
    simpl in Hw; subst w0.
    exists (s_push_ch s c w' v); splits; eauto.
    rewrite<- s_ch_new_push.
    rewrite s_ch_fresh_channel_push; auto.
    intro; subst c.
    apply Hnosend; auto.
    rewrite s_ch_new_push.
Qed.

    edestruct step_s_new_ch_terminates as [s'0[?[?[??]]]]; eauto.
    subst.
    exists s'0; split; auto.
    apply step_single.
    eapply st_newch; eauto.
    rewrite H0. reflexivity.
*)

  Lemma cannot_communicate_channels_par_assoc: forall s b1 b2 b3,
    cannot_communicate_channels s (b1 || (b2||b3)) <-> cannot_communicate_channels s ((b1||b2)||b3).
  Proof.
    split; introv [w[?[?[??]]]].
    exists w; splits; auto.
    apply join_wterm_worlds in H.
    apply join_wterm_worlds.
    eapply join_list_perm; eauto.
    simpl. permut3.
    destruct H2.
    split; auto; split; auto.
    exists w; splits; auto.
    apply join_wterm_worlds in H.
    apply join_wterm_worlds.
    eapply join_list_perm; eauto.
    simpl. permut3.
    destruct H1.
    split; auto; split; auto.
  Qed.

  Lemma valid_machine_state_par_assoc: forall wC s b1 b2 b3,
    valid_machine_state wC s (b1 || (b2||b3)) <-> valid_machine_state wC s ((b1||b2)||b3).
  Proof.
    split; introv [wS[wb[ww[?[H0 ?]]]]]; exists wS wb ww; splits; auto;
      apply join_wterm_worlds in H0; apply join_wterm_worlds;
      eapply join_list_perm; eauto.
    simpl; permut3.
    simpl; permut3.
  Qed.

  Lemma noninterfering_wterm_2par1: forall b1 b2 b3 w,
    noninterfering_wterm b1 (b2||b3) ->
    join_wterms (b1 || (b2 || b3))%wterm w ->
    noninterfering_wterm b1 b2.
  Proof.
    unfold noninterfering_wterm; intros.
    destruct H0 as [w1_[w23[?[??]]]].
    replace w1_ with w1 in *.
    eapply noninterfering_sub2; eauto.
    destruct H3 as [w2_[w3[?[??]]]].
    replace w2_ with w2 in *.
    exists w3; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma cannot_communicate_channels_par_assoc_swap2: forall s b1 b2 b3,
    cannot_communicate_channels s (b1 || (b2||b3)) <-> cannot_communicate_channels s ((b1||b3)||b2).
  Proof.
    split; introv [w[H[?[??]]]]; exists w; splits; auto.
    apply join_wterm_worlds in H; apply join_wterm_worlds.
    eapply join_list_perm; eauto.
    simpl; permut3.
    destruct H2.
    split; auto; split; auto.
    apply join_wterm_worlds in H; apply join_wterm_worlds.
    eapply join_list_perm; eauto.
    simpl; permut3.
    destruct H1.
    split; auto; split; auto.
  Qed.

  Lemma valid_machine_state_par_assoc_swap2: forall wC s b1 b2 b3,
    valid_machine_state wC s (b1 || (b2||b3)) <-> valid_machine_state wC s ((b1||b3)||b2).
  Proof.
    split; introv [wS[wb[ww[?[H0 ?]]]]]; exists wS wb ww; splits; auto;
      apply join_wterm_worlds in H0; apply join_wterm_worlds;
      eapply join_list_perm; eauto.
    simpl; permut3.
    simpl; permut3.
  Qed.

  Lemma noninterfering_wterm_2par2: forall b1 b2 b3 w,
    noninterfering_wterm b1 (b2||b3) ->
    join_wterms (b1 || (b2 || b3))%wterm w ->
    noninterfering_wterm b1 b3.
  Proof.
    unfold noninterfering_wterm; intros.
    rename w2 into w3.
    destruct H0 as [w1_[w23[?[??]]]].
    replace w1_ with w1 in *.
    eapply noninterfering_sub2; eauto.
    destruct H3 as [w2[w3_[?[??]]]].
    replace w3_ with w3 in *.
    exists w2; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
  Qed.


  Lemma noninterfering_schedule_step': forall s w1 b1 w2 b2 s' b2' s'' b1'',
    noninterfering_wterm b1 b2 ->
    join_wterms b1 w1 ->
    join_wterms b2 w2 ->
    terminates s w2 b1 ->
    cannot_communicate_channels s (b1||b2) ->
    valid_machine_state gwemp s (b1||b2) ->
    step s b2 s' b2' ->
    step s' b1 s'' b1'' ->
    exists s'0,
      step s b1 s'0 b1'' /\ step_star s'0 b2 s'' b2'.
  Proof.
    introv Hni Hw1 Hw2 Hterm1 Hcc Hvms Hs2 Hs1.
    assert (Hni': noninterfering_wterm b1 b2').
      eapply noninterfering_step2; eauto.
      eapply cannot_communicate_channels_par2; eauto.
    revert w2 Hw1 Hw2 Hterm1 Hcc Hvms Hni Hni'.
    step_cases (induction Hs2; intros; subst) Case constructor;
      try solve [exists s''; split; auto; apply step_single; eapply constructor; eauto].

    - Case "st_while_go".
    exists s''; split; auto; apply step_single; eapply constructor; eauto.
    eapply IterationBounds.step_go_preservation; eauto.
    - Case "st_send". clear constructor.
    simpl in Hw2; subst w2.
    edestruct no_send_conflict_step_s_push_ch
      with (wO:=w_w') as [s'0[?[[?| [R_[wc1[wl1[wl2[wc2[c1[vl1[vl2[v2[?[?[??]]]]]]]]]]]] ]?]]]; eauto.
    eapply Hni; eauto. reflexivity.
    eapply cannot_communicate_channels_par1; eauto.
    congruence.
    exists s'0; split; auto.
    apply step_single.
    eapply st_send with (R:=R) (v:=v); eauto.
    rewrite H6; eauto.
    rewrite H1 in H8. inversion H8; subst R_ wl q; clear H8.
    exists s'0; split; auto.
    apply step_single.
    eapply st_send with (R:=R) (v:=v); eauto.
    - Case "st_recv".
    clear constructor.
    simpl in Hw2; subst w.
    subst.
    edestruct no_send_conflict_step_s_pop_ch with (wO:=w2)
      as [s'0[?[??]]]; eauto.
    left.
    eapply Hni; eauto. reflexivity.
    edestruct valid_machine_state_join_wterms as [w12 ?]; eauto.
    destruct H4 as [w1_[w2_[?[??]]]].
    simpl in H6; subst w2_.
    replace w1_ with w1 in *.
    exists w12; auto.
    eapply join_wterms_eq; eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply valid_machine_state_par_observer with (wO1:=gwemp) (wO2:=w2); eauto.
    reflexivity. apply gjoin_bot2.
    exists s'0; split; auto.
    apply step_single.
    destruct H6 as [?| [?[wc2[v2 ?]]] ].
    rewrite<- H6 in H5.
    eapply st_recv with (R:=R); eauto.
    subst.
    apply s_ch_get_push_same with (w:=wc2) (v:=v2) in H5.
    rewrite H5 in H8.
    eapply st_recv with (R:=R); eauto.

    - Case "st_recv_blocked".
    exists s''; splits; auto.
    apply step_nil.
    - Case "st_newch".
    clear constructor; simpl in *; subst.
    unfold noninterfering_wterm,noninterfering in Hni'.
admit.

    - Case "st_endch".
    clear constructor; simpl in *; subst.
    edestruct no_send_conflict_step_s_rm_ch as [s'0[?[??]]]; eauto.
    intro.
    eapply has_total_endpoint_send_conflict; eauto. apply H0.
    edestruct valid_machine_state_join_wterms as [w12 ?]; eauto.
    destruct H4 as [w1_[w2_[?[??]]]].
    simpl in H5; subst w2_.
    replace w1_ with w1 in *.
    eapply gjoins_sub1 with (w1':=w2); eauto. exists w12; auto.
    exists w; auto.
    eapply join_wterms_eq; eauto.
    congruence.
    rewrite H4 in H2.
    exists s'0; split; auto.
    apply step_single.
    eapply st_endch; eauto.
    apply H.

    - Case "st_seq1".
    rename b0 into b2.
    destruct IHHs2 with w2 as [s'0[??]]; auto.
    destruct Hcc as [w12[?[?[?[??]]]]]; exists w12; splits; auto.
    split; auto.
    exists s'0; split; eauto.
    eapply step_star_seq_app; eauto.
    - Case "st_par1".
    destruct Hw2 as [w0'[w3[?[? ?]]]].
    destruct IHHs2 with w0' as [s'0[??]]; auto.
    eapply terminates_sub; eauto. exists w3; auto.
    eapply valid_machine_state_par_observer with (wO2:=w2) (bO:=(b0||b2)%wterm); eauto.
    exists w0' w3; splits; auto.
    apply gjoin_bot2.
    apply cannot_communicate_channels_par_assoc in Hcc.
    eapply cannot_communicate_channels_par1; eauto.
    apply valid_machine_state_par_assoc in Hvms.
    eapply valid_machine_state_par1_sub; eauto.
    edestruct valid_machine_state_join_wterms as [w ?]; eauto.
    eapply noninterfering_wterm_2par1; eauto.
    assert (valid_machine_state gwemp s' (b1||(b1'||b2))).
      eapply valid_machine_state_step; eauto.
      apply st_par2; apply st_par1; auto.
    edestruct valid_machine_state_join_wterms with (s:=s') as [w' ?]; eauto.
    eapply noninterfering_wterm_2par1; eauto.
    exists s'0; split; auto. apply step_star_par1_step; auto.
    - Case "st_par2".
    destruct Hw2 as [w0'[w3[?[? ?]]]].
    destruct IHHs2 with w3 as [s'0[??]]; auto.
    eapply terminates_sub; eauto. exists w0'; auto.
    eapply valid_machine_state_par_observer with (wO2:=w2) (bO:=(b0||b2)%wterm); eauto.
    exists w0' w3; splits; auto.
    apply gjoin_bot2.
    apply cannot_communicate_channels_par_assoc_swap2 in Hcc.
    eapply cannot_communicate_channels_par1; eauto.
    apply valid_machine_state_par_assoc_swap2 in Hvms.
    eapply valid_machine_state_par1_sub; eauto.
    edestruct valid_machine_state_join_wterms as [w ?]; eauto.
    eapply noninterfering_wterm_2par2; eauto.
    assert (valid_machine_state gwemp s' (b1||(b0||b2'))).
      eapply valid_machine_state_step; eauto.
    edestruct valid_machine_state_join_wterms with (s:=s') as [w' ?]; eauto.
    eapply noninterfering_wterm_2par2; eauto.
    exists s'0; split; auto. apply step_star_par2_step; auto.
  Qed.



(*
  Lemma step_s_new_ch_terminates: forall w R s b s' b',
    step (s_new_ch s (fresh_channel s w) R) b s' b' ->
    ~can_send (FreshChannels.fc_c (fresh_channel s w)) w ->
    ~can_recv (FreshChannels.fc_c (fresh_channel s w)) w ->
    join_wterms b w ->
    exists s'0,
      s' = (s_new_ch s'0 (fresh_channel s w) R) /\
      (fresh_channel s'0 w = fresh_channel s w \/
      exists c wc v,
        fresh_channel s'0 w = fresh_channel (s_push_ch s c wc v) w) /\
      s_new_ch s (fresh_channel s w) R = s_new_ch s'0 (fresh_channel s'0 w) R /\
      step s b s'0 b'.
  Proof.
    introv H Hnosend Hnorecv Hw.
    remember (fresh_channel s w) as fresh_ch.
    remember (s_new_ch _ _ _) as s0.
    revert s w Heqs0 Heqfresh_ch Hnosend Hnorecv Hw.
    step_cases (induction H) Case constructor; intros; subst; jauto.
    - Case "st_while_go".
    rewrite s_iter_get_new_ch in H.
    eexists; splits; eauto; eapply constructor; eauto.
    - Case "st_send".
    simpl in Hw; subst w0.
    exists (s_push_ch s c w' v); splits; eauto.
    rewrite<- s_ch_new_push.
    rewrite s_ch_fresh_channel_push; auto.
    intro; subst c.
    apply Hnosend; auto.
    rewrite s_ch_new_push.
Qed.

    edestruct step_s_new_ch_terminates as [s'0[?[?[??]]]]; eauto.
    subst.
    exists s'0; split; auto.
    apply step_single.
    eapply st_newch; eauto.
    rewrite H0. reflexivity.
*)



  Lemma noninterfering_rev_context_step: forall s wC C w2 b2 s' C' s'' b2'',
    context_world C wC ->
    join_wterms b2 w2 ->
    no_send_conflict wC w2 ->
    terminates s wC b2 ->
    cannot_communicate_channels s b2 ->
    valid_machine_state wC s b2 ->
    cstep s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) ->
    step s' b2 s'' b2'' ->
    exists s'0,
      step s b2 s'0 b2'' /\
      cstep_star s'0 C (gwemp:tm_skip) s'' C' (gwemp:tm_skip).
  Proof.
    introv HjC Hjb2 Hnosend Hterm Hcc Hvms HcsC Hs2.
    remember (gwemp : tm_skip)%wterm as b1 in HcsC at 1.
    remember (gwemp : tm_skip)%wterm as b1' in HcsC.
    revert wC Heqb1 Heqb1' HjC Hnosend Hterm Hcc Hvms.
    induction HcsC; intros; subst.
    - simpl in HjC.
    destruct IHHcsC with wC as [s'0[??]]; auto.
    exists s'0; split; auto.
    replace (cx_seq C t2) with (compose_contexts (cx_seq cx_hole t2) C); auto.
    replace (cx_seq C' t2) with (compose_contexts (cx_seq cx_hole t2) C'); auto.
    apply cstep_star_composed_context_step2 with gwemp; auto.
    - destruct HjC as [wCa[wCb[?[??]]]].
    destruct IHHcsC with wCa as [s'0[??]]; auto.
    eapply no_send_conflict_sub; eauto.
    exists wCb; auto. apply join_sub_refl.
    eapply terminates_sub; eauto. exists wCb; auto.
    eapply valid_machine_state_subO; eauto.
    exists wCb; auto.
    exists s'0; split; auto.
    eapply cstep_star_par_cx_par1.
    apply cstep_star_skip_subst with gwemp; auto.
    - unfold set_context,join_wterms in HjC; fold set_context in HjC; fold join_wterms in HjC.
    destruct HjC as [wCa[wCb[?[??]]]].
    destruct IHHcsC with wCb as [s'0[??]]; auto.
    eapply no_send_conflict_sub; eauto.
    exists wCa; auto. apply join_sub_refl.
    eapply terminates_sub; eauto.
    exists wCa; auto.
    eapply valid_machine_state_subO; eauto.
    exists wCa; auto.
    exists s'0; split; auto.
    eapply cstep_star_par_cx_par2.
    apply cstep_star_skip_subst with gwemp; auto.
    - clear Heqb1'.
    rename b0 into b1; rename b2' into b1'.
    destruct HjC as [wCa[wCb[?[??]]]].
    edestruct noninterfering_schedule_step with (w1:=w2) (w2:=wCb) (b1:=b2) (b2:=b1) as [s'0[??]]; eauto.
    eapply no_send_conflict_sub with wC w2; auto.
    exists wCa; auto.
    apply join_sub_refl.
    destruct Hvms as [wS[w2_[ww[?[??]]]]].
    replace w2_ with w2 in *.
    destruct H5 as [wC2[? _]].
    apply gjoin_join_list in H5.
    edestruct gjoin_assoc with (a:=wCa) (b:=wCb) as [?[??]]; eauto.
    apply gjoin_com; eauto.
    exists x; auto.
    eapply join_wterms_eq; eauto.
    eapply terminates_sub; eauto.
    exists wCa; auto.
    eapply valid_machine_state_subO; eauto. exists wCa; auto.
    exists s'0; split; auto.
    eapply cstep_star_par_cx_par1.
    apply step_single_cstep_star. apply st_par2; auto.
    - unfold set_context,join_wterms in HjC; fold set_context in HjC; fold join_wterms in HjC.
    destruct HjC as [wCa[wCb[?[??]]]].
    edestruct noninterfering_schedule_step with (w1:=w2) (w2:=wCa) (b1:=b2) (b2:=b1) as [s'0[??]]; eauto.
    eapply no_send_conflict_sub with wC w2; auto.
    exists wCb; auto.
    apply join_sub_refl.
    destruct Hvms as [wS[w2_[ww[?[??]]]]].
    replace w2_ with w2 in *.
    destruct H5 as [wC2[? _]].
    apply gjoin_join_list in H5.
    apply gjoin_com in H2.
    apply gjoin_com in H5.
    edestruct gjoin_assoc with (a:=wCb) (b:=wCa) as [?[??]]; eauto.
    exists x; auto.
    eapply join_wterms_eq; eauto.
    eapply terminates_sub; eauto.
    exists wCb; auto.
    eapply valid_machine_state_subO; eauto. exists wCb; auto.
    exists s'0; split; auto.
    eapply cstep_star_par_cx_par2.
    apply step_single_cstep_star. apply st_par1; auto.
    - apply step_spin_no_change in H. subst s'.
    exists s''; split; auto.
    apply cstep_nil.
  Qed.

(*
  Lemma preservation_no_send_to_context: forall s C b s' C' b',
    no_send_to_context C b ->
    cstep_star s C b s' C' b' ->
    no_send_to_context C' b'.
  Proof.
(* THIS IS ONLY TRUE IF WE DO NOT ALLOW FIRST-CLASS CHANNELS *)
(* Otherwise, we also need to state that any recv of b does not have the resource invariant to get a send-permission *)
  Admitted.

  Lemma preservation_no_send_to_context1: forall s C b s' C',
    no_send_to_context C b ->
    cstep_star s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) ->
    no_send_to_context C' b.
  Proof.
    intros. eapply preservation_no_send_to_context; eauto.
    apply cstep_star_skip_subst with gwemp; eauto.
  Qed.

  Lemma preservation_no_send_to_context1': forall s C b1 b s' C' b1',
    no_send_to_context C b ->
    cstep_star s C b1 s' C' b1' ->
    no_send_to_context C' b.
  Proof.
  (* THIS IS ONLY TRUE IF WE DO NOT ALLOW FIRST-CLASS CHANNELS *)
  (* This is used as a placeholder in the bisimulation proof to satisfy the
      Guard. We need a more precise lemma that prevents b1 from giving a recv
      permission to the context. bleh. *)
Admitted.

  Lemma preservation_no_send_to_context_par: forall s C b1 b s' C' b1' b',
    no_send_to_context C b ->
    cstep_star s C (b1||b) s' C' (b1'||b') ->
    no_send_to_context C' b'.
  Proof.
  (* THIS IS ONLY TRUE IF WE DO NOT ALLOW FIRST-CLASS CHANNELS *)
  (* This is used as a placeholder in the bisimulation proof to satisfy the
      Guard. We need a more precise lemma that prevents b1 from giving a recv
      permission to the context. bleh. *)
Admitted.

  Lemma preservation_no_send_to_context2: forall s C b s' b',
    no_send_to_context C b ->
    step s b s' b' ->
    no_send_to_context C b'.
  Proof.
    intros. eapply preservation_no_send_to_context; eauto.
    eapply step_single_cstep_star; eauto.
  Qed.

  Lemma preservation_no_send_to_context2': forall s C b s'' b'',
    no_send_to_context C b ->
    step_star s b s'' b'' ->
    no_send_to_context C b''.
  Proof.
    intros. induction H0; auto.
    apply IHstep_star.
    eapply preservation_no_send_to_context2; eauto.
  Qed.
*)

Inductive cstep_star_snoc_I s C b: state->context->wterm->Prop :=
| cstep_snoc_nil: 
    cstep_star_snoc_I s C b s C b
| cstep_snoc_cons: forall s' s'' b' b'' C' C'',
    cstep_star_snoc_I s C b s' C' b' ->
    cstep s' C' b' s'' C'' b'' ->
    cstep_star_snoc_I s C b s'' C'' b''
| cstep_snoc_s_iter: forall s' s'' b' C',
    cstep_star_snoc_I s C b s' C' b' ->
    s_iter s' s'' ->
    cstep_star_snoc_I s C b s'' C' b'.

  Lemma cstep_cons_snoc: forall s C b s' C' b' s'' C'' b'',
    cstep s C b s' C' b' ->
    cstep_star_snoc_I s' C' b' s'' C'' b'' ->
    cstep_star_snoc_I s C b s'' C'' b''.
  Proof.
    intros.
    induction H0.
    eapply cstep_snoc_cons; eauto.
    apply cstep_snoc_nil.
    eapply cstep_snoc_cons; eauto.
    eapply cstep_snoc_s_iter; eauto.
  Qed.

  Lemma cstep_snoc_cons_s_iter: forall s C b s' s'' C' b',
    s_iter s s' ->
    cstep_star_snoc_I s' C b s'' C' b' ->
    cstep_star_snoc_I s C b s'' C' b'.
  Proof.
    intros.
    induction H0.
    eapply cstep_snoc_s_iter; eauto.
    apply cstep_snoc_nil.
    eapply cstep_snoc_cons; eauto.
    eapply cstep_snoc_s_iter; eauto.
  Qed.

  Lemma cstep_star_snoc_equiv: forall s C b s'' C'' b'',
    cstep_star s C b s'' C'' b'' <-> cstep_star_snoc_I s C b s'' C'' b''.
  Proof.
    split; intros.
    * induction H.
    - apply cstep_snoc_nil.
    - eapply cstep_cons_snoc; eauto.
    - eapply cstep_snoc_cons_s_iter; eauto.
    * induction H.
    - apply cstep_nil.
    - eapply cstep_star_snoc; eauto.
    - eapply cstep_star_app; eauto.
    eapply cstep_iteration; eauto.
    apply cstep_nil.
  Qed.


  Lemma terminates_s_iter_star: forall wC s b s',
    terminates s wC b ->
    s_iter_star s s' ->
    terminates s' wC b.
  Proof.
    intros.
    induction H0; auto.
    apply IHs_iter_star.
    eapply terminates_s_iter; eauto.
  Qed.

  Lemma cannot_communicate_channels_s_iter_star: forall s b s',
    cannot_communicate_channels s b ->
    s_iter_star s s' ->
    cannot_communicate_channels s' b.
  Proof.
    intros.
    induction H0; auto.
    apply IHs_iter_star.
    eapply cannot_communicate_channels_s_iter; eauto.
  Qed.

  Lemma valid_machine_state_s_iter_star: forall wO s b s',
    valid_machine_state wO s b ->
    s_iter_star s s' ->
    valid_machine_state wO s' b.
  Proof.
    intros.
    induction H0; auto.
    apply IHs_iter_star.
    eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma noninterfering_schedule_step_star_step: forall wO wO2 s b1 w2 b2 s' b2' s'' b1',
    noninterfering_wterm b1 b2 ->
    join_wterms b2 w2 ->
    gjoin wO w2 wO2 ->
    terminates s wO2 b1 ->
    step_star s b2 s' b2' ->
    step s' b1 s'' b1' ->
    cannot_communicate_channels s (b1||b2) ->
    valid_machine_state wO s (b1||b2) ->
    exists s'0 s'1,
      s_iter_star s s'0 /\
      step s'0 b1 s'1 b1' /\ step_star s'1 b2 s'' b2'.
  Proof.
    introv H Hjb2 HwO2 Hterm1 H0 H1 Hcc Hvms.
    revert s'' w2 wO2 Hjb2 H1 H HwO2 Hcc Hterm1 Hvms.
    induction H0; intros.
    - exists s s''; splits; auto. apply s_iter_nil. apply step_nil.
    - rename b into b2; rename b' into b2'; rename b'' into b2''.
    assert (Hvms': valid_machine_state wO s' (b1 || b2')).
      eapply valid_machine_state_step; eauto.
      apply st_par2; eauto.
    assert (exists w1 w2' wO2', join_wterms b1 w1 /\ join_wterms b2' w2' /\ gjoin wO w2' wO2').
      destruct Hvms' as [wS'[wb'[ww'[?[??]]]]].
      destruct H4 as [w1[w2'[?[??]]]].
      apply gjoin_join_list in H7.
      jrewrite<- H7 in H5.
      edestruct join_list_app_inv with (wl1:=[wO,w2']) (wl2:=[wS',w1]) as [wO2'[wS'1[?[??]]]]; eauto.
      eapply join_list_perm; eauto.  permut4.
      exists w1 w2' wO2'; splits; auto.
      apply gjoin_join_list; auto.
    destruct H3 as [w1[w2'[wO2'[?[??]]]]].
    destruct IHstep_star with s''0 w2' wO2' as [s'2[s'3[?[??]]]]; clear IHstep_star; auto. 
    eapply noninterfering_step2; eauto.
    eapply cannot_communicate_channels_par2; eauto.
    eapply cannot_communicate_channels_step; eauto. apply st_par2; auto.
    eapply terminates_stepO' with (bO:=(wO:tm_skip||b2)%wterm); eauto.
    apply st_par2; eauto.
    exists wO w2; splits; auto; reflexivity.
    exists wO w2'; splits; auto; try reflexivity.
    eapply valid_machine_state_par_observer with (bO:=b2); eauto.
    edestruct step_commute_front_s_iter_star with (b:=b2) as [s'1[??]]; eauto.
    clear H H0.
    edestruct noninterfering_schedule_step' with (s:=s'1) as [s'4[??]]; eauto.
    eapply terminates_s_iter_star; eauto.
    eapply terminates_sub; eauto. exists wO; auto.
    eapply valid_machine_state_par_observer with (bO:=b2); eauto.
    eapply cannot_communicate_channels_s_iter_star; eauto.
    eapply valid_machine_state_s_iter_star; eauto.
    eapply valid_machine_state_subO; eauto.
    exists wO; apply gjoin_bot2.
    exists s'1 s'4; splits; auto.
    eapply step_star_app; eauto.
    - rename b into b2; rename b'' into b2''.
    destruct IHstep_star with s''0 w2 wO2 as [s'2[s'3[?[??]]]]; clear IHstep_star; auto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply terminates_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
    exists s'2 s'3; splits; auto.
    eapply s_iter_cons; eauto.
  Qed.

  Definition w_send_share w c := match w_prod w c with
  | Some (exist (exist p _) _) => fst p
  | _ => Share.bot
  end.

  Definition w_recv_share w c := match w_cons w c with
  | Some (exist (exist p _) _) => fst p
  | _ => Share.bot
  end.

  Lemma can_send_preserve_share_step: forall s w w1 w2 b1 c s' w' w1' b1',
    can_send c w ->
    cannot_communicate_channels s b1 ->
    join_wterms b1 w1 ->
    join_wterms b1' w1' ->
    gjoin w1 w2 w ->
    gjoin w1' w2 w' ->
    step s b1 s' b1' ->
    w_send_share w' c = w_send_share w c.
Admitted.

  Lemma can_send_preserve_step: forall s w w1 w2 b1 c s' w' w1' b1',
    can_send c w ->
    cannot_communicate_channels s b1 ->
    join_wterms b1 w1 ->
    join_wterms b1' w1' ->
    gjoin w1 w2 w ->
    gjoin w1' w2 w' ->
    step s b1 s' b1' ->
    can_send c w'.
Admitted. (* follows directly from can_send_preserve_share_step *)


  Lemma can_recv_preserve_share_step: forall s w w1 w2 b1 c s' w' w1' b1',
    can_recv c w ->
    cannot_communicate_channels s b1 ->
    join_wterms b1 w1 ->
    join_wterms b1' w1' ->
    gjoin w1 w2 w ->
    gjoin w1' w2 w' ->
    step s b1 s' b1' ->
    w_recv_share w' c = w_recv_share w c.
Admitted.

  Lemma can_recv_preserve_step: forall s w w1 w2 b1 c s' w' w1' b1',
    can_recv c w ->
    cannot_communicate_channels s b1 ->
    join_wterms b1 w1 ->
    join_wterms b1' w1' ->
    gjoin w1 w2 w ->
    gjoin w1' w2 w' ->
    step s b1 s' b1' ->
    can_recv c w'.
Admitted. (* follows from can_recv_preserve_share_step *)

  Lemma total_send_recv_step: forall s w w1 w2 b1 c s' w' w1' b1',
    total_send_recv c w ->
    cannot_communicate_channels s b1 ->
    join_wterms b1 w1 ->
    join_wterms b1' w1' ->
    gjoin w1 w2 w ->
    gjoin w1' w2 w' ->
    step s b1 s' b1' ->
    total_send_recv c w'.
Admitted. (* follows from can_send_preserve_share_step and can_recv_preserve_share_step *)


  Lemma private_communications_step2: forall s w1 w2 w b2 s' w2' w' b2',
    private_communications w1 w ->
    cannot_communicate_channels s b2 ->
    gjoin w1 w2 w ->
    join_wterms b2 w2 ->
    join_wterms b2' w2' ->
    gjoin w1 w2' w' ->
    step s b2 s' b2' ->
    private_communications w1 w'.
  Proof.
    introv Hprivate Hcc Hw Hw2 Hw2' Hw' Hs.
    red; intros.
    eapply total_send_recv_step with (b1:=b2) (w1:=w2) (b1':=b2') (w1':=w2') (w:=w); eauto.
    apply gjoin_com; eauto.
    apply gjoin_com; eauto.
  Qed.

  Lemma private_communications_step: forall s w1 w2 w12 w3 w b1 s' w1' w12' w' b1',
    private_communications w12 w ->
    cannot_communicate_channels s b1 ->
    gjoin w1 w2 w12 ->
    gjoin w12 w3 w ->
    join_wterms b1 w1 ->
    join_wterms b1' w1' ->
    gjoin w1' w2 w12' ->
    gjoin w12' w3 w' ->
    step s b1 s' b1' ->
    private_communications w12' w'.
  Proof.
    introv Hprivate Hcc Hw12 Hw Hw1 Hw1' Hw12' Hw' Hs.
    revert w1 w2 w12 w3 w w1' w12' w' Hprivate Hw Hw1 Hw1' Hw12 Hw12' Hw'.
    step_cases (induction Hs) Case; intros; eauto; simpl in *; subst; subst;
      try solve [replace w12' with w12 in *; [ replace w' with w0 in *; auto |]; eapply gjoin_eq; eauto].
    - Case "st_send".
admit. (* cannot_communicate_channels implies that channel resources do not change *)
    - Case "st_recv".
admit. (* cannot_communicate_channels implies that channel resources do not change *)
    - Case "st_newch".
admit. (* cannot_communicate_channels implies that channel resources do not change *)
    - Case "st_endch".
admit. (* cannot_communicate_channels implies that channel resources do not change *)
    - Case "st_assign".
admit. (* does not change channel resources *)
    - Case "st_load".
admit. (* does not change channel resources *)
    - Case "st_store".
admit. (* does not change channel resources *)
    - Case "st_local".
admit. (* does not change channel resources *)
    - Case "st_freelocal".
admit. (* does not change channel resources *)
    - Case "st_par".
    destruct Hw1' as [w1_[w2_[?[??]]]]; subst w1_ w2_.
    replace w1' with w0 in *.
    replace w12' with w12 in *.
    replace w' with w5 in *; auto.
    eapply gjoin_eq; eauto.
    eapply gjoin_eq; eauto.
    eapply gjoin_eq; eauto.
    - Case "st_seq1".
    eapply IHHs; eauto.
    eapply cannot_communicate_channels_seq; eauto.
    - Case "st_par1".
    destruct Hw1 as [w1a[w2a[?[??]]]].
    destruct Hw1' as [w1a'[w2a'[?[??]]]].
    replace w2a' with w2a in *.
    edestruct gjoin_assoc with (a:=w1a) (b:=w2a) as [?[??]]; eauto.
    eapply IHHs with (w1:=w1a) (w1':=w1a'); eauto.
    eapply cannot_communicate_channels_par1; eauto.
    edestruct gjoin_assoc with (a:=w1a') (b:=w2a) as [?[??]]; eauto.
    replace x0 with x in *; auto.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    - Case "st_par2".
    destruct Hw1 as [w1a[w2a[?[??]]]].
    destruct Hw1' as [w1a'[w2a'[?[??]]]].
    replace w1a' with w1a in *.
    apply gjoin_com in H1.
    edestruct gjoin_assoc with (a:=w2a) (b:=w1a) as [?[??]]; eauto.
    eapply IHHs with (w1:=w2a) (w1':=w2a'); eauto.
    eapply cannot_communicate_channels_par2; eauto.
    apply gjoin_com in H4.
    edestruct gjoin_assoc with (a:=w2a') (b:=w1a) as [?[??]]; eauto.
    replace x0 with x in *; auto.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    - Case "st_par_join".
    destruct Hw1 as [w1_[w2_[?[??]]]]; subst w1_ w2_.
    replace w1' with w0 in *.
    replace w12' with w12 in *.
    replace w' with w5 in *; auto.
    eapply gjoin_eq; eauto.
    eapply gjoin_eq; eauto.
    eapply gjoin_eq; eauto.
  Qed.


  Lemma private_communications_schedule_step_star_stepO: forall s wO bO wb w2 w b s' b' s'' bO'',
    join_wterms b w ->
    join_wterms bO wO ->
    gjoin w w2 wb ->
    private_communications w wb ->
    cannot_communicate_channels s b ->
    valid_machine_state wO s (b||w2:tm_skip) ->
    step_star s b s' b' ->
    step s' bO s'' bO'' ->
    exists s'0 s'1,
      s_iter_star s s'0 /\
      step s'0 bO s'1 bO'' /\
      step_star s'1 b s'' b'.
  Proof.
    introv Hw HwO Hwb Hprivate Hcc Hvms Hss Hs.
    revert s'' w2 w wb wO Hw HwO Hwb Hprivate Hcc Hvms Hs.
    induction Hss; intros.
    - exists s s''; splits; auto. apply s_iter_nil. apply step_nil.
    -
    assert (Hvms': valid_machine_state wO s' (b' || w2:tm_skip)).
      eapply valid_machine_state_step; eauto.
      apply st_par1; eauto.
    assert (exists w' wb' wOb', join_wterms b' w' /\ gjoin w' w2 wb' /\ gjoin wO wb' wOb').
      destruct Hvms' as [wS'[wb'[ww'[?[??]]]]].
      destruct H1 as [w'[w2_[?[??]]]].
      simpl in H3; subst w2_.
      edestruct join_list_app_inv with (wl1:=[wO,wb']) (wl2:=[wS']) as [wOb'[wS'1[?[??]]]]; eauto.
      eapply join_list_perm; eauto.  permut4.
      exists w' wb' wOb'; splits; auto.
      apply gjoin_join_list; auto.
    destruct H0 as [w'[wb'[wOb'[?[??]]]]].
    destruct IHHss with s''0 w2 w' wb' wO as [s'2[s'3[?[??]]]]; clear IHHss; auto. 
    eapply private_communications_step; eauto.
    apply gjoin_bot1. 
    apply gjoin_bot1. 
    apply gjoin_com in H2.
    edestruct gjoin_assoc with (a:=w') (b:=w2) as [?[??]]; eauto.
    apply gjoin_com in H3.
    eapply cannot_communicate_channels_step; eauto.
    eapply valid_machine_state_par_observer with (bO:=(w2:tm_skip)%wterm) (wO1:=wO) (wO2:=w2); eauto.
    reflexivity.    
    edestruct step_commute_front_s_iter_star with (b:=b) as [s'1[??]]; eauto.
    (**)
    edestruct no_conflict_schedule_stepO with (b:=b) (bO:=bO) (s:=s'1) as [s'4[??]]; eauto.
    exists w2; auto.
    destruct Hvms as [wS[wbw2[ww[?[??]]]]].
    destruct H9 as [w_[w2_[?[??]]]].
    simpl in H11; subst w2_.
    replace w_ with w in *.
    replace wbw2 with wb in *.
    edestruct join_list_app_inv with (wl1:=[wO,wb]) (wl2:=[wS]) (w:=ww); eauto.
    eapply join_list_perm; eauto.
    permut4.
    destruct H11 as [_[? _]].
    exists x; apply gjoin_join_list; auto.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    exists s'1 s'4; splits; auto.
    eapply step_cons; eauto.
    - destruct IHHss with s''0 w2 w wb wO as [s'2[s'3[?[??]]]]; clear IHHss; auto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
    exists s'2 s'3; splits; auto.
    eapply s_iter_cons; eauto.
  Qed.

  Fixpoint context_size C := match C with
  | cx_hole => 0
  | cx_seq C t => 1 + context_size C + term_size t
  | cx_par1 C b2 => 1 + (context_size C) + (wterm_size b2)
  | cx_par2 b1 C => 1 + (wterm_size b1) + (context_size C)
  end.

  Lemma context_size_intro: forall C1 C2,
    C1 = C2 -> context_size C1 = context_size C2.
  Proof.
    intros; rewrite H; auto.
  Qed.

  Lemma set_context_size: forall C b,
    wterm_size C[b] = context_size C + wterm_size b.
  Proof.
    induction C; intros; simpl; auto; rewrite IHC; omega.
  Qed.

Ltac absurd_rec_def :=
  match goal with
  | [ H : (?t = ?E) |- _ ] => match E with
    | context[t] => elimtype False;
        first [apply context_size_intro in H | apply wterm_size_intro in H | apply term_size_intro in H];
        repeat (simpl in H; rewrite set_context_size in H); simpl in H; omega
    end
  | [ H : (?E = ?t) |- _ ] => match E with
    | context[t] => elimtype False;
        first [apply context_size_intro in H | apply wterm_size_intro in H | apply term_size_intro in H];
        repeat (simpl in H; rewrite set_context_size in H); simpl in H; omega
    end
  end.



  Lemma context_schedule_cstep: forall w2 wb s w b s' wC C b' s'' C',
    join_wterms b w ->
    context_world C wC ->
    gjoin w w2 wb ->
    private_communications w wb ->
    cannot_communicate_channels s b ->
    valid_machine_state wC s (b || w2:tm_skip) ->
    step_star s b s' b' ->
    cstep s' C (gwemp:tm_skip) s'' C' (gwemp:tm_skip) ->
    exists s'0 s'1,
      s_iter_star s s'0 /\
      cstep s'0 C (gwemp:tm_skip) s'1 C' (gwemp:tm_skip) /\
      step_star s'1 b s'' b'.
  Proof.
    introv Hw HwC Hwb Hprivate Hcc Hvms Hss Hc.
    edestruct private_communications_schedule_step_star_stepO
      with (bO:=C[gwemp:tm_skip]) (b:=b) as [s'0[s'1[?[??]]]]; eauto.
    apply context_world_join_wterms; auto.
    apply step_simulates_cstep; eauto.
    exists s'0 s'1; splits; auto.
    eapply cstep_simulates_cstep_step; eauto.
  Qed.

  Lemma context_schedule_step_cstep: forall w2 wb s w b s' wC C b' s'' C',
    join_wterms b w ->
    context_world C wC ->
    gjoin w w2 wb ->
    gjoins wC wb ->
    private_communications w wb ->
    step s b s' b' ->
    cstep s' C (gwemp:tm_skip) s'' C' (gwemp:tm_skip) ->
    exists s'0,
      cstep s C (gwemp:tm_skip) s'0 C' (gwemp:tm_skip) /\
      step s'0 b s'' b'.
  Proof.
    introv Hw HwC Hwb HwCb Hprivate Hss Hc.
    edestruct no_conflict_schedule_stepO
      with (bO:=C[gwemp:tm_skip]) (b:=b) as [s'0[??]]; eauto.
    apply context_world_join_wterms; eauto.
    exists w2; auto.
    apply step_simulates_cstep; eauto.
    exists s'0; split; auto.
    eapply cstep_simulates_cstep_step; eauto.
  Qed.

  Lemma context_schedule_first: forall w2 wb s w b wC C b'' s'' C'',
    join_wterms b w ->
    context_world C wC ->
    gjoin w w2 wb ->
    private_communications w wb ->
    cannot_communicate_channels s b ->
    valid_machine_state wC s (b || w2:tm_skip) ->
    cstep_star s C b s'' C'' b'' ->
    exists s',
      cstep_star s C (gwemp:tm_skip) s' C'' (gwemp:tm_skip) /\
      step_star s' b s'' b''.
  Proof.
    introv Hw HwC Hwb Hprivate Hcc Hvms Hcs.
    revert w2 wb w wC Hw HwC Hwb Hprivate Hcc Hvms.
    apply cstep_star_snoc_equiv in Hcs.
    induction Hcs; intros.
    - exists s; split. apply cstep_nil. apply step_nil.
    - apply cstep_star_snoc_equiv in Hcs.
    edestruct IHHcs as [s'0[??]]; clear IHHcs; eauto.
    clear Hcs.
    edestruct cstep_inversion with (s:=s') as [[??]|[??]]; eauto; clear H.
    subst C''.
    exists s'0; split; auto.
      eapply step_star_snoc; eauto.
    subst b''.
    edestruct valid_machine_state_cstep_star with (b:=(b||w2:tm_skip)%wterm) as [wC'[? Hvms']]; eauto.
    eapply cstep_star_skip_subst; eauto.
    edestruct context_schedule_cstep with (s:=s'0) (s':=s') (C':=C'') as [s'1[s'2[?[??]]]]; eauto.
    eapply cannot_communicate_channels_cstep_star with (C:=C); eauto.
    eapply cstep_star_skip_subst; eauto.
    eapply valid_machine_state_par1_sub; eauto.
    exists s'2; split; auto.
      eapply cstep_star_app; eauto.
      eapply cstep_star_snoc; eauto.
      eapply cstep_star_s_iter_set; eauto.
    - apply cstep_star_snoc_equiv in Hcs.
    edestruct IHHcs as [s'0[??]]; clear IHHcs; eauto.
    exists s'0; split; auto.
    eapply step_star_app; eauto.
    eapply st_iteration; eauto; apply step_nil.
  Qed.


(*
  Lemma noninterfering_schedule_step_star: forall s b1 b2 s' b2' s'' b1',
    noninterfering_wterm b1 b2 ->
    b2 <> b2' ->
    step s b2 s' b2' ->
    step_star s' b1 s'' b1' ->
    exists s'0,
         step_star s b1 s'0 b1'
    /\ step s'0 b2 s'' b2'.
  Proof.
    introv H Hnospin2 H0 H1.
    revert s H0 H.
    induction H1; intros.
    exists s0; split; [apply step_nil|auto].
    rename t into b1; rename t' into b1'; rename t'' into b1''.
    rename s into s'0; rename s0 into s.
    edestruct noninterfering_schedule_step as [s'1[??]]; eauto.
    destruct IHstep_star with s'1 as [s'2[??]]; auto.
    eapply noninterfering_preservation1; eauto.
    apply step_single_cstep_star with (C:=cx_hole); eauto.
    exists s'2; split; auto.
    eapply step_cons; eauto.
  Qed.
*)





(*
    introv H Hjb2 H0 H1 Hvms.
    revert s'' w2 Hjb2 H1 H Hvms.
    induction H0; intros.
    - exists s s''; splits; auto. apply s_iter_nil. apply step_nil.
    - rename b into b2; rename b' into b2'; rename b'' into b2''.
    edestruct valid_machine_state_step with (b:=(b1||b2)%wterm) as [wS'[w12'[ww'[?[??]]]]]; eauto.
    apply st_par2; eauto.
    destruct H4 as [w1[w2'[?[??]]]].
    destruct IHstep_star with s''0 w2' as [s'2[s'3[?[??]]]]; clear IHstep_star; auto.
(*    eapply terminates_preservation with (b2:=b2); eauto.
    apply step_single_cstep_star; apply st_par2; auto.*)
    eapply noninterfering_preservation2; eauto.
    apply step_single_cstep_star with (C:=dummyContext wO); eauto.
    eapply valid_machine_state_step_star; eauto.
    apply step_single; apply st_par2; auto.
    edestruct step_commute_front_s_iter_star with (b:=b2) as [s'1[??]]; eauto.
    clear H H8.
Check noninterfering_schedule_step'.

noninterfering_schedule_step
     : forall (s : state) (w1 : world) (b1 : wterm) (w2 : world) (b2 : wterm)
         (s' : state) (b2' : wterm) (s'' : state) (b1'' : wterm),
       join_wterms b1 w1 ->
       join_wterms b2 w2 ->
       no_send_conflict w2 w1 ->
       step s b2 s' b2' ->
       step s' b1 s'' b1'' ->
       exists s'0, step s b1 s'0 b1'' /\ step s'0 b2 s'' b2'


Check noninterfering_schedule_step.
    edestruct noninterfering_schedule_step' with (s:=s'1) as [s'4[??]]; eauto.
(*    eapply terminates_preservation with (b2:=b2); eauto.
    eapply cstep_star_s_iter_set; eauto.*)
    edestruct valid_machine_state_join_wterms as [w ?]; eauto.
    destruct H as [w1_[w2_[?[??]]]]; eauto.
    replace w1_ with w1 in *.
    replace w2_ with w2 in *.
    exists w; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.

    exists s'1 s'4; splits; auto.
    eapply step_star_app; eauto.
    - rename b into b2; rename b'' into b2''.
    destruct IHstep_star with s''0 w2 as [s'2[s'3[?[??]]]]; clear IHstep_star; auto.
    eapply terminates_preservation with (b2:=b2); eauto.
    eapply cstep_iteration; eauto. apply cstep_nil.
    eapply valid_machine_state_s_iter; eauto.
    exists s'2 s'3; splits; auto.
    eapply s_iter_cons; eauto.
  Qed.
*)

  Lemma private_communications_step_star: forall s w1 w2 w b1 s' w1' w' b1',
    private_communications w1 w ->
    cannot_communicate_channels s b1 ->
    valid_machine_state w2 s b1 ->
    gjoin w1 w2 w ->
    join_wterms b1 w1 ->
    join_wterms b1' w1' ->
    gjoin w1' w2 w' ->
    step_star s b1 s' b1' ->
    private_communications w1' w'.
  Proof.
    introv Hprivate Hcc Hvms Hw Hw1 Hw1' Hw' Hss.
    revert w1 w2 w w1' w' Hprivate Hcc Hvms Hw Hw1 Hw1' Hw'.
    induction Hss; intros.
    - replace w1' with w1 in *.
    replace w' with w in *; auto.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    - renames b to b1, b' to b1', b'' to b1'', w1' to w1'', w' to w''.
    assert (Hvms': valid_machine_state w2 s' b1').
      eapply valid_machine_state_step; eauto.
    edestruct valid_machine_state_join_wterms as [w1' ?]; eauto.
    assert (exists w', gjoin w1' w2 w').
      destruct Hvms' as [wS'[w1'_[ww'[?[??]]]]].
      replace w1'_ with w1' in *.
      edestruct join_list_app_inv with (wl1:=[w1',w2]) (wl2:=[wS']) as [w'[wS'_[?[??]]]]; eauto.
      eapply join_list_perm; eauto. permut4.
      exists w'; apply gjoin_join_list; auto.
      eapply join_wterms_eq; eauto.
    destruct H1 as [w' ?].
    eapply IHHss with (w1:=w1') (w:=w'); eauto.
    eapply private_communications_step; eauto.
    apply gjoin_bot1. apply gjoin_bot1.
    eapply cannot_communicate_channels_step; eauto.
    - renames b to b1, b'' to b1'.
    eapply IHHss with (w1:=w1) (w:=w); eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma private_communications_step_star1a: forall s w w1 b1 b2 s' w' w1' b1',
    private_communications w1 w ->
    cannot_communicate_channels s b1 ->
    valid_machine_state gwemp s (b1||b2) ->
    join_wterms b1 w1 ->
    join_wterms b1' w1' ->
    join_wterms (b1||b2)%wterm w ->
    join_wterms (b1'||b2)%wterm w' ->
    step_star s b1 s' b1' ->
    private_communications w1' w'.
  Proof.
    introv Hprivate Hcc Hvms Hw1 Hw1' Hw Hw' Hss.
    destruct Hw as [w1_[w2[?[??]]]].
    replace w1_ with w1 in *. clear w1_.
    destruct Hw' as [w1'_[w2_[?[??]]]].
    replace w1'_ with w1' in *. clear w1'_.
    replace w2_ with w2 in *. clear w2_.
    edestruct valid_machine_state_par_observer' as [w2_[wO2[?[??]]]]; eauto.
    replace w2_ with w2 in *. clear w2_.
    apply gworld_unit_identity in H6; subst wO2.
    eapply private_communications_step_star with (w2:=w2) (w1:=w1) (w1':=w1'); eauto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma private_communications_step_star2: forall s w1 w2 w b2 s' w2' w' b2',
    private_communications w1 w ->
    cannot_communicate_channels s b2 ->
    valid_machine_state w1 s b2 ->
    gjoin w1 w2 w ->
    join_wterms b2 w2 ->
    join_wterms b2' w2' ->
    gjoin w1 w2' w' ->
    step_star s b2 s' b2' ->
    private_communications w1 w'.
  Proof.
    introv Hprivate Hcc Hvms Hw Hw2 Hw2' Hw' Hss.
    revert w1 w2 w w2' w' Hprivate Hcc Hvms Hw Hw2 Hw2' Hw'.
    induction Hss; intros.
    - replace w2' with w2 in *.
    replace w' with w in *; auto.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    - renames b to b2, b' to b2', b'' to b2'', w2' to w2'', w' to w''.
    assert (Hvms': valid_machine_state w1 s' b2').
      eapply valid_machine_state_step; eauto.
    edestruct valid_machine_state_join_wterms as [w2' ?]; eauto.
    assert (exists w', gjoin w1 w2' w').
      destruct Hvms' as [wS'[w2'_[ww'[?[??]]]]].
      replace w2'_ with w2' in *.
      edestruct join_list_app_inv with (wl1:=[w1,w2']) (wl2:=[wS']) as [w'[wS'_[?[??]]]]; eauto.
      eapply join_list_perm; eauto. permut4.
      exists w'; apply gjoin_join_list; auto.
      eapply join_wterms_eq; eauto.
    destruct H1 as [w' ?].
    eapply IHHss with (w2:=w2') (w:=w'); eauto.
    eapply private_communications_step2; eauto.
    eapply cannot_communicate_channels_step; eauto.
    - renames b to b1, b'' to b1'.
    eapply IHHss with (w1:=w1) (w:=w); eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma private_communications_step_star2a: forall s w w1 b1 b2 s' w' b2',
    private_communications w1 w ->
    cannot_communicate_channels s b2 ->
    valid_machine_state gwemp s (b1||b2) ->
    join_wterms b1 w1 ->
    join_wterms (b1||b2)%wterm w ->
    join_wterms (b1||b2')%wterm w' ->
    step_star s b2 s' b2' ->
    private_communications w1 w'.
  Proof.
    introv Hprivate Hcc Hvms Hw1 Hw Hw' Hss.
    destruct Hw as [w1_[w2[?[??]]]].
    replace w1_ with w1 in *. clear w1_.
    destruct Hw' as [w1_[w2'[?[??]]]].
    replace w1_ with w1 in *. clear w1_.
    edestruct valid_machine_state_par_observer' as [w2_[wO2[?[??]]]]; eauto.
    replace w2_ with w2 in *. clear w2_.
    apply gworld_unit_identity in H6; subst wO2.
    eapply private_communications_step_star2 with (w2:=w2) (w1:=w1) (w2':=w2'); eauto.
    eapply valid_machine_state_par_observer with (wO1:=gwemp) (wO2:=w1) (bO:=b1); auto.
    apply gjoin_bot2.
    apply valid_machine_state_par_swap; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma private_communications_step_star2b: forall s w w2 b1 b2 s' w' w2' b2',
    private_communications w2 w ->
    cannot_communicate_channels s b2 ->
    valid_machine_state gwemp s (b2||b1) ->
    join_wterms b2 w2 ->
    join_wterms b2' w2' ->
    join_wterms (b1||b2)%wterm w ->
    join_wterms (b1||b2')%wterm w' ->
    step_star s b2 s' b2' ->
    private_communications w2' w'.
  Proof.
    introv Hprivate Hcc Hvms Hw2 Hw2' Hw Hw' Hss.
    eapply private_communications_step_star1a; eauto.
    apply join_wterm_worlds in Hw; apply join_wterm_worlds.
    eapply join_list_perm; eauto. simpl; permut4.
    apply join_wterm_worlds in Hw'; apply join_wterm_worlds.
    eapply join_list_perm; eauto. simpl; permut4.
  Qed.

  Lemma private_communications_step_star1b: forall s w w2 b1 b2 s' w' b1',
    private_communications w2 w ->
    cannot_communicate_channels s b1 ->
    valid_machine_state gwemp s (b1||b2) ->
    join_wterms b2 w2 ->
    join_wterms (b1||b2)%wterm w ->
    join_wterms (b1'||b2)%wterm w' ->
    step_star s b1 s' b1' ->
    private_communications w2 w'.
  Proof.
    introv Hprivate Hcc Hvms Hw2 Hw Hw' Hss.
    destruct Hw as [w1[w2_[?[??]]]].
    replace w2_ with w2 in *. clear w2_.
    destruct Hw' as [w1'[w2_[?[??]]]].
    replace w2_ with w2 in *. clear w2_.
    apply valid_machine_state_par_swap in Hvms.
    edestruct valid_machine_state_par_observer' as [w1_[wO1[?[??]]]]; eauto.
    replace w1_ with w1 in *. clear w1_.
    apply gworld_unit_identity in H6; subst wO1.
    eapply private_communications_step_star2 with (w1:=w2) (w2:=w1) (w2':=w1'); eauto.
    eapply valid_machine_state_par_observer with (wO1:=gwemp) (wO2:=w2) (bO:=b2); auto.
    apply gjoin_bot2.
    apply valid_machine_state_par_swap; auto.
    apply gjoin_com; auto.
    apply gjoin_com; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
  Qed.


  Lemma noninterfering_schedule_step_star_cstep_star: forall wO wO2 w12 s w1 b1 w2 b2 s' C b2' s'' C' b1',
    noninterfering_wterm b1 b2 ->
    join_wterms b1 w1 ->
    join_wterms b2 w2 ->
    gjoin w1 w2 w12 ->
    gjoin wO w2 wO2 ->
    terminates s wO2 b1 ->
    step_star s b2 s' b2' ->
    cstep_star s' C b1 s'' C' b1' ->
    context_world C wO ->
    private_communications w2 w12 ->
    cannot_communicate_channels s (b1||b2) ->
    valid_machine_state wO s (b1||b2) ->
    exists s'0,
      cstep_star s C b1 s'0 C' b1' /\ step_star s'0 b2 s'' b2'.
  Proof.
    introv Hni Hw1 Hw2 Hw12 HwO2 Hterm1 Hss Hcs HwC.
    introv Hprivate Hcc Hvms.
    revert wO s w1 w2 w12 wO2 Hss Hni Hw1 Hw2 Hw12 HwO2 Hterm1 HwC Hprivate Hcc Hvms.
    induction Hcs; intros.
    - exists s0; split; auto. apply cstep_nil.
    - renames b to b1, b' to b1', b'' to b1''.
    renames s to s'0, s0 to s.
    edestruct cstep_inversion with (s:=s'0) as [[??]|[??]]; eauto; clear H; subst.
    + edestruct noninterfering_schedule_step_star_step as [s'1[s'2[?[??]]]]; eauto.
    clear Hss H0.
    assert (step_star s b1 s'2 b1').
      eapply step_star_app.
      eapply s_iter_step_star; eauto.
      apply step_single; auto.
    assert (Hvms': valid_machine_state wO s'2 (b1'||b2)).
      eapply valid_machine_state_step_star; eauto.
      apply step_star_par1_step; auto.
    edestruct valid_machine_state_join_wterms with (s:=s'2) as [w12' ?]; eauto.
    destruct H3 as [w1'[w2_[?[??]]]].
    replace w2_ with w2 in *.
    destruct IHHcs with wO s'2 w1' w2 w12' wO2 as [s'3[??]]; clear IHHcs; auto.
    eapply noninterfering_step_star1; eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply terminates_step_star; eauto.
    eapply private_communications_step_star1b with (b1:=b1) (b2:=b2) (b1':=b1'); eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply valid_machine_state_subO; eauto.
    exists wO; apply gjoin_bot2.
    exists w1 w2; splits; auto.
    exists w1' w2; splits; auto.
    eapply cannot_communicate_channels_step_star; eauto.
    apply step_star_par1_step; auto.
    exists s'3; split; auto.
    eapply cstep_star_app.
    apply cstep_star_s_iter_set; eauto.
    eapply step_cstep_star; eauto.
    eapply join_wterms_eq; eauto.
    + edestruct context_schedule_cstep with (C':=C') (b:=b2) (w:=w2) (wb:=w12) as [s'1[s'2[?[??]]]]; eauto.
    apply gjoin_com; eauto.
    eapply cannot_communicate_channels_par2; eauto.
    apply valid_machine_state_par_swap.
    eapply valid_machine_state_wterm; intros; eauto.
    exists w1 w2; splits; eauto.
    exists w1 w2; splits; eauto; reflexivity.
    clear Hss H0.
    edestruct valid_machine_state_cstep_star with (C:=C) (s:=s) (b:=(b1||b2)%wterm) as [wO'[??]]; eauto.
    eapply cstep_star_snoc.
    eapply cstep_star_s_iter_set; eauto.
    apply cstep_skip_subst with gwemp; eauto.
    assert (exists wO'2, gjoin wO' w2 wO'2).
      edestruct valid_machine_state_par_observer' with (s:=s'2) as [w2_[wO'2[?[??]]]]; eauto.
      replace w2_ with w2 in *.
      exists wO'2; auto.
      eapply join_wterms_eq; eauto.
    destruct H4 as [wO'2 ?].
    destruct IHHcs with wO' s'2 w1 w2 w12 wO'2 as [s'3[??]]; clear IHHcs; auto.
    eapply terminates_step_starO'; eauto.
    eapply step_star_snoc.
    apply s_iter_step_star; eauto.
    apply step_simulates_cstep.
    apply cstep_skip_subst with gwemp; eauto.
    eapply join_wterms_context2; eauto.
    eapply join_wterms_context2; eauto.
    eapply valid_machine_state_par_observer; eauto.
    eapply cannot_communicate_channels_cstep_star with (wC:=wO) (s:=s); eauto.
    eapply cstep_star_snoc.
    apply cstep_star_s_iter_set; eauto.
    apply cstep_skip_subst with gwemp; eauto.
    exists s'3; split; auto.
    eapply cstep_star_app.
    apply cstep_star_s_iter_set; eauto.
    eapply cstep_cons; eauto.
    apply cstep_skip_subst with gwemp; auto.
    - destruct IHHcs with wO s0 w1 w2 w12 wO2 as [s'2[??]]; clear IHHcs; eauto.
    eapply step_star_iter_snoc; eauto.
  Qed.

(*
  Lemma noninterfering_schedule_step_star': forall wC wC2 s b1 w2 b2 s' b2' s'' b1',
    noninterfering_wterm b1 b2 ->
    join_wterms b2 w2 ->
    no_send_conflict wC2 w2 ->
    join wC w2 wC2 ->
    terminates s wC2 b1 -> (*step b2 does not enable termination of b1*)
    step s b2 s' b2' ->
    step_star s' b1 s'' b1' ->
    valid_machine_state wC s (b1 || b2) ->
    exists s'0,
      step_star s b1 s'0 b1' /\ step_star s'0 b2 s'' b2'.
  Proof.
Admitted.
    introv H Hjb2 Hnsc HwC2 Hterm1 H0 H1 Hvms.
    edestruct noninterfering_schedule_step_star_cstep_star
      with (wO:=wC) (C':=dummyContext wC) as [s'0[??]]; eauto.
    red; intros.
    replace wb with w2 in *.
    apply context_world_join_wterms in H2.
    apply context_world_dummyContext2 in H2.
    subst; auto.
    eapply join_wterms_eq; eauto.
    apply step_single; eauto.
    apply step_star_add_context_cstep_star; eauto.
     apply context_world_dummyContext.
    exists s'0; split; auto.
    eapply cstep_star_dummyContext; eauto.
  Qed.
*)


  Lemma noninterfering_schedule_step_star_step_star': forall wO wO2 s b1 w2 b2 s' b2' s'' b1',
    noninterfering_wterm b1 b2 ->
    join_wterms b2 w2 ->
(* could be wO instead of wO2? *)
    gjoin wO w2 wO2 ->
    terminates s wO2 b1 -> (*step b2 does not enable termination of b1*)
    step_star s b2 s' b2' ->
    step_star s' b1 s'' b1' ->
    cannot_communicate_channels s (b1||b2) ->
    valid_machine_state wO s (b1||b2) ->
    exists s'0,
      step_star s b1 s'0 b1' /\ step_star s'0 b2 s'' b2'.
  Proof.
    introv H Hjb2 HwO2 Hterm1 H0 H1 Hcc Hvms.
    revert s Hjb2 HwO2 Hterm1 H0 H Hcc Hvms.
    induction H1; intros.
    - exists s0; split; auto. apply step_nil.
    - rename b into b1; rename b' into b1'; rename b'' into b1''.
    edestruct noninterfering_schedule_step_star_step as [s'1[s'2[?[??]]]]; eauto.
    assert (Hss: step_star s0 b1 s'2 b1').
      eapply step_star_app.
      apply s_iter_step_star; eauto.
      eapply step_single; auto.
    clear H3 H4.
    assert (Hcc': cannot_communicate_channels s'2 (b1'||b2)).
      eapply cannot_communicate_channels_step_star; eauto.
      apply step_star_par1_step; auto.
    destruct IHstep_star with s'2 as [s'3[??]]; auto.
    eapply terminates_step_star; eauto.
    eapply noninterfering_step_star1; eauto.
    apply cannot_communicate_channels_par1 with b2; eauto.
    eapply valid_machine_state_step_star; eauto.
      apply step_star_par1_step; auto.
    exists s'3; split; auto.
    eapply step_star_app; eauto.
    - renames b to b1, b'' to b1''.
    destruct IHstep_star with s0 as [s'3[??]]; eauto.
    eapply step_star_iter_snoc; eauto.
  Qed.

(*
  Lemma no_send_conflict_step: forall s wC w1 b1 s' w1' b1' w2 w w',
    no_send_conflict wC w ->
    join_wterms b1 w1 ->
    join_wterms b1' w1' ->
    gjoin w1 w2 w ->
    gjoin w1' w2 w' ->
    cannot_communicate_channels s b1 ->
    step s b1 s' b1' ->
    no_send_conflict wC w'.
  Proof.
    introv Hnosend Hw1 Hw1' Hw Hw' Hcc Hs.
    revert w1 w1' w2 Hnosend Hw1 Hw1' Hw Hw' Hcc.
    step_cases (induction Hs) Case; intros;
      simpl in *; subst; subst; eauto;
      try solve [replace w' with w in *; auto; eapply gjoin_eq; eauto].
    - Case "st_send".
admit. (* cannot communicate channel permissions *)
    - Case "st_recv".
admit. (* cannot communicate channel permissions *)
    - Case "st_newch".
admit. (* cannot communicate channel permissions *)
    - Case "st_endch".
admit. (* cannot communicate channel permissions *)
    - Case "st_assign".
admit. (* no change in channel permissions *)
    - Case "st_load".
admit. (* no change in channel permissions *)
    - Case "st_store".
admit. (* no change in channel permissions *)
    - Case "st_local".
admit. (* no change in channel permissions *)
    - Case "st_freelocal".
admit. (* no change in channel permissions *)
    - Case "st_par".
    destruct Hw1' as [w1_[w2_[?[??]]]]; subst w1_ w2_.
    replace w3 with w1' in *.
    replace w' with w in *; auto.
    eapply gjoin_eq; eauto.
    eapply gjoin_eq; eauto.
    - Case "st_seq1".
    eapply IHHs; eauto.
    eapply cannot_communicate_channels_seq; eauto.
    - Case "st_par1".
    renames w1 to w12, w1' to w12', w2 to w3.
    destruct Hw1 as [w1[w2[?[??]]]].
    destruct Hw1' as [w1'[w2_[?[??]]]].
    replace w2_ with w2 in *.
    edestruct gjoin_assoc with (a:=w1) (b:=w2) as [?[??]]; eauto.
    edestruct gjoin_assoc with (a:=w1') (b:=w2) as [?[??]]; eauto.
    replace x0 with x in *.
    apply IHHs with w1 w1' x; eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    - Case "st_par2".
    renames w1 to w12, w1' to w12', w2 to w3.
    destruct Hw1 as [w1[w2[?[??]]]].
    destruct Hw1' as [w1_[w2'[?[??]]]].
    replace w1_ with w1 in *.
    apply gjoin_com in H1.
    apply gjoin_com in H4.
    edestruct gjoin_assoc with (a:=w2) (b:=w1) as [?[??]]; eauto.
    edestruct gjoin_assoc with (a:=w2') (b:=w1) as [?[??]]; eauto.
    replace x0 with x in *.
    apply IHHs with w2 w2' x; eauto.
    eapply cannot_communicate_channels_par2; eauto.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    - Case "st_par_join".
    destruct Hw1 as [w1_[w2_[?[??]]]]; subst w1_ w2_.
    replace w3 with w1' in *.
    replace w' with w in *; auto.
    eapply gjoin_eq; eauto.
    eapply gjoin_eq; eauto.
  Qed.
*)
  Lemma private_communications_no_send_conflict: forall wO w1 w,
    private_communications w1 w ->
    gjoins wO w ->
    no_send_conflict wO w1.
  Proof.
    unfold private_communications, no_send_conflict; intros.
    edestruct H as [??]; eauto.
    split; intro.
    eapply has_total_endpoint_send_conflict; eauto; apply gjoins_com; auto.
    eapply has_total_endpoint_recv_conflict; eauto; apply gjoins_com; auto.
  Qed.

(*
  Lemma no_send_conflict_stepO: forall s wO bO w s' wO' bO',
    no_send_conflict wO w ->
    join_wterms bO wO ->
    join_wterms bO' wO' ->
    private_communications w ->
    step s bO s' bO' ->
    no_send_conflict wO' w.
  Proof.
    introv Hnosend HwO HwO' Hprivate Hs.
    revert wO w wO' Hnosend HwO HwO' Hprivate.
    step_cases (induction Hs) Case; intros;
      simpl in *; subst; subst; eauto.
    - Case "st_send".
admit. (* cannot communicate channel permissions *)
    - Case "st_recv".
admit. (* cannot communicate channel permissions *)
    - Case "st_newch".
admit. (* cannot communicate channel permissions *)
    - Case "st_endch".
admit. (* cannot communicate channel permissions *)
    - Case "st_assign".
admit. (* no change in channel permissions *)
    - Case "st_load".
admit. (* no change in channel permissions *)
    - Case "st_store".
admit. (* no change in channel permissions *)
    - Case "st_local".
admit. (* no change in channel permissions *)
    - Case "st_freelocal".
admit. (* no change in channel permissions *)
    - Case "st_par".
    destruct HwO' as [w1_[w2_[?[??]]]]; subst w1_ w2_.
    replace wO' with wO in *; auto.
    eapply gjoin_eq; eauto.
    - Case "st_par1".
    destruct HwO as [w1[w2[?[??]]]].
    destruct HwO' as [w1'[w2_[?[??]]]].
    replace w2_ with w2 in *.

    apply IHHs with w1; eauto.
    eapply no_send_conflict_sub; eauto.
    exists w2; auto. apply join_sub_refl.

    eapply cannot_communicate_channels_par1; eauto.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    - Case "st_par2".
    renames w1 to w12, w1' to w12', w2 to w3.
    destruct Hw1 as [w1[w2[?[??]]]].
    destruct Hw1' as [w1_[w2'[?[??]]]].
    replace w1_ with w1 in *.
    apply gjoin_com in H1.
    apply gjoin_com in H4.
    edestruct gjoin_assoc with (a:=w2) (b:=w1) as [?[??]]; eauto.
    edestruct gjoin_assoc with (a:=w2') (b:=w1) as [?[??]]; eauto.
    replace x0 with x in *.
    apply IHHs with w2 w2' x; eauto.
    eapply cannot_communicate_channels_par2; eauto.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    - Case "st_par_join".
    destruct Hw1 as [w1_[w2_[?[??]]]]; subst w1_ w2_.
    replace w3 with w1' in *.
    replace w' with w in *; auto.
    eapply gjoin_eq; eauto.
    eapply gjoin_eq; eauto.
  Qed.
*)

(*
  Lemma no_send_conflict_cstep_star: forall s wC C b1 w2 b2 s' wC' C' b1' w2' b2',
    no_send_conflict wC w2 ->
    context_world C wC ->
    context_world C' wC' ->
    join_wterms b2 w2 ->
    join_wterms b2' w2' ->
    cstep_star s C (b1 || b2) s' C' (b1' || b2') ->
    no_send_conflict wC' w2'.
(* no_send_conflict also needs to state that a conflict cannot be transferred through the channels *)
Admitted.


  Lemma no_send_conflict_no_send_to_context: forall wC C w b,
    no_send_conflict wC w ->
    context_world C wC ->
    join_wterms b w ->
    no_send_to_context C b.
  Proof.
    red; intros.
    apply context_world_join_wterms in H2.
    replace wC0 with wC in *. replace wb with w in *.
    auto. eapply join_wterms_eq; eauto. eapply context_world_eq; eauto.
  Qed.

  Lemma no_send_conflict_stepO: forall s wC bO w2 s' wC' bO',
    no_send_conflict wC w2 ->
    join_wterms bO wC ->
    join_wterms bO' wC' ->
    step s bO s' bO' ->
    no_send_conflict wC' w2.
  Proof.
    intros.
    eapply no_send_conflict_cstep_star
      with (C:=cx_par1 cx_hole bO) (C':=cx_par1 cx_hole bO') (b1:=(gwemp:tm_skip)%wterm) (b2:=(w2:tm_skip)%wterm) (b2':=(w2:tm_skip)%wterm); eauto.
    exists gwemp wC; splits; auto. reflexivity. apply gjoin_bot2.
    exists gwemp wC'; splits; auto. reflexivity. apply gjoin_bot2.
    reflexivity. reflexivity.
    apply cstep_single_cstep_star.
    apply cst_par1_step2; eauto.
  Qed.
*)

  Lemma cannot_communicate_channels_par_swap: forall s b1 b2,
    cannot_communicate_channels s (b1||b2) ->
    cannot_communicate_channels s (b2||b1).
  Proof.
    introv [w[?[?[??]]]].
    exists w; splits; auto.
    apply join_wterm_worlds in H; apply join_wterm_worlds.
    eapply join_list_perm; simpl in *; eauto. permut4.
    split; auto.
  Qed.

  Lemma private_communications_cstep1b: forall s C w w2 b1 b2 s' C' w' b1',
    private_communications w2 w ->
    cannot_communicate_channels s b1 ->
    valid_machine_state gwemp s (b1 || b2) ->
    join_wterms b2 w2 ->
    join_wterms (b1 || b2)%wterm w ->
    join_wterms (b1' || b2)%wterm w' ->
    cstep s C b1 s' C' b1' ->
    private_communications w2 w'.
  Proof.
    introv Hprivate2 Hcc1 Hvms Hw2 Hw Hw' Hc.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    destruct Hw as [w1[w2_[?[??]]]].
    replace w2_ with w2 in *. clear w2_ H1.
    destruct Hw' as [w1'[w2_[?[??]]]].
    replace w2_ with w2 in *. clear w2_ H3.
    eapply private_communications_step2 with (b2:=b1) (w2:=w1); eauto.
    apply gjoin_com; auto.
    apply gjoin_com; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    subst b1'.
    replace w' with w in *; auto.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma noninterfering_context_cstep_star_step: forall s wC wC1 w12 C w1 b1 w2 b2 s' C' b1' s'' b2'',
    context_world C wC ->
    join_wterms b1 w1 ->
    join_wterms b2 w2 ->
    gjoin w1 w2 w12 ->
    noninterfering_wterm b1 b2 ->
    gjoin wC w1 wC1 ->
    terminates s wC1 b2 ->
    cstep_star s C b1 s' C' b1' ->
    step s' b2 s'' b2'' ->
    private_communications w2 w12 ->
    cannot_communicate_channels s (b1||b2) ->
    valid_machine_state wC s (b1||b2) ->
    exists s'0 s'1,
      s_iter_star s s'0 /\
      step s'0 b2 s'1 b2'' /\ cstep_star s'1 C b1 s'' C' b1'.
  Proof.
    introv HwC Hw1 Hw2 Hw12 Hni HwC1 Hterm2 Hcs Hs.
    introv Hprivate2 Hcc Hvms.
    revert w1 w12 wC wC1 HwC Hw1 Hw12 HwC1 Hterm2 Hprivate2 Hcc Hvms.
    induction Hcs; intros.
    - exists s s''; splits; auto. apply s_iter_nil. apply cstep_nil.
    - renames b to b1, b' to b1', b'' to b1''.
    edestruct valid_machine_state_cstep_star with (b:=(b1||b2)%wterm) as [wC'[??]]; eauto.
    apply cstep_star_par1_step.
    apply cstep_single_cstep_star; eauto.
    assert (exists w1' w12' wC1', join_wterms b1' w1' /\ gjoin w1' w2 w12' /\ gjoin wC' w1' wC1').
      destruct H1 as [wS'[wb'[ww'[?[? ?]]]]].
      destruct H2 as [w1'[w2_[?[??]]]]; eauto.
      replace w2_ with w2 in *.
      apply gjoin_join_list in H5.
      jrewrite<- H5 in H3.
      edestruct join_list_app_inv with (wl1:=[wC',w1']) (wl2:=[wS',w2]) (w:=ww') as [wC1'[wS'2[?[??]]]]; eauto.
      eapply join_list_perm; eauto.
      permut3.
      exists w1' wb' wC1'; splits; auto.
      apply gjoin_join_list; auto.
      apply gjoin_join_list; auto.
      eapply join_wterms_eq; eauto.
    destruct H2 as [w1'[w12'[wC1'[?[??]]]]].
    destruct IHHcs with w1' w12' wC' wC1' as [s'0[s'1[?[??]]]]; clear IHHcs; eauto.
    eapply noninterfering_cstep1; eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply terminates_cstep' with (C:=cx_par1 C b1) (C':=cx_par1 C' b1'); eauto.
    apply cstep_par_cx_par1; apply cstep_par2_step; auto.
    exists wC w1; auto.
    exists wC' w1'; auto.
    eapply valid_machine_state_par_observer with (s:=s) (bO:=b1); eauto.
    apply valid_machine_state_par_swap; auto.
    eapply private_communications_cstep1b with (C:=C) (b1:=b1) (b2:=b2); eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply valid_machine_state_subO; eauto.
    exists wC; apply gjoin_bot2.
    exists w1 w2; auto.
    exists w1' w2; auto.
    eapply cannot_communicate_channels_cstep with (C:=C); eauto.
    apply cstep_par1_step; eauto.
    (**)
    edestruct cstep_commute_front_s_iter_star as [s'2[??]]; eauto.
    clear H Hcs.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    + subst C'.
    edestruct noninterfering_schedule_step' with (b1:=b2) (b2:=b1) as [s'3[??]]; eauto.
    apply noninterfering_wterm_symm; auto.
    eapply terminates_s_iter_star; eauto.
    eapply terminates_sub; eauto.
    exists wC; auto.
    eapply valid_machine_state_par_observer with (bO:=b1); eauto.
    eapply valid_machine_state_par_swap; auto.
    apply cannot_communicate_channels_par_swap; auto.
    eapply cannot_communicate_channels_s_iter_star; eauto.
    apply valid_machine_state_par_swap.
    eapply valid_machine_state_s_iter_star; eauto.
    eapply valid_machine_state_subO; eauto.
    exists wC; apply gjoin_bot2.
    exists s'2 s'3; splits; auto.
    eapply cstep_star_app; eauto.
    apply step_star_add_context_cstep_star; auto.
    + subst b1'.
    replace w1' with w1 in *.
    edestruct noninterfering_rev_context_step with (s:=s'2) (b2:=b2) (C:=C) as [s'3[??]]; eauto.
    eapply private_communications_no_send_conflict; eauto.
    destruct Hvms as [wS[w12_[ww[?[??]]]]].
    replace w12_ with w12 in *.
    edestruct join_list_app_inv with (wl1:=[wC,w12]) (wl2:=[wS]) (w:=ww) as [wC12[?[? _]]].
    eapply join_list_perm; eauto. permut3.
    exists wC12; apply gjoin_join_list; auto.
    destruct H11 as [w1_[w2_[?[??]]]].
    replace w1_ with w1 in *.
    replace w2_ with w2 in *.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    eapply terminates_s_iter_star; eauto.
    eapply terminates_sub; eauto.
    exists w1; auto.
    eapply valid_machine_state_par_observer with (bO:=b1); eauto.
    eapply valid_machine_state_par_swap; auto.
    apply cannot_communicate_channels_par2 with b1.
    eapply cannot_communicate_channels_s_iter_star; eauto.
    eapply valid_machine_state_subO with (wC:=wC1).
    eapply valid_machine_state_par_observer with (bO:=b1) (wO1:=wC) (wO:=wC1); eauto.
    apply valid_machine_state_par_swap.
    eapply valid_machine_state_s_iter_star; eauto.
    exists w1; auto.
    exists s'2 s'3; splits; auto.
    eapply cstep_star_app; eauto.
    apply cstep_star_skip_subst with gwemp; eauto.
    eapply join_wterms_eq; eauto.

    - edestruct IHHcs as [s'0[s'1[?[??]]]]; clear IHHcs; eauto.
    eapply terminates_s_iter; eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
    exists s'0 s'1; splits; auto.
    eapply s_iter_cons; eauto.
  Qed.


(*
  Lemma noninterfering_context_cstep_star_step_star: forall s wC C w1 b1 w2 b2 s' C' b1' s'' b2'',
    context_world C wC ->
    join_wterms b1 w1 ->
    join_wterms b2 w2 ->
    noninterfering_wterm b1 b2 ->
    no_send_to_context C b2 ->
    terminates s wC b2 ->
    cstep_star s C b1 s' C' b1' ->
    step_star s' b2 s'' b2'' ->
    valid_machine_state wC s (b1 || b2) ->
    exists s'0,
      step_star s b2 s'0 b2'' /\
      cstep_star s'0 C b1 s'' C' b1'.
  Proof. 
Admitted.
    introv HjC Hjb1 Hjb2 Hni Hno_send Hterm2 Hcs Hss Hvms.
    revert s w2 wC HjC Hjb2 Hterm2 Hcs Hvms.
    induction Hss; intros.
    - exists s0; split; auto. apply step_nil.
    - renames b to b2, b' to b2', b'' to b2''.
    edestruct noninterfering_context_cstep_star_step with (b1:=b1) (b2:=b2) as [s'0[s'1[?[??]]]]; eauto.
    assert (valid_machine_state wC s'1 (b1||b2')).
      eapply valid_machine_state_step_star; eauto.
      eapply step_star_snoc.
      apply s_iter_step_star; eauto.
      apply st_par2; auto.
    assert (exists w2', join_wterms b2' w2').
      destruct H3 as [_[wb'[_[_[? _]]]]].
      destruct H3 as [_[w2'[_[? _]]]]; eauto.
    destruct H4 as [w2' ?].
    destruct IHHss with s'1 w2' wC as [s'2[??]]; eauto.
    eapply noninterfering_preservation2; eauto.
    apply step_single_cstep_star with (C:=cx_hole); eauto.
    eapply preservation_no_send_to_context2'; eauto.
    apply step_single; eauto.
    eapply terminates_preservation with (w2:=w1) (b2:=b1) (b2':=b1); eauto.
    eapply cstep_star_app.
    apply cstep_star_s_iter_set; eauto.
    apply step_single_cstep_star; apply st_par1; eauto.
    exists s'2; split; auto.
    eapply step_star_app.
    apply s_iter_step_star; eauto.
    eapply step_cons; eauto.

    - destruct IHHss with s0 w2 wC as [s'2[??]]; eauto.
    eapply cstep_star_iter_snoc; eauto.
  Qed.
*)

(*  Lemma noninterfering_schedule_cstep_star: forall s C b1 b2 s' C' b2' s'' b1',
    no_send_to_context C b2 ->
    noninterfering_wterm b1 b2 ->
    b2 <> b2' ->
    step s b2 s' b2' ->
    cstep_star s' C b1 s'' C' b1' ->
    exists s'0,
         cstep_star s C b1 s'0 C' b1'
    /\ step s'0 b2 s'' b2'.
  Proof.
Admitted.

    introv Hno_send H Hnospin2 H0 H1.
    revert s H0 H.
    induction H1; intros.
    exists s0; split; [apply cstep_nil|auto].
    rename t into b1; rename t' into b1'; rename t'' into b1''.
    rename s into s'0; rename s0 into s.
    edestruct cstep_inversion with (s:=s'0) as [[??]|[??]]; eauto; clear H; subst.
    edestruct noninterfering_schedule_step as [s'1[??]]; eauto.
    destruct IHcstep_star with s'1 as [s'2[??]]; auto.
admit.
    exists s'2; split; auto.
    eapply cstep_cons; eauto.
    apply cstep_simulates_step; apply step_add_context; auto.

    edestruct context_schedule_step_cstep as [s'1[??]]; eauto.
    destruct IHcstep_star with s'1 as [s'2[??]]; auto.
    eapply preservation_no_send_to_context1; eauto.
      eapply cstep_cons; eauto. apply cstep_nil.
    exists s'2; split; auto.
    eapply cstep_cons; eauto.
    apply cstep_skip_subst with gwemp; auto.
  Qed.
*)

  Lemma noninterfering_schedule_cstep_star': forall wC wC2 w12 s C b1 w2 b2 s' C' b2' s'' b1',
    context_world C wC ->
    private_communications w2 w12 ->
    noninterfering_wterm b1 b2 ->
    join_wterms b2 w2 ->
    join_wterms (b1||b2)%wterm w12 ->
    gjoin wC w2 wC2 ->
    terminates s wC2 b1 -> (*step b2 does not enable termination of b1*)
    step s b2 s' b2' ->
    cstep_star s' C b1 s'' C' b1' ->
    cannot_communicate_channels s (b1||b2) ->
    valid_machine_state wC s (b1 || b2) ->
    exists s'0,
      cstep_star s C b1 s'0 C' b1' /\ step_star s'0 b2 s'' b2'.
  Proof.
    introv HwC Hprivate Hni Hw2 Hw12 HwC2 Hterm1 Hs Hcs Hvms.
    destruct Hw12 as [w1[w2_[?[??]]]].
    replace w2_ with w2 in *. clear w2_ H0.
    eapply noninterfering_schedule_step_star_cstep_star with (w1:=w1); eauto.
    apply step_single; auto.
    eapply join_wterms_eq; eauto.
  Qed.


  Lemma noninterfereing_seq_schedule: forall wC wC2 w12 s C b1 w2 b2 s'' C'' b1'' b2'', 
    context_world C wC ->
    private_communications w2 w12 ->
    noninterfering_wterm b1 b2 ->
    join_wterms b2 w2 ->
    join_wterms (b1||b2)%wterm w12 ->
    gjoin wC w2 wC2 ->
    terminates s wC2 b1 -> (*step b2 does not enable termination of b1*)
    cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'') ->
    cannot_communicate_channels s (b1||b2) ->
    valid_machine_state wC s (b1||b2) ->
    exists s',
      cstep_star s C b1 s' C'' b1'' /\ step_star s' b2 s'' b2''.
  Proof.
    introv HwC Hprivate2 Hni Hw2 Hw12 HwC2 Hterm1 Hcs Hcc Hvms.
    remember (b1||b2)%wterm as b.
    remember (b1'' || b2'')%wterm as b''.
    revert wC b1 b2 w2 w12 wC2 HwC Heqb Heqb'' Hw2 Hw12 HwC2 Hterm1 Hprivate2 Hni Hcc Hvms.
    induction Hcs; intros; subst.

    - inversion Heqb''; subst; clear Heqb''.
    exists s; split; [apply cstep_nil | apply step_nil].
    - edestruct valid_machine_state_cstep as [wC'[??]]; eauto.
    edestruct cstep_inversion as [[??]|[??]]; eauto.
    + subst C'.
    replace wC' with wC in *.
    inversion H2; subst; clear H2.
    * destruct Hw12 as [w1[w2_[?[??]]]].
    replace w2_ with w2 in *. clear w2_.
    assert (exists w1' w12', join_wterms b1' w1' /\ gjoin w1' w2 w12').
      edestruct valid_machine_state_join_wterms with (s:=s') as [w12' ?]; eauto.
      destruct H5 as [w1'[w2_[?[??]]]].
      replace w2_ with w2 in *. clear w2_.
      exists w1' w12'; split; auto.
      eapply join_wterms_eq; eauto.
    destruct H5 as [w1'[w12'[??]]].
    edestruct IHHcs with (w12:=w12') as [s'0[??]]; clear IHHcs; eauto.
    exists w1' w2; splits; auto.
    eapply terminates_step; eauto.
    eapply private_communications_step2 with (b2:=b1) (b2':=b1'); eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply gjoin_com; eauto.
    eapply gjoin_com; eauto.
    eapply noninterfering_step1; eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply cannot_communicate_channels_step; eauto. apply st_par1; auto.
    exists s'0; split; auto.
    eapply step_cstep_star; eauto.
    eapply join_wterms_eq; eauto.
    * clear Hcs H.
    assert (exists w2', join_wterms b2' w2').
      destruct H1 as [_[wb'[_[_[? _]]]]].
      destruct H as [_[w2'[_[? _]]]]; eauto.
    destruct H as [w2' ?].
    generalize H1; intro Hvms'.
    destruct H1 as [wS'[w12'[ww'[?[??]]]]].
    apply join_wterm_worlds in H2; simpl in H2.
    edestruct join_list_app_inv as [w1[w2'_[?[??]]]]; eauto.
    replace w2'_ with w2' in *.
    jrewrite<- H2 in H3.
    jrewrite H4 in H3.
    jrewrite H5 in H3.
    rewrite app_nil_r in H3.
    edestruct join_list_app_inv with (wl1:=[wC,w2']) (wl2:=[wS',w1]) (w:=ww') as [wC2'[wS'1[?[??]]]]; eauto.
    eapply join_list_perm; eauto.
    permut4.
    apply gjoin_join_list in H8.
    destruct Hw12 as [w1_[w2_[?[??]]]].
    replace w1_ with w1 in *. clear w1_.
    replace w2_ with w2 in *. clear w2_.
    edestruct IHHcs as [s'1[??]]; clear IHHcs; eauto.
    exists w1 w2'; splits; eauto.
    eapply terminates_stepO' with (bO:=(wC:tm_skip||b2)%wterm); eauto.
    apply st_par2; eauto.
    exists wC w2; splits; auto. reflexivity.
    exists wC w2'; splits; auto. reflexivity.
    eapply valid_machine_state_par_observer with (bO:=b2); eauto.
    eapply private_communications_step with (b1:=b2) (b1':=b2') (w2:=gwemp) (w12:=w2) (w12':=w2') (w3:=w1); eauto.
    eapply cannot_communicate_channels_par2; eauto.
    apply gjoin_bot1.
    apply gjoin_com; auto.
    apply gjoin_bot1.
    apply gjoin_com; auto.
    eapply noninterfering_step2; eauto.
    eapply cannot_communicate_channels_par2; eauto.
    eapply cannot_communicate_channels_step; eauto. apply st_par2; auto.
    (**)
    edestruct noninterfering_schedule_cstep_star' as [s'0[??]]; eauto.
    exists w1 w2; splits; auto.
    exists s'0; split; auto.
    eapply step_star_app; eauto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    apply join_wterm_worlds; eauto.
    eapply join_wterms_eq; eauto.
    apply join_wterm_worlds; eauto.
    * apply cstep_star_skip_resolve in Hcs.
    discriminate.
    * eapply context_world_eq; eauto.

    + subst b'.
    assert (exists wC2', gjoin wC' w2 wC2').
      destruct H1 as [wS'[wb'[ww'[?[??]]]]].
      apply join_wterm_worlds in H3; simpl in H3.
      jrewrite<- H3 in H4.
      jrewrite Hw2 in H4.
      rewrite app_nil_l in H4.
      edestruct join_list_app_inv with (wl1:=[wC',w2]) (wl2:=[wS']++wterm_worlds b1) (w:=ww') as [wC2'[wS'1[?[??]]]]; eauto.
      eapply join_list_perm; eauto.
      permut3.
      apply gjoin_join_list in H5.
      exists wC2'; auto.
    destruct H3 as [wC2' ?].
    edestruct IHHcs as [s'1[??]]; clear IHHcs; eauto.
    eapply terminates_cstep' with (C:=cx_par1 C b2) (C':=cx_par1 C' b2); eauto.
    apply cst_par1.
    apply cstep_skip_subst with gwemp; eauto.
    exists wC w2; split; auto.
    exists wC' w2; split; auto.
    eapply valid_machine_state_par_observer with (bO:=b2); eauto.
    eapply cannot_communicate_channels_stepO with (bO:=C[gwemp:tm_skip]); eauto.
    eapply step_simulates_cstep; eauto.
    apply context_world_join_wterms; auto.
    exists s'1; split; auto.
    eapply cstep_cons; eauto.
    apply cstep_skip_subst with gwemp; auto.

    - edestruct IHHcs as [s'1[??]]; clear IHHcs; eauto.
    eapply terminates_s_iter; eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
    exists s'1; split; auto.
    eapply cstep_iteration; eauto.
  Qed.


(*
  Definition term_impl' s wC b1 b2 := forall C s' C' w1' b2',
    context_world C wC ->
    cstep_star s C (b1 || b2) s' C' (w1':tm_skip || b2') ->
    terminates s' C' b2' w1'.
  Lemma term_impl'_preservation: forall s wC C b1 b2 s' wC' C' b1' b2',
    term_impl' s wC b1 b2 ->
    context_world C wC ->
    cstep_star s C (b1 || b2) s' C' (b1' || b2') ->
    context_world C' wC' ->
    valid_machine_state wC s (b1||b2) ->
    term_impl' s' wC' b1' b2'.
  Proof.
    unfold term_impl'; intros.
    edestruct compose_context_world_cstep_star
      with (C1:=C) (C2:=C0) as [wO''[C12[C12'[?[?[??]]]]]]; eauto.
(* need to update 'terminates' to use wC (context worlds) instead of contexts C *)
Admitted.
*)


  Definition term_ensures_pred s b P := forall C s' C' w',
    cstep_star s C b s' C' (w':tm_skip) -> w' |= P.

(*
  Definition noninterfering_wterm_pred b1 P2:= forall w1 w2,
    join_wterms b1 w1 -> w2|=P2 -> noninterfering w1 w2.
  Definition noninterfering_pred P1 P2:= forall w1 w2,
    w1|=P1 -> w2|=P2 -> noninterfering w1 w2.

  Lemma noninterfering_wterm_pred_preservation b1 P2: forall w1 s wC C s' C' b1',
    noninterfering_wterm_pred b1 P2 ->
    context_world C wC ->
    join_wterms b1 w1 ->
    cannot_communicate_channels s b1 ->
    valid_machine_state wC s b1 ->
    cstep_star s C b1 s' C' b1' ->
    noninterfering_wterm_pred b1' P2.
  Proof.
    unfold noninterfering_wterm_pred; introv Hni HwC Hw1 Hcc Hcs Hvms; intros.
    rename w0 into w1'.
    assert (noninterfering_wterm b1 (w2:tm_skip)%wterm).
      red; intros.
      replace w0 with w1 in *.
      simpl in H2; subst w3.
      auto.
      eapply join_wterms_eq; eauto.
    eapply noninterfering_cstep_star1; eauto.
    apply cstep_star_par1_step; eauto.
    

    unfold noninterfering in *.
    eapply H; auto.
    cut (noninterfering_wterm b1 (w2:tm_skip)); intro.
    assert (noninterfering_wterm b1' (w2:tm_skip)).
      eapply noninterfering_cstep_star; eauto.
      apply cstep_star_par1_step; eauto.
      eapply cannot_communicate_channels_cstep_star; eauto.
      apply cstep_star_par1_step; eauto.
    apply H4; simpl; auto.
    intros.
    simpl in H4; subst w3.
    apply H; auto.
  Qed.

  Lemma noninterfering_emp: forall t1 b2,
    noninterfering_wterm (gwemp:t1) b2.
  Admitted.
*)


  Lemma noninterfereing_rev_seq_schedule: forall s C wC wC1 w12 w1 b1 w2 b2 s'' C'' b1'' b2'', 
    join_wterms b1 w1 ->
    join_wterms b2 w2 ->
    gjoin w1 w2 w12 ->
    context_world C wC ->
    private_communications w2 w12 ->
    noninterfering_wterm b1 b2 ->
    gjoin wC w1 wC1 ->
    terminates s wC1 b2 -> (*step b2 does not enable termination of b1*)
    cstep_star s C (b1 || b2) s'' C'' (b1'' || b2'') ->
    cannot_communicate_channels s (b1||b2) ->
    valid_machine_state wC s (b1 || b2) ->
    exists s',
      step_star s b2 s' b2'' /\ cstep_star s' C b1 s'' C'' b1''.
  Proof.
    introv Hw1 Hw2 Hw12 HwC Hprivate2.
    introv Hni HwC1 Hterm2 Hcs Hcc Hvms.
    remember (b1||b2)%wterm as b.
    remember (b1'' || b2'')%wterm as b''.
    revert b1 b2 b1'' b2'' w1 w2 w12 Heqb Heqb'' Hw1 Hw2 Hw12 HwC1 Hterm2 Hprivate2 Hni Hvms.
    apply cstep_star_snoc_equiv in Hcs.
    induction Hcs; intros; subst.
    - inversion Heqb''; subst; clear Heqb''.
    exists s; split; [apply step_nil | apply cstep_nil].
    - apply cstep_star_snoc_equiv in Hcs.
    edestruct valid_machine_state_cstep_star as [wC'[??]]; eauto.
    edestruct cstep_star_par_resolve as [[b1'[b2' ?]]|[w1'[w2'[w'[?[??]]]]]]; eauto.
    + subst b'.
    edestruct IHHcs as [s'0[??]]; clear IHHcs; eauto.
    edestruct cstep_inversion as [[??]|[??]]; eauto.
    * subst C''.
    inversion H4; subst; clear H4.
    exists s'0; split; auto.
    eapply cstep_star_snoc; eauto.
    apply cstep_simulates_step; apply step_add_context; auto.
    assert (exists w2' w12', join_wterms b2' w2' /\ gjoin w1 w2' w12').
      edestruct valid_machine_state_step_star with (b:=(b1||b2)%wterm) as [wS'[w12'[ww'[?[??]]]]]; eauto.
      apply step_star_par2_step; eauto.
      destruct H5 as [w1_[w2'[?[??]]]].
      replace w1_ with w1 in *. clear w1_.
      exists w2' w12'; split; auto.
      eapply join_wterms_eq; eauto.
    destruct H4 as [w2'[w12'[??]]].
    edestruct noninterfering_context_cstep_star_step
      with (w1:=w1) (s:=s'0) (b1:=b1) (b2:=b2') (C:=C) as [s'1[s'2[?[??]]]]; eauto.
    eapply noninterfering_step_star2; eauto.
    eapply cannot_communicate_channels_par2; eauto.
    eapply terminates_step_star; eauto.
    eapply private_communications_step_star with (b1:=b2) (w2:=w1) (w1:=w2); eauto.
    eapply cannot_communicate_channels_par2; eauto.
    eapply valid_machine_state_par_observer with (wO1:=gwemp) (wO2:=w1) (bO:=b1); eauto.
    apply gjoin_bot2.
    eapply valid_machine_state_subO; eauto.
    apply valid_machine_state_par_swap; eauto.
    exists wC; apply gjoin_bot2.
    apply gjoin_com; auto.
    apply gjoin_com; auto.
    eapply cannot_communicate_channels_step_star; eauto.
    apply step_star_par2_step; auto.
    eapply valid_machine_state_step_star; eauto.
    apply step_star_par2_step; auto.
    exists s'2; split; auto.
    eapply step_star_app; eauto.
    eapply step_star_snoc; eauto.
    apply s_iter_step_star; auto.

    * inverts H5.
    exists s'0; split; auto.
    eapply cstep_star_snoc; eauto.
    apply cstep_skip_subst with gwemp; auto.

    + subst b'.
    apply cstep_skip_resolve in H. discriminate.

    - edestruct IHHcs as [s'0[??]]; eauto.
    exists s'0; split; auto.
    eapply cstep_star_iter_snoc; eauto.
  Qed.


(*
  Lemma noninterfering_step_star_seq_schedule: forall wC s C b1 w2 b2 s'' b1'' b2'', 
    noninterfering_wterm b1 b2 ->
    join_wterms b2 w2 ->
    context_world C wC ->
    no_send_conflict wC w2 ->
    terminates s wC b1 -> (*step b2 does not enable termination of b1*)
    step_star s (b1 || b2) s'' (b1'' || b2'') ->
    valid_machine_state wC s (b1||b2) ->
    exists s',
         step_star s b1 s' b1''
    /\ step_star s' b2 s'' b2''.
  Proof.
Admitted.
    introv Hni Hjb2 HwC Hnsc Hterm1 Hss Hvms.
    remember (b1||b2)%wterm as b.
    remember (b1'' || b2'')%wterm as b''.
    revert b1 b2 w2 wC HwC Hnsc Heqb Heqb'' Hjb2 Hterm1 Hni Hvms.
    induction Hss; intros; subst.
    - inverts Heqb''; exists s; split; apply step_nil.
    - assert (Hvms': valid_machine_state wC s' b').
      eapply valid_machine_state_step; eauto.
    inverts H.
    + edestruct IHHss as [s'0[??]]; eauto.
    eapply terminates_preservation; eauto.
    apply cstep_single_cstep_star; auto.
    apply cstep_simulates_step.
    apply step_add_context. apply st_par1; auto.
    eapply noninterfering_preservation1 with (C:=C); eauto.
    apply step_single_cstep_star; eauto.
    exists s'0; split; auto.
    eapply step_cons; eauto.
    + clear Hss.
    assert (exists w2', join_wterms b2' w2').
      destruct Hvms' as [_[wb'[_[_[? _]]]]].
      destruct H as [_[w2'[_[? _]]]]; eauto.
    destruct H as [w2' ?].
    edestruct IHHss with (w2:=w2') as [s'0[??]]; clear IHHss; eauto.
admit.
    eapply terminates_preservation with (w2:=w2); eauto.
    apply step_single_cstep_star; apply st_par2; auto.
    eapply noninterfering_preservation2 with (C:=C); eauto.
    apply step_single_cstep_star; eauto.
    edestruct noninterfering_schedule_step_star' as [s'1[??]]; eauto.
    eapply terminates_dummyContext; eauto.
    exists s'1; split; auto.
    eapply step_star_app; eauto.

    + false; clear -Hss.
    edestruct step_star_skip_resolve; eauto; discriminate.

    - edestruct IHHss as [s'0[??]]; clear IHHss; eauto.
    eapply terminates_preservation; eauto.
    eapply cstep_iteration; eauto; apply cstep_nil.
    eapply valid_machine_state_s_iter; eauto.
    exists s'0; split; auto.
    eapply st_iteration; eauto.
Qed.
*)



  Lemma term_ensures_pred_s_iter: forall s s' b P,
    s_iter s s' ->
    term_ensures_pred s b P ->
    term_ensures_pred s' b P.
  Proof.
    unfold term_ensures_pred; intros.
    eapply H0. eapply cstep_iteration; eauto.
  Qed.

  Lemma term_ensures_pred_stepO: forall s s' b P bO bO',
    term_ensures_pred s b P ->
    step s bO s' bO' ->
    term_ensures_pred s' b P.
  Proof.
    unfold term_ensures_pred; intros.
    eapply H.
    eapply cstep_cons with (C':=cx_par1 C bO'); eauto.
    apply cst_par1_step2; eauto.
    eapply cstep_star_par_cx_par1.
    apply cstep_star_par1_step; eauto.
  Qed.

(*
  Lemma term_impl'_preservation_stepO: forall s s' wO wO' bO bO' b1 b2,
    step s bO s' bO' ->
    join_wterms bO wO ->
    join_wterms bO' wO' ->
    term_impl' s wO b1 b2 ->
    valid_machine_state wO s (b1||b2) ->
    term_impl' s' wO' b1 b2.
  Proof.
    intros.
    eapply term_impl'_preservation with (b2:=b2) (C:=cx_par1 cx_hole bO); eauto.
    exists gwemp wO; splits; simpl; auto. apply gjoin_bot2.
    eapply cstep_single_cstep_star.
    apply cst_par1_step2; eauto.
    exists gwemp wO'; splits; simpl; auto. apply gjoin_bot2.
  Qed.
*)


  Lemma term_ensures_pred_par1_cstep_star: forall P s C b1 b2 s' C' b1' b2',
    term_ensures_pred s b1 P ->
    cstep_star s C (b1 || b2) s' C' (b1' || b2') ->
    term_ensures_pred s' b1' P.
  Proof.
    intros; red; intros.
    eapply H with (C:=compose_contexts (compose_contexts C C0) (cx_par1 cx_hole b2)).
    apply cstep_star_par_compose_contexts1.
    eapply cstep_star_app.
    apply cstep_star_composed_context_step1'; eauto.
    apply cstep_star_composed_context_step2'.
    apply cstep_star_par1_step; eauto.
  Qed.

  Lemma term_ensures_pred_par2_cstep_star: forall P s C b1 b2 s' C' b1' b2',
    term_ensures_pred s b2 P ->
    cstep_star s C (b1 || b2) s' C' (b1' || b2') ->
    term_ensures_pred s' b2' P.
  Proof.
    intros; eapply term_ensures_pred_par1_cstep_star; eauto.
    apply cstep_star_par_swap; eauto.
  Qed.


  Lemma step_star_sans_s_iter_par1_step: forall s b1 b2 s' b1',
    step_star_sans_s_iter s b1 s' b1' ->
    step_star_sans_s_iter s (b1 || b2) s' (b1' || b2).
  Proof. intros; induction H; econstructor; eauto; apply st_par1; auto. Qed.

  Lemma step_star_sans_s_iter_par2_step: forall s b1 b2 s' b2',
    step_star_sans_s_iter s b2 s' b2' ->
    step_star_sans_s_iter s (b1 || b2) s' (b1 || b2').
  Proof. intros; induction H; econstructor; eauto; apply st_par2; auto. Qed.

  Lemma step_star_sans_s_iter_seq_app: forall s b t s' b',
    step_star_sans_s_iter s b s' b' ->
    step_star_sans_s_iter s (b;; t) s' (b';; t).
  Proof. intros; induction H; econstructor; eauto; apply st_seq1; auto. Qed.

  Lemma step_star_sans_s_iter_seq_app': forall s b t s' w',
    step_star_sans_s_iter s b s' (w': tm_skip) ->
    step_star_sans_s_iter s (b;; t) s' (w': t).
  Proof.
    intros;
    eapply step_star_sans_s_iter_app.
    apply step_star_sans_s_iter_seq_app; eauto.
    eapply step_cons_; [apply st_seq_skip | apply step_nil_].
  Qed.

  Lemma valid_machine_state_gjoin_par1: forall wO s b1 b2,
    valid_machine_state wO s (b1 || b2) ->
    exists w1 wO1,
      join_wterms b1 w1 /\ gjoin wO w1 wO1.
  Proof.
    intros.
    destruct H as [wS[w[ww[?[??]]]]].
    destruct H0 as [w1[w2[?[??]]]].
    apply gjoin_join_list in H3.
    jrewrite<- H3 in H1.
    eapply join_list_perm with (wl2:=[wO,w1]++[wS,w2]) in H1.
    edestruct join_list_app_inv as [wO1[wS2[?[??]]]]; eauto.
    exists w1 wO1; split; auto.
    apply gjoin_join_list; auto.
    permut3.
  Qed.

  Lemma valid_machine_state_gjoin_par2: forall wO s b1 b2,
    valid_machine_state wO s (b1 || b2) ->
    exists w2 wO2,
      join_wterms b2 w2 /\ gjoin wO w2 wO2.
  Proof.
    intros.
    eapply valid_machine_state_gjoin_par1; eauto.
    apply valid_machine_state_par_swap; eauto.
  Qed.

  Lemma private_communications_cstep_b: forall s C w w2 b1 b2 s' C' w' w2' b1' b2',
    private_communications w2 w ->
    cannot_communicate_channels s (b1 || b2) ->
    valid_machine_state gwemp s (b1 || b2) ->
    join_wterms b2 w2 ->
    join_wterms b2' w2' ->
    join_wterms (b1 || b2)%wterm w ->
    join_wterms (b1' || b2')%wterm w' ->
    cstep s C (b1||b2) s' C' (b1'||b2') ->
    private_communications w2' w'.
  Proof.
    introv Hprivate Hcc Hvms Hw2 Hw2' Hw Hw' Hc.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'. clear Hc.
    destruct Hw as [w1[w2_[?[??]]]].
    replace w2_ with w2 in *. clear w2_ H1.
    destruct Hw' as [w1'[w2'_[?[??]]]].
    replace w2'_ with w2' in *. clear w2'_ H3.
    inverts H.
    replace w2' with w2 in *. clear w2' Hw2'.
    eapply private_communications_step2 with (b2:=b1) (b2':=b1'); eauto.
    eapply cannot_communicate_channels_par1; eauto.
    apply gjoin_com; auto.
    apply gjoin_com; auto.
    eapply join_wterms_eq; eauto.
    replace w1' with w1 in *. clear w1' H0.
    eapply private_communications_step with (w2:=gwemp) (w3:=w1) (b1:=b2) (b1':=b2'); eauto.
    eapply cannot_communicate_channels_par2; eauto.
    apply gjoin_bot1.
    apply gjoin_com; auto.
    apply gjoin_bot1.
    apply gjoin_com; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    (**)
    inverts H0.
    replace w' with w in *.
    replace w2' with w2 in *; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma private_communications_cstep_star_b: forall s C wC w w2 b1 b2 s' C' w' w2' b1' b2',
    private_communications w2 w ->
    cannot_communicate_channels s (b1 || b2) ->
    valid_machine_state wC s (b1 || b2) ->
    context_world C wC ->
    join_wterms b2 w2 ->
    join_wterms b2' w2' ->
    join_wterms (b1 || b2)%wterm w ->
    join_wterms (b1' || b2')%wterm w' ->
    cstep_star s C (b1||b2) s' C' (b1'||b2') ->
    private_communications w2' w'.
  Proof.
    introv Hprivate Hcc Hvms HwC Hw2 Hw2' Hw Hw' Hcs.
    remember (b1||b2)%wterm as b.
    remember (b1'||b2')%wterm as b'.
    revert wC w2 w w2' w' b1 b2 b1' b2' Heqb Heqb' Hprivate Hcc Hvms HwC Hw2 Hw2' Hw Hw'.
    induction Hcs; intros; subst.
    - inverts Heqb'.
    replace w' with w in *.
    replace w2' with w2 in *; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    - renames b1' to b1'', b2' to b2'', w2' to w2'', w' to w''.
    edestruct cstep_par_resolve' as [ [b1'[b2' ?]] | [w1_[w2_[w_[?[?[??]]]]]] ]; eauto.
    subst b'.
    edestruct valid_machine_state_cstep as [wC'[Hvms' ?]]; eauto.
    assert (exists w2' w', join_wterms b2' w2' /\ join_wterms (b1'||b2')%wterm w').
      edestruct valid_machine_state_join_wterms with (s:=s') as [w12' ?]; eauto.
      destruct H1 as [w1'[w2'[?[??]]]].
      exists w2' w12'; split; auto. exists w1' w2'; splits; auto.
    destruct H1 as [w2'[w'[Hw2'' Hw'']]].
    eapply IHHcs with (w2:=w2'); eauto.
    eapply private_communications_cstep_b; eauto.
    eapply valid_machine_state_subO; eauto.
    exists wC; apply gjoin_bot2.
    eapply cannot_communicate_channels_cstep; eauto.
    (**)
    false; subst.
    eapply cstep_star_skip_resolve in Hcs; discriminate.
    - eapply IHHcs; eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma cannot_communicate_channels_par2_seq_app: forall s b1 b2 t,
    cannot_communicate_channels s (b1 || b2;; t) ->
    cannot_communicate_channels s (b1 || b2).
  Proof.
    intros.
    destruct H as [w12[?[??]]].
    destruct H as [w1[w2[?[??]]]].
    destruct H1 as [?[??]].
    exists w12; splits; auto.
    exists w1 w2; splits; auto. split; auto.
  Qed.

  Lemma cannot_communicate_channels_par1_seq_app: forall s b1 b2 t,
    cannot_communicate_channels s (b1;; t || b2) ->
    cannot_communicate_channels s (b1 || b2).
  Proof.
    intros.
    destruct H as [w12[?[??]]].
    destruct H as [w1[w2[?[??]]]].
    destruct H1 as [[??]?].
    exists w12; splits; auto.
    exists w1 w2; splits; auto. split; auto.
  Qed.


  Definition frameable_bisim_Cs (b:wterm) := forall w wF,
    join_wterms b w ->
    gjoins w wF ->
    bisimilar_Cs (extend_world b wF)  (b || wF:tm_skip).

  Lemma bisim_Cs_frameable_extend_world: forall b1 w1 w2,
    frameable_bisim_Cs b1 ->
    join_wterms b1 w1 -> gjoins w1 w2 ->
    bisimilar_Cs (extend_world b1 w2) (b1||w2:tm_skip).
  Proof.
    intros.
    spec H w1 w2 H0 H1.
    auto.
  Qed.

  Lemma frameable_bisim_Cs_frame: forall wF1 wb t,
    frameable_bisim_Cs (wb:t) ->
    frameable_bisim_Cs (extend_world (wb:t) wF1).
  Proof.
Admitted.

  Lemma frameable_bisim_Cs_par: forall w P1 P2 t1 t2,
    gprecise P1 -> gprecise P2 ->
    w|= gsep_con P1 P2 ->
    (forall w1, w1|=P1 -> frameable_bisim_Cs (w1: t1)) ->
    (forall w2, w2|=P2 -> frameable_bisim_Cs (w2: t2)) ->
    frameable_bisim_Cs (w: tm_par P1 t1 P2 t2).
  Proof.
Admitted.



  Definition seq_par_R P3 P4 t3 t4 :=
  (fun wC (s : state) ba bb =>
  (exists w1 w2 w12 b1 b2,
     ba = ((b1 || b2);; tm_par P3 t3 P4 t4)%wterm /\
     bb = (b1;; t3 || (b2;;t4))%wterm /\
     (forall w3, w3|=P3 -> frameable_bisim_Cs (w3:t3)) /\
     (forall w4, w4|=P4 -> frameable_bisim_Cs (w4:t4)) /\
     term_ensures_pred s b1 (gsep_con P3 TT) /\
     term_ensures_pred s b2 (gsep_con P4 TT) /\
     cannot_communicate_channels s (b1;;t3||b2;;t4) /\
     valid_machine_state wC s (b1 || b2) /\
     term_impl s wC b2 b1 /\
     term_impl s wC b1 b2 /\
     join_wterms b1 w1 /\
     join_wterms b2 w2 /\
     gjoin w1 w2 w12 /\
     private_communications w2 w12 /\
     noninterfering_wterm b1 b2)
    \/ bisimilar s ba bb).


  Lemma bisim_seq_par: forall P3 P4 t3 t4,
    gprecise P3 -> gprecise P4 ->
    pbisimulation (seq_par_R P3 P4 t3 t4).
  Proof.
    introv HprecP3 HprecP4.
    unfold seq_par_R.
    intros wC s ba bb H.
    destruct H as [[w1[w2[w12[b1[b2[Hba[Hbb[HfP3[HfP4[HtP3[HtP4[Hcc[Hvms[Hterm21[Hterm12[Hw1[Hw2[Hw12[Hprivate2 Hni]]]]]]]]]]]]]]]]]]] | Hb].
    subst.
    constructor.

    - introv [?]; discriminate.
    - introv [?]; discriminate.
    - intros.
    left; exists w1 w2 w12 b1 b2; splits; auto.
    eapply term_ensures_pred_s_iter; eauto.
    eapply term_ensures_pred_s_iter; eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    splits; auto.
    eapply valid_machine_state_s_iter; eauto.
    eapply term_impl_s_iter; eauto.
    eapply term_impl_s_iter; eauto.
    - intros.
    left; exists w1 w2 w12 b1 b2; splits; auto.
    eapply term_ensures_pred_stepO; eauto.
    eapply term_ensures_pred_stepO; eauto.
    eapply cannot_communicate_channels_stepO; eauto.
    splits; auto.
    eapply valid_machine_state_stepO'; eauto.
    eapply term_impl_stepO'; eauto.
    apply valid_machine_state_par_swap; auto.
    eapply term_impl_stepO'; eauto.
    - (* trivial: the parallelized program can simulate sequentialized program *)
    intros s'' b'' ?.
    inverts H. inverts H4.
    + eexists; split.
    apply step_single; apply st_par1; apply st_seq1; eauto.
    assert (Hvms': valid_machine_state wC s'' (b1'0 || b2)).
      eapply valid_machine_state_step; eauto. apply st_par1; auto.
    assert (exists w1'0 w12', join_wterms b1'0 w1'0 /\ gjoin w1'0 w2 w12').
      edestruct valid_machine_state_join_wterms with (s:=s'') as [w12' ?]; eauto.
      destruct H as [w1'0[w2_[?[??]]]].
      replace w2_ with w2 in *.
      eauto.
      eapply join_wterms_eq; eauto.
    destruct H as [w1'0[w12'[??]]].
    left; do 5 eexists; splits; eauto.
    red; intros; eapply HtP3; eauto. eapply step_cstep_star with (C':=C); eauto.
    eapply term_ensures_pred_stepO; eauto.
    eapply cannot_communicate_channels_step; eauto. apply st_par1; apply st_seq1; auto.
    splits; eauto.
    red; intros; eapply Hterm21; eauto. eapply step_cstep_star; eauto. apply st_par2; auto.
    red; intros; eapply Hterm12; eauto. eapply step_cstep_star; eauto. apply st_par1; auto.
    eapply private_communications_step2 with (b2:=b1) (w2:=w1); eauto.
    eapply cannot_communicate_channels_par1.
    eapply cannot_communicate_channels_par1_seq_app; eauto.
    apply gjoin_com; auto. apply gjoin_com; auto.
    eapply noninterfering_step1; eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply cannot_communicate_channels_par1_seq_app; eauto.
    + eexists; split.
    apply step_single; apply st_par2; apply st_seq1; eauto.
    assert (valid_machine_state wC s'' (b1 || b2')).
      eapply valid_machine_state_step; eauto. apply st_par2; auto.
    assert (exists w2' w12', join_wterms b2' w2' /\ gjoin w1 w2' w12').
      edestruct valid_machine_state_join_wterms as [w12' H0]; eauto.
      destruct H0 as [w1_[w2'[?[? ?]]]]; eauto.
      replace w1_ with w1 in *; eauto.
      eapply join_wterms_eq; eauto.
    destruct H0 as [w2'[w12'[??]]].
    left; exists w1 w2' w12'; do 2 eexists; splits; eauto.
    eapply term_ensures_pred_stepO; eauto.
    red; intros; eapply HtP4; eauto. eapply step_cstep_star; eauto.
    eapply cannot_communicate_channels_step; eauto. apply st_par2; apply st_seq1; auto.
    splits; auto.
    red; intros; eapply Hterm21; eauto. eapply step_cstep_star; eauto. apply st_par1; auto.
    red; intros; eapply Hterm12; eauto. eapply step_cstep_star; eauto. apply st_par2; auto.
    eapply private_communications_step with (b1:=b2) (w12:=w2) (w2:=gwemp) (w3:=w1); eauto.
    eapply cannot_communicate_channels_seq.
    eapply cannot_communicate_channels_par2; eauto.
    apply gjoin_bot1. apply gjoin_com; auto. apply gjoin_bot1. apply gjoin_com; auto.
    eapply noninterfering_step2; eauto.
    eapply cannot_communicate_channels_seq.
    eapply cannot_communicate_channels_par2; eauto.
    + eexists; split.
    apply step_nil.
    right.
    setoid_rewrite bisim_Cs_seq_skip1'.
    apply bisim_Cs_bisim.
(* TODO: use frameable_bisim_Cs to rearrange excess (framed) resources between w0 and w3 *)
    apply bisim_Cs_tm_par_par_precise; auto.
admit.
admit.
(*
    eapply HtP3. apply cstep_nil with (C:=cx_hole).
    eapply HtP4. apply cstep_nil with (C:=cx_hole).
*)

    (* parallel steps *)
    - intros C s' wO' C' b' ? ? ?.
    edestruct cstep_star_par1_seq_inversion
        as [[b1'[b24'[H2 H3]]] | [s'1[C'1[w3[b24'[H2 H3]]]]]]; eauto;
        [ SCase "b1 does not terminate" | SCase "b1 terminates" ].
    SCase "b1 does not terminate".
      subst. clear H.
      edestruct cstep_star_par_seq_inversion
        as [[b1'1[b2'[H H3]]] | [s'1[C'1[b1'1[w4[H Hcs'1]]]]]]; eauto;
        [ SSCase "b2 does not terminate" | SSCase "b2 terminates" ].
      SSCase "b2 does not terminate".
        inversion H3; subst b1'1 b24'; clear H3 H2.
        left; exists ((b1' || b2');; tm_par P3 t3 P4 t4)%wterm; split; eauto.
        apply cstep_star_seq_append; auto.
        assert (Hvms': valid_machine_state wO' s' (b1' || b2')).
          eapply valid_machine_state_cstep_star'; eauto.
        assert (H2: exists w1' w2' w12', join_wterms b1' w1' /\ join_wterms b2' w2' /\ gjoin w1' w2' w12').
          destruct Hvms' as [_[w12'[_[_[H2 _]]]]].
          destruct H2 as [w1'[w2'[?[? ?]]]]; eauto.
          exists w1' w2' w12'; splits; auto.
        destruct H2 as [w1'[w2'[w12'[Hw1'[Hw2' Hw12']]]]].
        left; exists w1' w2' w12' b1' b2'; splits; eauto.
        eapply term_ensures_pred_par1_cstep_star; eauto.
        eapply term_ensures_pred_par2_cstep_star; eauto.
        eapply cannot_communicate_channels_cstep_star with (C:=C); eauto.
          eapply cstep_star_par1_inner_seq_append.
          eapply cstep_star_par2_inner_seq_append; eauto.
        splits; auto.
        eapply term_impl_cstep_star' with (b1:=b2); eauto.
          apply cstep_star_par_swap; eauto.
          apply valid_machine_state_par_swap; auto.
        eapply term_impl_cstep_star' with (b1:=b1); eauto.
        eapply private_communications_cstep_star_b with (b2:=b2); eauto.
        eapply cannot_communicate_channels_par1_seq_app; eauto.
        eapply cannot_communicate_channels_par2_seq_app; eauto.
        exists w1 w2; splits; auto.
        exists w1' w2'; splits; auto.
        eapply noninterfering_cstep_star; eauto.
        eapply cannot_communicate_channels_par1.
        eapply cannot_communicate_channels_par1_seq_app; eauto.
        eapply cannot_communicate_channels_par2; eauto.
        eapply cannot_communicate_channels_par2_seq_app; eauto.
      SSCase "b2 terminates".
        clear H2.
        assert (Hcs: cstep_star s C (b1 || b2;; t4) s'1 C'1 (b1'1 || w4: t4)).
          eapply cstep_star_snoc.
          apply cstep_star_par2_inner_seq_append; eauto. apply step_cstep; apply st_par2; apply st_seq_skip.
        edestruct valid_machine_state_cstep_star with (C':=C'1) as [wO'1[H4 Hvms']]; eauto.
        edestruct valid_machine_state_gjoin_par2 with (s:=s'1) as [w4_[wO'1w3[??]]]; eauto.
        simpl in H2; subst w4_.
        edestruct valid_machine_state_join_wterms with (s:=s'1) as [w1'1w4 Hw1'1w4]; eauto.
        edestruct noninterfereing_seq_schedule
            with (s:=s'1) (C:=C'1) (b1:=b1'1) (b2:=(w4:t4)%wterm) (w2:=w4) as [s'2[H6 H7]]; eauto.
          eapply private_communications_cstep_star_b with (b1:=b1) (b2':=(w4:tm_skip)%wterm); eauto.
          eapply cannot_communicate_channels_par1_seq_app.
          eapply cannot_communicate_channels_par2_seq_app; eauto.
          reflexivity. exists w1 w2; splits; simpl; auto.
          eapply noninterfering_cstep_star with (b1:=b1) (b2:=(b2;;t4)%wterm); eauto.
          eapply cannot_communicate_channels_par1.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
          eapply cannot_communicate_channels_par2; eauto.
          reflexivity.
          eapply Hterm21; eauto. apply cstep_star_par_swap; auto.
          eapply cannot_communicate_channels_cstep_star with (b:=(b1||b2;;t4)%wterm) (C:=C); eauto.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
        clear Hcs Hcs'1 Hvms'.
        assert (Hcs: cstep_star s C (b1||b2) s'2 C' (b1'||w4:tm_skip)).
          eapply cstep_star_app; eauto.
          apply cstep_star_par1_step; auto.
        clear H H6 H4.
        assert (Hvms': valid_machine_state wO' s'2 (b1'||w4:t4)).
          eapply valid_machine_state_cstep_star' with (C:=C) (b:=(b1||b2;;t4)%wterm); eauto.
          eapply cstep_star_app.
          eapply cstep_star_par2_inner_seq_append; eauto.
          apply step_single_cstep_star. apply st_par2; apply st_seq_skip.
        assert (HwO'4: exists wO'4, gjoin wO' w4 wO'4).
          destruct Hvms' as [wS'2[wb'[ww'[HwS'1[Hwb' Hww']]]]]; eauto.
          jrewrite<- Hwb' in Hww'.
          eapply join_list_perm with (wl2:=([w4,wO']++([wS'2]++wterm_worlds b1'))%list) in Hww'.
          edestruct join_list_app_inv with (w:=ww') as [wO'4[_[HwO'4 _]]]; eauto.
          apply gjoin_join_list in HwO'4.
          exists wO'4; apply gjoin_com; auto.
          permut4.
        assert (Hvms'': valid_machine_state wO' s' (b1'||b24')).
          eapply valid_machine_state_cstep_star' with (s:=s'2) (C:=C'); eauto.
          apply cstep_star_par2_step.
          apply step_star_add_context_cstep_star; auto.
        assert (HwO'24': exists w24' wO'24', join_wterms b24' w24' /\ gjoin wO' w24' wO'24').
          destruct Hvms'' as [wS'[wb''[ww''[?[??]]]]].
          jrewrite<- H2 in H4.
          eapply join_list_perm with (wl2:=(((wterm_worlds b24') ++ [wO'])++([wS']++wterm_worlds b1'))%list) in H4.
          edestruct join_list_app_inv with (w:=ww'') as [wO'24'[_[? _]]]; eauto.
          edestruct join_list_app_inv with (w:=wO'24') as [w24'[wO'_[?[??]]]]; eauto.
          apply join_wterm_worlds in H6.
          apply join_list_single in H8; subst wO'_.
          exists w24' wO'24'; split; auto. apply gjoin_com; auto.
          permut4.
        destruct HwO'4 as [wO'4 HwO'4].
        destruct HwO'24' as [w24'[wO'24'[Hw24' HwO'24']]].
        edestruct terminates_now with (s:=s') (b:=b1') as [s''[w3 H8]].
          eapply terminates_step_starO' with (bO:=(w4:t4||wO':tm_skip)%wterm) (wO:=wO'4) (bO':=(b24'||wO':tm_skip)%wterm); eauto.
          eapply Hterm21 with (C':=C'); eauto.
          apply cstep_star_par_swap; eauto.
          apply step_star_par1_step; auto.
          exists w4 wO'; splits; auto; try reflexivity. apply gjoin_com; auto.
          exists w24' wO'; splits; auto; try reflexivity. apply gjoin_com; eauto.
          eapply valid_machine_state_par_observer with (bO:=(w4:t4)%wterm); eauto.
          reflexivity.
        edestruct noninterfering_schedule_step_star_step_star'
            with (b2:=(w4:t4)%wterm) (w2:=w4) (b1:=b1') as [s'3[??]]; eauto.
          eapply noninterfering_cstep_star with (C:=C) (b1:=b1) (b2:=(b2;;t4)%wterm); eauto.
          eapply cstep_star_app.
          apply cstep_star_par2_inner_seq_append; eauto.
          apply step_single_cstep_star. apply st_par2. apply st_seq_skip.
          eapply cannot_communicate_channels_par1; eauto.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
          eapply cannot_communicate_channels_par2; eauto.
          reflexivity.
          eapply Hterm21 with (C:=C) (C':=C'); eauto.
          apply cstep_star_par_swap; auto.
          apply step_star_step_star_sans_s_iter; eauto.
          eapply cannot_communicate_channels_cstep_star with (C:=C) (b:=(b1||b2;;t4)%wterm); eauto.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
          eapply cstep_star_app.
          apply cstep_star_par2_inner_seq_append; eauto.
          apply step_single_cstep_star. apply st_par2. apply st_seq_skip.
        assert (Hcs_s_s'3: cstep_star s C (b1||b2) s'3 C' (w3:tm_skip||w4:tm_skip)).
          eapply cstep_star_app; eauto.
          apply cstep_star_par1_step. apply step_star_add_context_cstep_star; eauto.
        clear Hvms'' Hvms' H H7 Hcs.
        assert (Hvms': valid_machine_state wO' s'3 (w3:tm_skip||w4:tm_skip)%wterm).
          eapply valid_machine_state_cstep_star'; eauto.
        assert (Hw34: gjoins w3 w4).
          edestruct valid_machine_state_join_wterms as [wb H12]; eauto.
          destruct H12 as [w3_[w4_[?[??]]]].
          simpl in *; subst w3_ w4_.
          exists wb; auto.
        destruct Hw34 as [w34 Hw34].
        assert (HP3: w3 |= gsep_con P3 TT).
          eapply HtP3 with (C:=compose_contexts C (cx_par1 cx_hole b2)).
          apply cstep_star_par_compose_contexts1; eauto.
        assert (HP4: w4 |= gsep_con P4 TT).
          eapply HtP4 with (C:=compose_contexts (cx_par2 b1 cx_hole) C).
          apply cstep_star_par_compose_contexts2'; eauto.
        right; exists s'' (w3: t3 || b24')%wterm%wterm; splits; auto.
        eapply cstep_star_seq_join.
        eapply cstep_star_app; eauto.
        apply step_single_cstep_star. apply st_par_join; auto.
        eapply Hw34.
        eapply step_cstep_star.
        apply st_par; eauto.
        eapply cstep_star_par2_step; auto.
        apply step_star_add_context_cstep_star; auto.
        apply step_star_sans_s_iter_par1_step.
        apply step_star_sans_s_iter_seq_app'; auto.
        right; apply bisim_refl.
      SCase "b1 terminates".
        clear H.
        rename H2 into Hcs_s_s'1.
        rename H3 into Hcs_s'1_s'.
        edestruct cstep_star_par_seq_inversion
            as [[b1'2[b2'[H H4]]] | [s'2[C'2[b1'2[w4[H H4]]]]]]; eauto;
            [ SSCase "b2 does not terminate" | SSCase "b2 terminates" ].
        SSCase "b2 does not terminate".
          inversion H4; subst b1'2 b24'; clear H4 Hcs_s_s'1.
          rename H into Hcs_s_s'1.
          edestruct cstep_star_par_seq_inversion
              as [[b3'3[b2'3[H2 H4]]] | [s'3[C'3[b3'3[w4[H2 H4]]]]]]; eauto;
              [ SSSCase "b2' does not terminate" | SSSCase "b2' terminates" ].
          SSSCase "b2' does not terminate".
            subst. clear Hcs_s'1_s'.
            rename H2 into Hcs_s'1_s'.
            assert (Hvms'1: exists wO'1, context_world C'1 wO'1 /\ valid_machine_state wO'1 s'1 (w3:t3 || b2')).
              edestruct valid_machine_state_cstep_star with (C':=C'1) as [wO'1[H3 Hvms'1]]; eauto.
            destruct Hvms'1 as [wO'1[HwO'1 Hvms'1]].
            assert (Hw2'Hw3w2': exists w2' w3w2', join_wterms b2' w2' /\ gjoin w3 w2' w3w2').
              edestruct valid_machine_state_join_wterms with (b:=(w3:tm_skip||b2')%wterm) as [wb' H]; eauto.
              destruct H as [w3_[w2'[?[??]]]].
              simpl in H; subst w3_.
              exists w2' wb'; split; auto.
            destruct Hw2'Hw3w2' as [w2'[w3w2'[Hw2' Hw3w2']]].
            assert (HwO'1w3: exists wO'1w3, gjoin wO'1 w3 wO'1w3).
              edestruct valid_machine_state_par_observer' with (s:= s'1) (bO:=(w3:t3)%wterm) as [w3_[wO'1w3[?[??]]]].
              eapply valid_machine_state_par_swap; eauto.
              replace w3_ with w3 in *; eauto.
            destruct HwO'1w3 as [wO'1w3 HwO'1w3].
            assert (Hvms': valid_machine_state wO' s' (b3'3 || b2'3)).
              eapply valid_machine_state_cstep_star'; eauto.
            assert (HwO'b3'3: exists w3'3 wO'b3'3, join_wterms b3'3 w3'3 /\ gjoin wO' w3'3 wO'b3'3).
              edestruct valid_machine_state_par_observer' with (bO:=b3'3) as [w3'3[wO'b3'3[?[??]]]]; eauto.
              apply valid_machine_state_par_swap; eauto.
            destruct HwO'b3'3 as [w3'3[wO'b3'3[Hw3'3 HwO'b3'3]]].
            assert (Htermb2': terminates s'1 wO'1w3 b2').
              eapply Hterm12; eauto.
            edestruct terminates_now with (s:=s') as [s''[w4 H7]]; eauto.
              eapply terminates_cstep_star' with (s:=s'1) (C:=cx_par1 C'1 (w3:t3)%wterm) (C':=cx_par1 C' b3'3); eauto.
              apply cstep_star_par_cx_par1.
              apply cstep_star_par_swap; eauto.
              exists wO'1 w3; splits; auto; reflexivity.
              exists wO' w3'3; splits; eauto; reflexivity.
              eapply valid_machine_state_par_observer with (bO:=(w3:t3)%wterm); eauto.
              reflexivity.
              apply valid_machine_state_par_swap; auto.
            assert (Hcs_s'1_s'': cstep_star s'1 C'1 (w3:t3||b2') s'' C' (b3'3||w4:tm_skip)).
              eapply cstep_star_app; eauto.
              apply cstep_star_par2_step; auto.
              apply step_star_add_context_cstep_star; auto.
              apply step_star_step_star_sans_s_iter; auto.
            edestruct noninterfereing_rev_seq_schedule
                with (C:=C'1) (b1:=(w3:t3)%wterm) (b2:=b2') as [s'2[H10 H11]]; eauto.
              reflexivity.
              eapply private_communications_cstep_star_b with (b1':=(w3:tm_skip)%wterm) (b2:=b2); eauto.
              eapply cannot_communicate_channels_par1_seq_app; eauto.
              eapply cannot_communicate_channels_par2_seq_app; eauto.
              exists w1 w2; splits; auto.
              exists w3 w2'; splits; auto; reflexivity.
              eapply noninterfering_cstep_star with (C:=C) (b1:=(b1;;t3)%wterm) (b2:=b2); eauto.
              eapply cstep_star_app.
              apply cstep_star_par1_inner_seq_append; eauto.
              apply step_single_cstep_star. apply st_par1. apply st_seq_skip.
              eapply cannot_communicate_channels_par1; eauto.
              eapply cannot_communicate_channels_par2.
              eapply cannot_communicate_channels_par2_seq_app; eauto.
              eapply cannot_communicate_channels_cstep_star with (C:=C) (b:=(b1;;t3||b2)%wterm) (b':=(w3:t3||b2')%wterm); eauto.
              eapply cannot_communicate_channels_par2_seq_app; eauto.
              eapply cstep_star_snoc.
              apply cstep_star_par1_inner_seq_append; eauto.
              apply step_cstep; apply st_par1; apply st_seq_skip.
            clear Hcs_s'1_s'' Hcs_s'1_s'.
            rename H10 into Hss_s'1_s'2.
            rename H11 into Hcs_s'2_s''.
            assert (Hcs_s_s'2: cstep_star s C (b1||b2) s'2 C'1 (w3:tm_skip||w4:tm_skip)).
              eapply cstep_star_app; eauto.
              apply cstep_star_par2_step. apply step_star_add_context_cstep_star. auto.
            clear Hcs_s_s'1 Hss_s'1_s'2 Hvms' Hvms'1.
            assert (Hvms': valid_machine_state wO'1 s'2 (w3:tm_skip || w4:tm_skip)).
              eapply valid_machine_state_cstep_star'; eauto.
            assert (Hw34: gjoins w3 w4).
              edestruct valid_machine_state_join_wterms with (s:=s'2) as [wb' ?]; eauto.
              destruct H as [w3_[w4_[?[??]]]]; simpl in *; subst w3_ w4_.
              exists wb'; auto.
            destruct Hw34 as [w34 Hw34].
            assert (HP3: w3 |= gsep_con P3 TT).
              eapply HtP3 with (C:=compose_contexts C (cx_par1 cx_hole b2)).
              apply cstep_star_par_compose_contexts1; eauto.
            assert (HP4: w4 |= gsep_con P4 TT).
              eapply HtP4 with (C:=compose_contexts (cx_par2 b1 cx_hole) C).
              apply cstep_star_par_compose_contexts2'; eauto.
            right; exists s'' (b3'3 || w4:t4)%wterm; splits; auto.
            eapply cstep_star_seq_join; eauto.
            eapply cstep_star_app; eauto. apply step_single_cstep_star.
            apply st_par_join; eauto.
            eapply step_cstep_star.
            apply st_par; eauto.
            apply cstep_star_par1_step; auto.
            apply step_star_sans_s_iter_par2_step.
            apply step_star_sans_s_iter_seq_app'; auto.
            right; apply bisim_refl.
      SSSCase "b2' terminates".
        clear Hcs_s'1_s'.
        rename H2 into Hcs_s'1_s'3.
        rename H4 into Hcs_s'3_s'.
        assert (Hvms': exists wC'1, context_world C'1 wC'1 /\ valid_machine_state wC'1 s'1 (w3:t3 || b2')).
          edestruct valid_machine_state_cstep_star with (C:=C) as [wC'1[? Hvms']]; eauto.
        destruct Hvms' as [wC'1[? Hvms']].
        assert (HwC'1w3: exists wC'1w3, gjoin wC'1 w3 wC'1w3).
          edestruct valid_machine_state_par_observer' with (bO:=(w3:t3)%wterm) as [w3_[wC'1w3[?[??]]]].
          eapply valid_machine_state_par_swap; eauto.
          simpl in H2; subst w3_; eauto.
        destruct HwC'1w3 as [wC'1w3 HwC'1w3].
        assert (Hterm_b2': terminates s'1 wC'1w3 b2').
          eapply Hterm12; eauto.
        assert (Hw2'Hw32': exists w2' w32', join_wterms b2' w2' /\ gjoin w3 w2' w32').
          edestruct valid_machine_state_join_wterms with (s:=s'1) as [wb' ?]; eauto.
          destruct H2 as [w3_[w2'[?[??]]]]; eauto.
          simpl in H2; subst w3_.
          exists w2' wb'; split; auto.
        destruct Hw2'Hw32' as [w2'[w32'[Hw2' Hw32']]].
        edestruct noninterfereing_rev_seq_schedule
            with (b1:=(w3:t3)%wterm) (b2:=b2') as [s'4[Hss_s'1_s'4 Hcs_s'4_s'3]]; eauto.
          reflexivity.
          eapply private_communications_cstep_star_b with (b1:=b1) (b2:=b2); eauto.
          eapply cannot_communicate_channels_par1_seq_app.
          eapply cannot_communicate_channels_par2_seq_app; eauto.
          exists w1 w2; splits; auto.
          exists w3 w2'; splits; auto.
          reflexivity.
          eapply noninterfering_cstep_star with (C:=C) (b1:=(b1;;t3)%wterm) (b2:=b2); eauto.
          eapply cstep_star_snoc.
          apply cstep_star_par1_inner_seq_append; eauto.
          apply step_cstep; apply st_par1; apply st_seq_skip.
          eapply cannot_communicate_channels_par1; eauto.
          eapply cannot_communicate_channels_par2.
          eapply cannot_communicate_channels_par2_seq_app; eauto.
          eapply cannot_communicate_channels_cstep_star with (C:=C) (b:=(b1;;t3||b2)%wterm); eauto.
          eapply cannot_communicate_channels_par2_seq_app; eauto.
          eapply cstep_star_snoc.
          apply cstep_star_par1_inner_seq_append; eauto.
          apply step_cstep; apply st_par1; apply st_seq_skip.
        clear Hcs_s'1_s'3.
        assert (Hcs_s_s'4: cstep_star s C (b1||b2) s'4 C'1 (w3:tm_skip||w4:tm_skip)).
          eapply cstep_star_app; eauto.
          apply cstep_star_par2_step; auto. apply step_star_add_context_cstep_star; auto.
        clear Hcs_s_s'1 Hss_s'1_s'4 Hvms'.
        assert (Hvms': valid_machine_state wC'1 s'4 (w3:tm_skip || w4:tm_skip)%wterm).
          eapply valid_machine_state_cstep_star'; eauto.
        assert (Hw34: gjoins w3 w4).
          edestruct valid_machine_state_join_wterms with (s:=s'4) as [w34 ?]; eauto.
          destruct H2 as [w3_[w4_[?[??]]]]; simpl in *; subst w3_ w4_.
          exists w34; auto.
        destruct Hw34 as [w34 Hw34].
        assert (Hcs_s'4_s': cstep_star s'4 C'1 (w3:t3||w4:t4) s' C' b').
          eapply cstep_star_app; eauto.
          apply cstep_star_par1_step; auto.
        clear Hcs_s'4_s'3 Hcs_s'3_s'.
        assert (HP3: w3 |= gsep_con P3 TT).
          eapply HtP3 with (C:=(cx_par1 C b2)).
          apply cstep_star_par_cx_par1; eauto.
        assert (HP4: w4 |= gsep_con P4 TT).
          eapply HtP4 with (C:=cx_par2 b1 C).
          apply cstep_star_par_cx_par2; eauto.
        right; exists s' b'; splits; auto.
        eapply cstep_star_app; eauto.
        eapply cstep_star_app.
        eapply cstep_star_seq_append. eauto.
        eapply step_cstep_star. apply st_seq1. apply st_par_join. eauto.
        eapply step_cstep_star. apply st_seq_skip.
        eapply step_single_cstep_star. apply st_par; auto.
        apply step_nil_.
        right; apply bisim_refl.
      SSCase "b2 terminates".
        clear Hcs_s_s'1.
        rename H into Hcs_s_s'2.
        rename H4 into Hcs_s'2_s'1.
        assert (Hvms'2: exists wC'2, context_world C'2 wC'2 /\ valid_machine_state wC'2 s'2 (b1'2 || w4:t4)).
          edestruct valid_machine_state_cstep_star with (C:=C) as [wC'2[? Hvms'2]]; eauto.
        destruct Hvms'2 as [wC'2[HwC'2 Hvms'2]].
        assert (HwC'2w4: exists wC'2w4, gjoin wC'2 w4 wC'2w4).
          edestruct valid_machine_state_par_observer' with (bO:=(w4:t4)%wterm) as [w4_[wC'2w4[?[??]]]]; eauto.
          simpl in H; subst w4_; eauto.
        destruct HwC'2w4 as [wC'2w4 HwC'2w4].
        assert (Hterm_b1'2: terminates s'2 wC'2w4 b1'2).
          eapply Hterm21; eauto; apply cstep_star_par_swap; eauto.
        assert (Hw1'2Hw1'2w4: exists w1'2 w1'2w4, join_wterms b1'2 w1'2 /\ gjoin w1'2 w4 w1'2w4).
          edestruct valid_machine_state_join_wterms with (s:=s'2) as [wb' H6]; eauto.
          destruct H6 as [w1'2[w4_[?[??]]]]; simpl in *; subst w4_; eauto.
        destruct Hw1'2Hw1'2w4 as [w1'2[w1'2w4[Hw1'2 Hw1'2w4]]].
        edestruct noninterfereing_seq_schedule with (b1:=b1'2) (b2:=(w4:t4)%wterm) (w2:=w4) (w12:=w1'2w4) as [s'3[Hcs_s'2_s'3 Hss_s'3_s'1]]; eauto.
          eapply private_communications_cstep_star_b with (b1:=b1) (b2:=b2) (b1':=b1'2) (b2':=(w4:tm_skip)%wterm) ; eauto.
          eapply cannot_communicate_channels_par1_seq_app.
          eapply cannot_communicate_channels_par2_seq_app; eauto.
          reflexivity.
          exists w1 w2; splits; auto.
          exists w1'2 w4; splits; auto.
          reflexivity.
          eapply noninterfering_cstep_star with (C:=C) (b1:=b1) (b2:=(b2;;t4)%wterm) (b2':=(w4:t4)%wterm); eauto.
          eapply cstep_star_snoc.
          apply cstep_star_par2_inner_seq_append; eauto.
          apply step_cstep; apply st_par2; apply st_seq_skip.
          eapply cannot_communicate_channels_par1.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
          eapply cannot_communicate_channels_par2; eauto.
          reflexivity.
          exists w1'2 w4; splits; auto.
          reflexivity.
          eapply cannot_communicate_channels_cstep_star with (C:=C) (b:=(b1||b2;;t4)%wterm) (b':=(b1'2||w4:t4)%wterm); eauto.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
          eapply cstep_star_snoc.
          apply cstep_star_par2_inner_seq_append; eauto.
          apply step_cstep; apply st_par2; apply st_seq_skip.
        clear Hcs_s'2_s'1. 
        assert (Hcs_s_s'3: cstep_star s C (b1||b2) s'3 C'1 (w3:tm_skip||w4:tm_skip)).
          eapply cstep_star_app; eauto. apply cstep_star_par1_step; auto.
        clear wC'2 w1'2 b1'2 s'2 C'2 Hcs_s_s'2 Hcs_s'2_s'3 Hw1'2 Hvms'2 HwC'2 HwC'2w4 Hterm_b1'2 Hw1'2w4.
        edestruct valid_machine_state_cstep_star with (s:=s) as [wC'1[HwC'1 Hvms']]; eauto.
        assert (Hcs_s'3_s': cstep_star s'3 C'1 (w3:t3||w4:t4) s' C' b').
          eapply cstep_star_app; eauto. apply cstep_star_par2_step.
          apply step_star_add_context_cstep_star. auto.
        assert (Hw34: gjoins w3 w4).
          edestruct valid_machine_state_join_wterms with (s:=s'3) as [wb' H5]; eauto.
          destruct H5 as [w3_[w4_[?[??]]]]; simpl in *; subst w3_ w4_.
          exists wb'; auto.
        destruct Hw34 as [w34 Hw34].
        assert (HP3: w3 |= gsep_con P3 TT).
          eapply HtP3 with (C:=(cx_par1 C b2)).
          apply cstep_star_par_cx_par1; eauto.
        assert (HP4: w4 |= gsep_con P4 TT).
          eapply HtP4 with (C:=cx_par2 b1 C).
          apply cstep_star_par_cx_par2; eauto.
        right; exists s' b'; splits; auto.
        eapply cstep_star_app; eauto.
        eapply cstep_star_app.
        eapply cstep_star_seq_append. eauto.
        eapply step_cstep_star. apply st_seq1. apply st_par_join. eauto.
        eapply step_cstep_star. apply st_seq_skip.
        eapply step_single_cstep_star. apply st_par; auto.
        apply step_nil_.
        right; apply bisim_refl.

    - constructor.
    + split; intros.
    eapply bisim_halted1; eauto.
    right; apply bisim_refl.
    + split; intros.
    eapply bisim_halted2; eauto.
    right; apply bisim_refl.
    + intros; right; eapply bisim_s_iter; eauto.
    + intros; right; eapply bisim_context; eauto.
    + intros.
    edestruct step_bisim1 as [bq'[??]]; eauto.
    + intros.
    edestruct cstep_star_bisim2 as [ba'[??]]; eauto.
  Qed.



(*
  Lemma dbisim_seq_par: forall wO s w12 b1 w2 b2 P3 t3 P4 t4,
    gprecise P3 -> gprecise P4 ->
    term_ensures_pred s b1 P3 ->
    term_ensures_pred s b2 P4 ->
    valid_machine_state wO s (b1 || b2) ->
    term_impl s wO b2 b1 ->
    term_impl s wO b1 b2 ->
    join_wterms b2 w2 ->
    join_wterms (b1||b2)%wterm w12 ->
    private_communications w2 w12 ->
    cannot_communicate_channels s ((b1;; t3) || (b2;; t4)) ->
    noninterfering_wterm b1 b2 ->
    (*-------------------------------------*)
    dbisimilar wO s ((b1 || b2);; tm_par P3 t3 P4 t4)%wterm s ((b1;; t3) || (b2;; t4))%wterm.
  Proof.
    introv HprecP3 HprecP4.
    introv HtermP3 HtermP4 Hvms Hterm21 Hterm12 Hw2 Hw12 Hprivate Hcc Hni.
pose   (fun wC s ba sb bb => exists w1 w2 w12 b1 b2,
     step_star_sans_s_iter sb bb s (b1;; t3 || (b2;;t4))%wterm /\
     ba = ((b1 || b2);; tm_par P3 t3 P4 t4)%wterm /\
     term_ensures_pred s b1 P3 /\
     term_ensures_pred s b2 P4 /\
     cannot_communicate_channels s (b1;;t3||b2;;t4) /\
     valid_machine_state wC s (b1 || b2) /\
     term_impl s wC b2 b1 /\
     term_impl s wC b1 b2 /\
     join_wterms b1 w1 /\
     join_wterms b2 w2 /\
     gjoin w1 w2 w12 /\
     private_communications w2 w12 /\
     noninterfering_wterm b1 b2) as R1.
pose   (fun wC s ba sb bb => exists w1 w2 w12 b1 b2,
     sb = s /\
     ba = ((b1 || b2);; tm_par P3 t3 P4 t4)%wterm /\
     bb = (b1;; t3 || (b2;;t4))%wterm /\
     term_ensures_pred s b1 P3 /\
     term_ensures_pred s b2 P4 /\
     cannot_communicate_channels s (b1;;t3||b2;;t4) /\
     valid_machine_state wC s (b1 || b2) /\
     term_impl s wC b2 b1 /\
     term_impl s wC b1 b2 /\
     join_wterms b1 w1 /\
     join_wterms b2 w2 /\
     gjoin w1 w2 w12 /\
     private_communications w2 w12 /\
     noninterfering_wterm b1 b2) as R2.

    exists (fun wC s ba sb bb =>
   (R1 wC s ba sb bb \/ R2 wC sb bb s ba)).

    splits.
    * clear -HprecP3 HprecP4.
    intros wC s ba sb bb [H0|H0]; subst R1 R2;
      [ destruct H0 as [w1[w2[w12[b1[b2[Hs[Hba[HtermP3[HtermP4[Hcc[Hvms[Hterm21[Hterm12[Hw1[Hw2[Hw12[Hprivate2 Hni]]]]]]]]]]]]]]]]]
      | auto
      ]; subst.
    + splits.
    - introv [?]; discriminate.
    - apply step_star_sans_s_iter_respects_s_i in Hs; congruence.
    - intros s' ?.
admit.
    - intros bO s' wO' bO' Hs_s_s' HwC HwO'.
    exists s'; splits; eauto.
    left; do 5 eexists; splits; eauto.
    eapply term_ensures_pred_stepO; eauto.
    eapply term_ensures_pred_stepO; eauto.
    eapply cannot_communicate_channels_stepO; eauto.
    eapply valid_machine_state_stepO'; eauto.
    splits; eauto.
    eapply term_impl_stepO'; eauto.
    eapply valid_machine_state_par_swap; eauto.
    eapply term_impl_stepO'; eauto.
    - admit.
    - intros s' ba' Hs.
    
destruct Hw12 as [w1[w2_[?[??]]]]. 
replace w2_ with w2 in *.
subst R1 R2.
split; auto; left; do 5 eexists; splits; eauto.
splits; eauto.
eapply join_wterms_eq; eauto.
- reflexivity.
- 
destruct Hw12 as [w1[w2_[?[??]]]]. 
replace w2_ with w2 in *.
subst R1 R2.
auto; right; do 5 eexists; splits; eauto.
splits; eauto.
eapply join_wterms_eq; eauto.

Qed.



  (fun wC (s : state) ba bb =>
  (exists w1 w2 w12 b1 b2,
     ba = ((b1 || b2);; tm_par P3 t3 P4 t4)%wterm /\
     bb = (b1;; t3 || (b2;;t4))%wterm /\
     term_ensures_pred s b1 P3 /\
     term_ensures_pred s b2 P4 /\
     cannot_communicate_channels s (b1;;t3||b2;;t4) /\
     valid_machine_state wC s (b1 || b2) /\
     term_impl s wC b2 b1 /\
     term_impl s wC b1 b2 /\
     join_wterms b1 w1 /\
     join_wterms b2 w2 /\
     gjoin w1 w2 w12 /\
     private_communications w2 w12 /\
     noninterfering_wterm b1 b2)
    \/ bisimilar s ba bb).


    exists (
    unfold seq_par_R.
    intros wC s ba bb H.
    destruct H as [[w1[w2[w12[b1[b2[Hba[Hbb[HtP3[HtP4[Hcc[Hvms[Hterm21[Hterm12[Hw1[Hw2[Hw12[Hprivate2 Hni]]]]]]]]]]]]]]]]] | Hb].
    subst.
    constructor.

    - introv [?]; discriminate.
    - introv [?]; discriminate.
    - intros.
    left; exists w1 w2 w12 b1 b2; splits; auto.
    eapply term_ensures_pred_s_iter; eauto.
    eapply term_ensures_pred_s_iter; eauto.
    eapply cannot_communicate_channels_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
    eapply term_impl_s_iter; eauto.
    splits; auto.
    eapply term_impl_s_iter; eauto.
    - intros.
    left; exists w1 w2 w12 b1 b2; splits; auto.
    eapply term_ensures_pred_stepO; eauto.
    eapply term_ensures_pred_stepO; eauto.
    eapply cannot_communicate_channels_stepO; eauto.
    eapply valid_machine_state_stepO'; eauto.
    eapply term_impl_stepO'; eauto.
    apply valid_machine_state_par_swap; auto.
    splits; auto.
    eapply term_impl_stepO'; eauto.
    - (* trivial: the parallelized program can simulate sequentialized program *)
    intros s'' b'' ?.
    inverts H. inverts H4.
    + eexists; split.
    apply step_single; apply st_par1; apply st_seq1; eauto.
    assert (Hvms': valid_machine_state wC s'' (b1'0 || b2)).
      eapply valid_machine_state_step; eauto. apply st_par1; auto.
    assert (exists w1'0 w12', join_wterms b1'0 w1'0 /\ gjoin w1'0 w2 w12').
      edestruct valid_machine_state_join_wterms with (s:=s'') as [w12' ?]; eauto.
      destruct H as [w1'0[w2_[?[??]]]].
      replace w2_ with w2 in *.
      eauto.
      eapply join_wterms_eq; eauto.
    destruct H as [w1'0[w12'[??]]].
    left; do 5 eexists; splits; eauto.
    red; intros; eapply HtP3; eauto. eapply step_cstep_star with (C':=C); eauto.
    eapply term_ensures_pred_stepO; eauto.
    eapply cannot_communicate_channels_step; eauto. apply st_par1; apply st_seq1; auto.
    red; intros; eapply Hterm21; eauto. eapply step_cstep_star; eauto. apply st_par2; auto.
    splits; eauto.
    red; intros; eapply Hterm12; eauto. eapply step_cstep_star; eauto. apply st_par1; auto.
    eapply private_communications_step2 with (b2:=b1) (w2:=w1); eauto.
    eapply cannot_communicate_channels_par1.
    eapply cannot_communicate_channels_par1_seq_app; eauto.
    apply gjoin_com; auto. apply gjoin_com; auto.
    eapply noninterfering_step1; eauto.
    eapply cannot_communicate_channels_par1; eauto.
    eapply cannot_communicate_channels_par1_seq_app; eauto.
    + eexists; split.
    apply step_single; apply st_par2; apply st_seq1; eauto.
    assert (valid_machine_state wC s'' (b1 || b2')).
      eapply valid_machine_state_step; eauto. apply st_par2; auto.
    assert (exists w2' w12', join_wterms b2' w2' /\ gjoin w1 w2' w12').
      edestruct valid_machine_state_join_wterms as [w12' H0]; eauto.
      destruct H0 as [w1_[w2'[?[? ?]]]]; eauto.
      replace w1_ with w1 in *; eauto.
      eapply join_wterms_eq; eauto.
    destruct H0 as [w2'[w12'[??]]].
    left; exists w1 w2' w12'; do 2 eexists; splits; eauto.
    eapply term_ensures_pred_stepO; eauto.
    red; intros; eapply HtP4; eauto. eapply step_cstep_star; eauto.
    eapply cannot_communicate_channels_step; eauto. apply st_par2; apply st_seq1; auto.
    red; intros; eapply Hterm21; eauto. eapply step_cstep_star; eauto. apply st_par1; auto.
    splits; auto.
    red; intros; eapply Hterm12; eauto. eapply step_cstep_star; eauto. apply st_par2; auto.
    eapply private_communications_step with (b1:=b2) (w12:=w2) (w2:=gwemp) (w3:=w1); eauto.
    eapply cannot_communicate_channels_seq.
    eapply cannot_communicate_channels_par2; eauto.
    apply gjoin_bot1. apply gjoin_com; auto. apply gjoin_bot1. apply gjoin_com; auto.
    eapply noninterfering_step2; eauto.
    eapply cannot_communicate_channels_seq.
    eapply cannot_communicate_channels_par2; eauto.
    + eexists; split.
    apply step_nil.
    right.
    setoid_rewrite bisim_Cs_seq_skip1'.
    apply bisim_Cs_bisim.
    apply bisim_Cs_tm_par_par_precise; auto.
    eapply HtP3. apply cstep_nil with (C:=cx_hole).
    eapply HtP4. apply cstep_nil with (C:=cx_hole).


    (* parallel steps *)
    - intros C s' wO' C' b' ? ? ?.
    edestruct cstep_star_par1_seq_inversion
        as [[b1'[b24'[H2 H3]]] | [s'1[C'1[w3[b24'[H2 H3]]]]]]; eauto;
        [ SCase "b1 does not terminate" | SCase "b1 terminates" ].
    SCase "b1 does not terminate".
      subst. clear H.
      edestruct cstep_star_par_seq_inversion
        as [[b1'1[b2'[H H3]]] | [s'1[C'1[b1'1[w4[H Hcs'1]]]]]]; eauto;
        [ SSCase "b2 does not terminate" | SSCase "b2 terminates" ].
      SSCase "b2 does not terminate".
        inversion H3; subst b1'1 b24'; clear H3 H2.
        left; exists ((b1' || b2');; tm_par P3 t3 P4 t4)%wterm; split; eauto.
        apply cstep_star_seq_append; auto.
        assert (Hvms': valid_machine_state wO' s' (b1' || b2')).
          eapply valid_machine_state_cstep_star'; eauto.
        assert (H2: exists w1' w2' w12', join_wterms b1' w1' /\ join_wterms b2' w2' /\ gjoin w1' w2' w12').
          destruct Hvms' as [_[w12'[_[_[H2 _]]]]].
          destruct H2 as [w1'[w2'[?[? ?]]]]; eauto.
          exists w1' w2' w12'; splits; auto.
        destruct H2 as [w1'[w2'[w12'[Hw1'[Hw2' Hw12']]]]].
        left; exists w1' w2' w12' b1' b2'; splits; eauto.
        eapply term_ensures_pred_par1_cstep_star; eauto.
        eapply term_ensures_pred_par2_cstep_star; eauto.
        eapply cannot_communicate_channels_cstep_star with (C:=C); eauto.
          eapply cstep_star_par1_inner_seq_append.
          eapply cstep_star_par2_inner_seq_append; eauto.
        eapply term_impl_cstep_star' with (b1:=b2); eauto.
          apply cstep_star_par_swap; eauto.
          apply valid_machine_state_par_swap; auto.
        splits; auto.
        eapply term_impl_cstep_star' with (b1:=b1); eauto.
        eapply private_communications_cstep_star_b with (b2:=b2); eauto.
        eapply cannot_communicate_channels_par1_seq_app; eauto.
        eapply cannot_communicate_channels_par2_seq_app; eauto.
        exists w1 w2; splits; auto.
        exists w1' w2'; splits; auto.
        eapply noninterfering_cstep_star; eauto.
        eapply cannot_communicate_channels_par1.
        eapply cannot_communicate_channels_par1_seq_app; eauto.
        eapply cannot_communicate_channels_par2; eauto.
        eapply cannot_communicate_channels_par2_seq_app; eauto.
      SSCase "b2 terminates".
        clear H2.
        assert (Hcs: cstep_star s C (b1 || b2;; t4) s'1 C'1 (b1'1 || w4: t4)).
          eapply cstep_star_snoc.
          apply cstep_star_par2_inner_seq_append; eauto. apply step_cstep; apply st_par2; apply st_seq_skip.
        edestruct valid_machine_state_cstep_star with (C':=C'1) as [wO'1[H4 Hvms']]; eauto.
        edestruct valid_machine_state_gjoin_par2 with (s:=s'1) as [w4_[wO'1w3[??]]]; eauto.
        simpl in H2; subst w4_.
        edestruct valid_machine_state_join_wterms with (s:=s'1) as [w1'1w4 Hw1'1w4]; eauto.
        edestruct noninterfereing_seq_schedule
            with (s:=s'1) (C:=C'1) (b1:=b1'1) (b2:=(w4:t4)%wterm) (w2:=w4) as [s'2[H6 H7]]; eauto.
          eapply private_communications_cstep_star_b with (b1:=b1) (b2':=(w4:tm_skip)%wterm); eauto.
          eapply cannot_communicate_channels_par1_seq_app.
          eapply cannot_communicate_channels_par2_seq_app; eauto.
          reflexivity. exists w1 w2; splits; simpl; auto.
          eapply noninterfering_cstep_star with (b1:=b1) (b2:=(b2;;t4)%wterm); eauto.
          eapply cannot_communicate_channels_par1.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
          eapply cannot_communicate_channels_par2; eauto.
          reflexivity.
          eapply Hterm21; eauto. apply cstep_star_par_swap; auto.
          eapply cannot_communicate_channels_cstep_star with (b:=(b1||b2;;t4)%wterm) (C:=C); eauto.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
        clear Hcs Hcs'1 Hvms'.
        assert (Hcs: cstep_star s C (b1||b2) s'2 C' (b1'||w4:tm_skip)).
          eapply cstep_star_app; eauto.
          apply cstep_star_par1_step; auto.
        clear H H6 H4.
        assert (Hvms': valid_machine_state wO' s'2 (b1'||w4:t4)).
          eapply valid_machine_state_cstep_star' with (C:=C) (b:=(b1||b2;;t4)%wterm); eauto.
          eapply cstep_star_app.
          eapply cstep_star_par2_inner_seq_append; eauto.
          apply step_single_cstep_star. apply st_par2; apply st_seq_skip.
        assert (HwO'4: exists wO'4, gjoin wO' w4 wO'4).
          destruct Hvms' as [wS'2[wb'[ww'[HwS'1[Hwb' Hww']]]]]; eauto.
          jrewrite<- Hwb' in Hww'.
          eapply join_list_perm with (wl2:=([w4,wO']++([wS'2]++wterm_worlds b1'))%list) in Hww'.
          edestruct join_list_app_inv with (w:=ww') as [wO'4[_[HwO'4 _]]]; eauto.
          apply gjoin_join_list in HwO'4.
          exists wO'4; apply gjoin_com; auto.
          permut4.
        assert (Hvms'': valid_machine_state wO' s' (b1'||b24')).
          eapply valid_machine_state_cstep_star' with (s:=s'2) (C:=C'); eauto.
          apply cstep_star_par2_step.
          apply step_star_add_context_cstep_star; auto.
        assert (HwO'24': exists w24' wO'24', join_wterms b24' w24' /\ gjoin wO' w24' wO'24').
          destruct Hvms'' as [wS'[wb''[ww''[?[??]]]]].
          jrewrite<- H2 in H4.
          eapply join_list_perm with (wl2:=(((wterm_worlds b24') ++ [wO'])++([wS']++wterm_worlds b1'))%list) in H4.
          edestruct join_list_app_inv with (w:=ww'') as [wO'24'[_[? _]]]; eauto.
          edestruct join_list_app_inv with (w:=wO'24') as [w24'[wO'_[?[??]]]]; eauto.
          apply join_wterm_worlds in H6.
          apply join_list_single in H8; subst wO'_.
          exists w24' wO'24'; split; auto. apply gjoin_com; auto.
          permut4.
        destruct HwO'4 as [wO'4 HwO'4].
        destruct HwO'24' as [w24'[wO'24'[Hw24' HwO'24']]].
        edestruct terminates_now with (s:=s') (b:=b1') as [s''[w3 H8]].
          eapply terminates_step_starO' with (bO:=(w4:t4||wO':tm_skip)%wterm) (wO:=wO'4) (bO':=(b24'||wO':tm_skip)%wterm); eauto.
          eapply Hterm21 with (C':=C'); eauto.
          apply cstep_star_par_swap; eauto.
          apply step_star_par1_step; auto.
          exists w4 wO'; splits; auto; try reflexivity. apply gjoin_com; auto.
          exists w24' wO'; splits; auto; try reflexivity. apply gjoin_com; eauto.
          eapply valid_machine_state_par_observer with (bO:=(w4:t4)%wterm); eauto.
          reflexivity.
        edestruct noninterfering_schedule_step_star_step_star'
            with (b2:=(w4:t4)%wterm) (w2:=w4) (b1:=b1') as [s'3[??]]; eauto.
          eapply noninterfering_cstep_star with (C:=C) (b1:=b1) (b2:=(b2;;t4)%wterm); eauto.
          eapply cstep_star_app.
          apply cstep_star_par2_inner_seq_append; eauto.
          apply step_single_cstep_star. apply st_par2. apply st_seq_skip.
          eapply cannot_communicate_channels_par1; eauto.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
          eapply cannot_communicate_channels_par2; eauto.
          reflexivity.
          eapply Hterm21 with (C:=C) (C':=C'); eauto.
          apply cstep_star_par_swap; auto.
          apply step_star_step_star_sans_s_iter; eauto.
          eapply cannot_communicate_channels_cstep_star with (C:=C) (b:=(b1||b2;;t4)%wterm); eauto.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
          eapply cstep_star_app.
          apply cstep_star_par2_inner_seq_append; eauto.
          apply step_single_cstep_star. apply st_par2. apply st_seq_skip.
        assert (Hcs_s_s'3: cstep_star s C (b1||b2) s'3 C' (w3:tm_skip||w4:tm_skip)).
          eapply cstep_star_app; eauto.
          apply cstep_star_par1_step. apply step_star_add_context_cstep_star; eauto.
        clear Hvms'' Hvms' H H7 Hcs.
        assert (Hvms': valid_machine_state wO' s'3 (w3:tm_skip||w4:tm_skip)%wterm).
          eapply valid_machine_state_cstep_star'; eauto.
        assert (Hw34: gjoins w3 w4).
          edestruct valid_machine_state_join_wterms as [wb H12]; eauto.
          destruct H12 as [w3_[w4_[?[??]]]].
          simpl in *; subst w3_ w4_.
          exists wb; auto.
        destruct Hw34 as [w34 Hw34].
        assert (HP3: w3 |= P3).
          eapply HtP3 with (C:=compose_contexts C (cx_par1 cx_hole b2)).
          apply cstep_star_par_compose_contexts1; eauto.
        assert (HP4: w4 |= P4).
          eapply HtP4 with (C:=compose_contexts (cx_par2 b1 cx_hole) C).
          apply cstep_star_par_compose_contexts2'; eauto.
        right; exists s'' (w3: t3 || b24')%wterm%wterm; splits; auto.
        eapply cstep_star_seq_join.
        eapply cstep_star_app; eauto.
        apply step_single_cstep_star. apply st_par_join; auto.
        eapply Hw34.
        eapply step_cstep_star.
        apply st_par; eauto.
        apply TT_gstar; auto.
        apply TT_gstar; auto.
        eapply cstep_star_par2_step; auto.
        apply step_star_add_context_cstep_star; auto.
        apply step_star_sans_s_iter_par1_step.
        apply step_star_sans_s_iter_seq_app'; auto.
        right; apply bisim_refl.
      SCase "b1 terminates".
        clear H.
        rename H2 into Hcs_s_s'1.
        rename H3 into Hcs_s'1_s'.
        edestruct cstep_star_par_seq_inversion
            as [[b1'2[b2'[H H4]]] | [s'2[C'2[b1'2[w4[H H4]]]]]]; eauto;
            [ SSCase "b2 does not terminate" | SSCase "b2 terminates" ].
        SSCase "b2 does not terminate".
          inversion H4; subst b1'2 b24'; clear H4 Hcs_s_s'1.
          rename H into Hcs_s_s'1.
          edestruct cstep_star_par_seq_inversion
              as [[b3'3[b2'3[H2 H4]]] | [s'3[C'3[b3'3[w4[H2 H4]]]]]]; eauto;
              [ SSSCase "b2' does not terminate" | SSSCase "b2' terminates" ].
          SSSCase "b2' does not terminate".
            subst. clear Hcs_s'1_s'.
            rename H2 into Hcs_s'1_s'.
            assert (Hvms'1: exists wO'1, context_world C'1 wO'1 /\ valid_machine_state wO'1 s'1 (w3:t3 || b2')).
              edestruct valid_machine_state_cstep_star with (C':=C'1) as [wO'1[H3 Hvms'1]]; eauto.
            destruct Hvms'1 as [wO'1[HwO'1 Hvms'1]].
            assert (Hw2'Hw3w2': exists w2' w3w2', join_wterms b2' w2' /\ gjoin w3 w2' w3w2').
              edestruct valid_machine_state_join_wterms with (b:=(w3:tm_skip||b2')%wterm) as [wb' H]; eauto.
              destruct H as [w3_[w2'[?[??]]]].
              simpl in H; subst w3_.
              exists w2' wb'; split; auto.
            destruct Hw2'Hw3w2' as [w2'[w3w2'[Hw2' Hw3w2']]].
            assert (HwO'1w3: exists wO'1w3, gjoin wO'1 w3 wO'1w3).
              edestruct valid_machine_state_par_observer' with (s:= s'1) (bO:=(w3:t3)%wterm) as [w3_[wO'1w3[?[??]]]].
              eapply valid_machine_state_par_swap; eauto.
              replace w3_ with w3 in *; eauto.
            destruct HwO'1w3 as [wO'1w3 HwO'1w3].
            assert (Hvms': valid_machine_state wO' s' (b3'3 || b2'3)).
              eapply valid_machine_state_cstep_star'; eauto.
            assert (HwO'b3'3: exists w3'3 wO'b3'3, join_wterms b3'3 w3'3 /\ gjoin wO' w3'3 wO'b3'3).
              edestruct valid_machine_state_par_observer' with (bO:=b3'3) as [w3'3[wO'b3'3[?[??]]]]; eauto.
              apply valid_machine_state_par_swap; eauto.
            destruct HwO'b3'3 as [w3'3[wO'b3'3[Hw3'3 HwO'b3'3]]].
            assert (Htermb2': terminates s'1 wO'1w3 b2').
              eapply Hterm12; eauto.
            edestruct terminates_now with (s:=s') as [s''[w4 H7]]; eauto.
              eapply terminates_cstep_star' with (s:=s'1) (C:=cx_par1 C'1 (w3:t3)%wterm) (C':=cx_par1 C' b3'3); eauto.
              apply cstep_star_par_cx_par1.
              apply cstep_star_par_swap; eauto.
              exists wO'1 w3; splits; auto; reflexivity.
              exists wO' w3'3; splits; eauto; reflexivity.
              eapply valid_machine_state_par_observer with (bO:=(w3:t3)%wterm); eauto.
              reflexivity.
              apply valid_machine_state_par_swap; auto.
            assert (Hcs_s'1_s'': cstep_star s'1 C'1 (w3:t3||b2') s'' C' (b3'3||w4:tm_skip)).
              eapply cstep_star_app; eauto.
              apply cstep_star_par2_step; auto.
              apply step_star_add_context_cstep_star; auto.
              apply step_star_step_star_sans_s_iter; auto.
            edestruct noninterfereing_rev_seq_schedule
                with (C:=C'1) (b1:=(w3:t3)%wterm) (b2:=b2') as [s'2[H10 H11]]; eauto.
              reflexivity.
              eapply private_communications_cstep_star_b with (b1':=(w3:tm_skip)%wterm) (b2:=b2); eauto.
              eapply cannot_communicate_channels_par1_seq_app; eauto.
              eapply cannot_communicate_channels_par2_seq_app; eauto.
              exists w1 w2; splits; auto.
              exists w3 w2'; splits; auto; reflexivity.
              eapply noninterfering_cstep_star with (C:=C) (b1:=(b1;;t3)%wterm) (b2:=b2); eauto.
              eapply cstep_star_app.
              apply cstep_star_par1_inner_seq_append; eauto.
              apply step_single_cstep_star. apply st_par1. apply st_seq_skip.
              eapply cannot_communicate_channels_par1; eauto.
              eapply cannot_communicate_channels_par2.
              eapply cannot_communicate_channels_par2_seq_app; eauto.
              eapply cannot_communicate_channels_cstep_star with (C:=C) (b:=(b1;;t3||b2)%wterm) (b':=(w3:t3||b2')%wterm); eauto.
              eapply cannot_communicate_channels_par2_seq_app; eauto.
              eapply cstep_star_snoc.
              apply cstep_star_par1_inner_seq_append; eauto.
              apply step_cstep; apply st_par1; apply st_seq_skip.
            clear Hcs_s'1_s'' Hcs_s'1_s'.
            rename H10 into Hss_s'1_s'2.
            rename H11 into Hcs_s'2_s''.
            assert (Hcs_s_s'2: cstep_star s C (b1||b2) s'2 C'1 (w3:tm_skip||w4:tm_skip)).
              eapply cstep_star_app; eauto.
              apply cstep_star_par2_step. apply step_star_add_context_cstep_star. auto.
            clear Hcs_s_s'1 Hss_s'1_s'2 Hvms' Hvms'1.
            assert (Hvms': valid_machine_state wO'1 s'2 (w3:tm_skip || w4:tm_skip)).
              eapply valid_machine_state_cstep_star'; eauto.
            assert (Hw34: gjoins w3 w4).
              edestruct valid_machine_state_join_wterms with (s:=s'2) as [wb' ?]; eauto.
              destruct H as [w3_[w4_[?[??]]]]; simpl in *; subst w3_ w4_.
              exists wb'; auto.
            destruct Hw34 as [w34 Hw34].
            assert (HP3: w3 |= P3).
              eapply HtP3 with (C:=compose_contexts C (cx_par1 cx_hole b2)).
              apply cstep_star_par_compose_contexts1; eauto.
            assert (HP4: w4 |= P4).
              eapply HtP4 with (C:=compose_contexts (cx_par2 b1 cx_hole) C).
              apply cstep_star_par_compose_contexts2'; eauto.
            right; exists s'' (b3'3 || w4:t4)%wterm; splits; auto.
            eapply cstep_star_seq_join; eauto.
            eapply cstep_star_app; eauto. apply step_single_cstep_star.
            apply st_par_join; eauto.
            eapply step_cstep_star.
            apply st_par; eauto.
            apply TT_gstar; auto.
            apply TT_gstar; auto.
            apply cstep_star_par1_step; auto.
            apply step_star_sans_s_iter_par2_step.
            apply step_star_sans_s_iter_seq_app'; auto.
            right; apply bisim_refl.
      SSSCase "b2' terminates".
        clear Hcs_s'1_s'.
        rename H2 into Hcs_s'1_s'3.
        rename H4 into Hcs_s'3_s'.
        assert (Hvms': exists wC'1, context_world C'1 wC'1 /\ valid_machine_state wC'1 s'1 (w3:t3 || b2')).
          edestruct valid_machine_state_cstep_star with (C:=C) as [wC'1[? Hvms']]; eauto.
        destruct Hvms' as [wC'1[? Hvms']].
        assert (HwC'1w3: exists wC'1w3, gjoin wC'1 w3 wC'1w3).
          edestruct valid_machine_state_par_observer' with (bO:=(w3:t3)%wterm) as [w3_[wC'1w3[?[??]]]].
          eapply valid_machine_state_par_swap; eauto.
          simpl in H2; subst w3_; eauto.
        destruct HwC'1w3 as [wC'1w3 HwC'1w3].
        assert (Hterm_b2': terminates s'1 wC'1w3 b2').
          eapply Hterm12; eauto.
        assert (Hw2'Hw32': exists w2' w32', join_wterms b2' w2' /\ gjoin w3 w2' w32').
          edestruct valid_machine_state_join_wterms with (s:=s'1) as [wb' ?]; eauto.
          destruct H2 as [w3_[w2'[?[??]]]]; eauto.
          simpl in H2; subst w3_.
          exists w2' wb'; split; auto.
        destruct Hw2'Hw32' as [w2'[w32'[Hw2' Hw32']]].
        edestruct noninterfereing_rev_seq_schedule
            with (b1:=(w3:t3)%wterm) (b2:=b2') as [s'4[Hss_s'1_s'4 Hcs_s'4_s'3]]; eauto.
          reflexivity.
          eapply private_communications_cstep_star_b with (b1:=b1) (b2:=b2); eauto.
          eapply cannot_communicate_channels_par1_seq_app.
          eapply cannot_communicate_channels_par2_seq_app; eauto.
          exists w1 w2; splits; auto.
          exists w3 w2'; splits; auto.
          reflexivity.
          eapply noninterfering_cstep_star with (C:=C) (b1:=(b1;;t3)%wterm) (b2:=b2); eauto.
          eapply cstep_star_snoc.
          apply cstep_star_par1_inner_seq_append; eauto.
          apply step_cstep; apply st_par1; apply st_seq_skip.
          eapply cannot_communicate_channels_par1; eauto.
          eapply cannot_communicate_channels_par2.
          eapply cannot_communicate_channels_par2_seq_app; eauto.
          eapply cannot_communicate_channels_cstep_star with (C:=C) (b:=(b1;;t3||b2)%wterm); eauto.
          eapply cannot_communicate_channels_par2_seq_app; eauto.
          eapply cstep_star_snoc.
          apply cstep_star_par1_inner_seq_append; eauto.
          apply step_cstep; apply st_par1; apply st_seq_skip.
        clear Hcs_s'1_s'3.
        assert (Hcs_s_s'4: cstep_star s C (b1||b2) s'4 C'1 (w3:tm_skip||w4:tm_skip)).
          eapply cstep_star_app; eauto.
          apply cstep_star_par2_step; auto. apply step_star_add_context_cstep_star; auto.
        clear Hcs_s_s'1 Hss_s'1_s'4 Hvms'.
        assert (Hvms': valid_machine_state wC'1 s'4 (w3:tm_skip || w4:tm_skip)%wterm).
          eapply valid_machine_state_cstep_star'; eauto.
        assert (Hw34: gjoins w3 w4).
          edestruct valid_machine_state_join_wterms with (s:=s'4) as [w34 ?]; eauto.
          destruct H2 as [w3_[w4_[?[??]]]]; simpl in *; subst w3_ w4_.
          exists w34; auto.
        destruct Hw34 as [w34 Hw34].
        assert (Hcs_s'4_s': cstep_star s'4 C'1 (w3:t3||w4:t4) s' C' b').
          eapply cstep_star_app; eauto.
          apply cstep_star_par1_step; auto.
        clear Hcs_s'4_s'3 Hcs_s'3_s'.
        assert (HP3: w3 |= P3).
          eapply HtP3 with (C:=(cx_par1 C b2)).
          apply cstep_star_par_cx_par1; eauto.
        assert (HP4: w4 |= P4).
          eapply HtP4 with (C:=cx_par2 b1 C).
          apply cstep_star_par_cx_par2; eauto.
        right; exists s' b'; splits; auto.
        eapply cstep_star_app; eauto.
        eapply cstep_star_app.
        eapply cstep_star_seq_append. eauto.
        eapply step_cstep_star. apply st_seq1. apply st_par_join. eauto.
        eapply step_cstep_star. apply st_seq_skip.
        eapply step_single_cstep_star. apply st_par; auto.
        apply TT_gstar; auto.
        apply TT_gstar; auto.
        apply step_nil_.
        right; apply bisim_refl.
      SSCase "b2 terminates".
        clear Hcs_s_s'1.
        rename H into Hcs_s_s'2.
        rename H4 into Hcs_s'2_s'1.
        assert (Hvms'2: exists wC'2, context_world C'2 wC'2 /\ valid_machine_state wC'2 s'2 (b1'2 || w4:t4)).
          edestruct valid_machine_state_cstep_star with (C:=C) as [wC'2[? Hvms'2]]; eauto.
        destruct Hvms'2 as [wC'2[HwC'2 Hvms'2]].
        assert (HwC'2w4: exists wC'2w4, gjoin wC'2 w4 wC'2w4).
          edestruct valid_machine_state_par_observer' with (bO:=(w4:t4)%wterm) as [w4_[wC'2w4[?[??]]]]; eauto.
          simpl in H; subst w4_; eauto.
        destruct HwC'2w4 as [wC'2w4 HwC'2w4].
        assert (Hterm_b1'2: terminates s'2 wC'2w4 b1'2).
          eapply Hterm21; eauto; apply cstep_star_par_swap; eauto.
        assert (Hw1'2Hw1'2w4: exists w1'2 w1'2w4, join_wterms b1'2 w1'2 /\ gjoin w1'2 w4 w1'2w4).
          edestruct valid_machine_state_join_wterms with (s:=s'2) as [wb' H6]; eauto.
          destruct H6 as [w1'2[w4_[?[??]]]]; simpl in *; subst w4_; eauto.
        destruct Hw1'2Hw1'2w4 as [w1'2[w1'2w4[Hw1'2 Hw1'2w4]]].
        edestruct noninterfereing_seq_schedule with (b1:=b1'2) (b2:=(w4:t4)%wterm) (w2:=w4) (w12:=w1'2w4) as [s'3[Hcs_s'2_s'3 Hss_s'3_s'1]]; eauto.
          eapply private_communications_cstep_star_b with (b1:=b1) (b2:=b2) (b1':=b1'2) (b2':=(w4:tm_skip)%wterm) ; eauto.
          eapply cannot_communicate_channels_par1_seq_app.
          eapply cannot_communicate_channels_par2_seq_app; eauto.
          reflexivity.
          exists w1 w2; splits; auto.
          exists w1'2 w4; splits; auto.
          reflexivity.
          eapply noninterfering_cstep_star with (C:=C) (b1:=b1) (b2:=(b2;;t4)%wterm) (b2':=(w4:t4)%wterm); eauto.
          eapply cstep_star_snoc.
          apply cstep_star_par2_inner_seq_append; eauto.
          apply step_cstep; apply st_par2; apply st_seq_skip.
          eapply cannot_communicate_channels_par1.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
          eapply cannot_communicate_channels_par2; eauto.
          reflexivity.
          exists w1'2 w4; splits; auto.
          reflexivity.
          eapply cannot_communicate_channels_cstep_star with (C:=C) (b:=(b1||b2;;t4)%wterm) (b':=(b1'2||w4:t4)%wterm); eauto.
          eapply cannot_communicate_channels_par1_seq_app; eauto.
          eapply cstep_star_snoc.
          apply cstep_star_par2_inner_seq_append; eauto.
          apply step_cstep; apply st_par2; apply st_seq_skip.
        clear Hcs_s'2_s'1. 
        assert (Hcs_s_s'3: cstep_star s C (b1||b2) s'3 C'1 (w3:tm_skip||w4:tm_skip)).
          eapply cstep_star_app; eauto. apply cstep_star_par1_step; auto.
        clear wC'2 w1'2 b1'2 s'2 C'2 Hcs_s_s'2 Hcs_s'2_s'3 Hw1'2 Hvms'2 HwC'2 HwC'2w4 Hterm_b1'2 Hw1'2w4.
        edestruct valid_machine_state_cstep_star with (s:=s) as [wC'1[HwC'1 Hvms']]; eauto.
        assert (Hcs_s'3_s': cstep_star s'3 C'1 (w3:t3||w4:t4) s' C' b').
          eapply cstep_star_app; eauto. apply cstep_star_par2_step.
          apply step_star_add_context_cstep_star. auto.
        assert (Hw34: gjoins w3 w4).
          edestruct valid_machine_state_join_wterms with (s:=s'3) as [wb' H5]; eauto.
          destruct H5 as [w3_[w4_[?[??]]]]; simpl in *; subst w3_ w4_.
          exists wb'; auto.
        destruct Hw34 as [w34 Hw34].
        assert (HP3: w3 |= P3).
          eapply HtP3 with (C:=(cx_par1 C b2)).
          apply cstep_star_par_cx_par1; eauto.
        assert (HP4: w4 |= P4).
          eapply HtP4 with (C:=cx_par2 b1 C).
          apply cstep_star_par_cx_par2; eauto.
        right; exists s' b'; splits; auto.
        eapply cstep_star_app; eauto.
        eapply cstep_star_app.
        eapply cstep_star_seq_append. eauto.
        eapply step_cstep_star. apply st_seq1. apply st_par_join. eauto.
        eapply step_cstep_star. apply st_seq_skip.
        eapply step_single_cstep_star. apply st_par; auto.
        apply TT_gstar; auto.
        apply TT_gstar; auto.
        apply step_nil_.
        right; apply bisim_refl.

    - constructor.
    + split; intros.
    eapply bisim_halted1; eauto.
    right; apply bisim_refl.
    + split; intros.
    eapply bisim_halted2; eauto.
    right; apply bisim_refl.
    + intros; right; eapply bisim_s_iter; eauto.
    + intros; right; eapply bisim_context; eauto.
    + intros.
    edestruct step_bisim1 as [bq'[??]]; eauto.
    + intros.
    edestruct cstep_star_bisim2 as [ba'[??]]; eauto.
  Qed.
*)

  Corollary cbisim_seq_par: forall wO s w12 b1 w2 b2 P3 t3 P4 t4,
    gprecise P3 -> gprecise P4 ->
    (forall w3, w3|=P3 -> frameable_bisim_Cs (w3:t3)) ->
    (forall w4, w4|=P4 -> frameable_bisim_Cs (w4:t4)) ->
    term_ensures_pred s b1 (gsep_con P3 TT) ->
    term_ensures_pred s b2 (gsep_con P4 TT) ->
    valid_machine_state wO s (b1 || b2) ->
    term_impl s wO b2 b1 ->
    term_impl s wO b1 b2 ->
    join_wterms b2 w2 ->
    join_wterms (b1||b2)%wterm w12 ->
    private_communications w2 w12 ->
    cannot_communicate_channels s ((b1;; t3) || (b2;; t4)) ->
    noninterfering_wterm b1 b2 ->
    (*-------------------------------------*)
    cbisimilar wO s ((b1 || b2);; tm_par P3 t3 P4 t4)%wterm s ((b1;; t3) || (b2;; t4))%wterm.
  Proof.
    introv HprecP3 HprecP4 HtermP3 HtermP4 Hvms HframeP3 HframeP4.
    introv Hterm21 Hterm12 Hw2 Hw12 Hprivate2 Hcc Hni.
    apply pbisim_cbisim.
    exists (seq_par_R P3 P4 t3 t4).
    split.
    apply bisim_seq_par; auto.
    destruct Hw12 as [w1[w2_[?[??]]]].
    replace w2_ with w2 in *.
    left; jauto.
    eapply join_wterms_eq; eauto.
    split; eapply valid_machine_state_wterm'; eauto.
  Qed.


End Lemmas.


